#AUTOGENERATED!!!

from enum import IntEnum

class ObjectIds(object):
    Null = 0
    Boolean = 1
    SByte = 2
    Byte = 3
    Int16 = 4
    UInt16 = 5
    Int32 = 6
    UInt32 = 7
    Int64 = 8
    UInt64 = 9
    Float = 10
    Double = 11
    String = 12
    DateTime = 13
    Guid = 14
    ByteString = 15
    XmlElement = 16
    NodeId = 17
    ExpandedNodeId = 18
    StatusCode = 19
    QualifiedName = 20
    LocalizedText = 21
    Structure = 22
    DataValue = 23
    BaseDataType = 24
    DiagnosticInfo = 25
    Number = 26
    Integer = 27
    UInteger = 28
    Enumeration = 29
    Image = 30
    References = 31
    NonHierarchicalReferences = 32
    HierarchicalReferences = 33
    HasChild = 34
    Organizes = 35
    HasEventSource = 36
    HasModellingRule = 37
    HasEncoding = 38
    HasDescription = 39
    HasTypeDefinition = 40
    GeneratesEvent = 41
    Aggregates = 44
    HasSubtype = 45
    HasProperty = 46
    HasComponent = 47
    HasNotifier = 48
    HasOrderedComponent = 49
    FromState = 51
    ToState = 52
    HasCause = 53
    HasEffect = 54
    HasHistoricalConfiguration = 56
    BaseObjectType = 58
    FolderType = 61
    BaseVariableType = 62
    BaseDataVariableType = 63
    PropertyType = 68
    DataTypeDescriptionType = 69
    DataTypeDictionaryType = 72
    DataTypeSystemType = 75
    DataTypeEncodingType = 76
    ModellingRuleType = 77
    ModellingRule_Mandatory = 78
    ModellingRule_MandatoryShared = 79
    ModellingRule_Optional = 80
    ModellingRule_ExposesItsArray = 83
    RootFolder = 84
    ObjectsFolder = 85
    TypesFolder = 86
    ViewsFolder = 87
    ObjectTypesFolder = 88
    VariableTypesFolder = 89
    DataTypesFolder = 90
    ReferenceTypesFolder = 91
    XmlSchema_TypeSystem = 92
    OPCBinarySchema_TypeSystem = 93
    DataTypeDescriptionType_DataTypeVersion = 104
    DataTypeDescriptionType_DictionaryFragment = 105
    DataTypeDictionaryType_DataTypeVersion = 106
    DataTypeDictionaryType_NamespaceUri = 107
    ModellingRuleType_NamingRule = 111
    ModellingRule_Mandatory_NamingRule = 112
    ModellingRule_Optional_NamingRule = 113
    ModellingRule_ExposesItsArray_NamingRule = 114
    ModellingRule_MandatoryShared_NamingRule = 116
    HasSubStateMachine = 117
    NamingRuleType = 120
    Decimal128 = 121
    IdType = 256
    NodeClass = 257
    Node = 258
    Node_Encoding_DefaultXml = 259
    Node_Encoding_DefaultBinary = 260
    ObjectNode = 261
    ObjectNode_Encoding_DefaultXml = 262
    ObjectNode_Encoding_DefaultBinary = 263
    ObjectTypeNode = 264
    ObjectTypeNode_Encoding_DefaultXml = 265
    ObjectTypeNode_Encoding_DefaultBinary = 266
    VariableNode = 267
    VariableNode_Encoding_DefaultXml = 268
    VariableNode_Encoding_DefaultBinary = 269
    VariableTypeNode = 270
    VariableTypeNode_Encoding_DefaultXml = 271
    VariableTypeNode_Encoding_DefaultBinary = 272
    ReferenceTypeNode = 273
    ReferenceTypeNode_Encoding_DefaultXml = 274
    ReferenceTypeNode_Encoding_DefaultBinary = 275
    MethodNode = 276
    MethodNode_Encoding_DefaultXml = 277
    MethodNode_Encoding_DefaultBinary = 278
    ViewNode = 279
    ViewNode_Encoding_DefaultXml = 280
    ViewNode_Encoding_DefaultBinary = 281
    DataTypeNode = 282
    DataTypeNode_Encoding_DefaultXml = 283
    DataTypeNode_Encoding_DefaultBinary = 284
    ReferenceNode = 285
    ReferenceNode_Encoding_DefaultXml = 286
    ReferenceNode_Encoding_DefaultBinary = 287
    IntegerId = 288
    Counter = 289
    Duration = 290
    NumericRange = 291
    Time = 292
    Date = 293
    UtcTime = 294
    LocaleId = 295
    Argument = 296
    Argument_Encoding_DefaultXml = 297
    Argument_Encoding_DefaultBinary = 298
    StatusResult = 299
    StatusResult_Encoding_DefaultXml = 300
    StatusResult_Encoding_DefaultBinary = 301
    MessageSecurityMode = 302
    UserTokenType = 303
    UserTokenPolicy = 304
    UserTokenPolicy_Encoding_DefaultXml = 305
    UserTokenPolicy_Encoding_DefaultBinary = 306
    ApplicationType = 307
    ApplicationDescription = 308
    ApplicationDescription_Encoding_DefaultXml = 309
    ApplicationDescription_Encoding_DefaultBinary = 310
    ApplicationInstanceCertificate = 311
    EndpointDescription = 312
    EndpointDescription_Encoding_DefaultXml = 313
    EndpointDescription_Encoding_DefaultBinary = 314
    SecurityTokenRequestType = 315
    UserIdentityToken = 316
    UserIdentityToken_Encoding_DefaultXml = 317
    UserIdentityToken_Encoding_DefaultBinary = 318
    AnonymousIdentityToken = 319
    AnonymousIdentityToken_Encoding_DefaultXml = 320
    AnonymousIdentityToken_Encoding_DefaultBinary = 321
    UserNameIdentityToken = 322
    UserNameIdentityToken_Encoding_DefaultXml = 323
    UserNameIdentityToken_Encoding_DefaultBinary = 324
    X509IdentityToken = 325
    X509IdentityToken_Encoding_DefaultXml = 326
    X509IdentityToken_Encoding_DefaultBinary = 327
    EndpointConfiguration = 331
    EndpointConfiguration_Encoding_DefaultXml = 332
    EndpointConfiguration_Encoding_DefaultBinary = 333
    ComplianceLevel = 334
    SupportedProfile = 335
    SupportedProfile_Encoding_DefaultXml = 336
    SupportedProfile_Encoding_DefaultBinary = 337
    BuildInfo = 338
    BuildInfo_Encoding_DefaultXml = 339
    BuildInfo_Encoding_DefaultBinary = 340
    SoftwareCertificate = 341
    SoftwareCertificate_Encoding_DefaultXml = 342
    SoftwareCertificate_Encoding_DefaultBinary = 343
    SignedSoftwareCertificate = 344
    SignedSoftwareCertificate_Encoding_DefaultXml = 345
    SignedSoftwareCertificate_Encoding_DefaultBinary = 346
    AttributeWriteMask = 347
    NodeAttributesMask = 348
    NodeAttributes = 349
    NodeAttributes_Encoding_DefaultXml = 350
    NodeAttributes_Encoding_DefaultBinary = 351
    ObjectAttributes = 352
    ObjectAttributes_Encoding_DefaultXml = 353
    ObjectAttributes_Encoding_DefaultBinary = 354
    VariableAttributes = 355
    VariableAttributes_Encoding_DefaultXml = 356
    VariableAttributes_Encoding_DefaultBinary = 357
    MethodAttributes = 358
    MethodAttributes_Encoding_DefaultXml = 359
    MethodAttributes_Encoding_DefaultBinary = 360
    ObjectTypeAttributes = 361
    ObjectTypeAttributes_Encoding_DefaultXml = 362
    ObjectTypeAttributes_Encoding_DefaultBinary = 363
    VariableTypeAttributes = 364
    VariableTypeAttributes_Encoding_DefaultXml = 365
    VariableTypeAttributes_Encoding_DefaultBinary = 366
    ReferenceTypeAttributes = 367
    ReferenceTypeAttributes_Encoding_DefaultXml = 368
    ReferenceTypeAttributes_Encoding_DefaultBinary = 369
    DataTypeAttributes = 370
    DataTypeAttributes_Encoding_DefaultXml = 371
    DataTypeAttributes_Encoding_DefaultBinary = 372
    ViewAttributes = 373
    ViewAttributes_Encoding_DefaultXml = 374
    ViewAttributes_Encoding_DefaultBinary = 375
    AddNodesItem = 376
    AddNodesItem_Encoding_DefaultXml = 377
    AddNodesItem_Encoding_DefaultBinary = 378
    AddReferencesItem = 379
    AddReferencesItem_Encoding_DefaultXml = 380
    AddReferencesItem_Encoding_DefaultBinary = 381
    DeleteNodesItem = 382
    DeleteNodesItem_Encoding_DefaultXml = 383
    DeleteNodesItem_Encoding_DefaultBinary = 384
    DeleteReferencesItem = 385
    DeleteReferencesItem_Encoding_DefaultXml = 386
    DeleteReferencesItem_Encoding_DefaultBinary = 387
    SessionAuthenticationToken = 388
    RequestHeader = 389
    RequestHeader_Encoding_DefaultXml = 390
    RequestHeader_Encoding_DefaultBinary = 391
    ResponseHeader = 392
    ResponseHeader_Encoding_DefaultXml = 393
    ResponseHeader_Encoding_DefaultBinary = 394
    ServiceFault = 395
    ServiceFault_Encoding_DefaultXml = 396
    ServiceFault_Encoding_DefaultBinary = 397
    EnumeratedTestType = 398
    FindServersRequest = 420
    FindServersRequest_Encoding_DefaultXml = 421
    FindServersRequest_Encoding_DefaultBinary = 422
    FindServersResponse = 423
    FindServersResponse_Encoding_DefaultXml = 424
    FindServersResponse_Encoding_DefaultBinary = 425
    GetEndpointsRequest = 426
    GetEndpointsRequest_Encoding_DefaultXml = 427
    GetEndpointsRequest_Encoding_DefaultBinary = 428
    GetEndpointsResponse = 429
    GetEndpointsResponse_Encoding_DefaultXml = 430
    GetEndpointsResponse_Encoding_DefaultBinary = 431
    RegisteredServer = 432
    RegisteredServer_Encoding_DefaultXml = 433
    RegisteredServer_Encoding_DefaultBinary = 434
    RegisterServerRequest = 435
    RegisterServerRequest_Encoding_DefaultXml = 436
    RegisterServerRequest_Encoding_DefaultBinary = 437
    RegisterServerResponse = 438
    RegisterServerResponse_Encoding_DefaultXml = 439
    RegisterServerResponse_Encoding_DefaultBinary = 440
    ChannelSecurityToken = 441
    ChannelSecurityToken_Encoding_DefaultXml = 442
    ChannelSecurityToken_Encoding_DefaultBinary = 443
    OpenSecureChannelRequest = 444
    OpenSecureChannelRequest_Encoding_DefaultXml = 445
    OpenSecureChannelRequest_Encoding_DefaultBinary = 446
    OpenSecureChannelResponse = 447
    OpenSecureChannelResponse_Encoding_DefaultXml = 448
    OpenSecureChannelResponse_Encoding_DefaultBinary = 449
    CloseSecureChannelRequest = 450
    CloseSecureChannelRequest_Encoding_DefaultXml = 451
    CloseSecureChannelRequest_Encoding_DefaultBinary = 452
    CloseSecureChannelResponse = 453
    CloseSecureChannelResponse_Encoding_DefaultXml = 454
    CloseSecureChannelResponse_Encoding_DefaultBinary = 455
    SignatureData = 456
    SignatureData_Encoding_DefaultXml = 457
    SignatureData_Encoding_DefaultBinary = 458
    CreateSessionRequest = 459
    CreateSessionRequest_Encoding_DefaultXml = 460
    CreateSessionRequest_Encoding_DefaultBinary = 461
    CreateSessionResponse = 462
    CreateSessionResponse_Encoding_DefaultXml = 463
    CreateSessionResponse_Encoding_DefaultBinary = 464
    ActivateSessionRequest = 465
    ActivateSessionRequest_Encoding_DefaultXml = 466
    ActivateSessionRequest_Encoding_DefaultBinary = 467
    ActivateSessionResponse = 468
    ActivateSessionResponse_Encoding_DefaultXml = 469
    ActivateSessionResponse_Encoding_DefaultBinary = 470
    CloseSessionRequest = 471
    CloseSessionRequest_Encoding_DefaultXml = 472
    CloseSessionRequest_Encoding_DefaultBinary = 473
    CloseSessionResponse = 474
    CloseSessionResponse_Encoding_DefaultXml = 475
    CloseSessionResponse_Encoding_DefaultBinary = 476
    CancelRequest = 477
    CancelRequest_Encoding_DefaultXml = 478
    CancelRequest_Encoding_DefaultBinary = 479
    CancelResponse = 480
    CancelResponse_Encoding_DefaultXml = 481
    CancelResponse_Encoding_DefaultBinary = 482
    AddNodesResult = 483
    AddNodesResult_Encoding_DefaultXml = 484
    AddNodesResult_Encoding_DefaultBinary = 485
    AddNodesRequest = 486
    AddNodesRequest_Encoding_DefaultXml = 487
    AddNodesRequest_Encoding_DefaultBinary = 488
    AddNodesResponse = 489
    AddNodesResponse_Encoding_DefaultXml = 490
    AddNodesResponse_Encoding_DefaultBinary = 491
    AddReferencesRequest = 492
    AddReferencesRequest_Encoding_DefaultXml = 493
    AddReferencesRequest_Encoding_DefaultBinary = 494
    AddReferencesResponse = 495
    AddReferencesResponse_Encoding_DefaultXml = 496
    AddReferencesResponse_Encoding_DefaultBinary = 497
    DeleteNodesRequest = 498
    DeleteNodesRequest_Encoding_DefaultXml = 499
    DeleteNodesRequest_Encoding_DefaultBinary = 500
    DeleteNodesResponse = 501
    DeleteNodesResponse_Encoding_DefaultXml = 502
    DeleteNodesResponse_Encoding_DefaultBinary = 503
    DeleteReferencesRequest = 504
    DeleteReferencesRequest_Encoding_DefaultXml = 505
    DeleteReferencesRequest_Encoding_DefaultBinary = 506
    DeleteReferencesResponse = 507
    DeleteReferencesResponse_Encoding_DefaultXml = 508
    DeleteReferencesResponse_Encoding_DefaultBinary = 509
    BrowseDirection = 510
    ViewDescription = 511
    ViewDescription_Encoding_DefaultXml = 512
    ViewDescription_Encoding_DefaultBinary = 513
    BrowseDescription = 514
    BrowseDescription_Encoding_DefaultXml = 515
    BrowseDescription_Encoding_DefaultBinary = 516
    BrowseResultMask = 517
    ReferenceDescription = 518
    ReferenceDescription_Encoding_DefaultXml = 519
    ReferenceDescription_Encoding_DefaultBinary = 520
    ContinuationPoint = 521
    BrowseResult = 522
    BrowseResult_Encoding_DefaultXml = 523
    BrowseResult_Encoding_DefaultBinary = 524
    BrowseRequest = 525
    BrowseRequest_Encoding_DefaultXml = 526
    BrowseRequest_Encoding_DefaultBinary = 527
    BrowseResponse = 528
    BrowseResponse_Encoding_DefaultXml = 529
    BrowseResponse_Encoding_DefaultBinary = 530
    BrowseNextRequest = 531
    BrowseNextRequest_Encoding_DefaultXml = 532
    BrowseNextRequest_Encoding_DefaultBinary = 533
    BrowseNextResponse = 534
    BrowseNextResponse_Encoding_DefaultXml = 535
    BrowseNextResponse_Encoding_DefaultBinary = 536
    RelativePathElement = 537
    RelativePathElement_Encoding_DefaultXml = 538
    RelativePathElement_Encoding_DefaultBinary = 539
    RelativePath = 540
    RelativePath_Encoding_DefaultXml = 541
    RelativePath_Encoding_DefaultBinary = 542
    BrowsePath = 543
    BrowsePath_Encoding_DefaultXml = 544
    BrowsePath_Encoding_DefaultBinary = 545
    BrowsePathTarget = 546
    BrowsePathTarget_Encoding_DefaultXml = 547
    BrowsePathTarget_Encoding_DefaultBinary = 548
    BrowsePathResult = 549
    BrowsePathResult_Encoding_DefaultXml = 550
    BrowsePathResult_Encoding_DefaultBinary = 551
    TranslateBrowsePathsToNodeIdsRequest = 552
    TranslateBrowsePathsToNodeIdsRequest_Encoding_DefaultXml = 553
    TranslateBrowsePathsToNodeIdsRequest_Encoding_DefaultBinary = 554
    TranslateBrowsePathsToNodeIdsResponse = 555
    TranslateBrowsePathsToNodeIdsResponse_Encoding_DefaultXml = 556
    TranslateBrowsePathsToNodeIdsResponse_Encoding_DefaultBinary = 557
    RegisterNodesRequest = 558
    RegisterNodesRequest_Encoding_DefaultXml = 559
    RegisterNodesRequest_Encoding_DefaultBinary = 560
    RegisterNodesResponse = 561
    RegisterNodesResponse_Encoding_DefaultXml = 562
    RegisterNodesResponse_Encoding_DefaultBinary = 563
    UnregisterNodesRequest = 564
    UnregisterNodesRequest_Encoding_DefaultXml = 565
    UnregisterNodesRequest_Encoding_DefaultBinary = 566
    UnregisterNodesResponse = 567
    UnregisterNodesResponse_Encoding_DefaultXml = 568
    UnregisterNodesResponse_Encoding_DefaultBinary = 569
    QueryDataDescription = 570
    QueryDataDescription_Encoding_DefaultXml = 571
    QueryDataDescription_Encoding_DefaultBinary = 572
    NodeTypeDescription = 573
    NodeTypeDescription_Encoding_DefaultXml = 574
    NodeTypeDescription_Encoding_DefaultBinary = 575
    FilterOperator = 576
    QueryDataSet = 577
    QueryDataSet_Encoding_DefaultXml = 578
    QueryDataSet_Encoding_DefaultBinary = 579
    NodeReference = 580
    NodeReference_Encoding_DefaultXml = 581
    NodeReference_Encoding_DefaultBinary = 582
    ContentFilterElement = 583
    ContentFilterElement_Encoding_DefaultXml = 584
    ContentFilterElement_Encoding_DefaultBinary = 585
    ContentFilter = 586
    ContentFilter_Encoding_DefaultXml = 587
    ContentFilter_Encoding_DefaultBinary = 588
    FilterOperand = 589
    FilterOperand_Encoding_DefaultXml = 590
    FilterOperand_Encoding_DefaultBinary = 591
    ElementOperand = 592
    ElementOperand_Encoding_DefaultXml = 593
    ElementOperand_Encoding_DefaultBinary = 594
    LiteralOperand = 595
    LiteralOperand_Encoding_DefaultXml = 596
    LiteralOperand_Encoding_DefaultBinary = 597
    AttributeOperand = 598
    AttributeOperand_Encoding_DefaultXml = 599
    AttributeOperand_Encoding_DefaultBinary = 600
    SimpleAttributeOperand = 601
    SimpleAttributeOperand_Encoding_DefaultXml = 602
    SimpleAttributeOperand_Encoding_DefaultBinary = 603
    ContentFilterElementResult = 604
    ContentFilterElementResult_Encoding_DefaultXml = 605
    ContentFilterElementResult_Encoding_DefaultBinary = 606
    ContentFilterResult = 607
    ContentFilterResult_Encoding_DefaultXml = 608
    ContentFilterResult_Encoding_DefaultBinary = 609
    ParsingResult = 610
    ParsingResult_Encoding_DefaultXml = 611
    ParsingResult_Encoding_DefaultBinary = 612
    QueryFirstRequest = 613
    QueryFirstRequest_Encoding_DefaultXml = 614
    QueryFirstRequest_Encoding_DefaultBinary = 615
    QueryFirstResponse = 616
    QueryFirstResponse_Encoding_DefaultXml = 617
    QueryFirstResponse_Encoding_DefaultBinary = 618
    QueryNextRequest = 619
    QueryNextRequest_Encoding_DefaultXml = 620
    QueryNextRequest_Encoding_DefaultBinary = 621
    QueryNextResponse = 622
    QueryNextResponse_Encoding_DefaultXml = 623
    QueryNextResponse_Encoding_DefaultBinary = 624
    TimestampsToReturn = 625
    ReadValueId = 626
    ReadValueId_Encoding_DefaultXml = 627
    ReadValueId_Encoding_DefaultBinary = 628
    ReadRequest = 629
    ReadRequest_Encoding_DefaultXml = 630
    ReadRequest_Encoding_DefaultBinary = 631
    ReadResponse = 632
    ReadResponse_Encoding_DefaultXml = 633
    ReadResponse_Encoding_DefaultBinary = 634
    HistoryReadValueId = 635
    HistoryReadValueId_Encoding_DefaultXml = 636
    HistoryReadValueId_Encoding_DefaultBinary = 637
    HistoryReadResult = 638
    HistoryReadResult_Encoding_DefaultXml = 639
    HistoryReadResult_Encoding_DefaultBinary = 640
    HistoryReadDetails = 641
    HistoryReadDetails_Encoding_DefaultXml = 642
    HistoryReadDetails_Encoding_DefaultBinary = 643
    ReadEventDetails = 644
    ReadEventDetails_Encoding_DefaultXml = 645
    ReadEventDetails_Encoding_DefaultBinary = 646
    ReadRawModifiedDetails = 647
    ReadRawModifiedDetails_Encoding_DefaultXml = 648
    ReadRawModifiedDetails_Encoding_DefaultBinary = 649
    ReadProcessedDetails = 650
    ReadProcessedDetails_Encoding_DefaultXml = 651
    ReadProcessedDetails_Encoding_DefaultBinary = 652
    ReadAtTimeDetails = 653
    ReadAtTimeDetails_Encoding_DefaultXml = 654
    ReadAtTimeDetails_Encoding_DefaultBinary = 655
    HistoryData = 656
    HistoryData_Encoding_DefaultXml = 657
    HistoryData_Encoding_DefaultBinary = 658
    HistoryEvent = 659
    HistoryEvent_Encoding_DefaultXml = 660
    HistoryEvent_Encoding_DefaultBinary = 661
    HistoryReadRequest = 662
    HistoryReadRequest_Encoding_DefaultXml = 663
    HistoryReadRequest_Encoding_DefaultBinary = 664
    HistoryReadResponse = 665
    HistoryReadResponse_Encoding_DefaultXml = 666
    HistoryReadResponse_Encoding_DefaultBinary = 667
    WriteValue = 668
    WriteValue_Encoding_DefaultXml = 669
    WriteValue_Encoding_DefaultBinary = 670
    WriteRequest = 671
    WriteRequest_Encoding_DefaultXml = 672
    WriteRequest_Encoding_DefaultBinary = 673
    WriteResponse = 674
    WriteResponse_Encoding_DefaultXml = 675
    WriteResponse_Encoding_DefaultBinary = 676
    HistoryUpdateDetails = 677
    HistoryUpdateDetails_Encoding_DefaultXml = 678
    HistoryUpdateDetails_Encoding_DefaultBinary = 679
    UpdateDataDetails = 680
    UpdateDataDetails_Encoding_DefaultXml = 681
    UpdateDataDetails_Encoding_DefaultBinary = 682
    UpdateEventDetails = 683
    UpdateEventDetails_Encoding_DefaultXml = 684
    UpdateEventDetails_Encoding_DefaultBinary = 685
    DeleteRawModifiedDetails = 686
    DeleteRawModifiedDetails_Encoding_DefaultXml = 687
    DeleteRawModifiedDetails_Encoding_DefaultBinary = 688
    DeleteAtTimeDetails = 689
    DeleteAtTimeDetails_Encoding_DefaultXml = 690
    DeleteAtTimeDetails_Encoding_DefaultBinary = 691
    DeleteEventDetails = 692
    DeleteEventDetails_Encoding_DefaultXml = 693
    DeleteEventDetails_Encoding_DefaultBinary = 694
    HistoryUpdateResult = 695
    HistoryUpdateResult_Encoding_DefaultXml = 696
    HistoryUpdateResult_Encoding_DefaultBinary = 697
    HistoryUpdateRequest = 698
    HistoryUpdateRequest_Encoding_DefaultXml = 699
    HistoryUpdateRequest_Encoding_DefaultBinary = 700
    HistoryUpdateResponse = 701
    HistoryUpdateResponse_Encoding_DefaultXml = 702
    HistoryUpdateResponse_Encoding_DefaultBinary = 703
    CallMethodRequest = 704
    CallMethodRequest_Encoding_DefaultXml = 705
    CallMethodRequest_Encoding_DefaultBinary = 706
    CallMethodResult = 707
    CallMethodResult_Encoding_DefaultXml = 708
    CallMethodResult_Encoding_DefaultBinary = 709
    CallRequest = 710
    CallRequest_Encoding_DefaultXml = 711
    CallRequest_Encoding_DefaultBinary = 712
    CallResponse = 713
    CallResponse_Encoding_DefaultXml = 714
    CallResponse_Encoding_DefaultBinary = 715
    MonitoringMode = 716
    DataChangeTrigger = 717
    DeadbandType = 718
    MonitoringFilter = 719
    MonitoringFilter_Encoding_DefaultXml = 720
    MonitoringFilter_Encoding_DefaultBinary = 721
    DataChangeFilter = 722
    DataChangeFilter_Encoding_DefaultXml = 723
    DataChangeFilter_Encoding_DefaultBinary = 724
    EventFilter = 725
    EventFilter_Encoding_DefaultXml = 726
    EventFilter_Encoding_DefaultBinary = 727
    AggregateFilter = 728
    AggregateFilter_Encoding_DefaultXml = 729
    AggregateFilter_Encoding_DefaultBinary = 730
    MonitoringFilterResult = 731
    MonitoringFilterResult_Encoding_DefaultXml = 732
    MonitoringFilterResult_Encoding_DefaultBinary = 733
    EventFilterResult = 734
    EventFilterResult_Encoding_DefaultXml = 735
    EventFilterResult_Encoding_DefaultBinary = 736
    AggregateFilterResult = 737
    AggregateFilterResult_Encoding_DefaultXml = 738
    AggregateFilterResult_Encoding_DefaultBinary = 739
    MonitoringParameters = 740
    MonitoringParameters_Encoding_DefaultXml = 741
    MonitoringParameters_Encoding_DefaultBinary = 742
    MonitoredItemCreateRequest = 743
    MonitoredItemCreateRequest_Encoding_DefaultXml = 744
    MonitoredItemCreateRequest_Encoding_DefaultBinary = 745
    MonitoredItemCreateResult = 746
    MonitoredItemCreateResult_Encoding_DefaultXml = 747
    MonitoredItemCreateResult_Encoding_DefaultBinary = 748
    CreateMonitoredItemsRequest = 749
    CreateMonitoredItemsRequest_Encoding_DefaultXml = 750
    CreateMonitoredItemsRequest_Encoding_DefaultBinary = 751
    CreateMonitoredItemsResponse = 752
    CreateMonitoredItemsResponse_Encoding_DefaultXml = 753
    CreateMonitoredItemsResponse_Encoding_DefaultBinary = 754
    MonitoredItemModifyRequest = 755
    MonitoredItemModifyRequest_Encoding_DefaultXml = 756
    MonitoredItemModifyRequest_Encoding_DefaultBinary = 757
    MonitoredItemModifyResult = 758
    MonitoredItemModifyResult_Encoding_DefaultXml = 759
    MonitoredItemModifyResult_Encoding_DefaultBinary = 760
    ModifyMonitoredItemsRequest = 761
    ModifyMonitoredItemsRequest_Encoding_DefaultXml = 762
    ModifyMonitoredItemsRequest_Encoding_DefaultBinary = 763
    ModifyMonitoredItemsResponse = 764
    ModifyMonitoredItemsResponse_Encoding_DefaultXml = 765
    ModifyMonitoredItemsResponse_Encoding_DefaultBinary = 766
    SetMonitoringModeRequest = 767
    SetMonitoringModeRequest_Encoding_DefaultXml = 768
    SetMonitoringModeRequest_Encoding_DefaultBinary = 769
    SetMonitoringModeResponse = 770
    SetMonitoringModeResponse_Encoding_DefaultXml = 771
    SetMonitoringModeResponse_Encoding_DefaultBinary = 772
    SetTriggeringRequest = 773
    SetTriggeringRequest_Encoding_DefaultXml = 774
    SetTriggeringRequest_Encoding_DefaultBinary = 775
    SetTriggeringResponse = 776
    SetTriggeringResponse_Encoding_DefaultXml = 777
    SetTriggeringResponse_Encoding_DefaultBinary = 778
    DeleteMonitoredItemsRequest = 779
    DeleteMonitoredItemsRequest_Encoding_DefaultXml = 780
    DeleteMonitoredItemsRequest_Encoding_DefaultBinary = 781
    DeleteMonitoredItemsResponse = 782
    DeleteMonitoredItemsResponse_Encoding_DefaultXml = 783
    DeleteMonitoredItemsResponse_Encoding_DefaultBinary = 784
    CreateSubscriptionRequest = 785
    CreateSubscriptionRequest_Encoding_DefaultXml = 786
    CreateSubscriptionRequest_Encoding_DefaultBinary = 787
    CreateSubscriptionResponse = 788
    CreateSubscriptionResponse_Encoding_DefaultXml = 789
    CreateSubscriptionResponse_Encoding_DefaultBinary = 790
    ModifySubscriptionRequest = 791
    ModifySubscriptionRequest_Encoding_DefaultXml = 792
    ModifySubscriptionRequest_Encoding_DefaultBinary = 793
    ModifySubscriptionResponse = 794
    ModifySubscriptionResponse_Encoding_DefaultXml = 795
    ModifySubscriptionResponse_Encoding_DefaultBinary = 796
    SetPublishingModeRequest = 797
    SetPublishingModeRequest_Encoding_DefaultXml = 798
    SetPublishingModeRequest_Encoding_DefaultBinary = 799
    SetPublishingModeResponse = 800
    SetPublishingModeResponse_Encoding_DefaultXml = 801
    SetPublishingModeResponse_Encoding_DefaultBinary = 802
    NotificationMessage = 803
    NotificationMessage_Encoding_DefaultXml = 804
    NotificationMessage_Encoding_DefaultBinary = 805
    MonitoredItemNotification = 806
    MonitoredItemNotification_Encoding_DefaultXml = 807
    MonitoredItemNotification_Encoding_DefaultBinary = 808
    DataChangeNotification = 809
    DataChangeNotification_Encoding_DefaultXml = 810
    DataChangeNotification_Encoding_DefaultBinary = 811
    StatusChangeNotification = 818
    StatusChangeNotification_Encoding_DefaultXml = 819
    StatusChangeNotification_Encoding_DefaultBinary = 820
    SubscriptionAcknowledgement = 821
    SubscriptionAcknowledgement_Encoding_DefaultXml = 822
    SubscriptionAcknowledgement_Encoding_DefaultBinary = 823
    PublishRequest = 824
    PublishRequest_Encoding_DefaultXml = 825
    PublishRequest_Encoding_DefaultBinary = 826
    PublishResponse = 827
    PublishResponse_Encoding_DefaultXml = 828
    PublishResponse_Encoding_DefaultBinary = 829
    RepublishRequest = 830
    RepublishRequest_Encoding_DefaultXml = 831
    RepublishRequest_Encoding_DefaultBinary = 832
    RepublishResponse = 833
    RepublishResponse_Encoding_DefaultXml = 834
    RepublishResponse_Encoding_DefaultBinary = 835
    TransferResult = 836
    TransferResult_Encoding_DefaultXml = 837
    TransferResult_Encoding_DefaultBinary = 838
    TransferSubscriptionsRequest = 839
    TransferSubscriptionsRequest_Encoding_DefaultXml = 840
    TransferSubscriptionsRequest_Encoding_DefaultBinary = 841
    TransferSubscriptionsResponse = 842
    TransferSubscriptionsResponse_Encoding_DefaultXml = 843
    TransferSubscriptionsResponse_Encoding_DefaultBinary = 844
    DeleteSubscriptionsRequest = 845
    DeleteSubscriptionsRequest_Encoding_DefaultXml = 846
    DeleteSubscriptionsRequest_Encoding_DefaultBinary = 847
    DeleteSubscriptionsResponse = 848
    DeleteSubscriptionsResponse_Encoding_DefaultXml = 849
    DeleteSubscriptionsResponse_Encoding_DefaultBinary = 850
    RedundancySupport = 851
    ServerState = 852
    RedundantServerDataType = 853
    RedundantServerDataType_Encoding_DefaultXml = 854
    RedundantServerDataType_Encoding_DefaultBinary = 855
    SamplingIntervalDiagnosticsDataType = 856
    SamplingIntervalDiagnosticsDataType_Encoding_DefaultXml = 857
    SamplingIntervalDiagnosticsDataType_Encoding_DefaultBinary = 858
    ServerDiagnosticsSummaryDataType = 859
    ServerDiagnosticsSummaryDataType_Encoding_DefaultXml = 860
    ServerDiagnosticsSummaryDataType_Encoding_DefaultBinary = 861
    ServerStatusDataType = 862
    ServerStatusDataType_Encoding_DefaultXml = 863
    ServerStatusDataType_Encoding_DefaultBinary = 864
    SessionDiagnosticsDataType = 865
    SessionDiagnosticsDataType_Encoding_DefaultXml = 866
    SessionDiagnosticsDataType_Encoding_DefaultBinary = 867
    SessionSecurityDiagnosticsDataType = 868
    SessionSecurityDiagnosticsDataType_Encoding_DefaultXml = 869
    SessionSecurityDiagnosticsDataType_Encoding_DefaultBinary = 870
    ServiceCounterDataType = 871
    ServiceCounterDataType_Encoding_DefaultXml = 872
    ServiceCounterDataType_Encoding_DefaultBinary = 873
    SubscriptionDiagnosticsDataType = 874
    SubscriptionDiagnosticsDataType_Encoding_DefaultXml = 875
    SubscriptionDiagnosticsDataType_Encoding_DefaultBinary = 876
    ModelChangeStructureDataType = 877
    ModelChangeStructureDataType_Encoding_DefaultXml = 878
    ModelChangeStructureDataType_Encoding_DefaultBinary = 879
    Range = 884
    Range_Encoding_DefaultXml = 885
    Range_Encoding_DefaultBinary = 886
    EUInformation = 887
    EUInformation_Encoding_DefaultXml = 888
    EUInformation_Encoding_DefaultBinary = 889
    ExceptionDeviationFormat = 890
    Annotation = 891
    Annotation_Encoding_DefaultXml = 892
    Annotation_Encoding_DefaultBinary = 893
    ProgramDiagnosticDataType = 894
    ProgramDiagnosticDataType_Encoding_DefaultXml = 895
    ProgramDiagnosticDataType_Encoding_DefaultBinary = 896
    SemanticChangeStructureDataType = 897
    SemanticChangeStructureDataType_Encoding_DefaultXml = 898
    SemanticChangeStructureDataType_Encoding_DefaultBinary = 899
    EventNotificationList = 914
    EventNotificationList_Encoding_DefaultXml = 915
    EventNotificationList_Encoding_DefaultBinary = 916
    EventFieldList = 917
    EventFieldList_Encoding_DefaultXml = 918
    EventFieldList_Encoding_DefaultBinary = 919
    HistoryEventFieldList = 920
    HistoryEventFieldList_Encoding_DefaultXml = 921
    HistoryEventFieldList_Encoding_DefaultBinary = 922
    IssuedIdentityToken = 938
    IssuedIdentityToken_Encoding_DefaultXml = 939
    IssuedIdentityToken_Encoding_DefaultBinary = 940
    NotificationData = 945
    NotificationData_Encoding_DefaultXml = 946
    NotificationData_Encoding_DefaultBinary = 947
    AggregateConfiguration = 948
    AggregateConfiguration_Encoding_DefaultXml = 949
    AggregateConfiguration_Encoding_DefaultBinary = 950
    ImageBMP = 2000
    ImageGIF = 2001
    ImageJPG = 2002
    ImagePNG = 2003
    ServerType = 2004
    ServerType_ServerArray = 2005
    ServerType_NamespaceArray = 2006
    ServerType_ServerStatus = 2007
    ServerType_ServiceLevel = 2008
    ServerType_ServerCapabilities = 2009
    ServerType_ServerDiagnostics = 2010
    ServerType_VendorServerInfo = 2011
    ServerType_ServerRedundancy = 2012
    ServerCapabilitiesType = 2013
    ServerCapabilitiesType_ServerProfileArray = 2014
    ServerCapabilitiesType_LocaleIdArray = 2016
    ServerCapabilitiesType_MinSupportedSampleRate = 2017
    ServerCapabilitiesType_ModellingRules = 2019
    ServerDiagnosticsType = 2020
    ServerDiagnosticsType_ServerDiagnosticsSummary = 2021
    ServerDiagnosticsType_SamplingIntervalDiagnosticsArray = 2022
    ServerDiagnosticsType_SubscriptionDiagnosticsArray = 2023
    ServerDiagnosticsType_EnabledFlag = 2025
    SessionsDiagnosticsSummaryType = 2026
    SessionsDiagnosticsSummaryType_SessionDiagnosticsArray = 2027
    SessionsDiagnosticsSummaryType_SessionSecurityDiagnosticsArray = 2028
    SessionDiagnosticsObjectType = 2029
    SessionDiagnosticsObjectType_SessionDiagnostics = 2030
    SessionDiagnosticsObjectType_SessionSecurityDiagnostics = 2031
    SessionDiagnosticsObjectType_SubscriptionDiagnosticsArray = 2032
    VendorServerInfoType = 2033
    ServerRedundancyType = 2034
    ServerRedundancyType_RedundancySupport = 2035
    TransparentRedundancyType = 2036
    TransparentRedundancyType_CurrentServerId = 2037
    TransparentRedundancyType_RedundantServerArray = 2038
    NonTransparentRedundancyType = 2039
    NonTransparentRedundancyType_ServerUriArray = 2040
    BaseEventType = 2041
    BaseEventType_EventId = 2042
    BaseEventType_EventType = 2043
    BaseEventType_SourceNode = 2044
    BaseEventType_SourceName = 2045
    BaseEventType_Time = 2046
    BaseEventType_ReceiveTime = 2047
    BaseEventType_Message = 2050
    BaseEventType_Severity = 2051
    AuditEventType = 2052
    AuditEventType_ActionTimeStamp = 2053
    AuditEventType_Status = 2054
    AuditEventType_ServerId = 2055
    AuditEventType_ClientAuditEntryId = 2056
    AuditEventType_ClientUserId = 2057
    AuditSecurityEventType = 2058
    AuditChannelEventType = 2059
    AuditOpenSecureChannelEventType = 2060
    AuditOpenSecureChannelEventType_ClientCertificate = 2061
    AuditOpenSecureChannelEventType_RequestType = 2062
    AuditOpenSecureChannelEventType_SecurityPolicyUri = 2063
    AuditOpenSecureChannelEventType_SecurityMode = 2065
    AuditOpenSecureChannelEventType_RequestedLifetime = 2066
    AuditSessionEventType = 2069
    AuditSessionEventType_SessionId = 2070
    AuditCreateSessionEventType = 2071
    AuditCreateSessionEventType_SecureChannelId = 2072
    AuditCreateSessionEventType_ClientCertificate = 2073
    AuditCreateSessionEventType_RevisedSessionTimeout = 2074
    AuditActivateSessionEventType = 2075
    AuditActivateSessionEventType_ClientSoftwareCertificates = 2076
    AuditActivateSessionEventType_UserIdentityToken = 2077
    AuditCancelEventType = 2078
    AuditCancelEventType_RequestHandle = 2079
    AuditCertificateEventType = 2080
    AuditCertificateEventType_Certificate = 2081
    AuditCertificateDataMismatchEventType = 2082
    AuditCertificateDataMismatchEventType_InvalidHostname = 2083
    AuditCertificateDataMismatchEventType_InvalidUri = 2084
    AuditCertificateExpiredEventType = 2085
    AuditCertificateInvalidEventType = 2086
    AuditCertificateUntrustedEventType = 2087
    AuditCertificateRevokedEventType = 2088
    AuditCertificateMismatchEventType = 2089
    AuditNodeManagementEventType = 2090
    AuditAddNodesEventType = 2091
    AuditAddNodesEventType_NodesToAdd = 2092
    AuditDeleteNodesEventType = 2093
    AuditDeleteNodesEventType_NodesToDelete = 2094
    AuditAddReferencesEventType = 2095
    AuditAddReferencesEventType_ReferencesToAdd = 2096
    AuditDeleteReferencesEventType = 2097
    AuditDeleteReferencesEventType_ReferencesToDelete = 2098
    AuditUpdateEventType = 2099
    AuditWriteUpdateEventType = 2100
    AuditWriteUpdateEventType_IndexRange = 2101
    AuditWriteUpdateEventType_OldValue = 2102
    AuditWriteUpdateEventType_NewValue = 2103
    AuditHistoryUpdateEventType = 2104
    AuditUpdateMethodEventType = 2127
    AuditUpdateMethodEventType_MethodId = 2128
    AuditUpdateMethodEventType_InputArguments = 2129
    SystemEventType = 2130
    DeviceFailureEventType = 2131
    BaseModelChangeEventType = 2132
    GeneralModelChangeEventType = 2133
    GeneralModelChangeEventType_Changes = 2134
    ServerVendorCapabilityType = 2137
    ServerStatusType = 2138
    ServerStatusType_StartTime = 2139
    ServerStatusType_CurrentTime = 2140
    ServerStatusType_State = 2141
    ServerStatusType_BuildInfo = 2142
    ServerDiagnosticsSummaryType = 2150
    ServerDiagnosticsSummaryType_ServerViewCount = 2151
    ServerDiagnosticsSummaryType_CurrentSessionCount = 2152
    ServerDiagnosticsSummaryType_CumulatedSessionCount = 2153
    ServerDiagnosticsSummaryType_SecurityRejectedSessionCount = 2154
    ServerDiagnosticsSummaryType_RejectedSessionCount = 2155
    ServerDiagnosticsSummaryType_SessionTimeoutCount = 2156
    ServerDiagnosticsSummaryType_SessionAbortCount = 2157
    ServerDiagnosticsSummaryType_PublishingIntervalCount = 2159
    ServerDiagnosticsSummaryType_CurrentSubscriptionCount = 2160
    ServerDiagnosticsSummaryType_CumulatedSubscriptionCount = 2161
    ServerDiagnosticsSummaryType_SecurityRejectedRequestsCount = 2162
    ServerDiagnosticsSummaryType_RejectedRequestsCount = 2163
    SamplingIntervalDiagnosticsArrayType = 2164
    SamplingIntervalDiagnosticsType = 2165
    SamplingIntervalDiagnosticsType_SamplingInterval = 2166
    SubscriptionDiagnosticsArrayType = 2171
    SubscriptionDiagnosticsType = 2172
    SubscriptionDiagnosticsType_SessionId = 2173
    SubscriptionDiagnosticsType_SubscriptionId = 2174
    SubscriptionDiagnosticsType_Priority = 2175
    SubscriptionDiagnosticsType_PublishingInterval = 2176
    SubscriptionDiagnosticsType_MaxKeepAliveCount = 2177
    SubscriptionDiagnosticsType_MaxNotificationsPerPublish = 2179
    SubscriptionDiagnosticsType_PublishingEnabled = 2180
    SubscriptionDiagnosticsType_ModifyCount = 2181
    SubscriptionDiagnosticsType_EnableCount = 2182
    SubscriptionDiagnosticsType_DisableCount = 2183
    SubscriptionDiagnosticsType_RepublishRequestCount = 2184
    SubscriptionDiagnosticsType_RepublishMessageRequestCount = 2185
    SubscriptionDiagnosticsType_RepublishMessageCount = 2186
    SubscriptionDiagnosticsType_TransferRequestCount = 2187
    SubscriptionDiagnosticsType_TransferredToAltClientCount = 2188
    SubscriptionDiagnosticsType_TransferredToSameClientCount = 2189
    SubscriptionDiagnosticsType_PublishRequestCount = 2190
    SubscriptionDiagnosticsType_DataChangeNotificationsCount = 2191
    SubscriptionDiagnosticsType_NotificationsCount = 2193
    SessionDiagnosticsArrayType = 2196
    SessionDiagnosticsVariableType = 2197
    SessionDiagnosticsVariableType_SessionId = 2198
    SessionDiagnosticsVariableType_SessionName = 2199
    SessionDiagnosticsVariableType_ClientDescription = 2200
    SessionDiagnosticsVariableType_ServerUri = 2201
    SessionDiagnosticsVariableType_EndpointUrl = 2202
    SessionDiagnosticsVariableType_LocaleIds = 2203
    SessionDiagnosticsVariableType_ActualSessionTimeout = 2204
    SessionDiagnosticsVariableType_ClientConnectionTime = 2205
    SessionDiagnosticsVariableType_ClientLastContactTime = 2206
    SessionDiagnosticsVariableType_CurrentSubscriptionsCount = 2207
    SessionDiagnosticsVariableType_CurrentMonitoredItemsCount = 2208
    SessionDiagnosticsVariableType_CurrentPublishRequestsInQueue = 2209
    SessionDiagnosticsVariableType_ReadCount = 2217
    SessionDiagnosticsVariableType_HistoryReadCount = 2218
    SessionDiagnosticsVariableType_WriteCount = 2219
    SessionDiagnosticsVariableType_HistoryUpdateCount = 2220
    SessionDiagnosticsVariableType_CallCount = 2221
    SessionDiagnosticsVariableType_CreateMonitoredItemsCount = 2222
    SessionDiagnosticsVariableType_ModifyMonitoredItemsCount = 2223
    SessionDiagnosticsVariableType_SetMonitoringModeCount = 2224
    SessionDiagnosticsVariableType_SetTriggeringCount = 2225
    SessionDiagnosticsVariableType_DeleteMonitoredItemsCount = 2226
    SessionDiagnosticsVariableType_CreateSubscriptionCount = 2227
    SessionDiagnosticsVariableType_ModifySubscriptionCount = 2228
    SessionDiagnosticsVariableType_SetPublishingModeCount = 2229
    SessionDiagnosticsVariableType_PublishCount = 2230
    SessionDiagnosticsVariableType_RepublishCount = 2231
    SessionDiagnosticsVariableType_TransferSubscriptionsCount = 2232
    SessionDiagnosticsVariableType_DeleteSubscriptionsCount = 2233
    SessionDiagnosticsVariableType_AddNodesCount = 2234
    SessionDiagnosticsVariableType_AddReferencesCount = 2235
    SessionDiagnosticsVariableType_DeleteNodesCount = 2236
    SessionDiagnosticsVariableType_DeleteReferencesCount = 2237
    SessionDiagnosticsVariableType_BrowseCount = 2238
    SessionDiagnosticsVariableType_BrowseNextCount = 2239
    SessionDiagnosticsVariableType_TranslateBrowsePathsToNodeIdsCount = 2240
    SessionDiagnosticsVariableType_QueryFirstCount = 2241
    SessionDiagnosticsVariableType_QueryNextCount = 2242
    SessionSecurityDiagnosticsArrayType = 2243
    SessionSecurityDiagnosticsType = 2244
    SessionSecurityDiagnosticsType_SessionId = 2245
    SessionSecurityDiagnosticsType_ClientUserIdOfSession = 2246
    SessionSecurityDiagnosticsType_ClientUserIdHistory = 2247
    SessionSecurityDiagnosticsType_AuthenticationMechanism = 2248
    SessionSecurityDiagnosticsType_Encoding = 2249
    SessionSecurityDiagnosticsType_TransportProtocol = 2250
    SessionSecurityDiagnosticsType_SecurityMode = 2251
    SessionSecurityDiagnosticsType_SecurityPolicyUri = 2252
    Server = 2253
    Server_ServerArray = 2254
    Server_NamespaceArray = 2255
    Server_ServerStatus = 2256
    Server_ServerStatus_StartTime = 2257
    Server_ServerStatus_CurrentTime = 2258
    Server_ServerStatus_State = 2259
    Server_ServerStatus_BuildInfo = 2260
    Server_ServerStatus_BuildInfo_ProductName = 2261
    Server_ServerStatus_BuildInfo_ProductUri = 2262
    Server_ServerStatus_BuildInfo_ManufacturerName = 2263
    Server_ServerStatus_BuildInfo_SoftwareVersion = 2264
    Server_ServerStatus_BuildInfo_BuildNumber = 2265
    Server_ServerStatus_BuildInfo_BuildDate = 2266
    Server_ServiceLevel = 2267
    Server_ServerCapabilities = 2268
    Server_ServerCapabilities_ServerProfileArray = 2269
    Server_ServerCapabilities_LocaleIdArray = 2271
    Server_ServerCapabilities_MinSupportedSampleRate = 2272
    Server_ServerDiagnostics = 2274
    Server_ServerDiagnostics_ServerDiagnosticsSummary = 2275
    Server_ServerDiagnostics_ServerDiagnosticsSummary_ServerViewCount = 2276
    Server_ServerDiagnostics_ServerDiagnosticsSummary_CurrentSessionCount = 2277
    Server_ServerDiagnostics_ServerDiagnosticsSummary_CumulatedSessionCount = 2278
    Server_ServerDiagnostics_ServerDiagnosticsSummary_SecurityRejectedSessionCount = 2279
    Server_ServerDiagnostics_ServerDiagnosticsSummary_SessionTimeoutCount = 2281
    Server_ServerDiagnostics_ServerDiagnosticsSummary_SessionAbortCount = 2282
    Server_ServerDiagnostics_ServerDiagnosticsSummary_PublishingIntervalCount = 2284
    Server_ServerDiagnostics_ServerDiagnosticsSummary_CurrentSubscriptionCount = 2285
    Server_ServerDiagnostics_ServerDiagnosticsSummary_CumulatedSubscriptionCount = 2286
    Server_ServerDiagnostics_ServerDiagnosticsSummary_SecurityRejectedRequestsCount = 2287
    Server_ServerDiagnostics_ServerDiagnosticsSummary_RejectedRequestsCount = 2288
    Server_ServerDiagnostics_SamplingIntervalDiagnosticsArray = 2289
    Server_ServerDiagnostics_SubscriptionDiagnosticsArray = 2290
    Server_ServerDiagnostics_EnabledFlag = 2294
    Server_VendorServerInfo = 2295
    Server_ServerRedundancy = 2296
    StateMachineType = 2299
    StateType = 2307
    StateType_StateNumber = 2308
    InitialStateType = 2309
    TransitionType = 2310
    TransitionEventType = 2311
    TransitionType_TransitionNumber = 2312
    AuditUpdateStateEventType = 2315
    HistoricalDataConfigurationType = 2318
    HistoricalDataConfigurationType_Stepped = 2323
    HistoricalDataConfigurationType_Definition = 2324
    HistoricalDataConfigurationType_MaxTimeInterval = 2325
    HistoricalDataConfigurationType_MinTimeInterval = 2326
    HistoricalDataConfigurationType_ExceptionDeviation = 2327
    HistoricalDataConfigurationType_ExceptionDeviationFormat = 2328
    HistoryServerCapabilitiesType = 2330
    HistoryServerCapabilitiesType_AccessHistoryDataCapability = 2331
    HistoryServerCapabilitiesType_AccessHistoryEventsCapability = 2332
    HistoryServerCapabilitiesType_InsertDataCapability = 2334
    HistoryServerCapabilitiesType_ReplaceDataCapability = 2335
    HistoryServerCapabilitiesType_UpdateDataCapability = 2336
    HistoryServerCapabilitiesType_DeleteRawCapability = 2337
    HistoryServerCapabilitiesType_DeleteAtTimeCapability = 2338
    AggregateFunctionType = 2340
    AggregateFunction_Interpolative = 2341
    AggregateFunction_Average = 2342
    AggregateFunction_TimeAverage = 2343
    AggregateFunction_Total = 2344
    AggregateFunction_Minimum = 2346
    AggregateFunction_Maximum = 2347
    AggregateFunction_MinimumActualTime = 2348
    AggregateFunction_MaximumActualTime = 2349
    AggregateFunction_Range = 2350
    AggregateFunction_AnnotationCount = 2351
    AggregateFunction_Count = 2352
    AggregateFunction_NumberOfTransitions = 2355
    AggregateFunction_Start = 2357
    AggregateFunction_End = 2358
    AggregateFunction_Delta = 2359
    AggregateFunction_DurationGood = 2360
    AggregateFunction_DurationBad = 2361
    AggregateFunction_PercentGood = 2362
    AggregateFunction_PercentBad = 2363
    AggregateFunction_WorstQuality = 2364
    DataItemType = 2365
    DataItemType_Definition = 2366
    DataItemType_ValuePrecision = 2367
    AnalogItemType = 2368
    AnalogItemType_EURange = 2369
    AnalogItemType_InstrumentRange = 2370
    AnalogItemType_EngineeringUnits = 2371
    DiscreteItemType = 2372
    TwoStateDiscreteType = 2373
    TwoStateDiscreteType_FalseState = 2374
    TwoStateDiscreteType_TrueState = 2375
    MultiStateDiscreteType = 2376
    MultiStateDiscreteType_EnumStrings = 2377
    ProgramTransitionEventType = 2378
    ProgramTransitionEventType_IntermediateResult = 2379
    ProgramDiagnosticType = 2380
    ProgramDiagnosticType_CreateSessionId = 2381
    ProgramDiagnosticType_CreateClientName = 2382
    ProgramDiagnosticType_InvocationCreationTime = 2383
    ProgramDiagnosticType_LastTransitionTime = 2384
    ProgramDiagnosticType_LastMethodCall = 2385
    ProgramDiagnosticType_LastMethodSessionId = 2386
    ProgramDiagnosticType_LastMethodInputArguments = 2387
    ProgramDiagnosticType_LastMethodOutputArguments = 2388
    ProgramDiagnosticType_LastMethodCallTime = 2389
    ProgramDiagnosticType_LastMethodReturnStatus = 2390
    ProgramStateMachineType = 2391
    ProgramStateMachineType_Creatable = 2392
    ProgramStateMachineType_Deletable = 2393
    ProgramStateMachineType_AutoDelete = 2394
    ProgramStateMachineType_RecycleCount = 2395
    ProgramStateMachineType_InstanceCount = 2396
    ProgramStateMachineType_MaxInstanceCount = 2397
    ProgramStateMachineType_MaxRecycleCount = 2398
    ProgramStateMachineType_ProgramDiagnostics = 2399
    ProgramStateMachineType_Ready = 2400
    ProgramStateMachineType_Ready_StateNumber = 2401
    ProgramStateMachineType_Running = 2402
    ProgramStateMachineType_Running_StateNumber = 2403
    ProgramStateMachineType_Suspended = 2404
    ProgramStateMachineType_Suspended_StateNumber = 2405
    ProgramStateMachineType_Halted = 2406
    ProgramStateMachineType_Halted_StateNumber = 2407
    ProgramStateMachineType_HaltedToReady = 2408
    ProgramStateMachineType_HaltedToReady_TransitionNumber = 2409
    ProgramStateMachineType_ReadyToRunning = 2410
    ProgramStateMachineType_ReadyToRunning_TransitionNumber = 2411
    ProgramStateMachineType_RunningToHalted = 2412
    ProgramStateMachineType_RunningToHalted_TransitionNumber = 2413
    ProgramStateMachineType_RunningToReady = 2414
    ProgramStateMachineType_RunningToReady_TransitionNumber = 2415
    ProgramStateMachineType_RunningToSuspended = 2416
    ProgramStateMachineType_RunningToSuspended_TransitionNumber = 2417
    ProgramStateMachineType_SuspendedToRunning = 2418
    ProgramStateMachineType_SuspendedToRunning_TransitionNumber = 2419
    ProgramStateMachineType_SuspendedToHalted = 2420
    ProgramStateMachineType_SuspendedToHalted_TransitionNumber = 2421
    ProgramStateMachineType_SuspendedToReady = 2422
    ProgramStateMachineType_SuspendedToReady_TransitionNumber = 2423
    ProgramStateMachineType_ReadyToHalted = 2424
    ProgramStateMachineType_ReadyToHalted_TransitionNumber = 2425
    ProgramStateMachineType_Start = 2426
    ProgramStateMachineType_Suspend = 2427
    ProgramStateMachineType_Resume = 2428
    ProgramStateMachineType_Halt = 2429
    ProgramStateMachineType_Reset = 2430
    SessionDiagnosticsVariableType_RegisterNodesCount = 2730
    SessionDiagnosticsVariableType_UnregisterNodesCount = 2731
    ServerCapabilitiesType_MaxBrowseContinuationPoints = 2732
    ServerCapabilitiesType_MaxQueryContinuationPoints = 2733
    ServerCapabilitiesType_MaxHistoryContinuationPoints = 2734
    Server_ServerCapabilities_MaxBrowseContinuationPoints = 2735
    Server_ServerCapabilities_MaxQueryContinuationPoints = 2736
    Server_ServerCapabilities_MaxHistoryContinuationPoints = 2737
    SemanticChangeEventType = 2738
    SemanticChangeEventType_Changes = 2739
    ServerType_Auditing = 2742
    ServerDiagnosticsType_SessionsDiagnosticsSummary = 2744
    AuditChannelEventType_SecureChannelId = 2745
    AuditOpenSecureChannelEventType_ClientCertificateThumbprint = 2746
    AuditCreateSessionEventType_ClientCertificateThumbprint = 2747
    AuditUrlMismatchEventType = 2748
    AuditUrlMismatchEventType_EndpointUrl = 2749
    AuditWriteUpdateEventType_AttributeId = 2750
    AuditHistoryUpdateEventType_ParameterDataTypeId = 2751
    ServerStatusType_SecondsTillShutdown = 2752
    ServerStatusType_ShutdownReason = 2753
    ServerCapabilitiesType_AggregateFunctions = 2754
    StateVariableType = 2755
    StateVariableType_Id = 2756
    StateVariableType_Name = 2757
    StateVariableType_Number = 2758
    StateVariableType_EffectiveDisplayName = 2759
    FiniteStateVariableType = 2760
    FiniteStateVariableType_Id = 2761
    TransitionVariableType = 2762
    TransitionVariableType_Id = 2763
    TransitionVariableType_Name = 2764
    TransitionVariableType_Number = 2765
    TransitionVariableType_TransitionTime = 2766
    FiniteTransitionVariableType = 2767
    FiniteTransitionVariableType_Id = 2768
    StateMachineType_CurrentState = 2769
    StateMachineType_LastTransition = 2770
    FiniteStateMachineType = 2771
    FiniteStateMachineType_CurrentState = 2772
    FiniteStateMachineType_LastTransition = 2773
    TransitionEventType_Transition = 2774
    TransitionEventType_FromState = 2775
    TransitionEventType_ToState = 2776
    AuditUpdateStateEventType_OldStateId = 2777
    AuditUpdateStateEventType_NewStateId = 2778
    ConditionType = 2782
    RefreshStartEventType = 2787
    RefreshEndEventType = 2788
    RefreshRequiredEventType = 2789
    AuditConditionEventType = 2790
    AuditConditionEnableEventType = 2803
    AuditConditionCommentEventType = 2829
    DialogConditionType = 2830
    DialogConditionType_Prompt = 2831
    AcknowledgeableConditionType = 2881
    AlarmConditionType = 2915
    ShelvedStateMachineType = 2929
    ShelvedStateMachineType_Unshelved = 2930
    ShelvedStateMachineType_TimedShelved = 2932
    ShelvedStateMachineType_OneShotShelved = 2933
    ShelvedStateMachineType_UnshelvedToTimedShelved = 2935
    ShelvedStateMachineType_UnshelvedToOneShotShelved = 2936
    ShelvedStateMachineType_TimedShelvedToUnshelved = 2940
    ShelvedStateMachineType_TimedShelvedToOneShotShelved = 2942
    ShelvedStateMachineType_OneShotShelvedToUnshelved = 2943
    ShelvedStateMachineType_OneShotShelvedToTimedShelved = 2945
    ShelvedStateMachineType_Unshelve = 2947
    ShelvedStateMachineType_OneShotShelve = 2948
    ShelvedStateMachineType_TimedShelve = 2949
    LimitAlarmType = 2955
    ShelvedStateMachineType_TimedShelve_InputArguments = 2991
    Server_ServerStatus_SecondsTillShutdown = 2992
    Server_ServerStatus_ShutdownReason = 2993
    Server_Auditing = 2994
    Server_ServerCapabilities_ModellingRules = 2996
    Server_ServerCapabilities_AggregateFunctions = 2997
    SubscriptionDiagnosticsType_EventNotificationsCount = 2998
    AuditHistoryEventUpdateEventType = 2999
    AuditHistoryEventUpdateEventType_Filter = 3003
    AuditHistoryValueUpdateEventType = 3006
    AuditHistoryDeleteEventType = 3012
    AuditHistoryRawModifyDeleteEventType = 3014
    AuditHistoryRawModifyDeleteEventType_IsDeleteModified = 3015
    AuditHistoryRawModifyDeleteEventType_StartTime = 3016
    AuditHistoryRawModifyDeleteEventType_EndTime = 3017
    AuditHistoryAtTimeDeleteEventType = 3019
    AuditHistoryAtTimeDeleteEventType_ReqTimes = 3020
    AuditHistoryAtTimeDeleteEventType_OldValues = 3021
    AuditHistoryEventDeleteEventType = 3022
    AuditHistoryEventDeleteEventType_EventIds = 3023
    AuditHistoryEventDeleteEventType_OldValues = 3024
    AuditHistoryEventUpdateEventType_UpdatedNode = 3025
    AuditHistoryValueUpdateEventType_UpdatedNode = 3026
    AuditHistoryDeleteEventType_UpdatedNode = 3027
    AuditHistoryEventUpdateEventType_PerformInsertReplace = 3028
    AuditHistoryEventUpdateEventType_NewValues = 3029
    AuditHistoryEventUpdateEventType_OldValues = 3030
    AuditHistoryValueUpdateEventType_PerformInsertReplace = 3031
    AuditHistoryValueUpdateEventType_NewValues = 3032
    AuditHistoryValueUpdateEventType_OldValues = 3033
    AuditHistoryRawModifyDeleteEventType_OldValues = 3034
    EventQueueOverflowEventType = 3035
    EventTypesFolder = 3048
    ServerCapabilitiesType_SoftwareCertificates = 3049
    SessionDiagnosticsVariableType_MaxResponseMessageSize = 3050
    BuildInfoType = 3051
    BuildInfoType_ProductUri = 3052
    BuildInfoType_ManufacturerName = 3053
    BuildInfoType_ProductName = 3054
    BuildInfoType_SoftwareVersion = 3055
    BuildInfoType_BuildNumber = 3056
    BuildInfoType_BuildDate = 3057
    SessionSecurityDiagnosticsType_ClientCertificate = 3058
    HistoricalDataConfigurationType_AggregateConfiguration = 3059
    DefaultBinary = 3062
    DefaultXml = 3063
    AlwaysGeneratesEvent = 3065
    Icon = 3067
    NodeVersion = 3068
    LocalTime = 3069
    AllowNulls = 3070
    EnumValues = 3071
    InputArguments = 3072
    OutputArguments = 3073
    ServerType_ServerStatus_StartTime = 3074
    ServerType_ServerStatus_CurrentTime = 3075
    ServerType_ServerStatus_State = 3076
    ServerType_ServerStatus_BuildInfo = 3077
    ServerType_ServerStatus_BuildInfo_ProductUri = 3078
    ServerType_ServerStatus_BuildInfo_ManufacturerName = 3079
    ServerType_ServerStatus_BuildInfo_ProductName = 3080
    ServerType_ServerStatus_BuildInfo_SoftwareVersion = 3081
    ServerType_ServerStatus_BuildInfo_BuildNumber = 3082
    ServerType_ServerStatus_BuildInfo_BuildDate = 3083
    ServerType_ServerStatus_SecondsTillShutdown = 3084
    ServerType_ServerStatus_ShutdownReason = 3085
    ServerType_ServerCapabilities_ServerProfileArray = 3086
    ServerType_ServerCapabilities_LocaleIdArray = 3087
    ServerType_ServerCapabilities_MinSupportedSampleRate = 3088
    ServerType_ServerCapabilities_MaxBrowseContinuationPoints = 3089
    ServerType_ServerCapabilities_MaxQueryContinuationPoints = 3090
    ServerType_ServerCapabilities_MaxHistoryContinuationPoints = 3091
    ServerType_ServerCapabilities_SoftwareCertificates = 3092
    ServerType_ServerCapabilities_ModellingRules = 3093
    ServerType_ServerCapabilities_AggregateFunctions = 3094
    ServerType_ServerDiagnostics_ServerDiagnosticsSummary = 3095
    ServerType_ServerDiagnostics_ServerDiagnosticsSummary_ServerViewCount = 3096
    ServerType_ServerDiagnostics_ServerDiagnosticsSummary_CurrentSessionCount = 3097
    ServerType_ServerDiagnostics_ServerDiagnosticsSummary_CumulatedSessionCount = 3098
    ServerType_ServerDiagnostics_ServerDiagnosticsSummary_SecurityRejectedSessionCount = 3099
    ServerType_ServerDiagnostics_ServerDiagnosticsSummary_RejectedSessionCount = 3100
    ServerType_ServerDiagnostics_ServerDiagnosticsSummary_SessionTimeoutCount = 3101
    ServerType_ServerDiagnostics_ServerDiagnosticsSummary_SessionAbortCount = 3102
    ServerType_ServerDiagnostics_ServerDiagnosticsSummary_PublishingIntervalCount = 3104
    ServerType_ServerDiagnostics_ServerDiagnosticsSummary_CurrentSubscriptionCount = 3105
    ServerType_ServerDiagnostics_ServerDiagnosticsSummary_CumulatedSubscriptionCount = 3106
    ServerType_ServerDiagnostics_ServerDiagnosticsSummary_SecurityRejectedRequestsCount = 3107
    ServerType_ServerDiagnostics_ServerDiagnosticsSummary_RejectedRequestsCount = 3108
    ServerType_ServerDiagnostics_SamplingIntervalDiagnosticsArray = 3109
    ServerType_ServerDiagnostics_SubscriptionDiagnosticsArray = 3110
    ServerType_ServerDiagnostics_SessionsDiagnosticsSummary = 3111
    ServerType_ServerDiagnostics_SessionsDiagnosticsSummary_SessionDiagnosticsArray = 3112
    ServerType_ServerDiagnostics_SessionsDiagnosticsSummary_SessionSecurityDiagnosticsArray = 3113
    ServerType_ServerDiagnostics_EnabledFlag = 3114
    ServerType_ServerRedundancy_RedundancySupport = 3115
    ServerDiagnosticsType_ServerDiagnosticsSummary_ServerViewCount = 3116
    ServerDiagnosticsType_ServerDiagnosticsSummary_CurrentSessionCount = 3117
    ServerDiagnosticsType_ServerDiagnosticsSummary_CumulatedSessionCount = 3118
    ServerDiagnosticsType_ServerDiagnosticsSummary_SecurityRejectedSessionCount = 3119
    ServerDiagnosticsType_ServerDiagnosticsSummary_RejectedSessionCount = 3120
    ServerDiagnosticsType_ServerDiagnosticsSummary_SessionTimeoutCount = 3121
    ServerDiagnosticsType_ServerDiagnosticsSummary_SessionAbortCount = 3122
    ServerDiagnosticsType_ServerDiagnosticsSummary_PublishingIntervalCount = 3124
    ServerDiagnosticsType_ServerDiagnosticsSummary_CurrentSubscriptionCount = 3125
    ServerDiagnosticsType_ServerDiagnosticsSummary_CumulatedSubscriptionCount = 3126
    ServerDiagnosticsType_ServerDiagnosticsSummary_SecurityRejectedRequestsCount = 3127
    ServerDiagnosticsType_ServerDiagnosticsSummary_RejectedRequestsCount = 3128
    ServerDiagnosticsType_SessionsDiagnosticsSummary_SessionDiagnosticsArray = 3129
    ServerDiagnosticsType_SessionsDiagnosticsSummary_SessionSecurityDiagnosticsArray = 3130
    SessionDiagnosticsObjectType_SessionDiagnostics_SessionId = 3131
    SessionDiagnosticsObjectType_SessionDiagnostics_SessionName = 3132
    SessionDiagnosticsObjectType_SessionDiagnostics_ClientDescription = 3133
    SessionDiagnosticsObjectType_SessionDiagnostics_ServerUri = 3134
    SessionDiagnosticsObjectType_SessionDiagnostics_EndpointUrl = 3135
    SessionDiagnosticsObjectType_SessionDiagnostics_LocaleIds = 3136
    SessionDiagnosticsObjectType_SessionDiagnostics_ActualSessionTimeout = 3137
    SessionDiagnosticsObjectType_SessionDiagnostics_MaxResponseMessageSize = 3138
    SessionDiagnosticsObjectType_SessionDiagnostics_ClientConnectionTime = 3139
    SessionDiagnosticsObjectType_SessionDiagnostics_ClientLastContactTime = 3140
    SessionDiagnosticsObjectType_SessionDiagnostics_CurrentSubscriptionsCount = 3141
    SessionDiagnosticsObjectType_SessionDiagnostics_CurrentMonitoredItemsCount = 3142
    SessionDiagnosticsObjectType_SessionDiagnostics_CurrentPublishRequestsInQueue = 3143
    SessionDiagnosticsObjectType_SessionDiagnostics_ReadCount = 3151
    SessionDiagnosticsObjectType_SessionDiagnostics_HistoryReadCount = 3152
    SessionDiagnosticsObjectType_SessionDiagnostics_WriteCount = 3153
    SessionDiagnosticsObjectType_SessionDiagnostics_HistoryUpdateCount = 3154
    SessionDiagnosticsObjectType_SessionDiagnostics_CallCount = 3155
    SessionDiagnosticsObjectType_SessionDiagnostics_CreateMonitoredItemsCount = 3156
    SessionDiagnosticsObjectType_SessionDiagnostics_ModifyMonitoredItemsCount = 3157
    SessionDiagnosticsObjectType_SessionDiagnostics_SetMonitoringModeCount = 3158
    SessionDiagnosticsObjectType_SessionDiagnostics_SetTriggeringCount = 3159
    SessionDiagnosticsObjectType_SessionDiagnostics_DeleteMonitoredItemsCount = 3160
    SessionDiagnosticsObjectType_SessionDiagnostics_CreateSubscriptionCount = 3161
    SessionDiagnosticsObjectType_SessionDiagnostics_ModifySubscriptionCount = 3162
    SessionDiagnosticsObjectType_SessionDiagnostics_SetPublishingModeCount = 3163
    SessionDiagnosticsObjectType_SessionDiagnostics_PublishCount = 3164
    SessionDiagnosticsObjectType_SessionDiagnostics_RepublishCount = 3165
    SessionDiagnosticsObjectType_SessionDiagnostics_TransferSubscriptionsCount = 3166
    SessionDiagnosticsObjectType_SessionDiagnostics_DeleteSubscriptionsCount = 3167
    SessionDiagnosticsObjectType_SessionDiagnostics_AddNodesCount = 3168
    SessionDiagnosticsObjectType_SessionDiagnostics_AddReferencesCount = 3169
    SessionDiagnosticsObjectType_SessionDiagnostics_DeleteNodesCount = 3170
    SessionDiagnosticsObjectType_SessionDiagnostics_DeleteReferencesCount = 3171
    SessionDiagnosticsObjectType_SessionDiagnostics_BrowseCount = 3172
    SessionDiagnosticsObjectType_SessionDiagnostics_BrowseNextCount = 3173
    SessionDiagnosticsObjectType_SessionDiagnostics_TranslateBrowsePathsToNodeIdsCount = 3174
    SessionDiagnosticsObjectType_SessionDiagnostics_QueryFirstCount = 3175
    SessionDiagnosticsObjectType_SessionDiagnostics_QueryNextCount = 3176
    SessionDiagnosticsObjectType_SessionDiagnostics_RegisterNodesCount = 3177
    SessionDiagnosticsObjectType_SessionDiagnostics_UnregisterNodesCount = 3178
    SessionDiagnosticsObjectType_SessionSecurityDiagnostics_SessionId = 3179
    SessionDiagnosticsObjectType_SessionSecurityDiagnostics_ClientUserIdOfSession = 3180
    SessionDiagnosticsObjectType_SessionSecurityDiagnostics_ClientUserIdHistory = 3181
    SessionDiagnosticsObjectType_SessionSecurityDiagnostics_AuthenticationMechanism = 3182
    SessionDiagnosticsObjectType_SessionSecurityDiagnostics_Encoding = 3183
    SessionDiagnosticsObjectType_SessionSecurityDiagnostics_TransportProtocol = 3184
    SessionDiagnosticsObjectType_SessionSecurityDiagnostics_SecurityMode = 3185
    SessionDiagnosticsObjectType_SessionSecurityDiagnostics_SecurityPolicyUri = 3186
    SessionDiagnosticsObjectType_SessionSecurityDiagnostics_ClientCertificate = 3187
    TransparentRedundancyType_RedundancySupport = 3188
    NonTransparentRedundancyType_RedundancySupport = 3189
    BaseEventType_LocalTime = 3190
    EventQueueOverflowEventType_EventId = 3191
    EventQueueOverflowEventType_EventType = 3192
    EventQueueOverflowEventType_SourceNode = 3193
    EventQueueOverflowEventType_SourceName = 3194
    EventQueueOverflowEventType_Time = 3195
    EventQueueOverflowEventType_ReceiveTime = 3196
    EventQueueOverflowEventType_LocalTime = 3197
    EventQueueOverflowEventType_Message = 3198
    EventQueueOverflowEventType_Severity = 3199
    AuditEventType_EventId = 3200
    AuditEventType_EventType = 3201
    AuditEventType_SourceNode = 3202
    AuditEventType_SourceName = 3203
    AuditEventType_Time = 3204
    AuditEventType_ReceiveTime = 3205
    AuditEventType_LocalTime = 3206
    AuditEventType_Message = 3207
    AuditEventType_Severity = 3208
    AuditSecurityEventType_EventId = 3209
    AuditSecurityEventType_EventType = 3210
    AuditSecurityEventType_SourceNode = 3211
    AuditSecurityEventType_SourceName = 3212
    AuditSecurityEventType_Time = 3213
    AuditSecurityEventType_ReceiveTime = 3214
    AuditSecurityEventType_LocalTime = 3215
    AuditSecurityEventType_Message = 3216
    AuditSecurityEventType_Severity = 3217
    AuditSecurityEventType_ActionTimeStamp = 3218
    AuditSecurityEventType_Status = 3219
    AuditSecurityEventType_ServerId = 3220
    AuditSecurityEventType_ClientAuditEntryId = 3221
    AuditSecurityEventType_ClientUserId = 3222
    AuditChannelEventType_EventId = 3223
    AuditChannelEventType_EventType = 3224
    AuditChannelEventType_SourceNode = 3225
    AuditChannelEventType_SourceName = 3226
    AuditChannelEventType_Time = 3227
    AuditChannelEventType_ReceiveTime = 3228
    AuditChannelEventType_LocalTime = 3229
    AuditChannelEventType_Message = 3230
    AuditChannelEventType_Severity = 3231
    AuditChannelEventType_ActionTimeStamp = 3232
    AuditChannelEventType_Status = 3233
    AuditChannelEventType_ServerId = 3234
    AuditChannelEventType_ClientAuditEntryId = 3235
    AuditChannelEventType_ClientUserId = 3236
    AuditOpenSecureChannelEventType_EventId = 3237
    AuditOpenSecureChannelEventType_EventType = 3238
    AuditOpenSecureChannelEventType_SourceNode = 3239
    AuditOpenSecureChannelEventType_SourceName = 3240
    AuditOpenSecureChannelEventType_Time = 3241
    AuditOpenSecureChannelEventType_ReceiveTime = 3242
    AuditOpenSecureChannelEventType_LocalTime = 3243
    AuditOpenSecureChannelEventType_Message = 3244
    AuditOpenSecureChannelEventType_Severity = 3245
    AuditOpenSecureChannelEventType_ActionTimeStamp = 3246
    AuditOpenSecureChannelEventType_Status = 3247
    AuditOpenSecureChannelEventType_ServerId = 3248
    AuditOpenSecureChannelEventType_ClientAuditEntryId = 3249
    AuditOpenSecureChannelEventType_ClientUserId = 3250
    AuditOpenSecureChannelEventType_SecureChannelId = 3251
    AuditSessionEventType_EventId = 3252
    AuditSessionEventType_EventType = 3253
    AuditSessionEventType_SourceNode = 3254
    AuditSessionEventType_SourceName = 3255
    AuditSessionEventType_Time = 3256
    AuditSessionEventType_ReceiveTime = 3257
    AuditSessionEventType_LocalTime = 3258
    AuditSessionEventType_Message = 3259
    AuditSessionEventType_Severity = 3260
    AuditSessionEventType_ActionTimeStamp = 3261
    AuditSessionEventType_Status = 3262
    AuditSessionEventType_ServerId = 3263
    AuditSessionEventType_ClientAuditEntryId = 3264
    AuditSessionEventType_ClientUserId = 3265
    AuditCreateSessionEventType_EventId = 3266
    AuditCreateSessionEventType_EventType = 3267
    AuditCreateSessionEventType_SourceNode = 3268
    AuditCreateSessionEventType_SourceName = 3269
    AuditCreateSessionEventType_Time = 3270
    AuditCreateSessionEventType_ReceiveTime = 3271
    AuditCreateSessionEventType_LocalTime = 3272
    AuditCreateSessionEventType_Message = 3273
    AuditCreateSessionEventType_Severity = 3274
    AuditCreateSessionEventType_ActionTimeStamp = 3275
    AuditCreateSessionEventType_Status = 3276
    AuditCreateSessionEventType_ServerId = 3277
    AuditCreateSessionEventType_ClientAuditEntryId = 3278
    AuditCreateSessionEventType_ClientUserId = 3279
    AuditUrlMismatchEventType_EventId = 3281
    AuditUrlMismatchEventType_EventType = 3282
    AuditUrlMismatchEventType_SourceNode = 3283
    AuditUrlMismatchEventType_SourceName = 3284
    AuditUrlMismatchEventType_Time = 3285
    AuditUrlMismatchEventType_ReceiveTime = 3286
    AuditUrlMismatchEventType_LocalTime = 3287
    AuditUrlMismatchEventType_Message = 3288
    AuditUrlMismatchEventType_Severity = 3289
    AuditUrlMismatchEventType_ActionTimeStamp = 3290
    AuditUrlMismatchEventType_Status = 3291
    AuditUrlMismatchEventType_ServerId = 3292
    AuditUrlMismatchEventType_ClientAuditEntryId = 3293
    AuditUrlMismatchEventType_ClientUserId = 3294
    AuditUrlMismatchEventType_SecureChannelId = 3296
    AuditUrlMismatchEventType_ClientCertificate = 3297
    AuditUrlMismatchEventType_ClientCertificateThumbprint = 3298
    AuditUrlMismatchEventType_RevisedSessionTimeout = 3299
    AuditActivateSessionEventType_EventId = 3300
    AuditActivateSessionEventType_EventType = 3301
    AuditActivateSessionEventType_SourceNode = 3302
    AuditActivateSessionEventType_SourceName = 3303
    AuditActivateSessionEventType_Time = 3304
    AuditActivateSessionEventType_ReceiveTime = 3305
    AuditActivateSessionEventType_LocalTime = 3306
    AuditActivateSessionEventType_Message = 3307
    AuditActivateSessionEventType_Severity = 3308
    AuditActivateSessionEventType_ActionTimeStamp = 3309
    AuditActivateSessionEventType_Status = 3310
    AuditActivateSessionEventType_ServerId = 3311
    AuditActivateSessionEventType_ClientAuditEntryId = 3312
    AuditActivateSessionEventType_ClientUserId = 3313
    AuditActivateSessionEventType_SessionId = 3314
    AuditCancelEventType_EventId = 3315
    AuditCancelEventType_EventType = 3316
    AuditCancelEventType_SourceNode = 3317
    AuditCancelEventType_SourceName = 3318
    AuditCancelEventType_Time = 3319
    AuditCancelEventType_ReceiveTime = 3320
    AuditCancelEventType_LocalTime = 3321
    AuditCancelEventType_Message = 3322
    AuditCancelEventType_Severity = 3323
    AuditCancelEventType_ActionTimeStamp = 3324
    AuditCancelEventType_Status = 3325
    AuditCancelEventType_ServerId = 3326
    AuditCancelEventType_ClientAuditEntryId = 3327
    AuditCancelEventType_ClientUserId = 3328
    AuditCancelEventType_SessionId = 3329
    AuditCertificateEventType_EventId = 3330
    AuditCertificateEventType_EventType = 3331
    AuditCertificateEventType_SourceNode = 3332
    AuditCertificateEventType_SourceName = 3333
    AuditCertificateEventType_Time = 3334
    AuditCertificateEventType_ReceiveTime = 3335
    AuditCertificateEventType_LocalTime = 3336
    AuditCertificateEventType_Message = 3337
    AuditCertificateEventType_Severity = 3338
    AuditCertificateEventType_ActionTimeStamp = 3339
    AuditCertificateEventType_Status = 3340
    AuditCertificateEventType_ServerId = 3341
    AuditCertificateEventType_ClientAuditEntryId = 3342
    AuditCertificateEventType_ClientUserId = 3343
    AuditCertificateDataMismatchEventType_EventId = 3344
    AuditCertificateDataMismatchEventType_EventType = 3345
    AuditCertificateDataMismatchEventType_SourceNode = 3346
    AuditCertificateDataMismatchEventType_SourceName = 3347
    AuditCertificateDataMismatchEventType_Time = 3348
    AuditCertificateDataMismatchEventType_ReceiveTime = 3349
    AuditCertificateDataMismatchEventType_LocalTime = 3350
    AuditCertificateDataMismatchEventType_Message = 3351
    AuditCertificateDataMismatchEventType_Severity = 3352
    AuditCertificateDataMismatchEventType_ActionTimeStamp = 3353
    AuditCertificateDataMismatchEventType_Status = 3354
    AuditCertificateDataMismatchEventType_ServerId = 3355
    AuditCertificateDataMismatchEventType_ClientAuditEntryId = 3356
    AuditCertificateDataMismatchEventType_ClientUserId = 3357
    AuditCertificateDataMismatchEventType_Certificate = 3358
    AuditCertificateExpiredEventType_EventId = 3359
    AuditCertificateExpiredEventType_EventType = 3360
    AuditCertificateExpiredEventType_SourceNode = 3361
    AuditCertificateExpiredEventType_SourceName = 3362
    AuditCertificateExpiredEventType_Time = 3363
    AuditCertificateExpiredEventType_ReceiveTime = 3364
    AuditCertificateExpiredEventType_LocalTime = 3365
    AuditCertificateExpiredEventType_Message = 3366
    AuditCertificateExpiredEventType_Severity = 3367
    AuditCertificateExpiredEventType_ActionTimeStamp = 3368
    AuditCertificateExpiredEventType_Status = 3369
    AuditCertificateExpiredEventType_ServerId = 3370
    AuditCertificateExpiredEventType_ClientAuditEntryId = 3371
    AuditCertificateExpiredEventType_ClientUserId = 3372
    AuditCertificateExpiredEventType_Certificate = 3373
    AuditCertificateInvalidEventType_EventId = 3374
    AuditCertificateInvalidEventType_EventType = 3375
    AuditCertificateInvalidEventType_SourceNode = 3376
    AuditCertificateInvalidEventType_SourceName = 3377
    AuditCertificateInvalidEventType_Time = 3378
    AuditCertificateInvalidEventType_ReceiveTime = 3379
    AuditCertificateInvalidEventType_LocalTime = 3380
    AuditCertificateInvalidEventType_Message = 3381
    AuditCertificateInvalidEventType_Severity = 3382
    AuditCertificateInvalidEventType_ActionTimeStamp = 3383
    AuditCertificateInvalidEventType_Status = 3384
    AuditCertificateInvalidEventType_ServerId = 3385
    AuditCertificateInvalidEventType_ClientAuditEntryId = 3386
    AuditCertificateInvalidEventType_ClientUserId = 3387
    AuditCertificateInvalidEventType_Certificate = 3388
    AuditCertificateUntrustedEventType_EventId = 3389
    AuditCertificateUntrustedEventType_EventType = 3390
    AuditCertificateUntrustedEventType_SourceNode = 3391
    AuditCertificateUntrustedEventType_SourceName = 3392
    AuditCertificateUntrustedEventType_Time = 3393
    AuditCertificateUntrustedEventType_ReceiveTime = 3394
    AuditCertificateUntrustedEventType_LocalTime = 3395
    AuditCertificateUntrustedEventType_Message = 3396
    AuditCertificateUntrustedEventType_Severity = 3397
    AuditCertificateUntrustedEventType_ActionTimeStamp = 3398
    AuditCertificateUntrustedEventType_Status = 3399
    AuditCertificateUntrustedEventType_ServerId = 3400
    AuditCertificateUntrustedEventType_ClientAuditEntryId = 3401
    AuditCertificateUntrustedEventType_ClientUserId = 3402
    AuditCertificateUntrustedEventType_Certificate = 3403
    AuditCertificateRevokedEventType_EventId = 3404
    AuditCertificateRevokedEventType_EventType = 3405
    AuditCertificateRevokedEventType_SourceNode = 3406
    AuditCertificateRevokedEventType_SourceName = 3407
    AuditCertificateRevokedEventType_Time = 3408
    AuditCertificateRevokedEventType_ReceiveTime = 3409
    AuditCertificateRevokedEventType_LocalTime = 3410
    AuditCertificateRevokedEventType_Message = 3411
    AuditCertificateRevokedEventType_Severity = 3412
    AuditCertificateRevokedEventType_ActionTimeStamp = 3413
    AuditCertificateRevokedEventType_Status = 3414
    AuditCertificateRevokedEventType_ServerId = 3415
    AuditCertificateRevokedEventType_ClientAuditEntryId = 3416
    AuditCertificateRevokedEventType_ClientUserId = 3417
    AuditCertificateRevokedEventType_Certificate = 3418
    AuditCertificateMismatchEventType_EventId = 3419
    AuditCertificateMismatchEventType_EventType = 3420
    AuditCertificateMismatchEventType_SourceNode = 3421
    AuditCertificateMismatchEventType_SourceName = 3422
    AuditCertificateMismatchEventType_Time = 3423
    AuditCertificateMismatchEventType_ReceiveTime = 3424
    AuditCertificateMismatchEventType_LocalTime = 3425
    AuditCertificateMismatchEventType_Message = 3426
    AuditCertificateMismatchEventType_Severity = 3427
    AuditCertificateMismatchEventType_ActionTimeStamp = 3428
    AuditCertificateMismatchEventType_Status = 3429
    AuditCertificateMismatchEventType_ServerId = 3430
    AuditCertificateMismatchEventType_ClientAuditEntryId = 3431
    AuditCertificateMismatchEventType_ClientUserId = 3432
    AuditCertificateMismatchEventType_Certificate = 3433
    AuditNodeManagementEventType_EventId = 3434
    AuditNodeManagementEventType_EventType = 3435
    AuditNodeManagementEventType_SourceNode = 3436
    AuditNodeManagementEventType_SourceName = 3437
    AuditNodeManagementEventType_Time = 3438
    AuditNodeManagementEventType_ReceiveTime = 3439
    AuditNodeManagementEventType_LocalTime = 3440
    AuditNodeManagementEventType_Message = 3441
    AuditNodeManagementEventType_Severity = 3442
    AuditNodeManagementEventType_ActionTimeStamp = 3443
    AuditNodeManagementEventType_Status = 3444
    AuditNodeManagementEventType_ServerId = 3445
    AuditNodeManagementEventType_ClientAuditEntryId = 3446
    AuditNodeManagementEventType_ClientUserId = 3447
    AuditAddNodesEventType_EventId = 3448
    AuditAddNodesEventType_EventType = 3449
    AuditAddNodesEventType_SourceNode = 3450
    AuditAddNodesEventType_SourceName = 3451
    AuditAddNodesEventType_Time = 3452
    AuditAddNodesEventType_ReceiveTime = 3453
    AuditAddNodesEventType_LocalTime = 3454
    AuditAddNodesEventType_Message = 3455
    AuditAddNodesEventType_Severity = 3456
    AuditAddNodesEventType_ActionTimeStamp = 3457
    AuditAddNodesEventType_Status = 3458
    AuditAddNodesEventType_ServerId = 3459
    AuditAddNodesEventType_ClientAuditEntryId = 3460
    AuditAddNodesEventType_ClientUserId = 3461
    AuditDeleteNodesEventType_EventId = 3462
    AuditDeleteNodesEventType_EventType = 3463
    AuditDeleteNodesEventType_SourceNode = 3464
    AuditDeleteNodesEventType_SourceName = 3465
    AuditDeleteNodesEventType_Time = 3466
    AuditDeleteNodesEventType_ReceiveTime = 3467
    AuditDeleteNodesEventType_LocalTime = 3468
    AuditDeleteNodesEventType_Message = 3469
    AuditDeleteNodesEventType_Severity = 3470
    AuditDeleteNodesEventType_ActionTimeStamp = 3471
    AuditDeleteNodesEventType_Status = 3472
    AuditDeleteNodesEventType_ServerId = 3473
    AuditDeleteNodesEventType_ClientAuditEntryId = 3474
    AuditDeleteNodesEventType_ClientUserId = 3475
    AuditAddReferencesEventType_EventId = 3476
    AuditAddReferencesEventType_EventType = 3477
    AuditAddReferencesEventType_SourceNode = 3478
    AuditAddReferencesEventType_SourceName = 3479
    AuditAddReferencesEventType_Time = 3480
    AuditAddReferencesEventType_ReceiveTime = 3481
    AuditAddReferencesEventType_LocalTime = 3482
    AuditAddReferencesEventType_Message = 3483
    AuditAddReferencesEventType_Severity = 3484
    AuditAddReferencesEventType_ActionTimeStamp = 3485
    AuditAddReferencesEventType_Status = 3486
    AuditAddReferencesEventType_ServerId = 3487
    AuditAddReferencesEventType_ClientAuditEntryId = 3488
    AuditAddReferencesEventType_ClientUserId = 3489
    AuditDeleteReferencesEventType_EventId = 3490
    AuditDeleteReferencesEventType_EventType = 3491
    AuditDeleteReferencesEventType_SourceNode = 3492
    AuditDeleteReferencesEventType_SourceName = 3493
    AuditDeleteReferencesEventType_Time = 3494
    AuditDeleteReferencesEventType_ReceiveTime = 3495
    AuditDeleteReferencesEventType_LocalTime = 3496
    AuditDeleteReferencesEventType_Message = 3497
    AuditDeleteReferencesEventType_Severity = 3498
    AuditDeleteReferencesEventType_ActionTimeStamp = 3499
    AuditDeleteReferencesEventType_Status = 3500
    AuditDeleteReferencesEventType_ServerId = 3501
    AuditDeleteReferencesEventType_ClientAuditEntryId = 3502
    AuditDeleteReferencesEventType_ClientUserId = 3503
    AuditUpdateEventType_EventId = 3504
    AuditUpdateEventType_EventType = 3505
    AuditUpdateEventType_SourceNode = 3506
    AuditUpdateEventType_SourceName = 3507
    AuditUpdateEventType_Time = 3508
    AuditUpdateEventType_ReceiveTime = 3509
    AuditUpdateEventType_LocalTime = 3510
    AuditUpdateEventType_Message = 3511
    AuditUpdateEventType_Severity = 3512
    AuditUpdateEventType_ActionTimeStamp = 3513
    AuditUpdateEventType_Status = 3514
    AuditUpdateEventType_ServerId = 3515
    AuditUpdateEventType_ClientAuditEntryId = 3516
    AuditUpdateEventType_ClientUserId = 3517
    AuditWriteUpdateEventType_EventId = 3518
    AuditWriteUpdateEventType_EventType = 3519
    AuditWriteUpdateEventType_SourceNode = 3520
    AuditWriteUpdateEventType_SourceName = 3521
    AuditWriteUpdateEventType_Time = 3522
    AuditWriteUpdateEventType_ReceiveTime = 3523
    AuditWriteUpdateEventType_LocalTime = 3524
    AuditWriteUpdateEventType_Message = 3525
    AuditWriteUpdateEventType_Severity = 3526
    AuditWriteUpdateEventType_ActionTimeStamp = 3527
    AuditWriteUpdateEventType_Status = 3528
    AuditWriteUpdateEventType_ServerId = 3529
    AuditWriteUpdateEventType_ClientAuditEntryId = 3530
    AuditWriteUpdateEventType_ClientUserId = 3531
    AuditHistoryUpdateEventType_EventId = 3532
    AuditHistoryUpdateEventType_EventType = 3533
    AuditHistoryUpdateEventType_SourceNode = 3534
    AuditHistoryUpdateEventType_SourceName = 3535
    AuditHistoryUpdateEventType_Time = 3536
    AuditHistoryUpdateEventType_ReceiveTime = 3537
    AuditHistoryUpdateEventType_LocalTime = 3538
    AuditHistoryUpdateEventType_Message = 3539
    AuditHistoryUpdateEventType_Severity = 3540
    AuditHistoryUpdateEventType_ActionTimeStamp = 3541
    AuditHistoryUpdateEventType_Status = 3542
    AuditHistoryUpdateEventType_ServerId = 3543
    AuditHistoryUpdateEventType_ClientAuditEntryId = 3544
    AuditHistoryUpdateEventType_ClientUserId = 3545
    AuditHistoryEventUpdateEventType_EventId = 3546
    AuditHistoryEventUpdateEventType_EventType = 3547
    AuditHistoryEventUpdateEventType_SourceNode = 3548
    AuditHistoryEventUpdateEventType_SourceName = 3549
    AuditHistoryEventUpdateEventType_Time = 3550
    AuditHistoryEventUpdateEventType_ReceiveTime = 3551
    AuditHistoryEventUpdateEventType_LocalTime = 3552
    AuditHistoryEventUpdateEventType_Message = 3553
    AuditHistoryEventUpdateEventType_Severity = 3554
    AuditHistoryEventUpdateEventType_ActionTimeStamp = 3555
    AuditHistoryEventUpdateEventType_Status = 3556
    AuditHistoryEventUpdateEventType_ServerId = 3557
    AuditHistoryEventUpdateEventType_ClientAuditEntryId = 3558
    AuditHistoryEventUpdateEventType_ClientUserId = 3559
    AuditHistoryEventUpdateEventType_ParameterDataTypeId = 3560
    AuditHistoryValueUpdateEventType_EventId = 3561
    AuditHistoryValueUpdateEventType_EventType = 3562
    AuditHistoryValueUpdateEventType_SourceNode = 3563
    AuditHistoryValueUpdateEventType_SourceName = 3564
    AuditHistoryValueUpdateEventType_Time = 3565
    AuditHistoryValueUpdateEventType_ReceiveTime = 3566
    AuditHistoryValueUpdateEventType_LocalTime = 3567
    AuditHistoryValueUpdateEventType_Message = 3568
    AuditHistoryValueUpdateEventType_Severity = 3569
    AuditHistoryValueUpdateEventType_ActionTimeStamp = 3570
    AuditHistoryValueUpdateEventType_Status = 3571
    AuditHistoryValueUpdateEventType_ServerId = 3572
    AuditHistoryValueUpdateEventType_ClientAuditEntryId = 3573
    AuditHistoryValueUpdateEventType_ClientUserId = 3574
    AuditHistoryValueUpdateEventType_ParameterDataTypeId = 3575
    AuditHistoryDeleteEventType_EventId = 3576
    AuditHistoryDeleteEventType_EventType = 3577
    AuditHistoryDeleteEventType_SourceNode = 3578
    AuditHistoryDeleteEventType_SourceName = 3579
    AuditHistoryDeleteEventType_Time = 3580
    AuditHistoryDeleteEventType_ReceiveTime = 3581
    AuditHistoryDeleteEventType_LocalTime = 3582
    AuditHistoryDeleteEventType_Message = 3583
    AuditHistoryDeleteEventType_Severity = 3584
    AuditHistoryDeleteEventType_ActionTimeStamp = 3585
    AuditHistoryDeleteEventType_Status = 3586
    AuditHistoryDeleteEventType_ServerId = 3587
    AuditHistoryDeleteEventType_ClientAuditEntryId = 3588
    AuditHistoryDeleteEventType_ClientUserId = 3589
    AuditHistoryDeleteEventType_ParameterDataTypeId = 3590
    AuditHistoryRawModifyDeleteEventType_EventId = 3591
    AuditHistoryRawModifyDeleteEventType_EventType = 3592
    AuditHistoryRawModifyDeleteEventType_SourceNode = 3593
    AuditHistoryRawModifyDeleteEventType_SourceName = 3594
    AuditHistoryRawModifyDeleteEventType_Time = 3595
    AuditHistoryRawModifyDeleteEventType_ReceiveTime = 3596
    AuditHistoryRawModifyDeleteEventType_LocalTime = 3597
    AuditHistoryRawModifyDeleteEventType_Message = 3598
    AuditHistoryRawModifyDeleteEventType_Severity = 3599
    AuditHistoryRawModifyDeleteEventType_ActionTimeStamp = 3600
    AuditHistoryRawModifyDeleteEventType_Status = 3601
    AuditHistoryRawModifyDeleteEventType_ServerId = 3602
    AuditHistoryRawModifyDeleteEventType_ClientAuditEntryId = 3603
    AuditHistoryRawModifyDeleteEventType_ClientUserId = 3604
    AuditHistoryRawModifyDeleteEventType_ParameterDataTypeId = 3605
    AuditHistoryRawModifyDeleteEventType_UpdatedNode = 3606
    AuditHistoryAtTimeDeleteEventType_EventId = 3607
    AuditHistoryAtTimeDeleteEventType_EventType = 3608
    AuditHistoryAtTimeDeleteEventType_SourceNode = 3609
    AuditHistoryAtTimeDeleteEventType_SourceName = 3610
    AuditHistoryAtTimeDeleteEventType_Time = 3611
    AuditHistoryAtTimeDeleteEventType_ReceiveTime = 3612
    AuditHistoryAtTimeDeleteEventType_LocalTime = 3613
    AuditHistoryAtTimeDeleteEventType_Message = 3614
    AuditHistoryAtTimeDeleteEventType_Severity = 3615
    AuditHistoryAtTimeDeleteEventType_ActionTimeStamp = 3616
    AuditHistoryAtTimeDeleteEventType_Status = 3617
    AuditHistoryAtTimeDeleteEventType_ServerId = 3618
    AuditHistoryAtTimeDeleteEventType_ClientAuditEntryId = 3619
    AuditHistoryAtTimeDeleteEventType_ClientUserId = 3620
    AuditHistoryAtTimeDeleteEventType_ParameterDataTypeId = 3621
    AuditHistoryAtTimeDeleteEventType_UpdatedNode = 3622
    AuditHistoryEventDeleteEventType_EventId = 3623
    AuditHistoryEventDeleteEventType_EventType = 3624
    AuditHistoryEventDeleteEventType_SourceNode = 3625
    AuditHistoryEventDeleteEventType_SourceName = 3626
    AuditHistoryEventDeleteEventType_Time = 3627
    AuditHistoryEventDeleteEventType_ReceiveTime = 3628
    AuditHistoryEventDeleteEventType_LocalTime = 3629
    AuditHistoryEventDeleteEventType_Message = 3630
    AuditHistoryEventDeleteEventType_Severity = 3631
    AuditHistoryEventDeleteEventType_ActionTimeStamp = 3632
    AuditHistoryEventDeleteEventType_Status = 3633
    AuditHistoryEventDeleteEventType_ServerId = 3634
    AuditHistoryEventDeleteEventType_ClientAuditEntryId = 3635
    AuditHistoryEventDeleteEventType_ClientUserId = 3636
    AuditHistoryEventDeleteEventType_ParameterDataTypeId = 3637
    AuditHistoryEventDeleteEventType_UpdatedNode = 3638
    AuditUpdateMethodEventType_EventId = 3639
    AuditUpdateMethodEventType_EventType = 3640
    AuditUpdateMethodEventType_SourceNode = 3641
    AuditUpdateMethodEventType_SourceName = 3642
    AuditUpdateMethodEventType_Time = 3643
    AuditUpdateMethodEventType_ReceiveTime = 3644
    AuditUpdateMethodEventType_LocalTime = 3645
    AuditUpdateMethodEventType_Message = 3646
    AuditUpdateMethodEventType_Severity = 3647
    AuditUpdateMethodEventType_ActionTimeStamp = 3648
    AuditUpdateMethodEventType_Status = 3649
    AuditUpdateMethodEventType_ServerId = 3650
    AuditUpdateMethodEventType_ClientAuditEntryId = 3651
    AuditUpdateMethodEventType_ClientUserId = 3652
    SystemEventType_EventId = 3653
    SystemEventType_EventType = 3654
    SystemEventType_SourceNode = 3655
    SystemEventType_SourceName = 3656
    SystemEventType_Time = 3657
    SystemEventType_ReceiveTime = 3658
    SystemEventType_LocalTime = 3659
    SystemEventType_Message = 3660
    SystemEventType_Severity = 3661
    DeviceFailureEventType_EventId = 3662
    DeviceFailureEventType_EventType = 3663
    DeviceFailureEventType_SourceNode = 3664
    DeviceFailureEventType_SourceName = 3665
    DeviceFailureEventType_Time = 3666
    DeviceFailureEventType_ReceiveTime = 3667
    DeviceFailureEventType_LocalTime = 3668
    DeviceFailureEventType_Message = 3669
    DeviceFailureEventType_Severity = 3670
    BaseModelChangeEventType_EventId = 3671
    BaseModelChangeEventType_EventType = 3672
    BaseModelChangeEventType_SourceNode = 3673
    BaseModelChangeEventType_SourceName = 3674
    BaseModelChangeEventType_Time = 3675
    BaseModelChangeEventType_ReceiveTime = 3676
    BaseModelChangeEventType_LocalTime = 3677
    BaseModelChangeEventType_Message = 3678
    BaseModelChangeEventType_Severity = 3679
    GeneralModelChangeEventType_EventId = 3680
    GeneralModelChangeEventType_EventType = 3681
    GeneralModelChangeEventType_SourceNode = 3682
    GeneralModelChangeEventType_SourceName = 3683
    GeneralModelChangeEventType_Time = 3684
    GeneralModelChangeEventType_ReceiveTime = 3685
    GeneralModelChangeEventType_LocalTime = 3686
    GeneralModelChangeEventType_Message = 3687
    GeneralModelChangeEventType_Severity = 3688
    SemanticChangeEventType_EventId = 3689
    SemanticChangeEventType_EventType = 3690
    SemanticChangeEventType_SourceNode = 3691
    SemanticChangeEventType_SourceName = 3692
    SemanticChangeEventType_Time = 3693
    SemanticChangeEventType_ReceiveTime = 3694
    SemanticChangeEventType_LocalTime = 3695
    SemanticChangeEventType_Message = 3696
    SemanticChangeEventType_Severity = 3697
    ServerStatusType_BuildInfo_ProductUri = 3698
    ServerStatusType_BuildInfo_ManufacturerName = 3699
    ServerStatusType_BuildInfo_ProductName = 3700
    ServerStatusType_BuildInfo_SoftwareVersion = 3701
    ServerStatusType_BuildInfo_BuildNumber = 3702
    ServerStatusType_BuildInfo_BuildDate = 3703
    Server_ServerCapabilities_SoftwareCertificates = 3704
    Server_ServerDiagnostics_ServerDiagnosticsSummary_RejectedSessionCount = 3705
    Server_ServerDiagnostics_SessionsDiagnosticsSummary = 3706
    Server_ServerDiagnostics_SessionsDiagnosticsSummary_SessionDiagnosticsArray = 3707
    Server_ServerDiagnostics_SessionsDiagnosticsSummary_SessionSecurityDiagnosticsArray = 3708
    Server_ServerRedundancy_RedundancySupport = 3709
    FiniteStateVariableType_Name = 3714
    FiniteStateVariableType_Number = 3715
    FiniteStateVariableType_EffectiveDisplayName = 3716
    FiniteTransitionVariableType_Name = 3717
    FiniteTransitionVariableType_Number = 3718
    FiniteTransitionVariableType_TransitionTime = 3719
    StateMachineType_CurrentState_Id = 3720
    StateMachineType_CurrentState_Name = 3721
    StateMachineType_CurrentState_Number = 3722
    StateMachineType_CurrentState_EffectiveDisplayName = 3723
    StateMachineType_LastTransition_Id = 3724
    StateMachineType_LastTransition_Name = 3725
    StateMachineType_LastTransition_Number = 3726
    StateMachineType_LastTransition_TransitionTime = 3727
    FiniteStateMachineType_CurrentState_Id = 3728
    FiniteStateMachineType_CurrentState_Name = 3729
    FiniteStateMachineType_CurrentState_Number = 3730
    FiniteStateMachineType_CurrentState_EffectiveDisplayName = 3731
    FiniteStateMachineType_LastTransition_Id = 3732
    FiniteStateMachineType_LastTransition_Name = 3733
    FiniteStateMachineType_LastTransition_Number = 3734
    FiniteStateMachineType_LastTransition_TransitionTime = 3735
    InitialStateType_StateNumber = 3736
    TransitionEventType_EventId = 3737
    TransitionEventType_EventType = 3738
    TransitionEventType_SourceNode = 3739
    TransitionEventType_SourceName = 3740
    TransitionEventType_Time = 3741
    TransitionEventType_ReceiveTime = 3742
    TransitionEventType_LocalTime = 3743
    TransitionEventType_Message = 3744
    TransitionEventType_Severity = 3745
    TransitionEventType_FromState_Id = 3746
    TransitionEventType_FromState_Name = 3747
    TransitionEventType_FromState_Number = 3748
    TransitionEventType_FromState_EffectiveDisplayName = 3749
    TransitionEventType_ToState_Id = 3750
    TransitionEventType_ToState_Name = 3751
    TransitionEventType_ToState_Number = 3752
    TransitionEventType_ToState_EffectiveDisplayName = 3753
    TransitionEventType_Transition_Id = 3754
    TransitionEventType_Transition_Name = 3755
    TransitionEventType_Transition_Number = 3756
    TransitionEventType_Transition_TransitionTime = 3757
    AuditUpdateStateEventType_EventId = 3758
    AuditUpdateStateEventType_EventType = 3759
    AuditUpdateStateEventType_SourceNode = 3760
    AuditUpdateStateEventType_SourceName = 3761
    AuditUpdateStateEventType_Time = 3762
    AuditUpdateStateEventType_ReceiveTime = 3763
    AuditUpdateStateEventType_LocalTime = 3764
    AuditUpdateStateEventType_Message = 3765
    AuditUpdateStateEventType_Severity = 3766
    AuditUpdateStateEventType_ActionTimeStamp = 3767
    AuditUpdateStateEventType_Status = 3768
    AuditUpdateStateEventType_ServerId = 3769
    AuditUpdateStateEventType_ClientAuditEntryId = 3770
    AuditUpdateStateEventType_ClientUserId = 3771
    AuditUpdateStateEventType_MethodId = 3772
    AuditUpdateStateEventType_InputArguments = 3773
    AnalogItemType_Definition = 3774
    AnalogItemType_ValuePrecision = 3775
    DiscreteItemType_Definition = 3776
    DiscreteItemType_ValuePrecision = 3777
    TwoStateDiscreteType_Definition = 3778
    TwoStateDiscreteType_ValuePrecision = 3779
    MultiStateDiscreteType_Definition = 3780
    MultiStateDiscreteType_ValuePrecision = 3781
    ProgramTransitionEventType_EventId = 3782
    ProgramTransitionEventType_EventType = 3783
    ProgramTransitionEventType_SourceNode = 3784
    ProgramTransitionEventType_SourceName = 3785
    ProgramTransitionEventType_Time = 3786
    ProgramTransitionEventType_ReceiveTime = 3787
    ProgramTransitionEventType_LocalTime = 3788
    ProgramTransitionEventType_Message = 3789
    ProgramTransitionEventType_Severity = 3790
    ProgramTransitionEventType_FromState = 3791
    ProgramTransitionEventType_FromState_Id = 3792
    ProgramTransitionEventType_FromState_Name = 3793
    ProgramTransitionEventType_FromState_Number = 3794
    ProgramTransitionEventType_FromState_EffectiveDisplayName = 3795
    ProgramTransitionEventType_ToState = 3796
    ProgramTransitionEventType_ToState_Id = 3797
    ProgramTransitionEventType_ToState_Name = 3798
    ProgramTransitionEventType_ToState_Number = 3799
    ProgramTransitionEventType_ToState_EffectiveDisplayName = 3800
    ProgramTransitionEventType_Transition = 3801
    ProgramTransitionEventType_Transition_Id = 3802
    ProgramTransitionEventType_Transition_Name = 3803
    ProgramTransitionEventType_Transition_Number = 3804
    ProgramTransitionEventType_Transition_TransitionTime = 3805
    ProgramTransitionAuditEventType = 3806
    ProgramTransitionAuditEventType_EventId = 3807
    ProgramTransitionAuditEventType_EventType = 3808
    ProgramTransitionAuditEventType_SourceNode = 3809
    ProgramTransitionAuditEventType_SourceName = 3810
    ProgramTransitionAuditEventType_Time = 3811
    ProgramTransitionAuditEventType_ReceiveTime = 3812
    ProgramTransitionAuditEventType_LocalTime = 3813
    ProgramTransitionAuditEventType_Message = 3814
    ProgramTransitionAuditEventType_Severity = 3815
    ProgramTransitionAuditEventType_ActionTimeStamp = 3816
    ProgramTransitionAuditEventType_Status = 3817
    ProgramTransitionAuditEventType_ServerId = 3818
    ProgramTransitionAuditEventType_ClientAuditEntryId = 3819
    ProgramTransitionAuditEventType_ClientUserId = 3820
    ProgramTransitionAuditEventType_MethodId = 3821
    ProgramTransitionAuditEventType_InputArguments = 3822
    ProgramTransitionAuditEventType_OldStateId = 3823
    ProgramTransitionAuditEventType_NewStateId = 3824
    ProgramTransitionAuditEventType_Transition = 3825
    ProgramTransitionAuditEventType_Transition_Id = 3826
    ProgramTransitionAuditEventType_Transition_Name = 3827
    ProgramTransitionAuditEventType_Transition_Number = 3828
    ProgramTransitionAuditEventType_Transition_TransitionTime = 3829
    ProgramStateMachineType_CurrentState = 3830
    ProgramStateMachineType_CurrentState_Id = 3831
    ProgramStateMachineType_CurrentState_Name = 3832
    ProgramStateMachineType_CurrentState_Number = 3833
    ProgramStateMachineType_CurrentState_EffectiveDisplayName = 3834
    ProgramStateMachineType_LastTransition = 3835
    ProgramStateMachineType_LastTransition_Id = 3836
    ProgramStateMachineType_LastTransition_Name = 3837
    ProgramStateMachineType_LastTransition_Number = 3838
    ProgramStateMachineType_LastTransition_TransitionTime = 3839
    ProgramStateMachineType_ProgramDiagnostics_CreateSessionId = 3840
    ProgramStateMachineType_ProgramDiagnostics_CreateClientName = 3841
    ProgramStateMachineType_ProgramDiagnostics_InvocationCreationTime = 3842
    ProgramStateMachineType_ProgramDiagnostics_LastTransitionTime = 3843
    ProgramStateMachineType_ProgramDiagnostics_LastMethodCall = 3844
    ProgramStateMachineType_ProgramDiagnostics_LastMethodSessionId = 3845
    ProgramStateMachineType_ProgramDiagnostics_LastMethodInputArguments = 3846
    ProgramStateMachineType_ProgramDiagnostics_LastMethodOutputArguments = 3847
    ProgramStateMachineType_ProgramDiagnostics_LastMethodCallTime = 3848
    ProgramStateMachineType_ProgramDiagnostics_LastMethodReturnStatus = 3849
    ProgramStateMachineType_FinalResultData = 3850
    AddCommentMethodType = 3863
    AddCommentMethodType_InputArguments = 3864
    ConditionType_EventId = 3865
    ConditionType_EventType = 3866
    ConditionType_SourceNode = 3867
    ConditionType_SourceName = 3868
    ConditionType_Time = 3869
    ConditionType_ReceiveTime = 3870
    ConditionType_LocalTime = 3871
    ConditionType_Message = 3872
    ConditionType_Severity = 3873
    ConditionType_Retain = 3874
    ConditionType_ConditionRefresh = 3875
    ConditionType_ConditionRefresh_InputArguments = 3876
    RefreshStartEventType_EventId = 3969
    RefreshStartEventType_EventType = 3970
    RefreshStartEventType_SourceNode = 3971
    RefreshStartEventType_SourceName = 3972
    RefreshStartEventType_Time = 3973
    RefreshStartEventType_ReceiveTime = 3974
    RefreshStartEventType_LocalTime = 3975
    RefreshStartEventType_Message = 3976
    RefreshStartEventType_Severity = 3977
    RefreshEndEventType_EventId = 3978
    RefreshEndEventType_EventType = 3979
    RefreshEndEventType_SourceNode = 3980
    RefreshEndEventType_SourceName = 3981
    RefreshEndEventType_Time = 3982
    RefreshEndEventType_ReceiveTime = 3983
    RefreshEndEventType_LocalTime = 3984
    RefreshEndEventType_Message = 3985
    RefreshEndEventType_Severity = 3986
    RefreshRequiredEventType_EventId = 3987
    RefreshRequiredEventType_EventType = 3988
    RefreshRequiredEventType_SourceNode = 3989
    RefreshRequiredEventType_SourceName = 3990
    RefreshRequiredEventType_Time = 3991
    RefreshRequiredEventType_ReceiveTime = 3992
    RefreshRequiredEventType_LocalTime = 3993
    RefreshRequiredEventType_Message = 3994
    RefreshRequiredEventType_Severity = 3995
    AuditConditionEventType_EventId = 3996
    AuditConditionEventType_EventType = 3997
    AuditConditionEventType_SourceNode = 3998
    AuditConditionEventType_SourceName = 3999
    AuditConditionEventType_Time = 4000
    AuditConditionEventType_ReceiveTime = 4001
    AuditConditionEventType_LocalTime = 4002
    AuditConditionEventType_Message = 4003
    AuditConditionEventType_Severity = 4004
    AuditConditionEventType_ActionTimeStamp = 4005
    AuditConditionEventType_Status = 4006
    AuditConditionEventType_ServerId = 4007
    AuditConditionEventType_ClientAuditEntryId = 4008
    AuditConditionEventType_ClientUserId = 4009
    AuditConditionEventType_MethodId = 4010
    AuditConditionEventType_InputArguments = 4011
    AuditConditionEnableEventType_EventId = 4106
    AuditConditionEnableEventType_EventType = 4107
    AuditConditionEnableEventType_SourceNode = 4108
    AuditConditionEnableEventType_SourceName = 4109
    AuditConditionEnableEventType_Time = 4110
    AuditConditionEnableEventType_ReceiveTime = 4111
    AuditConditionEnableEventType_LocalTime = 4112
    AuditConditionEnableEventType_Message = 4113
    AuditConditionEnableEventType_Severity = 4114
    AuditConditionEnableEventType_ActionTimeStamp = 4115
    AuditConditionEnableEventType_Status = 4116
    AuditConditionEnableEventType_ServerId = 4117
    AuditConditionEnableEventType_ClientAuditEntryId = 4118
    AuditConditionEnableEventType_ClientUserId = 4119
    AuditConditionEnableEventType_MethodId = 4120
    AuditConditionEnableEventType_InputArguments = 4121
    AuditConditionCommentEventType_EventId = 4170
    AuditConditionCommentEventType_EventType = 4171
    AuditConditionCommentEventType_SourceNode = 4172
    AuditConditionCommentEventType_SourceName = 4173
    AuditConditionCommentEventType_Time = 4174
    AuditConditionCommentEventType_ReceiveTime = 4175
    AuditConditionCommentEventType_LocalTime = 4176
    AuditConditionCommentEventType_Message = 4177
    AuditConditionCommentEventType_Severity = 4178
    AuditConditionCommentEventType_ActionTimeStamp = 4179
    AuditConditionCommentEventType_Status = 4180
    AuditConditionCommentEventType_ServerId = 4181
    AuditConditionCommentEventType_ClientAuditEntryId = 4182
    AuditConditionCommentEventType_ClientUserId = 4183
    AuditConditionCommentEventType_MethodId = 4184
    AuditConditionCommentEventType_InputArguments = 4185
    DialogConditionType_EventId = 4188
    DialogConditionType_EventType = 4189
    DialogConditionType_SourceNode = 4190
    DialogConditionType_SourceName = 4191
    DialogConditionType_Time = 4192
    DialogConditionType_ReceiveTime = 4193
    DialogConditionType_LocalTime = 4194
    DialogConditionType_Message = 4195
    DialogConditionType_Severity = 4196
    DialogConditionType_Retain = 4197
    DialogConditionType_ConditionRefresh = 4198
    DialogConditionType_ConditionRefresh_InputArguments = 4199
    AcknowledgeableConditionType_EventId = 5113
    AcknowledgeableConditionType_EventType = 5114
    AcknowledgeableConditionType_SourceNode = 5115
    AcknowledgeableConditionType_SourceName = 5116
    AcknowledgeableConditionType_Time = 5117
    AcknowledgeableConditionType_ReceiveTime = 5118
    AcknowledgeableConditionType_LocalTime = 5119
    AcknowledgeableConditionType_Message = 5120
    AcknowledgeableConditionType_Severity = 5121
    AcknowledgeableConditionType_Retain = 5122
    AcknowledgeableConditionType_ConditionRefresh = 5123
    AcknowledgeableConditionType_ConditionRefresh_InputArguments = 5124
    AlarmConditionType_EventId = 5540
    AlarmConditionType_EventType = 5541
    AlarmConditionType_SourceNode = 5542
    AlarmConditionType_SourceName = 5543
    AlarmConditionType_Time = 5544
    AlarmConditionType_ReceiveTime = 5545
    AlarmConditionType_LocalTime = 5546
    AlarmConditionType_Message = 5547
    AlarmConditionType_Severity = 5548
    AlarmConditionType_Retain = 5549
    AlarmConditionType_ConditionRefresh = 5550
    AlarmConditionType_ConditionRefresh_InputArguments = 5551
    ShelvedStateMachineType_CurrentState = 6088
    ShelvedStateMachineType_CurrentState_Id = 6089
    ShelvedStateMachineType_CurrentState_Name = 6090
    ShelvedStateMachineType_CurrentState_Number = 6091
    ShelvedStateMachineType_CurrentState_EffectiveDisplayName = 6092
    ShelvedStateMachineType_LastTransition = 6093
    ShelvedStateMachineType_LastTransition_Id = 6094
    ShelvedStateMachineType_LastTransition_Name = 6095
    ShelvedStateMachineType_LastTransition_Number = 6096
    ShelvedStateMachineType_LastTransition_TransitionTime = 6097
    ShelvedStateMachineType_Unshelved_StateNumber = 6098
    ShelvedStateMachineType_TimedShelved_StateNumber = 6100
    ShelvedStateMachineType_OneShotShelved_StateNumber = 6101
    TimedShelveMethodType = 6102
    TimedShelveMethodType_InputArguments = 6103
    LimitAlarmType_EventId = 6116
    LimitAlarmType_EventType = 6117
    LimitAlarmType_SourceNode = 6118
    LimitAlarmType_SourceName = 6119
    LimitAlarmType_Time = 6120
    LimitAlarmType_ReceiveTime = 6121
    LimitAlarmType_LocalTime = 6122
    LimitAlarmType_Message = 6123
    LimitAlarmType_Severity = 6124
    LimitAlarmType_Retain = 6125
    LimitAlarmType_ConditionRefresh = 6126
    LimitAlarmType_ConditionRefresh_InputArguments = 6127
    IdType_EnumStrings = 7591
    EnumValueType = 7594
    MessageSecurityMode_EnumStrings = 7595
    UserTokenType_EnumStrings = 7596
    ApplicationType_EnumStrings = 7597
    SecurityTokenRequestType_EnumStrings = 7598
    ComplianceLevel_EnumStrings = 7599
    BrowseDirection_EnumStrings = 7603
    FilterOperator_EnumStrings = 7605
    TimestampsToReturn_EnumStrings = 7606
    MonitoringMode_EnumStrings = 7608
    DataChangeTrigger_EnumStrings = 7609
    DeadbandType_EnumStrings = 7610
    RedundancySupport_EnumStrings = 7611
    ServerState_EnumStrings = 7612
    ExceptionDeviationFormat_EnumStrings = 7614
    EnumValueType_Encoding_DefaultXml = 7616
    OpcUa_BinarySchema = 7617
    OpcUa_BinarySchema_DataTypeVersion = 7618
    OpcUa_BinarySchema_NamespaceUri = 7619
    OpcUa_BinarySchema_Argument = 7650
    OpcUa_BinarySchema_Argument_DataTypeVersion = 7651
    OpcUa_BinarySchema_Argument_DictionaryFragment = 7652
    OpcUa_BinarySchema_EnumValueType = 7656
    OpcUa_BinarySchema_EnumValueType_DataTypeVersion = 7657
    OpcUa_BinarySchema_EnumValueType_DictionaryFragment = 7658
    OpcUa_BinarySchema_StatusResult = 7659
    OpcUa_BinarySchema_StatusResult_DataTypeVersion = 7660
    OpcUa_BinarySchema_StatusResult_DictionaryFragment = 7661
    OpcUa_BinarySchema_UserTokenPolicy = 7662
    OpcUa_BinarySchema_UserTokenPolicy_DataTypeVersion = 7663
    OpcUa_BinarySchema_UserTokenPolicy_DictionaryFragment = 7664
    OpcUa_BinarySchema_ApplicationDescription = 7665
    OpcUa_BinarySchema_ApplicationDescription_DataTypeVersion = 7666
    OpcUa_BinarySchema_ApplicationDescription_DictionaryFragment = 7667
    OpcUa_BinarySchema_EndpointDescription = 7668
    OpcUa_BinarySchema_EndpointDescription_DataTypeVersion = 7669
    OpcUa_BinarySchema_EndpointDescription_DictionaryFragment = 7670
    OpcUa_BinarySchema_UserIdentityToken = 7671
    OpcUa_BinarySchema_UserIdentityToken_DataTypeVersion = 7672
    OpcUa_BinarySchema_UserIdentityToken_DictionaryFragment = 7673
    OpcUa_BinarySchema_AnonymousIdentityToken = 7674
    OpcUa_BinarySchema_AnonymousIdentityToken_DataTypeVersion = 7675
    OpcUa_BinarySchema_AnonymousIdentityToken_DictionaryFragment = 7676
    OpcUa_BinarySchema_UserNameIdentityToken = 7677
    OpcUa_BinarySchema_UserNameIdentityToken_DataTypeVersion = 7678
    OpcUa_BinarySchema_UserNameIdentityToken_DictionaryFragment = 7679
    OpcUa_BinarySchema_X509IdentityToken = 7680
    OpcUa_BinarySchema_X509IdentityToken_DataTypeVersion = 7681
    OpcUa_BinarySchema_X509IdentityToken_DictionaryFragment = 7682
    OpcUa_BinarySchema_IssuedIdentityToken = 7683
    OpcUa_BinarySchema_IssuedIdentityToken_DataTypeVersion = 7684
    OpcUa_BinarySchema_IssuedIdentityToken_DictionaryFragment = 7685
    OpcUa_BinarySchema_EndpointConfiguration = 7686
    OpcUa_BinarySchema_EndpointConfiguration_DataTypeVersion = 7687
    OpcUa_BinarySchema_EndpointConfiguration_DictionaryFragment = 7688
    OpcUa_BinarySchema_SupportedProfile = 7689
    OpcUa_BinarySchema_SupportedProfile_DataTypeVersion = 7690
    OpcUa_BinarySchema_SupportedProfile_DictionaryFragment = 7691
    OpcUa_BinarySchema_BuildInfo = 7692
    OpcUa_BinarySchema_BuildInfo_DataTypeVersion = 7693
    OpcUa_BinarySchema_BuildInfo_DictionaryFragment = 7694
    OpcUa_BinarySchema_SoftwareCertificate = 7695
    OpcUa_BinarySchema_SoftwareCertificate_DataTypeVersion = 7696
    OpcUa_BinarySchema_SoftwareCertificate_DictionaryFragment = 7697
    OpcUa_BinarySchema_SignedSoftwareCertificate = 7698
    OpcUa_BinarySchema_SignedSoftwareCertificate_DataTypeVersion = 7699
    OpcUa_BinarySchema_SignedSoftwareCertificate_DictionaryFragment = 7700
    OpcUa_BinarySchema_AddNodesItem = 7728
    OpcUa_BinarySchema_AddNodesItem_DataTypeVersion = 7729
    OpcUa_BinarySchema_AddNodesItem_DictionaryFragment = 7730
    OpcUa_BinarySchema_AddReferencesItem = 7731
    OpcUa_BinarySchema_AddReferencesItem_DataTypeVersion = 7732
    OpcUa_BinarySchema_AddReferencesItem_DictionaryFragment = 7733
    OpcUa_BinarySchema_DeleteNodesItem = 7734
    OpcUa_BinarySchema_DeleteNodesItem_DataTypeVersion = 7735
    OpcUa_BinarySchema_DeleteNodesItem_DictionaryFragment = 7736
    OpcUa_BinarySchema_DeleteReferencesItem = 7737
    OpcUa_BinarySchema_DeleteReferencesItem_DataTypeVersion = 7738
    OpcUa_BinarySchema_DeleteReferencesItem_DictionaryFragment = 7739
    OpcUa_BinarySchema_RegisteredServer = 7782
    OpcUa_BinarySchema_RegisteredServer_DataTypeVersion = 7783
    OpcUa_BinarySchema_RegisteredServer_DictionaryFragment = 7784
    OpcUa_BinarySchema_ContentFilterElement = 7929
    OpcUa_BinarySchema_ContentFilterElement_DataTypeVersion = 7930
    OpcUa_BinarySchema_ContentFilterElement_DictionaryFragment = 7931
    OpcUa_BinarySchema_ContentFilter = 7932
    OpcUa_BinarySchema_ContentFilter_DataTypeVersion = 7933
    OpcUa_BinarySchema_ContentFilter_DictionaryFragment = 7934
    OpcUa_BinarySchema_FilterOperand = 7935
    OpcUa_BinarySchema_FilterOperand_DataTypeVersion = 7936
    OpcUa_BinarySchema_FilterOperand_DictionaryFragment = 7937
    OpcUa_BinarySchema_ElementOperand = 7938
    OpcUa_BinarySchema_ElementOperand_DataTypeVersion = 7939
    OpcUa_BinarySchema_ElementOperand_DictionaryFragment = 7940
    OpcUa_BinarySchema_LiteralOperand = 7941
    OpcUa_BinarySchema_LiteralOperand_DataTypeVersion = 7942
    OpcUa_BinarySchema_LiteralOperand_DictionaryFragment = 7943
    OpcUa_BinarySchema_AttributeOperand = 7944
    OpcUa_BinarySchema_AttributeOperand_DataTypeVersion = 7945
    OpcUa_BinarySchema_AttributeOperand_DictionaryFragment = 7946
    OpcUa_BinarySchema_SimpleAttributeOperand = 7947
    OpcUa_BinarySchema_SimpleAttributeOperand_DataTypeVersion = 7948
    OpcUa_BinarySchema_SimpleAttributeOperand_DictionaryFragment = 7949
    OpcUa_BinarySchema_HistoryEvent = 8004
    OpcUa_BinarySchema_HistoryEvent_DataTypeVersion = 8005
    OpcUa_BinarySchema_HistoryEvent_DictionaryFragment = 8006
    OpcUa_BinarySchema_MonitoringFilter = 8067
    OpcUa_BinarySchema_MonitoringFilter_DataTypeVersion = 8068
    OpcUa_BinarySchema_MonitoringFilter_DictionaryFragment = 8069
    OpcUa_BinarySchema_EventFilter = 8073
    OpcUa_BinarySchema_EventFilter_DataTypeVersion = 8074
    OpcUa_BinarySchema_EventFilter_DictionaryFragment = 8075
    OpcUa_BinarySchema_AggregateConfiguration = 8076
    OpcUa_BinarySchema_AggregateConfiguration_DataTypeVersion = 8077
    OpcUa_BinarySchema_AggregateConfiguration_DictionaryFragment = 8078
    OpcUa_BinarySchema_HistoryEventFieldList = 8172
    OpcUa_BinarySchema_HistoryEventFieldList_DataTypeVersion = 8173
    OpcUa_BinarySchema_HistoryEventFieldList_DictionaryFragment = 8174
    OpcUa_BinarySchema_RedundantServerDataType = 8208
    OpcUa_BinarySchema_RedundantServerDataType_DataTypeVersion = 8209
    OpcUa_BinarySchema_RedundantServerDataType_DictionaryFragment = 8210
    OpcUa_BinarySchema_SamplingIntervalDiagnosticsDataType = 8211
    OpcUa_BinarySchema_SamplingIntervalDiagnosticsDataType_DataTypeVersion = 8212
    OpcUa_BinarySchema_SamplingIntervalDiagnosticsDataType_DictionaryFragment = 8213
    OpcUa_BinarySchema_ServerDiagnosticsSummaryDataType = 8214
    OpcUa_BinarySchema_ServerDiagnosticsSummaryDataType_DataTypeVersion = 8215
    OpcUa_BinarySchema_ServerDiagnosticsSummaryDataType_DictionaryFragment = 8216
    OpcUa_BinarySchema_ServerStatusDataType = 8217
    OpcUa_BinarySchema_ServerStatusDataType_DataTypeVersion = 8218
    OpcUa_BinarySchema_ServerStatusDataType_DictionaryFragment = 8219
    OpcUa_BinarySchema_SessionDiagnosticsDataType = 8220
    OpcUa_BinarySchema_SessionDiagnosticsDataType_DataTypeVersion = 8221
    OpcUa_BinarySchema_SessionDiagnosticsDataType_DictionaryFragment = 8222
    OpcUa_BinarySchema_SessionSecurityDiagnosticsDataType = 8223
    OpcUa_BinarySchema_SessionSecurityDiagnosticsDataType_DataTypeVersion = 8224
    OpcUa_BinarySchema_SessionSecurityDiagnosticsDataType_DictionaryFragment = 8225
    OpcUa_BinarySchema_ServiceCounterDataType = 8226
    OpcUa_BinarySchema_ServiceCounterDataType_DataTypeVersion = 8227
    OpcUa_BinarySchema_ServiceCounterDataType_DictionaryFragment = 8228
    OpcUa_BinarySchema_SubscriptionDiagnosticsDataType = 8229
    OpcUa_BinarySchema_SubscriptionDiagnosticsDataType_DataTypeVersion = 8230
    OpcUa_BinarySchema_SubscriptionDiagnosticsDataType_DictionaryFragment = 8231
    OpcUa_BinarySchema_ModelChangeStructureDataType = 8232
    OpcUa_BinarySchema_ModelChangeStructureDataType_DataTypeVersion = 8233
    OpcUa_BinarySchema_ModelChangeStructureDataType_DictionaryFragment = 8234
    OpcUa_BinarySchema_SemanticChangeStructureDataType = 8235
    OpcUa_BinarySchema_SemanticChangeStructureDataType_DataTypeVersion = 8236
    OpcUa_BinarySchema_SemanticChangeStructureDataType_DictionaryFragment = 8237
    OpcUa_BinarySchema_Range = 8238
    OpcUa_BinarySchema_Range_DataTypeVersion = 8239
    OpcUa_BinarySchema_Range_DictionaryFragment = 8240
    OpcUa_BinarySchema_EUInformation = 8241
    OpcUa_BinarySchema_EUInformation_DataTypeVersion = 8242
    OpcUa_BinarySchema_EUInformation_DictionaryFragment = 8243
    OpcUa_BinarySchema_Annotation = 8244
    OpcUa_BinarySchema_Annotation_DataTypeVersion = 8245
    OpcUa_BinarySchema_Annotation_DictionaryFragment = 8246
    OpcUa_BinarySchema_ProgramDiagnosticDataType = 8247
    OpcUa_BinarySchema_ProgramDiagnosticDataType_DataTypeVersion = 8248
    OpcUa_BinarySchema_ProgramDiagnosticDataType_DictionaryFragment = 8249
    EnumValueType_Encoding_DefaultBinary = 8251
    OpcUa_XmlSchema = 8252
    OpcUa_XmlSchema_DataTypeVersion = 8253
    OpcUa_XmlSchema_NamespaceUri = 8254
    OpcUa_XmlSchema_Argument = 8285
    OpcUa_XmlSchema_Argument_DataTypeVersion = 8286
    OpcUa_XmlSchema_Argument_DictionaryFragment = 8287
    OpcUa_XmlSchema_EnumValueType = 8291
    OpcUa_XmlSchema_EnumValueType_DataTypeVersion = 8292
    OpcUa_XmlSchema_EnumValueType_DictionaryFragment = 8293
    OpcUa_XmlSchema_StatusResult = 8294
    OpcUa_XmlSchema_StatusResult_DataTypeVersion = 8295
    OpcUa_XmlSchema_StatusResult_DictionaryFragment = 8296
    OpcUa_XmlSchema_UserTokenPolicy = 8297
    OpcUa_XmlSchema_UserTokenPolicy_DataTypeVersion = 8298
    OpcUa_XmlSchema_UserTokenPolicy_DictionaryFragment = 8299
    OpcUa_XmlSchema_ApplicationDescription = 8300
    OpcUa_XmlSchema_ApplicationDescription_DataTypeVersion = 8301
    OpcUa_XmlSchema_ApplicationDescription_DictionaryFragment = 8302
    OpcUa_XmlSchema_EndpointDescription = 8303
    OpcUa_XmlSchema_EndpointDescription_DataTypeVersion = 8304
    OpcUa_XmlSchema_EndpointDescription_DictionaryFragment = 8305
    OpcUa_XmlSchema_UserIdentityToken = 8306
    OpcUa_XmlSchema_UserIdentityToken_DataTypeVersion = 8307
    OpcUa_XmlSchema_UserIdentityToken_DictionaryFragment = 8308
    OpcUa_XmlSchema_AnonymousIdentityToken = 8309
    OpcUa_XmlSchema_AnonymousIdentityToken_DataTypeVersion = 8310
    OpcUa_XmlSchema_AnonymousIdentityToken_DictionaryFragment = 8311
    OpcUa_XmlSchema_UserNameIdentityToken = 8312
    OpcUa_XmlSchema_UserNameIdentityToken_DataTypeVersion = 8313
    OpcUa_XmlSchema_UserNameIdentityToken_DictionaryFragment = 8314
    OpcUa_XmlSchema_X509IdentityToken = 8315
    OpcUa_XmlSchema_X509IdentityToken_DataTypeVersion = 8316
    OpcUa_XmlSchema_X509IdentityToken_DictionaryFragment = 8317
    OpcUa_XmlSchema_IssuedIdentityToken = 8318
    OpcUa_XmlSchema_IssuedIdentityToken_DataTypeVersion = 8319
    OpcUa_XmlSchema_IssuedIdentityToken_DictionaryFragment = 8320
    OpcUa_XmlSchema_EndpointConfiguration = 8321
    OpcUa_XmlSchema_EndpointConfiguration_DataTypeVersion = 8322
    OpcUa_XmlSchema_EndpointConfiguration_DictionaryFragment = 8323
    OpcUa_XmlSchema_SupportedProfile = 8324
    OpcUa_XmlSchema_SupportedProfile_DataTypeVersion = 8325
    OpcUa_XmlSchema_SupportedProfile_DictionaryFragment = 8326
    OpcUa_XmlSchema_BuildInfo = 8327
    OpcUa_XmlSchema_BuildInfo_DataTypeVersion = 8328
    OpcUa_XmlSchema_BuildInfo_DictionaryFragment = 8329
    OpcUa_XmlSchema_SoftwareCertificate = 8330
    OpcUa_XmlSchema_SoftwareCertificate_DataTypeVersion = 8331
    OpcUa_XmlSchema_SoftwareCertificate_DictionaryFragment = 8332
    OpcUa_XmlSchema_SignedSoftwareCertificate = 8333
    OpcUa_XmlSchema_SignedSoftwareCertificate_DataTypeVersion = 8334
    OpcUa_XmlSchema_SignedSoftwareCertificate_DictionaryFragment = 8335
    OpcUa_XmlSchema_AddNodesItem = 8363
    OpcUa_XmlSchema_AddNodesItem_DataTypeVersion = 8364
    OpcUa_XmlSchema_AddNodesItem_DictionaryFragment = 8365
    OpcUa_XmlSchema_AddReferencesItem = 8366
    OpcUa_XmlSchema_AddReferencesItem_DataTypeVersion = 8367
    OpcUa_XmlSchema_AddReferencesItem_DictionaryFragment = 8368
    OpcUa_XmlSchema_DeleteNodesItem = 8369
    OpcUa_XmlSchema_DeleteNodesItem_DataTypeVersion = 8370
    OpcUa_XmlSchema_DeleteNodesItem_DictionaryFragment = 8371
    OpcUa_XmlSchema_DeleteReferencesItem = 8372
    OpcUa_XmlSchema_DeleteReferencesItem_DataTypeVersion = 8373
    OpcUa_XmlSchema_DeleteReferencesItem_DictionaryFragment = 8374
    OpcUa_XmlSchema_RegisteredServer = 8417
    OpcUa_XmlSchema_RegisteredServer_DataTypeVersion = 8418
    OpcUa_XmlSchema_RegisteredServer_DictionaryFragment = 8419
    OpcUa_XmlSchema_ContentFilterElement = 8564
    OpcUa_XmlSchema_ContentFilterElement_DataTypeVersion = 8565
    OpcUa_XmlSchema_ContentFilterElement_DictionaryFragment = 8566
    OpcUa_XmlSchema_ContentFilter = 8567
    OpcUa_XmlSchema_ContentFilter_DataTypeVersion = 8568
    OpcUa_XmlSchema_ContentFilter_DictionaryFragment = 8569
    OpcUa_XmlSchema_FilterOperand = 8570
    OpcUa_XmlSchema_FilterOperand_DataTypeVersion = 8571
    OpcUa_XmlSchema_FilterOperand_DictionaryFragment = 8572
    OpcUa_XmlSchema_ElementOperand = 8573
    OpcUa_XmlSchema_ElementOperand_DataTypeVersion = 8574
    OpcUa_XmlSchema_ElementOperand_DictionaryFragment = 8575
    OpcUa_XmlSchema_LiteralOperand = 8576
    OpcUa_XmlSchema_LiteralOperand_DataTypeVersion = 8577
    OpcUa_XmlSchema_LiteralOperand_DictionaryFragment = 8578
    OpcUa_XmlSchema_AttributeOperand = 8579
    OpcUa_XmlSchema_AttributeOperand_DataTypeVersion = 8580
    OpcUa_XmlSchema_AttributeOperand_DictionaryFragment = 8581
    OpcUa_XmlSchema_SimpleAttributeOperand = 8582
    OpcUa_XmlSchema_SimpleAttributeOperand_DataTypeVersion = 8583
    OpcUa_XmlSchema_SimpleAttributeOperand_DictionaryFragment = 8584
    OpcUa_XmlSchema_HistoryEvent = 8639
    OpcUa_XmlSchema_HistoryEvent_DataTypeVersion = 8640
    OpcUa_XmlSchema_HistoryEvent_DictionaryFragment = 8641
    OpcUa_XmlSchema_MonitoringFilter = 8702
    OpcUa_XmlSchema_MonitoringFilter_DataTypeVersion = 8703
    OpcUa_XmlSchema_MonitoringFilter_DictionaryFragment = 8704
    OpcUa_XmlSchema_EventFilter = 8708
    OpcUa_XmlSchema_EventFilter_DataTypeVersion = 8709
    OpcUa_XmlSchema_EventFilter_DictionaryFragment = 8710
    OpcUa_XmlSchema_AggregateConfiguration = 8711
    OpcUa_XmlSchema_AggregateConfiguration_DataTypeVersion = 8712
    OpcUa_XmlSchema_AggregateConfiguration_DictionaryFragment = 8713
    OpcUa_XmlSchema_HistoryEventFieldList = 8807
    OpcUa_XmlSchema_HistoryEventFieldList_DataTypeVersion = 8808
    OpcUa_XmlSchema_HistoryEventFieldList_DictionaryFragment = 8809
    OpcUa_XmlSchema_RedundantServerDataType = 8843
    OpcUa_XmlSchema_RedundantServerDataType_DataTypeVersion = 8844
    OpcUa_XmlSchema_RedundantServerDataType_DictionaryFragment = 8845
    OpcUa_XmlSchema_SamplingIntervalDiagnosticsDataType = 8846
    OpcUa_XmlSchema_SamplingIntervalDiagnosticsDataType_DataTypeVersion = 8847
    OpcUa_XmlSchema_SamplingIntervalDiagnosticsDataType_DictionaryFragment = 8848
    OpcUa_XmlSchema_ServerDiagnosticsSummaryDataType = 8849
    OpcUa_XmlSchema_ServerDiagnosticsSummaryDataType_DataTypeVersion = 8850
    OpcUa_XmlSchema_ServerDiagnosticsSummaryDataType_DictionaryFragment = 8851
    OpcUa_XmlSchema_ServerStatusDataType = 8852
    OpcUa_XmlSchema_ServerStatusDataType_DataTypeVersion = 8853
    OpcUa_XmlSchema_ServerStatusDataType_DictionaryFragment = 8854
    OpcUa_XmlSchema_SessionDiagnosticsDataType = 8855
    OpcUa_XmlSchema_SessionDiagnosticsDataType_DataTypeVersion = 8856
    OpcUa_XmlSchema_SessionDiagnosticsDataType_DictionaryFragment = 8857
    OpcUa_XmlSchema_SessionSecurityDiagnosticsDataType = 8858
    OpcUa_XmlSchema_SessionSecurityDiagnosticsDataType_DataTypeVersion = 8859
    OpcUa_XmlSchema_SessionSecurityDiagnosticsDataType_DictionaryFragment = 8860
    OpcUa_XmlSchema_ServiceCounterDataType = 8861
    OpcUa_XmlSchema_ServiceCounterDataType_DataTypeVersion = 8862
    OpcUa_XmlSchema_ServiceCounterDataType_DictionaryFragment = 8863
    OpcUa_XmlSchema_SubscriptionDiagnosticsDataType = 8864
    OpcUa_XmlSchema_SubscriptionDiagnosticsDataType_DataTypeVersion = 8865
    OpcUa_XmlSchema_SubscriptionDiagnosticsDataType_DictionaryFragment = 8866
    OpcUa_XmlSchema_ModelChangeStructureDataType = 8867
    OpcUa_XmlSchema_ModelChangeStructureDataType_DataTypeVersion = 8868
    OpcUa_XmlSchema_ModelChangeStructureDataType_DictionaryFragment = 8869
    OpcUa_XmlSchema_SemanticChangeStructureDataType = 8870
    OpcUa_XmlSchema_SemanticChangeStructureDataType_DataTypeVersion = 8871
    OpcUa_XmlSchema_SemanticChangeStructureDataType_DictionaryFragment = 8872
    OpcUa_XmlSchema_Range = 8873
    OpcUa_XmlSchema_Range_DataTypeVersion = 8874
    OpcUa_XmlSchema_Range_DictionaryFragment = 8875
    OpcUa_XmlSchema_EUInformation = 8876
    OpcUa_XmlSchema_EUInformation_DataTypeVersion = 8877
    OpcUa_XmlSchema_EUInformation_DictionaryFragment = 8878
    OpcUa_XmlSchema_Annotation = 8879
    OpcUa_XmlSchema_Annotation_DataTypeVersion = 8880
    OpcUa_XmlSchema_Annotation_DictionaryFragment = 8881
    OpcUa_XmlSchema_ProgramDiagnosticDataType = 8882
    OpcUa_XmlSchema_ProgramDiagnosticDataType_DataTypeVersion = 8883
    OpcUa_XmlSchema_ProgramDiagnosticDataType_DictionaryFragment = 8884
    SubscriptionDiagnosticsType_MaxLifetimeCount = 8888
    SubscriptionDiagnosticsType_LatePublishRequestCount = 8889
    SubscriptionDiagnosticsType_CurrentKeepAliveCount = 8890
    SubscriptionDiagnosticsType_CurrentLifetimeCount = 8891
    SubscriptionDiagnosticsType_UnacknowledgedMessageCount = 8892
    SubscriptionDiagnosticsType_DiscardedMessageCount = 8893
    SubscriptionDiagnosticsType_MonitoredItemCount = 8894
    SubscriptionDiagnosticsType_DisabledMonitoredItemCount = 8895
    SubscriptionDiagnosticsType_MonitoringQueueOverflowCount = 8896
    SubscriptionDiagnosticsType_NextSequenceNumber = 8897
    SessionDiagnosticsObjectType_SessionDiagnostics_TotalRequestCount = 8898
    SessionDiagnosticsVariableType_TotalRequestCount = 8900
    SubscriptionDiagnosticsType_EventQueueOverFlowCount = 8902
    TimeZoneDataType = 8912
    TimeZoneDataType_Encoding_DefaultXml = 8913
    OpcUa_BinarySchema_TimeZoneDataType = 8914
    OpcUa_BinarySchema_TimeZoneDataType_DataTypeVersion = 8915
    OpcUa_BinarySchema_TimeZoneDataType_DictionaryFragment = 8916
    TimeZoneDataType_Encoding_DefaultBinary = 8917
    OpcUa_XmlSchema_TimeZoneDataType = 8918
    OpcUa_XmlSchema_TimeZoneDataType_DataTypeVersion = 8919
    OpcUa_XmlSchema_TimeZoneDataType_DictionaryFragment = 8920
    AuditConditionRespondEventType = 8927
    AuditConditionRespondEventType_EventId = 8928
    AuditConditionRespondEventType_EventType = 8929
    AuditConditionRespondEventType_SourceNode = 8930
    AuditConditionRespondEventType_SourceName = 8931
    AuditConditionRespondEventType_Time = 8932
    AuditConditionRespondEventType_ReceiveTime = 8933
    AuditConditionRespondEventType_LocalTime = 8934
    AuditConditionRespondEventType_Message = 8935
    AuditConditionRespondEventType_Severity = 8936
    AuditConditionRespondEventType_ActionTimeStamp = 8937
    AuditConditionRespondEventType_Status = 8938
    AuditConditionRespondEventType_ServerId = 8939
    AuditConditionRespondEventType_ClientAuditEntryId = 8940
    AuditConditionRespondEventType_ClientUserId = 8941
    AuditConditionRespondEventType_MethodId = 8942
    AuditConditionRespondEventType_InputArguments = 8943
    AuditConditionAcknowledgeEventType = 8944
    AuditConditionAcknowledgeEventType_EventId = 8945
    AuditConditionAcknowledgeEventType_EventType = 8946
    AuditConditionAcknowledgeEventType_SourceNode = 8947
    AuditConditionAcknowledgeEventType_SourceName = 8948
    AuditConditionAcknowledgeEventType_Time = 8949
    AuditConditionAcknowledgeEventType_ReceiveTime = 8950
    AuditConditionAcknowledgeEventType_LocalTime = 8951
    AuditConditionAcknowledgeEventType_Message = 8952
    AuditConditionAcknowledgeEventType_Severity = 8953
    AuditConditionAcknowledgeEventType_ActionTimeStamp = 8954
    AuditConditionAcknowledgeEventType_Status = 8955
    AuditConditionAcknowledgeEventType_ServerId = 8956
    AuditConditionAcknowledgeEventType_ClientAuditEntryId = 8957
    AuditConditionAcknowledgeEventType_ClientUserId = 8958
    AuditConditionAcknowledgeEventType_MethodId = 8959
    AuditConditionAcknowledgeEventType_InputArguments = 8960
    AuditConditionConfirmEventType = 8961
    AuditConditionConfirmEventType_EventId = 8962
    AuditConditionConfirmEventType_EventType = 8963
    AuditConditionConfirmEventType_SourceNode = 8964
    AuditConditionConfirmEventType_SourceName = 8965
    AuditConditionConfirmEventType_Time = 8966
    AuditConditionConfirmEventType_ReceiveTime = 8967
    AuditConditionConfirmEventType_LocalTime = 8968
    AuditConditionConfirmEventType_Message = 8969
    AuditConditionConfirmEventType_Severity = 8970
    AuditConditionConfirmEventType_ActionTimeStamp = 8971
    AuditConditionConfirmEventType_Status = 8972
    AuditConditionConfirmEventType_ServerId = 8973
    AuditConditionConfirmEventType_ClientAuditEntryId = 8974
    AuditConditionConfirmEventType_ClientUserId = 8975
    AuditConditionConfirmEventType_MethodId = 8976
    AuditConditionConfirmEventType_InputArguments = 8977
    TwoStateVariableType = 8995
    TwoStateVariableType_Id = 8996
    TwoStateVariableType_Name = 8997
    TwoStateVariableType_Number = 8998
    TwoStateVariableType_EffectiveDisplayName = 8999
    TwoStateVariableType_TransitionTime = 9000
    TwoStateVariableType_EffectiveTransitionTime = 9001
    ConditionVariableType = 9002
    ConditionVariableType_SourceTimestamp = 9003
    HasTrueSubState = 9004
    HasFalseSubState = 9005
    HasCondition = 9006
    ConditionRefreshMethodType = 9007
    ConditionRefreshMethodType_InputArguments = 9008
    ConditionType_ConditionName = 9009
    ConditionType_BranchId = 9010
    ConditionType_EnabledState = 9011
    ConditionType_EnabledState_Id = 9012
    ConditionType_EnabledState_Name = 9013
    ConditionType_EnabledState_Number = 9014
    ConditionType_EnabledState_EffectiveDisplayName = 9015
    ConditionType_EnabledState_TransitionTime = 9016
    ConditionType_EnabledState_EffectiveTransitionTime = 9017
    ConditionType_EnabledState_TrueState = 9018
    ConditionType_EnabledState_FalseState = 9019
    ConditionType_Quality = 9020
    ConditionType_Quality_SourceTimestamp = 9021
    ConditionType_LastSeverity = 9022
    ConditionType_LastSeverity_SourceTimestamp = 9023
    ConditionType_Comment = 9024
    ConditionType_Comment_SourceTimestamp = 9025
    ConditionType_ClientUserId = 9026
    ConditionType_Enable = 9027
    ConditionType_Disable = 9028
    ConditionType_AddComment = 9029
    ConditionType_AddComment_InputArguments = 9030
    DialogResponseMethodType = 9031
    DialogResponseMethodType_InputArguments = 9032
    DialogConditionType_ConditionName = 9033
    DialogConditionType_BranchId = 9034
    DialogConditionType_EnabledState = 9035
    DialogConditionType_EnabledState_Id = 9036
    DialogConditionType_EnabledState_Name = 9037
    DialogConditionType_EnabledState_Number = 9038
    DialogConditionType_EnabledState_EffectiveDisplayName = 9039
    DialogConditionType_EnabledState_TransitionTime = 9040
    DialogConditionType_EnabledState_EffectiveTransitionTime = 9041
    DialogConditionType_EnabledState_TrueState = 9042
    DialogConditionType_EnabledState_FalseState = 9043
    DialogConditionType_Quality = 9044
    DialogConditionType_Quality_SourceTimestamp = 9045
    DialogConditionType_LastSeverity = 9046
    DialogConditionType_LastSeverity_SourceTimestamp = 9047
    DialogConditionType_Comment = 9048
    DialogConditionType_Comment_SourceTimestamp = 9049
    DialogConditionType_ClientUserId = 9050
    DialogConditionType_Enable = 9051
    DialogConditionType_Disable = 9052
    DialogConditionType_AddComment = 9053
    DialogConditionType_AddComment_InputArguments = 9054
    DialogConditionType_DialogState = 9055
    DialogConditionType_DialogState_Id = 9056
    DialogConditionType_DialogState_Name = 9057
    DialogConditionType_DialogState_Number = 9058
    DialogConditionType_DialogState_EffectiveDisplayName = 9059
    DialogConditionType_DialogState_TransitionTime = 9060
    DialogConditionType_DialogState_EffectiveTransitionTime = 9061
    DialogConditionType_DialogState_TrueState = 9062
    DialogConditionType_DialogState_FalseState = 9063
    DialogConditionType_ResponseOptionSet = 9064
    DialogConditionType_DefaultResponse = 9065
    DialogConditionType_OkResponse = 9066
    DialogConditionType_CancelResponse = 9067
    DialogConditionType_LastResponse = 9068
    DialogConditionType_Respond = 9069
    DialogConditionType_Respond_InputArguments = 9070
    AcknowledgeableConditionType_ConditionName = 9071
    AcknowledgeableConditionType_BranchId = 9072
    AcknowledgeableConditionType_EnabledState = 9073
    AcknowledgeableConditionType_EnabledState_Id = 9074
    AcknowledgeableConditionType_EnabledState_Name = 9075
    AcknowledgeableConditionType_EnabledState_Number = 9076
    AcknowledgeableConditionType_EnabledState_EffectiveDisplayName = 9077
    AcknowledgeableConditionType_EnabledState_TransitionTime = 9078
    AcknowledgeableConditionType_EnabledState_EffectiveTransitionTime = 9079
    AcknowledgeableConditionType_EnabledState_TrueState = 9080
    AcknowledgeableConditionType_EnabledState_FalseState = 9081
    AcknowledgeableConditionType_Quality = 9082
    AcknowledgeableConditionType_Quality_SourceTimestamp = 9083
    AcknowledgeableConditionType_LastSeverity = 9084
    AcknowledgeableConditionType_LastSeverity_SourceTimestamp = 9085
    AcknowledgeableConditionType_Comment = 9086
    AcknowledgeableConditionType_Comment_SourceTimestamp = 9087
    AcknowledgeableConditionType_ClientUserId = 9088
    AcknowledgeableConditionType_Enable = 9089
    AcknowledgeableConditionType_Disable = 9090
    AcknowledgeableConditionType_AddComment = 9091
    AcknowledgeableConditionType_AddComment_InputArguments = 9092
    AcknowledgeableConditionType_AckedState = 9093
    AcknowledgeableConditionType_AckedState_Id = 9094
    AcknowledgeableConditionType_AckedState_Name = 9095
    AcknowledgeableConditionType_AckedState_Number = 9096
    AcknowledgeableConditionType_AckedState_EffectiveDisplayName = 9097
    AcknowledgeableConditionType_AckedState_TransitionTime = 9098
    AcknowledgeableConditionType_AckedState_EffectiveTransitionTime = 9099
    AcknowledgeableConditionType_AckedState_TrueState = 9100
    AcknowledgeableConditionType_AckedState_FalseState = 9101
    AcknowledgeableConditionType_ConfirmedState = 9102
    AcknowledgeableConditionType_ConfirmedState_Id = 9103
    AcknowledgeableConditionType_ConfirmedState_Name = 9104
    AcknowledgeableConditionType_ConfirmedState_Number = 9105
    AcknowledgeableConditionType_ConfirmedState_EffectiveDisplayName = 9106
    AcknowledgeableConditionType_ConfirmedState_TransitionTime = 9107
    AcknowledgeableConditionType_ConfirmedState_EffectiveTransitionTime = 9108
    AcknowledgeableConditionType_ConfirmedState_TrueState = 9109
    AcknowledgeableConditionType_ConfirmedState_FalseState = 9110
    AcknowledgeableConditionType_Acknowledge = 9111
    AcknowledgeableConditionType_Acknowledge_InputArguments = 9112
    AcknowledgeableConditionType_Confirm = 9113
    AcknowledgeableConditionType_Confirm_InputArguments = 9114
    ShelvedStateMachineType_UnshelveTime = 9115
    AlarmConditionType_ConditionName = 9116
    AlarmConditionType_BranchId = 9117
    AlarmConditionType_EnabledState = 9118
    AlarmConditionType_EnabledState_Id = 9119
    AlarmConditionType_EnabledState_Name = 9120
    AlarmConditionType_EnabledState_Number = 9121
    AlarmConditionType_EnabledState_EffectiveDisplayName = 9122
    AlarmConditionType_EnabledState_TransitionTime = 9123
    AlarmConditionType_EnabledState_EffectiveTransitionTime = 9124
    AlarmConditionType_EnabledState_TrueState = 9125
    AlarmConditionType_EnabledState_FalseState = 9126
    AlarmConditionType_Quality = 9127
    AlarmConditionType_Quality_SourceTimestamp = 9128
    AlarmConditionType_LastSeverity = 9129
    AlarmConditionType_LastSeverity_SourceTimestamp = 9130
    AlarmConditionType_Comment = 9131
    AlarmConditionType_Comment_SourceTimestamp = 9132
    AlarmConditionType_ClientUserId = 9133
    AlarmConditionType_Enable = 9134
    AlarmConditionType_Disable = 9135
    AlarmConditionType_AddComment = 9136
    AlarmConditionType_AddComment_InputArguments = 9137
    AlarmConditionType_AckedState = 9138
    AlarmConditionType_AckedState_Id = 9139
    AlarmConditionType_AckedState_Name = 9140
    AlarmConditionType_AckedState_Number = 9141
    AlarmConditionType_AckedState_EffectiveDisplayName = 9142
    AlarmConditionType_AckedState_TransitionTime = 9143
    AlarmConditionType_AckedState_EffectiveTransitionTime = 9144
    AlarmConditionType_AckedState_TrueState = 9145
    AlarmConditionType_AckedState_FalseState = 9146
    AlarmConditionType_ConfirmedState = 9147
    AlarmConditionType_ConfirmedState_Id = 9148
    AlarmConditionType_ConfirmedState_Name = 9149
    AlarmConditionType_ConfirmedState_Number = 9150
    AlarmConditionType_ConfirmedState_EffectiveDisplayName = 9151
    AlarmConditionType_ConfirmedState_TransitionTime = 9152
    AlarmConditionType_ConfirmedState_EffectiveTransitionTime = 9153
    AlarmConditionType_ConfirmedState_TrueState = 9154
    AlarmConditionType_ConfirmedState_FalseState = 9155
    AlarmConditionType_Acknowledge = 9156
    AlarmConditionType_Acknowledge_InputArguments = 9157
    AlarmConditionType_Confirm = 9158
    AlarmConditionType_Confirm_InputArguments = 9159
    AlarmConditionType_ActiveState = 9160
    AlarmConditionType_ActiveState_Id = 9161
    AlarmConditionType_ActiveState_Name = 9162
    AlarmConditionType_ActiveState_Number = 9163
    AlarmConditionType_ActiveState_EffectiveDisplayName = 9164
    AlarmConditionType_ActiveState_TransitionTime = 9165
    AlarmConditionType_ActiveState_EffectiveTransitionTime = 9166
    AlarmConditionType_ActiveState_TrueState = 9167
    AlarmConditionType_ActiveState_FalseState = 9168
    AlarmConditionType_SuppressedState = 9169
    AlarmConditionType_SuppressedState_Id = 9170
    AlarmConditionType_SuppressedState_Name = 9171
    AlarmConditionType_SuppressedState_Number = 9172
    AlarmConditionType_SuppressedState_EffectiveDisplayName = 9173
    AlarmConditionType_SuppressedState_TransitionTime = 9174
    AlarmConditionType_SuppressedState_EffectiveTransitionTime = 9175
    AlarmConditionType_SuppressedState_TrueState = 9176
    AlarmConditionType_SuppressedState_FalseState = 9177
    AlarmConditionType_ShelvingState = 9178
    AlarmConditionType_ShelvingState_CurrentState = 9179
    AlarmConditionType_ShelvingState_CurrentState_Id = 9180
    AlarmConditionType_ShelvingState_CurrentState_Name = 9181
    AlarmConditionType_ShelvingState_CurrentState_Number = 9182
    AlarmConditionType_ShelvingState_CurrentState_EffectiveDisplayName = 9183
    AlarmConditionType_ShelvingState_LastTransition = 9184
    AlarmConditionType_ShelvingState_LastTransition_Id = 9185
    AlarmConditionType_ShelvingState_LastTransition_Name = 9186
    AlarmConditionType_ShelvingState_LastTransition_Number = 9187
    AlarmConditionType_ShelvingState_LastTransition_TransitionTime = 9188
    AlarmConditionType_ShelvingState_UnshelveTime = 9189
    AlarmConditionType_ShelvingState_Unshelve = 9211
    AlarmConditionType_ShelvingState_OneShotShelve = 9212
    AlarmConditionType_ShelvingState_TimedShelve = 9213
    AlarmConditionType_ShelvingState_TimedShelve_InputArguments = 9214
    AlarmConditionType_SuppressedOrShelved = 9215
    AlarmConditionType_MaxTimeShelved = 9216
    LimitAlarmType_ConditionName = 9217
    LimitAlarmType_BranchId = 9218
    LimitAlarmType_EnabledState = 9219
    LimitAlarmType_EnabledState_Id = 9220
    LimitAlarmType_EnabledState_Name = 9221
    LimitAlarmType_EnabledState_Number = 9222
    LimitAlarmType_EnabledState_EffectiveDisplayName = 9223
    LimitAlarmType_EnabledState_TransitionTime = 9224
    LimitAlarmType_EnabledState_EffectiveTransitionTime = 9225
    LimitAlarmType_EnabledState_TrueState = 9226
    LimitAlarmType_EnabledState_FalseState = 9227
    LimitAlarmType_Quality = 9228
    LimitAlarmType_Quality_SourceTimestamp = 9229
    LimitAlarmType_LastSeverity = 9230
    LimitAlarmType_LastSeverity_SourceTimestamp = 9231
    LimitAlarmType_Comment = 9232
    LimitAlarmType_Comment_SourceTimestamp = 9233
    LimitAlarmType_ClientUserId = 9234
    LimitAlarmType_Enable = 9235
    LimitAlarmType_Disable = 9236
    LimitAlarmType_AddComment = 9237
    LimitAlarmType_AddComment_InputArguments = 9238
    LimitAlarmType_AckedState = 9239
    LimitAlarmType_AckedState_Id = 9240
    LimitAlarmType_AckedState_Name = 9241
    LimitAlarmType_AckedState_Number = 9242
    LimitAlarmType_AckedState_EffectiveDisplayName = 9243
    LimitAlarmType_AckedState_TransitionTime = 9244
    LimitAlarmType_AckedState_EffectiveTransitionTime = 9245
    LimitAlarmType_AckedState_TrueState = 9246
    LimitAlarmType_AckedState_FalseState = 9247
    LimitAlarmType_ConfirmedState = 9248
    LimitAlarmType_ConfirmedState_Id = 9249
    LimitAlarmType_ConfirmedState_Name = 9250
    LimitAlarmType_ConfirmedState_Number = 9251
    LimitAlarmType_ConfirmedState_EffectiveDisplayName = 9252
    LimitAlarmType_ConfirmedState_TransitionTime = 9253
    LimitAlarmType_ConfirmedState_EffectiveTransitionTime = 9254
    LimitAlarmType_ConfirmedState_TrueState = 9255
    LimitAlarmType_ConfirmedState_FalseState = 9256
    LimitAlarmType_Acknowledge = 9257
    LimitAlarmType_Acknowledge_InputArguments = 9258
    LimitAlarmType_Confirm = 9259
    LimitAlarmType_Confirm_InputArguments = 9260
    LimitAlarmType_ActiveState = 9261
    LimitAlarmType_ActiveState_Id = 9262
    LimitAlarmType_ActiveState_Name = 9263
    LimitAlarmType_ActiveState_Number = 9264
    LimitAlarmType_ActiveState_EffectiveDisplayName = 9265
    LimitAlarmType_ActiveState_TransitionTime = 9266
    LimitAlarmType_ActiveState_EffectiveTransitionTime = 9267
    LimitAlarmType_ActiveState_TrueState = 9268
    LimitAlarmType_ActiveState_FalseState = 9269
    LimitAlarmType_SuppressedState = 9270
    LimitAlarmType_SuppressedState_Id = 9271
    LimitAlarmType_SuppressedState_Name = 9272
    LimitAlarmType_SuppressedState_Number = 9273
    LimitAlarmType_SuppressedState_EffectiveDisplayName = 9274
    LimitAlarmType_SuppressedState_TransitionTime = 9275
    LimitAlarmType_SuppressedState_EffectiveTransitionTime = 9276
    LimitAlarmType_SuppressedState_TrueState = 9277
    LimitAlarmType_SuppressedState_FalseState = 9278
    LimitAlarmType_ShelvingState = 9279
    LimitAlarmType_ShelvingState_CurrentState = 9280
    LimitAlarmType_ShelvingState_CurrentState_Id = 9281
    LimitAlarmType_ShelvingState_CurrentState_Name = 9282
    LimitAlarmType_ShelvingState_CurrentState_Number = 9283
    LimitAlarmType_ShelvingState_CurrentState_EffectiveDisplayName = 9284
    LimitAlarmType_ShelvingState_LastTransition = 9285
    LimitAlarmType_ShelvingState_LastTransition_Id = 9286
    LimitAlarmType_ShelvingState_LastTransition_Name = 9287
    LimitAlarmType_ShelvingState_LastTransition_Number = 9288
    LimitAlarmType_ShelvingState_LastTransition_TransitionTime = 9289
    LimitAlarmType_ShelvingState_UnshelveTime = 9290
    LimitAlarmType_ShelvingState_Unshelve = 9312
    LimitAlarmType_ShelvingState_OneShotShelve = 9313
    LimitAlarmType_ShelvingState_TimedShelve = 9314
    LimitAlarmType_ShelvingState_TimedShelve_InputArguments = 9315
    LimitAlarmType_SuppressedOrShelved = 9316
    LimitAlarmType_MaxTimeShelved = 9317
    ExclusiveLimitStateMachineType = 9318
    ExclusiveLimitStateMachineType_CurrentState = 9319
    ExclusiveLimitStateMachineType_CurrentState_Id = 9320
    ExclusiveLimitStateMachineType_CurrentState_Name = 9321
    ExclusiveLimitStateMachineType_CurrentState_Number = 9322
    ExclusiveLimitStateMachineType_CurrentState_EffectiveDisplayName = 9323
    ExclusiveLimitStateMachineType_LastTransition = 9324
    ExclusiveLimitStateMachineType_LastTransition_Id = 9325
    ExclusiveLimitStateMachineType_LastTransition_Name = 9326
    ExclusiveLimitStateMachineType_LastTransition_Number = 9327
    ExclusiveLimitStateMachineType_LastTransition_TransitionTime = 9328
    ExclusiveLimitStateMachineType_HighHigh = 9329
    ExclusiveLimitStateMachineType_HighHigh_StateNumber = 9330
    ExclusiveLimitStateMachineType_High = 9331
    ExclusiveLimitStateMachineType_High_StateNumber = 9332
    ExclusiveLimitStateMachineType_Low = 9333
    ExclusiveLimitStateMachineType_Low_StateNumber = 9334
    ExclusiveLimitStateMachineType_LowLow = 9335
    ExclusiveLimitStateMachineType_LowLow_StateNumber = 9336
    ExclusiveLimitStateMachineType_LowLowToLow = 9337
    ExclusiveLimitStateMachineType_LowToLowLow = 9338
    ExclusiveLimitStateMachineType_HighHighToHigh = 9339
    ExclusiveLimitStateMachineType_HighToHighHigh = 9340
    ExclusiveLimitAlarmType = 9341
    ExclusiveLimitAlarmType_EventId = 9342
    ExclusiveLimitAlarmType_EventType = 9343
    ExclusiveLimitAlarmType_SourceNode = 9344
    ExclusiveLimitAlarmType_SourceName = 9345
    ExclusiveLimitAlarmType_Time = 9346
    ExclusiveLimitAlarmType_ReceiveTime = 9347
    ExclusiveLimitAlarmType_LocalTime = 9348
    ExclusiveLimitAlarmType_Message = 9349
    ExclusiveLimitAlarmType_Severity = 9350
    ExclusiveLimitAlarmType_ConditionName = 9351
    ExclusiveLimitAlarmType_BranchId = 9352
    ExclusiveLimitAlarmType_Retain = 9353
    ExclusiveLimitAlarmType_EnabledState = 9354
    ExclusiveLimitAlarmType_EnabledState_Id = 9355
    ExclusiveLimitAlarmType_EnabledState_Name = 9356
    ExclusiveLimitAlarmType_EnabledState_Number = 9357
    ExclusiveLimitAlarmType_EnabledState_EffectiveDisplayName = 9358
    ExclusiveLimitAlarmType_EnabledState_TransitionTime = 9359
    ExclusiveLimitAlarmType_EnabledState_EffectiveTransitionTime = 9360
    ExclusiveLimitAlarmType_EnabledState_TrueState = 9361
    ExclusiveLimitAlarmType_EnabledState_FalseState = 9362
    ExclusiveLimitAlarmType_Quality = 9363
    ExclusiveLimitAlarmType_Quality_SourceTimestamp = 9364
    ExclusiveLimitAlarmType_LastSeverity = 9365
    ExclusiveLimitAlarmType_LastSeverity_SourceTimestamp = 9366
    ExclusiveLimitAlarmType_Comment = 9367
    ExclusiveLimitAlarmType_Comment_SourceTimestamp = 9368
    ExclusiveLimitAlarmType_ClientUserId = 9369
    ExclusiveLimitAlarmType_Enable = 9370
    ExclusiveLimitAlarmType_Disable = 9371
    ExclusiveLimitAlarmType_AddComment = 9372
    ExclusiveLimitAlarmType_AddComment_InputArguments = 9373
    ExclusiveLimitAlarmType_ConditionRefresh = 9374
    ExclusiveLimitAlarmType_ConditionRefresh_InputArguments = 9375
    ExclusiveLimitAlarmType_AckedState = 9376
    ExclusiveLimitAlarmType_AckedState_Id = 9377
    ExclusiveLimitAlarmType_AckedState_Name = 9378
    ExclusiveLimitAlarmType_AckedState_Number = 9379
    ExclusiveLimitAlarmType_AckedState_EffectiveDisplayName = 9380
    ExclusiveLimitAlarmType_AckedState_TransitionTime = 9381
    ExclusiveLimitAlarmType_AckedState_EffectiveTransitionTime = 9382
    ExclusiveLimitAlarmType_AckedState_TrueState = 9383
    ExclusiveLimitAlarmType_AckedState_FalseState = 9384
    ExclusiveLimitAlarmType_ConfirmedState = 9385
    ExclusiveLimitAlarmType_ConfirmedState_Id = 9386
    ExclusiveLimitAlarmType_ConfirmedState_Name = 9387
    ExclusiveLimitAlarmType_ConfirmedState_Number = 9388
    ExclusiveLimitAlarmType_ConfirmedState_EffectiveDisplayName = 9389
    ExclusiveLimitAlarmType_ConfirmedState_TransitionTime = 9390
    ExclusiveLimitAlarmType_ConfirmedState_EffectiveTransitionTime = 9391
    ExclusiveLimitAlarmType_ConfirmedState_TrueState = 9392
    ExclusiveLimitAlarmType_ConfirmedState_FalseState = 9393
    ExclusiveLimitAlarmType_Acknowledge = 9394
    ExclusiveLimitAlarmType_Acknowledge_InputArguments = 9395
    ExclusiveLimitAlarmType_Confirm = 9396
    ExclusiveLimitAlarmType_Confirm_InputArguments = 9397
    ExclusiveLimitAlarmType_ActiveState = 9398
    ExclusiveLimitAlarmType_ActiveState_Id = 9399
    ExclusiveLimitAlarmType_ActiveState_Name = 9400
    ExclusiveLimitAlarmType_ActiveState_Number = 9401
    ExclusiveLimitAlarmType_ActiveState_EffectiveDisplayName = 9402
    ExclusiveLimitAlarmType_ActiveState_TransitionTime = 9403
    ExclusiveLimitAlarmType_ActiveState_EffectiveTransitionTime = 9404
    ExclusiveLimitAlarmType_ActiveState_TrueState = 9405
    ExclusiveLimitAlarmType_ActiveState_FalseState = 9406
    ExclusiveLimitAlarmType_SuppressedState = 9407
    ExclusiveLimitAlarmType_SuppressedState_Id = 9408
    ExclusiveLimitAlarmType_SuppressedState_Name = 9409
    ExclusiveLimitAlarmType_SuppressedState_Number = 9410
    ExclusiveLimitAlarmType_SuppressedState_EffectiveDisplayName = 9411
    ExclusiveLimitAlarmType_SuppressedState_TransitionTime = 9412
    ExclusiveLimitAlarmType_SuppressedState_EffectiveTransitionTime = 9413
    ExclusiveLimitAlarmType_SuppressedState_TrueState = 9414
    ExclusiveLimitAlarmType_SuppressedState_FalseState = 9415
    ExclusiveLimitAlarmType_ShelvingState = 9416
    ExclusiveLimitAlarmType_ShelvingState_CurrentState = 9417
    ExclusiveLimitAlarmType_ShelvingState_CurrentState_Id = 9418
    ExclusiveLimitAlarmType_ShelvingState_CurrentState_Name = 9419
    ExclusiveLimitAlarmType_ShelvingState_CurrentState_Number = 9420
    ExclusiveLimitAlarmType_ShelvingState_CurrentState_EffectiveDisplayName = 9421
    ExclusiveLimitAlarmType_ShelvingState_LastTransition = 9422
    ExclusiveLimitAlarmType_ShelvingState_LastTransition_Id = 9423
    ExclusiveLimitAlarmType_ShelvingState_LastTransition_Name = 9424
    ExclusiveLimitAlarmType_ShelvingState_LastTransition_Number = 9425
    ExclusiveLimitAlarmType_ShelvingState_LastTransition_TransitionTime = 9426
    ExclusiveLimitAlarmType_ShelvingState_UnshelveTime = 9427
    ExclusiveLimitAlarmType_ShelvingState_Unshelve = 9449
    ExclusiveLimitAlarmType_ShelvingState_OneShotShelve = 9450
    ExclusiveLimitAlarmType_ShelvingState_TimedShelve = 9451
    ExclusiveLimitAlarmType_ShelvingState_TimedShelve_InputArguments = 9452
    ExclusiveLimitAlarmType_SuppressedOrShelved = 9453
    ExclusiveLimitAlarmType_MaxTimeShelved = 9454
    ExclusiveLimitAlarmType_LimitState = 9455
    ExclusiveLimitAlarmType_LimitState_CurrentState = 9456
    ExclusiveLimitAlarmType_LimitState_CurrentState_Id = 9457
    ExclusiveLimitAlarmType_LimitState_CurrentState_Name = 9458
    ExclusiveLimitAlarmType_LimitState_CurrentState_Number = 9459
    ExclusiveLimitAlarmType_LimitState_CurrentState_EffectiveDisplayName = 9460
    ExclusiveLimitAlarmType_LimitState_LastTransition = 9461
    ExclusiveLimitAlarmType_LimitState_LastTransition_Id = 9462
    ExclusiveLimitAlarmType_LimitState_LastTransition_Name = 9463
    ExclusiveLimitAlarmType_LimitState_LastTransition_Number = 9464
    ExclusiveLimitAlarmType_LimitState_LastTransition_TransitionTime = 9465
    ExclusiveLimitAlarmType_HighHighLimit = 9478
    ExclusiveLimitAlarmType_HighLimit = 9479
    ExclusiveLimitAlarmType_LowLimit = 9480
    ExclusiveLimitAlarmType_LowLowLimit = 9481
    ExclusiveLevelAlarmType = 9482
    ExclusiveLevelAlarmType_EventId = 9483
    ExclusiveLevelAlarmType_EventType = 9484
    ExclusiveLevelAlarmType_SourceNode = 9485
    ExclusiveLevelAlarmType_SourceName = 9486
    ExclusiveLevelAlarmType_Time = 9487
    ExclusiveLevelAlarmType_ReceiveTime = 9488
    ExclusiveLevelAlarmType_LocalTime = 9489
    ExclusiveLevelAlarmType_Message = 9490
    ExclusiveLevelAlarmType_Severity = 9491
    ExclusiveLevelAlarmType_ConditionName = 9492
    ExclusiveLevelAlarmType_BranchId = 9493
    ExclusiveLevelAlarmType_Retain = 9494
    ExclusiveLevelAlarmType_EnabledState = 9495
    ExclusiveLevelAlarmType_EnabledState_Id = 9496
    ExclusiveLevelAlarmType_EnabledState_Name = 9497
    ExclusiveLevelAlarmType_EnabledState_Number = 9498
    ExclusiveLevelAlarmType_EnabledState_EffectiveDisplayName = 9499
    ExclusiveLevelAlarmType_EnabledState_TransitionTime = 9500
    ExclusiveLevelAlarmType_EnabledState_EffectiveTransitionTime = 9501
    ExclusiveLevelAlarmType_EnabledState_TrueState = 9502
    ExclusiveLevelAlarmType_EnabledState_FalseState = 9503
    ExclusiveLevelAlarmType_Quality = 9504
    ExclusiveLevelAlarmType_Quality_SourceTimestamp = 9505
    ExclusiveLevelAlarmType_LastSeverity = 9506
    ExclusiveLevelAlarmType_LastSeverity_SourceTimestamp = 9507
    ExclusiveLevelAlarmType_Comment = 9508
    ExclusiveLevelAlarmType_Comment_SourceTimestamp = 9509
    ExclusiveLevelAlarmType_ClientUserId = 9510
    ExclusiveLevelAlarmType_Enable = 9511
    ExclusiveLevelAlarmType_Disable = 9512
    ExclusiveLevelAlarmType_AddComment = 9513
    ExclusiveLevelAlarmType_AddComment_InputArguments = 9514
    ExclusiveLevelAlarmType_ConditionRefresh = 9515
    ExclusiveLevelAlarmType_ConditionRefresh_InputArguments = 9516
    ExclusiveLevelAlarmType_AckedState = 9517
    ExclusiveLevelAlarmType_AckedState_Id = 9518
    ExclusiveLevelAlarmType_AckedState_Name = 9519
    ExclusiveLevelAlarmType_AckedState_Number = 9520
    ExclusiveLevelAlarmType_AckedState_EffectiveDisplayName = 9521
    ExclusiveLevelAlarmType_AckedState_TransitionTime = 9522
    ExclusiveLevelAlarmType_AckedState_EffectiveTransitionTime = 9523
    ExclusiveLevelAlarmType_AckedState_TrueState = 9524
    ExclusiveLevelAlarmType_AckedState_FalseState = 9525
    ExclusiveLevelAlarmType_ConfirmedState = 9526
    ExclusiveLevelAlarmType_ConfirmedState_Id = 9527
    ExclusiveLevelAlarmType_ConfirmedState_Name = 9528
    ExclusiveLevelAlarmType_ConfirmedState_Number = 9529
    ExclusiveLevelAlarmType_ConfirmedState_EffectiveDisplayName = 9530
    ExclusiveLevelAlarmType_ConfirmedState_TransitionTime = 9531
    ExclusiveLevelAlarmType_ConfirmedState_EffectiveTransitionTime = 9532
    ExclusiveLevelAlarmType_ConfirmedState_TrueState = 9533
    ExclusiveLevelAlarmType_ConfirmedState_FalseState = 9534
    ExclusiveLevelAlarmType_Acknowledge = 9535
    ExclusiveLevelAlarmType_Acknowledge_InputArguments = 9536
    ExclusiveLevelAlarmType_Confirm = 9537
    ExclusiveLevelAlarmType_Confirm_InputArguments = 9538
    ExclusiveLevelAlarmType_ActiveState = 9539
    ExclusiveLevelAlarmType_ActiveState_Id = 9540
    ExclusiveLevelAlarmType_ActiveState_Name = 9541
    ExclusiveLevelAlarmType_ActiveState_Number = 9542
    ExclusiveLevelAlarmType_ActiveState_EffectiveDisplayName = 9543
    ExclusiveLevelAlarmType_ActiveState_TransitionTime = 9544
    ExclusiveLevelAlarmType_ActiveState_EffectiveTransitionTime = 9545
    ExclusiveLevelAlarmType_ActiveState_TrueState = 9546
    ExclusiveLevelAlarmType_ActiveState_FalseState = 9547
    ExclusiveLevelAlarmType_SuppressedState = 9548
    ExclusiveLevelAlarmType_SuppressedState_Id = 9549
    ExclusiveLevelAlarmType_SuppressedState_Name = 9550
    ExclusiveLevelAlarmType_SuppressedState_Number = 9551
    ExclusiveLevelAlarmType_SuppressedState_EffectiveDisplayName = 9552
    ExclusiveLevelAlarmType_SuppressedState_TransitionTime = 9553
    ExclusiveLevelAlarmType_SuppressedState_EffectiveTransitionTime = 9554
    ExclusiveLevelAlarmType_SuppressedState_TrueState = 9555
    ExclusiveLevelAlarmType_SuppressedState_FalseState = 9556
    ExclusiveLevelAlarmType_ShelvingState = 9557
    ExclusiveLevelAlarmType_ShelvingState_CurrentState = 9558
    ExclusiveLevelAlarmType_ShelvingState_CurrentState_Id = 9559
    ExclusiveLevelAlarmType_ShelvingState_CurrentState_Name = 9560
    ExclusiveLevelAlarmType_ShelvingState_CurrentState_Number = 9561
    ExclusiveLevelAlarmType_ShelvingState_CurrentState_EffectiveDisplayName = 9562
    ExclusiveLevelAlarmType_ShelvingState_LastTransition = 9563
    ExclusiveLevelAlarmType_ShelvingState_LastTransition_Id = 9564
    ExclusiveLevelAlarmType_ShelvingState_LastTransition_Name = 9565
    ExclusiveLevelAlarmType_ShelvingState_LastTransition_Number = 9566
    ExclusiveLevelAlarmType_ShelvingState_LastTransition_TransitionTime = 9567
    ExclusiveLevelAlarmType_ShelvingState_UnshelveTime = 9568
    ExclusiveLevelAlarmType_ShelvingState_Unshelve = 9590
    ExclusiveLevelAlarmType_ShelvingState_OneShotShelve = 9591
    ExclusiveLevelAlarmType_ShelvingState_TimedShelve = 9592
    ExclusiveLevelAlarmType_ShelvingState_TimedShelve_InputArguments = 9593
    ExclusiveLevelAlarmType_SuppressedOrShelved = 9594
    ExclusiveLevelAlarmType_MaxTimeShelved = 9595
    ExclusiveLevelAlarmType_LimitState = 9596
    ExclusiveLevelAlarmType_LimitState_CurrentState = 9597
    ExclusiveLevelAlarmType_LimitState_CurrentState_Id = 9598
    ExclusiveLevelAlarmType_LimitState_CurrentState_Name = 9599
    ExclusiveLevelAlarmType_LimitState_CurrentState_Number = 9600
    ExclusiveLevelAlarmType_LimitState_CurrentState_EffectiveDisplayName = 9601
    ExclusiveLevelAlarmType_LimitState_LastTransition = 9602
    ExclusiveLevelAlarmType_LimitState_LastTransition_Id = 9603
    ExclusiveLevelAlarmType_LimitState_LastTransition_Name = 9604
    ExclusiveLevelAlarmType_LimitState_LastTransition_Number = 9605
    ExclusiveLevelAlarmType_LimitState_LastTransition_TransitionTime = 9606
    ExclusiveLevelAlarmType_HighHighLimit = 9619
    ExclusiveLevelAlarmType_HighLimit = 9620
    ExclusiveLevelAlarmType_LowLimit = 9621
    ExclusiveLevelAlarmType_LowLowLimit = 9622
    ExclusiveRateOfChangeAlarmType = 9623
    ExclusiveRateOfChangeAlarmType_EventId = 9624
    ExclusiveRateOfChangeAlarmType_EventType = 9625
    ExclusiveRateOfChangeAlarmType_SourceNode = 9626
    ExclusiveRateOfChangeAlarmType_SourceName = 9627
    ExclusiveRateOfChangeAlarmType_Time = 9628
    ExclusiveRateOfChangeAlarmType_ReceiveTime = 9629
    ExclusiveRateOfChangeAlarmType_LocalTime = 9630
    ExclusiveRateOfChangeAlarmType_Message = 9631
    ExclusiveRateOfChangeAlarmType_Severity = 9632
    ExclusiveRateOfChangeAlarmType_ConditionName = 9633
    ExclusiveRateOfChangeAlarmType_BranchId = 9634
    ExclusiveRateOfChangeAlarmType_Retain = 9635
    ExclusiveRateOfChangeAlarmType_EnabledState = 9636
    ExclusiveRateOfChangeAlarmType_EnabledState_Id = 9637
    ExclusiveRateOfChangeAlarmType_EnabledState_Name = 9638
    ExclusiveRateOfChangeAlarmType_EnabledState_Number = 9639
    ExclusiveRateOfChangeAlarmType_EnabledState_EffectiveDisplayName = 9640
    ExclusiveRateOfChangeAlarmType_EnabledState_TransitionTime = 9641
    ExclusiveRateOfChangeAlarmType_EnabledState_EffectiveTransitionTime = 9642
    ExclusiveRateOfChangeAlarmType_EnabledState_TrueState = 9643
    ExclusiveRateOfChangeAlarmType_EnabledState_FalseState = 9644
    ExclusiveRateOfChangeAlarmType_Quality = 9645
    ExclusiveRateOfChangeAlarmType_Quality_SourceTimestamp = 9646
    ExclusiveRateOfChangeAlarmType_LastSeverity = 9647
    ExclusiveRateOfChangeAlarmType_LastSeverity_SourceTimestamp = 9648
    ExclusiveRateOfChangeAlarmType_Comment = 9649
    ExclusiveRateOfChangeAlarmType_Comment_SourceTimestamp = 9650
    ExclusiveRateOfChangeAlarmType_ClientUserId = 9651
    ExclusiveRateOfChangeAlarmType_Enable = 9652
    ExclusiveRateOfChangeAlarmType_Disable = 9653
    ExclusiveRateOfChangeAlarmType_AddComment = 9654
    ExclusiveRateOfChangeAlarmType_AddComment_InputArguments = 9655
    ExclusiveRateOfChangeAlarmType_ConditionRefresh = 9656
    ExclusiveRateOfChangeAlarmType_ConditionRefresh_InputArguments = 9657
    ExclusiveRateOfChangeAlarmType_AckedState = 9658
    ExclusiveRateOfChangeAlarmType_AckedState_Id = 9659
    ExclusiveRateOfChangeAlarmType_AckedState_Name = 9660
    ExclusiveRateOfChangeAlarmType_AckedState_Number = 9661
    ExclusiveRateOfChangeAlarmType_AckedState_EffectiveDisplayName = 9662
    ExclusiveRateOfChangeAlarmType_AckedState_TransitionTime = 9663
    ExclusiveRateOfChangeAlarmType_AckedState_EffectiveTransitionTime = 9664
    ExclusiveRateOfChangeAlarmType_AckedState_TrueState = 9665
    ExclusiveRateOfChangeAlarmType_AckedState_FalseState = 9666
    ExclusiveRateOfChangeAlarmType_ConfirmedState = 9667
    ExclusiveRateOfChangeAlarmType_ConfirmedState_Id = 9668
    ExclusiveRateOfChangeAlarmType_ConfirmedState_Name = 9669
    ExclusiveRateOfChangeAlarmType_ConfirmedState_Number = 9670
    ExclusiveRateOfChangeAlarmType_ConfirmedState_EffectiveDisplayName = 9671
    ExclusiveRateOfChangeAlarmType_ConfirmedState_TransitionTime = 9672
    ExclusiveRateOfChangeAlarmType_ConfirmedState_EffectiveTransitionTime = 9673
    ExclusiveRateOfChangeAlarmType_ConfirmedState_TrueState = 9674
    ExclusiveRateOfChangeAlarmType_ConfirmedState_FalseState = 9675
    ExclusiveRateOfChangeAlarmType_Acknowledge = 9676
    ExclusiveRateOfChangeAlarmType_Acknowledge_InputArguments = 9677
    ExclusiveRateOfChangeAlarmType_Confirm = 9678
    ExclusiveRateOfChangeAlarmType_Confirm_InputArguments = 9679
    ExclusiveRateOfChangeAlarmType_ActiveState = 9680
    ExclusiveRateOfChangeAlarmType_ActiveState_Id = 9681
    ExclusiveRateOfChangeAlarmType_ActiveState_Name = 9682
    ExclusiveRateOfChangeAlarmType_ActiveState_Number = 9683
    ExclusiveRateOfChangeAlarmType_ActiveState_EffectiveDisplayName = 9684
    ExclusiveRateOfChangeAlarmType_ActiveState_TransitionTime = 9685
    ExclusiveRateOfChangeAlarmType_ActiveState_EffectiveTransitionTime = 9686
    ExclusiveRateOfChangeAlarmType_ActiveState_TrueState = 9687
    ExclusiveRateOfChangeAlarmType_ActiveState_FalseState = 9688
    ExclusiveRateOfChangeAlarmType_SuppressedState = 9689
    ExclusiveRateOfChangeAlarmType_SuppressedState_Id = 9690
    ExclusiveRateOfChangeAlarmType_SuppressedState_Name = 9691
    ExclusiveRateOfChangeAlarmType_SuppressedState_Number = 9692
    ExclusiveRateOfChangeAlarmType_SuppressedState_EffectiveDisplayName = 9693
    ExclusiveRateOfChangeAlarmType_SuppressedState_TransitionTime = 9694
    ExclusiveRateOfChangeAlarmType_SuppressedState_EffectiveTransitionTime = 9695
    ExclusiveRateOfChangeAlarmType_SuppressedState_TrueState = 9696
    ExclusiveRateOfChangeAlarmType_SuppressedState_FalseState = 9697
    ExclusiveRateOfChangeAlarmType_ShelvingState = 9698
    ExclusiveRateOfChangeAlarmType_ShelvingState_CurrentState = 9699
    ExclusiveRateOfChangeAlarmType_ShelvingState_CurrentState_Id = 9700
    ExclusiveRateOfChangeAlarmType_ShelvingState_CurrentState_Name = 9701
    ExclusiveRateOfChangeAlarmType_ShelvingState_CurrentState_Number = 9702
    ExclusiveRateOfChangeAlarmType_ShelvingState_CurrentState_EffectiveDisplayName = 9703
    ExclusiveRateOfChangeAlarmType_ShelvingState_LastTransition = 9704
    ExclusiveRateOfChangeAlarmType_ShelvingState_LastTransition_Id = 9705
    ExclusiveRateOfChangeAlarmType_ShelvingState_LastTransition_Name = 9706
    ExclusiveRateOfChangeAlarmType_ShelvingState_LastTransition_Number = 9707
    ExclusiveRateOfChangeAlarmType_ShelvingState_LastTransition_TransitionTime = 9708
    ExclusiveRateOfChangeAlarmType_ShelvingState_UnshelveTime = 9709
    ExclusiveRateOfChangeAlarmType_ShelvingState_Unshelve = 9731
    ExclusiveRateOfChangeAlarmType_ShelvingState_OneShotShelve = 9732
    ExclusiveRateOfChangeAlarmType_ShelvingState_TimedShelve = 9733
    ExclusiveRateOfChangeAlarmType_ShelvingState_TimedShelve_InputArguments = 9734
    ExclusiveRateOfChangeAlarmType_SuppressedOrShelved = 9735
    ExclusiveRateOfChangeAlarmType_MaxTimeShelved = 9736
    ExclusiveRateOfChangeAlarmType_LimitState = 9737
    ExclusiveRateOfChangeAlarmType_LimitState_CurrentState = 9738
    ExclusiveRateOfChangeAlarmType_LimitState_CurrentState_Id = 9739
    ExclusiveRateOfChangeAlarmType_LimitState_CurrentState_Name = 9740
    ExclusiveRateOfChangeAlarmType_LimitState_CurrentState_Number = 9741
    ExclusiveRateOfChangeAlarmType_LimitState_CurrentState_EffectiveDisplayName = 9742
    ExclusiveRateOfChangeAlarmType_LimitState_LastTransition = 9743
    ExclusiveRateOfChangeAlarmType_LimitState_LastTransition_Id = 9744
    ExclusiveRateOfChangeAlarmType_LimitState_LastTransition_Name = 9745
    ExclusiveRateOfChangeAlarmType_LimitState_LastTransition_Number = 9746
    ExclusiveRateOfChangeAlarmType_LimitState_LastTransition_TransitionTime = 9747
    ExclusiveRateOfChangeAlarmType_HighHighLimit = 9760
    ExclusiveRateOfChangeAlarmType_HighLimit = 9761
    ExclusiveRateOfChangeAlarmType_LowLimit = 9762
    ExclusiveRateOfChangeAlarmType_LowLowLimit = 9763
    ExclusiveDeviationAlarmType = 9764
    ExclusiveDeviationAlarmType_EventId = 9765
    ExclusiveDeviationAlarmType_EventType = 9766
    ExclusiveDeviationAlarmType_SourceNode = 9767
    ExclusiveDeviationAlarmType_SourceName = 9768
    ExclusiveDeviationAlarmType_Time = 9769
    ExclusiveDeviationAlarmType_ReceiveTime = 9770
    ExclusiveDeviationAlarmType_LocalTime = 9771
    ExclusiveDeviationAlarmType_Message = 9772
    ExclusiveDeviationAlarmType_Severity = 9773
    ExclusiveDeviationAlarmType_ConditionName = 9774
    ExclusiveDeviationAlarmType_BranchId = 9775
    ExclusiveDeviationAlarmType_Retain = 9776
    ExclusiveDeviationAlarmType_EnabledState = 9777
    ExclusiveDeviationAlarmType_EnabledState_Id = 9778
    ExclusiveDeviationAlarmType_EnabledState_Name = 9779
    ExclusiveDeviationAlarmType_EnabledState_Number = 9780
    ExclusiveDeviationAlarmType_EnabledState_EffectiveDisplayName = 9781
    ExclusiveDeviationAlarmType_EnabledState_TransitionTime = 9782
    ExclusiveDeviationAlarmType_EnabledState_EffectiveTransitionTime = 9783
    ExclusiveDeviationAlarmType_EnabledState_TrueState = 9784
    ExclusiveDeviationAlarmType_EnabledState_FalseState = 9785
    ExclusiveDeviationAlarmType_Quality = 9786
    ExclusiveDeviationAlarmType_Quality_SourceTimestamp = 9787
    ExclusiveDeviationAlarmType_LastSeverity = 9788
    ExclusiveDeviationAlarmType_LastSeverity_SourceTimestamp = 9789
    ExclusiveDeviationAlarmType_Comment = 9790
    ExclusiveDeviationAlarmType_Comment_SourceTimestamp = 9791
    ExclusiveDeviationAlarmType_ClientUserId = 9792
    ExclusiveDeviationAlarmType_Enable = 9793
    ExclusiveDeviationAlarmType_Disable = 9794
    ExclusiveDeviationAlarmType_AddComment = 9795
    ExclusiveDeviationAlarmType_AddComment_InputArguments = 9796
    ExclusiveDeviationAlarmType_ConditionRefresh = 9797
    ExclusiveDeviationAlarmType_ConditionRefresh_InputArguments = 9798
    ExclusiveDeviationAlarmType_AckedState = 9799
    ExclusiveDeviationAlarmType_AckedState_Id = 9800
    ExclusiveDeviationAlarmType_AckedState_Name = 9801
    ExclusiveDeviationAlarmType_AckedState_Number = 9802
    ExclusiveDeviationAlarmType_AckedState_EffectiveDisplayName = 9803
    ExclusiveDeviationAlarmType_AckedState_TransitionTime = 9804
    ExclusiveDeviationAlarmType_AckedState_EffectiveTransitionTime = 9805
    ExclusiveDeviationAlarmType_AckedState_TrueState = 9806
    ExclusiveDeviationAlarmType_AckedState_FalseState = 9807
    ExclusiveDeviationAlarmType_ConfirmedState = 9808
    ExclusiveDeviationAlarmType_ConfirmedState_Id = 9809
    ExclusiveDeviationAlarmType_ConfirmedState_Name = 9810
    ExclusiveDeviationAlarmType_ConfirmedState_Number = 9811
    ExclusiveDeviationAlarmType_ConfirmedState_EffectiveDisplayName = 9812
    ExclusiveDeviationAlarmType_ConfirmedState_TransitionTime = 9813
    ExclusiveDeviationAlarmType_ConfirmedState_EffectiveTransitionTime = 9814
    ExclusiveDeviationAlarmType_ConfirmedState_TrueState = 9815
    ExclusiveDeviationAlarmType_ConfirmedState_FalseState = 9816
    ExclusiveDeviationAlarmType_Acknowledge = 9817
    ExclusiveDeviationAlarmType_Acknowledge_InputArguments = 9818
    ExclusiveDeviationAlarmType_Confirm = 9819
    ExclusiveDeviationAlarmType_Confirm_InputArguments = 9820
    ExclusiveDeviationAlarmType_ActiveState = 9821
    ExclusiveDeviationAlarmType_ActiveState_Id = 9822
    ExclusiveDeviationAlarmType_ActiveState_Name = 9823
    ExclusiveDeviationAlarmType_ActiveState_Number = 9824
    ExclusiveDeviationAlarmType_ActiveState_EffectiveDisplayName = 9825
    ExclusiveDeviationAlarmType_ActiveState_TransitionTime = 9826
    ExclusiveDeviationAlarmType_ActiveState_EffectiveTransitionTime = 9827
    ExclusiveDeviationAlarmType_ActiveState_TrueState = 9828
    ExclusiveDeviationAlarmType_ActiveState_FalseState = 9829
    ExclusiveDeviationAlarmType_SuppressedState = 9830
    ExclusiveDeviationAlarmType_SuppressedState_Id = 9831
    ExclusiveDeviationAlarmType_SuppressedState_Name = 9832
    ExclusiveDeviationAlarmType_SuppressedState_Number = 9833
    ExclusiveDeviationAlarmType_SuppressedState_EffectiveDisplayName = 9834
    ExclusiveDeviationAlarmType_SuppressedState_TransitionTime = 9835
    ExclusiveDeviationAlarmType_SuppressedState_EffectiveTransitionTime = 9836
    ExclusiveDeviationAlarmType_SuppressedState_TrueState = 9837
    ExclusiveDeviationAlarmType_SuppressedState_FalseState = 9838
    ExclusiveDeviationAlarmType_ShelvingState = 9839
    ExclusiveDeviationAlarmType_ShelvingState_CurrentState = 9840
    ExclusiveDeviationAlarmType_ShelvingState_CurrentState_Id = 9841
    ExclusiveDeviationAlarmType_ShelvingState_CurrentState_Name = 9842
    ExclusiveDeviationAlarmType_ShelvingState_CurrentState_Number = 9843
    ExclusiveDeviationAlarmType_ShelvingState_CurrentState_EffectiveDisplayName = 9844
    ExclusiveDeviationAlarmType_ShelvingState_LastTransition = 9845
    ExclusiveDeviationAlarmType_ShelvingState_LastTransition_Id = 9846
    ExclusiveDeviationAlarmType_ShelvingState_LastTransition_Name = 9847
    ExclusiveDeviationAlarmType_ShelvingState_LastTransition_Number = 9848
    ExclusiveDeviationAlarmType_ShelvingState_LastTransition_TransitionTime = 9849
    ExclusiveDeviationAlarmType_ShelvingState_UnshelveTime = 9850
    ExclusiveDeviationAlarmType_ShelvingState_Unshelve = 9872
    ExclusiveDeviationAlarmType_ShelvingState_OneShotShelve = 9873
    ExclusiveDeviationAlarmType_ShelvingState_TimedShelve = 9874
    ExclusiveDeviationAlarmType_ShelvingState_TimedShelve_InputArguments = 9875
    ExclusiveDeviationAlarmType_SuppressedOrShelved = 9876
    ExclusiveDeviationAlarmType_MaxTimeShelved = 9877
    ExclusiveDeviationAlarmType_LimitState = 9878
    ExclusiveDeviationAlarmType_LimitState_CurrentState = 9879
    ExclusiveDeviationAlarmType_LimitState_CurrentState_Id = 9880
    ExclusiveDeviationAlarmType_LimitState_CurrentState_Name = 9881
    ExclusiveDeviationAlarmType_LimitState_CurrentState_Number = 9882
    ExclusiveDeviationAlarmType_LimitState_CurrentState_EffectiveDisplayName = 9883
    ExclusiveDeviationAlarmType_LimitState_LastTransition = 9884
    ExclusiveDeviationAlarmType_LimitState_LastTransition_Id = 9885
    ExclusiveDeviationAlarmType_LimitState_LastTransition_Name = 9886
    ExclusiveDeviationAlarmType_LimitState_LastTransition_Number = 9887
    ExclusiveDeviationAlarmType_LimitState_LastTransition_TransitionTime = 9888
    ExclusiveDeviationAlarmType_HighHighLimit = 9901
    ExclusiveDeviationAlarmType_HighLimit = 9902
    ExclusiveDeviationAlarmType_LowLimit = 9903
    ExclusiveDeviationAlarmType_LowLowLimit = 9904
    ExclusiveDeviationAlarmType_SetpointNode = 9905
    NonExclusiveLimitAlarmType = 9906
    NonExclusiveLimitAlarmType_EventId = 9907
    NonExclusiveLimitAlarmType_EventType = 9908
    NonExclusiveLimitAlarmType_SourceNode = 9909
    NonExclusiveLimitAlarmType_SourceName = 9910
    NonExclusiveLimitAlarmType_Time = 9911
    NonExclusiveLimitAlarmType_ReceiveTime = 9912
    NonExclusiveLimitAlarmType_LocalTime = 9913
    NonExclusiveLimitAlarmType_Message = 9914
    NonExclusiveLimitAlarmType_Severity = 9915
    NonExclusiveLimitAlarmType_ConditionName = 9916
    NonExclusiveLimitAlarmType_BranchId = 9917
    NonExclusiveLimitAlarmType_Retain = 9918
    NonExclusiveLimitAlarmType_EnabledState = 9919
    NonExclusiveLimitAlarmType_EnabledState_Id = 9920
    NonExclusiveLimitAlarmType_EnabledState_Name = 9921
    NonExclusiveLimitAlarmType_EnabledState_Number = 9922
    NonExclusiveLimitAlarmType_EnabledState_EffectiveDisplayName = 9923
    NonExclusiveLimitAlarmType_EnabledState_TransitionTime = 9924
    NonExclusiveLimitAlarmType_EnabledState_EffectiveTransitionTime = 9925
    NonExclusiveLimitAlarmType_EnabledState_TrueState = 9926
    NonExclusiveLimitAlarmType_EnabledState_FalseState = 9927
    NonExclusiveLimitAlarmType_Quality = 9928
    NonExclusiveLimitAlarmType_Quality_SourceTimestamp = 9929
    NonExclusiveLimitAlarmType_LastSeverity = 9930
    NonExclusiveLimitAlarmType_LastSeverity_SourceTimestamp = 9931
    NonExclusiveLimitAlarmType_Comment = 9932
    NonExclusiveLimitAlarmType_Comment_SourceTimestamp = 9933
    NonExclusiveLimitAlarmType_ClientUserId = 9934
    NonExclusiveLimitAlarmType_Enable = 9935
    NonExclusiveLimitAlarmType_Disable = 9936
    NonExclusiveLimitAlarmType_AddComment = 9937
    NonExclusiveLimitAlarmType_AddComment_InputArguments = 9938
    NonExclusiveLimitAlarmType_ConditionRefresh = 9939
    NonExclusiveLimitAlarmType_ConditionRefresh_InputArguments = 9940
    NonExclusiveLimitAlarmType_AckedState = 9941
    NonExclusiveLimitAlarmType_AckedState_Id = 9942
    NonExclusiveLimitAlarmType_AckedState_Name = 9943
    NonExclusiveLimitAlarmType_AckedState_Number = 9944
    NonExclusiveLimitAlarmType_AckedState_EffectiveDisplayName = 9945
    NonExclusiveLimitAlarmType_AckedState_TransitionTime = 9946
    NonExclusiveLimitAlarmType_AckedState_EffectiveTransitionTime = 9947
    NonExclusiveLimitAlarmType_AckedState_TrueState = 9948
    NonExclusiveLimitAlarmType_AckedState_FalseState = 9949
    NonExclusiveLimitAlarmType_ConfirmedState = 9950
    NonExclusiveLimitAlarmType_ConfirmedState_Id = 9951
    NonExclusiveLimitAlarmType_ConfirmedState_Name = 9952
    NonExclusiveLimitAlarmType_ConfirmedState_Number = 9953
    NonExclusiveLimitAlarmType_ConfirmedState_EffectiveDisplayName = 9954
    NonExclusiveLimitAlarmType_ConfirmedState_TransitionTime = 9955
    NonExclusiveLimitAlarmType_ConfirmedState_EffectiveTransitionTime = 9956
    NonExclusiveLimitAlarmType_ConfirmedState_TrueState = 9957
    NonExclusiveLimitAlarmType_ConfirmedState_FalseState = 9958
    NonExclusiveLimitAlarmType_Acknowledge = 9959
    NonExclusiveLimitAlarmType_Acknowledge_InputArguments = 9960
    NonExclusiveLimitAlarmType_Confirm = 9961
    NonExclusiveLimitAlarmType_Confirm_InputArguments = 9962
    NonExclusiveLimitAlarmType_ActiveState = 9963
    NonExclusiveLimitAlarmType_ActiveState_Id = 9964
    NonExclusiveLimitAlarmType_ActiveState_Name = 9965
    NonExclusiveLimitAlarmType_ActiveState_Number = 9966
    NonExclusiveLimitAlarmType_ActiveState_EffectiveDisplayName = 9967
    NonExclusiveLimitAlarmType_ActiveState_TransitionTime = 9968
    NonExclusiveLimitAlarmType_ActiveState_EffectiveTransitionTime = 9969
    NonExclusiveLimitAlarmType_ActiveState_TrueState = 9970
    NonExclusiveLimitAlarmType_ActiveState_FalseState = 9971
    NonExclusiveLimitAlarmType_SuppressedState = 9972
    NonExclusiveLimitAlarmType_SuppressedState_Id = 9973
    NonExclusiveLimitAlarmType_SuppressedState_Name = 9974
    NonExclusiveLimitAlarmType_SuppressedState_Number = 9975
    NonExclusiveLimitAlarmType_SuppressedState_EffectiveDisplayName = 9976
    NonExclusiveLimitAlarmType_SuppressedState_TransitionTime = 9977
    NonExclusiveLimitAlarmType_SuppressedState_EffectiveTransitionTime = 9978
    NonExclusiveLimitAlarmType_SuppressedState_TrueState = 9979
    NonExclusiveLimitAlarmType_SuppressedState_FalseState = 9980
    NonExclusiveLimitAlarmType_ShelvingState = 9981
    NonExclusiveLimitAlarmType_ShelvingState_CurrentState = 9982
    NonExclusiveLimitAlarmType_ShelvingState_CurrentState_Id = 9983
    NonExclusiveLimitAlarmType_ShelvingState_CurrentState_Name = 9984
    NonExclusiveLimitAlarmType_ShelvingState_CurrentState_Number = 9985
    NonExclusiveLimitAlarmType_ShelvingState_CurrentState_EffectiveDisplayName = 9986
    NonExclusiveLimitAlarmType_ShelvingState_LastTransition = 9987
    NonExclusiveLimitAlarmType_ShelvingState_LastTransition_Id = 9988
    NonExclusiveLimitAlarmType_ShelvingState_LastTransition_Name = 9989
    NonExclusiveLimitAlarmType_ShelvingState_LastTransition_Number = 9990
    NonExclusiveLimitAlarmType_ShelvingState_LastTransition_TransitionTime = 9991
    NonExclusiveLimitAlarmType_ShelvingState_UnshelveTime = 9992
    NonExclusiveLimitAlarmType_ShelvingState_Unshelve = 10014
    NonExclusiveLimitAlarmType_ShelvingState_OneShotShelve = 10015
    NonExclusiveLimitAlarmType_ShelvingState_TimedShelve = 10016
    NonExclusiveLimitAlarmType_ShelvingState_TimedShelve_InputArguments = 10017
    NonExclusiveLimitAlarmType_SuppressedOrShelved = 10018
    NonExclusiveLimitAlarmType_MaxTimeShelved = 10019
    NonExclusiveLimitAlarmType_HighHighState = 10020
    NonExclusiveLimitAlarmType_HighHighState_Id = 10021
    NonExclusiveLimitAlarmType_HighHighState_Name = 10022
    NonExclusiveLimitAlarmType_HighHighState_Number = 10023
    NonExclusiveLimitAlarmType_HighHighState_EffectiveDisplayName = 10024
    NonExclusiveLimitAlarmType_HighHighState_TransitionTime = 10025
    NonExclusiveLimitAlarmType_HighHighState_EffectiveTransitionTime = 10026
    NonExclusiveLimitAlarmType_HighHighState_TrueState = 10027
    NonExclusiveLimitAlarmType_HighHighState_FalseState = 10028
    NonExclusiveLimitAlarmType_HighState = 10029
    NonExclusiveLimitAlarmType_HighState_Id = 10030
    NonExclusiveLimitAlarmType_HighState_Name = 10031
    NonExclusiveLimitAlarmType_HighState_Number = 10032
    NonExclusiveLimitAlarmType_HighState_EffectiveDisplayName = 10033
    NonExclusiveLimitAlarmType_HighState_TransitionTime = 10034
    NonExclusiveLimitAlarmType_HighState_EffectiveTransitionTime = 10035
    NonExclusiveLimitAlarmType_HighState_TrueState = 10036
    NonExclusiveLimitAlarmType_HighState_FalseState = 10037
    NonExclusiveLimitAlarmType_LowState = 10038
    NonExclusiveLimitAlarmType_LowState_Id = 10039
    NonExclusiveLimitAlarmType_LowState_Name = 10040
    NonExclusiveLimitAlarmType_LowState_Number = 10041
    NonExclusiveLimitAlarmType_LowState_EffectiveDisplayName = 10042
    NonExclusiveLimitAlarmType_LowState_TransitionTime = 10043
    NonExclusiveLimitAlarmType_LowState_EffectiveTransitionTime = 10044
    NonExclusiveLimitAlarmType_LowState_TrueState = 10045
    NonExclusiveLimitAlarmType_LowState_FalseState = 10046
    NonExclusiveLimitAlarmType_LowLowState = 10047
    NonExclusiveLimitAlarmType_LowLowState_Id = 10048
    NonExclusiveLimitAlarmType_LowLowState_Name = 10049
    NonExclusiveLimitAlarmType_LowLowState_Number = 10050
    NonExclusiveLimitAlarmType_LowLowState_EffectiveDisplayName = 10051
    NonExclusiveLimitAlarmType_LowLowState_TransitionTime = 10052
    NonExclusiveLimitAlarmType_LowLowState_EffectiveTransitionTime = 10053
    NonExclusiveLimitAlarmType_LowLowState_TrueState = 10054
    NonExclusiveLimitAlarmType_LowLowState_FalseState = 10055
    NonExclusiveLimitAlarmType_HighHighLimit = 10056
    NonExclusiveLimitAlarmType_HighLimit = 10057
    NonExclusiveLimitAlarmType_LowLimit = 10058
    NonExclusiveLimitAlarmType_LowLowLimit = 10059
    NonExclusiveLevelAlarmType = 10060
    NonExclusiveLevelAlarmType_EventId = 10061
    NonExclusiveLevelAlarmType_EventType = 10062
    NonExclusiveLevelAlarmType_SourceNode = 10063
    NonExclusiveLevelAlarmType_SourceName = 10064
    NonExclusiveLevelAlarmType_Time = 10065
    NonExclusiveLevelAlarmType_ReceiveTime = 10066
    NonExclusiveLevelAlarmType_LocalTime = 10067
    NonExclusiveLevelAlarmType_Message = 10068
    NonExclusiveLevelAlarmType_Severity = 10069
    NonExclusiveLevelAlarmType_ConditionName = 10070
    NonExclusiveLevelAlarmType_BranchId = 10071
    NonExclusiveLevelAlarmType_Retain = 10072
    NonExclusiveLevelAlarmType_EnabledState = 10073
    NonExclusiveLevelAlarmType_EnabledState_Id = 10074
    NonExclusiveLevelAlarmType_EnabledState_Name = 10075
    NonExclusiveLevelAlarmType_EnabledState_Number = 10076
    NonExclusiveLevelAlarmType_EnabledState_EffectiveDisplayName = 10077
    NonExclusiveLevelAlarmType_EnabledState_TransitionTime = 10078
    NonExclusiveLevelAlarmType_EnabledState_EffectiveTransitionTime = 10079
    NonExclusiveLevelAlarmType_EnabledState_TrueState = 10080
    NonExclusiveLevelAlarmType_EnabledState_FalseState = 10081
    NonExclusiveLevelAlarmType_Quality = 10082
    NonExclusiveLevelAlarmType_Quality_SourceTimestamp = 10083
    NonExclusiveLevelAlarmType_LastSeverity = 10084
    NonExclusiveLevelAlarmType_LastSeverity_SourceTimestamp = 10085
    NonExclusiveLevelAlarmType_Comment = 10086
    NonExclusiveLevelAlarmType_Comment_SourceTimestamp = 10087
    NonExclusiveLevelAlarmType_ClientUserId = 10088
    NonExclusiveLevelAlarmType_Enable = 10089
    NonExclusiveLevelAlarmType_Disable = 10090
    NonExclusiveLevelAlarmType_AddComment = 10091
    NonExclusiveLevelAlarmType_AddComment_InputArguments = 10092
    NonExclusiveLevelAlarmType_ConditionRefresh = 10093
    NonExclusiveLevelAlarmType_ConditionRefresh_InputArguments = 10094
    NonExclusiveLevelAlarmType_AckedState = 10095
    NonExclusiveLevelAlarmType_AckedState_Id = 10096
    NonExclusiveLevelAlarmType_AckedState_Name = 10097
    NonExclusiveLevelAlarmType_AckedState_Number = 10098
    NonExclusiveLevelAlarmType_AckedState_EffectiveDisplayName = 10099
    NonExclusiveLevelAlarmType_AckedState_TransitionTime = 10100
    NonExclusiveLevelAlarmType_AckedState_EffectiveTransitionTime = 10101
    NonExclusiveLevelAlarmType_AckedState_TrueState = 10102
    NonExclusiveLevelAlarmType_AckedState_FalseState = 10103
    NonExclusiveLevelAlarmType_ConfirmedState = 10104
    NonExclusiveLevelAlarmType_ConfirmedState_Id = 10105
    NonExclusiveLevelAlarmType_ConfirmedState_Name = 10106
    NonExclusiveLevelAlarmType_ConfirmedState_Number = 10107
    NonExclusiveLevelAlarmType_ConfirmedState_EffectiveDisplayName = 10108
    NonExclusiveLevelAlarmType_ConfirmedState_TransitionTime = 10109
    NonExclusiveLevelAlarmType_ConfirmedState_EffectiveTransitionTime = 10110
    NonExclusiveLevelAlarmType_ConfirmedState_TrueState = 10111
    NonExclusiveLevelAlarmType_ConfirmedState_FalseState = 10112
    NonExclusiveLevelAlarmType_Acknowledge = 10113
    NonExclusiveLevelAlarmType_Acknowledge_InputArguments = 10114
    NonExclusiveLevelAlarmType_Confirm = 10115
    NonExclusiveLevelAlarmType_Confirm_InputArguments = 10116
    NonExclusiveLevelAlarmType_ActiveState = 10117
    NonExclusiveLevelAlarmType_ActiveState_Id = 10118
    NonExclusiveLevelAlarmType_ActiveState_Name = 10119
    NonExclusiveLevelAlarmType_ActiveState_Number = 10120
    NonExclusiveLevelAlarmType_ActiveState_EffectiveDisplayName = 10121
    NonExclusiveLevelAlarmType_ActiveState_TransitionTime = 10122
    NonExclusiveLevelAlarmType_ActiveState_EffectiveTransitionTime = 10123
    NonExclusiveLevelAlarmType_ActiveState_TrueState = 10124
    NonExclusiveLevelAlarmType_ActiveState_FalseState = 10125
    NonExclusiveLevelAlarmType_SuppressedState = 10126
    NonExclusiveLevelAlarmType_SuppressedState_Id = 10127
    NonExclusiveLevelAlarmType_SuppressedState_Name = 10128
    NonExclusiveLevelAlarmType_SuppressedState_Number = 10129
    NonExclusiveLevelAlarmType_SuppressedState_EffectiveDisplayName = 10130
    NonExclusiveLevelAlarmType_SuppressedState_TransitionTime = 10131
    NonExclusiveLevelAlarmType_SuppressedState_EffectiveTransitionTime = 10132
    NonExclusiveLevelAlarmType_SuppressedState_TrueState = 10133
    NonExclusiveLevelAlarmType_SuppressedState_FalseState = 10134
    NonExclusiveLevelAlarmType_ShelvingState = 10135
    NonExclusiveLevelAlarmType_ShelvingState_CurrentState = 10136
    NonExclusiveLevelAlarmType_ShelvingState_CurrentState_Id = 10137
    NonExclusiveLevelAlarmType_ShelvingState_CurrentState_Name = 10138
    NonExclusiveLevelAlarmType_ShelvingState_CurrentState_Number = 10139
    NonExclusiveLevelAlarmType_ShelvingState_CurrentState_EffectiveDisplayName = 10140
    NonExclusiveLevelAlarmType_ShelvingState_LastTransition = 10141
    NonExclusiveLevelAlarmType_ShelvingState_LastTransition_Id = 10142
    NonExclusiveLevelAlarmType_ShelvingState_LastTransition_Name = 10143
    NonExclusiveLevelAlarmType_ShelvingState_LastTransition_Number = 10144
    NonExclusiveLevelAlarmType_ShelvingState_LastTransition_TransitionTime = 10145
    NonExclusiveLevelAlarmType_ShelvingState_UnshelveTime = 10146
    NonExclusiveLevelAlarmType_ShelvingState_Unshelve = 10168
    NonExclusiveLevelAlarmType_ShelvingState_OneShotShelve = 10169
    NonExclusiveLevelAlarmType_ShelvingState_TimedShelve = 10170
    NonExclusiveLevelAlarmType_ShelvingState_TimedShelve_InputArguments = 10171
    NonExclusiveLevelAlarmType_SuppressedOrShelved = 10172
    NonExclusiveLevelAlarmType_MaxTimeShelved = 10173
    NonExclusiveLevelAlarmType_HighHighState = 10174
    NonExclusiveLevelAlarmType_HighHighState_Id = 10175
    NonExclusiveLevelAlarmType_HighHighState_Name = 10176
    NonExclusiveLevelAlarmType_HighHighState_Number = 10177
    NonExclusiveLevelAlarmType_HighHighState_EffectiveDisplayName = 10178
    NonExclusiveLevelAlarmType_HighHighState_TransitionTime = 10179
    NonExclusiveLevelAlarmType_HighHighState_EffectiveTransitionTime = 10180
    NonExclusiveLevelAlarmType_HighHighState_TrueState = 10181
    NonExclusiveLevelAlarmType_HighHighState_FalseState = 10182
    NonExclusiveLevelAlarmType_HighState = 10183
    NonExclusiveLevelAlarmType_HighState_Id = 10184
    NonExclusiveLevelAlarmType_HighState_Name = 10185
    NonExclusiveLevelAlarmType_HighState_Number = 10186
    NonExclusiveLevelAlarmType_HighState_EffectiveDisplayName = 10187
    NonExclusiveLevelAlarmType_HighState_TransitionTime = 10188
    NonExclusiveLevelAlarmType_HighState_EffectiveTransitionTime = 10189
    NonExclusiveLevelAlarmType_HighState_TrueState = 10190
    NonExclusiveLevelAlarmType_HighState_FalseState = 10191
    NonExclusiveLevelAlarmType_LowState = 10192
    NonExclusiveLevelAlarmType_LowState_Id = 10193
    NonExclusiveLevelAlarmType_LowState_Name = 10194
    NonExclusiveLevelAlarmType_LowState_Number = 10195
    NonExclusiveLevelAlarmType_LowState_EffectiveDisplayName = 10196
    NonExclusiveLevelAlarmType_LowState_TransitionTime = 10197
    NonExclusiveLevelAlarmType_LowState_EffectiveTransitionTime = 10198
    NonExclusiveLevelAlarmType_LowState_TrueState = 10199
    NonExclusiveLevelAlarmType_LowState_FalseState = 10200
    NonExclusiveLevelAlarmType_LowLowState = 10201
    NonExclusiveLevelAlarmType_LowLowState_Id = 10202
    NonExclusiveLevelAlarmType_LowLowState_Name = 10203
    NonExclusiveLevelAlarmType_LowLowState_Number = 10204
    NonExclusiveLevelAlarmType_LowLowState_EffectiveDisplayName = 10205
    NonExclusiveLevelAlarmType_LowLowState_TransitionTime = 10206
    NonExclusiveLevelAlarmType_LowLowState_EffectiveTransitionTime = 10207
    NonExclusiveLevelAlarmType_LowLowState_TrueState = 10208
    NonExclusiveLevelAlarmType_LowLowState_FalseState = 10209
    NonExclusiveLevelAlarmType_HighHighLimit = 10210
    NonExclusiveLevelAlarmType_HighLimit = 10211
    NonExclusiveLevelAlarmType_LowLimit = 10212
    NonExclusiveLevelAlarmType_LowLowLimit = 10213
    NonExclusiveRateOfChangeAlarmType = 10214
    NonExclusiveRateOfChangeAlarmType_EventId = 10215
    NonExclusiveRateOfChangeAlarmType_EventType = 10216
    NonExclusiveRateOfChangeAlarmType_SourceNode = 10217
    NonExclusiveRateOfChangeAlarmType_SourceName = 10218
    NonExclusiveRateOfChangeAlarmType_Time = 10219
    NonExclusiveRateOfChangeAlarmType_ReceiveTime = 10220
    NonExclusiveRateOfChangeAlarmType_LocalTime = 10221
    NonExclusiveRateOfChangeAlarmType_Message = 10222
    NonExclusiveRateOfChangeAlarmType_Severity = 10223
    NonExclusiveRateOfChangeAlarmType_ConditionName = 10224
    NonExclusiveRateOfChangeAlarmType_BranchId = 10225
    NonExclusiveRateOfChangeAlarmType_Retain = 10226
    NonExclusiveRateOfChangeAlarmType_EnabledState = 10227
    NonExclusiveRateOfChangeAlarmType_EnabledState_Id = 10228
    NonExclusiveRateOfChangeAlarmType_EnabledState_Name = 10229
    NonExclusiveRateOfChangeAlarmType_EnabledState_Number = 10230
    NonExclusiveRateOfChangeAlarmType_EnabledState_EffectiveDisplayName = 10231
    NonExclusiveRateOfChangeAlarmType_EnabledState_TransitionTime = 10232
    NonExclusiveRateOfChangeAlarmType_EnabledState_EffectiveTransitionTime = 10233
    NonExclusiveRateOfChangeAlarmType_EnabledState_TrueState = 10234
    NonExclusiveRateOfChangeAlarmType_EnabledState_FalseState = 10235
    NonExclusiveRateOfChangeAlarmType_Quality = 10236
    NonExclusiveRateOfChangeAlarmType_Quality_SourceTimestamp = 10237
    NonExclusiveRateOfChangeAlarmType_LastSeverity = 10238
    NonExclusiveRateOfChangeAlarmType_LastSeverity_SourceTimestamp = 10239
    NonExclusiveRateOfChangeAlarmType_Comment = 10240
    NonExclusiveRateOfChangeAlarmType_Comment_SourceTimestamp = 10241
    NonExclusiveRateOfChangeAlarmType_ClientUserId = 10242
    NonExclusiveRateOfChangeAlarmType_Enable = 10243
    NonExclusiveRateOfChangeAlarmType_Disable = 10244
    NonExclusiveRateOfChangeAlarmType_AddComment = 10245
    NonExclusiveRateOfChangeAlarmType_AddComment_InputArguments = 10246
    NonExclusiveRateOfChangeAlarmType_ConditionRefresh = 10247
    NonExclusiveRateOfChangeAlarmType_ConditionRefresh_InputArguments = 10248
    NonExclusiveRateOfChangeAlarmType_AckedState = 10249
    NonExclusiveRateOfChangeAlarmType_AckedState_Id = 10250
    NonExclusiveRateOfChangeAlarmType_AckedState_Name = 10251
    NonExclusiveRateOfChangeAlarmType_AckedState_Number = 10252
    NonExclusiveRateOfChangeAlarmType_AckedState_EffectiveDisplayName = 10253
    NonExclusiveRateOfChangeAlarmType_AckedState_TransitionTime = 10254
    NonExclusiveRateOfChangeAlarmType_AckedState_EffectiveTransitionTime = 10255
    NonExclusiveRateOfChangeAlarmType_AckedState_TrueState = 10256
    NonExclusiveRateOfChangeAlarmType_AckedState_FalseState = 10257
    NonExclusiveRateOfChangeAlarmType_ConfirmedState = 10258
    NonExclusiveRateOfChangeAlarmType_ConfirmedState_Id = 10259
    NonExclusiveRateOfChangeAlarmType_ConfirmedState_Name = 10260
    NonExclusiveRateOfChangeAlarmType_ConfirmedState_Number = 10261
    NonExclusiveRateOfChangeAlarmType_ConfirmedState_EffectiveDisplayName = 10262
    NonExclusiveRateOfChangeAlarmType_ConfirmedState_TransitionTime = 10263
    NonExclusiveRateOfChangeAlarmType_ConfirmedState_EffectiveTransitionTime = 10264
    NonExclusiveRateOfChangeAlarmType_ConfirmedState_TrueState = 10265
    NonExclusiveRateOfChangeAlarmType_ConfirmedState_FalseState = 10266
    NonExclusiveRateOfChangeAlarmType_Acknowledge = 10267
    NonExclusiveRateOfChangeAlarmType_Acknowledge_InputArguments = 10268
    NonExclusiveRateOfChangeAlarmType_Confirm = 10269
    NonExclusiveRateOfChangeAlarmType_Confirm_InputArguments = 10270
    NonExclusiveRateOfChangeAlarmType_ActiveState = 10271
    NonExclusiveRateOfChangeAlarmType_ActiveState_Id = 10272
    NonExclusiveRateOfChangeAlarmType_ActiveState_Name = 10273
    NonExclusiveRateOfChangeAlarmType_ActiveState_Number = 10274
    NonExclusiveRateOfChangeAlarmType_ActiveState_EffectiveDisplayName = 10275
    NonExclusiveRateOfChangeAlarmType_ActiveState_TransitionTime = 10276
    NonExclusiveRateOfChangeAlarmType_ActiveState_EffectiveTransitionTime = 10277
    NonExclusiveRateOfChangeAlarmType_ActiveState_TrueState = 10278
    NonExclusiveRateOfChangeAlarmType_ActiveState_FalseState = 10279
    NonExclusiveRateOfChangeAlarmType_SuppressedState = 10280
    NonExclusiveRateOfChangeAlarmType_SuppressedState_Id = 10281
    NonExclusiveRateOfChangeAlarmType_SuppressedState_Name = 10282
    NonExclusiveRateOfChangeAlarmType_SuppressedState_Number = 10283
    NonExclusiveRateOfChangeAlarmType_SuppressedState_EffectiveDisplayName = 10284
    NonExclusiveRateOfChangeAlarmType_SuppressedState_TransitionTime = 10285
    NonExclusiveRateOfChangeAlarmType_SuppressedState_EffectiveTransitionTime = 10286
    NonExclusiveRateOfChangeAlarmType_SuppressedState_TrueState = 10287
    NonExclusiveRateOfChangeAlarmType_SuppressedState_FalseState = 10288
    NonExclusiveRateOfChangeAlarmType_ShelvingState = 10289
    NonExclusiveRateOfChangeAlarmType_ShelvingState_CurrentState = 10290
    NonExclusiveRateOfChangeAlarmType_ShelvingState_CurrentState_Id = 10291
    NonExclusiveRateOfChangeAlarmType_ShelvingState_CurrentState_Name = 10292
    NonExclusiveRateOfChangeAlarmType_ShelvingState_CurrentState_Number = 10293
    NonExclusiveRateOfChangeAlarmType_ShelvingState_CurrentState_EffectiveDisplayName = 10294
    NonExclusiveRateOfChangeAlarmType_ShelvingState_LastTransition = 10295
    NonExclusiveRateOfChangeAlarmType_ShelvingState_LastTransition_Id = 10296
    NonExclusiveRateOfChangeAlarmType_ShelvingState_LastTransition_Name = 10297
    NonExclusiveRateOfChangeAlarmType_ShelvingState_LastTransition_Number = 10298
    NonExclusiveRateOfChangeAlarmType_ShelvingState_LastTransition_TransitionTime = 10299
    NonExclusiveRateOfChangeAlarmType_ShelvingState_UnshelveTime = 10300
    NonExclusiveRateOfChangeAlarmType_ShelvingState_Unshelve = 10322
    NonExclusiveRateOfChangeAlarmType_ShelvingState_OneShotShelve = 10323
    NonExclusiveRateOfChangeAlarmType_ShelvingState_TimedShelve = 10324
    NonExclusiveRateOfChangeAlarmType_ShelvingState_TimedShelve_InputArguments = 10325
    NonExclusiveRateOfChangeAlarmType_SuppressedOrShelved = 10326
    NonExclusiveRateOfChangeAlarmType_MaxTimeShelved = 10327
    NonExclusiveRateOfChangeAlarmType_HighHighState = 10328
    NonExclusiveRateOfChangeAlarmType_HighHighState_Id = 10329
    NonExclusiveRateOfChangeAlarmType_HighHighState_Name = 10330
    NonExclusiveRateOfChangeAlarmType_HighHighState_Number = 10331
    NonExclusiveRateOfChangeAlarmType_HighHighState_EffectiveDisplayName = 10332
    NonExclusiveRateOfChangeAlarmType_HighHighState_TransitionTime = 10333
    NonExclusiveRateOfChangeAlarmType_HighHighState_EffectiveTransitionTime = 10334
    NonExclusiveRateOfChangeAlarmType_HighHighState_TrueState = 10335
    NonExclusiveRateOfChangeAlarmType_HighHighState_FalseState = 10336
    NonExclusiveRateOfChangeAlarmType_HighState = 10337
    NonExclusiveRateOfChangeAlarmType_HighState_Id = 10338
    NonExclusiveRateOfChangeAlarmType_HighState_Name = 10339
    NonExclusiveRateOfChangeAlarmType_HighState_Number = 10340
    NonExclusiveRateOfChangeAlarmType_HighState_EffectiveDisplayName = 10341
    NonExclusiveRateOfChangeAlarmType_HighState_TransitionTime = 10342
    NonExclusiveRateOfChangeAlarmType_HighState_EffectiveTransitionTime = 10343
    NonExclusiveRateOfChangeAlarmType_HighState_TrueState = 10344
    NonExclusiveRateOfChangeAlarmType_HighState_FalseState = 10345
    NonExclusiveRateOfChangeAlarmType_LowState = 10346
    NonExclusiveRateOfChangeAlarmType_LowState_Id = 10347
    NonExclusiveRateOfChangeAlarmType_LowState_Name = 10348
    NonExclusiveRateOfChangeAlarmType_LowState_Number = 10349
    NonExclusiveRateOfChangeAlarmType_LowState_EffectiveDisplayName = 10350
    NonExclusiveRateOfChangeAlarmType_LowState_TransitionTime = 10351
    NonExclusiveRateOfChangeAlarmType_LowState_EffectiveTransitionTime = 10352
    NonExclusiveRateOfChangeAlarmType_LowState_TrueState = 10353
    NonExclusiveRateOfChangeAlarmType_LowState_FalseState = 10354
    NonExclusiveRateOfChangeAlarmType_LowLowState = 10355
    NonExclusiveRateOfChangeAlarmType_LowLowState_Id = 10356
    NonExclusiveRateOfChangeAlarmType_LowLowState_Name = 10357
    NonExclusiveRateOfChangeAlarmType_LowLowState_Number = 10358
    NonExclusiveRateOfChangeAlarmType_LowLowState_EffectiveDisplayName = 10359
    NonExclusiveRateOfChangeAlarmType_LowLowState_TransitionTime = 10360
    NonExclusiveRateOfChangeAlarmType_LowLowState_EffectiveTransitionTime = 10361
    NonExclusiveRateOfChangeAlarmType_LowLowState_TrueState = 10362
    NonExclusiveRateOfChangeAlarmType_LowLowState_FalseState = 10363
    NonExclusiveRateOfChangeAlarmType_HighHighLimit = 10364
    NonExclusiveRateOfChangeAlarmType_HighLimit = 10365
    NonExclusiveRateOfChangeAlarmType_LowLimit = 10366
    NonExclusiveRateOfChangeAlarmType_LowLowLimit = 10367
    NonExclusiveDeviationAlarmType = 10368
    NonExclusiveDeviationAlarmType_EventId = 10369
    NonExclusiveDeviationAlarmType_EventType = 10370
    NonExclusiveDeviationAlarmType_SourceNode = 10371
    NonExclusiveDeviationAlarmType_SourceName = 10372
    NonExclusiveDeviationAlarmType_Time = 10373
    NonExclusiveDeviationAlarmType_ReceiveTime = 10374
    NonExclusiveDeviationAlarmType_LocalTime = 10375
    NonExclusiveDeviationAlarmType_Message = 10376
    NonExclusiveDeviationAlarmType_Severity = 10377
    NonExclusiveDeviationAlarmType_ConditionName = 10378
    NonExclusiveDeviationAlarmType_BranchId = 10379
    NonExclusiveDeviationAlarmType_Retain = 10380
    NonExclusiveDeviationAlarmType_EnabledState = 10381
    NonExclusiveDeviationAlarmType_EnabledState_Id = 10382
    NonExclusiveDeviationAlarmType_EnabledState_Name = 10383
    NonExclusiveDeviationAlarmType_EnabledState_Number = 10384
    NonExclusiveDeviationAlarmType_EnabledState_EffectiveDisplayName = 10385
    NonExclusiveDeviationAlarmType_EnabledState_TransitionTime = 10386
    NonExclusiveDeviationAlarmType_EnabledState_EffectiveTransitionTime = 10387
    NonExclusiveDeviationAlarmType_EnabledState_TrueState = 10388
    NonExclusiveDeviationAlarmType_EnabledState_FalseState = 10389
    NonExclusiveDeviationAlarmType_Quality = 10390
    NonExclusiveDeviationAlarmType_Quality_SourceTimestamp = 10391
    NonExclusiveDeviationAlarmType_LastSeverity = 10392
    NonExclusiveDeviationAlarmType_LastSeverity_SourceTimestamp = 10393
    NonExclusiveDeviationAlarmType_Comment = 10394
    NonExclusiveDeviationAlarmType_Comment_SourceTimestamp = 10395
    NonExclusiveDeviationAlarmType_ClientUserId = 10396
    NonExclusiveDeviationAlarmType_Enable = 10397
    NonExclusiveDeviationAlarmType_Disable = 10398
    NonExclusiveDeviationAlarmType_AddComment = 10399
    NonExclusiveDeviationAlarmType_AddComment_InputArguments = 10400
    NonExclusiveDeviationAlarmType_ConditionRefresh = 10401
    NonExclusiveDeviationAlarmType_ConditionRefresh_InputArguments = 10402
    NonExclusiveDeviationAlarmType_AckedState = 10403
    NonExclusiveDeviationAlarmType_AckedState_Id = 10404
    NonExclusiveDeviationAlarmType_AckedState_Name = 10405
    NonExclusiveDeviationAlarmType_AckedState_Number = 10406
    NonExclusiveDeviationAlarmType_AckedState_EffectiveDisplayName = 10407
    NonExclusiveDeviationAlarmType_AckedState_TransitionTime = 10408
    NonExclusiveDeviationAlarmType_AckedState_EffectiveTransitionTime = 10409
    NonExclusiveDeviationAlarmType_AckedState_TrueState = 10410
    NonExclusiveDeviationAlarmType_AckedState_FalseState = 10411
    NonExclusiveDeviationAlarmType_ConfirmedState = 10412
    NonExclusiveDeviationAlarmType_ConfirmedState_Id = 10413
    NonExclusiveDeviationAlarmType_ConfirmedState_Name = 10414
    NonExclusiveDeviationAlarmType_ConfirmedState_Number = 10415
    NonExclusiveDeviationAlarmType_ConfirmedState_EffectiveDisplayName = 10416
    NonExclusiveDeviationAlarmType_ConfirmedState_TransitionTime = 10417
    NonExclusiveDeviationAlarmType_ConfirmedState_EffectiveTransitionTime = 10418
    NonExclusiveDeviationAlarmType_ConfirmedState_TrueState = 10419
    NonExclusiveDeviationAlarmType_ConfirmedState_FalseState = 10420
    NonExclusiveDeviationAlarmType_Acknowledge = 10421
    NonExclusiveDeviationAlarmType_Acknowledge_InputArguments = 10422
    NonExclusiveDeviationAlarmType_Confirm = 10423
    NonExclusiveDeviationAlarmType_Confirm_InputArguments = 10424
    NonExclusiveDeviationAlarmType_ActiveState = 10425
    NonExclusiveDeviationAlarmType_ActiveState_Id = 10426
    NonExclusiveDeviationAlarmType_ActiveState_Name = 10427
    NonExclusiveDeviationAlarmType_ActiveState_Number = 10428
    NonExclusiveDeviationAlarmType_ActiveState_EffectiveDisplayName = 10429
    NonExclusiveDeviationAlarmType_ActiveState_TransitionTime = 10430
    NonExclusiveDeviationAlarmType_ActiveState_EffectiveTransitionTime = 10431
    NonExclusiveDeviationAlarmType_ActiveState_TrueState = 10432
    NonExclusiveDeviationAlarmType_ActiveState_FalseState = 10433
    NonExclusiveDeviationAlarmType_SuppressedState = 10434
    NonExclusiveDeviationAlarmType_SuppressedState_Id = 10435
    NonExclusiveDeviationAlarmType_SuppressedState_Name = 10436
    NonExclusiveDeviationAlarmType_SuppressedState_Number = 10437
    NonExclusiveDeviationAlarmType_SuppressedState_EffectiveDisplayName = 10438
    NonExclusiveDeviationAlarmType_SuppressedState_TransitionTime = 10439
    NonExclusiveDeviationAlarmType_SuppressedState_EffectiveTransitionTime = 10440
    NonExclusiveDeviationAlarmType_SuppressedState_TrueState = 10441
    NonExclusiveDeviationAlarmType_SuppressedState_FalseState = 10442
    NonExclusiveDeviationAlarmType_ShelvingState = 10443
    NonExclusiveDeviationAlarmType_ShelvingState_CurrentState = 10444
    NonExclusiveDeviationAlarmType_ShelvingState_CurrentState_Id = 10445
    NonExclusiveDeviationAlarmType_ShelvingState_CurrentState_Name = 10446
    NonExclusiveDeviationAlarmType_ShelvingState_CurrentState_Number = 10447
    NonExclusiveDeviationAlarmType_ShelvingState_CurrentState_EffectiveDisplayName = 10448
    NonExclusiveDeviationAlarmType_ShelvingState_LastTransition = 10449
    NonExclusiveDeviationAlarmType_ShelvingState_LastTransition_Id = 10450
    NonExclusiveDeviationAlarmType_ShelvingState_LastTransition_Name = 10451
    NonExclusiveDeviationAlarmType_ShelvingState_LastTransition_Number = 10452
    NonExclusiveDeviationAlarmType_ShelvingState_LastTransition_TransitionTime = 10453
    NonExclusiveDeviationAlarmType_ShelvingState_UnshelveTime = 10454
    NonExclusiveDeviationAlarmType_ShelvingState_Unshelve = 10476
    NonExclusiveDeviationAlarmType_ShelvingState_OneShotShelve = 10477
    NonExclusiveDeviationAlarmType_ShelvingState_TimedShelve = 10478
    NonExclusiveDeviationAlarmType_ShelvingState_TimedShelve_InputArguments = 10479
    NonExclusiveDeviationAlarmType_SuppressedOrShelved = 10480
    NonExclusiveDeviationAlarmType_MaxTimeShelved = 10481
    NonExclusiveDeviationAlarmType_HighHighState = 10482
    NonExclusiveDeviationAlarmType_HighHighState_Id = 10483
    NonExclusiveDeviationAlarmType_HighHighState_Name = 10484
    NonExclusiveDeviationAlarmType_HighHighState_Number = 10485
    NonExclusiveDeviationAlarmType_HighHighState_EffectiveDisplayName = 10486
    NonExclusiveDeviationAlarmType_HighHighState_TransitionTime = 10487
    NonExclusiveDeviationAlarmType_HighHighState_EffectiveTransitionTime = 10488
    NonExclusiveDeviationAlarmType_HighHighState_TrueState = 10489
    NonExclusiveDeviationAlarmType_HighHighState_FalseState = 10490
    NonExclusiveDeviationAlarmType_HighState = 10491
    NonExclusiveDeviationAlarmType_HighState_Id = 10492
    NonExclusiveDeviationAlarmType_HighState_Name = 10493
    NonExclusiveDeviationAlarmType_HighState_Number = 10494
    NonExclusiveDeviationAlarmType_HighState_EffectiveDisplayName = 10495
    NonExclusiveDeviationAlarmType_HighState_TransitionTime = 10496
    NonExclusiveDeviationAlarmType_HighState_EffectiveTransitionTime = 10497
    NonExclusiveDeviationAlarmType_HighState_TrueState = 10498
    NonExclusiveDeviationAlarmType_HighState_FalseState = 10499
    NonExclusiveDeviationAlarmType_LowState = 10500
    NonExclusiveDeviationAlarmType_LowState_Id = 10501
    NonExclusiveDeviationAlarmType_LowState_Name = 10502
    NonExclusiveDeviationAlarmType_LowState_Number = 10503
    NonExclusiveDeviationAlarmType_LowState_EffectiveDisplayName = 10504
    NonExclusiveDeviationAlarmType_LowState_TransitionTime = 10505
    NonExclusiveDeviationAlarmType_LowState_EffectiveTransitionTime = 10506
    NonExclusiveDeviationAlarmType_LowState_TrueState = 10507
    NonExclusiveDeviationAlarmType_LowState_FalseState = 10508
    NonExclusiveDeviationAlarmType_LowLowState = 10509
    NonExclusiveDeviationAlarmType_LowLowState_Id = 10510
    NonExclusiveDeviationAlarmType_LowLowState_Name = 10511
    NonExclusiveDeviationAlarmType_LowLowState_Number = 10512
    NonExclusiveDeviationAlarmType_LowLowState_EffectiveDisplayName = 10513
    NonExclusiveDeviationAlarmType_LowLowState_TransitionTime = 10514
    NonExclusiveDeviationAlarmType_LowLowState_EffectiveTransitionTime = 10515
    NonExclusiveDeviationAlarmType_LowLowState_TrueState = 10516
    NonExclusiveDeviationAlarmType_LowLowState_FalseState = 10517
    NonExclusiveDeviationAlarmType_HighHighLimit = 10518
    NonExclusiveDeviationAlarmType_HighLimit = 10519
    NonExclusiveDeviationAlarmType_LowLimit = 10520
    NonExclusiveDeviationAlarmType_LowLowLimit = 10521
    NonExclusiveDeviationAlarmType_SetpointNode = 10522
    DiscreteAlarmType = 10523
    DiscreteAlarmType_EventId = 10524
    DiscreteAlarmType_EventType = 10525
    DiscreteAlarmType_SourceNode = 10526
    DiscreteAlarmType_SourceName = 10527
    DiscreteAlarmType_Time = 10528
    DiscreteAlarmType_ReceiveTime = 10529
    DiscreteAlarmType_LocalTime = 10530
    DiscreteAlarmType_Message = 10531
    DiscreteAlarmType_Severity = 10532
    DiscreteAlarmType_ConditionName = 10533
    DiscreteAlarmType_BranchId = 10534
    DiscreteAlarmType_Retain = 10535
    DiscreteAlarmType_EnabledState = 10536
    DiscreteAlarmType_EnabledState_Id = 10537
    DiscreteAlarmType_EnabledState_Name = 10538
    DiscreteAlarmType_EnabledState_Number = 10539
    DiscreteAlarmType_EnabledState_EffectiveDisplayName = 10540
    DiscreteAlarmType_EnabledState_TransitionTime = 10541
    DiscreteAlarmType_EnabledState_EffectiveTransitionTime = 10542
    DiscreteAlarmType_EnabledState_TrueState = 10543
    DiscreteAlarmType_EnabledState_FalseState = 10544
    DiscreteAlarmType_Quality = 10545
    DiscreteAlarmType_Quality_SourceTimestamp = 10546
    DiscreteAlarmType_LastSeverity = 10547
    DiscreteAlarmType_LastSeverity_SourceTimestamp = 10548
    DiscreteAlarmType_Comment = 10549
    DiscreteAlarmType_Comment_SourceTimestamp = 10550
    DiscreteAlarmType_ClientUserId = 10551
    DiscreteAlarmType_Enable = 10552
    DiscreteAlarmType_Disable = 10553
    DiscreteAlarmType_AddComment = 10554
    DiscreteAlarmType_AddComment_InputArguments = 10555
    DiscreteAlarmType_ConditionRefresh = 10556
    DiscreteAlarmType_ConditionRefresh_InputArguments = 10557
    DiscreteAlarmType_AckedState = 10558
    DiscreteAlarmType_AckedState_Id = 10559
    DiscreteAlarmType_AckedState_Name = 10560
    DiscreteAlarmType_AckedState_Number = 10561
    DiscreteAlarmType_AckedState_EffectiveDisplayName = 10562
    DiscreteAlarmType_AckedState_TransitionTime = 10563
    DiscreteAlarmType_AckedState_EffectiveTransitionTime = 10564
    DiscreteAlarmType_AckedState_TrueState = 10565
    DiscreteAlarmType_AckedState_FalseState = 10566
    DiscreteAlarmType_ConfirmedState = 10567
    DiscreteAlarmType_ConfirmedState_Id = 10568
    DiscreteAlarmType_ConfirmedState_Name = 10569
    DiscreteAlarmType_ConfirmedState_Number = 10570
    DiscreteAlarmType_ConfirmedState_EffectiveDisplayName = 10571
    DiscreteAlarmType_ConfirmedState_TransitionTime = 10572
    DiscreteAlarmType_ConfirmedState_EffectiveTransitionTime = 10573
    DiscreteAlarmType_ConfirmedState_TrueState = 10574
    DiscreteAlarmType_ConfirmedState_FalseState = 10575
    DiscreteAlarmType_Acknowledge = 10576
    DiscreteAlarmType_Acknowledge_InputArguments = 10577
    DiscreteAlarmType_Confirm = 10578
    DiscreteAlarmType_Confirm_InputArguments = 10579
    DiscreteAlarmType_ActiveState = 10580
    DiscreteAlarmType_ActiveState_Id = 10581
    DiscreteAlarmType_ActiveState_Name = 10582
    DiscreteAlarmType_ActiveState_Number = 10583
    DiscreteAlarmType_ActiveState_EffectiveDisplayName = 10584
    DiscreteAlarmType_ActiveState_TransitionTime = 10585
    DiscreteAlarmType_ActiveState_EffectiveTransitionTime = 10586
    DiscreteAlarmType_ActiveState_TrueState = 10587
    DiscreteAlarmType_ActiveState_FalseState = 10588
    DiscreteAlarmType_SuppressedState = 10589
    DiscreteAlarmType_SuppressedState_Id = 10590
    DiscreteAlarmType_SuppressedState_Name = 10591
    DiscreteAlarmType_SuppressedState_Number = 10592
    DiscreteAlarmType_SuppressedState_EffectiveDisplayName = 10593
    DiscreteAlarmType_SuppressedState_TransitionTime = 10594
    DiscreteAlarmType_SuppressedState_EffectiveTransitionTime = 10595
    DiscreteAlarmType_SuppressedState_TrueState = 10596
    DiscreteAlarmType_SuppressedState_FalseState = 10597
    DiscreteAlarmType_ShelvingState = 10598
    DiscreteAlarmType_ShelvingState_CurrentState = 10599
    DiscreteAlarmType_ShelvingState_CurrentState_Id = 10600
    DiscreteAlarmType_ShelvingState_CurrentState_Name = 10601
    DiscreteAlarmType_ShelvingState_CurrentState_Number = 10602
    DiscreteAlarmType_ShelvingState_CurrentState_EffectiveDisplayName = 10603
    DiscreteAlarmType_ShelvingState_LastTransition = 10604
    DiscreteAlarmType_ShelvingState_LastTransition_Id = 10605
    DiscreteAlarmType_ShelvingState_LastTransition_Name = 10606
    DiscreteAlarmType_ShelvingState_LastTransition_Number = 10607
    DiscreteAlarmType_ShelvingState_LastTransition_TransitionTime = 10608
    DiscreteAlarmType_ShelvingState_UnshelveTime = 10609
    DiscreteAlarmType_ShelvingState_Unshelve = 10631
    DiscreteAlarmType_ShelvingState_OneShotShelve = 10632
    DiscreteAlarmType_ShelvingState_TimedShelve = 10633
    DiscreteAlarmType_ShelvingState_TimedShelve_InputArguments = 10634
    DiscreteAlarmType_SuppressedOrShelved = 10635
    DiscreteAlarmType_MaxTimeShelved = 10636
    OffNormalAlarmType = 10637
    OffNormalAlarmType_EventId = 10638
    OffNormalAlarmType_EventType = 10639
    OffNormalAlarmType_SourceNode = 10640
    OffNormalAlarmType_SourceName = 10641
    OffNormalAlarmType_Time = 10642
    OffNormalAlarmType_ReceiveTime = 10643
    OffNormalAlarmType_LocalTime = 10644
    OffNormalAlarmType_Message = 10645
    OffNormalAlarmType_Severity = 10646
    OffNormalAlarmType_ConditionName = 10647
    OffNormalAlarmType_BranchId = 10648
    OffNormalAlarmType_Retain = 10649
    OffNormalAlarmType_EnabledState = 10650
    OffNormalAlarmType_EnabledState_Id = 10651
    OffNormalAlarmType_EnabledState_Name = 10652
    OffNormalAlarmType_EnabledState_Number = 10653
    OffNormalAlarmType_EnabledState_EffectiveDisplayName = 10654
    OffNormalAlarmType_EnabledState_TransitionTime = 10655
    OffNormalAlarmType_EnabledState_EffectiveTransitionTime = 10656
    OffNormalAlarmType_EnabledState_TrueState = 10657
    OffNormalAlarmType_EnabledState_FalseState = 10658
    OffNormalAlarmType_Quality = 10659
    OffNormalAlarmType_Quality_SourceTimestamp = 10660
    OffNormalAlarmType_LastSeverity = 10661
    OffNormalAlarmType_LastSeverity_SourceTimestamp = 10662
    OffNormalAlarmType_Comment = 10663
    OffNormalAlarmType_Comment_SourceTimestamp = 10664
    OffNormalAlarmType_ClientUserId = 10665
    OffNormalAlarmType_Enable = 10666
    OffNormalAlarmType_Disable = 10667
    OffNormalAlarmType_AddComment = 10668
    OffNormalAlarmType_AddComment_InputArguments = 10669
    OffNormalAlarmType_ConditionRefresh = 10670
    OffNormalAlarmType_ConditionRefresh_InputArguments = 10671
    OffNormalAlarmType_AckedState = 10672
    OffNormalAlarmType_AckedState_Id = 10673
    OffNormalAlarmType_AckedState_Name = 10674
    OffNormalAlarmType_AckedState_Number = 10675
    OffNormalAlarmType_AckedState_EffectiveDisplayName = 10676
    OffNormalAlarmType_AckedState_TransitionTime = 10677
    OffNormalAlarmType_AckedState_EffectiveTransitionTime = 10678
    OffNormalAlarmType_AckedState_TrueState = 10679
    OffNormalAlarmType_AckedState_FalseState = 10680
    OffNormalAlarmType_ConfirmedState = 10681
    OffNormalAlarmType_ConfirmedState_Id = 10682
    OffNormalAlarmType_ConfirmedState_Name = 10683
    OffNormalAlarmType_ConfirmedState_Number = 10684
    OffNormalAlarmType_ConfirmedState_EffectiveDisplayName = 10685
    OffNormalAlarmType_ConfirmedState_TransitionTime = 10686
    OffNormalAlarmType_ConfirmedState_EffectiveTransitionTime = 10687
    OffNormalAlarmType_ConfirmedState_TrueState = 10688
    OffNormalAlarmType_ConfirmedState_FalseState = 10689
    OffNormalAlarmType_Acknowledge = 10690
    OffNormalAlarmType_Acknowledge_InputArguments = 10691
    OffNormalAlarmType_Confirm = 10692
    OffNormalAlarmType_Confirm_InputArguments = 10693
    OffNormalAlarmType_ActiveState = 10694
    OffNormalAlarmType_ActiveState_Id = 10695
    OffNormalAlarmType_ActiveState_Name = 10696
    OffNormalAlarmType_ActiveState_Number = 10697
    OffNormalAlarmType_ActiveState_EffectiveDisplayName = 10698
    OffNormalAlarmType_ActiveState_TransitionTime = 10699
    OffNormalAlarmType_ActiveState_EffectiveTransitionTime = 10700
    OffNormalAlarmType_ActiveState_TrueState = 10701
    OffNormalAlarmType_ActiveState_FalseState = 10702
    OffNormalAlarmType_SuppressedState = 10703
    OffNormalAlarmType_SuppressedState_Id = 10704
    OffNormalAlarmType_SuppressedState_Name = 10705
    OffNormalAlarmType_SuppressedState_Number = 10706
    OffNormalAlarmType_SuppressedState_EffectiveDisplayName = 10707
    OffNormalAlarmType_SuppressedState_TransitionTime = 10708
    OffNormalAlarmType_SuppressedState_EffectiveTransitionTime = 10709
    OffNormalAlarmType_SuppressedState_TrueState = 10710
    OffNormalAlarmType_SuppressedState_FalseState = 10711
    OffNormalAlarmType_ShelvingState = 10712
    OffNormalAlarmType_ShelvingState_CurrentState = 10713
    OffNormalAlarmType_ShelvingState_CurrentState_Id = 10714
    OffNormalAlarmType_ShelvingState_CurrentState_Name = 10715
    OffNormalAlarmType_ShelvingState_CurrentState_Number = 10716
    OffNormalAlarmType_ShelvingState_CurrentState_EffectiveDisplayName = 10717
    OffNormalAlarmType_ShelvingState_LastTransition = 10718
    OffNormalAlarmType_ShelvingState_LastTransition_Id = 10719
    OffNormalAlarmType_ShelvingState_LastTransition_Name = 10720
    OffNormalAlarmType_ShelvingState_LastTransition_Number = 10721
    OffNormalAlarmType_ShelvingState_LastTransition_TransitionTime = 10722
    OffNormalAlarmType_ShelvingState_UnshelveTime = 10723
    OffNormalAlarmType_ShelvingState_Unshelve = 10745
    OffNormalAlarmType_ShelvingState_OneShotShelve = 10746
    OffNormalAlarmType_ShelvingState_TimedShelve = 10747
    OffNormalAlarmType_ShelvingState_TimedShelve_InputArguments = 10748
    OffNormalAlarmType_SuppressedOrShelved = 10749
    OffNormalAlarmType_MaxTimeShelved = 10750
    TripAlarmType = 10751
    TripAlarmType_EventId = 10752
    TripAlarmType_EventType = 10753
    TripAlarmType_SourceNode = 10754
    TripAlarmType_SourceName = 10755
    TripAlarmType_Time = 10756
    TripAlarmType_ReceiveTime = 10757
    TripAlarmType_LocalTime = 10758
    TripAlarmType_Message = 10759
    TripAlarmType_Severity = 10760
    TripAlarmType_ConditionName = 10761
    TripAlarmType_BranchId = 10762
    TripAlarmType_Retain = 10763
    TripAlarmType_EnabledState = 10764
    TripAlarmType_EnabledState_Id = 10765
    TripAlarmType_EnabledState_Name = 10766
    TripAlarmType_EnabledState_Number = 10767
    TripAlarmType_EnabledState_EffectiveDisplayName = 10768
    TripAlarmType_EnabledState_TransitionTime = 10769
    TripAlarmType_EnabledState_EffectiveTransitionTime = 10770
    TripAlarmType_EnabledState_TrueState = 10771
    TripAlarmType_EnabledState_FalseState = 10772
    TripAlarmType_Quality = 10773
    TripAlarmType_Quality_SourceTimestamp = 10774
    TripAlarmType_LastSeverity = 10775
    TripAlarmType_LastSeverity_SourceTimestamp = 10776
    TripAlarmType_Comment = 10777
    TripAlarmType_Comment_SourceTimestamp = 10778
    TripAlarmType_ClientUserId = 10779
    TripAlarmType_Enable = 10780
    TripAlarmType_Disable = 10781
    TripAlarmType_AddComment = 10782
    TripAlarmType_AddComment_InputArguments = 10783
    TripAlarmType_ConditionRefresh = 10784
    TripAlarmType_ConditionRefresh_InputArguments = 10785
    TripAlarmType_AckedState = 10786
    TripAlarmType_AckedState_Id = 10787
    TripAlarmType_AckedState_Name = 10788
    TripAlarmType_AckedState_Number = 10789
    TripAlarmType_AckedState_EffectiveDisplayName = 10790
    TripAlarmType_AckedState_TransitionTime = 10791
    TripAlarmType_AckedState_EffectiveTransitionTime = 10792
    TripAlarmType_AckedState_TrueState = 10793
    TripAlarmType_AckedState_FalseState = 10794
    TripAlarmType_ConfirmedState = 10795
    TripAlarmType_ConfirmedState_Id = 10796
    TripAlarmType_ConfirmedState_Name = 10797
    TripAlarmType_ConfirmedState_Number = 10798
    TripAlarmType_ConfirmedState_EffectiveDisplayName = 10799
    TripAlarmType_ConfirmedState_TransitionTime = 10800
    TripAlarmType_ConfirmedState_EffectiveTransitionTime = 10801
    TripAlarmType_ConfirmedState_TrueState = 10802
    TripAlarmType_ConfirmedState_FalseState = 10803
    TripAlarmType_Acknowledge = 10804
    TripAlarmType_Acknowledge_InputArguments = 10805
    TripAlarmType_Confirm = 10806
    TripAlarmType_Confirm_InputArguments = 10807
    TripAlarmType_ActiveState = 10808
    TripAlarmType_ActiveState_Id = 10809
    TripAlarmType_ActiveState_Name = 10810
    TripAlarmType_ActiveState_Number = 10811
    TripAlarmType_ActiveState_EffectiveDisplayName = 10812
    TripAlarmType_ActiveState_TransitionTime = 10813
    TripAlarmType_ActiveState_EffectiveTransitionTime = 10814
    TripAlarmType_ActiveState_TrueState = 10815
    TripAlarmType_ActiveState_FalseState = 10816
    TripAlarmType_SuppressedState = 10817
    TripAlarmType_SuppressedState_Id = 10818
    TripAlarmType_SuppressedState_Name = 10819
    TripAlarmType_SuppressedState_Number = 10820
    TripAlarmType_SuppressedState_EffectiveDisplayName = 10821
    TripAlarmType_SuppressedState_TransitionTime = 10822
    TripAlarmType_SuppressedState_EffectiveTransitionTime = 10823
    TripAlarmType_SuppressedState_TrueState = 10824
    TripAlarmType_SuppressedState_FalseState = 10825
    TripAlarmType_ShelvingState = 10826
    TripAlarmType_ShelvingState_CurrentState = 10827
    TripAlarmType_ShelvingState_CurrentState_Id = 10828
    TripAlarmType_ShelvingState_CurrentState_Name = 10829
    TripAlarmType_ShelvingState_CurrentState_Number = 10830
    TripAlarmType_ShelvingState_CurrentState_EffectiveDisplayName = 10831
    TripAlarmType_ShelvingState_LastTransition = 10832
    TripAlarmType_ShelvingState_LastTransition_Id = 10833
    TripAlarmType_ShelvingState_LastTransition_Name = 10834
    TripAlarmType_ShelvingState_LastTransition_Number = 10835
    TripAlarmType_ShelvingState_LastTransition_TransitionTime = 10836
    TripAlarmType_ShelvingState_UnshelveTime = 10837
    TripAlarmType_ShelvingState_Unshelve = 10859
    TripAlarmType_ShelvingState_OneShotShelve = 10860
    TripAlarmType_ShelvingState_TimedShelve = 10861
    TripAlarmType_ShelvingState_TimedShelve_InputArguments = 10862
    TripAlarmType_SuppressedOrShelved = 10863
    TripAlarmType_MaxTimeShelved = 10864
    AuditConditionShelvingEventType = 11093
    AuditConditionShelvingEventType_EventId = 11094
    AuditConditionShelvingEventType_EventType = 11095
    AuditConditionShelvingEventType_SourceNode = 11096
    AuditConditionShelvingEventType_SourceName = 11097
    AuditConditionShelvingEventType_Time = 11098
    AuditConditionShelvingEventType_ReceiveTime = 11099
    AuditConditionShelvingEventType_LocalTime = 11100
    AuditConditionShelvingEventType_Message = 11101
    AuditConditionShelvingEventType_Severity = 11102
    AuditConditionShelvingEventType_ActionTimeStamp = 11103
    AuditConditionShelvingEventType_Status = 11104
    AuditConditionShelvingEventType_ServerId = 11105
    AuditConditionShelvingEventType_ClientAuditEntryId = 11106
    AuditConditionShelvingEventType_ClientUserId = 11107
    AuditConditionShelvingEventType_MethodId = 11108
    AuditConditionShelvingEventType_InputArguments = 11109
    TwoStateVariableType_TrueState = 11110
    TwoStateVariableType_FalseState = 11111
    ConditionType_ConditionClassId = 11112
    ConditionType_ConditionClassName = 11113
    DialogConditionType_ConditionClassId = 11114
    DialogConditionType_ConditionClassName = 11115
    AcknowledgeableConditionType_ConditionClassId = 11116
    AcknowledgeableConditionType_ConditionClassName = 11117
    AlarmConditionType_ConditionClassId = 11118
    AlarmConditionType_ConditionClassName = 11119
    AlarmConditionType_InputNode = 11120
    LimitAlarmType_ConditionClassId = 11121
    LimitAlarmType_ConditionClassName = 11122
    LimitAlarmType_InputNode = 11123
    LimitAlarmType_HighHighLimit = 11124
    LimitAlarmType_HighLimit = 11125
    LimitAlarmType_LowLimit = 11126
    LimitAlarmType_LowLowLimit = 11127
    ExclusiveLimitAlarmType_ConditionClassId = 11128
    ExclusiveLimitAlarmType_ConditionClassName = 11129
    ExclusiveLimitAlarmType_InputNode = 11130
    ExclusiveLevelAlarmType_ConditionClassId = 11131
    ExclusiveLevelAlarmType_ConditionClassName = 11132
    ExclusiveLevelAlarmType_InputNode = 11133
    ExclusiveRateOfChangeAlarmType_ConditionClassId = 11134
    ExclusiveRateOfChangeAlarmType_ConditionClassName = 11135
    ExclusiveRateOfChangeAlarmType_InputNode = 11136
    ExclusiveDeviationAlarmType_ConditionClassId = 11137
    ExclusiveDeviationAlarmType_ConditionClassName = 11138
    ExclusiveDeviationAlarmType_InputNode = 11139
    NonExclusiveLimitAlarmType_ConditionClassId = 11140
    NonExclusiveLimitAlarmType_ConditionClassName = 11141
    NonExclusiveLimitAlarmType_InputNode = 11142
    NonExclusiveLevelAlarmType_ConditionClassId = 11143
    NonExclusiveLevelAlarmType_ConditionClassName = 11144
    NonExclusiveLevelAlarmType_InputNode = 11145
    NonExclusiveRateOfChangeAlarmType_ConditionClassId = 11146
    NonExclusiveRateOfChangeAlarmType_ConditionClassName = 11147
    NonExclusiveRateOfChangeAlarmType_InputNode = 11148
    NonExclusiveDeviationAlarmType_ConditionClassId = 11149
    NonExclusiveDeviationAlarmType_ConditionClassName = 11150
    NonExclusiveDeviationAlarmType_InputNode = 11151
    DiscreteAlarmType_ConditionClassId = 11152
    DiscreteAlarmType_ConditionClassName = 11153
    DiscreteAlarmType_InputNode = 11154
    OffNormalAlarmType_ConditionClassId = 11155
    OffNormalAlarmType_ConditionClassName = 11156
    OffNormalAlarmType_InputNode = 11157
    OffNormalAlarmType_NormalState = 11158
    TripAlarmType_ConditionClassId = 11159
    TripAlarmType_ConditionClassName = 11160
    TripAlarmType_InputNode = 11161
    TripAlarmType_NormalState = 11162
    BaseConditionClassType = 11163
    ProcessConditionClassType = 11164
    MaintenanceConditionClassType = 11165
    SystemConditionClassType = 11166
    HistoricalDataConfigurationType_AggregateConfiguration_TreatUncertainAsBad = 11168
    HistoricalDataConfigurationType_AggregateConfiguration_PercentDataBad = 11169
    HistoricalDataConfigurationType_AggregateConfiguration_PercentDataGood = 11170
    HistoricalDataConfigurationType_AggregateConfiguration_UseSlopedExtrapolation = 11171
    HistoryServerCapabilitiesType_AggregateFunctions = 11172
    AggregateConfigurationType = 11187
    AggregateConfigurationType_TreatUncertainAsBad = 11188
    AggregateConfigurationType_PercentDataBad = 11189
    AggregateConfigurationType_PercentDataGood = 11190
    AggregateConfigurationType_UseSlopedExtrapolation = 11191
    HistoryServerCapabilities = 11192
    HistoryServerCapabilities_AccessHistoryDataCapability = 11193
    HistoryServerCapabilities_InsertDataCapability = 11196
    HistoryServerCapabilities_ReplaceDataCapability = 11197
    HistoryServerCapabilities_UpdateDataCapability = 11198
    HistoryServerCapabilities_DeleteRawCapability = 11199
    HistoryServerCapabilities_DeleteAtTimeCapability = 11200
    HistoryServerCapabilities_AggregateFunctions = 11201
    HAConfiguration = 11202
    HAConfiguration_AggregateConfiguration = 11203
    HAConfiguration_AggregateConfiguration_TreatUncertainAsBad = 11204
    HAConfiguration_AggregateConfiguration_PercentDataBad = 11205
    HAConfiguration_AggregateConfiguration_PercentDataGood = 11206
    HAConfiguration_AggregateConfiguration_UseSlopedExtrapolation = 11207
    HAConfiguration_Stepped = 11208
    HAConfiguration_Definition = 11209
    HAConfiguration_MaxTimeInterval = 11210
    HAConfiguration_MinTimeInterval = 11211
    HAConfiguration_ExceptionDeviation = 11212
    HAConfiguration_ExceptionDeviationFormat = 11213
    Annotations = 11214
    HistoricalEventFilter = 11215
    ModificationInfo = 11216
    HistoryModifiedData = 11217
    ModificationInfo_Encoding_DefaultXml = 11218
    HistoryModifiedData_Encoding_DefaultXml = 11219
    ModificationInfo_Encoding_DefaultBinary = 11226
    HistoryModifiedData_Encoding_DefaultBinary = 11227
    HistoryUpdateType = 11234
    MultiStateValueDiscreteType = 11238
    MultiStateValueDiscreteType_Definition = 11239
    MultiStateValueDiscreteType_ValuePrecision = 11240
    MultiStateValueDiscreteType_EnumValues = 11241
    HistoryServerCapabilities_AccessHistoryEventsCapability = 11242
    HistoryServerCapabilitiesType_MaxReturnDataValues = 11268
    HistoryServerCapabilitiesType_MaxReturnEventValues = 11269
    HistoryServerCapabilitiesType_InsertAnnotationCapability = 11270
    HistoryServerCapabilities_MaxReturnDataValues = 11273
    HistoryServerCapabilities_MaxReturnEventValues = 11274
    HistoryServerCapabilities_InsertAnnotationCapability = 11275
    HistoryServerCapabilitiesType_InsertEventCapability = 11278
    HistoryServerCapabilitiesType_ReplaceEventCapability = 11279
    HistoryServerCapabilitiesType_UpdateEventCapability = 11280
    HistoryServerCapabilities_InsertEventCapability = 11281
    HistoryServerCapabilities_ReplaceEventCapability = 11282
    HistoryServerCapabilities_UpdateEventCapability = 11283
    AggregateFunction_TimeAverage2 = 11285
    AggregateFunction_Minimum2 = 11286
    AggregateFunction_Maximum2 = 11287
    AggregateFunction_Range2 = 11288
    AggregateFunction_WorstQuality2 = 11292
    PerformUpdateType = 11293
    UpdateStructureDataDetails = 11295
    UpdateStructureDataDetails_Encoding_DefaultXml = 11296
    UpdateStructureDataDetails_Encoding_DefaultBinary = 11300
    AggregateFunction_Total2 = 11304
    AggregateFunction_MinimumActualTime2 = 11305
    AggregateFunction_MaximumActualTime2 = 11306
    AggregateFunction_DurationInStateZero = 11307
    AggregateFunction_DurationInStateNonZero = 11308
    Server_ServerRedundancy_CurrentServerId = 11312
    Server_ServerRedundancy_RedundantServerArray = 11313
    Server_ServerRedundancy_ServerUriArray = 11314
    ShelvedStateMachineType_UnshelvedToTimedShelved_TransitionNumber = 11322
    ShelvedStateMachineType_UnshelvedToOneShotShelved_TransitionNumber = 11323
    ShelvedStateMachineType_TimedShelvedToUnshelved_TransitionNumber = 11324
    ShelvedStateMachineType_TimedShelvedToOneShotShelved_TransitionNumber = 11325
    ShelvedStateMachineType_OneShotShelvedToUnshelved_TransitionNumber = 11326
    ShelvedStateMachineType_OneShotShelvedToTimedShelved_TransitionNumber = 11327
    ExclusiveLimitStateMachineType_LowLowToLow_TransitionNumber = 11340
    ExclusiveLimitStateMachineType_LowToLowLow_TransitionNumber = 11341
    ExclusiveLimitStateMachineType_HighHighToHigh_TransitionNumber = 11342
    ExclusiveLimitStateMachineType_HighToHighHigh_TransitionNumber = 11343
    AggregateFunction_StandardDeviationSample = 11426
    AggregateFunction_StandardDeviationPopulation = 11427
    AggregateFunction_VarianceSample = 11428
    AggregateFunction_VariancePopulation = 11429
    EnumStrings = 11432
    ValueAsText = 11433
    ProgressEventType = 11436
    ProgressEventType_EventId = 11437
    ProgressEventType_EventType = 11438
    ProgressEventType_SourceNode = 11439
    ProgressEventType_SourceName = 11440
    ProgressEventType_Time = 11441
    ProgressEventType_ReceiveTime = 11442
    ProgressEventType_LocalTime = 11443
    ProgressEventType_Message = 11444
    ProgressEventType_Severity = 11445
    SystemStatusChangeEventType = 11446
    SystemStatusChangeEventType_EventId = 11447
    SystemStatusChangeEventType_EventType = 11448
    SystemStatusChangeEventType_SourceNode = 11449
    SystemStatusChangeEventType_SourceName = 11450
    SystemStatusChangeEventType_Time = 11451
    SystemStatusChangeEventType_ReceiveTime = 11452
    SystemStatusChangeEventType_LocalTime = 11453
    SystemStatusChangeEventType_Message = 11454
    SystemStatusChangeEventType_Severity = 11455
    TransitionVariableType_EffectiveTransitionTime = 11456
    FiniteTransitionVariableType_EffectiveTransitionTime = 11457
    StateMachineType_LastTransition_EffectiveTransitionTime = 11458
    FiniteStateMachineType_LastTransition_EffectiveTransitionTime = 11459
    TransitionEventType_Transition_EffectiveTransitionTime = 11460
    MultiStateValueDiscreteType_ValueAsText = 11461
    ProgramTransitionEventType_Transition_EffectiveTransitionTime = 11462
    ProgramTransitionAuditEventType_Transition_EffectiveTransitionTime = 11463
    ProgramStateMachineType_LastTransition_EffectiveTransitionTime = 11464
    ShelvedStateMachineType_LastTransition_EffectiveTransitionTime = 11465
    AlarmConditionType_ShelvingState_LastTransition_EffectiveTransitionTime = 11466
    LimitAlarmType_ShelvingState_LastTransition_EffectiveTransitionTime = 11467
    ExclusiveLimitStateMachineType_LastTransition_EffectiveTransitionTime = 11468
    ExclusiveLimitAlarmType_ShelvingState_LastTransition_EffectiveTransitionTime = 11469
    ExclusiveLimitAlarmType_LimitState_LastTransition_EffectiveTransitionTime = 11470
    ExclusiveLevelAlarmType_ShelvingState_LastTransition_EffectiveTransitionTime = 11471
    ExclusiveLevelAlarmType_LimitState_LastTransition_EffectiveTransitionTime = 11472
    ExclusiveRateOfChangeAlarmType_ShelvingState_LastTransition_EffectiveTransitionTime = 11473
    ExclusiveRateOfChangeAlarmType_LimitState_LastTransition_EffectiveTransitionTime = 11474
    ExclusiveDeviationAlarmType_ShelvingState_LastTransition_EffectiveTransitionTime = 11475
    ExclusiveDeviationAlarmType_LimitState_LastTransition_EffectiveTransitionTime = 11476
    NonExclusiveLimitAlarmType_ShelvingState_LastTransition_EffectiveTransitionTime = 11477
    NonExclusiveLevelAlarmType_ShelvingState_LastTransition_EffectiveTransitionTime = 11478
    NonExclusiveRateOfChangeAlarmType_ShelvingState_LastTransition_EffectiveTransitionTime = 11479
    NonExclusiveDeviationAlarmType_ShelvingState_LastTransition_EffectiveTransitionTime = 11480
    DiscreteAlarmType_ShelvingState_LastTransition_EffectiveTransitionTime = 11481
    OffNormalAlarmType_ShelvingState_LastTransition_EffectiveTransitionTime = 11482
    TripAlarmType_ShelvingState_LastTransition_EffectiveTransitionTime = 11483
    AuditActivateSessionEventType_SecureChannelId = 11485
    OptionSetType = 11487
    OptionSetType_OptionSetValues = 11488
    ServerType_GetMonitoredItems = 11489
    ServerType_GetMonitoredItems_InputArguments = 11490
    ServerType_GetMonitoredItems_OutputArguments = 11491
    Server_GetMonitoredItems = 11492
    Server_GetMonitoredItems_InputArguments = 11493
    Server_GetMonitoredItems_OutputArguments = 11494
    GetMonitoredItemsMethodType = 11495
    GetMonitoredItemsMethodType_InputArguments = 11496
    GetMonitoredItemsMethodType_OutputArguments = 11497
    MaxStringLength = 11498
    HistoricalDataConfigurationType_StartOfArchive = 11499
    HistoricalDataConfigurationType_StartOfOnlineArchive = 11500
    HistoryServerCapabilitiesType_DeleteEventCapability = 11501
    HistoryServerCapabilities_DeleteEventCapability = 11502
    HAConfiguration_StartOfArchive = 11503
    HAConfiguration_StartOfOnlineArchive = 11504
    AggregateFunction_StartBound = 11505
    AggregateFunction_EndBound = 11506
    AggregateFunction_DeltaBounds = 11507
    ModellingRule_OptionalPlaceholder = 11508
    ModellingRule_OptionalPlaceholder_NamingRule = 11509
    ModellingRule_MandatoryPlaceholder = 11510
    ModellingRule_MandatoryPlaceholder_NamingRule = 11511
    MaxArrayLength = 11512
    EngineeringUnits = 11513
    ServerType_ServerCapabilities_MaxArrayLength = 11514
    ServerType_ServerCapabilities_MaxStringLength = 11515
    ServerType_ServerCapabilities_OperationLimits = 11516
    ServerType_ServerCapabilities_OperationLimits_MaxNodesPerRead = 11517
    ServerType_ServerCapabilities_OperationLimits_MaxNodesPerWrite = 11519
    ServerType_ServerCapabilities_OperationLimits_MaxNodesPerMethodCall = 11521
    ServerType_ServerCapabilities_OperationLimits_MaxNodesPerBrowse = 11522
    ServerType_ServerCapabilities_OperationLimits_MaxNodesPerRegisterNodes = 11523
    ServerType_ServerCapabilities_OperationLimits_MaxNodesPerTranslateBrowsePathsToNodeIds = 11524
    ServerType_ServerCapabilities_OperationLimits_MaxNodesPerNodeManagement = 11525
    ServerType_ServerCapabilities_OperationLimits_MaxMonitoredItemsPerCall = 11526
    ServerType_Namespaces = 11527
    ServerType_Namespaces_AddressSpaceFile = 11528
    ServerType_Namespaces_AddressSpaceFile_Size = 11529
    ServerType_Namespaces_AddressSpaceFile_OpenCount = 11532
    ServerType_Namespaces_AddressSpaceFile_Open = 11533
    ServerType_Namespaces_AddressSpaceFile_Open_InputArguments = 11534
    ServerType_Namespaces_AddressSpaceFile_Open_OutputArguments = 11535
    ServerType_Namespaces_AddressSpaceFile_Close = 11536
    ServerType_Namespaces_AddressSpaceFile_Close_InputArguments = 11537
    ServerType_Namespaces_AddressSpaceFile_Read = 11538
    ServerType_Namespaces_AddressSpaceFile_Read_InputArguments = 11539
    ServerType_Namespaces_AddressSpaceFile_Read_OutputArguments = 11540
    ServerType_Namespaces_AddressSpaceFile_Write = 11541
    ServerType_Namespaces_AddressSpaceFile_Write_InputArguments = 11542
    ServerType_Namespaces_AddressSpaceFile_GetPosition = 11543
    ServerType_Namespaces_AddressSpaceFile_GetPosition_InputArguments = 11544
    ServerType_Namespaces_AddressSpaceFile_GetPosition_OutputArguments = 11545
    ServerType_Namespaces_AddressSpaceFile_SetPosition = 11546
    ServerType_Namespaces_AddressSpaceFile_SetPosition_InputArguments = 11547
    ServerType_Namespaces_AddressSpaceFile_ExportNamespace = 11548
    ServerCapabilitiesType_MaxArrayLength = 11549
    ServerCapabilitiesType_MaxStringLength = 11550
    ServerCapabilitiesType_OperationLimits = 11551
    ServerCapabilitiesType_OperationLimits_MaxNodesPerRead = 11552
    ServerCapabilitiesType_OperationLimits_MaxNodesPerWrite = 11554
    ServerCapabilitiesType_OperationLimits_MaxNodesPerMethodCall = 11556
    ServerCapabilitiesType_OperationLimits_MaxNodesPerBrowse = 11557
    ServerCapabilitiesType_OperationLimits_MaxNodesPerRegisterNodes = 11558
    ServerCapabilitiesType_OperationLimits_MaxNodesPerTranslateBrowsePathsToNodeIds = 11559
    ServerCapabilitiesType_OperationLimits_MaxNodesPerNodeManagement = 11560
    ServerCapabilitiesType_OperationLimits_MaxMonitoredItemsPerCall = 11561
    ServerCapabilitiesType_VendorCapability = 11562
    OperationLimitsType = 11564
    OperationLimitsType_MaxNodesPerRead = 11565
    OperationLimitsType_MaxNodesPerWrite = 11567
    OperationLimitsType_MaxNodesPerMethodCall = 11569
    OperationLimitsType_MaxNodesPerBrowse = 11570
    OperationLimitsType_MaxNodesPerRegisterNodes = 11571
    OperationLimitsType_MaxNodesPerTranslateBrowsePathsToNodeIds = 11572
    OperationLimitsType_MaxNodesPerNodeManagement = 11573
    OperationLimitsType_MaxMonitoredItemsPerCall = 11574
    FileType = 11575
    FileType_Size = 11576
    FileType_OpenCount = 11579
    FileType_Open = 11580
    FileType_Open_InputArguments = 11581
    FileType_Open_OutputArguments = 11582
    FileType_Close = 11583
    FileType_Close_InputArguments = 11584
    FileType_Read = 11585
    FileType_Read_InputArguments = 11586
    FileType_Read_OutputArguments = 11587
    FileType_Write = 11588
    FileType_Write_InputArguments = 11589
    FileType_GetPosition = 11590
    FileType_GetPosition_InputArguments = 11591
    FileType_GetPosition_OutputArguments = 11592
    FileType_SetPosition = 11593
    FileType_SetPosition_InputArguments = 11594
    AddressSpaceFileType = 11595
    AddressSpaceFileType_Size = 11596
    AddressSpaceFileType_OpenCount = 11599
    AddressSpaceFileType_Open = 11600
    AddressSpaceFileType_Open_InputArguments = 11601
    AddressSpaceFileType_Open_OutputArguments = 11602
    AddressSpaceFileType_Close = 11603
    AddressSpaceFileType_Close_InputArguments = 11604
    AddressSpaceFileType_Read = 11605
    AddressSpaceFileType_Read_InputArguments = 11606
    AddressSpaceFileType_Read_OutputArguments = 11607
    AddressSpaceFileType_Write = 11608
    AddressSpaceFileType_Write_InputArguments = 11609
    AddressSpaceFileType_GetPosition = 11610
    AddressSpaceFileType_GetPosition_InputArguments = 11611
    AddressSpaceFileType_GetPosition_OutputArguments = 11612
    AddressSpaceFileType_SetPosition = 11613
    AddressSpaceFileType_SetPosition_InputArguments = 11614
    AddressSpaceFileType_ExportNamespace = 11615
    NamespaceMetadataType = 11616
    NamespaceMetadataType_NamespaceUri = 11617
    NamespaceMetadataType_NamespaceVersion = 11618
    NamespaceMetadataType_NamespacePublicationDate = 11619
    NamespaceMetadataType_IsNamespaceSubset = 11620
    NamespaceMetadataType_StaticNodeIdIdentifierTypes = 11621
    NamespaceMetadataType_StaticNumericNodeIdRange = 11622
    NamespaceMetadataType_StaticStringNodeIdPattern = 11623
    NamespaceMetadataType_NamespaceFile = 11624
    NamespaceMetadataType_NamespaceFile_Size = 11625
    NamespaceMetadataType_NamespaceFile_OpenCount = 11628
    NamespaceMetadataType_NamespaceFile_Open = 11629
    NamespaceMetadataType_NamespaceFile_Open_InputArguments = 11630
    NamespaceMetadataType_NamespaceFile_Open_OutputArguments = 11631
    NamespaceMetadataType_NamespaceFile_Close = 11632
    NamespaceMetadataType_NamespaceFile_Close_InputArguments = 11633
    NamespaceMetadataType_NamespaceFile_Read = 11634
    NamespaceMetadataType_NamespaceFile_Read_InputArguments = 11635
    NamespaceMetadataType_NamespaceFile_Read_OutputArguments = 11636
    NamespaceMetadataType_NamespaceFile_Write = 11637
    NamespaceMetadataType_NamespaceFile_Write_InputArguments = 11638
    NamespaceMetadataType_NamespaceFile_GetPosition = 11639
    NamespaceMetadataType_NamespaceFile_GetPosition_InputArguments = 11640
    NamespaceMetadataType_NamespaceFile_GetPosition_OutputArguments = 11641
    NamespaceMetadataType_NamespaceFile_SetPosition = 11642
    NamespaceMetadataType_NamespaceFile_SetPosition_InputArguments = 11643
    NamespaceMetadataType_NamespaceFile_ExportNamespace = 11644
    NamespacesType = 11645
    NamespacesType_NamespaceIdentifier = 11646
    NamespacesType_NamespaceIdentifier_NamespaceUri = 11647
    NamespacesType_NamespaceIdentifier_NamespaceVersion = 11648
    NamespacesType_NamespaceIdentifier_NamespacePublicationDate = 11649
    NamespacesType_NamespaceIdentifier_IsNamespaceSubset = 11650
    NamespacesType_NamespaceIdentifier_StaticNodeIdIdentifierTypes = 11651
    NamespacesType_NamespaceIdentifier_StaticNumericNodeIdRange = 11652
    NamespacesType_NamespaceIdentifier_StaticStringNodeIdPattern = 11653
    NamespacesType_NamespaceIdentifier_NamespaceFile = 11654
    NamespacesType_NamespaceIdentifier_NamespaceFile_Size = 11655
    NamespacesType_NamespaceIdentifier_NamespaceFile_OpenCount = 11658
    NamespacesType_NamespaceIdentifier_NamespaceFile_Open = 11659
    NamespacesType_NamespaceIdentifier_NamespaceFile_Open_InputArguments = 11660
    NamespacesType_NamespaceIdentifier_NamespaceFile_Open_OutputArguments = 11661
    NamespacesType_NamespaceIdentifier_NamespaceFile_Close = 11662
    NamespacesType_NamespaceIdentifier_NamespaceFile_Close_InputArguments = 11663
    NamespacesType_NamespaceIdentifier_NamespaceFile_Read = 11664
    NamespacesType_NamespaceIdentifier_NamespaceFile_Read_InputArguments = 11665
    NamespacesType_NamespaceIdentifier_NamespaceFile_Read_OutputArguments = 11666
    NamespacesType_NamespaceIdentifier_NamespaceFile_Write = 11667
    NamespacesType_NamespaceIdentifier_NamespaceFile_Write_InputArguments = 11668
    NamespacesType_NamespaceIdentifier_NamespaceFile_GetPosition = 11669
    NamespacesType_NamespaceIdentifier_NamespaceFile_GetPosition_InputArguments = 11670
    NamespacesType_NamespaceIdentifier_NamespaceFile_GetPosition_OutputArguments = 11671
    NamespacesType_NamespaceIdentifier_NamespaceFile_SetPosition = 11672
    NamespacesType_NamespaceIdentifier_NamespaceFile_SetPosition_InputArguments = 11673
    NamespacesType_NamespaceIdentifier_NamespaceFile_ExportNamespace = 11674
    NamespacesType_AddressSpaceFile = 11675
    NamespacesType_AddressSpaceFile_Size = 11676
    NamespacesType_AddressSpaceFile_OpenCount = 11679
    NamespacesType_AddressSpaceFile_Open = 11680
    NamespacesType_AddressSpaceFile_Open_InputArguments = 11681
    NamespacesType_AddressSpaceFile_Open_OutputArguments = 11682
    NamespacesType_AddressSpaceFile_Close = 11683
    NamespacesType_AddressSpaceFile_Close_InputArguments = 11684
    NamespacesType_AddressSpaceFile_Read = 11685
    NamespacesType_AddressSpaceFile_Read_InputArguments = 11686
    NamespacesType_AddressSpaceFile_Read_OutputArguments = 11687
    NamespacesType_AddressSpaceFile_Write = 11688
    NamespacesType_AddressSpaceFile_Write_InputArguments = 11689
    NamespacesType_AddressSpaceFile_GetPosition = 11690
    NamespacesType_AddressSpaceFile_GetPosition_InputArguments = 11691
    NamespacesType_AddressSpaceFile_GetPosition_OutputArguments = 11692
    NamespacesType_AddressSpaceFile_SetPosition = 11693
    NamespacesType_AddressSpaceFile_SetPosition_InputArguments = 11694
    NamespacesType_AddressSpaceFile_ExportNamespace = 11695
    SystemStatusChangeEventType_SystemState = 11696
    SamplingIntervalDiagnosticsType_SampledMonitoredItemsCount = 11697
    SamplingIntervalDiagnosticsType_MaxSampledMonitoredItemsCount = 11698
    SamplingIntervalDiagnosticsType_DisabledMonitoredItemsSamplingCount = 11699
    OptionSetType_BitMask = 11701
    Server_ServerCapabilities_MaxArrayLength = 11702
    Server_ServerCapabilities_MaxStringLength = 11703
    Server_ServerCapabilities_OperationLimits = 11704
    Server_ServerCapabilities_OperationLimits_MaxNodesPerRead = 11705
    Server_ServerCapabilities_OperationLimits_MaxNodesPerWrite = 11707
    Server_ServerCapabilities_OperationLimits_MaxNodesPerMethodCall = 11709
    Server_ServerCapabilities_OperationLimits_MaxNodesPerBrowse = 11710
    Server_ServerCapabilities_OperationLimits_MaxNodesPerRegisterNodes = 11711
    Server_ServerCapabilities_OperationLimits_MaxNodesPerTranslateBrowsePathsToNodeIds = 11712
    Server_ServerCapabilities_OperationLimits_MaxNodesPerNodeManagement = 11713
    Server_ServerCapabilities_OperationLimits_MaxMonitoredItemsPerCall = 11714
    Server_Namespaces = 11715
    Server_Namespaces_AddressSpaceFile = 11716
    Server_Namespaces_AddressSpaceFile_Size = 11717
    Server_Namespaces_AddressSpaceFile_OpenCount = 11720
    Server_Namespaces_AddressSpaceFile_Open = 11721
    Server_Namespaces_AddressSpaceFile_Open_InputArguments = 11722
    Server_Namespaces_AddressSpaceFile_Open_OutputArguments = 11723
    Server_Namespaces_AddressSpaceFile_Close = 11724
    Server_Namespaces_AddressSpaceFile_Close_InputArguments = 11725
    Server_Namespaces_AddressSpaceFile_Read = 11726
    Server_Namespaces_AddressSpaceFile_Read_InputArguments = 11727
    Server_Namespaces_AddressSpaceFile_Read_OutputArguments = 11728
    Server_Namespaces_AddressSpaceFile_Write = 11729
    Server_Namespaces_AddressSpaceFile_Write_InputArguments = 11730
    Server_Namespaces_AddressSpaceFile_GetPosition = 11731
    Server_Namespaces_AddressSpaceFile_GetPosition_InputArguments = 11732
    Server_Namespaces_AddressSpaceFile_GetPosition_OutputArguments = 11733
    Server_Namespaces_AddressSpaceFile_SetPosition = 11734
    Server_Namespaces_AddressSpaceFile_SetPosition_InputArguments = 11735
    Server_Namespaces_AddressSpaceFile_ExportNamespace = 11736
    BitFieldMaskDataType = 11737
    OpenMethodType = 11738
    OpenMethodType_InputArguments = 11739
    OpenMethodType_OutputArguments = 11740
    CloseMethodType = 11741
    CloseMethodType_InputArguments = 11742
    ReadMethodType = 11743
    ReadMethodType_InputArguments = 11744
    ReadMethodType_OutputArguments = 11745
    WriteMethodType = 11746
    WriteMethodType_InputArguments = 11747
    GetPositionMethodType = 11748
    GetPositionMethodType_InputArguments = 11749
    GetPositionMethodType_OutputArguments = 11750
    SetPositionMethodType = 11751
    SetPositionMethodType_InputArguments = 11752
    SystemOffNormalAlarmType = 11753
    SystemOffNormalAlarmType_EventId = 11754
    SystemOffNormalAlarmType_EventType = 11755
    SystemOffNormalAlarmType_SourceNode = 11756
    SystemOffNormalAlarmType_SourceName = 11757
    SystemOffNormalAlarmType_Time = 11758
    SystemOffNormalAlarmType_ReceiveTime = 11759
    SystemOffNormalAlarmType_LocalTime = 11760
    SystemOffNormalAlarmType_Message = 11761
    SystemOffNormalAlarmType_Severity = 11762
    SystemOffNormalAlarmType_ConditionClassId = 11763
    SystemOffNormalAlarmType_ConditionClassName = 11764
    SystemOffNormalAlarmType_ConditionName = 11765
    SystemOffNormalAlarmType_BranchId = 11766
    SystemOffNormalAlarmType_Retain = 11767
    SystemOffNormalAlarmType_EnabledState = 11768
    SystemOffNormalAlarmType_EnabledState_Id = 11769
    SystemOffNormalAlarmType_EnabledState_Name = 11770
    SystemOffNormalAlarmType_EnabledState_Number = 11771
    SystemOffNormalAlarmType_EnabledState_EffectiveDisplayName = 11772
    SystemOffNormalAlarmType_EnabledState_TransitionTime = 11773
    SystemOffNormalAlarmType_EnabledState_EffectiveTransitionTime = 11774
    SystemOffNormalAlarmType_EnabledState_TrueState = 11775
    SystemOffNormalAlarmType_EnabledState_FalseState = 11776
    SystemOffNormalAlarmType_Quality = 11777
    SystemOffNormalAlarmType_Quality_SourceTimestamp = 11778
    SystemOffNormalAlarmType_LastSeverity = 11779
    SystemOffNormalAlarmType_LastSeverity_SourceTimestamp = 11780
    SystemOffNormalAlarmType_Comment = 11781
    SystemOffNormalAlarmType_Comment_SourceTimestamp = 11782
    SystemOffNormalAlarmType_ClientUserId = 11783
    SystemOffNormalAlarmType_Disable = 11784
    SystemOffNormalAlarmType_Enable = 11785
    SystemOffNormalAlarmType_AddComment = 11786
    SystemOffNormalAlarmType_AddComment_InputArguments = 11787
    SystemOffNormalAlarmType_ConditionRefresh = 11788
    SystemOffNormalAlarmType_ConditionRefresh_InputArguments = 11789
    SystemOffNormalAlarmType_AckedState = 11790
    SystemOffNormalAlarmType_AckedState_Id = 11791
    SystemOffNormalAlarmType_AckedState_Name = 11792
    SystemOffNormalAlarmType_AckedState_Number = 11793
    SystemOffNormalAlarmType_AckedState_EffectiveDisplayName = 11794
    SystemOffNormalAlarmType_AckedState_TransitionTime = 11795
    SystemOffNormalAlarmType_AckedState_EffectiveTransitionTime = 11796
    SystemOffNormalAlarmType_AckedState_TrueState = 11797
    SystemOffNormalAlarmType_AckedState_FalseState = 11798
    SystemOffNormalAlarmType_ConfirmedState = 11799
    SystemOffNormalAlarmType_ConfirmedState_Id = 11800
    SystemOffNormalAlarmType_ConfirmedState_Name = 11801
    SystemOffNormalAlarmType_ConfirmedState_Number = 11802
    SystemOffNormalAlarmType_ConfirmedState_EffectiveDisplayName = 11803
    SystemOffNormalAlarmType_ConfirmedState_TransitionTime = 11804
    SystemOffNormalAlarmType_ConfirmedState_EffectiveTransitionTime = 11805
    SystemOffNormalAlarmType_ConfirmedState_TrueState = 11806
    SystemOffNormalAlarmType_ConfirmedState_FalseState = 11807
    SystemOffNormalAlarmType_Acknowledge = 11808
    SystemOffNormalAlarmType_Acknowledge_InputArguments = 11809
    SystemOffNormalAlarmType_Confirm = 11810
    SystemOffNormalAlarmType_Confirm_InputArguments = 11811
    SystemOffNormalAlarmType_ActiveState = 11812
    SystemOffNormalAlarmType_ActiveState_Id = 11813
    SystemOffNormalAlarmType_ActiveState_Name = 11814
    SystemOffNormalAlarmType_ActiveState_Number = 11815
    SystemOffNormalAlarmType_ActiveState_EffectiveDisplayName = 11816
    SystemOffNormalAlarmType_ActiveState_TransitionTime = 11817
    SystemOffNormalAlarmType_ActiveState_EffectiveTransitionTime = 11818
    SystemOffNormalAlarmType_ActiveState_TrueState = 11819
    SystemOffNormalAlarmType_ActiveState_FalseState = 11820
    SystemOffNormalAlarmType_InputNode = 11821
    SystemOffNormalAlarmType_SuppressedState = 11822
    SystemOffNormalAlarmType_SuppressedState_Id = 11823
    SystemOffNormalAlarmType_SuppressedState_Name = 11824
    SystemOffNormalAlarmType_SuppressedState_Number = 11825
    SystemOffNormalAlarmType_SuppressedState_EffectiveDisplayName = 11826
    SystemOffNormalAlarmType_SuppressedState_TransitionTime = 11827
    SystemOffNormalAlarmType_SuppressedState_EffectiveTransitionTime = 11828
    SystemOffNormalAlarmType_SuppressedState_TrueState = 11829
    SystemOffNormalAlarmType_SuppressedState_FalseState = 11830
    SystemOffNormalAlarmType_ShelvingState = 11831
    SystemOffNormalAlarmType_ShelvingState_CurrentState = 11832
    SystemOffNormalAlarmType_ShelvingState_CurrentState_Id = 11833
    SystemOffNormalAlarmType_ShelvingState_CurrentState_Name = 11834
    SystemOffNormalAlarmType_ShelvingState_CurrentState_Number = 11835
    SystemOffNormalAlarmType_ShelvingState_CurrentState_EffectiveDisplayName = 11836
    SystemOffNormalAlarmType_ShelvingState_LastTransition = 11837
    SystemOffNormalAlarmType_ShelvingState_LastTransition_Id = 11838
    SystemOffNormalAlarmType_ShelvingState_LastTransition_Name = 11839
    SystemOffNormalAlarmType_ShelvingState_LastTransition_Number = 11840
    SystemOffNormalAlarmType_ShelvingState_LastTransition_TransitionTime = 11841
    SystemOffNormalAlarmType_ShelvingState_LastTransition_EffectiveTransitionTime = 11842
    SystemOffNormalAlarmType_ShelvingState_UnshelveTime = 11843
    SystemOffNormalAlarmType_ShelvingState_Unshelve = 11844
    SystemOffNormalAlarmType_ShelvingState_OneShotShelve = 11845
    SystemOffNormalAlarmType_ShelvingState_TimedShelve = 11846
    SystemOffNormalAlarmType_ShelvingState_TimedShelve_InputArguments = 11847
    SystemOffNormalAlarmType_SuppressedOrShelved = 11848
    SystemOffNormalAlarmType_MaxTimeShelved = 11849
    SystemOffNormalAlarmType_NormalState = 11850
    AuditConditionCommentEventType_Comment = 11851
    AuditConditionRespondEventType_SelectedResponse = 11852
    AuditConditionAcknowledgeEventType_Comment = 11853
    AuditConditionConfirmEventType_Comment = 11854
    AuditConditionShelvingEventType_ShelvingTime = 11855
    AuditProgramTransitionEventType = 11856
    AuditProgramTransitionEventType_EventId = 11857
    AuditProgramTransitionEventType_EventType = 11858
    AuditProgramTransitionEventType_SourceNode = 11859
    AuditProgramTransitionEventType_SourceName = 11860
    AuditProgramTransitionEventType_Time = 11861
    AuditProgramTransitionEventType_ReceiveTime = 11862
    AuditProgramTransitionEventType_LocalTime = 11863
    AuditProgramTransitionEventType_Message = 11864
    AuditProgramTransitionEventType_Severity = 11865
    AuditProgramTransitionEventType_ActionTimeStamp = 11866
    AuditProgramTransitionEventType_Status = 11867
    AuditProgramTransitionEventType_ServerId = 11868
    AuditProgramTransitionEventType_ClientAuditEntryId = 11869
    AuditProgramTransitionEventType_ClientUserId = 11870
    AuditProgramTransitionEventType_MethodId = 11871
    AuditProgramTransitionEventType_InputArguments = 11872
    AuditProgramTransitionEventType_OldStateId = 11873
    AuditProgramTransitionEventType_NewStateId = 11874
    AuditProgramTransitionEventType_TransitionNumber = 11875
    HistoricalDataConfigurationType_AggregateFunctions = 11876
    HAConfiguration_AggregateFunctions = 11877
    NodeClass_EnumValues = 11878
    InstanceNode = 11879
    TypeNode = 11880
    NodeAttributesMask_EnumValues = 11881
    AttributeWriteMask_EnumValues = 11882
    BrowseResultMask_EnumValues = 11883
    HistoryUpdateType_EnumValues = 11884
    PerformUpdateType_EnumValues = 11885
    EnumeratedTestType_EnumValues = 11886
    InstanceNode_Encoding_DefaultXml = 11887
    TypeNode_Encoding_DefaultXml = 11888
    InstanceNode_Encoding_DefaultBinary = 11889
    TypeNode_Encoding_DefaultBinary = 11890
    SessionDiagnosticsObjectType_SessionDiagnostics_UnauthorizedRequestCount = 11891
    SessionDiagnosticsVariableType_UnauthorizedRequestCount = 11892
    OpenFileMode = 11939
    OpenFileMode_EnumValues = 11940
    ModelChangeStructureVerbMask = 11941
    ModelChangeStructureVerbMask_EnumValues = 11942
    EndpointUrlListDataType = 11943
    NetworkGroupDataType = 11944
    NonTransparentNetworkRedundancyType = 11945
    NonTransparentNetworkRedundancyType_RedundancySupport = 11946
    NonTransparentNetworkRedundancyType_ServerUriArray = 11947
    NonTransparentNetworkRedundancyType_ServerNetworkGroups = 11948
    EndpointUrlListDataType_Encoding_DefaultXml = 11949
    NetworkGroupDataType_Encoding_DefaultXml = 11950
    OpcUa_XmlSchema_EndpointUrlListDataType = 11951
    OpcUa_XmlSchema_EndpointUrlListDataType_DataTypeVersion = 11952
    OpcUa_XmlSchema_EndpointUrlListDataType_DictionaryFragment = 11953
    OpcUa_XmlSchema_NetworkGroupDataType = 11954
    OpcUa_XmlSchema_NetworkGroupDataType_DataTypeVersion = 11955
    OpcUa_XmlSchema_NetworkGroupDataType_DictionaryFragment = 11956
    EndpointUrlListDataType_Encoding_DefaultBinary = 11957
    NetworkGroupDataType_Encoding_DefaultBinary = 11958
    OpcUa_BinarySchema_EndpointUrlListDataType = 11959
    OpcUa_BinarySchema_EndpointUrlListDataType_DataTypeVersion = 11960
    OpcUa_BinarySchema_EndpointUrlListDataType_DictionaryFragment = 11961
    OpcUa_BinarySchema_NetworkGroupDataType = 11962
    OpcUa_BinarySchema_NetworkGroupDataType_DataTypeVersion = 11963
    OpcUa_BinarySchema_NetworkGroupDataType_DictionaryFragment = 11964
    ArrayItemType = 12021
    ArrayItemType_Definition = 12022
    ArrayItemType_ValuePrecision = 12023
    ArrayItemType_InstrumentRange = 12024
    ArrayItemType_EURange = 12025
    ArrayItemType_EngineeringUnits = 12026
    ArrayItemType_Title = 12027
    ArrayItemType_AxisScaleType = 12028
    YArrayItemType = 12029
    YArrayItemType_Definition = 12030
    YArrayItemType_ValuePrecision = 12031
    YArrayItemType_InstrumentRange = 12032
    YArrayItemType_EURange = 12033
    YArrayItemType_EngineeringUnits = 12034
    YArrayItemType_Title = 12035
    YArrayItemType_AxisScaleType = 12036
    YArrayItemType_XAxisDefinition = 12037
    XYArrayItemType = 12038
    XYArrayItemType_Definition = 12039
    XYArrayItemType_ValuePrecision = 12040
    XYArrayItemType_InstrumentRange = 12041
    XYArrayItemType_EURange = 12042
    XYArrayItemType_EngineeringUnits = 12043
    XYArrayItemType_Title = 12044
    XYArrayItemType_AxisScaleType = 12045
    XYArrayItemType_XAxisDefinition = 12046
    ImageItemType = 12047
    ImageItemType_Definition = 12048
    ImageItemType_ValuePrecision = 12049
    ImageItemType_InstrumentRange = 12050
    ImageItemType_EURange = 12051
    ImageItemType_EngineeringUnits = 12052
    ImageItemType_Title = 12053
    ImageItemType_AxisScaleType = 12054
    ImageItemType_XAxisDefinition = 12055
    ImageItemType_YAxisDefinition = 12056
    CubeItemType = 12057
    CubeItemType_Definition = 12058
    CubeItemType_ValuePrecision = 12059
    CubeItemType_InstrumentRange = 12060
    CubeItemType_EURange = 12061
    CubeItemType_EngineeringUnits = 12062
    CubeItemType_Title = 12063
    CubeItemType_AxisScaleType = 12064
    CubeItemType_XAxisDefinition = 12065
    CubeItemType_YAxisDefinition = 12066
    CubeItemType_ZAxisDefinition = 12067
    NDimensionArrayItemType = 12068
    NDimensionArrayItemType_Definition = 12069
    NDimensionArrayItemType_ValuePrecision = 12070
    NDimensionArrayItemType_InstrumentRange = 12071
    NDimensionArrayItemType_EURange = 12072
    NDimensionArrayItemType_EngineeringUnits = 12073
    NDimensionArrayItemType_Title = 12074
    NDimensionArrayItemType_AxisScaleType = 12075
    NDimensionArrayItemType_AxisDefinition = 12076
    AxisScaleEnumeration = 12077
    AxisScaleEnumeration_EnumStrings = 12078
    AxisInformation = 12079
    XVType = 12080
    AxisInformation_Encoding_DefaultXml = 12081
    XVType_Encoding_DefaultXml = 12082
    OpcUa_XmlSchema_AxisInformation = 12083
    OpcUa_XmlSchema_AxisInformation_DataTypeVersion = 12084
    OpcUa_XmlSchema_AxisInformation_DictionaryFragment = 12085
    OpcUa_XmlSchema_XVType = 12086
    OpcUa_XmlSchema_XVType_DataTypeVersion = 12087
    OpcUa_XmlSchema_XVType_DictionaryFragment = 12088
    AxisInformation_Encoding_DefaultBinary = 12089
    XVType_Encoding_DefaultBinary = 12090
    OpcUa_BinarySchema_AxisInformation = 12091
    OpcUa_BinarySchema_AxisInformation_DataTypeVersion = 12092
    OpcUa_BinarySchema_AxisInformation_DictionaryFragment = 12093
    OpcUa_BinarySchema_XVType = 12094
    OpcUa_BinarySchema_XVType_DataTypeVersion = 12095
    OpcUa_BinarySchema_XVType_DictionaryFragment = 12096
    SessionsDiagnosticsSummaryType_SessionPlaceholder = 12097
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics = 12098
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_SessionId = 12099
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_SessionName = 12100
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_ClientDescription = 12101
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_ServerUri = 12102
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_EndpointUrl = 12103
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_LocaleIds = 12104
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_ActualSessionTimeout = 12105
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_MaxResponseMessageSize = 12106
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_ClientConnectionTime = 12107
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_ClientLastContactTime = 12108
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_CurrentSubscriptionsCount = 12109
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_CurrentMonitoredItemsCount = 12110
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_CurrentPublishRequestsInQueue = 12111
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_TotalRequestCount = 12112
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_UnauthorizedRequestCount = 12113
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_ReadCount = 12114
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_HistoryReadCount = 12115
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_WriteCount = 12116
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_HistoryUpdateCount = 12117
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_CallCount = 12118
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_CreateMonitoredItemsCount = 12119
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_ModifyMonitoredItemsCount = 12120
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_SetMonitoringModeCount = 12121
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_SetTriggeringCount = 12122
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_DeleteMonitoredItemsCount = 12123
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_CreateSubscriptionCount = 12124
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_ModifySubscriptionCount = 12125
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_SetPublishingModeCount = 12126
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_PublishCount = 12127
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_RepublishCount = 12128
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_TransferSubscriptionsCount = 12129
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_DeleteSubscriptionsCount = 12130
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_AddNodesCount = 12131
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_AddReferencesCount = 12132
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_DeleteNodesCount = 12133
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_DeleteReferencesCount = 12134
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_BrowseCount = 12135
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_BrowseNextCount = 12136
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_TranslateBrowsePathsToNodeIdsCount = 12137
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_QueryFirstCount = 12138
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_QueryNextCount = 12139
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_RegisterNodesCount = 12140
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_UnregisterNodesCount = 12141
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionSecurityDiagnostics = 12142
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionSecurityDiagnostics_SessionId = 12143
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionSecurityDiagnostics_ClientUserIdOfSession = 12144
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionSecurityDiagnostics_ClientUserIdHistory = 12145
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionSecurityDiagnostics_AuthenticationMechanism = 12146
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionSecurityDiagnostics_Encoding = 12147
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionSecurityDiagnostics_TransportProtocol = 12148
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionSecurityDiagnostics_SecurityMode = 12149
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionSecurityDiagnostics_SecurityPolicyUri = 12150
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionSecurityDiagnostics_ClientCertificate = 12151
    SessionsDiagnosticsSummaryType_SessionPlaceholder_SubscriptionDiagnosticsArray = 12152
    ServerType_ServerCapabilities_OperationLimits_MaxNodesPerHistoryReadData = 12153
    ServerType_ServerCapabilities_OperationLimits_MaxNodesPerHistoryReadEvents = 12154
    ServerType_ServerCapabilities_OperationLimits_MaxNodesPerHistoryUpdateData = 12155
    ServerType_ServerCapabilities_OperationLimits_MaxNodesPerHistoryUpdateEvents = 12156
    ServerCapabilitiesType_OperationLimits_MaxNodesPerHistoryReadData = 12157
    ServerCapabilitiesType_OperationLimits_MaxNodesPerHistoryReadEvents = 12158
    ServerCapabilitiesType_OperationLimits_MaxNodesPerHistoryUpdateData = 12159
    ServerCapabilitiesType_OperationLimits_MaxNodesPerHistoryUpdateEvents = 12160
    OperationLimitsType_MaxNodesPerHistoryReadData = 12161
    OperationLimitsType_MaxNodesPerHistoryReadEvents = 12162
    OperationLimitsType_MaxNodesPerHistoryUpdateData = 12163
    OperationLimitsType_MaxNodesPerHistoryUpdateEvents = 12164
    Server_ServerCapabilities_OperationLimits_MaxNodesPerHistoryReadData = 12165
    Server_ServerCapabilities_OperationLimits_MaxNodesPerHistoryReadEvents = 12166
    Server_ServerCapabilities_OperationLimits_MaxNodesPerHistoryUpdateData = 12167
    Server_ServerCapabilities_OperationLimits_MaxNodesPerHistoryUpdateEvents = 12168
    NamingRuleType_EnumValues = 12169
    ViewVersion = 12170
    ComplexNumberType = 12171
    DoubleComplexNumberType = 12172
    ComplexNumberType_Encoding_DefaultXml = 12173
    DoubleComplexNumberType_Encoding_DefaultXml = 12174
    OpcUa_XmlSchema_ComplexNumberType = 12175
    OpcUa_XmlSchema_ComplexNumberType_DataTypeVersion = 12176
    OpcUa_XmlSchema_ComplexNumberType_DictionaryFragment = 12177
    OpcUa_XmlSchema_DoubleComplexNumberType = 12178
    OpcUa_XmlSchema_DoubleComplexNumberType_DataTypeVersion = 12179
    OpcUa_XmlSchema_DoubleComplexNumberType_DictionaryFragment = 12180
    ComplexNumberType_Encoding_DefaultBinary = 12181
    DoubleComplexNumberType_Encoding_DefaultBinary = 12182
    OpcUa_BinarySchema_ComplexNumberType = 12183
    OpcUa_BinarySchema_ComplexNumberType_DataTypeVersion = 12184
    OpcUa_BinarySchema_ComplexNumberType_DictionaryFragment = 12185
    OpcUa_BinarySchema_DoubleComplexNumberType = 12186
    OpcUa_BinarySchema_DoubleComplexNumberType_DataTypeVersion = 12187
    OpcUa_BinarySchema_DoubleComplexNumberType_DictionaryFragment = 12188
    ServerOnNetwork = 12189
    FindServersOnNetworkRequest = 12190
    FindServersOnNetworkResponse = 12191
    RegisterServer2Request = 12193
    RegisterServer2Response = 12194
    ServerOnNetwork_Encoding_DefaultXml = 12195
    FindServersOnNetworkRequest_Encoding_DefaultXml = 12196
    FindServersOnNetworkResponse_Encoding_DefaultXml = 12197
    RegisterServer2Request_Encoding_DefaultXml = 12199
    RegisterServer2Response_Encoding_DefaultXml = 12200
    OpcUa_XmlSchema_ServerOnNetwork = 12201
    OpcUa_XmlSchema_ServerOnNetwork_DataTypeVersion = 12202
    OpcUa_XmlSchema_ServerOnNetwork_DictionaryFragment = 12203
    ServerOnNetwork_Encoding_DefaultBinary = 12207
    FindServersOnNetworkRequest_Encoding_DefaultBinary = 12208
    FindServersOnNetworkResponse_Encoding_DefaultBinary = 12209
    RegisterServer2Request_Encoding_DefaultBinary = 12211
    RegisterServer2Response_Encoding_DefaultBinary = 12212
    OpcUa_BinarySchema_ServerOnNetwork = 12213
    OpcUa_BinarySchema_ServerOnNetwork_DataTypeVersion = 12214
    OpcUa_BinarySchema_ServerOnNetwork_DictionaryFragment = 12215
    ProgressEventType_Context = 12502
    ProgressEventType_Progress = 12503
    KerberosIdentityToken = 12504
    KerberosIdentityToken_Encoding_DefaultXml = 12505
    OpcUa_XmlSchema_KerberosIdentityToken = 12506
    OpcUa_XmlSchema_KerberosIdentityToken_DataTypeVersion = 12507
    OpcUa_XmlSchema_KerberosIdentityToken_DictionaryFragment = 12508
    KerberosIdentityToken_Encoding_DefaultBinary = 12509
    OpcUa_BinarySchema_KerberosIdentityToken = 12510
    OpcUa_BinarySchema_KerberosIdentityToken_DataTypeVersion = 12511
    OpcUa_BinarySchema_KerberosIdentityToken_DictionaryFragment = 12512
    OpenWithMasksMethodType = 12513
    OpenWithMasksMethodType_InputArguments = 12514
    OpenWithMasksMethodType_OutputArguments = 12515
    CloseAndUpdateMethodType = 12516
    CloseAndUpdateMethodType_OutputArguments = 12517
    AddCertificateMethodType = 12518
    AddCertificateMethodType_InputArguments = 12519
    RemoveCertificateMethodType = 12520
    RemoveCertificateMethodType_InputArguments = 12521
    TrustListType = 12522
    TrustListType_Size = 12523
    TrustListType_OpenCount = 12526
    TrustListType_Open = 12527
    TrustListType_Open_InputArguments = 12528
    TrustListType_Open_OutputArguments = 12529
    TrustListType_Close = 12530
    TrustListType_Close_InputArguments = 12531
    TrustListType_Read = 12532
    TrustListType_Read_InputArguments = 12533
    TrustListType_Read_OutputArguments = 12534
    TrustListType_Write = 12535
    TrustListType_Write_InputArguments = 12536
    TrustListType_GetPosition = 12537
    TrustListType_GetPosition_InputArguments = 12538
    TrustListType_GetPosition_OutputArguments = 12539
    TrustListType_SetPosition = 12540
    TrustListType_SetPosition_InputArguments = 12541
    TrustListType_LastUpdateTime = 12542
    TrustListType_OpenWithMasks = 12543
    TrustListType_OpenWithMasks_InputArguments = 12544
    TrustListType_OpenWithMasks_OutputArguments = 12545
    TrustListType_CloseAndUpdate = 12546
    TrustListType_CloseAndUpdate_OutputArguments = 12547
    TrustListType_AddCertificate = 12548
    TrustListType_AddCertificate_InputArguments = 12549
    TrustListType_RemoveCertificate = 12550
    TrustListType_RemoveCertificate_InputArguments = 12551
    TrustListMasks = 12552
    TrustListMasks_EnumValues = 12553
    TrustListDataType = 12554
    CertificateGroupType = 12555
    CertificateType = 12556
    ApplicationCertificateType = 12557
    HttpsCertificateType = 12558
    RsaMinApplicationCertificateType = 12559
    RsaSha256ApplicationCertificateType = 12560
    TrustListUpdatedAuditEventType = 12561
    TrustListUpdatedAuditEventType_EventId = 12562
    TrustListUpdatedAuditEventType_EventType = 12563
    TrustListUpdatedAuditEventType_SourceNode = 12564
    TrustListUpdatedAuditEventType_SourceName = 12565
    TrustListUpdatedAuditEventType_Time = 12566
    TrustListUpdatedAuditEventType_ReceiveTime = 12567
    TrustListUpdatedAuditEventType_LocalTime = 12568
    TrustListUpdatedAuditEventType_Message = 12569
    TrustListUpdatedAuditEventType_Severity = 12570
    TrustListUpdatedAuditEventType_ActionTimeStamp = 12571
    TrustListUpdatedAuditEventType_Status = 12572
    TrustListUpdatedAuditEventType_ServerId = 12573
    TrustListUpdatedAuditEventType_ClientAuditEntryId = 12574
    TrustListUpdatedAuditEventType_ClientUserId = 12575
    TrustListUpdatedAuditEventType_MethodId = 12576
    TrustListUpdatedAuditEventType_InputArguments = 12577
    UpdateCertificateMethodType = 12578
    UpdateCertificateMethodType_InputArguments = 12579
    UpdateCertificateMethodType_OutputArguments = 12580
    ServerConfigurationType = 12581
    ServerConfigurationType_SupportedPrivateKeyFormats = 12583
    ServerConfigurationType_MaxTrustListSize = 12584
    ServerConfigurationType_MulticastDnsEnabled = 12585
    ServerConfigurationType_UpdateCertificate = 12616
    ServerConfigurationType_UpdateCertificate_InputArguments = 12617
    ServerConfigurationType_UpdateCertificate_OutputArguments = 12618
    CertificateUpdatedAuditEventType = 12620
    CertificateUpdatedAuditEventType_EventId = 12621
    CertificateUpdatedAuditEventType_EventType = 12622
    CertificateUpdatedAuditEventType_SourceNode = 12623
    CertificateUpdatedAuditEventType_SourceName = 12624
    CertificateUpdatedAuditEventType_Time = 12625
    CertificateUpdatedAuditEventType_ReceiveTime = 12626
    CertificateUpdatedAuditEventType_LocalTime = 12627
    CertificateUpdatedAuditEventType_Message = 12628
    CertificateUpdatedAuditEventType_Severity = 12629
    CertificateUpdatedAuditEventType_ActionTimeStamp = 12630
    CertificateUpdatedAuditEventType_Status = 12631
    CertificateUpdatedAuditEventType_ServerId = 12632
    CertificateUpdatedAuditEventType_ClientAuditEntryId = 12633
    CertificateUpdatedAuditEventType_ClientUserId = 12634
    CertificateUpdatedAuditEventType_MethodId = 12635
    CertificateUpdatedAuditEventType_InputArguments = 12636
    ServerConfiguration = 12637
    ServerConfiguration_SupportedPrivateKeyFormats = 12639
    ServerConfiguration_MaxTrustListSize = 12640
    ServerConfiguration_MulticastDnsEnabled = 12641
    ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList = 12642
    ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_Size = 12643
    ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_OpenCount = 12646
    ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_Open = 12647
    ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_Open_InputArguments = 12648
    ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_Open_OutputArguments = 12649
    ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_Close = 12650
    ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_Close_InputArguments = 12651
    ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_Read = 12652
    ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_Read_InputArguments = 12653
    ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_Read_OutputArguments = 12654
    ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_Write = 12655
    ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_Write_InputArguments = 12656
    ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_GetPosition = 12657
    ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_GetPosition_InputArguments = 12658
    ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_GetPosition_OutputArguments = 12659
    ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_SetPosition = 12660
    ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_SetPosition_InputArguments = 12661
    ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_LastUpdateTime = 12662
    ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_OpenWithMasks = 12663
    ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_OpenWithMasks_InputArguments = 12664
    ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_OpenWithMasks_OutputArguments = 12665
    ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_CloseAndUpdate = 12666
    ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_CloseAndUpdate_OutputArguments = 12667
    ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_AddCertificate = 12668
    ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_AddCertificate_InputArguments = 12669
    ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_RemoveCertificate = 12670
    ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_RemoveCertificate_InputArguments = 12671
    TrustListDataType_Encoding_DefaultXml = 12676
    OpcUa_XmlSchema_TrustListDataType = 12677
    OpcUa_XmlSchema_TrustListDataType_DataTypeVersion = 12678
    OpcUa_XmlSchema_TrustListDataType_DictionaryFragment = 12679
    TrustListDataType_Encoding_DefaultBinary = 12680
    OpcUa_BinarySchema_TrustListDataType = 12681
    OpcUa_BinarySchema_TrustListDataType_DataTypeVersion = 12682
    OpcUa_BinarySchema_TrustListDataType_DictionaryFragment = 12683
    ServerType_Namespaces_AddressSpaceFile_Writable = 12684
    ServerType_Namespaces_AddressSpaceFile_UserWritable = 12685
    FileType_Writable = 12686
    FileType_UserWritable = 12687
    AddressSpaceFileType_Writable = 12688
    AddressSpaceFileType_UserWritable = 12689
    NamespaceMetadataType_NamespaceFile_Writable = 12690
    NamespaceMetadataType_NamespaceFile_UserWritable = 12691
    NamespacesType_NamespaceIdentifier_NamespaceFile_Writable = 12692
    NamespacesType_NamespaceIdentifier_NamespaceFile_UserWritable = 12693
    NamespacesType_AddressSpaceFile_Writable = 12694
    NamespacesType_AddressSpaceFile_UserWritable = 12695
    Server_Namespaces_AddressSpaceFile_Writable = 12696
    Server_Namespaces_AddressSpaceFile_UserWritable = 12697
    TrustListType_Writable = 12698
    TrustListType_UserWritable = 12699
    CloseAndUpdateMethodType_InputArguments = 12704
    TrustListType_CloseAndUpdate_InputArguments = 12705
    ServerConfigurationType_ServerCapabilities = 12708
    ServerConfiguration_ServerCapabilities = 12710
    OpcUa_XmlSchema_RelativePathElement = 12712
    OpcUa_XmlSchema_RelativePathElement_DataTypeVersion = 12713
    OpcUa_XmlSchema_RelativePathElement_DictionaryFragment = 12714
    OpcUa_XmlSchema_RelativePath = 12715
    OpcUa_XmlSchema_RelativePath_DataTypeVersion = 12716
    OpcUa_XmlSchema_RelativePath_DictionaryFragment = 12717
    OpcUa_BinarySchema_RelativePathElement = 12718
    OpcUa_BinarySchema_RelativePathElement_DataTypeVersion = 12719
    OpcUa_BinarySchema_RelativePathElement_DictionaryFragment = 12720
    OpcUa_BinarySchema_RelativePath = 12721
    OpcUa_BinarySchema_RelativePath_DataTypeVersion = 12722
    OpcUa_BinarySchema_RelativePath_DictionaryFragment = 12723
    ServerConfigurationType_CreateSigningRequest = 12731
    ServerConfigurationType_CreateSigningRequest_InputArguments = 12732
    ServerConfigurationType_CreateSigningRequest_OutputArguments = 12733
    ServerConfigurationType_ApplyChanges = 12734
    ServerConfiguration_CreateSigningRequest = 12737
    ServerConfiguration_CreateSigningRequest_InputArguments = 12738
    ServerConfiguration_CreateSigningRequest_OutputArguments = 12739
    ServerConfiguration_ApplyChanges = 12740
    CreateSigningRequestMethodType = 12741
    CreateSigningRequestMethodType_InputArguments = 12742
    CreateSigningRequestMethodType_OutputArguments = 12743
    OptionSetValues = 12745
    ServerType_SetSubscriptionDurable = 12746
    ServerType_SetSubscriptionDurable_InputArguments = 12747
    ServerType_SetSubscriptionDurable_OutputArguments = 12748
    Server_SetSubscriptionDurable = 12749
    Server_SetSubscriptionDurable_InputArguments = 12750
    Server_SetSubscriptionDurable_OutputArguments = 12751
    SetSubscriptionDurableMethodType = 12752
    SetSubscriptionDurableMethodType_InputArguments = 12753
    SetSubscriptionDurableMethodType_OutputArguments = 12754
    OptionSet = 12755
    Union = 12756
    OptionSet_Encoding_DefaultXml = 12757
    Union_Encoding_DefaultXml = 12758
    OpcUa_XmlSchema_OptionSet = 12759
    OpcUa_XmlSchema_OptionSet_DataTypeVersion = 12760
    OpcUa_XmlSchema_OptionSet_DictionaryFragment = 12761
    OpcUa_XmlSchema_Union = 12762
    OpcUa_XmlSchema_Union_DataTypeVersion = 12763
    OpcUa_XmlSchema_Union_DictionaryFragment = 12764
    OptionSet_Encoding_DefaultBinary = 12765
    Union_Encoding_DefaultBinary = 12766
    OpcUa_BinarySchema_OptionSet = 12767
    OpcUa_BinarySchema_OptionSet_DataTypeVersion = 12768
    OpcUa_BinarySchema_OptionSet_DictionaryFragment = 12769
    OpcUa_BinarySchema_Union = 12770
    OpcUa_BinarySchema_Union_DataTypeVersion = 12771
    OpcUa_BinarySchema_Union_DictionaryFragment = 12772
    GetRejectedListMethodType = 12773
    GetRejectedListMethodType_OutputArguments = 12774
    ServerConfigurationType_GetRejectedList = 12775
    ServerConfigurationType_GetRejectedList_OutputArguments = 12776
    ServerConfiguration_GetRejectedList = 12777
    ServerConfiguration_GetRejectedList_OutputArguments = 12778
    SamplingIntervalDiagnosticsArrayType_SamplingIntervalDiagnostics = 12779
    SamplingIntervalDiagnosticsArrayType_SamplingIntervalDiagnostics_SamplingInterval = 12780
    SamplingIntervalDiagnosticsArrayType_SamplingIntervalDiagnostics_SampledMonitoredItemsCount = 12781
    SamplingIntervalDiagnosticsArrayType_SamplingIntervalDiagnostics_MaxSampledMonitoredItemsCount = 12782
    SamplingIntervalDiagnosticsArrayType_SamplingIntervalDiagnostics_DisabledMonitoredItemsSamplingCount = 12783
    SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics = 12784
    SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_SessionId = 12785
    SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_SubscriptionId = 12786
    SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_Priority = 12787
    SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_PublishingInterval = 12788
    SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_MaxKeepAliveCount = 12789
    SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_MaxLifetimeCount = 12790
    SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_MaxNotificationsPerPublish = 12791
    SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_PublishingEnabled = 12792
    SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_ModifyCount = 12793
    SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_EnableCount = 12794
    SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_DisableCount = 12795
    SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_RepublishRequestCount = 12796
    SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_RepublishMessageRequestCount = 12797
    SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_RepublishMessageCount = 12798
    SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_TransferRequestCount = 12799
    SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_TransferredToAltClientCount = 12800
    SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_TransferredToSameClientCount = 12801
    SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_PublishRequestCount = 12802
    SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_DataChangeNotificationsCount = 12803
    SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_EventNotificationsCount = 12804
    SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_NotificationsCount = 12805
    SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_LatePublishRequestCount = 12806
    SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_CurrentKeepAliveCount = 12807
    SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_CurrentLifetimeCount = 12808
    SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_UnacknowledgedMessageCount = 12809
    SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_DiscardedMessageCount = 12810
    SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_MonitoredItemCount = 12811
    SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_DisabledMonitoredItemCount = 12812
    SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_MonitoringQueueOverflowCount = 12813
    SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_NextSequenceNumber = 12814
    SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_EventQueueOverFlowCount = 12815
    SessionDiagnosticsArrayType_SessionDiagnostics = 12816
    SessionDiagnosticsArrayType_SessionDiagnostics_SessionId = 12817
    SessionDiagnosticsArrayType_SessionDiagnostics_SessionName = 12818
    SessionDiagnosticsArrayType_SessionDiagnostics_ClientDescription = 12819
    SessionDiagnosticsArrayType_SessionDiagnostics_ServerUri = 12820
    SessionDiagnosticsArrayType_SessionDiagnostics_EndpointUrl = 12821
    SessionDiagnosticsArrayType_SessionDiagnostics_LocaleIds = 12822
    SessionDiagnosticsArrayType_SessionDiagnostics_ActualSessionTimeout = 12823
    SessionDiagnosticsArrayType_SessionDiagnostics_MaxResponseMessageSize = 12824
    SessionDiagnosticsArrayType_SessionDiagnostics_ClientConnectionTime = 12825
    SessionDiagnosticsArrayType_SessionDiagnostics_ClientLastContactTime = 12826
    SessionDiagnosticsArrayType_SessionDiagnostics_CurrentSubscriptionsCount = 12827
    SessionDiagnosticsArrayType_SessionDiagnostics_CurrentMonitoredItemsCount = 12828
    SessionDiagnosticsArrayType_SessionDiagnostics_CurrentPublishRequestsInQueue = 12829
    SessionDiagnosticsArrayType_SessionDiagnostics_TotalRequestCount = 12830
    SessionDiagnosticsArrayType_SessionDiagnostics_UnauthorizedRequestCount = 12831
    SessionDiagnosticsArrayType_SessionDiagnostics_ReadCount = 12832
    SessionDiagnosticsArrayType_SessionDiagnostics_HistoryReadCount = 12833
    SessionDiagnosticsArrayType_SessionDiagnostics_WriteCount = 12834
    SessionDiagnosticsArrayType_SessionDiagnostics_HistoryUpdateCount = 12835
    SessionDiagnosticsArrayType_SessionDiagnostics_CallCount = 12836
    SessionDiagnosticsArrayType_SessionDiagnostics_CreateMonitoredItemsCount = 12837
    SessionDiagnosticsArrayType_SessionDiagnostics_ModifyMonitoredItemsCount = 12838
    SessionDiagnosticsArrayType_SessionDiagnostics_SetMonitoringModeCount = 12839
    SessionDiagnosticsArrayType_SessionDiagnostics_SetTriggeringCount = 12840
    SessionDiagnosticsArrayType_SessionDiagnostics_DeleteMonitoredItemsCount = 12841
    SessionDiagnosticsArrayType_SessionDiagnostics_CreateSubscriptionCount = 12842
    SessionDiagnosticsArrayType_SessionDiagnostics_ModifySubscriptionCount = 12843
    SessionDiagnosticsArrayType_SessionDiagnostics_SetPublishingModeCount = 12844
    SessionDiagnosticsArrayType_SessionDiagnostics_PublishCount = 12845
    SessionDiagnosticsArrayType_SessionDiagnostics_RepublishCount = 12846
    SessionDiagnosticsArrayType_SessionDiagnostics_TransferSubscriptionsCount = 12847
    SessionDiagnosticsArrayType_SessionDiagnostics_DeleteSubscriptionsCount = 12848
    SessionDiagnosticsArrayType_SessionDiagnostics_AddNodesCount = 12849
    SessionDiagnosticsArrayType_SessionDiagnostics_AddReferencesCount = 12850
    SessionDiagnosticsArrayType_SessionDiagnostics_DeleteNodesCount = 12851
    SessionDiagnosticsArrayType_SessionDiagnostics_DeleteReferencesCount = 12852
    SessionDiagnosticsArrayType_SessionDiagnostics_BrowseCount = 12853
    SessionDiagnosticsArrayType_SessionDiagnostics_BrowseNextCount = 12854
    SessionDiagnosticsArrayType_SessionDiagnostics_TranslateBrowsePathsToNodeIdsCount = 12855
    SessionDiagnosticsArrayType_SessionDiagnostics_QueryFirstCount = 12856
    SessionDiagnosticsArrayType_SessionDiagnostics_QueryNextCount = 12857
    SessionDiagnosticsArrayType_SessionDiagnostics_RegisterNodesCount = 12858
    SessionDiagnosticsArrayType_SessionDiagnostics_UnregisterNodesCount = 12859
    SessionSecurityDiagnosticsArrayType_SessionSecurityDiagnostics = 12860
    SessionSecurityDiagnosticsArrayType_SessionSecurityDiagnostics_SessionId = 12861
    SessionSecurityDiagnosticsArrayType_SessionSecurityDiagnostics_ClientUserIdOfSession = 12862
    SessionSecurityDiagnosticsArrayType_SessionSecurityDiagnostics_ClientUserIdHistory = 12863
    SessionSecurityDiagnosticsArrayType_SessionSecurityDiagnostics_AuthenticationMechanism = 12864
    SessionSecurityDiagnosticsArrayType_SessionSecurityDiagnostics_Encoding = 12865
    SessionSecurityDiagnosticsArrayType_SessionSecurityDiagnostics_TransportProtocol = 12866
    SessionSecurityDiagnosticsArrayType_SessionSecurityDiagnostics_SecurityMode = 12867
    SessionSecurityDiagnosticsArrayType_SessionSecurityDiagnostics_SecurityPolicyUri = 12868
    SessionSecurityDiagnosticsArrayType_SessionSecurityDiagnostics_ClientCertificate = 12869
    ServerType_ResendData = 12871
    ServerType_ResendData_InputArguments = 12872
    Server_ResendData = 12873
    Server_ResendData_InputArguments = 12874
    ResendDataMethodType = 12875
    ResendDataMethodType_InputArguments = 12876
    NormalizedString = 12877
    DecimalString = 12878
    DurationString = 12879
    TimeString = 12880
    DateString = 12881
    ServerType_EstimatedReturnTime = 12882
    ServerType_RequestServerStateChange = 12883
    ServerType_RequestServerStateChange_InputArguments = 12884
    Server_EstimatedReturnTime = 12885
    Server_RequestServerStateChange = 12886
    Server_RequestServerStateChange_InputArguments = 12887
    RequestServerStateChangeMethodType = 12888
    RequestServerStateChangeMethodType_InputArguments = 12889
    DiscoveryConfiguration = 12890
    MdnsDiscoveryConfiguration = 12891
    DiscoveryConfiguration_Encoding_DefaultXml = 12892
    MdnsDiscoveryConfiguration_Encoding_DefaultXml = 12893
    OpcUa_XmlSchema_DiscoveryConfiguration = 12894
    OpcUa_XmlSchema_DiscoveryConfiguration_DataTypeVersion = 12895
    OpcUa_XmlSchema_DiscoveryConfiguration_DictionaryFragment = 12896
    OpcUa_XmlSchema_MdnsDiscoveryConfiguration = 12897
    OpcUa_XmlSchema_MdnsDiscoveryConfiguration_DataTypeVersion = 12898
    OpcUa_XmlSchema_MdnsDiscoveryConfiguration_DictionaryFragment = 12899
    DiscoveryConfiguration_Encoding_DefaultBinary = 12900
    MdnsDiscoveryConfiguration_Encoding_DefaultBinary = 12901
    OpcUa_BinarySchema_DiscoveryConfiguration = 12902
    OpcUa_BinarySchema_DiscoveryConfiguration_DataTypeVersion = 12903
    OpcUa_BinarySchema_DiscoveryConfiguration_DictionaryFragment = 12904
    OpcUa_BinarySchema_MdnsDiscoveryConfiguration = 12905
    OpcUa_BinarySchema_MdnsDiscoveryConfiguration_DataTypeVersion = 12906
    OpcUa_BinarySchema_MdnsDiscoveryConfiguration_DictionaryFragment = 12907
    MaxByteStringLength = 12908
    ServerType_ServerCapabilities_MaxByteStringLength = 12909
    ServerCapabilitiesType_MaxByteStringLength = 12910
    Server_ServerCapabilities_MaxByteStringLength = 12911
    ConditionType_ConditionRefresh2 = 12912
    ConditionType_ConditionRefresh2_InputArguments = 12913
    ConditionRefresh2MethodType = 12914
    ConditionRefresh2MethodType_InputArguments = 12915
    DialogConditionType_ConditionRefresh2 = 12916
    DialogConditionType_ConditionRefresh2_InputArguments = 12917
    AcknowledgeableConditionType_ConditionRefresh2 = 12918
    AcknowledgeableConditionType_ConditionRefresh2_InputArguments = 12919
    AlarmConditionType_ConditionRefresh2 = 12984
    AlarmConditionType_ConditionRefresh2_InputArguments = 12985
    LimitAlarmType_ConditionRefresh2 = 12986
    LimitAlarmType_ConditionRefresh2_InputArguments = 12987
    ExclusiveLimitAlarmType_ConditionRefresh2 = 12988
    ExclusiveLimitAlarmType_ConditionRefresh2_InputArguments = 12989
    NonExclusiveLimitAlarmType_ConditionRefresh2 = 12990
    NonExclusiveLimitAlarmType_ConditionRefresh2_InputArguments = 12991
    NonExclusiveLevelAlarmType_ConditionRefresh2 = 12992
    NonExclusiveLevelAlarmType_ConditionRefresh2_InputArguments = 12993
    ExclusiveLevelAlarmType_ConditionRefresh2 = 12994
    ExclusiveLevelAlarmType_ConditionRefresh2_InputArguments = 12995
    NonExclusiveDeviationAlarmType_ConditionRefresh2 = 12996
    NonExclusiveDeviationAlarmType_ConditionRefresh2_InputArguments = 12997
    ExclusiveDeviationAlarmType_ConditionRefresh2 = 12998
    ExclusiveDeviationAlarmType_ConditionRefresh2_InputArguments = 12999
    NonExclusiveRateOfChangeAlarmType_ConditionRefresh2 = 13000
    NonExclusiveRateOfChangeAlarmType_ConditionRefresh2_InputArguments = 13001
    ExclusiveRateOfChangeAlarmType_ConditionRefresh2 = 13002
    ExclusiveRateOfChangeAlarmType_ConditionRefresh2_InputArguments = 13003
    DiscreteAlarmType_ConditionRefresh2 = 13004
    DiscreteAlarmType_ConditionRefresh2_InputArguments = 13005
    OffNormalAlarmType_ConditionRefresh2 = 13006
    OffNormalAlarmType_ConditionRefresh2_InputArguments = 13007
    SystemOffNormalAlarmType_ConditionRefresh2 = 13008
    SystemOffNormalAlarmType_ConditionRefresh2_InputArguments = 13009
    TripAlarmType_ConditionRefresh2 = 13010
    TripAlarmType_ConditionRefresh2_InputArguments = 13011
    CertificateExpirationAlarmType = 13225
    CertificateExpirationAlarmType_EventId = 13226
    CertificateExpirationAlarmType_EventType = 13227
    CertificateExpirationAlarmType_SourceNode = 13228
    CertificateExpirationAlarmType_SourceName = 13229
    CertificateExpirationAlarmType_Time = 13230
    CertificateExpirationAlarmType_ReceiveTime = 13231
    CertificateExpirationAlarmType_LocalTime = 13232
    CertificateExpirationAlarmType_Message = 13233
    CertificateExpirationAlarmType_Severity = 13234
    CertificateExpirationAlarmType_ConditionClassId = 13235
    CertificateExpirationAlarmType_ConditionClassName = 13236
    CertificateExpirationAlarmType_ConditionName = 13237
    CertificateExpirationAlarmType_BranchId = 13238
    CertificateExpirationAlarmType_Retain = 13239
    CertificateExpirationAlarmType_EnabledState = 13240
    CertificateExpirationAlarmType_EnabledState_Id = 13241
    CertificateExpirationAlarmType_EnabledState_Name = 13242
    CertificateExpirationAlarmType_EnabledState_Number = 13243
    CertificateExpirationAlarmType_EnabledState_EffectiveDisplayName = 13244
    CertificateExpirationAlarmType_EnabledState_TransitionTime = 13245
    CertificateExpirationAlarmType_EnabledState_EffectiveTransitionTime = 13246
    CertificateExpirationAlarmType_EnabledState_TrueState = 13247
    CertificateExpirationAlarmType_EnabledState_FalseState = 13248
    CertificateExpirationAlarmType_Quality = 13249
    CertificateExpirationAlarmType_Quality_SourceTimestamp = 13250
    CertificateExpirationAlarmType_LastSeverity = 13251
    CertificateExpirationAlarmType_LastSeverity_SourceTimestamp = 13252
    CertificateExpirationAlarmType_Comment = 13253
    CertificateExpirationAlarmType_Comment_SourceTimestamp = 13254
    CertificateExpirationAlarmType_ClientUserId = 13255
    CertificateExpirationAlarmType_Disable = 13256
    CertificateExpirationAlarmType_Enable = 13257
    CertificateExpirationAlarmType_AddComment = 13258
    CertificateExpirationAlarmType_AddComment_InputArguments = 13259
    CertificateExpirationAlarmType_ConditionRefresh = 13260
    CertificateExpirationAlarmType_ConditionRefresh_InputArguments = 13261
    CertificateExpirationAlarmType_ConditionRefresh2 = 13262
    CertificateExpirationAlarmType_ConditionRefresh2_InputArguments = 13263
    CertificateExpirationAlarmType_AckedState = 13264
    CertificateExpirationAlarmType_AckedState_Id = 13265
    CertificateExpirationAlarmType_AckedState_Name = 13266
    CertificateExpirationAlarmType_AckedState_Number = 13267
    CertificateExpirationAlarmType_AckedState_EffectiveDisplayName = 13268
    CertificateExpirationAlarmType_AckedState_TransitionTime = 13269
    CertificateExpirationAlarmType_AckedState_EffectiveTransitionTime = 13270
    CertificateExpirationAlarmType_AckedState_TrueState = 13271
    CertificateExpirationAlarmType_AckedState_FalseState = 13272
    CertificateExpirationAlarmType_ConfirmedState = 13273
    CertificateExpirationAlarmType_ConfirmedState_Id = 13274
    CertificateExpirationAlarmType_ConfirmedState_Name = 13275
    CertificateExpirationAlarmType_ConfirmedState_Number = 13276
    CertificateExpirationAlarmType_ConfirmedState_EffectiveDisplayName = 13277
    CertificateExpirationAlarmType_ConfirmedState_TransitionTime = 13278
    CertificateExpirationAlarmType_ConfirmedState_EffectiveTransitionTime = 13279
    CertificateExpirationAlarmType_ConfirmedState_TrueState = 13280
    CertificateExpirationAlarmType_ConfirmedState_FalseState = 13281
    CertificateExpirationAlarmType_Acknowledge = 13282
    CertificateExpirationAlarmType_Acknowledge_InputArguments = 13283
    CertificateExpirationAlarmType_Confirm = 13284
    CertificateExpirationAlarmType_Confirm_InputArguments = 13285
    CertificateExpirationAlarmType_ActiveState = 13286
    CertificateExpirationAlarmType_ActiveState_Id = 13287
    CertificateExpirationAlarmType_ActiveState_Name = 13288
    CertificateExpirationAlarmType_ActiveState_Number = 13289
    CertificateExpirationAlarmType_ActiveState_EffectiveDisplayName = 13290
    CertificateExpirationAlarmType_ActiveState_TransitionTime = 13291
    CertificateExpirationAlarmType_ActiveState_EffectiveTransitionTime = 13292
    CertificateExpirationAlarmType_ActiveState_TrueState = 13293
    CertificateExpirationAlarmType_ActiveState_FalseState = 13294
    CertificateExpirationAlarmType_InputNode = 13295
    CertificateExpirationAlarmType_SuppressedState = 13296
    CertificateExpirationAlarmType_SuppressedState_Id = 13297
    CertificateExpirationAlarmType_SuppressedState_Name = 13298
    CertificateExpirationAlarmType_SuppressedState_Number = 13299
    CertificateExpirationAlarmType_SuppressedState_EffectiveDisplayName = 13300
    CertificateExpirationAlarmType_SuppressedState_TransitionTime = 13301
    CertificateExpirationAlarmType_SuppressedState_EffectiveTransitionTime = 13302
    CertificateExpirationAlarmType_SuppressedState_TrueState = 13303
    CertificateExpirationAlarmType_SuppressedState_FalseState = 13304
    CertificateExpirationAlarmType_ShelvingState = 13305
    CertificateExpirationAlarmType_ShelvingState_CurrentState = 13306
    CertificateExpirationAlarmType_ShelvingState_CurrentState_Id = 13307
    CertificateExpirationAlarmType_ShelvingState_CurrentState_Name = 13308
    CertificateExpirationAlarmType_ShelvingState_CurrentState_Number = 13309
    CertificateExpirationAlarmType_ShelvingState_CurrentState_EffectiveDisplayName = 13310
    CertificateExpirationAlarmType_ShelvingState_LastTransition = 13311
    CertificateExpirationAlarmType_ShelvingState_LastTransition_Id = 13312
    CertificateExpirationAlarmType_ShelvingState_LastTransition_Name = 13313
    CertificateExpirationAlarmType_ShelvingState_LastTransition_Number = 13314
    CertificateExpirationAlarmType_ShelvingState_LastTransition_TransitionTime = 13315
    CertificateExpirationAlarmType_ShelvingState_LastTransition_EffectiveTransitionTime = 13316
    CertificateExpirationAlarmType_ShelvingState_UnshelveTime = 13317
    CertificateExpirationAlarmType_ShelvingState_Unshelve = 13318
    CertificateExpirationAlarmType_ShelvingState_OneShotShelve = 13319
    CertificateExpirationAlarmType_ShelvingState_TimedShelve = 13320
    CertificateExpirationAlarmType_ShelvingState_TimedShelve_InputArguments = 13321
    CertificateExpirationAlarmType_SuppressedOrShelved = 13322
    CertificateExpirationAlarmType_MaxTimeShelved = 13323
    CertificateExpirationAlarmType_NormalState = 13324
    CertificateExpirationAlarmType_ExpirationDate = 13325
    CertificateExpirationAlarmType_CertificateType = 13326
    CertificateExpirationAlarmType_Certificate = 13327
    ServerType_Namespaces_AddressSpaceFile_MimeType = 13340
    FileType_MimeType = 13341
    CreateDirectoryMethodType = 13342
    CreateDirectoryMethodType_InputArguments = 13343
    CreateDirectoryMethodType_OutputArguments = 13344
    CreateFileMethodType = 13345
    CreateFileMethodType_InputArguments = 13346
    CreateFileMethodType_OutputArguments = 13347
    DeleteFileMethodType = 13348
    DeleteFileMethodType_InputArguments = 13349
    MoveOrCopyMethodType = 13350
    MoveOrCopyMethodType_InputArguments = 13351
    MoveOrCopyMethodType_OutputArguments = 13352
    FileDirectoryType = 13353
    FileDirectoryType_xFileDirectoryNamex = 13354
    FileDirectoryType_xFileDirectoryNamex_CreateDirectory = 13355
    FileDirectoryType_xFileDirectoryNamex_CreateDirectory_InputArguments = 13356
    FileDirectoryType_xFileDirectoryNamex_CreateDirectory_OutputArguments = 13357
    FileDirectoryType_xFileDirectoryNamex_CreateFile = 13358
    FileDirectoryType_xFileDirectoryNamex_CreateFile_InputArguments = 13359
    FileDirectoryType_xFileDirectoryNamex_CreateFile_OutputArguments = 13360
    FileDirectoryType_xFileDirectoryNamex_Delete = 13361
    FileDirectoryType_xFileDirectoryNamex_Delete_InputArguments = 13362
    FileDirectoryType_xFileDirectoryNamex_MoveOrCopy = 13363
    FileDirectoryType_xFileDirectoryNamex_MoveOrCopy_InputArguments = 13364
    FileDirectoryType_xFileDirectoryNamex_MoveOrCopy_OutputArguments = 13365
    FileDirectoryType_xFileNamex = 13366
    FileDirectoryType_xFileNamex_Size = 13367
    FileDirectoryType_xFileNamex_Writable = 13368
    FileDirectoryType_xFileNamex_UserWritable = 13369
    FileDirectoryType_xFileNamex_OpenCount = 13370
    FileDirectoryType_xFileNamex_MimeType = 13371
    FileDirectoryType_xFileNamex_Open = 13372
    FileDirectoryType_xFileNamex_Open_InputArguments = 13373
    FileDirectoryType_xFileNamex_Open_OutputArguments = 13374
    FileDirectoryType_xFileNamex_Close = 13375
    FileDirectoryType_xFileNamex_Close_InputArguments = 13376
    FileDirectoryType_xFileNamex_Read = 13377
    FileDirectoryType_xFileNamex_Read_InputArguments = 13378
    FileDirectoryType_xFileNamex_Read_OutputArguments = 13379
    FileDirectoryType_xFileNamex_Write = 13380
    FileDirectoryType_xFileNamex_Write_InputArguments = 13381
    FileDirectoryType_xFileNamex_GetPosition = 13382
    FileDirectoryType_xFileNamex_GetPosition_InputArguments = 13383
    FileDirectoryType_xFileNamex_GetPosition_OutputArguments = 13384
    FileDirectoryType_xFileNamex_SetPosition = 13385
    FileDirectoryType_xFileNamex_SetPosition_InputArguments = 13386
    FileDirectoryType_CreateDirectory = 13387
    FileDirectoryType_CreateDirectory_InputArguments = 13388
    FileDirectoryType_CreateDirectory_OutputArguments = 13389
    FileDirectoryType_CreateFile = 13390
    FileDirectoryType_CreateFile_InputArguments = 13391
    FileDirectoryType_CreateFile_OutputArguments = 13392
    FileDirectoryType_Delete = 13393
    FileDirectoryType_Delete_InputArguments = 13394
    FileDirectoryType_MoveOrCopy = 13395
    FileDirectoryType_MoveOrCopy_InputArguments = 13396
    FileDirectoryType_MoveOrCopy_OutputArguments = 13397
    AddressSpaceFileType_MimeType = 13398
    NamespaceMetadataType_NamespaceFile_MimeType = 13399
    NamespacesType_NamespaceIdentifier_NamespaceFile_MimeType = 13400
    NamespacesType_AddressSpaceFile_MimeType = 13401
    Server_Namespaces_AddressSpaceFile_MimeType = 13402
    TrustListType_MimeType = 13403
    CertificateGroupType_TrustList = 13599
    CertificateGroupType_TrustList_Size = 13600
    CertificateGroupType_TrustList_Writable = 13601
    CertificateGroupType_TrustList_UserWritable = 13602
    CertificateGroupType_TrustList_OpenCount = 13603
    CertificateGroupType_TrustList_MimeType = 13604
    CertificateGroupType_TrustList_Open = 13605
    CertificateGroupType_TrustList_Open_InputArguments = 13606
    CertificateGroupType_TrustList_Open_OutputArguments = 13607
    CertificateGroupType_TrustList_Close = 13608
    CertificateGroupType_TrustList_Close_InputArguments = 13609
    CertificateGroupType_TrustList_Read = 13610
    CertificateGroupType_TrustList_Read_InputArguments = 13611
    CertificateGroupType_TrustList_Read_OutputArguments = 13612
    CertificateGroupType_TrustList_Write = 13613
    CertificateGroupType_TrustList_Write_InputArguments = 13614
    CertificateGroupType_TrustList_GetPosition = 13615
    CertificateGroupType_TrustList_GetPosition_InputArguments = 13616
    CertificateGroupType_TrustList_GetPosition_OutputArguments = 13617
    CertificateGroupType_TrustList_SetPosition = 13618
    CertificateGroupType_TrustList_SetPosition_InputArguments = 13619
    CertificateGroupType_TrustList_LastUpdateTime = 13620
    CertificateGroupType_TrustList_OpenWithMasks = 13621
    CertificateGroupType_TrustList_OpenWithMasks_InputArguments = 13622
    CertificateGroupType_TrustList_OpenWithMasks_OutputArguments = 13623
    CertificateGroupType_TrustList_CloseAndUpdate = 13624
    CertificateGroupType_TrustList_CloseAndUpdate_InputArguments = 13625
    CertificateGroupType_TrustList_CloseAndUpdate_OutputArguments = 13626
    CertificateGroupType_TrustList_AddCertificate = 13627
    CertificateGroupType_TrustList_AddCertificate_InputArguments = 13628
    CertificateGroupType_TrustList_RemoveCertificate = 13629
    CertificateGroupType_TrustList_RemoveCertificate_InputArguments = 13630
    CertificateGroupType_CertificateTypes = 13631
    CertificateUpdatedAuditEventType_CertificateGroup = 13735
    CertificateUpdatedAuditEventType_CertificateType = 13736
    ServerConfiguration_UpdateCertificate = 13737
    ServerConfiguration_UpdateCertificate_InputArguments = 13738
    ServerConfiguration_UpdateCertificate_OutputArguments = 13739
    CertificateGroupFolderType = 13813
    CertificateGroupFolderType_DefaultApplicationGroup = 13814
    CertificateGroupFolderType_DefaultApplicationGroup_TrustList = 13815
    CertificateGroupFolderType_DefaultApplicationGroup_TrustList_Size = 13816
    CertificateGroupFolderType_DefaultApplicationGroup_TrustList_Writable = 13817
    CertificateGroupFolderType_DefaultApplicationGroup_TrustList_UserWritable = 13818
    CertificateGroupFolderType_DefaultApplicationGroup_TrustList_OpenCount = 13819
    CertificateGroupFolderType_DefaultApplicationGroup_TrustList_MimeType = 13820
    CertificateGroupFolderType_DefaultApplicationGroup_TrustList_Open = 13821
    CertificateGroupFolderType_DefaultApplicationGroup_TrustList_Open_InputArguments = 13822
    CertificateGroupFolderType_DefaultApplicationGroup_TrustList_Open_OutputArguments = 13823
    CertificateGroupFolderType_DefaultApplicationGroup_TrustList_Close = 13824
    CertificateGroupFolderType_DefaultApplicationGroup_TrustList_Close_InputArguments = 13825
    CertificateGroupFolderType_DefaultApplicationGroup_TrustList_Read = 13826
    CertificateGroupFolderType_DefaultApplicationGroup_TrustList_Read_InputArguments = 13827
    CertificateGroupFolderType_DefaultApplicationGroup_TrustList_Read_OutputArguments = 13828
    CertificateGroupFolderType_DefaultApplicationGroup_TrustList_Write = 13829
    CertificateGroupFolderType_DefaultApplicationGroup_TrustList_Write_InputArguments = 13830
    CertificateGroupFolderType_DefaultApplicationGroup_TrustList_GetPosition = 13831
    CertificateGroupFolderType_DefaultApplicationGroup_TrustList_GetPosition_InputArguments = 13832
    CertificateGroupFolderType_DefaultApplicationGroup_TrustList_GetPosition_OutputArguments = 13833
    CertificateGroupFolderType_DefaultApplicationGroup_TrustList_SetPosition = 13834
    CertificateGroupFolderType_DefaultApplicationGroup_TrustList_SetPosition_InputArguments = 13835
    CertificateGroupFolderType_DefaultApplicationGroup_TrustList_LastUpdateTime = 13836
    CertificateGroupFolderType_DefaultApplicationGroup_TrustList_OpenWithMasks = 13837
    CertificateGroupFolderType_DefaultApplicationGroup_TrustList_OpenWithMasks_InputArguments = 13838
    CertificateGroupFolderType_DefaultApplicationGroup_TrustList_OpenWithMasks_OutputArguments = 13839
    CertificateGroupFolderType_DefaultApplicationGroup_TrustList_CloseAndUpdate = 13840
    CertificateGroupFolderType_DefaultApplicationGroup_TrustList_CloseAndUpdate_InputArguments = 13841
    CertificateGroupFolderType_DefaultApplicationGroup_TrustList_CloseAndUpdate_OutputArguments = 13842
    CertificateGroupFolderType_DefaultApplicationGroup_TrustList_AddCertificate = 13843
    CertificateGroupFolderType_DefaultApplicationGroup_TrustList_AddCertificate_InputArguments = 13844
    CertificateGroupFolderType_DefaultApplicationGroup_TrustList_RemoveCertificate = 13845
    CertificateGroupFolderType_DefaultApplicationGroup_TrustList_RemoveCertificate_InputArguments = 13846
    CertificateGroupFolderType_DefaultApplicationGroup_CertificateTypes = 13847
    CertificateGroupFolderType_DefaultHttpsGroup = 13848
    CertificateGroupFolderType_DefaultHttpsGroup_TrustList = 13849
    CertificateGroupFolderType_DefaultHttpsGroup_TrustList_Size = 13850
    CertificateGroupFolderType_DefaultHttpsGroup_TrustList_Writable = 13851
    CertificateGroupFolderType_DefaultHttpsGroup_TrustList_UserWritable = 13852
    CertificateGroupFolderType_DefaultHttpsGroup_TrustList_OpenCount = 13853
    CertificateGroupFolderType_DefaultHttpsGroup_TrustList_MimeType = 13854
    CertificateGroupFolderType_DefaultHttpsGroup_TrustList_Open = 13855
    CertificateGroupFolderType_DefaultHttpsGroup_TrustList_Open_InputArguments = 13856
    CertificateGroupFolderType_DefaultHttpsGroup_TrustList_Open_OutputArguments = 13857
    CertificateGroupFolderType_DefaultHttpsGroup_TrustList_Close = 13858
    CertificateGroupFolderType_DefaultHttpsGroup_TrustList_Close_InputArguments = 13859
    CertificateGroupFolderType_DefaultHttpsGroup_TrustList_Read = 13860
    CertificateGroupFolderType_DefaultHttpsGroup_TrustList_Read_InputArguments = 13861
    CertificateGroupFolderType_DefaultHttpsGroup_TrustList_Read_OutputArguments = 13862
    CertificateGroupFolderType_DefaultHttpsGroup_TrustList_Write = 13863
    CertificateGroupFolderType_DefaultHttpsGroup_TrustList_Write_InputArguments = 13864
    CertificateGroupFolderType_DefaultHttpsGroup_TrustList_GetPosition = 13865
    CertificateGroupFolderType_DefaultHttpsGroup_TrustList_GetPosition_InputArguments = 13866
    CertificateGroupFolderType_DefaultHttpsGroup_TrustList_GetPosition_OutputArguments = 13867
    CertificateGroupFolderType_DefaultHttpsGroup_TrustList_SetPosition = 13868
    CertificateGroupFolderType_DefaultHttpsGroup_TrustList_SetPosition_InputArguments = 13869
    CertificateGroupFolderType_DefaultHttpsGroup_TrustList_LastUpdateTime = 13870
    CertificateGroupFolderType_DefaultHttpsGroup_TrustList_OpenWithMasks = 13871
    CertificateGroupFolderType_DefaultHttpsGroup_TrustList_OpenWithMasks_InputArguments = 13872
    CertificateGroupFolderType_DefaultHttpsGroup_TrustList_OpenWithMasks_OutputArguments = 13873
    CertificateGroupFolderType_DefaultHttpsGroup_TrustList_CloseAndUpdate = 13874
    CertificateGroupFolderType_DefaultHttpsGroup_TrustList_CloseAndUpdate_InputArguments = 13875
    CertificateGroupFolderType_DefaultHttpsGroup_TrustList_CloseAndUpdate_OutputArguments = 13876
    CertificateGroupFolderType_DefaultHttpsGroup_TrustList_AddCertificate = 13877
    CertificateGroupFolderType_DefaultHttpsGroup_TrustList_AddCertificate_InputArguments = 13878
    CertificateGroupFolderType_DefaultHttpsGroup_TrustList_RemoveCertificate = 13879
    CertificateGroupFolderType_DefaultHttpsGroup_TrustList_RemoveCertificate_InputArguments = 13880
    CertificateGroupFolderType_DefaultHttpsGroup_CertificateTypes = 13881
    CertificateGroupFolderType_DefaultUserTokenGroup = 13882
    CertificateGroupFolderType_DefaultUserTokenGroup_TrustList = 13883
    CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_Size = 13884
    CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_Writable = 13885
    CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_UserWritable = 13886
    CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_OpenCount = 13887
    CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_MimeType = 13888
    CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_Open = 13889
    CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_Open_InputArguments = 13890
    CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_Open_OutputArguments = 13891
    CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_Close = 13892
    CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_Close_InputArguments = 13893
    CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_Read = 13894
    CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_Read_InputArguments = 13895
    CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_Read_OutputArguments = 13896
    CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_Write = 13897
    CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_Write_InputArguments = 13898
    CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_GetPosition = 13899
    CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_GetPosition_InputArguments = 13900
    CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_GetPosition_OutputArguments = 13901
    CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_SetPosition = 13902
    CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_SetPosition_InputArguments = 13903
    CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_LastUpdateTime = 13904
    CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_OpenWithMasks = 13905
    CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_OpenWithMasks_InputArguments = 13906
    CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_OpenWithMasks_OutputArguments = 13907
    CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_CloseAndUpdate = 13908
    CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_CloseAndUpdate_InputArguments = 13909
    CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_CloseAndUpdate_OutputArguments = 13910
    CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_AddCertificate = 13911
    CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_AddCertificate_InputArguments = 13912
    CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_RemoveCertificate = 13913
    CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_RemoveCertificate_InputArguments = 13914
    CertificateGroupFolderType_DefaultUserTokenGroup_CertificateTypes = 13915
    CertificateGroupFolderType_xCertificateGroupx = 13916
    CertificateGroupFolderType_xCertificateGroupx_TrustList = 13917
    CertificateGroupFolderType_xCertificateGroupx_TrustList_Size = 13918
    CertificateGroupFolderType_xCertificateGroupx_TrustList_Writable = 13919
    CertificateGroupFolderType_xCertificateGroupx_TrustList_UserWritable = 13920
    CertificateGroupFolderType_xCertificateGroupx_TrustList_OpenCount = 13921
    CertificateGroupFolderType_xCertificateGroupx_TrustList_MimeType = 13922
    CertificateGroupFolderType_xCertificateGroupx_TrustList_Open = 13923
    CertificateGroupFolderType_xCertificateGroupx_TrustList_Open_InputArguments = 13924
    CertificateGroupFolderType_xCertificateGroupx_TrustList_Open_OutputArguments = 13925
    CertificateGroupFolderType_xCertificateGroupx_TrustList_Close = 13926
    CertificateGroupFolderType_xCertificateGroupx_TrustList_Close_InputArguments = 13927
    CertificateGroupFolderType_xCertificateGroupx_TrustList_Read = 13928
    CertificateGroupFolderType_xCertificateGroupx_TrustList_Read_InputArguments = 13929
    CertificateGroupFolderType_xCertificateGroupx_TrustList_Read_OutputArguments = 13930
    CertificateGroupFolderType_xCertificateGroupx_TrustList_Write = 13931
    CertificateGroupFolderType_xCertificateGroupx_TrustList_Write_InputArguments = 13932
    CertificateGroupFolderType_xCertificateGroupx_TrustList_GetPosition = 13933
    CertificateGroupFolderType_xCertificateGroupx_TrustList_GetPosition_InputArguments = 13934
    CertificateGroupFolderType_xCertificateGroupx_TrustList_GetPosition_OutputArguments = 13935
    CertificateGroupFolderType_xCertificateGroupx_TrustList_SetPosition = 13936
    CertificateGroupFolderType_xCertificateGroupx_TrustList_SetPosition_InputArguments = 13937
    CertificateGroupFolderType_xCertificateGroupx_TrustList_LastUpdateTime = 13938
    CertificateGroupFolderType_xCertificateGroupx_TrustList_OpenWithMasks = 13939
    CertificateGroupFolderType_xCertificateGroupx_TrustList_OpenWithMasks_InputArguments = 13940
    CertificateGroupFolderType_xCertificateGroupx_TrustList_OpenWithMasks_OutputArguments = 13941
    CertificateGroupFolderType_xCertificateGroupx_TrustList_CloseAndUpdate = 13942
    CertificateGroupFolderType_xCertificateGroupx_TrustList_CloseAndUpdate_InputArguments = 13943
    CertificateGroupFolderType_xCertificateGroupx_TrustList_CloseAndUpdate_OutputArguments = 13944
    CertificateGroupFolderType_xCertificateGroupx_TrustList_AddCertificate = 13945
    CertificateGroupFolderType_xCertificateGroupx_TrustList_AddCertificate_InputArguments = 13946
    CertificateGroupFolderType_xCertificateGroupx_TrustList_RemoveCertificate = 13947
    CertificateGroupFolderType_xCertificateGroupx_TrustList_RemoveCertificate_InputArguments = 13948
    CertificateGroupFolderType_xCertificateGroupx_CertificateTypes = 13949
    ServerConfigurationType_CertificateGroups = 13950
    ServerConfigurationType_CertificateGroups_DefaultApplicationGroup = 13951
    ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList = 13952
    ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_Size = 13953
    ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_Writable = 13954
    ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_UserWritable = 13955
    ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_OpenCount = 13956
    ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_MimeType = 13957
    ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_Open = 13958
    ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_Open_InputArguments = 13959
    ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_Open_OutputArguments = 13960
    ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_Close = 13961
    ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_Close_InputArguments = 13962
    ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_Read = 13963
    ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_Read_InputArguments = 13964
    ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_Read_OutputArguments = 13965
    ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_Write = 13966
    ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_Write_InputArguments = 13967
    ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_GetPosition = 13968
    ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_GetPosition_InputArguments = 13969
    ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_GetPosition_OutputArguments = 13970
    ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_SetPosition = 13971
    ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_SetPosition_InputArguments = 13972
    ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_LastUpdateTime = 13973
    ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_OpenWithMasks = 13974
    ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_OpenWithMasks_InputArguments = 13975
    ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_OpenWithMasks_OutputArguments = 13976
    ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_CloseAndUpdate = 13977
    ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_CloseAndUpdate_InputArguments = 13978
    ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_CloseAndUpdate_OutputArguments = 13979
    ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_AddCertificate = 13980
    ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_AddCertificate_InputArguments = 13981
    ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_RemoveCertificate = 13982
    ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_RemoveCertificate_InputArguments = 13983
    ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_CertificateTypes = 13984
    ServerConfigurationType_CertificateGroups_DefaultHttpsGroup = 13985
    ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList = 13986
    ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_Size = 13987
    ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_Writable = 13988
    ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_UserWritable = 13989
    ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_OpenCount = 13990
    ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_MimeType = 13991
    ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_Open = 13992
    ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_Open_InputArguments = 13993
    ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_Open_OutputArguments = 13994
    ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_Close = 13995
    ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_Close_InputArguments = 13996
    ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_Read = 13997
    ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_Read_InputArguments = 13998
    ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_Read_OutputArguments = 13999
    ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_Write = 14000
    ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_Write_InputArguments = 14001
    ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_GetPosition = 14002
    ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_GetPosition_InputArguments = 14003
    ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_GetPosition_OutputArguments = 14004
    ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_SetPosition = 14005
    ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_SetPosition_InputArguments = 14006
    ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_LastUpdateTime = 14007
    ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_OpenWithMasks = 14008
    ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_OpenWithMasks_InputArguments = 14009
    ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_OpenWithMasks_OutputArguments = 14010
    ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_CloseAndUpdate = 14011
    ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_CloseAndUpdate_InputArguments = 14012
    ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_CloseAndUpdate_OutputArguments = 14013
    ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_AddCertificate = 14014
    ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_AddCertificate_InputArguments = 14015
    ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_RemoveCertificate = 14016
    ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_RemoveCertificate_InputArguments = 14017
    ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_CertificateTypes = 14018
    ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup = 14019
    ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList = 14020
    ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_Size = 14021
    ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_Writable = 14022
    ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_UserWritable = 14023
    ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_OpenCount = 14024
    ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_MimeType = 14025
    ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_Open = 14026
    ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_Open_InputArguments = 14027
    ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_Open_OutputArguments = 14028
    ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_Close = 14029
    ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_Close_InputArguments = 14030
    ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_Read = 14031
    ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_Read_InputArguments = 14032
    ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_Read_OutputArguments = 14033
    ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_Write = 14034
    ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_Write_InputArguments = 14035
    ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_GetPosition = 14036
    ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_GetPosition_InputArguments = 14037
    ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_GetPosition_OutputArguments = 14038
    ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_SetPosition = 14039
    ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_SetPosition_InputArguments = 14040
    ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_LastUpdateTime = 14041
    ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_OpenWithMasks = 14042
    ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_OpenWithMasks_InputArguments = 14043
    ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_OpenWithMasks_OutputArguments = 14044
    ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_CloseAndUpdate = 14045
    ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_CloseAndUpdate_InputArguments = 14046
    ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_CloseAndUpdate_OutputArguments = 14047
    ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_AddCertificate = 14048
    ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_AddCertificate_InputArguments = 14049
    ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_RemoveCertificate = 14050
    ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_RemoveCertificate_InputArguments = 14051
    ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_CertificateTypes = 14052
    ServerConfiguration_CertificateGroups = 14053
    ServerConfiguration_CertificateGroups_DefaultHttpsGroup = 14088
    ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList = 14089
    ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_Size = 14090
    ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_Writable = 14091
    ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_UserWritable = 14092
    ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_OpenCount = 14093
    ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_MimeType = 14094
    ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_Open = 14095
    ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_Open_InputArguments = 14096
    ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_Open_OutputArguments = 14097
    ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_Close = 14098
    ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_Close_InputArguments = 14099
    ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_Read = 14100
    ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_Read_InputArguments = 14101
    ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_Read_OutputArguments = 14102
    ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_Write = 14103
    ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_Write_InputArguments = 14104
    ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_GetPosition = 14105
    ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_GetPosition_InputArguments = 14106
    ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_GetPosition_OutputArguments = 14107
    ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_SetPosition = 14108
    ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_SetPosition_InputArguments = 14109
    ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_LastUpdateTime = 14110
    ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_OpenWithMasks = 14111
    ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_OpenWithMasks_InputArguments = 14112
    ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_OpenWithMasks_OutputArguments = 14113
    ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_CloseAndUpdate = 14114
    ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_CloseAndUpdate_InputArguments = 14115
    ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_CloseAndUpdate_OutputArguments = 14116
    ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_AddCertificate = 14117
    ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_AddCertificate_InputArguments = 14118
    ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_RemoveCertificate = 14119
    ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_RemoveCertificate_InputArguments = 14120
    ServerConfiguration_CertificateGroups_DefaultHttpsGroup_CertificateTypes = 14121
    ServerConfiguration_CertificateGroups_DefaultUserTokenGroup = 14122
    ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList = 14123
    ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_Size = 14124
    ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_Writable = 14125
    ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_UserWritable = 14126
    ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_OpenCount = 14127
    ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_MimeType = 14128
    ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_Open = 14129
    ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_Open_InputArguments = 14130
    ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_Open_OutputArguments = 14131
    ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_Close = 14132
    ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_Close_InputArguments = 14133
    ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_Read = 14134
    ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_Read_InputArguments = 14135
    ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_Read_OutputArguments = 14136
    ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_Write = 14137
    ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_Write_InputArguments = 14138
    ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_GetPosition = 14139
    ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_GetPosition_InputArguments = 14140
    ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_GetPosition_OutputArguments = 14141
    ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_SetPosition = 14142
    ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_SetPosition_InputArguments = 14143
    ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_LastUpdateTime = 14144
    ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_OpenWithMasks = 14145
    ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_OpenWithMasks_InputArguments = 14146
    ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_OpenWithMasks_OutputArguments = 14147
    ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_CloseAndUpdate = 14148
    ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_CloseAndUpdate_InputArguments = 14149
    ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_CloseAndUpdate_OutputArguments = 14150
    ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_AddCertificate = 14151
    ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_AddCertificate_InputArguments = 14152
    ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_RemoveCertificate = 14153
    ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_RemoveCertificate_InputArguments = 14154
    ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_CertificateTypes = 14155
    ServerConfiguration_CertificateGroups_DefaultApplicationGroup = 14156
    ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_Writable = 14157
    ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_UserWritable = 14158
    ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_MimeType = 14159
    ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_CloseAndUpdate_InputArguments = 14160
    ServerConfiguration_CertificateGroups_DefaultApplicationGroup_CertificateTypes = 14161
    AuditCreateSessionEventType_SessionId = 14413
    AuditUrlMismatchEventType_SessionId = 14414
    Server_ServerRedundancy_ServerNetworkGroups = 14415


ObjectIdNames = {}
ObjectIdNames[0] = 'Null'
ObjectIdNames[1] = 'Boolean'
ObjectIdNames[2] = 'SByte'
ObjectIdNames[3] = 'Byte'
ObjectIdNames[4] = 'Int16'
ObjectIdNames[5] = 'UInt16'
ObjectIdNames[6] = 'Int32'
ObjectIdNames[7] = 'UInt32'
ObjectIdNames[8] = 'Int64'
ObjectIdNames[9] = 'UInt64'
ObjectIdNames[10] = 'Float'
ObjectIdNames[11] = 'Double'
ObjectIdNames[12] = 'String'
ObjectIdNames[13] = 'DateTime'
ObjectIdNames[14] = 'Guid'
ObjectIdNames[15] = 'ByteString'
ObjectIdNames[16] = 'XmlElement'
ObjectIdNames[17] = 'NodeId'
ObjectIdNames[18] = 'ExpandedNodeId'
ObjectIdNames[19] = 'StatusCode'
ObjectIdNames[20] = 'QualifiedName'
ObjectIdNames[21] = 'LocalizedText'
ObjectIdNames[22] = 'Structure'
ObjectIdNames[23] = 'DataValue'
ObjectIdNames[24] = 'BaseDataType'
ObjectIdNames[25] = 'DiagnosticInfo'
ObjectIdNames[26] = 'Number'
ObjectIdNames[27] = 'Integer'
ObjectIdNames[28] = 'UInteger'
ObjectIdNames[29] = 'Enumeration'
ObjectIdNames[30] = 'Image'
ObjectIdNames[31] = 'References'
ObjectIdNames[32] = 'NonHierarchicalReferences'
ObjectIdNames[33] = 'HierarchicalReferences'
ObjectIdNames[34] = 'HasChild'
ObjectIdNames[35] = 'Organizes'
ObjectIdNames[36] = 'HasEventSource'
ObjectIdNames[37] = 'HasModellingRule'
ObjectIdNames[38] = 'HasEncoding'
ObjectIdNames[39] = 'HasDescription'
ObjectIdNames[40] = 'HasTypeDefinition'
ObjectIdNames[41] = 'GeneratesEvent'
ObjectIdNames[44] = 'Aggregates'
ObjectIdNames[45] = 'HasSubtype'
ObjectIdNames[46] = 'HasProperty'
ObjectIdNames[47] = 'HasComponent'
ObjectIdNames[48] = 'HasNotifier'
ObjectIdNames[49] = 'HasOrderedComponent'
ObjectIdNames[51] = 'FromState'
ObjectIdNames[52] = 'ToState'
ObjectIdNames[53] = 'HasCause'
ObjectIdNames[54] = 'HasEffect'
ObjectIdNames[56] = 'HasHistoricalConfiguration'
ObjectIdNames[58] = 'BaseObjectType'
ObjectIdNames[61] = 'FolderType'
ObjectIdNames[62] = 'BaseVariableType'
ObjectIdNames[63] = 'BaseDataVariableType'
ObjectIdNames[68] = 'PropertyType'
ObjectIdNames[69] = 'DataTypeDescriptionType'
ObjectIdNames[72] = 'DataTypeDictionaryType'
ObjectIdNames[75] = 'DataTypeSystemType'
ObjectIdNames[76] = 'DataTypeEncodingType'
ObjectIdNames[77] = 'ModellingRuleType'
ObjectIdNames[78] = 'ModellingRule_Mandatory'
ObjectIdNames[79] = 'ModellingRule_MandatoryShared'
ObjectIdNames[80] = 'ModellingRule_Optional'
ObjectIdNames[83] = 'ModellingRule_ExposesItsArray'
ObjectIdNames[84] = 'RootFolder'
ObjectIdNames[85] = 'ObjectsFolder'
ObjectIdNames[86] = 'TypesFolder'
ObjectIdNames[87] = 'ViewsFolder'
ObjectIdNames[88] = 'ObjectTypesFolder'
ObjectIdNames[89] = 'VariableTypesFolder'
ObjectIdNames[90] = 'DataTypesFolder'
ObjectIdNames[91] = 'ReferenceTypesFolder'
ObjectIdNames[92] = 'XmlSchema_TypeSystem'
ObjectIdNames[93] = 'OPCBinarySchema_TypeSystem'
ObjectIdNames[104] = 'DataTypeDescriptionType_DataTypeVersion'
ObjectIdNames[105] = 'DataTypeDescriptionType_DictionaryFragment'
ObjectIdNames[106] = 'DataTypeDictionaryType_DataTypeVersion'
ObjectIdNames[107] = 'DataTypeDictionaryType_NamespaceUri'
ObjectIdNames[111] = 'ModellingRuleType_NamingRule'
ObjectIdNames[112] = 'ModellingRule_Mandatory_NamingRule'
ObjectIdNames[113] = 'ModellingRule_Optional_NamingRule'
ObjectIdNames[114] = 'ModellingRule_ExposesItsArray_NamingRule'
ObjectIdNames[116] = 'ModellingRule_MandatoryShared_NamingRule'
ObjectIdNames[117] = 'HasSubStateMachine'
ObjectIdNames[120] = 'NamingRuleType'
ObjectIdNames[121] = 'Decimal128'
ObjectIdNames[256] = 'IdType'
ObjectIdNames[257] = 'NodeClass'
ObjectIdNames[258] = 'Node'
ObjectIdNames[259] = 'Node_Encoding_DefaultXml'
ObjectIdNames[260] = 'Node_Encoding_DefaultBinary'
ObjectIdNames[261] = 'ObjectNode'
ObjectIdNames[262] = 'ObjectNode_Encoding_DefaultXml'
ObjectIdNames[263] = 'ObjectNode_Encoding_DefaultBinary'
ObjectIdNames[264] = 'ObjectTypeNode'
ObjectIdNames[265] = 'ObjectTypeNode_Encoding_DefaultXml'
ObjectIdNames[266] = 'ObjectTypeNode_Encoding_DefaultBinary'
ObjectIdNames[267] = 'VariableNode'
ObjectIdNames[268] = 'VariableNode_Encoding_DefaultXml'
ObjectIdNames[269] = 'VariableNode_Encoding_DefaultBinary'
ObjectIdNames[270] = 'VariableTypeNode'
ObjectIdNames[271] = 'VariableTypeNode_Encoding_DefaultXml'
ObjectIdNames[272] = 'VariableTypeNode_Encoding_DefaultBinary'
ObjectIdNames[273] = 'ReferenceTypeNode'
ObjectIdNames[274] = 'ReferenceTypeNode_Encoding_DefaultXml'
ObjectIdNames[275] = 'ReferenceTypeNode_Encoding_DefaultBinary'
ObjectIdNames[276] = 'MethodNode'
ObjectIdNames[277] = 'MethodNode_Encoding_DefaultXml'
ObjectIdNames[278] = 'MethodNode_Encoding_DefaultBinary'
ObjectIdNames[279] = 'ViewNode'
ObjectIdNames[280] = 'ViewNode_Encoding_DefaultXml'
ObjectIdNames[281] = 'ViewNode_Encoding_DefaultBinary'
ObjectIdNames[282] = 'DataTypeNode'
ObjectIdNames[283] = 'DataTypeNode_Encoding_DefaultXml'
ObjectIdNames[284] = 'DataTypeNode_Encoding_DefaultBinary'
ObjectIdNames[285] = 'ReferenceNode'
ObjectIdNames[286] = 'ReferenceNode_Encoding_DefaultXml'
ObjectIdNames[287] = 'ReferenceNode_Encoding_DefaultBinary'
ObjectIdNames[288] = 'IntegerId'
ObjectIdNames[289] = 'Counter'
ObjectIdNames[290] = 'Duration'
ObjectIdNames[291] = 'NumericRange'
ObjectIdNames[292] = 'Time'
ObjectIdNames[293] = 'Date'
ObjectIdNames[294] = 'UtcTime'
ObjectIdNames[295] = 'LocaleId'
ObjectIdNames[296] = 'Argument'
ObjectIdNames[297] = 'Argument_Encoding_DefaultXml'
ObjectIdNames[298] = 'Argument_Encoding_DefaultBinary'
ObjectIdNames[299] = 'StatusResult'
ObjectIdNames[300] = 'StatusResult_Encoding_DefaultXml'
ObjectIdNames[301] = 'StatusResult_Encoding_DefaultBinary'
ObjectIdNames[302] = 'MessageSecurityMode'
ObjectIdNames[303] = 'UserTokenType'
ObjectIdNames[304] = 'UserTokenPolicy'
ObjectIdNames[305] = 'UserTokenPolicy_Encoding_DefaultXml'
ObjectIdNames[306] = 'UserTokenPolicy_Encoding_DefaultBinary'
ObjectIdNames[307] = 'ApplicationType'
ObjectIdNames[308] = 'ApplicationDescription'
ObjectIdNames[309] = 'ApplicationDescription_Encoding_DefaultXml'
ObjectIdNames[310] = 'ApplicationDescription_Encoding_DefaultBinary'
ObjectIdNames[311] = 'ApplicationInstanceCertificate'
ObjectIdNames[312] = 'EndpointDescription'
ObjectIdNames[313] = 'EndpointDescription_Encoding_DefaultXml'
ObjectIdNames[314] = 'EndpointDescription_Encoding_DefaultBinary'
ObjectIdNames[315] = 'SecurityTokenRequestType'
ObjectIdNames[316] = 'UserIdentityToken'
ObjectIdNames[317] = 'UserIdentityToken_Encoding_DefaultXml'
ObjectIdNames[318] = 'UserIdentityToken_Encoding_DefaultBinary'
ObjectIdNames[319] = 'AnonymousIdentityToken'
ObjectIdNames[320] = 'AnonymousIdentityToken_Encoding_DefaultXml'
ObjectIdNames[321] = 'AnonymousIdentityToken_Encoding_DefaultBinary'
ObjectIdNames[322] = 'UserNameIdentityToken'
ObjectIdNames[323] = 'UserNameIdentityToken_Encoding_DefaultXml'
ObjectIdNames[324] = 'UserNameIdentityToken_Encoding_DefaultBinary'
ObjectIdNames[325] = 'X509IdentityToken'
ObjectIdNames[326] = 'X509IdentityToken_Encoding_DefaultXml'
ObjectIdNames[327] = 'X509IdentityToken_Encoding_DefaultBinary'
ObjectIdNames[331] = 'EndpointConfiguration'
ObjectIdNames[332] = 'EndpointConfiguration_Encoding_DefaultXml'
ObjectIdNames[333] = 'EndpointConfiguration_Encoding_DefaultBinary'
ObjectIdNames[334] = 'ComplianceLevel'
ObjectIdNames[335] = 'SupportedProfile'
ObjectIdNames[336] = 'SupportedProfile_Encoding_DefaultXml'
ObjectIdNames[337] = 'SupportedProfile_Encoding_DefaultBinary'
ObjectIdNames[338] = 'BuildInfo'
ObjectIdNames[339] = 'BuildInfo_Encoding_DefaultXml'
ObjectIdNames[340] = 'BuildInfo_Encoding_DefaultBinary'
ObjectIdNames[341] = 'SoftwareCertificate'
ObjectIdNames[342] = 'SoftwareCertificate_Encoding_DefaultXml'
ObjectIdNames[343] = 'SoftwareCertificate_Encoding_DefaultBinary'
ObjectIdNames[344] = 'SignedSoftwareCertificate'
ObjectIdNames[345] = 'SignedSoftwareCertificate_Encoding_DefaultXml'
ObjectIdNames[346] = 'SignedSoftwareCertificate_Encoding_DefaultBinary'
ObjectIdNames[347] = 'AttributeWriteMask'
ObjectIdNames[348] = 'NodeAttributesMask'
ObjectIdNames[349] = 'NodeAttributes'
ObjectIdNames[350] = 'NodeAttributes_Encoding_DefaultXml'
ObjectIdNames[351] = 'NodeAttributes_Encoding_DefaultBinary'
ObjectIdNames[352] = 'ObjectAttributes'
ObjectIdNames[353] = 'ObjectAttributes_Encoding_DefaultXml'
ObjectIdNames[354] = 'ObjectAttributes_Encoding_DefaultBinary'
ObjectIdNames[355] = 'VariableAttributes'
ObjectIdNames[356] = 'VariableAttributes_Encoding_DefaultXml'
ObjectIdNames[357] = 'VariableAttributes_Encoding_DefaultBinary'
ObjectIdNames[358] = 'MethodAttributes'
ObjectIdNames[359] = 'MethodAttributes_Encoding_DefaultXml'
ObjectIdNames[360] = 'MethodAttributes_Encoding_DefaultBinary'
ObjectIdNames[361] = 'ObjectTypeAttributes'
ObjectIdNames[362] = 'ObjectTypeAttributes_Encoding_DefaultXml'
ObjectIdNames[363] = 'ObjectTypeAttributes_Encoding_DefaultBinary'
ObjectIdNames[364] = 'VariableTypeAttributes'
ObjectIdNames[365] = 'VariableTypeAttributes_Encoding_DefaultXml'
ObjectIdNames[366] = 'VariableTypeAttributes_Encoding_DefaultBinary'
ObjectIdNames[367] = 'ReferenceTypeAttributes'
ObjectIdNames[368] = 'ReferenceTypeAttributes_Encoding_DefaultXml'
ObjectIdNames[369] = 'ReferenceTypeAttributes_Encoding_DefaultBinary'
ObjectIdNames[370] = 'DataTypeAttributes'
ObjectIdNames[371] = 'DataTypeAttributes_Encoding_DefaultXml'
ObjectIdNames[372] = 'DataTypeAttributes_Encoding_DefaultBinary'
ObjectIdNames[373] = 'ViewAttributes'
ObjectIdNames[374] = 'ViewAttributes_Encoding_DefaultXml'
ObjectIdNames[375] = 'ViewAttributes_Encoding_DefaultBinary'
ObjectIdNames[376] = 'AddNodesItem'
ObjectIdNames[377] = 'AddNodesItem_Encoding_DefaultXml'
ObjectIdNames[378] = 'AddNodesItem_Encoding_DefaultBinary'
ObjectIdNames[379] = 'AddReferencesItem'
ObjectIdNames[380] = 'AddReferencesItem_Encoding_DefaultXml'
ObjectIdNames[381] = 'AddReferencesItem_Encoding_DefaultBinary'
ObjectIdNames[382] = 'DeleteNodesItem'
ObjectIdNames[383] = 'DeleteNodesItem_Encoding_DefaultXml'
ObjectIdNames[384] = 'DeleteNodesItem_Encoding_DefaultBinary'
ObjectIdNames[385] = 'DeleteReferencesItem'
ObjectIdNames[386] = 'DeleteReferencesItem_Encoding_DefaultXml'
ObjectIdNames[387] = 'DeleteReferencesItem_Encoding_DefaultBinary'
ObjectIdNames[388] = 'SessionAuthenticationToken'
ObjectIdNames[389] = 'RequestHeader'
ObjectIdNames[390] = 'RequestHeader_Encoding_DefaultXml'
ObjectIdNames[391] = 'RequestHeader_Encoding_DefaultBinary'
ObjectIdNames[392] = 'ResponseHeader'
ObjectIdNames[393] = 'ResponseHeader_Encoding_DefaultXml'
ObjectIdNames[394] = 'ResponseHeader_Encoding_DefaultBinary'
ObjectIdNames[395] = 'ServiceFault'
ObjectIdNames[396] = 'ServiceFault_Encoding_DefaultXml'
ObjectIdNames[397] = 'ServiceFault_Encoding_DefaultBinary'
ObjectIdNames[398] = 'EnumeratedTestType'
ObjectIdNames[420] = 'FindServersRequest'
ObjectIdNames[421] = 'FindServersRequest_Encoding_DefaultXml'
ObjectIdNames[422] = 'FindServersRequest_Encoding_DefaultBinary'
ObjectIdNames[423] = 'FindServersResponse'
ObjectIdNames[424] = 'FindServersResponse_Encoding_DefaultXml'
ObjectIdNames[425] = 'FindServersResponse_Encoding_DefaultBinary'
ObjectIdNames[426] = 'GetEndpointsRequest'
ObjectIdNames[427] = 'GetEndpointsRequest_Encoding_DefaultXml'
ObjectIdNames[428] = 'GetEndpointsRequest_Encoding_DefaultBinary'
ObjectIdNames[429] = 'GetEndpointsResponse'
ObjectIdNames[430] = 'GetEndpointsResponse_Encoding_DefaultXml'
ObjectIdNames[431] = 'GetEndpointsResponse_Encoding_DefaultBinary'
ObjectIdNames[432] = 'RegisteredServer'
ObjectIdNames[433] = 'RegisteredServer_Encoding_DefaultXml'
ObjectIdNames[434] = 'RegisteredServer_Encoding_DefaultBinary'
ObjectIdNames[435] = 'RegisterServerRequest'
ObjectIdNames[436] = 'RegisterServerRequest_Encoding_DefaultXml'
ObjectIdNames[437] = 'RegisterServerRequest_Encoding_DefaultBinary'
ObjectIdNames[438] = 'RegisterServerResponse'
ObjectIdNames[439] = 'RegisterServerResponse_Encoding_DefaultXml'
ObjectIdNames[440] = 'RegisterServerResponse_Encoding_DefaultBinary'
ObjectIdNames[441] = 'ChannelSecurityToken'
ObjectIdNames[442] = 'ChannelSecurityToken_Encoding_DefaultXml'
ObjectIdNames[443] = 'ChannelSecurityToken_Encoding_DefaultBinary'
ObjectIdNames[444] = 'OpenSecureChannelRequest'
ObjectIdNames[445] = 'OpenSecureChannelRequest_Encoding_DefaultXml'
ObjectIdNames[446] = 'OpenSecureChannelRequest_Encoding_DefaultBinary'
ObjectIdNames[447] = 'OpenSecureChannelResponse'
ObjectIdNames[448] = 'OpenSecureChannelResponse_Encoding_DefaultXml'
ObjectIdNames[449] = 'OpenSecureChannelResponse_Encoding_DefaultBinary'
ObjectIdNames[450] = 'CloseSecureChannelRequest'
ObjectIdNames[451] = 'CloseSecureChannelRequest_Encoding_DefaultXml'
ObjectIdNames[452] = 'CloseSecureChannelRequest_Encoding_DefaultBinary'
ObjectIdNames[453] = 'CloseSecureChannelResponse'
ObjectIdNames[454] = 'CloseSecureChannelResponse_Encoding_DefaultXml'
ObjectIdNames[455] = 'CloseSecureChannelResponse_Encoding_DefaultBinary'
ObjectIdNames[456] = 'SignatureData'
ObjectIdNames[457] = 'SignatureData_Encoding_DefaultXml'
ObjectIdNames[458] = 'SignatureData_Encoding_DefaultBinary'
ObjectIdNames[459] = 'CreateSessionRequest'
ObjectIdNames[460] = 'CreateSessionRequest_Encoding_DefaultXml'
ObjectIdNames[461] = 'CreateSessionRequest_Encoding_DefaultBinary'
ObjectIdNames[462] = 'CreateSessionResponse'
ObjectIdNames[463] = 'CreateSessionResponse_Encoding_DefaultXml'
ObjectIdNames[464] = 'CreateSessionResponse_Encoding_DefaultBinary'
ObjectIdNames[465] = 'ActivateSessionRequest'
ObjectIdNames[466] = 'ActivateSessionRequest_Encoding_DefaultXml'
ObjectIdNames[467] = 'ActivateSessionRequest_Encoding_DefaultBinary'
ObjectIdNames[468] = 'ActivateSessionResponse'
ObjectIdNames[469] = 'ActivateSessionResponse_Encoding_DefaultXml'
ObjectIdNames[470] = 'ActivateSessionResponse_Encoding_DefaultBinary'
ObjectIdNames[471] = 'CloseSessionRequest'
ObjectIdNames[472] = 'CloseSessionRequest_Encoding_DefaultXml'
ObjectIdNames[473] = 'CloseSessionRequest_Encoding_DefaultBinary'
ObjectIdNames[474] = 'CloseSessionResponse'
ObjectIdNames[475] = 'CloseSessionResponse_Encoding_DefaultXml'
ObjectIdNames[476] = 'CloseSessionResponse_Encoding_DefaultBinary'
ObjectIdNames[477] = 'CancelRequest'
ObjectIdNames[478] = 'CancelRequest_Encoding_DefaultXml'
ObjectIdNames[479] = 'CancelRequest_Encoding_DefaultBinary'
ObjectIdNames[480] = 'CancelResponse'
ObjectIdNames[481] = 'CancelResponse_Encoding_DefaultXml'
ObjectIdNames[482] = 'CancelResponse_Encoding_DefaultBinary'
ObjectIdNames[483] = 'AddNodesResult'
ObjectIdNames[484] = 'AddNodesResult_Encoding_DefaultXml'
ObjectIdNames[485] = 'AddNodesResult_Encoding_DefaultBinary'
ObjectIdNames[486] = 'AddNodesRequest'
ObjectIdNames[487] = 'AddNodesRequest_Encoding_DefaultXml'
ObjectIdNames[488] = 'AddNodesRequest_Encoding_DefaultBinary'
ObjectIdNames[489] = 'AddNodesResponse'
ObjectIdNames[490] = 'AddNodesResponse_Encoding_DefaultXml'
ObjectIdNames[491] = 'AddNodesResponse_Encoding_DefaultBinary'
ObjectIdNames[492] = 'AddReferencesRequest'
ObjectIdNames[493] = 'AddReferencesRequest_Encoding_DefaultXml'
ObjectIdNames[494] = 'AddReferencesRequest_Encoding_DefaultBinary'
ObjectIdNames[495] = 'AddReferencesResponse'
ObjectIdNames[496] = 'AddReferencesResponse_Encoding_DefaultXml'
ObjectIdNames[497] = 'AddReferencesResponse_Encoding_DefaultBinary'
ObjectIdNames[498] = 'DeleteNodesRequest'
ObjectIdNames[499] = 'DeleteNodesRequest_Encoding_DefaultXml'
ObjectIdNames[500] = 'DeleteNodesRequest_Encoding_DefaultBinary'
ObjectIdNames[501] = 'DeleteNodesResponse'
ObjectIdNames[502] = 'DeleteNodesResponse_Encoding_DefaultXml'
ObjectIdNames[503] = 'DeleteNodesResponse_Encoding_DefaultBinary'
ObjectIdNames[504] = 'DeleteReferencesRequest'
ObjectIdNames[505] = 'DeleteReferencesRequest_Encoding_DefaultXml'
ObjectIdNames[506] = 'DeleteReferencesRequest_Encoding_DefaultBinary'
ObjectIdNames[507] = 'DeleteReferencesResponse'
ObjectIdNames[508] = 'DeleteReferencesResponse_Encoding_DefaultXml'
ObjectIdNames[509] = 'DeleteReferencesResponse_Encoding_DefaultBinary'
ObjectIdNames[510] = 'BrowseDirection'
ObjectIdNames[511] = 'ViewDescription'
ObjectIdNames[512] = 'ViewDescription_Encoding_DefaultXml'
ObjectIdNames[513] = 'ViewDescription_Encoding_DefaultBinary'
ObjectIdNames[514] = 'BrowseDescription'
ObjectIdNames[515] = 'BrowseDescription_Encoding_DefaultXml'
ObjectIdNames[516] = 'BrowseDescription_Encoding_DefaultBinary'
ObjectIdNames[517] = 'BrowseResultMask'
ObjectIdNames[518] = 'ReferenceDescription'
ObjectIdNames[519] = 'ReferenceDescription_Encoding_DefaultXml'
ObjectIdNames[520] = 'ReferenceDescription_Encoding_DefaultBinary'
ObjectIdNames[521] = 'ContinuationPoint'
ObjectIdNames[522] = 'BrowseResult'
ObjectIdNames[523] = 'BrowseResult_Encoding_DefaultXml'
ObjectIdNames[524] = 'BrowseResult_Encoding_DefaultBinary'
ObjectIdNames[525] = 'BrowseRequest'
ObjectIdNames[526] = 'BrowseRequest_Encoding_DefaultXml'
ObjectIdNames[527] = 'BrowseRequest_Encoding_DefaultBinary'
ObjectIdNames[528] = 'BrowseResponse'
ObjectIdNames[529] = 'BrowseResponse_Encoding_DefaultXml'
ObjectIdNames[530] = 'BrowseResponse_Encoding_DefaultBinary'
ObjectIdNames[531] = 'BrowseNextRequest'
ObjectIdNames[532] = 'BrowseNextRequest_Encoding_DefaultXml'
ObjectIdNames[533] = 'BrowseNextRequest_Encoding_DefaultBinary'
ObjectIdNames[534] = 'BrowseNextResponse'
ObjectIdNames[535] = 'BrowseNextResponse_Encoding_DefaultXml'
ObjectIdNames[536] = 'BrowseNextResponse_Encoding_DefaultBinary'
ObjectIdNames[537] = 'RelativePathElement'
ObjectIdNames[538] = 'RelativePathElement_Encoding_DefaultXml'
ObjectIdNames[539] = 'RelativePathElement_Encoding_DefaultBinary'
ObjectIdNames[540] = 'RelativePath'
ObjectIdNames[541] = 'RelativePath_Encoding_DefaultXml'
ObjectIdNames[542] = 'RelativePath_Encoding_DefaultBinary'
ObjectIdNames[543] = 'BrowsePath'
ObjectIdNames[544] = 'BrowsePath_Encoding_DefaultXml'
ObjectIdNames[545] = 'BrowsePath_Encoding_DefaultBinary'
ObjectIdNames[546] = 'BrowsePathTarget'
ObjectIdNames[547] = 'BrowsePathTarget_Encoding_DefaultXml'
ObjectIdNames[548] = 'BrowsePathTarget_Encoding_DefaultBinary'
ObjectIdNames[549] = 'BrowsePathResult'
ObjectIdNames[550] = 'BrowsePathResult_Encoding_DefaultXml'
ObjectIdNames[551] = 'BrowsePathResult_Encoding_DefaultBinary'
ObjectIdNames[552] = 'TranslateBrowsePathsToNodeIdsRequest'
ObjectIdNames[553] = 'TranslateBrowsePathsToNodeIdsRequest_Encoding_DefaultXml'
ObjectIdNames[554] = 'TranslateBrowsePathsToNodeIdsRequest_Encoding_DefaultBinary'
ObjectIdNames[555] = 'TranslateBrowsePathsToNodeIdsResponse'
ObjectIdNames[556] = 'TranslateBrowsePathsToNodeIdsResponse_Encoding_DefaultXml'
ObjectIdNames[557] = 'TranslateBrowsePathsToNodeIdsResponse_Encoding_DefaultBinary'
ObjectIdNames[558] = 'RegisterNodesRequest'
ObjectIdNames[559] = 'RegisterNodesRequest_Encoding_DefaultXml'
ObjectIdNames[560] = 'RegisterNodesRequest_Encoding_DefaultBinary'
ObjectIdNames[561] = 'RegisterNodesResponse'
ObjectIdNames[562] = 'RegisterNodesResponse_Encoding_DefaultXml'
ObjectIdNames[563] = 'RegisterNodesResponse_Encoding_DefaultBinary'
ObjectIdNames[564] = 'UnregisterNodesRequest'
ObjectIdNames[565] = 'UnregisterNodesRequest_Encoding_DefaultXml'
ObjectIdNames[566] = 'UnregisterNodesRequest_Encoding_DefaultBinary'
ObjectIdNames[567] = 'UnregisterNodesResponse'
ObjectIdNames[568] = 'UnregisterNodesResponse_Encoding_DefaultXml'
ObjectIdNames[569] = 'UnregisterNodesResponse_Encoding_DefaultBinary'
ObjectIdNames[570] = 'QueryDataDescription'
ObjectIdNames[571] = 'QueryDataDescription_Encoding_DefaultXml'
ObjectIdNames[572] = 'QueryDataDescription_Encoding_DefaultBinary'
ObjectIdNames[573] = 'NodeTypeDescription'
ObjectIdNames[574] = 'NodeTypeDescription_Encoding_DefaultXml'
ObjectIdNames[575] = 'NodeTypeDescription_Encoding_DefaultBinary'
ObjectIdNames[576] = 'FilterOperator'
ObjectIdNames[577] = 'QueryDataSet'
ObjectIdNames[578] = 'QueryDataSet_Encoding_DefaultXml'
ObjectIdNames[579] = 'QueryDataSet_Encoding_DefaultBinary'
ObjectIdNames[580] = 'NodeReference'
ObjectIdNames[581] = 'NodeReference_Encoding_DefaultXml'
ObjectIdNames[582] = 'NodeReference_Encoding_DefaultBinary'
ObjectIdNames[583] = 'ContentFilterElement'
ObjectIdNames[584] = 'ContentFilterElement_Encoding_DefaultXml'
ObjectIdNames[585] = 'ContentFilterElement_Encoding_DefaultBinary'
ObjectIdNames[586] = 'ContentFilter'
ObjectIdNames[587] = 'ContentFilter_Encoding_DefaultXml'
ObjectIdNames[588] = 'ContentFilter_Encoding_DefaultBinary'
ObjectIdNames[589] = 'FilterOperand'
ObjectIdNames[590] = 'FilterOperand_Encoding_DefaultXml'
ObjectIdNames[591] = 'FilterOperand_Encoding_DefaultBinary'
ObjectIdNames[592] = 'ElementOperand'
ObjectIdNames[593] = 'ElementOperand_Encoding_DefaultXml'
ObjectIdNames[594] = 'ElementOperand_Encoding_DefaultBinary'
ObjectIdNames[595] = 'LiteralOperand'
ObjectIdNames[596] = 'LiteralOperand_Encoding_DefaultXml'
ObjectIdNames[597] = 'LiteralOperand_Encoding_DefaultBinary'
ObjectIdNames[598] = 'AttributeOperand'
ObjectIdNames[599] = 'AttributeOperand_Encoding_DefaultXml'
ObjectIdNames[600] = 'AttributeOperand_Encoding_DefaultBinary'
ObjectIdNames[601] = 'SimpleAttributeOperand'
ObjectIdNames[602] = 'SimpleAttributeOperand_Encoding_DefaultXml'
ObjectIdNames[603] = 'SimpleAttributeOperand_Encoding_DefaultBinary'
ObjectIdNames[604] = 'ContentFilterElementResult'
ObjectIdNames[605] = 'ContentFilterElementResult_Encoding_DefaultXml'
ObjectIdNames[606] = 'ContentFilterElementResult_Encoding_DefaultBinary'
ObjectIdNames[607] = 'ContentFilterResult'
ObjectIdNames[608] = 'ContentFilterResult_Encoding_DefaultXml'
ObjectIdNames[609] = 'ContentFilterResult_Encoding_DefaultBinary'
ObjectIdNames[610] = 'ParsingResult'
ObjectIdNames[611] = 'ParsingResult_Encoding_DefaultXml'
ObjectIdNames[612] = 'ParsingResult_Encoding_DefaultBinary'
ObjectIdNames[613] = 'QueryFirstRequest'
ObjectIdNames[614] = 'QueryFirstRequest_Encoding_DefaultXml'
ObjectIdNames[615] = 'QueryFirstRequest_Encoding_DefaultBinary'
ObjectIdNames[616] = 'QueryFirstResponse'
ObjectIdNames[617] = 'QueryFirstResponse_Encoding_DefaultXml'
ObjectIdNames[618] = 'QueryFirstResponse_Encoding_DefaultBinary'
ObjectIdNames[619] = 'QueryNextRequest'
ObjectIdNames[620] = 'QueryNextRequest_Encoding_DefaultXml'
ObjectIdNames[621] = 'QueryNextRequest_Encoding_DefaultBinary'
ObjectIdNames[622] = 'QueryNextResponse'
ObjectIdNames[623] = 'QueryNextResponse_Encoding_DefaultXml'
ObjectIdNames[624] = 'QueryNextResponse_Encoding_DefaultBinary'
ObjectIdNames[625] = 'TimestampsToReturn'
ObjectIdNames[626] = 'ReadValueId'
ObjectIdNames[627] = 'ReadValueId_Encoding_DefaultXml'
ObjectIdNames[628] = 'ReadValueId_Encoding_DefaultBinary'
ObjectIdNames[629] = 'ReadRequest'
ObjectIdNames[630] = 'ReadRequest_Encoding_DefaultXml'
ObjectIdNames[631] = 'ReadRequest_Encoding_DefaultBinary'
ObjectIdNames[632] = 'ReadResponse'
ObjectIdNames[633] = 'ReadResponse_Encoding_DefaultXml'
ObjectIdNames[634] = 'ReadResponse_Encoding_DefaultBinary'
ObjectIdNames[635] = 'HistoryReadValueId'
ObjectIdNames[636] = 'HistoryReadValueId_Encoding_DefaultXml'
ObjectIdNames[637] = 'HistoryReadValueId_Encoding_DefaultBinary'
ObjectIdNames[638] = 'HistoryReadResult'
ObjectIdNames[639] = 'HistoryReadResult_Encoding_DefaultXml'
ObjectIdNames[640] = 'HistoryReadResult_Encoding_DefaultBinary'
ObjectIdNames[641] = 'HistoryReadDetails'
ObjectIdNames[642] = 'HistoryReadDetails_Encoding_DefaultXml'
ObjectIdNames[643] = 'HistoryReadDetails_Encoding_DefaultBinary'
ObjectIdNames[644] = 'ReadEventDetails'
ObjectIdNames[645] = 'ReadEventDetails_Encoding_DefaultXml'
ObjectIdNames[646] = 'ReadEventDetails_Encoding_DefaultBinary'
ObjectIdNames[647] = 'ReadRawModifiedDetails'
ObjectIdNames[648] = 'ReadRawModifiedDetails_Encoding_DefaultXml'
ObjectIdNames[649] = 'ReadRawModifiedDetails_Encoding_DefaultBinary'
ObjectIdNames[650] = 'ReadProcessedDetails'
ObjectIdNames[651] = 'ReadProcessedDetails_Encoding_DefaultXml'
ObjectIdNames[652] = 'ReadProcessedDetails_Encoding_DefaultBinary'
ObjectIdNames[653] = 'ReadAtTimeDetails'
ObjectIdNames[654] = 'ReadAtTimeDetails_Encoding_DefaultXml'
ObjectIdNames[655] = 'ReadAtTimeDetails_Encoding_DefaultBinary'
ObjectIdNames[656] = 'HistoryData'
ObjectIdNames[657] = 'HistoryData_Encoding_DefaultXml'
ObjectIdNames[658] = 'HistoryData_Encoding_DefaultBinary'
ObjectIdNames[659] = 'HistoryEvent'
ObjectIdNames[660] = 'HistoryEvent_Encoding_DefaultXml'
ObjectIdNames[661] = 'HistoryEvent_Encoding_DefaultBinary'
ObjectIdNames[662] = 'HistoryReadRequest'
ObjectIdNames[663] = 'HistoryReadRequest_Encoding_DefaultXml'
ObjectIdNames[664] = 'HistoryReadRequest_Encoding_DefaultBinary'
ObjectIdNames[665] = 'HistoryReadResponse'
ObjectIdNames[666] = 'HistoryReadResponse_Encoding_DefaultXml'
ObjectIdNames[667] = 'HistoryReadResponse_Encoding_DefaultBinary'
ObjectIdNames[668] = 'WriteValue'
ObjectIdNames[669] = 'WriteValue_Encoding_DefaultXml'
ObjectIdNames[670] = 'WriteValue_Encoding_DefaultBinary'
ObjectIdNames[671] = 'WriteRequest'
ObjectIdNames[672] = 'WriteRequest_Encoding_DefaultXml'
ObjectIdNames[673] = 'WriteRequest_Encoding_DefaultBinary'
ObjectIdNames[674] = 'WriteResponse'
ObjectIdNames[675] = 'WriteResponse_Encoding_DefaultXml'
ObjectIdNames[676] = 'WriteResponse_Encoding_DefaultBinary'
ObjectIdNames[677] = 'HistoryUpdateDetails'
ObjectIdNames[678] = 'HistoryUpdateDetails_Encoding_DefaultXml'
ObjectIdNames[679] = 'HistoryUpdateDetails_Encoding_DefaultBinary'
ObjectIdNames[680] = 'UpdateDataDetails'
ObjectIdNames[681] = 'UpdateDataDetails_Encoding_DefaultXml'
ObjectIdNames[682] = 'UpdateDataDetails_Encoding_DefaultBinary'
ObjectIdNames[683] = 'UpdateEventDetails'
ObjectIdNames[684] = 'UpdateEventDetails_Encoding_DefaultXml'
ObjectIdNames[685] = 'UpdateEventDetails_Encoding_DefaultBinary'
ObjectIdNames[686] = 'DeleteRawModifiedDetails'
ObjectIdNames[687] = 'DeleteRawModifiedDetails_Encoding_DefaultXml'
ObjectIdNames[688] = 'DeleteRawModifiedDetails_Encoding_DefaultBinary'
ObjectIdNames[689] = 'DeleteAtTimeDetails'
ObjectIdNames[690] = 'DeleteAtTimeDetails_Encoding_DefaultXml'
ObjectIdNames[691] = 'DeleteAtTimeDetails_Encoding_DefaultBinary'
ObjectIdNames[692] = 'DeleteEventDetails'
ObjectIdNames[693] = 'DeleteEventDetails_Encoding_DefaultXml'
ObjectIdNames[694] = 'DeleteEventDetails_Encoding_DefaultBinary'
ObjectIdNames[695] = 'HistoryUpdateResult'
ObjectIdNames[696] = 'HistoryUpdateResult_Encoding_DefaultXml'
ObjectIdNames[697] = 'HistoryUpdateResult_Encoding_DefaultBinary'
ObjectIdNames[698] = 'HistoryUpdateRequest'
ObjectIdNames[699] = 'HistoryUpdateRequest_Encoding_DefaultXml'
ObjectIdNames[700] = 'HistoryUpdateRequest_Encoding_DefaultBinary'
ObjectIdNames[701] = 'HistoryUpdateResponse'
ObjectIdNames[702] = 'HistoryUpdateResponse_Encoding_DefaultXml'
ObjectIdNames[703] = 'HistoryUpdateResponse_Encoding_DefaultBinary'
ObjectIdNames[704] = 'CallMethodRequest'
ObjectIdNames[705] = 'CallMethodRequest_Encoding_DefaultXml'
ObjectIdNames[706] = 'CallMethodRequest_Encoding_DefaultBinary'
ObjectIdNames[707] = 'CallMethodResult'
ObjectIdNames[708] = 'CallMethodResult_Encoding_DefaultXml'
ObjectIdNames[709] = 'CallMethodResult_Encoding_DefaultBinary'
ObjectIdNames[710] = 'CallRequest'
ObjectIdNames[711] = 'CallRequest_Encoding_DefaultXml'
ObjectIdNames[712] = 'CallRequest_Encoding_DefaultBinary'
ObjectIdNames[713] = 'CallResponse'
ObjectIdNames[714] = 'CallResponse_Encoding_DefaultXml'
ObjectIdNames[715] = 'CallResponse_Encoding_DefaultBinary'
ObjectIdNames[716] = 'MonitoringMode'
ObjectIdNames[717] = 'DataChangeTrigger'
ObjectIdNames[718] = 'DeadbandType'
ObjectIdNames[719] = 'MonitoringFilter'
ObjectIdNames[720] = 'MonitoringFilter_Encoding_DefaultXml'
ObjectIdNames[721] = 'MonitoringFilter_Encoding_DefaultBinary'
ObjectIdNames[722] = 'DataChangeFilter'
ObjectIdNames[723] = 'DataChangeFilter_Encoding_DefaultXml'
ObjectIdNames[724] = 'DataChangeFilter_Encoding_DefaultBinary'
ObjectIdNames[725] = 'EventFilter'
ObjectIdNames[726] = 'EventFilter_Encoding_DefaultXml'
ObjectIdNames[727] = 'EventFilter_Encoding_DefaultBinary'
ObjectIdNames[728] = 'AggregateFilter'
ObjectIdNames[729] = 'AggregateFilter_Encoding_DefaultXml'
ObjectIdNames[730] = 'AggregateFilter_Encoding_DefaultBinary'
ObjectIdNames[731] = 'MonitoringFilterResult'
ObjectIdNames[732] = 'MonitoringFilterResult_Encoding_DefaultXml'
ObjectIdNames[733] = 'MonitoringFilterResult_Encoding_DefaultBinary'
ObjectIdNames[734] = 'EventFilterResult'
ObjectIdNames[735] = 'EventFilterResult_Encoding_DefaultXml'
ObjectIdNames[736] = 'EventFilterResult_Encoding_DefaultBinary'
ObjectIdNames[737] = 'AggregateFilterResult'
ObjectIdNames[738] = 'AggregateFilterResult_Encoding_DefaultXml'
ObjectIdNames[739] = 'AggregateFilterResult_Encoding_DefaultBinary'
ObjectIdNames[740] = 'MonitoringParameters'
ObjectIdNames[741] = 'MonitoringParameters_Encoding_DefaultXml'
ObjectIdNames[742] = 'MonitoringParameters_Encoding_DefaultBinary'
ObjectIdNames[743] = 'MonitoredItemCreateRequest'
ObjectIdNames[744] = 'MonitoredItemCreateRequest_Encoding_DefaultXml'
ObjectIdNames[745] = 'MonitoredItemCreateRequest_Encoding_DefaultBinary'
ObjectIdNames[746] = 'MonitoredItemCreateResult'
ObjectIdNames[747] = 'MonitoredItemCreateResult_Encoding_DefaultXml'
ObjectIdNames[748] = 'MonitoredItemCreateResult_Encoding_DefaultBinary'
ObjectIdNames[749] = 'CreateMonitoredItemsRequest'
ObjectIdNames[750] = 'CreateMonitoredItemsRequest_Encoding_DefaultXml'
ObjectIdNames[751] = 'CreateMonitoredItemsRequest_Encoding_DefaultBinary'
ObjectIdNames[752] = 'CreateMonitoredItemsResponse'
ObjectIdNames[753] = 'CreateMonitoredItemsResponse_Encoding_DefaultXml'
ObjectIdNames[754] = 'CreateMonitoredItemsResponse_Encoding_DefaultBinary'
ObjectIdNames[755] = 'MonitoredItemModifyRequest'
ObjectIdNames[756] = 'MonitoredItemModifyRequest_Encoding_DefaultXml'
ObjectIdNames[757] = 'MonitoredItemModifyRequest_Encoding_DefaultBinary'
ObjectIdNames[758] = 'MonitoredItemModifyResult'
ObjectIdNames[759] = 'MonitoredItemModifyResult_Encoding_DefaultXml'
ObjectIdNames[760] = 'MonitoredItemModifyResult_Encoding_DefaultBinary'
ObjectIdNames[761] = 'ModifyMonitoredItemsRequest'
ObjectIdNames[762] = 'ModifyMonitoredItemsRequest_Encoding_DefaultXml'
ObjectIdNames[763] = 'ModifyMonitoredItemsRequest_Encoding_DefaultBinary'
ObjectIdNames[764] = 'ModifyMonitoredItemsResponse'
ObjectIdNames[765] = 'ModifyMonitoredItemsResponse_Encoding_DefaultXml'
ObjectIdNames[766] = 'ModifyMonitoredItemsResponse_Encoding_DefaultBinary'
ObjectIdNames[767] = 'SetMonitoringModeRequest'
ObjectIdNames[768] = 'SetMonitoringModeRequest_Encoding_DefaultXml'
ObjectIdNames[769] = 'SetMonitoringModeRequest_Encoding_DefaultBinary'
ObjectIdNames[770] = 'SetMonitoringModeResponse'
ObjectIdNames[771] = 'SetMonitoringModeResponse_Encoding_DefaultXml'
ObjectIdNames[772] = 'SetMonitoringModeResponse_Encoding_DefaultBinary'
ObjectIdNames[773] = 'SetTriggeringRequest'
ObjectIdNames[774] = 'SetTriggeringRequest_Encoding_DefaultXml'
ObjectIdNames[775] = 'SetTriggeringRequest_Encoding_DefaultBinary'
ObjectIdNames[776] = 'SetTriggeringResponse'
ObjectIdNames[777] = 'SetTriggeringResponse_Encoding_DefaultXml'
ObjectIdNames[778] = 'SetTriggeringResponse_Encoding_DefaultBinary'
ObjectIdNames[779] = 'DeleteMonitoredItemsRequest'
ObjectIdNames[780] = 'DeleteMonitoredItemsRequest_Encoding_DefaultXml'
ObjectIdNames[781] = 'DeleteMonitoredItemsRequest_Encoding_DefaultBinary'
ObjectIdNames[782] = 'DeleteMonitoredItemsResponse'
ObjectIdNames[783] = 'DeleteMonitoredItemsResponse_Encoding_DefaultXml'
ObjectIdNames[784] = 'DeleteMonitoredItemsResponse_Encoding_DefaultBinary'
ObjectIdNames[785] = 'CreateSubscriptionRequest'
ObjectIdNames[786] = 'CreateSubscriptionRequest_Encoding_DefaultXml'
ObjectIdNames[787] = 'CreateSubscriptionRequest_Encoding_DefaultBinary'
ObjectIdNames[788] = 'CreateSubscriptionResponse'
ObjectIdNames[789] = 'CreateSubscriptionResponse_Encoding_DefaultXml'
ObjectIdNames[790] = 'CreateSubscriptionResponse_Encoding_DefaultBinary'
ObjectIdNames[791] = 'ModifySubscriptionRequest'
ObjectIdNames[792] = 'ModifySubscriptionRequest_Encoding_DefaultXml'
ObjectIdNames[793] = 'ModifySubscriptionRequest_Encoding_DefaultBinary'
ObjectIdNames[794] = 'ModifySubscriptionResponse'
ObjectIdNames[795] = 'ModifySubscriptionResponse_Encoding_DefaultXml'
ObjectIdNames[796] = 'ModifySubscriptionResponse_Encoding_DefaultBinary'
ObjectIdNames[797] = 'SetPublishingModeRequest'
ObjectIdNames[798] = 'SetPublishingModeRequest_Encoding_DefaultXml'
ObjectIdNames[799] = 'SetPublishingModeRequest_Encoding_DefaultBinary'
ObjectIdNames[800] = 'SetPublishingModeResponse'
ObjectIdNames[801] = 'SetPublishingModeResponse_Encoding_DefaultXml'
ObjectIdNames[802] = 'SetPublishingModeResponse_Encoding_DefaultBinary'
ObjectIdNames[803] = 'NotificationMessage'
ObjectIdNames[804] = 'NotificationMessage_Encoding_DefaultXml'
ObjectIdNames[805] = 'NotificationMessage_Encoding_DefaultBinary'
ObjectIdNames[806] = 'MonitoredItemNotification'
ObjectIdNames[807] = 'MonitoredItemNotification_Encoding_DefaultXml'
ObjectIdNames[808] = 'MonitoredItemNotification_Encoding_DefaultBinary'
ObjectIdNames[809] = 'DataChangeNotification'
ObjectIdNames[810] = 'DataChangeNotification_Encoding_DefaultXml'
ObjectIdNames[811] = 'DataChangeNotification_Encoding_DefaultBinary'
ObjectIdNames[818] = 'StatusChangeNotification'
ObjectIdNames[819] = 'StatusChangeNotification_Encoding_DefaultXml'
ObjectIdNames[820] = 'StatusChangeNotification_Encoding_DefaultBinary'
ObjectIdNames[821] = 'SubscriptionAcknowledgement'
ObjectIdNames[822] = 'SubscriptionAcknowledgement_Encoding_DefaultXml'
ObjectIdNames[823] = 'SubscriptionAcknowledgement_Encoding_DefaultBinary'
ObjectIdNames[824] = 'PublishRequest'
ObjectIdNames[825] = 'PublishRequest_Encoding_DefaultXml'
ObjectIdNames[826] = 'PublishRequest_Encoding_DefaultBinary'
ObjectIdNames[827] = 'PublishResponse'
ObjectIdNames[828] = 'PublishResponse_Encoding_DefaultXml'
ObjectIdNames[829] = 'PublishResponse_Encoding_DefaultBinary'
ObjectIdNames[830] = 'RepublishRequest'
ObjectIdNames[831] = 'RepublishRequest_Encoding_DefaultXml'
ObjectIdNames[832] = 'RepublishRequest_Encoding_DefaultBinary'
ObjectIdNames[833] = 'RepublishResponse'
ObjectIdNames[834] = 'RepublishResponse_Encoding_DefaultXml'
ObjectIdNames[835] = 'RepublishResponse_Encoding_DefaultBinary'
ObjectIdNames[836] = 'TransferResult'
ObjectIdNames[837] = 'TransferResult_Encoding_DefaultXml'
ObjectIdNames[838] = 'TransferResult_Encoding_DefaultBinary'
ObjectIdNames[839] = 'TransferSubscriptionsRequest'
ObjectIdNames[840] = 'TransferSubscriptionsRequest_Encoding_DefaultXml'
ObjectIdNames[841] = 'TransferSubscriptionsRequest_Encoding_DefaultBinary'
ObjectIdNames[842] = 'TransferSubscriptionsResponse'
ObjectIdNames[843] = 'TransferSubscriptionsResponse_Encoding_DefaultXml'
ObjectIdNames[844] = 'TransferSubscriptionsResponse_Encoding_DefaultBinary'
ObjectIdNames[845] = 'DeleteSubscriptionsRequest'
ObjectIdNames[846] = 'DeleteSubscriptionsRequest_Encoding_DefaultXml'
ObjectIdNames[847] = 'DeleteSubscriptionsRequest_Encoding_DefaultBinary'
ObjectIdNames[848] = 'DeleteSubscriptionsResponse'
ObjectIdNames[849] = 'DeleteSubscriptionsResponse_Encoding_DefaultXml'
ObjectIdNames[850] = 'DeleteSubscriptionsResponse_Encoding_DefaultBinary'
ObjectIdNames[851] = 'RedundancySupport'
ObjectIdNames[852] = 'ServerState'
ObjectIdNames[853] = 'RedundantServerDataType'
ObjectIdNames[854] = 'RedundantServerDataType_Encoding_DefaultXml'
ObjectIdNames[855] = 'RedundantServerDataType_Encoding_DefaultBinary'
ObjectIdNames[856] = 'SamplingIntervalDiagnosticsDataType'
ObjectIdNames[857] = 'SamplingIntervalDiagnosticsDataType_Encoding_DefaultXml'
ObjectIdNames[858] = 'SamplingIntervalDiagnosticsDataType_Encoding_DefaultBinary'
ObjectIdNames[859] = 'ServerDiagnosticsSummaryDataType'
ObjectIdNames[860] = 'ServerDiagnosticsSummaryDataType_Encoding_DefaultXml'
ObjectIdNames[861] = 'ServerDiagnosticsSummaryDataType_Encoding_DefaultBinary'
ObjectIdNames[862] = 'ServerStatusDataType'
ObjectIdNames[863] = 'ServerStatusDataType_Encoding_DefaultXml'
ObjectIdNames[864] = 'ServerStatusDataType_Encoding_DefaultBinary'
ObjectIdNames[865] = 'SessionDiagnosticsDataType'
ObjectIdNames[866] = 'SessionDiagnosticsDataType_Encoding_DefaultXml'
ObjectIdNames[867] = 'SessionDiagnosticsDataType_Encoding_DefaultBinary'
ObjectIdNames[868] = 'SessionSecurityDiagnosticsDataType'
ObjectIdNames[869] = 'SessionSecurityDiagnosticsDataType_Encoding_DefaultXml'
ObjectIdNames[870] = 'SessionSecurityDiagnosticsDataType_Encoding_DefaultBinary'
ObjectIdNames[871] = 'ServiceCounterDataType'
ObjectIdNames[872] = 'ServiceCounterDataType_Encoding_DefaultXml'
ObjectIdNames[873] = 'ServiceCounterDataType_Encoding_DefaultBinary'
ObjectIdNames[874] = 'SubscriptionDiagnosticsDataType'
ObjectIdNames[875] = 'SubscriptionDiagnosticsDataType_Encoding_DefaultXml'
ObjectIdNames[876] = 'SubscriptionDiagnosticsDataType_Encoding_DefaultBinary'
ObjectIdNames[877] = 'ModelChangeStructureDataType'
ObjectIdNames[878] = 'ModelChangeStructureDataType_Encoding_DefaultXml'
ObjectIdNames[879] = 'ModelChangeStructureDataType_Encoding_DefaultBinary'
ObjectIdNames[884] = 'Range'
ObjectIdNames[885] = 'Range_Encoding_DefaultXml'
ObjectIdNames[886] = 'Range_Encoding_DefaultBinary'
ObjectIdNames[887] = 'EUInformation'
ObjectIdNames[888] = 'EUInformation_Encoding_DefaultXml'
ObjectIdNames[889] = 'EUInformation_Encoding_DefaultBinary'
ObjectIdNames[890] = 'ExceptionDeviationFormat'
ObjectIdNames[891] = 'Annotation'
ObjectIdNames[892] = 'Annotation_Encoding_DefaultXml'
ObjectIdNames[893] = 'Annotation_Encoding_DefaultBinary'
ObjectIdNames[894] = 'ProgramDiagnosticDataType'
ObjectIdNames[895] = 'ProgramDiagnosticDataType_Encoding_DefaultXml'
ObjectIdNames[896] = 'ProgramDiagnosticDataType_Encoding_DefaultBinary'
ObjectIdNames[897] = 'SemanticChangeStructureDataType'
ObjectIdNames[898] = 'SemanticChangeStructureDataType_Encoding_DefaultXml'
ObjectIdNames[899] = 'SemanticChangeStructureDataType_Encoding_DefaultBinary'
ObjectIdNames[914] = 'EventNotificationList'
ObjectIdNames[915] = 'EventNotificationList_Encoding_DefaultXml'
ObjectIdNames[916] = 'EventNotificationList_Encoding_DefaultBinary'
ObjectIdNames[917] = 'EventFieldList'
ObjectIdNames[918] = 'EventFieldList_Encoding_DefaultXml'
ObjectIdNames[919] = 'EventFieldList_Encoding_DefaultBinary'
ObjectIdNames[920] = 'HistoryEventFieldList'
ObjectIdNames[921] = 'HistoryEventFieldList_Encoding_DefaultXml'
ObjectIdNames[922] = 'HistoryEventFieldList_Encoding_DefaultBinary'
ObjectIdNames[938] = 'IssuedIdentityToken'
ObjectIdNames[939] = 'IssuedIdentityToken_Encoding_DefaultXml'
ObjectIdNames[940] = 'IssuedIdentityToken_Encoding_DefaultBinary'
ObjectIdNames[945] = 'NotificationData'
ObjectIdNames[946] = 'NotificationData_Encoding_DefaultXml'
ObjectIdNames[947] = 'NotificationData_Encoding_DefaultBinary'
ObjectIdNames[948] = 'AggregateConfiguration'
ObjectIdNames[949] = 'AggregateConfiguration_Encoding_DefaultXml'
ObjectIdNames[950] = 'AggregateConfiguration_Encoding_DefaultBinary'
ObjectIdNames[2000] = 'ImageBMP'
ObjectIdNames[2001] = 'ImageGIF'
ObjectIdNames[2002] = 'ImageJPG'
ObjectIdNames[2003] = 'ImagePNG'
ObjectIdNames[2004] = 'ServerType'
ObjectIdNames[2005] = 'ServerType_ServerArray'
ObjectIdNames[2006] = 'ServerType_NamespaceArray'
ObjectIdNames[2007] = 'ServerType_ServerStatus'
ObjectIdNames[2008] = 'ServerType_ServiceLevel'
ObjectIdNames[2009] = 'ServerType_ServerCapabilities'
ObjectIdNames[2010] = 'ServerType_ServerDiagnostics'
ObjectIdNames[2011] = 'ServerType_VendorServerInfo'
ObjectIdNames[2012] = 'ServerType_ServerRedundancy'
ObjectIdNames[2013] = 'ServerCapabilitiesType'
ObjectIdNames[2014] = 'ServerCapabilitiesType_ServerProfileArray'
ObjectIdNames[2016] = 'ServerCapabilitiesType_LocaleIdArray'
ObjectIdNames[2017] = 'ServerCapabilitiesType_MinSupportedSampleRate'
ObjectIdNames[2019] = 'ServerCapabilitiesType_ModellingRules'
ObjectIdNames[2020] = 'ServerDiagnosticsType'
ObjectIdNames[2021] = 'ServerDiagnosticsType_ServerDiagnosticsSummary'
ObjectIdNames[2022] = 'ServerDiagnosticsType_SamplingIntervalDiagnosticsArray'
ObjectIdNames[2023] = 'ServerDiagnosticsType_SubscriptionDiagnosticsArray'
ObjectIdNames[2025] = 'ServerDiagnosticsType_EnabledFlag'
ObjectIdNames[2026] = 'SessionsDiagnosticsSummaryType'
ObjectIdNames[2027] = 'SessionsDiagnosticsSummaryType_SessionDiagnosticsArray'
ObjectIdNames[2028] = 'SessionsDiagnosticsSummaryType_SessionSecurityDiagnosticsArray'
ObjectIdNames[2029] = 'SessionDiagnosticsObjectType'
ObjectIdNames[2030] = 'SessionDiagnosticsObjectType_SessionDiagnostics'
ObjectIdNames[2031] = 'SessionDiagnosticsObjectType_SessionSecurityDiagnostics'
ObjectIdNames[2032] = 'SessionDiagnosticsObjectType_SubscriptionDiagnosticsArray'
ObjectIdNames[2033] = 'VendorServerInfoType'
ObjectIdNames[2034] = 'ServerRedundancyType'
ObjectIdNames[2035] = 'ServerRedundancyType_RedundancySupport'
ObjectIdNames[2036] = 'TransparentRedundancyType'
ObjectIdNames[2037] = 'TransparentRedundancyType_CurrentServerId'
ObjectIdNames[2038] = 'TransparentRedundancyType_RedundantServerArray'
ObjectIdNames[2039] = 'NonTransparentRedundancyType'
ObjectIdNames[2040] = 'NonTransparentRedundancyType_ServerUriArray'
ObjectIdNames[2041] = 'BaseEventType'
ObjectIdNames[2042] = 'BaseEventType_EventId'
ObjectIdNames[2043] = 'BaseEventType_EventType'
ObjectIdNames[2044] = 'BaseEventType_SourceNode'
ObjectIdNames[2045] = 'BaseEventType_SourceName'
ObjectIdNames[2046] = 'BaseEventType_Time'
ObjectIdNames[2047] = 'BaseEventType_ReceiveTime'
ObjectIdNames[2050] = 'BaseEventType_Message'
ObjectIdNames[2051] = 'BaseEventType_Severity'
ObjectIdNames[2052] = 'AuditEventType'
ObjectIdNames[2053] = 'AuditEventType_ActionTimeStamp'
ObjectIdNames[2054] = 'AuditEventType_Status'
ObjectIdNames[2055] = 'AuditEventType_ServerId'
ObjectIdNames[2056] = 'AuditEventType_ClientAuditEntryId'
ObjectIdNames[2057] = 'AuditEventType_ClientUserId'
ObjectIdNames[2058] = 'AuditSecurityEventType'
ObjectIdNames[2059] = 'AuditChannelEventType'
ObjectIdNames[2060] = 'AuditOpenSecureChannelEventType'
ObjectIdNames[2061] = 'AuditOpenSecureChannelEventType_ClientCertificate'
ObjectIdNames[2062] = 'AuditOpenSecureChannelEventType_RequestType'
ObjectIdNames[2063] = 'AuditOpenSecureChannelEventType_SecurityPolicyUri'
ObjectIdNames[2065] = 'AuditOpenSecureChannelEventType_SecurityMode'
ObjectIdNames[2066] = 'AuditOpenSecureChannelEventType_RequestedLifetime'
ObjectIdNames[2069] = 'AuditSessionEventType'
ObjectIdNames[2070] = 'AuditSessionEventType_SessionId'
ObjectIdNames[2071] = 'AuditCreateSessionEventType'
ObjectIdNames[2072] = 'AuditCreateSessionEventType_SecureChannelId'
ObjectIdNames[2073] = 'AuditCreateSessionEventType_ClientCertificate'
ObjectIdNames[2074] = 'AuditCreateSessionEventType_RevisedSessionTimeout'
ObjectIdNames[2075] = 'AuditActivateSessionEventType'
ObjectIdNames[2076] = 'AuditActivateSessionEventType_ClientSoftwareCertificates'
ObjectIdNames[2077] = 'AuditActivateSessionEventType_UserIdentityToken'
ObjectIdNames[2078] = 'AuditCancelEventType'
ObjectIdNames[2079] = 'AuditCancelEventType_RequestHandle'
ObjectIdNames[2080] = 'AuditCertificateEventType'
ObjectIdNames[2081] = 'AuditCertificateEventType_Certificate'
ObjectIdNames[2082] = 'AuditCertificateDataMismatchEventType'
ObjectIdNames[2083] = 'AuditCertificateDataMismatchEventType_InvalidHostname'
ObjectIdNames[2084] = 'AuditCertificateDataMismatchEventType_InvalidUri'
ObjectIdNames[2085] = 'AuditCertificateExpiredEventType'
ObjectIdNames[2086] = 'AuditCertificateInvalidEventType'
ObjectIdNames[2087] = 'AuditCertificateUntrustedEventType'
ObjectIdNames[2088] = 'AuditCertificateRevokedEventType'
ObjectIdNames[2089] = 'AuditCertificateMismatchEventType'
ObjectIdNames[2090] = 'AuditNodeManagementEventType'
ObjectIdNames[2091] = 'AuditAddNodesEventType'
ObjectIdNames[2092] = 'AuditAddNodesEventType_NodesToAdd'
ObjectIdNames[2093] = 'AuditDeleteNodesEventType'
ObjectIdNames[2094] = 'AuditDeleteNodesEventType_NodesToDelete'
ObjectIdNames[2095] = 'AuditAddReferencesEventType'
ObjectIdNames[2096] = 'AuditAddReferencesEventType_ReferencesToAdd'
ObjectIdNames[2097] = 'AuditDeleteReferencesEventType'
ObjectIdNames[2098] = 'AuditDeleteReferencesEventType_ReferencesToDelete'
ObjectIdNames[2099] = 'AuditUpdateEventType'
ObjectIdNames[2100] = 'AuditWriteUpdateEventType'
ObjectIdNames[2101] = 'AuditWriteUpdateEventType_IndexRange'
ObjectIdNames[2102] = 'AuditWriteUpdateEventType_OldValue'
ObjectIdNames[2103] = 'AuditWriteUpdateEventType_NewValue'
ObjectIdNames[2104] = 'AuditHistoryUpdateEventType'
ObjectIdNames[2127] = 'AuditUpdateMethodEventType'
ObjectIdNames[2128] = 'AuditUpdateMethodEventType_MethodId'
ObjectIdNames[2129] = 'AuditUpdateMethodEventType_InputArguments'
ObjectIdNames[2130] = 'SystemEventType'
ObjectIdNames[2131] = 'DeviceFailureEventType'
ObjectIdNames[2132] = 'BaseModelChangeEventType'
ObjectIdNames[2133] = 'GeneralModelChangeEventType'
ObjectIdNames[2134] = 'GeneralModelChangeEventType_Changes'
ObjectIdNames[2137] = 'ServerVendorCapabilityType'
ObjectIdNames[2138] = 'ServerStatusType'
ObjectIdNames[2139] = 'ServerStatusType_StartTime'
ObjectIdNames[2140] = 'ServerStatusType_CurrentTime'
ObjectIdNames[2141] = 'ServerStatusType_State'
ObjectIdNames[2142] = 'ServerStatusType_BuildInfo'
ObjectIdNames[2150] = 'ServerDiagnosticsSummaryType'
ObjectIdNames[2151] = 'ServerDiagnosticsSummaryType_ServerViewCount'
ObjectIdNames[2152] = 'ServerDiagnosticsSummaryType_CurrentSessionCount'
ObjectIdNames[2153] = 'ServerDiagnosticsSummaryType_CumulatedSessionCount'
ObjectIdNames[2154] = 'ServerDiagnosticsSummaryType_SecurityRejectedSessionCount'
ObjectIdNames[2155] = 'ServerDiagnosticsSummaryType_RejectedSessionCount'
ObjectIdNames[2156] = 'ServerDiagnosticsSummaryType_SessionTimeoutCount'
ObjectIdNames[2157] = 'ServerDiagnosticsSummaryType_SessionAbortCount'
ObjectIdNames[2159] = 'ServerDiagnosticsSummaryType_PublishingIntervalCount'
ObjectIdNames[2160] = 'ServerDiagnosticsSummaryType_CurrentSubscriptionCount'
ObjectIdNames[2161] = 'ServerDiagnosticsSummaryType_CumulatedSubscriptionCount'
ObjectIdNames[2162] = 'ServerDiagnosticsSummaryType_SecurityRejectedRequestsCount'
ObjectIdNames[2163] = 'ServerDiagnosticsSummaryType_RejectedRequestsCount'
ObjectIdNames[2164] = 'SamplingIntervalDiagnosticsArrayType'
ObjectIdNames[2165] = 'SamplingIntervalDiagnosticsType'
ObjectIdNames[2166] = 'SamplingIntervalDiagnosticsType_SamplingInterval'
ObjectIdNames[2171] = 'SubscriptionDiagnosticsArrayType'
ObjectIdNames[2172] = 'SubscriptionDiagnosticsType'
ObjectIdNames[2173] = 'SubscriptionDiagnosticsType_SessionId'
ObjectIdNames[2174] = 'SubscriptionDiagnosticsType_SubscriptionId'
ObjectIdNames[2175] = 'SubscriptionDiagnosticsType_Priority'
ObjectIdNames[2176] = 'SubscriptionDiagnosticsType_PublishingInterval'
ObjectIdNames[2177] = 'SubscriptionDiagnosticsType_MaxKeepAliveCount'
ObjectIdNames[2179] = 'SubscriptionDiagnosticsType_MaxNotificationsPerPublish'
ObjectIdNames[2180] = 'SubscriptionDiagnosticsType_PublishingEnabled'
ObjectIdNames[2181] = 'SubscriptionDiagnosticsType_ModifyCount'
ObjectIdNames[2182] = 'SubscriptionDiagnosticsType_EnableCount'
ObjectIdNames[2183] = 'SubscriptionDiagnosticsType_DisableCount'
ObjectIdNames[2184] = 'SubscriptionDiagnosticsType_RepublishRequestCount'
ObjectIdNames[2185] = 'SubscriptionDiagnosticsType_RepublishMessageRequestCount'
ObjectIdNames[2186] = 'SubscriptionDiagnosticsType_RepublishMessageCount'
ObjectIdNames[2187] = 'SubscriptionDiagnosticsType_TransferRequestCount'
ObjectIdNames[2188] = 'SubscriptionDiagnosticsType_TransferredToAltClientCount'
ObjectIdNames[2189] = 'SubscriptionDiagnosticsType_TransferredToSameClientCount'
ObjectIdNames[2190] = 'SubscriptionDiagnosticsType_PublishRequestCount'
ObjectIdNames[2191] = 'SubscriptionDiagnosticsType_DataChangeNotificationsCount'
ObjectIdNames[2193] = 'SubscriptionDiagnosticsType_NotificationsCount'
ObjectIdNames[2196] = 'SessionDiagnosticsArrayType'
ObjectIdNames[2197] = 'SessionDiagnosticsVariableType'
ObjectIdNames[2198] = 'SessionDiagnosticsVariableType_SessionId'
ObjectIdNames[2199] = 'SessionDiagnosticsVariableType_SessionName'
ObjectIdNames[2200] = 'SessionDiagnosticsVariableType_ClientDescription'
ObjectIdNames[2201] = 'SessionDiagnosticsVariableType_ServerUri'
ObjectIdNames[2202] = 'SessionDiagnosticsVariableType_EndpointUrl'
ObjectIdNames[2203] = 'SessionDiagnosticsVariableType_LocaleIds'
ObjectIdNames[2204] = 'SessionDiagnosticsVariableType_ActualSessionTimeout'
ObjectIdNames[2205] = 'SessionDiagnosticsVariableType_ClientConnectionTime'
ObjectIdNames[2206] = 'SessionDiagnosticsVariableType_ClientLastContactTime'
ObjectIdNames[2207] = 'SessionDiagnosticsVariableType_CurrentSubscriptionsCount'
ObjectIdNames[2208] = 'SessionDiagnosticsVariableType_CurrentMonitoredItemsCount'
ObjectIdNames[2209] = 'SessionDiagnosticsVariableType_CurrentPublishRequestsInQueue'
ObjectIdNames[2217] = 'SessionDiagnosticsVariableType_ReadCount'
ObjectIdNames[2218] = 'SessionDiagnosticsVariableType_HistoryReadCount'
ObjectIdNames[2219] = 'SessionDiagnosticsVariableType_WriteCount'
ObjectIdNames[2220] = 'SessionDiagnosticsVariableType_HistoryUpdateCount'
ObjectIdNames[2221] = 'SessionDiagnosticsVariableType_CallCount'
ObjectIdNames[2222] = 'SessionDiagnosticsVariableType_CreateMonitoredItemsCount'
ObjectIdNames[2223] = 'SessionDiagnosticsVariableType_ModifyMonitoredItemsCount'
ObjectIdNames[2224] = 'SessionDiagnosticsVariableType_SetMonitoringModeCount'
ObjectIdNames[2225] = 'SessionDiagnosticsVariableType_SetTriggeringCount'
ObjectIdNames[2226] = 'SessionDiagnosticsVariableType_DeleteMonitoredItemsCount'
ObjectIdNames[2227] = 'SessionDiagnosticsVariableType_CreateSubscriptionCount'
ObjectIdNames[2228] = 'SessionDiagnosticsVariableType_ModifySubscriptionCount'
ObjectIdNames[2229] = 'SessionDiagnosticsVariableType_SetPublishingModeCount'
ObjectIdNames[2230] = 'SessionDiagnosticsVariableType_PublishCount'
ObjectIdNames[2231] = 'SessionDiagnosticsVariableType_RepublishCount'
ObjectIdNames[2232] = 'SessionDiagnosticsVariableType_TransferSubscriptionsCount'
ObjectIdNames[2233] = 'SessionDiagnosticsVariableType_DeleteSubscriptionsCount'
ObjectIdNames[2234] = 'SessionDiagnosticsVariableType_AddNodesCount'
ObjectIdNames[2235] = 'SessionDiagnosticsVariableType_AddReferencesCount'
ObjectIdNames[2236] = 'SessionDiagnosticsVariableType_DeleteNodesCount'
ObjectIdNames[2237] = 'SessionDiagnosticsVariableType_DeleteReferencesCount'
ObjectIdNames[2238] = 'SessionDiagnosticsVariableType_BrowseCount'
ObjectIdNames[2239] = 'SessionDiagnosticsVariableType_BrowseNextCount'
ObjectIdNames[2240] = 'SessionDiagnosticsVariableType_TranslateBrowsePathsToNodeIdsCount'
ObjectIdNames[2241] = 'SessionDiagnosticsVariableType_QueryFirstCount'
ObjectIdNames[2242] = 'SessionDiagnosticsVariableType_QueryNextCount'
ObjectIdNames[2243] = 'SessionSecurityDiagnosticsArrayType'
ObjectIdNames[2244] = 'SessionSecurityDiagnosticsType'
ObjectIdNames[2245] = 'SessionSecurityDiagnosticsType_SessionId'
ObjectIdNames[2246] = 'SessionSecurityDiagnosticsType_ClientUserIdOfSession'
ObjectIdNames[2247] = 'SessionSecurityDiagnosticsType_ClientUserIdHistory'
ObjectIdNames[2248] = 'SessionSecurityDiagnosticsType_AuthenticationMechanism'
ObjectIdNames[2249] = 'SessionSecurityDiagnosticsType_Encoding'
ObjectIdNames[2250] = 'SessionSecurityDiagnosticsType_TransportProtocol'
ObjectIdNames[2251] = 'SessionSecurityDiagnosticsType_SecurityMode'
ObjectIdNames[2252] = 'SessionSecurityDiagnosticsType_SecurityPolicyUri'
ObjectIdNames[2253] = 'Server'
ObjectIdNames[2254] = 'Server_ServerArray'
ObjectIdNames[2255] = 'Server_NamespaceArray'
ObjectIdNames[2256] = 'Server_ServerStatus'
ObjectIdNames[2257] = 'Server_ServerStatus_StartTime'
ObjectIdNames[2258] = 'Server_ServerStatus_CurrentTime'
ObjectIdNames[2259] = 'Server_ServerStatus_State'
ObjectIdNames[2260] = 'Server_ServerStatus_BuildInfo'
ObjectIdNames[2261] = 'Server_ServerStatus_BuildInfo_ProductName'
ObjectIdNames[2262] = 'Server_ServerStatus_BuildInfo_ProductUri'
ObjectIdNames[2263] = 'Server_ServerStatus_BuildInfo_ManufacturerName'
ObjectIdNames[2264] = 'Server_ServerStatus_BuildInfo_SoftwareVersion'
ObjectIdNames[2265] = 'Server_ServerStatus_BuildInfo_BuildNumber'
ObjectIdNames[2266] = 'Server_ServerStatus_BuildInfo_BuildDate'
ObjectIdNames[2267] = 'Server_ServiceLevel'
ObjectIdNames[2268] = 'Server_ServerCapabilities'
ObjectIdNames[2269] = 'Server_ServerCapabilities_ServerProfileArray'
ObjectIdNames[2271] = 'Server_ServerCapabilities_LocaleIdArray'
ObjectIdNames[2272] = 'Server_ServerCapabilities_MinSupportedSampleRate'
ObjectIdNames[2274] = 'Server_ServerDiagnostics'
ObjectIdNames[2275] = 'Server_ServerDiagnostics_ServerDiagnosticsSummary'
ObjectIdNames[2276] = 'Server_ServerDiagnostics_ServerDiagnosticsSummary_ServerViewCount'
ObjectIdNames[2277] = 'Server_ServerDiagnostics_ServerDiagnosticsSummary_CurrentSessionCount'
ObjectIdNames[2278] = 'Server_ServerDiagnostics_ServerDiagnosticsSummary_CumulatedSessionCount'
ObjectIdNames[2279] = 'Server_ServerDiagnostics_ServerDiagnosticsSummary_SecurityRejectedSessionCount'
ObjectIdNames[2281] = 'Server_ServerDiagnostics_ServerDiagnosticsSummary_SessionTimeoutCount'
ObjectIdNames[2282] = 'Server_ServerDiagnostics_ServerDiagnosticsSummary_SessionAbortCount'
ObjectIdNames[2284] = 'Server_ServerDiagnostics_ServerDiagnosticsSummary_PublishingIntervalCount'
ObjectIdNames[2285] = 'Server_ServerDiagnostics_ServerDiagnosticsSummary_CurrentSubscriptionCount'
ObjectIdNames[2286] = 'Server_ServerDiagnostics_ServerDiagnosticsSummary_CumulatedSubscriptionCount'
ObjectIdNames[2287] = 'Server_ServerDiagnostics_ServerDiagnosticsSummary_SecurityRejectedRequestsCount'
ObjectIdNames[2288] = 'Server_ServerDiagnostics_ServerDiagnosticsSummary_RejectedRequestsCount'
ObjectIdNames[2289] = 'Server_ServerDiagnostics_SamplingIntervalDiagnosticsArray'
ObjectIdNames[2290] = 'Server_ServerDiagnostics_SubscriptionDiagnosticsArray'
ObjectIdNames[2294] = 'Server_ServerDiagnostics_EnabledFlag'
ObjectIdNames[2295] = 'Server_VendorServerInfo'
ObjectIdNames[2296] = 'Server_ServerRedundancy'
ObjectIdNames[2299] = 'StateMachineType'
ObjectIdNames[2307] = 'StateType'
ObjectIdNames[2308] = 'StateType_StateNumber'
ObjectIdNames[2309] = 'InitialStateType'
ObjectIdNames[2310] = 'TransitionType'
ObjectIdNames[2311] = 'TransitionEventType'
ObjectIdNames[2312] = 'TransitionType_TransitionNumber'
ObjectIdNames[2315] = 'AuditUpdateStateEventType'
ObjectIdNames[2318] = 'HistoricalDataConfigurationType'
ObjectIdNames[2323] = 'HistoricalDataConfigurationType_Stepped'
ObjectIdNames[2324] = 'HistoricalDataConfigurationType_Definition'
ObjectIdNames[2325] = 'HistoricalDataConfigurationType_MaxTimeInterval'
ObjectIdNames[2326] = 'HistoricalDataConfigurationType_MinTimeInterval'
ObjectIdNames[2327] = 'HistoricalDataConfigurationType_ExceptionDeviation'
ObjectIdNames[2328] = 'HistoricalDataConfigurationType_ExceptionDeviationFormat'
ObjectIdNames[2330] = 'HistoryServerCapabilitiesType'
ObjectIdNames[2331] = 'HistoryServerCapabilitiesType_AccessHistoryDataCapability'
ObjectIdNames[2332] = 'HistoryServerCapabilitiesType_AccessHistoryEventsCapability'
ObjectIdNames[2334] = 'HistoryServerCapabilitiesType_InsertDataCapability'
ObjectIdNames[2335] = 'HistoryServerCapabilitiesType_ReplaceDataCapability'
ObjectIdNames[2336] = 'HistoryServerCapabilitiesType_UpdateDataCapability'
ObjectIdNames[2337] = 'HistoryServerCapabilitiesType_DeleteRawCapability'
ObjectIdNames[2338] = 'HistoryServerCapabilitiesType_DeleteAtTimeCapability'
ObjectIdNames[2340] = 'AggregateFunctionType'
ObjectIdNames[2341] = 'AggregateFunction_Interpolative'
ObjectIdNames[2342] = 'AggregateFunction_Average'
ObjectIdNames[2343] = 'AggregateFunction_TimeAverage'
ObjectIdNames[2344] = 'AggregateFunction_Total'
ObjectIdNames[2346] = 'AggregateFunction_Minimum'
ObjectIdNames[2347] = 'AggregateFunction_Maximum'
ObjectIdNames[2348] = 'AggregateFunction_MinimumActualTime'
ObjectIdNames[2349] = 'AggregateFunction_MaximumActualTime'
ObjectIdNames[2350] = 'AggregateFunction_Range'
ObjectIdNames[2351] = 'AggregateFunction_AnnotationCount'
ObjectIdNames[2352] = 'AggregateFunction_Count'
ObjectIdNames[2355] = 'AggregateFunction_NumberOfTransitions'
ObjectIdNames[2357] = 'AggregateFunction_Start'
ObjectIdNames[2358] = 'AggregateFunction_End'
ObjectIdNames[2359] = 'AggregateFunction_Delta'
ObjectIdNames[2360] = 'AggregateFunction_DurationGood'
ObjectIdNames[2361] = 'AggregateFunction_DurationBad'
ObjectIdNames[2362] = 'AggregateFunction_PercentGood'
ObjectIdNames[2363] = 'AggregateFunction_PercentBad'
ObjectIdNames[2364] = 'AggregateFunction_WorstQuality'
ObjectIdNames[2365] = 'DataItemType'
ObjectIdNames[2366] = 'DataItemType_Definition'
ObjectIdNames[2367] = 'DataItemType_ValuePrecision'
ObjectIdNames[2368] = 'AnalogItemType'
ObjectIdNames[2369] = 'AnalogItemType_EURange'
ObjectIdNames[2370] = 'AnalogItemType_InstrumentRange'
ObjectIdNames[2371] = 'AnalogItemType_EngineeringUnits'
ObjectIdNames[2372] = 'DiscreteItemType'
ObjectIdNames[2373] = 'TwoStateDiscreteType'
ObjectIdNames[2374] = 'TwoStateDiscreteType_FalseState'
ObjectIdNames[2375] = 'TwoStateDiscreteType_TrueState'
ObjectIdNames[2376] = 'MultiStateDiscreteType'
ObjectIdNames[2377] = 'MultiStateDiscreteType_EnumStrings'
ObjectIdNames[2378] = 'ProgramTransitionEventType'
ObjectIdNames[2379] = 'ProgramTransitionEventType_IntermediateResult'
ObjectIdNames[2380] = 'ProgramDiagnosticType'
ObjectIdNames[2381] = 'ProgramDiagnosticType_CreateSessionId'
ObjectIdNames[2382] = 'ProgramDiagnosticType_CreateClientName'
ObjectIdNames[2383] = 'ProgramDiagnosticType_InvocationCreationTime'
ObjectIdNames[2384] = 'ProgramDiagnosticType_LastTransitionTime'
ObjectIdNames[2385] = 'ProgramDiagnosticType_LastMethodCall'
ObjectIdNames[2386] = 'ProgramDiagnosticType_LastMethodSessionId'
ObjectIdNames[2387] = 'ProgramDiagnosticType_LastMethodInputArguments'
ObjectIdNames[2388] = 'ProgramDiagnosticType_LastMethodOutputArguments'
ObjectIdNames[2389] = 'ProgramDiagnosticType_LastMethodCallTime'
ObjectIdNames[2390] = 'ProgramDiagnosticType_LastMethodReturnStatus'
ObjectIdNames[2391] = 'ProgramStateMachineType'
ObjectIdNames[2392] = 'ProgramStateMachineType_Creatable'
ObjectIdNames[2393] = 'ProgramStateMachineType_Deletable'
ObjectIdNames[2394] = 'ProgramStateMachineType_AutoDelete'
ObjectIdNames[2395] = 'ProgramStateMachineType_RecycleCount'
ObjectIdNames[2396] = 'ProgramStateMachineType_InstanceCount'
ObjectIdNames[2397] = 'ProgramStateMachineType_MaxInstanceCount'
ObjectIdNames[2398] = 'ProgramStateMachineType_MaxRecycleCount'
ObjectIdNames[2399] = 'ProgramStateMachineType_ProgramDiagnostics'
ObjectIdNames[2400] = 'ProgramStateMachineType_Ready'
ObjectIdNames[2401] = 'ProgramStateMachineType_Ready_StateNumber'
ObjectIdNames[2402] = 'ProgramStateMachineType_Running'
ObjectIdNames[2403] = 'ProgramStateMachineType_Running_StateNumber'
ObjectIdNames[2404] = 'ProgramStateMachineType_Suspended'
ObjectIdNames[2405] = 'ProgramStateMachineType_Suspended_StateNumber'
ObjectIdNames[2406] = 'ProgramStateMachineType_Halted'
ObjectIdNames[2407] = 'ProgramStateMachineType_Halted_StateNumber'
ObjectIdNames[2408] = 'ProgramStateMachineType_HaltedToReady'
ObjectIdNames[2409] = 'ProgramStateMachineType_HaltedToReady_TransitionNumber'
ObjectIdNames[2410] = 'ProgramStateMachineType_ReadyToRunning'
ObjectIdNames[2411] = 'ProgramStateMachineType_ReadyToRunning_TransitionNumber'
ObjectIdNames[2412] = 'ProgramStateMachineType_RunningToHalted'
ObjectIdNames[2413] = 'ProgramStateMachineType_RunningToHalted_TransitionNumber'
ObjectIdNames[2414] = 'ProgramStateMachineType_RunningToReady'
ObjectIdNames[2415] = 'ProgramStateMachineType_RunningToReady_TransitionNumber'
ObjectIdNames[2416] = 'ProgramStateMachineType_RunningToSuspended'
ObjectIdNames[2417] = 'ProgramStateMachineType_RunningToSuspended_TransitionNumber'
ObjectIdNames[2418] = 'ProgramStateMachineType_SuspendedToRunning'
ObjectIdNames[2419] = 'ProgramStateMachineType_SuspendedToRunning_TransitionNumber'
ObjectIdNames[2420] = 'ProgramStateMachineType_SuspendedToHalted'
ObjectIdNames[2421] = 'ProgramStateMachineType_SuspendedToHalted_TransitionNumber'
ObjectIdNames[2422] = 'ProgramStateMachineType_SuspendedToReady'
ObjectIdNames[2423] = 'ProgramStateMachineType_SuspendedToReady_TransitionNumber'
ObjectIdNames[2424] = 'ProgramStateMachineType_ReadyToHalted'
ObjectIdNames[2425] = 'ProgramStateMachineType_ReadyToHalted_TransitionNumber'
ObjectIdNames[2426] = 'ProgramStateMachineType_Start'
ObjectIdNames[2427] = 'ProgramStateMachineType_Suspend'
ObjectIdNames[2428] = 'ProgramStateMachineType_Resume'
ObjectIdNames[2429] = 'ProgramStateMachineType_Halt'
ObjectIdNames[2430] = 'ProgramStateMachineType_Reset'
ObjectIdNames[2730] = 'SessionDiagnosticsVariableType_RegisterNodesCount'
ObjectIdNames[2731] = 'SessionDiagnosticsVariableType_UnregisterNodesCount'
ObjectIdNames[2732] = 'ServerCapabilitiesType_MaxBrowseContinuationPoints'
ObjectIdNames[2733] = 'ServerCapabilitiesType_MaxQueryContinuationPoints'
ObjectIdNames[2734] = 'ServerCapabilitiesType_MaxHistoryContinuationPoints'
ObjectIdNames[2735] = 'Server_ServerCapabilities_MaxBrowseContinuationPoints'
ObjectIdNames[2736] = 'Server_ServerCapabilities_MaxQueryContinuationPoints'
ObjectIdNames[2737] = 'Server_ServerCapabilities_MaxHistoryContinuationPoints'
ObjectIdNames[2738] = 'SemanticChangeEventType'
ObjectIdNames[2739] = 'SemanticChangeEventType_Changes'
ObjectIdNames[2742] = 'ServerType_Auditing'
ObjectIdNames[2744] = 'ServerDiagnosticsType_SessionsDiagnosticsSummary'
ObjectIdNames[2745] = 'AuditChannelEventType_SecureChannelId'
ObjectIdNames[2746] = 'AuditOpenSecureChannelEventType_ClientCertificateThumbprint'
ObjectIdNames[2747] = 'AuditCreateSessionEventType_ClientCertificateThumbprint'
ObjectIdNames[2748] = 'AuditUrlMismatchEventType'
ObjectIdNames[2749] = 'AuditUrlMismatchEventType_EndpointUrl'
ObjectIdNames[2750] = 'AuditWriteUpdateEventType_AttributeId'
ObjectIdNames[2751] = 'AuditHistoryUpdateEventType_ParameterDataTypeId'
ObjectIdNames[2752] = 'ServerStatusType_SecondsTillShutdown'
ObjectIdNames[2753] = 'ServerStatusType_ShutdownReason'
ObjectIdNames[2754] = 'ServerCapabilitiesType_AggregateFunctions'
ObjectIdNames[2755] = 'StateVariableType'
ObjectIdNames[2756] = 'StateVariableType_Id'
ObjectIdNames[2757] = 'StateVariableType_Name'
ObjectIdNames[2758] = 'StateVariableType_Number'
ObjectIdNames[2759] = 'StateVariableType_EffectiveDisplayName'
ObjectIdNames[2760] = 'FiniteStateVariableType'
ObjectIdNames[2761] = 'FiniteStateVariableType_Id'
ObjectIdNames[2762] = 'TransitionVariableType'
ObjectIdNames[2763] = 'TransitionVariableType_Id'
ObjectIdNames[2764] = 'TransitionVariableType_Name'
ObjectIdNames[2765] = 'TransitionVariableType_Number'
ObjectIdNames[2766] = 'TransitionVariableType_TransitionTime'
ObjectIdNames[2767] = 'FiniteTransitionVariableType'
ObjectIdNames[2768] = 'FiniteTransitionVariableType_Id'
ObjectIdNames[2769] = 'StateMachineType_CurrentState'
ObjectIdNames[2770] = 'StateMachineType_LastTransition'
ObjectIdNames[2771] = 'FiniteStateMachineType'
ObjectIdNames[2772] = 'FiniteStateMachineType_CurrentState'
ObjectIdNames[2773] = 'FiniteStateMachineType_LastTransition'
ObjectIdNames[2774] = 'TransitionEventType_Transition'
ObjectIdNames[2775] = 'TransitionEventType_FromState'
ObjectIdNames[2776] = 'TransitionEventType_ToState'
ObjectIdNames[2777] = 'AuditUpdateStateEventType_OldStateId'
ObjectIdNames[2778] = 'AuditUpdateStateEventType_NewStateId'
ObjectIdNames[2782] = 'ConditionType'
ObjectIdNames[2787] = 'RefreshStartEventType'
ObjectIdNames[2788] = 'RefreshEndEventType'
ObjectIdNames[2789] = 'RefreshRequiredEventType'
ObjectIdNames[2790] = 'AuditConditionEventType'
ObjectIdNames[2803] = 'AuditConditionEnableEventType'
ObjectIdNames[2829] = 'AuditConditionCommentEventType'
ObjectIdNames[2830] = 'DialogConditionType'
ObjectIdNames[2831] = 'DialogConditionType_Prompt'
ObjectIdNames[2881] = 'AcknowledgeableConditionType'
ObjectIdNames[2915] = 'AlarmConditionType'
ObjectIdNames[2929] = 'ShelvedStateMachineType'
ObjectIdNames[2930] = 'ShelvedStateMachineType_Unshelved'
ObjectIdNames[2932] = 'ShelvedStateMachineType_TimedShelved'
ObjectIdNames[2933] = 'ShelvedStateMachineType_OneShotShelved'
ObjectIdNames[2935] = 'ShelvedStateMachineType_UnshelvedToTimedShelved'
ObjectIdNames[2936] = 'ShelvedStateMachineType_UnshelvedToOneShotShelved'
ObjectIdNames[2940] = 'ShelvedStateMachineType_TimedShelvedToUnshelved'
ObjectIdNames[2942] = 'ShelvedStateMachineType_TimedShelvedToOneShotShelved'
ObjectIdNames[2943] = 'ShelvedStateMachineType_OneShotShelvedToUnshelved'
ObjectIdNames[2945] = 'ShelvedStateMachineType_OneShotShelvedToTimedShelved'
ObjectIdNames[2947] = 'ShelvedStateMachineType_Unshelve'
ObjectIdNames[2948] = 'ShelvedStateMachineType_OneShotShelve'
ObjectIdNames[2949] = 'ShelvedStateMachineType_TimedShelve'
ObjectIdNames[2955] = 'LimitAlarmType'
ObjectIdNames[2991] = 'ShelvedStateMachineType_TimedShelve_InputArguments'
ObjectIdNames[2992] = 'Server_ServerStatus_SecondsTillShutdown'
ObjectIdNames[2993] = 'Server_ServerStatus_ShutdownReason'
ObjectIdNames[2994] = 'Server_Auditing'
ObjectIdNames[2996] = 'Server_ServerCapabilities_ModellingRules'
ObjectIdNames[2997] = 'Server_ServerCapabilities_AggregateFunctions'
ObjectIdNames[2998] = 'SubscriptionDiagnosticsType_EventNotificationsCount'
ObjectIdNames[2999] = 'AuditHistoryEventUpdateEventType'
ObjectIdNames[3003] = 'AuditHistoryEventUpdateEventType_Filter'
ObjectIdNames[3006] = 'AuditHistoryValueUpdateEventType'
ObjectIdNames[3012] = 'AuditHistoryDeleteEventType'
ObjectIdNames[3014] = 'AuditHistoryRawModifyDeleteEventType'
ObjectIdNames[3015] = 'AuditHistoryRawModifyDeleteEventType_IsDeleteModified'
ObjectIdNames[3016] = 'AuditHistoryRawModifyDeleteEventType_StartTime'
ObjectIdNames[3017] = 'AuditHistoryRawModifyDeleteEventType_EndTime'
ObjectIdNames[3019] = 'AuditHistoryAtTimeDeleteEventType'
ObjectIdNames[3020] = 'AuditHistoryAtTimeDeleteEventType_ReqTimes'
ObjectIdNames[3021] = 'AuditHistoryAtTimeDeleteEventType_OldValues'
ObjectIdNames[3022] = 'AuditHistoryEventDeleteEventType'
ObjectIdNames[3023] = 'AuditHistoryEventDeleteEventType_EventIds'
ObjectIdNames[3024] = 'AuditHistoryEventDeleteEventType_OldValues'
ObjectIdNames[3025] = 'AuditHistoryEventUpdateEventType_UpdatedNode'
ObjectIdNames[3026] = 'AuditHistoryValueUpdateEventType_UpdatedNode'
ObjectIdNames[3027] = 'AuditHistoryDeleteEventType_UpdatedNode'
ObjectIdNames[3028] = 'AuditHistoryEventUpdateEventType_PerformInsertReplace'
ObjectIdNames[3029] = 'AuditHistoryEventUpdateEventType_NewValues'
ObjectIdNames[3030] = 'AuditHistoryEventUpdateEventType_OldValues'
ObjectIdNames[3031] = 'AuditHistoryValueUpdateEventType_PerformInsertReplace'
ObjectIdNames[3032] = 'AuditHistoryValueUpdateEventType_NewValues'
ObjectIdNames[3033] = 'AuditHistoryValueUpdateEventType_OldValues'
ObjectIdNames[3034] = 'AuditHistoryRawModifyDeleteEventType_OldValues'
ObjectIdNames[3035] = 'EventQueueOverflowEventType'
ObjectIdNames[3048] = 'EventTypesFolder'
ObjectIdNames[3049] = 'ServerCapabilitiesType_SoftwareCertificates'
ObjectIdNames[3050] = 'SessionDiagnosticsVariableType_MaxResponseMessageSize'
ObjectIdNames[3051] = 'BuildInfoType'
ObjectIdNames[3052] = 'BuildInfoType_ProductUri'
ObjectIdNames[3053] = 'BuildInfoType_ManufacturerName'
ObjectIdNames[3054] = 'BuildInfoType_ProductName'
ObjectIdNames[3055] = 'BuildInfoType_SoftwareVersion'
ObjectIdNames[3056] = 'BuildInfoType_BuildNumber'
ObjectIdNames[3057] = 'BuildInfoType_BuildDate'
ObjectIdNames[3058] = 'SessionSecurityDiagnosticsType_ClientCertificate'
ObjectIdNames[3059] = 'HistoricalDataConfigurationType_AggregateConfiguration'
ObjectIdNames[3062] = 'DefaultBinary'
ObjectIdNames[3063] = 'DefaultXml'
ObjectIdNames[3065] = 'AlwaysGeneratesEvent'
ObjectIdNames[3067] = 'Icon'
ObjectIdNames[3068] = 'NodeVersion'
ObjectIdNames[3069] = 'LocalTime'
ObjectIdNames[3070] = 'AllowNulls'
ObjectIdNames[3071] = 'EnumValues'
ObjectIdNames[3072] = 'InputArguments'
ObjectIdNames[3073] = 'OutputArguments'
ObjectIdNames[3074] = 'ServerType_ServerStatus_StartTime'
ObjectIdNames[3075] = 'ServerType_ServerStatus_CurrentTime'
ObjectIdNames[3076] = 'ServerType_ServerStatus_State'
ObjectIdNames[3077] = 'ServerType_ServerStatus_BuildInfo'
ObjectIdNames[3078] = 'ServerType_ServerStatus_BuildInfo_ProductUri'
ObjectIdNames[3079] = 'ServerType_ServerStatus_BuildInfo_ManufacturerName'
ObjectIdNames[3080] = 'ServerType_ServerStatus_BuildInfo_ProductName'
ObjectIdNames[3081] = 'ServerType_ServerStatus_BuildInfo_SoftwareVersion'
ObjectIdNames[3082] = 'ServerType_ServerStatus_BuildInfo_BuildNumber'
ObjectIdNames[3083] = 'ServerType_ServerStatus_BuildInfo_BuildDate'
ObjectIdNames[3084] = 'ServerType_ServerStatus_SecondsTillShutdown'
ObjectIdNames[3085] = 'ServerType_ServerStatus_ShutdownReason'
ObjectIdNames[3086] = 'ServerType_ServerCapabilities_ServerProfileArray'
ObjectIdNames[3087] = 'ServerType_ServerCapabilities_LocaleIdArray'
ObjectIdNames[3088] = 'ServerType_ServerCapabilities_MinSupportedSampleRate'
ObjectIdNames[3089] = 'ServerType_ServerCapabilities_MaxBrowseContinuationPoints'
ObjectIdNames[3090] = 'ServerType_ServerCapabilities_MaxQueryContinuationPoints'
ObjectIdNames[3091] = 'ServerType_ServerCapabilities_MaxHistoryContinuationPoints'
ObjectIdNames[3092] = 'ServerType_ServerCapabilities_SoftwareCertificates'
ObjectIdNames[3093] = 'ServerType_ServerCapabilities_ModellingRules'
ObjectIdNames[3094] = 'ServerType_ServerCapabilities_AggregateFunctions'
ObjectIdNames[3095] = 'ServerType_ServerDiagnostics_ServerDiagnosticsSummary'
ObjectIdNames[3096] = 'ServerType_ServerDiagnostics_ServerDiagnosticsSummary_ServerViewCount'
ObjectIdNames[3097] = 'ServerType_ServerDiagnostics_ServerDiagnosticsSummary_CurrentSessionCount'
ObjectIdNames[3098] = 'ServerType_ServerDiagnostics_ServerDiagnosticsSummary_CumulatedSessionCount'
ObjectIdNames[3099] = 'ServerType_ServerDiagnostics_ServerDiagnosticsSummary_SecurityRejectedSessionCount'
ObjectIdNames[3100] = 'ServerType_ServerDiagnostics_ServerDiagnosticsSummary_RejectedSessionCount'
ObjectIdNames[3101] = 'ServerType_ServerDiagnostics_ServerDiagnosticsSummary_SessionTimeoutCount'
ObjectIdNames[3102] = 'ServerType_ServerDiagnostics_ServerDiagnosticsSummary_SessionAbortCount'
ObjectIdNames[3104] = 'ServerType_ServerDiagnostics_ServerDiagnosticsSummary_PublishingIntervalCount'
ObjectIdNames[3105] = 'ServerType_ServerDiagnostics_ServerDiagnosticsSummary_CurrentSubscriptionCount'
ObjectIdNames[3106] = 'ServerType_ServerDiagnostics_ServerDiagnosticsSummary_CumulatedSubscriptionCount'
ObjectIdNames[3107] = 'ServerType_ServerDiagnostics_ServerDiagnosticsSummary_SecurityRejectedRequestsCount'
ObjectIdNames[3108] = 'ServerType_ServerDiagnostics_ServerDiagnosticsSummary_RejectedRequestsCount'
ObjectIdNames[3109] = 'ServerType_ServerDiagnostics_SamplingIntervalDiagnosticsArray'
ObjectIdNames[3110] = 'ServerType_ServerDiagnostics_SubscriptionDiagnosticsArray'
ObjectIdNames[3111] = 'ServerType_ServerDiagnostics_SessionsDiagnosticsSummary'
ObjectIdNames[3112] = 'ServerType_ServerDiagnostics_SessionsDiagnosticsSummary_SessionDiagnosticsArray'
ObjectIdNames[3113] = 'ServerType_ServerDiagnostics_SessionsDiagnosticsSummary_SessionSecurityDiagnosticsArray'
ObjectIdNames[3114] = 'ServerType_ServerDiagnostics_EnabledFlag'
ObjectIdNames[3115] = 'ServerType_ServerRedundancy_RedundancySupport'
ObjectIdNames[3116] = 'ServerDiagnosticsType_ServerDiagnosticsSummary_ServerViewCount'
ObjectIdNames[3117] = 'ServerDiagnosticsType_ServerDiagnosticsSummary_CurrentSessionCount'
ObjectIdNames[3118] = 'ServerDiagnosticsType_ServerDiagnosticsSummary_CumulatedSessionCount'
ObjectIdNames[3119] = 'ServerDiagnosticsType_ServerDiagnosticsSummary_SecurityRejectedSessionCount'
ObjectIdNames[3120] = 'ServerDiagnosticsType_ServerDiagnosticsSummary_RejectedSessionCount'
ObjectIdNames[3121] = 'ServerDiagnosticsType_ServerDiagnosticsSummary_SessionTimeoutCount'
ObjectIdNames[3122] = 'ServerDiagnosticsType_ServerDiagnosticsSummary_SessionAbortCount'
ObjectIdNames[3124] = 'ServerDiagnosticsType_ServerDiagnosticsSummary_PublishingIntervalCount'
ObjectIdNames[3125] = 'ServerDiagnosticsType_ServerDiagnosticsSummary_CurrentSubscriptionCount'
ObjectIdNames[3126] = 'ServerDiagnosticsType_ServerDiagnosticsSummary_CumulatedSubscriptionCount'
ObjectIdNames[3127] = 'ServerDiagnosticsType_ServerDiagnosticsSummary_SecurityRejectedRequestsCount'
ObjectIdNames[3128] = 'ServerDiagnosticsType_ServerDiagnosticsSummary_RejectedRequestsCount'
ObjectIdNames[3129] = 'ServerDiagnosticsType_SessionsDiagnosticsSummary_SessionDiagnosticsArray'
ObjectIdNames[3130] = 'ServerDiagnosticsType_SessionsDiagnosticsSummary_SessionSecurityDiagnosticsArray'
ObjectIdNames[3131] = 'SessionDiagnosticsObjectType_SessionDiagnostics_SessionId'
ObjectIdNames[3132] = 'SessionDiagnosticsObjectType_SessionDiagnostics_SessionName'
ObjectIdNames[3133] = 'SessionDiagnosticsObjectType_SessionDiagnostics_ClientDescription'
ObjectIdNames[3134] = 'SessionDiagnosticsObjectType_SessionDiagnostics_ServerUri'
ObjectIdNames[3135] = 'SessionDiagnosticsObjectType_SessionDiagnostics_EndpointUrl'
ObjectIdNames[3136] = 'SessionDiagnosticsObjectType_SessionDiagnostics_LocaleIds'
ObjectIdNames[3137] = 'SessionDiagnosticsObjectType_SessionDiagnostics_ActualSessionTimeout'
ObjectIdNames[3138] = 'SessionDiagnosticsObjectType_SessionDiagnostics_MaxResponseMessageSize'
ObjectIdNames[3139] = 'SessionDiagnosticsObjectType_SessionDiagnostics_ClientConnectionTime'
ObjectIdNames[3140] = 'SessionDiagnosticsObjectType_SessionDiagnostics_ClientLastContactTime'
ObjectIdNames[3141] = 'SessionDiagnosticsObjectType_SessionDiagnostics_CurrentSubscriptionsCount'
ObjectIdNames[3142] = 'SessionDiagnosticsObjectType_SessionDiagnostics_CurrentMonitoredItemsCount'
ObjectIdNames[3143] = 'SessionDiagnosticsObjectType_SessionDiagnostics_CurrentPublishRequestsInQueue'
ObjectIdNames[3151] = 'SessionDiagnosticsObjectType_SessionDiagnostics_ReadCount'
ObjectIdNames[3152] = 'SessionDiagnosticsObjectType_SessionDiagnostics_HistoryReadCount'
ObjectIdNames[3153] = 'SessionDiagnosticsObjectType_SessionDiagnostics_WriteCount'
ObjectIdNames[3154] = 'SessionDiagnosticsObjectType_SessionDiagnostics_HistoryUpdateCount'
ObjectIdNames[3155] = 'SessionDiagnosticsObjectType_SessionDiagnostics_CallCount'
ObjectIdNames[3156] = 'SessionDiagnosticsObjectType_SessionDiagnostics_CreateMonitoredItemsCount'
ObjectIdNames[3157] = 'SessionDiagnosticsObjectType_SessionDiagnostics_ModifyMonitoredItemsCount'
ObjectIdNames[3158] = 'SessionDiagnosticsObjectType_SessionDiagnostics_SetMonitoringModeCount'
ObjectIdNames[3159] = 'SessionDiagnosticsObjectType_SessionDiagnostics_SetTriggeringCount'
ObjectIdNames[3160] = 'SessionDiagnosticsObjectType_SessionDiagnostics_DeleteMonitoredItemsCount'
ObjectIdNames[3161] = 'SessionDiagnosticsObjectType_SessionDiagnostics_CreateSubscriptionCount'
ObjectIdNames[3162] = 'SessionDiagnosticsObjectType_SessionDiagnostics_ModifySubscriptionCount'
ObjectIdNames[3163] = 'SessionDiagnosticsObjectType_SessionDiagnostics_SetPublishingModeCount'
ObjectIdNames[3164] = 'SessionDiagnosticsObjectType_SessionDiagnostics_PublishCount'
ObjectIdNames[3165] = 'SessionDiagnosticsObjectType_SessionDiagnostics_RepublishCount'
ObjectIdNames[3166] = 'SessionDiagnosticsObjectType_SessionDiagnostics_TransferSubscriptionsCount'
ObjectIdNames[3167] = 'SessionDiagnosticsObjectType_SessionDiagnostics_DeleteSubscriptionsCount'
ObjectIdNames[3168] = 'SessionDiagnosticsObjectType_SessionDiagnostics_AddNodesCount'
ObjectIdNames[3169] = 'SessionDiagnosticsObjectType_SessionDiagnostics_AddReferencesCount'
ObjectIdNames[3170] = 'SessionDiagnosticsObjectType_SessionDiagnostics_DeleteNodesCount'
ObjectIdNames[3171] = 'SessionDiagnosticsObjectType_SessionDiagnostics_DeleteReferencesCount'
ObjectIdNames[3172] = 'SessionDiagnosticsObjectType_SessionDiagnostics_BrowseCount'
ObjectIdNames[3173] = 'SessionDiagnosticsObjectType_SessionDiagnostics_BrowseNextCount'
ObjectIdNames[3174] = 'SessionDiagnosticsObjectType_SessionDiagnostics_TranslateBrowsePathsToNodeIdsCount'
ObjectIdNames[3175] = 'SessionDiagnosticsObjectType_SessionDiagnostics_QueryFirstCount'
ObjectIdNames[3176] = 'SessionDiagnosticsObjectType_SessionDiagnostics_QueryNextCount'
ObjectIdNames[3177] = 'SessionDiagnosticsObjectType_SessionDiagnostics_RegisterNodesCount'
ObjectIdNames[3178] = 'SessionDiagnosticsObjectType_SessionDiagnostics_UnregisterNodesCount'
ObjectIdNames[3179] = 'SessionDiagnosticsObjectType_SessionSecurityDiagnostics_SessionId'
ObjectIdNames[3180] = 'SessionDiagnosticsObjectType_SessionSecurityDiagnostics_ClientUserIdOfSession'
ObjectIdNames[3181] = 'SessionDiagnosticsObjectType_SessionSecurityDiagnostics_ClientUserIdHistory'
ObjectIdNames[3182] = 'SessionDiagnosticsObjectType_SessionSecurityDiagnostics_AuthenticationMechanism'
ObjectIdNames[3183] = 'SessionDiagnosticsObjectType_SessionSecurityDiagnostics_Encoding'
ObjectIdNames[3184] = 'SessionDiagnosticsObjectType_SessionSecurityDiagnostics_TransportProtocol'
ObjectIdNames[3185] = 'SessionDiagnosticsObjectType_SessionSecurityDiagnostics_SecurityMode'
ObjectIdNames[3186] = 'SessionDiagnosticsObjectType_SessionSecurityDiagnostics_SecurityPolicyUri'
ObjectIdNames[3187] = 'SessionDiagnosticsObjectType_SessionSecurityDiagnostics_ClientCertificate'
ObjectIdNames[3188] = 'TransparentRedundancyType_RedundancySupport'
ObjectIdNames[3189] = 'NonTransparentRedundancyType_RedundancySupport'
ObjectIdNames[3190] = 'BaseEventType_LocalTime'
ObjectIdNames[3191] = 'EventQueueOverflowEventType_EventId'
ObjectIdNames[3192] = 'EventQueueOverflowEventType_EventType'
ObjectIdNames[3193] = 'EventQueueOverflowEventType_SourceNode'
ObjectIdNames[3194] = 'EventQueueOverflowEventType_SourceName'
ObjectIdNames[3195] = 'EventQueueOverflowEventType_Time'
ObjectIdNames[3196] = 'EventQueueOverflowEventType_ReceiveTime'
ObjectIdNames[3197] = 'EventQueueOverflowEventType_LocalTime'
ObjectIdNames[3198] = 'EventQueueOverflowEventType_Message'
ObjectIdNames[3199] = 'EventQueueOverflowEventType_Severity'
ObjectIdNames[3200] = 'AuditEventType_EventId'
ObjectIdNames[3201] = 'AuditEventType_EventType'
ObjectIdNames[3202] = 'AuditEventType_SourceNode'
ObjectIdNames[3203] = 'AuditEventType_SourceName'
ObjectIdNames[3204] = 'AuditEventType_Time'
ObjectIdNames[3205] = 'AuditEventType_ReceiveTime'
ObjectIdNames[3206] = 'AuditEventType_LocalTime'
ObjectIdNames[3207] = 'AuditEventType_Message'
ObjectIdNames[3208] = 'AuditEventType_Severity'
ObjectIdNames[3209] = 'AuditSecurityEventType_EventId'
ObjectIdNames[3210] = 'AuditSecurityEventType_EventType'
ObjectIdNames[3211] = 'AuditSecurityEventType_SourceNode'
ObjectIdNames[3212] = 'AuditSecurityEventType_SourceName'
ObjectIdNames[3213] = 'AuditSecurityEventType_Time'
ObjectIdNames[3214] = 'AuditSecurityEventType_ReceiveTime'
ObjectIdNames[3215] = 'AuditSecurityEventType_LocalTime'
ObjectIdNames[3216] = 'AuditSecurityEventType_Message'
ObjectIdNames[3217] = 'AuditSecurityEventType_Severity'
ObjectIdNames[3218] = 'AuditSecurityEventType_ActionTimeStamp'
ObjectIdNames[3219] = 'AuditSecurityEventType_Status'
ObjectIdNames[3220] = 'AuditSecurityEventType_ServerId'
ObjectIdNames[3221] = 'AuditSecurityEventType_ClientAuditEntryId'
ObjectIdNames[3222] = 'AuditSecurityEventType_ClientUserId'
ObjectIdNames[3223] = 'AuditChannelEventType_EventId'
ObjectIdNames[3224] = 'AuditChannelEventType_EventType'
ObjectIdNames[3225] = 'AuditChannelEventType_SourceNode'
ObjectIdNames[3226] = 'AuditChannelEventType_SourceName'
ObjectIdNames[3227] = 'AuditChannelEventType_Time'
ObjectIdNames[3228] = 'AuditChannelEventType_ReceiveTime'
ObjectIdNames[3229] = 'AuditChannelEventType_LocalTime'
ObjectIdNames[3230] = 'AuditChannelEventType_Message'
ObjectIdNames[3231] = 'AuditChannelEventType_Severity'
ObjectIdNames[3232] = 'AuditChannelEventType_ActionTimeStamp'
ObjectIdNames[3233] = 'AuditChannelEventType_Status'
ObjectIdNames[3234] = 'AuditChannelEventType_ServerId'
ObjectIdNames[3235] = 'AuditChannelEventType_ClientAuditEntryId'
ObjectIdNames[3236] = 'AuditChannelEventType_ClientUserId'
ObjectIdNames[3237] = 'AuditOpenSecureChannelEventType_EventId'
ObjectIdNames[3238] = 'AuditOpenSecureChannelEventType_EventType'
ObjectIdNames[3239] = 'AuditOpenSecureChannelEventType_SourceNode'
ObjectIdNames[3240] = 'AuditOpenSecureChannelEventType_SourceName'
ObjectIdNames[3241] = 'AuditOpenSecureChannelEventType_Time'
ObjectIdNames[3242] = 'AuditOpenSecureChannelEventType_ReceiveTime'
ObjectIdNames[3243] = 'AuditOpenSecureChannelEventType_LocalTime'
ObjectIdNames[3244] = 'AuditOpenSecureChannelEventType_Message'
ObjectIdNames[3245] = 'AuditOpenSecureChannelEventType_Severity'
ObjectIdNames[3246] = 'AuditOpenSecureChannelEventType_ActionTimeStamp'
ObjectIdNames[3247] = 'AuditOpenSecureChannelEventType_Status'
ObjectIdNames[3248] = 'AuditOpenSecureChannelEventType_ServerId'
ObjectIdNames[3249] = 'AuditOpenSecureChannelEventType_ClientAuditEntryId'
ObjectIdNames[3250] = 'AuditOpenSecureChannelEventType_ClientUserId'
ObjectIdNames[3251] = 'AuditOpenSecureChannelEventType_SecureChannelId'
ObjectIdNames[3252] = 'AuditSessionEventType_EventId'
ObjectIdNames[3253] = 'AuditSessionEventType_EventType'
ObjectIdNames[3254] = 'AuditSessionEventType_SourceNode'
ObjectIdNames[3255] = 'AuditSessionEventType_SourceName'
ObjectIdNames[3256] = 'AuditSessionEventType_Time'
ObjectIdNames[3257] = 'AuditSessionEventType_ReceiveTime'
ObjectIdNames[3258] = 'AuditSessionEventType_LocalTime'
ObjectIdNames[3259] = 'AuditSessionEventType_Message'
ObjectIdNames[3260] = 'AuditSessionEventType_Severity'
ObjectIdNames[3261] = 'AuditSessionEventType_ActionTimeStamp'
ObjectIdNames[3262] = 'AuditSessionEventType_Status'
ObjectIdNames[3263] = 'AuditSessionEventType_ServerId'
ObjectIdNames[3264] = 'AuditSessionEventType_ClientAuditEntryId'
ObjectIdNames[3265] = 'AuditSessionEventType_ClientUserId'
ObjectIdNames[3266] = 'AuditCreateSessionEventType_EventId'
ObjectIdNames[3267] = 'AuditCreateSessionEventType_EventType'
ObjectIdNames[3268] = 'AuditCreateSessionEventType_SourceNode'
ObjectIdNames[3269] = 'AuditCreateSessionEventType_SourceName'
ObjectIdNames[3270] = 'AuditCreateSessionEventType_Time'
ObjectIdNames[3271] = 'AuditCreateSessionEventType_ReceiveTime'
ObjectIdNames[3272] = 'AuditCreateSessionEventType_LocalTime'
ObjectIdNames[3273] = 'AuditCreateSessionEventType_Message'
ObjectIdNames[3274] = 'AuditCreateSessionEventType_Severity'
ObjectIdNames[3275] = 'AuditCreateSessionEventType_ActionTimeStamp'
ObjectIdNames[3276] = 'AuditCreateSessionEventType_Status'
ObjectIdNames[3277] = 'AuditCreateSessionEventType_ServerId'
ObjectIdNames[3278] = 'AuditCreateSessionEventType_ClientAuditEntryId'
ObjectIdNames[3279] = 'AuditCreateSessionEventType_ClientUserId'
ObjectIdNames[3281] = 'AuditUrlMismatchEventType_EventId'
ObjectIdNames[3282] = 'AuditUrlMismatchEventType_EventType'
ObjectIdNames[3283] = 'AuditUrlMismatchEventType_SourceNode'
ObjectIdNames[3284] = 'AuditUrlMismatchEventType_SourceName'
ObjectIdNames[3285] = 'AuditUrlMismatchEventType_Time'
ObjectIdNames[3286] = 'AuditUrlMismatchEventType_ReceiveTime'
ObjectIdNames[3287] = 'AuditUrlMismatchEventType_LocalTime'
ObjectIdNames[3288] = 'AuditUrlMismatchEventType_Message'
ObjectIdNames[3289] = 'AuditUrlMismatchEventType_Severity'
ObjectIdNames[3290] = 'AuditUrlMismatchEventType_ActionTimeStamp'
ObjectIdNames[3291] = 'AuditUrlMismatchEventType_Status'
ObjectIdNames[3292] = 'AuditUrlMismatchEventType_ServerId'
ObjectIdNames[3293] = 'AuditUrlMismatchEventType_ClientAuditEntryId'
ObjectIdNames[3294] = 'AuditUrlMismatchEventType_ClientUserId'
ObjectIdNames[3296] = 'AuditUrlMismatchEventType_SecureChannelId'
ObjectIdNames[3297] = 'AuditUrlMismatchEventType_ClientCertificate'
ObjectIdNames[3298] = 'AuditUrlMismatchEventType_ClientCertificateThumbprint'
ObjectIdNames[3299] = 'AuditUrlMismatchEventType_RevisedSessionTimeout'
ObjectIdNames[3300] = 'AuditActivateSessionEventType_EventId'
ObjectIdNames[3301] = 'AuditActivateSessionEventType_EventType'
ObjectIdNames[3302] = 'AuditActivateSessionEventType_SourceNode'
ObjectIdNames[3303] = 'AuditActivateSessionEventType_SourceName'
ObjectIdNames[3304] = 'AuditActivateSessionEventType_Time'
ObjectIdNames[3305] = 'AuditActivateSessionEventType_ReceiveTime'
ObjectIdNames[3306] = 'AuditActivateSessionEventType_LocalTime'
ObjectIdNames[3307] = 'AuditActivateSessionEventType_Message'
ObjectIdNames[3308] = 'AuditActivateSessionEventType_Severity'
ObjectIdNames[3309] = 'AuditActivateSessionEventType_ActionTimeStamp'
ObjectIdNames[3310] = 'AuditActivateSessionEventType_Status'
ObjectIdNames[3311] = 'AuditActivateSessionEventType_ServerId'
ObjectIdNames[3312] = 'AuditActivateSessionEventType_ClientAuditEntryId'
ObjectIdNames[3313] = 'AuditActivateSessionEventType_ClientUserId'
ObjectIdNames[3314] = 'AuditActivateSessionEventType_SessionId'
ObjectIdNames[3315] = 'AuditCancelEventType_EventId'
ObjectIdNames[3316] = 'AuditCancelEventType_EventType'
ObjectIdNames[3317] = 'AuditCancelEventType_SourceNode'
ObjectIdNames[3318] = 'AuditCancelEventType_SourceName'
ObjectIdNames[3319] = 'AuditCancelEventType_Time'
ObjectIdNames[3320] = 'AuditCancelEventType_ReceiveTime'
ObjectIdNames[3321] = 'AuditCancelEventType_LocalTime'
ObjectIdNames[3322] = 'AuditCancelEventType_Message'
ObjectIdNames[3323] = 'AuditCancelEventType_Severity'
ObjectIdNames[3324] = 'AuditCancelEventType_ActionTimeStamp'
ObjectIdNames[3325] = 'AuditCancelEventType_Status'
ObjectIdNames[3326] = 'AuditCancelEventType_ServerId'
ObjectIdNames[3327] = 'AuditCancelEventType_ClientAuditEntryId'
ObjectIdNames[3328] = 'AuditCancelEventType_ClientUserId'
ObjectIdNames[3329] = 'AuditCancelEventType_SessionId'
ObjectIdNames[3330] = 'AuditCertificateEventType_EventId'
ObjectIdNames[3331] = 'AuditCertificateEventType_EventType'
ObjectIdNames[3332] = 'AuditCertificateEventType_SourceNode'
ObjectIdNames[3333] = 'AuditCertificateEventType_SourceName'
ObjectIdNames[3334] = 'AuditCertificateEventType_Time'
ObjectIdNames[3335] = 'AuditCertificateEventType_ReceiveTime'
ObjectIdNames[3336] = 'AuditCertificateEventType_LocalTime'
ObjectIdNames[3337] = 'AuditCertificateEventType_Message'
ObjectIdNames[3338] = 'AuditCertificateEventType_Severity'
ObjectIdNames[3339] = 'AuditCertificateEventType_ActionTimeStamp'
ObjectIdNames[3340] = 'AuditCertificateEventType_Status'
ObjectIdNames[3341] = 'AuditCertificateEventType_ServerId'
ObjectIdNames[3342] = 'AuditCertificateEventType_ClientAuditEntryId'
ObjectIdNames[3343] = 'AuditCertificateEventType_ClientUserId'
ObjectIdNames[3344] = 'AuditCertificateDataMismatchEventType_EventId'
ObjectIdNames[3345] = 'AuditCertificateDataMismatchEventType_EventType'
ObjectIdNames[3346] = 'AuditCertificateDataMismatchEventType_SourceNode'
ObjectIdNames[3347] = 'AuditCertificateDataMismatchEventType_SourceName'
ObjectIdNames[3348] = 'AuditCertificateDataMismatchEventType_Time'
ObjectIdNames[3349] = 'AuditCertificateDataMismatchEventType_ReceiveTime'
ObjectIdNames[3350] = 'AuditCertificateDataMismatchEventType_LocalTime'
ObjectIdNames[3351] = 'AuditCertificateDataMismatchEventType_Message'
ObjectIdNames[3352] = 'AuditCertificateDataMismatchEventType_Severity'
ObjectIdNames[3353] = 'AuditCertificateDataMismatchEventType_ActionTimeStamp'
ObjectIdNames[3354] = 'AuditCertificateDataMismatchEventType_Status'
ObjectIdNames[3355] = 'AuditCertificateDataMismatchEventType_ServerId'
ObjectIdNames[3356] = 'AuditCertificateDataMismatchEventType_ClientAuditEntryId'
ObjectIdNames[3357] = 'AuditCertificateDataMismatchEventType_ClientUserId'
ObjectIdNames[3358] = 'AuditCertificateDataMismatchEventType_Certificate'
ObjectIdNames[3359] = 'AuditCertificateExpiredEventType_EventId'
ObjectIdNames[3360] = 'AuditCertificateExpiredEventType_EventType'
ObjectIdNames[3361] = 'AuditCertificateExpiredEventType_SourceNode'
ObjectIdNames[3362] = 'AuditCertificateExpiredEventType_SourceName'
ObjectIdNames[3363] = 'AuditCertificateExpiredEventType_Time'
ObjectIdNames[3364] = 'AuditCertificateExpiredEventType_ReceiveTime'
ObjectIdNames[3365] = 'AuditCertificateExpiredEventType_LocalTime'
ObjectIdNames[3366] = 'AuditCertificateExpiredEventType_Message'
ObjectIdNames[3367] = 'AuditCertificateExpiredEventType_Severity'
ObjectIdNames[3368] = 'AuditCertificateExpiredEventType_ActionTimeStamp'
ObjectIdNames[3369] = 'AuditCertificateExpiredEventType_Status'
ObjectIdNames[3370] = 'AuditCertificateExpiredEventType_ServerId'
ObjectIdNames[3371] = 'AuditCertificateExpiredEventType_ClientAuditEntryId'
ObjectIdNames[3372] = 'AuditCertificateExpiredEventType_ClientUserId'
ObjectIdNames[3373] = 'AuditCertificateExpiredEventType_Certificate'
ObjectIdNames[3374] = 'AuditCertificateInvalidEventType_EventId'
ObjectIdNames[3375] = 'AuditCertificateInvalidEventType_EventType'
ObjectIdNames[3376] = 'AuditCertificateInvalidEventType_SourceNode'
ObjectIdNames[3377] = 'AuditCertificateInvalidEventType_SourceName'
ObjectIdNames[3378] = 'AuditCertificateInvalidEventType_Time'
ObjectIdNames[3379] = 'AuditCertificateInvalidEventType_ReceiveTime'
ObjectIdNames[3380] = 'AuditCertificateInvalidEventType_LocalTime'
ObjectIdNames[3381] = 'AuditCertificateInvalidEventType_Message'
ObjectIdNames[3382] = 'AuditCertificateInvalidEventType_Severity'
ObjectIdNames[3383] = 'AuditCertificateInvalidEventType_ActionTimeStamp'
ObjectIdNames[3384] = 'AuditCertificateInvalidEventType_Status'
ObjectIdNames[3385] = 'AuditCertificateInvalidEventType_ServerId'
ObjectIdNames[3386] = 'AuditCertificateInvalidEventType_ClientAuditEntryId'
ObjectIdNames[3387] = 'AuditCertificateInvalidEventType_ClientUserId'
ObjectIdNames[3388] = 'AuditCertificateInvalidEventType_Certificate'
ObjectIdNames[3389] = 'AuditCertificateUntrustedEventType_EventId'
ObjectIdNames[3390] = 'AuditCertificateUntrustedEventType_EventType'
ObjectIdNames[3391] = 'AuditCertificateUntrustedEventType_SourceNode'
ObjectIdNames[3392] = 'AuditCertificateUntrustedEventType_SourceName'
ObjectIdNames[3393] = 'AuditCertificateUntrustedEventType_Time'
ObjectIdNames[3394] = 'AuditCertificateUntrustedEventType_ReceiveTime'
ObjectIdNames[3395] = 'AuditCertificateUntrustedEventType_LocalTime'
ObjectIdNames[3396] = 'AuditCertificateUntrustedEventType_Message'
ObjectIdNames[3397] = 'AuditCertificateUntrustedEventType_Severity'
ObjectIdNames[3398] = 'AuditCertificateUntrustedEventType_ActionTimeStamp'
ObjectIdNames[3399] = 'AuditCertificateUntrustedEventType_Status'
ObjectIdNames[3400] = 'AuditCertificateUntrustedEventType_ServerId'
ObjectIdNames[3401] = 'AuditCertificateUntrustedEventType_ClientAuditEntryId'
ObjectIdNames[3402] = 'AuditCertificateUntrustedEventType_ClientUserId'
ObjectIdNames[3403] = 'AuditCertificateUntrustedEventType_Certificate'
ObjectIdNames[3404] = 'AuditCertificateRevokedEventType_EventId'
ObjectIdNames[3405] = 'AuditCertificateRevokedEventType_EventType'
ObjectIdNames[3406] = 'AuditCertificateRevokedEventType_SourceNode'
ObjectIdNames[3407] = 'AuditCertificateRevokedEventType_SourceName'
ObjectIdNames[3408] = 'AuditCertificateRevokedEventType_Time'
ObjectIdNames[3409] = 'AuditCertificateRevokedEventType_ReceiveTime'
ObjectIdNames[3410] = 'AuditCertificateRevokedEventType_LocalTime'
ObjectIdNames[3411] = 'AuditCertificateRevokedEventType_Message'
ObjectIdNames[3412] = 'AuditCertificateRevokedEventType_Severity'
ObjectIdNames[3413] = 'AuditCertificateRevokedEventType_ActionTimeStamp'
ObjectIdNames[3414] = 'AuditCertificateRevokedEventType_Status'
ObjectIdNames[3415] = 'AuditCertificateRevokedEventType_ServerId'
ObjectIdNames[3416] = 'AuditCertificateRevokedEventType_ClientAuditEntryId'
ObjectIdNames[3417] = 'AuditCertificateRevokedEventType_ClientUserId'
ObjectIdNames[3418] = 'AuditCertificateRevokedEventType_Certificate'
ObjectIdNames[3419] = 'AuditCertificateMismatchEventType_EventId'
ObjectIdNames[3420] = 'AuditCertificateMismatchEventType_EventType'
ObjectIdNames[3421] = 'AuditCertificateMismatchEventType_SourceNode'
ObjectIdNames[3422] = 'AuditCertificateMismatchEventType_SourceName'
ObjectIdNames[3423] = 'AuditCertificateMismatchEventType_Time'
ObjectIdNames[3424] = 'AuditCertificateMismatchEventType_ReceiveTime'
ObjectIdNames[3425] = 'AuditCertificateMismatchEventType_LocalTime'
ObjectIdNames[3426] = 'AuditCertificateMismatchEventType_Message'
ObjectIdNames[3427] = 'AuditCertificateMismatchEventType_Severity'
ObjectIdNames[3428] = 'AuditCertificateMismatchEventType_ActionTimeStamp'
ObjectIdNames[3429] = 'AuditCertificateMismatchEventType_Status'
ObjectIdNames[3430] = 'AuditCertificateMismatchEventType_ServerId'
ObjectIdNames[3431] = 'AuditCertificateMismatchEventType_ClientAuditEntryId'
ObjectIdNames[3432] = 'AuditCertificateMismatchEventType_ClientUserId'
ObjectIdNames[3433] = 'AuditCertificateMismatchEventType_Certificate'
ObjectIdNames[3434] = 'AuditNodeManagementEventType_EventId'
ObjectIdNames[3435] = 'AuditNodeManagementEventType_EventType'
ObjectIdNames[3436] = 'AuditNodeManagementEventType_SourceNode'
ObjectIdNames[3437] = 'AuditNodeManagementEventType_SourceName'
ObjectIdNames[3438] = 'AuditNodeManagementEventType_Time'
ObjectIdNames[3439] = 'AuditNodeManagementEventType_ReceiveTime'
ObjectIdNames[3440] = 'AuditNodeManagementEventType_LocalTime'
ObjectIdNames[3441] = 'AuditNodeManagementEventType_Message'
ObjectIdNames[3442] = 'AuditNodeManagementEventType_Severity'
ObjectIdNames[3443] = 'AuditNodeManagementEventType_ActionTimeStamp'
ObjectIdNames[3444] = 'AuditNodeManagementEventType_Status'
ObjectIdNames[3445] = 'AuditNodeManagementEventType_ServerId'
ObjectIdNames[3446] = 'AuditNodeManagementEventType_ClientAuditEntryId'
ObjectIdNames[3447] = 'AuditNodeManagementEventType_ClientUserId'
ObjectIdNames[3448] = 'AuditAddNodesEventType_EventId'
ObjectIdNames[3449] = 'AuditAddNodesEventType_EventType'
ObjectIdNames[3450] = 'AuditAddNodesEventType_SourceNode'
ObjectIdNames[3451] = 'AuditAddNodesEventType_SourceName'
ObjectIdNames[3452] = 'AuditAddNodesEventType_Time'
ObjectIdNames[3453] = 'AuditAddNodesEventType_ReceiveTime'
ObjectIdNames[3454] = 'AuditAddNodesEventType_LocalTime'
ObjectIdNames[3455] = 'AuditAddNodesEventType_Message'
ObjectIdNames[3456] = 'AuditAddNodesEventType_Severity'
ObjectIdNames[3457] = 'AuditAddNodesEventType_ActionTimeStamp'
ObjectIdNames[3458] = 'AuditAddNodesEventType_Status'
ObjectIdNames[3459] = 'AuditAddNodesEventType_ServerId'
ObjectIdNames[3460] = 'AuditAddNodesEventType_ClientAuditEntryId'
ObjectIdNames[3461] = 'AuditAddNodesEventType_ClientUserId'
ObjectIdNames[3462] = 'AuditDeleteNodesEventType_EventId'
ObjectIdNames[3463] = 'AuditDeleteNodesEventType_EventType'
ObjectIdNames[3464] = 'AuditDeleteNodesEventType_SourceNode'
ObjectIdNames[3465] = 'AuditDeleteNodesEventType_SourceName'
ObjectIdNames[3466] = 'AuditDeleteNodesEventType_Time'
ObjectIdNames[3467] = 'AuditDeleteNodesEventType_ReceiveTime'
ObjectIdNames[3468] = 'AuditDeleteNodesEventType_LocalTime'
ObjectIdNames[3469] = 'AuditDeleteNodesEventType_Message'
ObjectIdNames[3470] = 'AuditDeleteNodesEventType_Severity'
ObjectIdNames[3471] = 'AuditDeleteNodesEventType_ActionTimeStamp'
ObjectIdNames[3472] = 'AuditDeleteNodesEventType_Status'
ObjectIdNames[3473] = 'AuditDeleteNodesEventType_ServerId'
ObjectIdNames[3474] = 'AuditDeleteNodesEventType_ClientAuditEntryId'
ObjectIdNames[3475] = 'AuditDeleteNodesEventType_ClientUserId'
ObjectIdNames[3476] = 'AuditAddReferencesEventType_EventId'
ObjectIdNames[3477] = 'AuditAddReferencesEventType_EventType'
ObjectIdNames[3478] = 'AuditAddReferencesEventType_SourceNode'
ObjectIdNames[3479] = 'AuditAddReferencesEventType_SourceName'
ObjectIdNames[3480] = 'AuditAddReferencesEventType_Time'
ObjectIdNames[3481] = 'AuditAddReferencesEventType_ReceiveTime'
ObjectIdNames[3482] = 'AuditAddReferencesEventType_LocalTime'
ObjectIdNames[3483] = 'AuditAddReferencesEventType_Message'
ObjectIdNames[3484] = 'AuditAddReferencesEventType_Severity'
ObjectIdNames[3485] = 'AuditAddReferencesEventType_ActionTimeStamp'
ObjectIdNames[3486] = 'AuditAddReferencesEventType_Status'
ObjectIdNames[3487] = 'AuditAddReferencesEventType_ServerId'
ObjectIdNames[3488] = 'AuditAddReferencesEventType_ClientAuditEntryId'
ObjectIdNames[3489] = 'AuditAddReferencesEventType_ClientUserId'
ObjectIdNames[3490] = 'AuditDeleteReferencesEventType_EventId'
ObjectIdNames[3491] = 'AuditDeleteReferencesEventType_EventType'
ObjectIdNames[3492] = 'AuditDeleteReferencesEventType_SourceNode'
ObjectIdNames[3493] = 'AuditDeleteReferencesEventType_SourceName'
ObjectIdNames[3494] = 'AuditDeleteReferencesEventType_Time'
ObjectIdNames[3495] = 'AuditDeleteReferencesEventType_ReceiveTime'
ObjectIdNames[3496] = 'AuditDeleteReferencesEventType_LocalTime'
ObjectIdNames[3497] = 'AuditDeleteReferencesEventType_Message'
ObjectIdNames[3498] = 'AuditDeleteReferencesEventType_Severity'
ObjectIdNames[3499] = 'AuditDeleteReferencesEventType_ActionTimeStamp'
ObjectIdNames[3500] = 'AuditDeleteReferencesEventType_Status'
ObjectIdNames[3501] = 'AuditDeleteReferencesEventType_ServerId'
ObjectIdNames[3502] = 'AuditDeleteReferencesEventType_ClientAuditEntryId'
ObjectIdNames[3503] = 'AuditDeleteReferencesEventType_ClientUserId'
ObjectIdNames[3504] = 'AuditUpdateEventType_EventId'
ObjectIdNames[3505] = 'AuditUpdateEventType_EventType'
ObjectIdNames[3506] = 'AuditUpdateEventType_SourceNode'
ObjectIdNames[3507] = 'AuditUpdateEventType_SourceName'
ObjectIdNames[3508] = 'AuditUpdateEventType_Time'
ObjectIdNames[3509] = 'AuditUpdateEventType_ReceiveTime'
ObjectIdNames[3510] = 'AuditUpdateEventType_LocalTime'
ObjectIdNames[3511] = 'AuditUpdateEventType_Message'
ObjectIdNames[3512] = 'AuditUpdateEventType_Severity'
ObjectIdNames[3513] = 'AuditUpdateEventType_ActionTimeStamp'
ObjectIdNames[3514] = 'AuditUpdateEventType_Status'
ObjectIdNames[3515] = 'AuditUpdateEventType_ServerId'
ObjectIdNames[3516] = 'AuditUpdateEventType_ClientAuditEntryId'
ObjectIdNames[3517] = 'AuditUpdateEventType_ClientUserId'
ObjectIdNames[3518] = 'AuditWriteUpdateEventType_EventId'
ObjectIdNames[3519] = 'AuditWriteUpdateEventType_EventType'
ObjectIdNames[3520] = 'AuditWriteUpdateEventType_SourceNode'
ObjectIdNames[3521] = 'AuditWriteUpdateEventType_SourceName'
ObjectIdNames[3522] = 'AuditWriteUpdateEventType_Time'
ObjectIdNames[3523] = 'AuditWriteUpdateEventType_ReceiveTime'
ObjectIdNames[3524] = 'AuditWriteUpdateEventType_LocalTime'
ObjectIdNames[3525] = 'AuditWriteUpdateEventType_Message'
ObjectIdNames[3526] = 'AuditWriteUpdateEventType_Severity'
ObjectIdNames[3527] = 'AuditWriteUpdateEventType_ActionTimeStamp'
ObjectIdNames[3528] = 'AuditWriteUpdateEventType_Status'
ObjectIdNames[3529] = 'AuditWriteUpdateEventType_ServerId'
ObjectIdNames[3530] = 'AuditWriteUpdateEventType_ClientAuditEntryId'
ObjectIdNames[3531] = 'AuditWriteUpdateEventType_ClientUserId'
ObjectIdNames[3532] = 'AuditHistoryUpdateEventType_EventId'
ObjectIdNames[3533] = 'AuditHistoryUpdateEventType_EventType'
ObjectIdNames[3534] = 'AuditHistoryUpdateEventType_SourceNode'
ObjectIdNames[3535] = 'AuditHistoryUpdateEventType_SourceName'
ObjectIdNames[3536] = 'AuditHistoryUpdateEventType_Time'
ObjectIdNames[3537] = 'AuditHistoryUpdateEventType_ReceiveTime'
ObjectIdNames[3538] = 'AuditHistoryUpdateEventType_LocalTime'
ObjectIdNames[3539] = 'AuditHistoryUpdateEventType_Message'
ObjectIdNames[3540] = 'AuditHistoryUpdateEventType_Severity'
ObjectIdNames[3541] = 'AuditHistoryUpdateEventType_ActionTimeStamp'
ObjectIdNames[3542] = 'AuditHistoryUpdateEventType_Status'
ObjectIdNames[3543] = 'AuditHistoryUpdateEventType_ServerId'
ObjectIdNames[3544] = 'AuditHistoryUpdateEventType_ClientAuditEntryId'
ObjectIdNames[3545] = 'AuditHistoryUpdateEventType_ClientUserId'
ObjectIdNames[3546] = 'AuditHistoryEventUpdateEventType_EventId'
ObjectIdNames[3547] = 'AuditHistoryEventUpdateEventType_EventType'
ObjectIdNames[3548] = 'AuditHistoryEventUpdateEventType_SourceNode'
ObjectIdNames[3549] = 'AuditHistoryEventUpdateEventType_SourceName'
ObjectIdNames[3550] = 'AuditHistoryEventUpdateEventType_Time'
ObjectIdNames[3551] = 'AuditHistoryEventUpdateEventType_ReceiveTime'
ObjectIdNames[3552] = 'AuditHistoryEventUpdateEventType_LocalTime'
ObjectIdNames[3553] = 'AuditHistoryEventUpdateEventType_Message'
ObjectIdNames[3554] = 'AuditHistoryEventUpdateEventType_Severity'
ObjectIdNames[3555] = 'AuditHistoryEventUpdateEventType_ActionTimeStamp'
ObjectIdNames[3556] = 'AuditHistoryEventUpdateEventType_Status'
ObjectIdNames[3557] = 'AuditHistoryEventUpdateEventType_ServerId'
ObjectIdNames[3558] = 'AuditHistoryEventUpdateEventType_ClientAuditEntryId'
ObjectIdNames[3559] = 'AuditHistoryEventUpdateEventType_ClientUserId'
ObjectIdNames[3560] = 'AuditHistoryEventUpdateEventType_ParameterDataTypeId'
ObjectIdNames[3561] = 'AuditHistoryValueUpdateEventType_EventId'
ObjectIdNames[3562] = 'AuditHistoryValueUpdateEventType_EventType'
ObjectIdNames[3563] = 'AuditHistoryValueUpdateEventType_SourceNode'
ObjectIdNames[3564] = 'AuditHistoryValueUpdateEventType_SourceName'
ObjectIdNames[3565] = 'AuditHistoryValueUpdateEventType_Time'
ObjectIdNames[3566] = 'AuditHistoryValueUpdateEventType_ReceiveTime'
ObjectIdNames[3567] = 'AuditHistoryValueUpdateEventType_LocalTime'
ObjectIdNames[3568] = 'AuditHistoryValueUpdateEventType_Message'
ObjectIdNames[3569] = 'AuditHistoryValueUpdateEventType_Severity'
ObjectIdNames[3570] = 'AuditHistoryValueUpdateEventType_ActionTimeStamp'
ObjectIdNames[3571] = 'AuditHistoryValueUpdateEventType_Status'
ObjectIdNames[3572] = 'AuditHistoryValueUpdateEventType_ServerId'
ObjectIdNames[3573] = 'AuditHistoryValueUpdateEventType_ClientAuditEntryId'
ObjectIdNames[3574] = 'AuditHistoryValueUpdateEventType_ClientUserId'
ObjectIdNames[3575] = 'AuditHistoryValueUpdateEventType_ParameterDataTypeId'
ObjectIdNames[3576] = 'AuditHistoryDeleteEventType_EventId'
ObjectIdNames[3577] = 'AuditHistoryDeleteEventType_EventType'
ObjectIdNames[3578] = 'AuditHistoryDeleteEventType_SourceNode'
ObjectIdNames[3579] = 'AuditHistoryDeleteEventType_SourceName'
ObjectIdNames[3580] = 'AuditHistoryDeleteEventType_Time'
ObjectIdNames[3581] = 'AuditHistoryDeleteEventType_ReceiveTime'
ObjectIdNames[3582] = 'AuditHistoryDeleteEventType_LocalTime'
ObjectIdNames[3583] = 'AuditHistoryDeleteEventType_Message'
ObjectIdNames[3584] = 'AuditHistoryDeleteEventType_Severity'
ObjectIdNames[3585] = 'AuditHistoryDeleteEventType_ActionTimeStamp'
ObjectIdNames[3586] = 'AuditHistoryDeleteEventType_Status'
ObjectIdNames[3587] = 'AuditHistoryDeleteEventType_ServerId'
ObjectIdNames[3588] = 'AuditHistoryDeleteEventType_ClientAuditEntryId'
ObjectIdNames[3589] = 'AuditHistoryDeleteEventType_ClientUserId'
ObjectIdNames[3590] = 'AuditHistoryDeleteEventType_ParameterDataTypeId'
ObjectIdNames[3591] = 'AuditHistoryRawModifyDeleteEventType_EventId'
ObjectIdNames[3592] = 'AuditHistoryRawModifyDeleteEventType_EventType'
ObjectIdNames[3593] = 'AuditHistoryRawModifyDeleteEventType_SourceNode'
ObjectIdNames[3594] = 'AuditHistoryRawModifyDeleteEventType_SourceName'
ObjectIdNames[3595] = 'AuditHistoryRawModifyDeleteEventType_Time'
ObjectIdNames[3596] = 'AuditHistoryRawModifyDeleteEventType_ReceiveTime'
ObjectIdNames[3597] = 'AuditHistoryRawModifyDeleteEventType_LocalTime'
ObjectIdNames[3598] = 'AuditHistoryRawModifyDeleteEventType_Message'
ObjectIdNames[3599] = 'AuditHistoryRawModifyDeleteEventType_Severity'
ObjectIdNames[3600] = 'AuditHistoryRawModifyDeleteEventType_ActionTimeStamp'
ObjectIdNames[3601] = 'AuditHistoryRawModifyDeleteEventType_Status'
ObjectIdNames[3602] = 'AuditHistoryRawModifyDeleteEventType_ServerId'
ObjectIdNames[3603] = 'AuditHistoryRawModifyDeleteEventType_ClientAuditEntryId'
ObjectIdNames[3604] = 'AuditHistoryRawModifyDeleteEventType_ClientUserId'
ObjectIdNames[3605] = 'AuditHistoryRawModifyDeleteEventType_ParameterDataTypeId'
ObjectIdNames[3606] = 'AuditHistoryRawModifyDeleteEventType_UpdatedNode'
ObjectIdNames[3607] = 'AuditHistoryAtTimeDeleteEventType_EventId'
ObjectIdNames[3608] = 'AuditHistoryAtTimeDeleteEventType_EventType'
ObjectIdNames[3609] = 'AuditHistoryAtTimeDeleteEventType_SourceNode'
ObjectIdNames[3610] = 'AuditHistoryAtTimeDeleteEventType_SourceName'
ObjectIdNames[3611] = 'AuditHistoryAtTimeDeleteEventType_Time'
ObjectIdNames[3612] = 'AuditHistoryAtTimeDeleteEventType_ReceiveTime'
ObjectIdNames[3613] = 'AuditHistoryAtTimeDeleteEventType_LocalTime'
ObjectIdNames[3614] = 'AuditHistoryAtTimeDeleteEventType_Message'
ObjectIdNames[3615] = 'AuditHistoryAtTimeDeleteEventType_Severity'
ObjectIdNames[3616] = 'AuditHistoryAtTimeDeleteEventType_ActionTimeStamp'
ObjectIdNames[3617] = 'AuditHistoryAtTimeDeleteEventType_Status'
ObjectIdNames[3618] = 'AuditHistoryAtTimeDeleteEventType_ServerId'
ObjectIdNames[3619] = 'AuditHistoryAtTimeDeleteEventType_ClientAuditEntryId'
ObjectIdNames[3620] = 'AuditHistoryAtTimeDeleteEventType_ClientUserId'
ObjectIdNames[3621] = 'AuditHistoryAtTimeDeleteEventType_ParameterDataTypeId'
ObjectIdNames[3622] = 'AuditHistoryAtTimeDeleteEventType_UpdatedNode'
ObjectIdNames[3623] = 'AuditHistoryEventDeleteEventType_EventId'
ObjectIdNames[3624] = 'AuditHistoryEventDeleteEventType_EventType'
ObjectIdNames[3625] = 'AuditHistoryEventDeleteEventType_SourceNode'
ObjectIdNames[3626] = 'AuditHistoryEventDeleteEventType_SourceName'
ObjectIdNames[3627] = 'AuditHistoryEventDeleteEventType_Time'
ObjectIdNames[3628] = 'AuditHistoryEventDeleteEventType_ReceiveTime'
ObjectIdNames[3629] = 'AuditHistoryEventDeleteEventType_LocalTime'
ObjectIdNames[3630] = 'AuditHistoryEventDeleteEventType_Message'
ObjectIdNames[3631] = 'AuditHistoryEventDeleteEventType_Severity'
ObjectIdNames[3632] = 'AuditHistoryEventDeleteEventType_ActionTimeStamp'
ObjectIdNames[3633] = 'AuditHistoryEventDeleteEventType_Status'
ObjectIdNames[3634] = 'AuditHistoryEventDeleteEventType_ServerId'
ObjectIdNames[3635] = 'AuditHistoryEventDeleteEventType_ClientAuditEntryId'
ObjectIdNames[3636] = 'AuditHistoryEventDeleteEventType_ClientUserId'
ObjectIdNames[3637] = 'AuditHistoryEventDeleteEventType_ParameterDataTypeId'
ObjectIdNames[3638] = 'AuditHistoryEventDeleteEventType_UpdatedNode'
ObjectIdNames[3639] = 'AuditUpdateMethodEventType_EventId'
ObjectIdNames[3640] = 'AuditUpdateMethodEventType_EventType'
ObjectIdNames[3641] = 'AuditUpdateMethodEventType_SourceNode'
ObjectIdNames[3642] = 'AuditUpdateMethodEventType_SourceName'
ObjectIdNames[3643] = 'AuditUpdateMethodEventType_Time'
ObjectIdNames[3644] = 'AuditUpdateMethodEventType_ReceiveTime'
ObjectIdNames[3645] = 'AuditUpdateMethodEventType_LocalTime'
ObjectIdNames[3646] = 'AuditUpdateMethodEventType_Message'
ObjectIdNames[3647] = 'AuditUpdateMethodEventType_Severity'
ObjectIdNames[3648] = 'AuditUpdateMethodEventType_ActionTimeStamp'
ObjectIdNames[3649] = 'AuditUpdateMethodEventType_Status'
ObjectIdNames[3650] = 'AuditUpdateMethodEventType_ServerId'
ObjectIdNames[3651] = 'AuditUpdateMethodEventType_ClientAuditEntryId'
ObjectIdNames[3652] = 'AuditUpdateMethodEventType_ClientUserId'
ObjectIdNames[3653] = 'SystemEventType_EventId'
ObjectIdNames[3654] = 'SystemEventType_EventType'
ObjectIdNames[3655] = 'SystemEventType_SourceNode'
ObjectIdNames[3656] = 'SystemEventType_SourceName'
ObjectIdNames[3657] = 'SystemEventType_Time'
ObjectIdNames[3658] = 'SystemEventType_ReceiveTime'
ObjectIdNames[3659] = 'SystemEventType_LocalTime'
ObjectIdNames[3660] = 'SystemEventType_Message'
ObjectIdNames[3661] = 'SystemEventType_Severity'
ObjectIdNames[3662] = 'DeviceFailureEventType_EventId'
ObjectIdNames[3663] = 'DeviceFailureEventType_EventType'
ObjectIdNames[3664] = 'DeviceFailureEventType_SourceNode'
ObjectIdNames[3665] = 'DeviceFailureEventType_SourceName'
ObjectIdNames[3666] = 'DeviceFailureEventType_Time'
ObjectIdNames[3667] = 'DeviceFailureEventType_ReceiveTime'
ObjectIdNames[3668] = 'DeviceFailureEventType_LocalTime'
ObjectIdNames[3669] = 'DeviceFailureEventType_Message'
ObjectIdNames[3670] = 'DeviceFailureEventType_Severity'
ObjectIdNames[3671] = 'BaseModelChangeEventType_EventId'
ObjectIdNames[3672] = 'BaseModelChangeEventType_EventType'
ObjectIdNames[3673] = 'BaseModelChangeEventType_SourceNode'
ObjectIdNames[3674] = 'BaseModelChangeEventType_SourceName'
ObjectIdNames[3675] = 'BaseModelChangeEventType_Time'
ObjectIdNames[3676] = 'BaseModelChangeEventType_ReceiveTime'
ObjectIdNames[3677] = 'BaseModelChangeEventType_LocalTime'
ObjectIdNames[3678] = 'BaseModelChangeEventType_Message'
ObjectIdNames[3679] = 'BaseModelChangeEventType_Severity'
ObjectIdNames[3680] = 'GeneralModelChangeEventType_EventId'
ObjectIdNames[3681] = 'GeneralModelChangeEventType_EventType'
ObjectIdNames[3682] = 'GeneralModelChangeEventType_SourceNode'
ObjectIdNames[3683] = 'GeneralModelChangeEventType_SourceName'
ObjectIdNames[3684] = 'GeneralModelChangeEventType_Time'
ObjectIdNames[3685] = 'GeneralModelChangeEventType_ReceiveTime'
ObjectIdNames[3686] = 'GeneralModelChangeEventType_LocalTime'
ObjectIdNames[3687] = 'GeneralModelChangeEventType_Message'
ObjectIdNames[3688] = 'GeneralModelChangeEventType_Severity'
ObjectIdNames[3689] = 'SemanticChangeEventType_EventId'
ObjectIdNames[3690] = 'SemanticChangeEventType_EventType'
ObjectIdNames[3691] = 'SemanticChangeEventType_SourceNode'
ObjectIdNames[3692] = 'SemanticChangeEventType_SourceName'
ObjectIdNames[3693] = 'SemanticChangeEventType_Time'
ObjectIdNames[3694] = 'SemanticChangeEventType_ReceiveTime'
ObjectIdNames[3695] = 'SemanticChangeEventType_LocalTime'
ObjectIdNames[3696] = 'SemanticChangeEventType_Message'
ObjectIdNames[3697] = 'SemanticChangeEventType_Severity'
ObjectIdNames[3698] = 'ServerStatusType_BuildInfo_ProductUri'
ObjectIdNames[3699] = 'ServerStatusType_BuildInfo_ManufacturerName'
ObjectIdNames[3700] = 'ServerStatusType_BuildInfo_ProductName'
ObjectIdNames[3701] = 'ServerStatusType_BuildInfo_SoftwareVersion'
ObjectIdNames[3702] = 'ServerStatusType_BuildInfo_BuildNumber'
ObjectIdNames[3703] = 'ServerStatusType_BuildInfo_BuildDate'
ObjectIdNames[3704] = 'Server_ServerCapabilities_SoftwareCertificates'
ObjectIdNames[3705] = 'Server_ServerDiagnostics_ServerDiagnosticsSummary_RejectedSessionCount'
ObjectIdNames[3706] = 'Server_ServerDiagnostics_SessionsDiagnosticsSummary'
ObjectIdNames[3707] = 'Server_ServerDiagnostics_SessionsDiagnosticsSummary_SessionDiagnosticsArray'
ObjectIdNames[3708] = 'Server_ServerDiagnostics_SessionsDiagnosticsSummary_SessionSecurityDiagnosticsArray'
ObjectIdNames[3709] = 'Server_ServerRedundancy_RedundancySupport'
ObjectIdNames[3714] = 'FiniteStateVariableType_Name'
ObjectIdNames[3715] = 'FiniteStateVariableType_Number'
ObjectIdNames[3716] = 'FiniteStateVariableType_EffectiveDisplayName'
ObjectIdNames[3717] = 'FiniteTransitionVariableType_Name'
ObjectIdNames[3718] = 'FiniteTransitionVariableType_Number'
ObjectIdNames[3719] = 'FiniteTransitionVariableType_TransitionTime'
ObjectIdNames[3720] = 'StateMachineType_CurrentState_Id'
ObjectIdNames[3721] = 'StateMachineType_CurrentState_Name'
ObjectIdNames[3722] = 'StateMachineType_CurrentState_Number'
ObjectIdNames[3723] = 'StateMachineType_CurrentState_EffectiveDisplayName'
ObjectIdNames[3724] = 'StateMachineType_LastTransition_Id'
ObjectIdNames[3725] = 'StateMachineType_LastTransition_Name'
ObjectIdNames[3726] = 'StateMachineType_LastTransition_Number'
ObjectIdNames[3727] = 'StateMachineType_LastTransition_TransitionTime'
ObjectIdNames[3728] = 'FiniteStateMachineType_CurrentState_Id'
ObjectIdNames[3729] = 'FiniteStateMachineType_CurrentState_Name'
ObjectIdNames[3730] = 'FiniteStateMachineType_CurrentState_Number'
ObjectIdNames[3731] = 'FiniteStateMachineType_CurrentState_EffectiveDisplayName'
ObjectIdNames[3732] = 'FiniteStateMachineType_LastTransition_Id'
ObjectIdNames[3733] = 'FiniteStateMachineType_LastTransition_Name'
ObjectIdNames[3734] = 'FiniteStateMachineType_LastTransition_Number'
ObjectIdNames[3735] = 'FiniteStateMachineType_LastTransition_TransitionTime'
ObjectIdNames[3736] = 'InitialStateType_StateNumber'
ObjectIdNames[3737] = 'TransitionEventType_EventId'
ObjectIdNames[3738] = 'TransitionEventType_EventType'
ObjectIdNames[3739] = 'TransitionEventType_SourceNode'
ObjectIdNames[3740] = 'TransitionEventType_SourceName'
ObjectIdNames[3741] = 'TransitionEventType_Time'
ObjectIdNames[3742] = 'TransitionEventType_ReceiveTime'
ObjectIdNames[3743] = 'TransitionEventType_LocalTime'
ObjectIdNames[3744] = 'TransitionEventType_Message'
ObjectIdNames[3745] = 'TransitionEventType_Severity'
ObjectIdNames[3746] = 'TransitionEventType_FromState_Id'
ObjectIdNames[3747] = 'TransitionEventType_FromState_Name'
ObjectIdNames[3748] = 'TransitionEventType_FromState_Number'
ObjectIdNames[3749] = 'TransitionEventType_FromState_EffectiveDisplayName'
ObjectIdNames[3750] = 'TransitionEventType_ToState_Id'
ObjectIdNames[3751] = 'TransitionEventType_ToState_Name'
ObjectIdNames[3752] = 'TransitionEventType_ToState_Number'
ObjectIdNames[3753] = 'TransitionEventType_ToState_EffectiveDisplayName'
ObjectIdNames[3754] = 'TransitionEventType_Transition_Id'
ObjectIdNames[3755] = 'TransitionEventType_Transition_Name'
ObjectIdNames[3756] = 'TransitionEventType_Transition_Number'
ObjectIdNames[3757] = 'TransitionEventType_Transition_TransitionTime'
ObjectIdNames[3758] = 'AuditUpdateStateEventType_EventId'
ObjectIdNames[3759] = 'AuditUpdateStateEventType_EventType'
ObjectIdNames[3760] = 'AuditUpdateStateEventType_SourceNode'
ObjectIdNames[3761] = 'AuditUpdateStateEventType_SourceName'
ObjectIdNames[3762] = 'AuditUpdateStateEventType_Time'
ObjectIdNames[3763] = 'AuditUpdateStateEventType_ReceiveTime'
ObjectIdNames[3764] = 'AuditUpdateStateEventType_LocalTime'
ObjectIdNames[3765] = 'AuditUpdateStateEventType_Message'
ObjectIdNames[3766] = 'AuditUpdateStateEventType_Severity'
ObjectIdNames[3767] = 'AuditUpdateStateEventType_ActionTimeStamp'
ObjectIdNames[3768] = 'AuditUpdateStateEventType_Status'
ObjectIdNames[3769] = 'AuditUpdateStateEventType_ServerId'
ObjectIdNames[3770] = 'AuditUpdateStateEventType_ClientAuditEntryId'
ObjectIdNames[3771] = 'AuditUpdateStateEventType_ClientUserId'
ObjectIdNames[3772] = 'AuditUpdateStateEventType_MethodId'
ObjectIdNames[3773] = 'AuditUpdateStateEventType_InputArguments'
ObjectIdNames[3774] = 'AnalogItemType_Definition'
ObjectIdNames[3775] = 'AnalogItemType_ValuePrecision'
ObjectIdNames[3776] = 'DiscreteItemType_Definition'
ObjectIdNames[3777] = 'DiscreteItemType_ValuePrecision'
ObjectIdNames[3778] = 'TwoStateDiscreteType_Definition'
ObjectIdNames[3779] = 'TwoStateDiscreteType_ValuePrecision'
ObjectIdNames[3780] = 'MultiStateDiscreteType_Definition'
ObjectIdNames[3781] = 'MultiStateDiscreteType_ValuePrecision'
ObjectIdNames[3782] = 'ProgramTransitionEventType_EventId'
ObjectIdNames[3783] = 'ProgramTransitionEventType_EventType'
ObjectIdNames[3784] = 'ProgramTransitionEventType_SourceNode'
ObjectIdNames[3785] = 'ProgramTransitionEventType_SourceName'
ObjectIdNames[3786] = 'ProgramTransitionEventType_Time'
ObjectIdNames[3787] = 'ProgramTransitionEventType_ReceiveTime'
ObjectIdNames[3788] = 'ProgramTransitionEventType_LocalTime'
ObjectIdNames[3789] = 'ProgramTransitionEventType_Message'
ObjectIdNames[3790] = 'ProgramTransitionEventType_Severity'
ObjectIdNames[3791] = 'ProgramTransitionEventType_FromState'
ObjectIdNames[3792] = 'ProgramTransitionEventType_FromState_Id'
ObjectIdNames[3793] = 'ProgramTransitionEventType_FromState_Name'
ObjectIdNames[3794] = 'ProgramTransitionEventType_FromState_Number'
ObjectIdNames[3795] = 'ProgramTransitionEventType_FromState_EffectiveDisplayName'
ObjectIdNames[3796] = 'ProgramTransitionEventType_ToState'
ObjectIdNames[3797] = 'ProgramTransitionEventType_ToState_Id'
ObjectIdNames[3798] = 'ProgramTransitionEventType_ToState_Name'
ObjectIdNames[3799] = 'ProgramTransitionEventType_ToState_Number'
ObjectIdNames[3800] = 'ProgramTransitionEventType_ToState_EffectiveDisplayName'
ObjectIdNames[3801] = 'ProgramTransitionEventType_Transition'
ObjectIdNames[3802] = 'ProgramTransitionEventType_Transition_Id'
ObjectIdNames[3803] = 'ProgramTransitionEventType_Transition_Name'
ObjectIdNames[3804] = 'ProgramTransitionEventType_Transition_Number'
ObjectIdNames[3805] = 'ProgramTransitionEventType_Transition_TransitionTime'
ObjectIdNames[3806] = 'ProgramTransitionAuditEventType'
ObjectIdNames[3807] = 'ProgramTransitionAuditEventType_EventId'
ObjectIdNames[3808] = 'ProgramTransitionAuditEventType_EventType'
ObjectIdNames[3809] = 'ProgramTransitionAuditEventType_SourceNode'
ObjectIdNames[3810] = 'ProgramTransitionAuditEventType_SourceName'
ObjectIdNames[3811] = 'ProgramTransitionAuditEventType_Time'
ObjectIdNames[3812] = 'ProgramTransitionAuditEventType_ReceiveTime'
ObjectIdNames[3813] = 'ProgramTransitionAuditEventType_LocalTime'
ObjectIdNames[3814] = 'ProgramTransitionAuditEventType_Message'
ObjectIdNames[3815] = 'ProgramTransitionAuditEventType_Severity'
ObjectIdNames[3816] = 'ProgramTransitionAuditEventType_ActionTimeStamp'
ObjectIdNames[3817] = 'ProgramTransitionAuditEventType_Status'
ObjectIdNames[3818] = 'ProgramTransitionAuditEventType_ServerId'
ObjectIdNames[3819] = 'ProgramTransitionAuditEventType_ClientAuditEntryId'
ObjectIdNames[3820] = 'ProgramTransitionAuditEventType_ClientUserId'
ObjectIdNames[3821] = 'ProgramTransitionAuditEventType_MethodId'
ObjectIdNames[3822] = 'ProgramTransitionAuditEventType_InputArguments'
ObjectIdNames[3823] = 'ProgramTransitionAuditEventType_OldStateId'
ObjectIdNames[3824] = 'ProgramTransitionAuditEventType_NewStateId'
ObjectIdNames[3825] = 'ProgramTransitionAuditEventType_Transition'
ObjectIdNames[3826] = 'ProgramTransitionAuditEventType_Transition_Id'
ObjectIdNames[3827] = 'ProgramTransitionAuditEventType_Transition_Name'
ObjectIdNames[3828] = 'ProgramTransitionAuditEventType_Transition_Number'
ObjectIdNames[3829] = 'ProgramTransitionAuditEventType_Transition_TransitionTime'
ObjectIdNames[3830] = 'ProgramStateMachineType_CurrentState'
ObjectIdNames[3831] = 'ProgramStateMachineType_CurrentState_Id'
ObjectIdNames[3832] = 'ProgramStateMachineType_CurrentState_Name'
ObjectIdNames[3833] = 'ProgramStateMachineType_CurrentState_Number'
ObjectIdNames[3834] = 'ProgramStateMachineType_CurrentState_EffectiveDisplayName'
ObjectIdNames[3835] = 'ProgramStateMachineType_LastTransition'
ObjectIdNames[3836] = 'ProgramStateMachineType_LastTransition_Id'
ObjectIdNames[3837] = 'ProgramStateMachineType_LastTransition_Name'
ObjectIdNames[3838] = 'ProgramStateMachineType_LastTransition_Number'
ObjectIdNames[3839] = 'ProgramStateMachineType_LastTransition_TransitionTime'
ObjectIdNames[3840] = 'ProgramStateMachineType_ProgramDiagnostics_CreateSessionId'
ObjectIdNames[3841] = 'ProgramStateMachineType_ProgramDiagnostics_CreateClientName'
ObjectIdNames[3842] = 'ProgramStateMachineType_ProgramDiagnostics_InvocationCreationTime'
ObjectIdNames[3843] = 'ProgramStateMachineType_ProgramDiagnostics_LastTransitionTime'
ObjectIdNames[3844] = 'ProgramStateMachineType_ProgramDiagnostics_LastMethodCall'
ObjectIdNames[3845] = 'ProgramStateMachineType_ProgramDiagnostics_LastMethodSessionId'
ObjectIdNames[3846] = 'ProgramStateMachineType_ProgramDiagnostics_LastMethodInputArguments'
ObjectIdNames[3847] = 'ProgramStateMachineType_ProgramDiagnostics_LastMethodOutputArguments'
ObjectIdNames[3848] = 'ProgramStateMachineType_ProgramDiagnostics_LastMethodCallTime'
ObjectIdNames[3849] = 'ProgramStateMachineType_ProgramDiagnostics_LastMethodReturnStatus'
ObjectIdNames[3850] = 'ProgramStateMachineType_FinalResultData'
ObjectIdNames[3863] = 'AddCommentMethodType'
ObjectIdNames[3864] = 'AddCommentMethodType_InputArguments'
ObjectIdNames[3865] = 'ConditionType_EventId'
ObjectIdNames[3866] = 'ConditionType_EventType'
ObjectIdNames[3867] = 'ConditionType_SourceNode'
ObjectIdNames[3868] = 'ConditionType_SourceName'
ObjectIdNames[3869] = 'ConditionType_Time'
ObjectIdNames[3870] = 'ConditionType_ReceiveTime'
ObjectIdNames[3871] = 'ConditionType_LocalTime'
ObjectIdNames[3872] = 'ConditionType_Message'
ObjectIdNames[3873] = 'ConditionType_Severity'
ObjectIdNames[3874] = 'ConditionType_Retain'
ObjectIdNames[3875] = 'ConditionType_ConditionRefresh'
ObjectIdNames[3876] = 'ConditionType_ConditionRefresh_InputArguments'
ObjectIdNames[3969] = 'RefreshStartEventType_EventId'
ObjectIdNames[3970] = 'RefreshStartEventType_EventType'
ObjectIdNames[3971] = 'RefreshStartEventType_SourceNode'
ObjectIdNames[3972] = 'RefreshStartEventType_SourceName'
ObjectIdNames[3973] = 'RefreshStartEventType_Time'
ObjectIdNames[3974] = 'RefreshStartEventType_ReceiveTime'
ObjectIdNames[3975] = 'RefreshStartEventType_LocalTime'
ObjectIdNames[3976] = 'RefreshStartEventType_Message'
ObjectIdNames[3977] = 'RefreshStartEventType_Severity'
ObjectIdNames[3978] = 'RefreshEndEventType_EventId'
ObjectIdNames[3979] = 'RefreshEndEventType_EventType'
ObjectIdNames[3980] = 'RefreshEndEventType_SourceNode'
ObjectIdNames[3981] = 'RefreshEndEventType_SourceName'
ObjectIdNames[3982] = 'RefreshEndEventType_Time'
ObjectIdNames[3983] = 'RefreshEndEventType_ReceiveTime'
ObjectIdNames[3984] = 'RefreshEndEventType_LocalTime'
ObjectIdNames[3985] = 'RefreshEndEventType_Message'
ObjectIdNames[3986] = 'RefreshEndEventType_Severity'
ObjectIdNames[3987] = 'RefreshRequiredEventType_EventId'
ObjectIdNames[3988] = 'RefreshRequiredEventType_EventType'
ObjectIdNames[3989] = 'RefreshRequiredEventType_SourceNode'
ObjectIdNames[3990] = 'RefreshRequiredEventType_SourceName'
ObjectIdNames[3991] = 'RefreshRequiredEventType_Time'
ObjectIdNames[3992] = 'RefreshRequiredEventType_ReceiveTime'
ObjectIdNames[3993] = 'RefreshRequiredEventType_LocalTime'
ObjectIdNames[3994] = 'RefreshRequiredEventType_Message'
ObjectIdNames[3995] = 'RefreshRequiredEventType_Severity'
ObjectIdNames[3996] = 'AuditConditionEventType_EventId'
ObjectIdNames[3997] = 'AuditConditionEventType_EventType'
ObjectIdNames[3998] = 'AuditConditionEventType_SourceNode'
ObjectIdNames[3999] = 'AuditConditionEventType_SourceName'
ObjectIdNames[4000] = 'AuditConditionEventType_Time'
ObjectIdNames[4001] = 'AuditConditionEventType_ReceiveTime'
ObjectIdNames[4002] = 'AuditConditionEventType_LocalTime'
ObjectIdNames[4003] = 'AuditConditionEventType_Message'
ObjectIdNames[4004] = 'AuditConditionEventType_Severity'
ObjectIdNames[4005] = 'AuditConditionEventType_ActionTimeStamp'
ObjectIdNames[4006] = 'AuditConditionEventType_Status'
ObjectIdNames[4007] = 'AuditConditionEventType_ServerId'
ObjectIdNames[4008] = 'AuditConditionEventType_ClientAuditEntryId'
ObjectIdNames[4009] = 'AuditConditionEventType_ClientUserId'
ObjectIdNames[4010] = 'AuditConditionEventType_MethodId'
ObjectIdNames[4011] = 'AuditConditionEventType_InputArguments'
ObjectIdNames[4106] = 'AuditConditionEnableEventType_EventId'
ObjectIdNames[4107] = 'AuditConditionEnableEventType_EventType'
ObjectIdNames[4108] = 'AuditConditionEnableEventType_SourceNode'
ObjectIdNames[4109] = 'AuditConditionEnableEventType_SourceName'
ObjectIdNames[4110] = 'AuditConditionEnableEventType_Time'
ObjectIdNames[4111] = 'AuditConditionEnableEventType_ReceiveTime'
ObjectIdNames[4112] = 'AuditConditionEnableEventType_LocalTime'
ObjectIdNames[4113] = 'AuditConditionEnableEventType_Message'
ObjectIdNames[4114] = 'AuditConditionEnableEventType_Severity'
ObjectIdNames[4115] = 'AuditConditionEnableEventType_ActionTimeStamp'
ObjectIdNames[4116] = 'AuditConditionEnableEventType_Status'
ObjectIdNames[4117] = 'AuditConditionEnableEventType_ServerId'
ObjectIdNames[4118] = 'AuditConditionEnableEventType_ClientAuditEntryId'
ObjectIdNames[4119] = 'AuditConditionEnableEventType_ClientUserId'
ObjectIdNames[4120] = 'AuditConditionEnableEventType_MethodId'
ObjectIdNames[4121] = 'AuditConditionEnableEventType_InputArguments'
ObjectIdNames[4170] = 'AuditConditionCommentEventType_EventId'
ObjectIdNames[4171] = 'AuditConditionCommentEventType_EventType'
ObjectIdNames[4172] = 'AuditConditionCommentEventType_SourceNode'
ObjectIdNames[4173] = 'AuditConditionCommentEventType_SourceName'
ObjectIdNames[4174] = 'AuditConditionCommentEventType_Time'
ObjectIdNames[4175] = 'AuditConditionCommentEventType_ReceiveTime'
ObjectIdNames[4176] = 'AuditConditionCommentEventType_LocalTime'
ObjectIdNames[4177] = 'AuditConditionCommentEventType_Message'
ObjectIdNames[4178] = 'AuditConditionCommentEventType_Severity'
ObjectIdNames[4179] = 'AuditConditionCommentEventType_ActionTimeStamp'
ObjectIdNames[4180] = 'AuditConditionCommentEventType_Status'
ObjectIdNames[4181] = 'AuditConditionCommentEventType_ServerId'
ObjectIdNames[4182] = 'AuditConditionCommentEventType_ClientAuditEntryId'
ObjectIdNames[4183] = 'AuditConditionCommentEventType_ClientUserId'
ObjectIdNames[4184] = 'AuditConditionCommentEventType_MethodId'
ObjectIdNames[4185] = 'AuditConditionCommentEventType_InputArguments'
ObjectIdNames[4188] = 'DialogConditionType_EventId'
ObjectIdNames[4189] = 'DialogConditionType_EventType'
ObjectIdNames[4190] = 'DialogConditionType_SourceNode'
ObjectIdNames[4191] = 'DialogConditionType_SourceName'
ObjectIdNames[4192] = 'DialogConditionType_Time'
ObjectIdNames[4193] = 'DialogConditionType_ReceiveTime'
ObjectIdNames[4194] = 'DialogConditionType_LocalTime'
ObjectIdNames[4195] = 'DialogConditionType_Message'
ObjectIdNames[4196] = 'DialogConditionType_Severity'
ObjectIdNames[4197] = 'DialogConditionType_Retain'
ObjectIdNames[4198] = 'DialogConditionType_ConditionRefresh'
ObjectIdNames[4199] = 'DialogConditionType_ConditionRefresh_InputArguments'
ObjectIdNames[5113] = 'AcknowledgeableConditionType_EventId'
ObjectIdNames[5114] = 'AcknowledgeableConditionType_EventType'
ObjectIdNames[5115] = 'AcknowledgeableConditionType_SourceNode'
ObjectIdNames[5116] = 'AcknowledgeableConditionType_SourceName'
ObjectIdNames[5117] = 'AcknowledgeableConditionType_Time'
ObjectIdNames[5118] = 'AcknowledgeableConditionType_ReceiveTime'
ObjectIdNames[5119] = 'AcknowledgeableConditionType_LocalTime'
ObjectIdNames[5120] = 'AcknowledgeableConditionType_Message'
ObjectIdNames[5121] = 'AcknowledgeableConditionType_Severity'
ObjectIdNames[5122] = 'AcknowledgeableConditionType_Retain'
ObjectIdNames[5123] = 'AcknowledgeableConditionType_ConditionRefresh'
ObjectIdNames[5124] = 'AcknowledgeableConditionType_ConditionRefresh_InputArguments'
ObjectIdNames[5540] = 'AlarmConditionType_EventId'
ObjectIdNames[5541] = 'AlarmConditionType_EventType'
ObjectIdNames[5542] = 'AlarmConditionType_SourceNode'
ObjectIdNames[5543] = 'AlarmConditionType_SourceName'
ObjectIdNames[5544] = 'AlarmConditionType_Time'
ObjectIdNames[5545] = 'AlarmConditionType_ReceiveTime'
ObjectIdNames[5546] = 'AlarmConditionType_LocalTime'
ObjectIdNames[5547] = 'AlarmConditionType_Message'
ObjectIdNames[5548] = 'AlarmConditionType_Severity'
ObjectIdNames[5549] = 'AlarmConditionType_Retain'
ObjectIdNames[5550] = 'AlarmConditionType_ConditionRefresh'
ObjectIdNames[5551] = 'AlarmConditionType_ConditionRefresh_InputArguments'
ObjectIdNames[6088] = 'ShelvedStateMachineType_CurrentState'
ObjectIdNames[6089] = 'ShelvedStateMachineType_CurrentState_Id'
ObjectIdNames[6090] = 'ShelvedStateMachineType_CurrentState_Name'
ObjectIdNames[6091] = 'ShelvedStateMachineType_CurrentState_Number'
ObjectIdNames[6092] = 'ShelvedStateMachineType_CurrentState_EffectiveDisplayName'
ObjectIdNames[6093] = 'ShelvedStateMachineType_LastTransition'
ObjectIdNames[6094] = 'ShelvedStateMachineType_LastTransition_Id'
ObjectIdNames[6095] = 'ShelvedStateMachineType_LastTransition_Name'
ObjectIdNames[6096] = 'ShelvedStateMachineType_LastTransition_Number'
ObjectIdNames[6097] = 'ShelvedStateMachineType_LastTransition_TransitionTime'
ObjectIdNames[6098] = 'ShelvedStateMachineType_Unshelved_StateNumber'
ObjectIdNames[6100] = 'ShelvedStateMachineType_TimedShelved_StateNumber'
ObjectIdNames[6101] = 'ShelvedStateMachineType_OneShotShelved_StateNumber'
ObjectIdNames[6102] = 'TimedShelveMethodType'
ObjectIdNames[6103] = 'TimedShelveMethodType_InputArguments'
ObjectIdNames[6116] = 'LimitAlarmType_EventId'
ObjectIdNames[6117] = 'LimitAlarmType_EventType'
ObjectIdNames[6118] = 'LimitAlarmType_SourceNode'
ObjectIdNames[6119] = 'LimitAlarmType_SourceName'
ObjectIdNames[6120] = 'LimitAlarmType_Time'
ObjectIdNames[6121] = 'LimitAlarmType_ReceiveTime'
ObjectIdNames[6122] = 'LimitAlarmType_LocalTime'
ObjectIdNames[6123] = 'LimitAlarmType_Message'
ObjectIdNames[6124] = 'LimitAlarmType_Severity'
ObjectIdNames[6125] = 'LimitAlarmType_Retain'
ObjectIdNames[6126] = 'LimitAlarmType_ConditionRefresh'
ObjectIdNames[6127] = 'LimitAlarmType_ConditionRefresh_InputArguments'
ObjectIdNames[7591] = 'IdType_EnumStrings'
ObjectIdNames[7594] = 'EnumValueType'
ObjectIdNames[7595] = 'MessageSecurityMode_EnumStrings'
ObjectIdNames[7596] = 'UserTokenType_EnumStrings'
ObjectIdNames[7597] = 'ApplicationType_EnumStrings'
ObjectIdNames[7598] = 'SecurityTokenRequestType_EnumStrings'
ObjectIdNames[7599] = 'ComplianceLevel_EnumStrings'
ObjectIdNames[7603] = 'BrowseDirection_EnumStrings'
ObjectIdNames[7605] = 'FilterOperator_EnumStrings'
ObjectIdNames[7606] = 'TimestampsToReturn_EnumStrings'
ObjectIdNames[7608] = 'MonitoringMode_EnumStrings'
ObjectIdNames[7609] = 'DataChangeTrigger_EnumStrings'
ObjectIdNames[7610] = 'DeadbandType_EnumStrings'
ObjectIdNames[7611] = 'RedundancySupport_EnumStrings'
ObjectIdNames[7612] = 'ServerState_EnumStrings'
ObjectIdNames[7614] = 'ExceptionDeviationFormat_EnumStrings'
ObjectIdNames[7616] = 'EnumValueType_Encoding_DefaultXml'
ObjectIdNames[7617] = 'OpcUa_BinarySchema'
ObjectIdNames[7618] = 'OpcUa_BinarySchema_DataTypeVersion'
ObjectIdNames[7619] = 'OpcUa_BinarySchema_NamespaceUri'
ObjectIdNames[7650] = 'OpcUa_BinarySchema_Argument'
ObjectIdNames[7651] = 'OpcUa_BinarySchema_Argument_DataTypeVersion'
ObjectIdNames[7652] = 'OpcUa_BinarySchema_Argument_DictionaryFragment'
ObjectIdNames[7656] = 'OpcUa_BinarySchema_EnumValueType'
ObjectIdNames[7657] = 'OpcUa_BinarySchema_EnumValueType_DataTypeVersion'
ObjectIdNames[7658] = 'OpcUa_BinarySchema_EnumValueType_DictionaryFragment'
ObjectIdNames[7659] = 'OpcUa_BinarySchema_StatusResult'
ObjectIdNames[7660] = 'OpcUa_BinarySchema_StatusResult_DataTypeVersion'
ObjectIdNames[7661] = 'OpcUa_BinarySchema_StatusResult_DictionaryFragment'
ObjectIdNames[7662] = 'OpcUa_BinarySchema_UserTokenPolicy'
ObjectIdNames[7663] = 'OpcUa_BinarySchema_UserTokenPolicy_DataTypeVersion'
ObjectIdNames[7664] = 'OpcUa_BinarySchema_UserTokenPolicy_DictionaryFragment'
ObjectIdNames[7665] = 'OpcUa_BinarySchema_ApplicationDescription'
ObjectIdNames[7666] = 'OpcUa_BinarySchema_ApplicationDescription_DataTypeVersion'
ObjectIdNames[7667] = 'OpcUa_BinarySchema_ApplicationDescription_DictionaryFragment'
ObjectIdNames[7668] = 'OpcUa_BinarySchema_EndpointDescription'
ObjectIdNames[7669] = 'OpcUa_BinarySchema_EndpointDescription_DataTypeVersion'
ObjectIdNames[7670] = 'OpcUa_BinarySchema_EndpointDescription_DictionaryFragment'
ObjectIdNames[7671] = 'OpcUa_BinarySchema_UserIdentityToken'
ObjectIdNames[7672] = 'OpcUa_BinarySchema_UserIdentityToken_DataTypeVersion'
ObjectIdNames[7673] = 'OpcUa_BinarySchema_UserIdentityToken_DictionaryFragment'
ObjectIdNames[7674] = 'OpcUa_BinarySchema_AnonymousIdentityToken'
ObjectIdNames[7675] = 'OpcUa_BinarySchema_AnonymousIdentityToken_DataTypeVersion'
ObjectIdNames[7676] = 'OpcUa_BinarySchema_AnonymousIdentityToken_DictionaryFragment'
ObjectIdNames[7677] = 'OpcUa_BinarySchema_UserNameIdentityToken'
ObjectIdNames[7678] = 'OpcUa_BinarySchema_UserNameIdentityToken_DataTypeVersion'
ObjectIdNames[7679] = 'OpcUa_BinarySchema_UserNameIdentityToken_DictionaryFragment'
ObjectIdNames[7680] = 'OpcUa_BinarySchema_X509IdentityToken'
ObjectIdNames[7681] = 'OpcUa_BinarySchema_X509IdentityToken_DataTypeVersion'
ObjectIdNames[7682] = 'OpcUa_BinarySchema_X509IdentityToken_DictionaryFragment'
ObjectIdNames[7683] = 'OpcUa_BinarySchema_IssuedIdentityToken'
ObjectIdNames[7684] = 'OpcUa_BinarySchema_IssuedIdentityToken_DataTypeVersion'
ObjectIdNames[7685] = 'OpcUa_BinarySchema_IssuedIdentityToken_DictionaryFragment'
ObjectIdNames[7686] = 'OpcUa_BinarySchema_EndpointConfiguration'
ObjectIdNames[7687] = 'OpcUa_BinarySchema_EndpointConfiguration_DataTypeVersion'
ObjectIdNames[7688] = 'OpcUa_BinarySchema_EndpointConfiguration_DictionaryFragment'
ObjectIdNames[7689] = 'OpcUa_BinarySchema_SupportedProfile'
ObjectIdNames[7690] = 'OpcUa_BinarySchema_SupportedProfile_DataTypeVersion'
ObjectIdNames[7691] = 'OpcUa_BinarySchema_SupportedProfile_DictionaryFragment'
ObjectIdNames[7692] = 'OpcUa_BinarySchema_BuildInfo'
ObjectIdNames[7693] = 'OpcUa_BinarySchema_BuildInfo_DataTypeVersion'
ObjectIdNames[7694] = 'OpcUa_BinarySchema_BuildInfo_DictionaryFragment'
ObjectIdNames[7695] = 'OpcUa_BinarySchema_SoftwareCertificate'
ObjectIdNames[7696] = 'OpcUa_BinarySchema_SoftwareCertificate_DataTypeVersion'
ObjectIdNames[7697] = 'OpcUa_BinarySchema_SoftwareCertificate_DictionaryFragment'
ObjectIdNames[7698] = 'OpcUa_BinarySchema_SignedSoftwareCertificate'
ObjectIdNames[7699] = 'OpcUa_BinarySchema_SignedSoftwareCertificate_DataTypeVersion'
ObjectIdNames[7700] = 'OpcUa_BinarySchema_SignedSoftwareCertificate_DictionaryFragment'
ObjectIdNames[7728] = 'OpcUa_BinarySchema_AddNodesItem'
ObjectIdNames[7729] = 'OpcUa_BinarySchema_AddNodesItem_DataTypeVersion'
ObjectIdNames[7730] = 'OpcUa_BinarySchema_AddNodesItem_DictionaryFragment'
ObjectIdNames[7731] = 'OpcUa_BinarySchema_AddReferencesItem'
ObjectIdNames[7732] = 'OpcUa_BinarySchema_AddReferencesItem_DataTypeVersion'
ObjectIdNames[7733] = 'OpcUa_BinarySchema_AddReferencesItem_DictionaryFragment'
ObjectIdNames[7734] = 'OpcUa_BinarySchema_DeleteNodesItem'
ObjectIdNames[7735] = 'OpcUa_BinarySchema_DeleteNodesItem_DataTypeVersion'
ObjectIdNames[7736] = 'OpcUa_BinarySchema_DeleteNodesItem_DictionaryFragment'
ObjectIdNames[7737] = 'OpcUa_BinarySchema_DeleteReferencesItem'
ObjectIdNames[7738] = 'OpcUa_BinarySchema_DeleteReferencesItem_DataTypeVersion'
ObjectIdNames[7739] = 'OpcUa_BinarySchema_DeleteReferencesItem_DictionaryFragment'
ObjectIdNames[7782] = 'OpcUa_BinarySchema_RegisteredServer'
ObjectIdNames[7783] = 'OpcUa_BinarySchema_RegisteredServer_DataTypeVersion'
ObjectIdNames[7784] = 'OpcUa_BinarySchema_RegisteredServer_DictionaryFragment'
ObjectIdNames[7929] = 'OpcUa_BinarySchema_ContentFilterElement'
ObjectIdNames[7930] = 'OpcUa_BinarySchema_ContentFilterElement_DataTypeVersion'
ObjectIdNames[7931] = 'OpcUa_BinarySchema_ContentFilterElement_DictionaryFragment'
ObjectIdNames[7932] = 'OpcUa_BinarySchema_ContentFilter'
ObjectIdNames[7933] = 'OpcUa_BinarySchema_ContentFilter_DataTypeVersion'
ObjectIdNames[7934] = 'OpcUa_BinarySchema_ContentFilter_DictionaryFragment'
ObjectIdNames[7935] = 'OpcUa_BinarySchema_FilterOperand'
ObjectIdNames[7936] = 'OpcUa_BinarySchema_FilterOperand_DataTypeVersion'
ObjectIdNames[7937] = 'OpcUa_BinarySchema_FilterOperand_DictionaryFragment'
ObjectIdNames[7938] = 'OpcUa_BinarySchema_ElementOperand'
ObjectIdNames[7939] = 'OpcUa_BinarySchema_ElementOperand_DataTypeVersion'
ObjectIdNames[7940] = 'OpcUa_BinarySchema_ElementOperand_DictionaryFragment'
ObjectIdNames[7941] = 'OpcUa_BinarySchema_LiteralOperand'
ObjectIdNames[7942] = 'OpcUa_BinarySchema_LiteralOperand_DataTypeVersion'
ObjectIdNames[7943] = 'OpcUa_BinarySchema_LiteralOperand_DictionaryFragment'
ObjectIdNames[7944] = 'OpcUa_BinarySchema_AttributeOperand'
ObjectIdNames[7945] = 'OpcUa_BinarySchema_AttributeOperand_DataTypeVersion'
ObjectIdNames[7946] = 'OpcUa_BinarySchema_AttributeOperand_DictionaryFragment'
ObjectIdNames[7947] = 'OpcUa_BinarySchema_SimpleAttributeOperand'
ObjectIdNames[7948] = 'OpcUa_BinarySchema_SimpleAttributeOperand_DataTypeVersion'
ObjectIdNames[7949] = 'OpcUa_BinarySchema_SimpleAttributeOperand_DictionaryFragment'
ObjectIdNames[8004] = 'OpcUa_BinarySchema_HistoryEvent'
ObjectIdNames[8005] = 'OpcUa_BinarySchema_HistoryEvent_DataTypeVersion'
ObjectIdNames[8006] = 'OpcUa_BinarySchema_HistoryEvent_DictionaryFragment'
ObjectIdNames[8067] = 'OpcUa_BinarySchema_MonitoringFilter'
ObjectIdNames[8068] = 'OpcUa_BinarySchema_MonitoringFilter_DataTypeVersion'
ObjectIdNames[8069] = 'OpcUa_BinarySchema_MonitoringFilter_DictionaryFragment'
ObjectIdNames[8073] = 'OpcUa_BinarySchema_EventFilter'
ObjectIdNames[8074] = 'OpcUa_BinarySchema_EventFilter_DataTypeVersion'
ObjectIdNames[8075] = 'OpcUa_BinarySchema_EventFilter_DictionaryFragment'
ObjectIdNames[8076] = 'OpcUa_BinarySchema_AggregateConfiguration'
ObjectIdNames[8077] = 'OpcUa_BinarySchema_AggregateConfiguration_DataTypeVersion'
ObjectIdNames[8078] = 'OpcUa_BinarySchema_AggregateConfiguration_DictionaryFragment'
ObjectIdNames[8172] = 'OpcUa_BinarySchema_HistoryEventFieldList'
ObjectIdNames[8173] = 'OpcUa_BinarySchema_HistoryEventFieldList_DataTypeVersion'
ObjectIdNames[8174] = 'OpcUa_BinarySchema_HistoryEventFieldList_DictionaryFragment'
ObjectIdNames[8208] = 'OpcUa_BinarySchema_RedundantServerDataType'
ObjectIdNames[8209] = 'OpcUa_BinarySchema_RedundantServerDataType_DataTypeVersion'
ObjectIdNames[8210] = 'OpcUa_BinarySchema_RedundantServerDataType_DictionaryFragment'
ObjectIdNames[8211] = 'OpcUa_BinarySchema_SamplingIntervalDiagnosticsDataType'
ObjectIdNames[8212] = 'OpcUa_BinarySchema_SamplingIntervalDiagnosticsDataType_DataTypeVersion'
ObjectIdNames[8213] = 'OpcUa_BinarySchema_SamplingIntervalDiagnosticsDataType_DictionaryFragment'
ObjectIdNames[8214] = 'OpcUa_BinarySchema_ServerDiagnosticsSummaryDataType'
ObjectIdNames[8215] = 'OpcUa_BinarySchema_ServerDiagnosticsSummaryDataType_DataTypeVersion'
ObjectIdNames[8216] = 'OpcUa_BinarySchema_ServerDiagnosticsSummaryDataType_DictionaryFragment'
ObjectIdNames[8217] = 'OpcUa_BinarySchema_ServerStatusDataType'
ObjectIdNames[8218] = 'OpcUa_BinarySchema_ServerStatusDataType_DataTypeVersion'
ObjectIdNames[8219] = 'OpcUa_BinarySchema_ServerStatusDataType_DictionaryFragment'
ObjectIdNames[8220] = 'OpcUa_BinarySchema_SessionDiagnosticsDataType'
ObjectIdNames[8221] = 'OpcUa_BinarySchema_SessionDiagnosticsDataType_DataTypeVersion'
ObjectIdNames[8222] = 'OpcUa_BinarySchema_SessionDiagnosticsDataType_DictionaryFragment'
ObjectIdNames[8223] = 'OpcUa_BinarySchema_SessionSecurityDiagnosticsDataType'
ObjectIdNames[8224] = 'OpcUa_BinarySchema_SessionSecurityDiagnosticsDataType_DataTypeVersion'
ObjectIdNames[8225] = 'OpcUa_BinarySchema_SessionSecurityDiagnosticsDataType_DictionaryFragment'
ObjectIdNames[8226] = 'OpcUa_BinarySchema_ServiceCounterDataType'
ObjectIdNames[8227] = 'OpcUa_BinarySchema_ServiceCounterDataType_DataTypeVersion'
ObjectIdNames[8228] = 'OpcUa_BinarySchema_ServiceCounterDataType_DictionaryFragment'
ObjectIdNames[8229] = 'OpcUa_BinarySchema_SubscriptionDiagnosticsDataType'
ObjectIdNames[8230] = 'OpcUa_BinarySchema_SubscriptionDiagnosticsDataType_DataTypeVersion'
ObjectIdNames[8231] = 'OpcUa_BinarySchema_SubscriptionDiagnosticsDataType_DictionaryFragment'
ObjectIdNames[8232] = 'OpcUa_BinarySchema_ModelChangeStructureDataType'
ObjectIdNames[8233] = 'OpcUa_BinarySchema_ModelChangeStructureDataType_DataTypeVersion'
ObjectIdNames[8234] = 'OpcUa_BinarySchema_ModelChangeStructureDataType_DictionaryFragment'
ObjectIdNames[8235] = 'OpcUa_BinarySchema_SemanticChangeStructureDataType'
ObjectIdNames[8236] = 'OpcUa_BinarySchema_SemanticChangeStructureDataType_DataTypeVersion'
ObjectIdNames[8237] = 'OpcUa_BinarySchema_SemanticChangeStructureDataType_DictionaryFragment'
ObjectIdNames[8238] = 'OpcUa_BinarySchema_Range'
ObjectIdNames[8239] = 'OpcUa_BinarySchema_Range_DataTypeVersion'
ObjectIdNames[8240] = 'OpcUa_BinarySchema_Range_DictionaryFragment'
ObjectIdNames[8241] = 'OpcUa_BinarySchema_EUInformation'
ObjectIdNames[8242] = 'OpcUa_BinarySchema_EUInformation_DataTypeVersion'
ObjectIdNames[8243] = 'OpcUa_BinarySchema_EUInformation_DictionaryFragment'
ObjectIdNames[8244] = 'OpcUa_BinarySchema_Annotation'
ObjectIdNames[8245] = 'OpcUa_BinarySchema_Annotation_DataTypeVersion'
ObjectIdNames[8246] = 'OpcUa_BinarySchema_Annotation_DictionaryFragment'
ObjectIdNames[8247] = 'OpcUa_BinarySchema_ProgramDiagnosticDataType'
ObjectIdNames[8248] = 'OpcUa_BinarySchema_ProgramDiagnosticDataType_DataTypeVersion'
ObjectIdNames[8249] = 'OpcUa_BinarySchema_ProgramDiagnosticDataType_DictionaryFragment'
ObjectIdNames[8251] = 'EnumValueType_Encoding_DefaultBinary'
ObjectIdNames[8252] = 'OpcUa_XmlSchema'
ObjectIdNames[8253] = 'OpcUa_XmlSchema_DataTypeVersion'
ObjectIdNames[8254] = 'OpcUa_XmlSchema_NamespaceUri'
ObjectIdNames[8285] = 'OpcUa_XmlSchema_Argument'
ObjectIdNames[8286] = 'OpcUa_XmlSchema_Argument_DataTypeVersion'
ObjectIdNames[8287] = 'OpcUa_XmlSchema_Argument_DictionaryFragment'
ObjectIdNames[8291] = 'OpcUa_XmlSchema_EnumValueType'
ObjectIdNames[8292] = 'OpcUa_XmlSchema_EnumValueType_DataTypeVersion'
ObjectIdNames[8293] = 'OpcUa_XmlSchema_EnumValueType_DictionaryFragment'
ObjectIdNames[8294] = 'OpcUa_XmlSchema_StatusResult'
ObjectIdNames[8295] = 'OpcUa_XmlSchema_StatusResult_DataTypeVersion'
ObjectIdNames[8296] = 'OpcUa_XmlSchema_StatusResult_DictionaryFragment'
ObjectIdNames[8297] = 'OpcUa_XmlSchema_UserTokenPolicy'
ObjectIdNames[8298] = 'OpcUa_XmlSchema_UserTokenPolicy_DataTypeVersion'
ObjectIdNames[8299] = 'OpcUa_XmlSchema_UserTokenPolicy_DictionaryFragment'
ObjectIdNames[8300] = 'OpcUa_XmlSchema_ApplicationDescription'
ObjectIdNames[8301] = 'OpcUa_XmlSchema_ApplicationDescription_DataTypeVersion'
ObjectIdNames[8302] = 'OpcUa_XmlSchema_ApplicationDescription_DictionaryFragment'
ObjectIdNames[8303] = 'OpcUa_XmlSchema_EndpointDescription'
ObjectIdNames[8304] = 'OpcUa_XmlSchema_EndpointDescription_DataTypeVersion'
ObjectIdNames[8305] = 'OpcUa_XmlSchema_EndpointDescription_DictionaryFragment'
ObjectIdNames[8306] = 'OpcUa_XmlSchema_UserIdentityToken'
ObjectIdNames[8307] = 'OpcUa_XmlSchema_UserIdentityToken_DataTypeVersion'
ObjectIdNames[8308] = 'OpcUa_XmlSchema_UserIdentityToken_DictionaryFragment'
ObjectIdNames[8309] = 'OpcUa_XmlSchema_AnonymousIdentityToken'
ObjectIdNames[8310] = 'OpcUa_XmlSchema_AnonymousIdentityToken_DataTypeVersion'
ObjectIdNames[8311] = 'OpcUa_XmlSchema_AnonymousIdentityToken_DictionaryFragment'
ObjectIdNames[8312] = 'OpcUa_XmlSchema_UserNameIdentityToken'
ObjectIdNames[8313] = 'OpcUa_XmlSchema_UserNameIdentityToken_DataTypeVersion'
ObjectIdNames[8314] = 'OpcUa_XmlSchema_UserNameIdentityToken_DictionaryFragment'
ObjectIdNames[8315] = 'OpcUa_XmlSchema_X509IdentityToken'
ObjectIdNames[8316] = 'OpcUa_XmlSchema_X509IdentityToken_DataTypeVersion'
ObjectIdNames[8317] = 'OpcUa_XmlSchema_X509IdentityToken_DictionaryFragment'
ObjectIdNames[8318] = 'OpcUa_XmlSchema_IssuedIdentityToken'
ObjectIdNames[8319] = 'OpcUa_XmlSchema_IssuedIdentityToken_DataTypeVersion'
ObjectIdNames[8320] = 'OpcUa_XmlSchema_IssuedIdentityToken_DictionaryFragment'
ObjectIdNames[8321] = 'OpcUa_XmlSchema_EndpointConfiguration'
ObjectIdNames[8322] = 'OpcUa_XmlSchema_EndpointConfiguration_DataTypeVersion'
ObjectIdNames[8323] = 'OpcUa_XmlSchema_EndpointConfiguration_DictionaryFragment'
ObjectIdNames[8324] = 'OpcUa_XmlSchema_SupportedProfile'
ObjectIdNames[8325] = 'OpcUa_XmlSchema_SupportedProfile_DataTypeVersion'
ObjectIdNames[8326] = 'OpcUa_XmlSchema_SupportedProfile_DictionaryFragment'
ObjectIdNames[8327] = 'OpcUa_XmlSchema_BuildInfo'
ObjectIdNames[8328] = 'OpcUa_XmlSchema_BuildInfo_DataTypeVersion'
ObjectIdNames[8329] = 'OpcUa_XmlSchema_BuildInfo_DictionaryFragment'
ObjectIdNames[8330] = 'OpcUa_XmlSchema_SoftwareCertificate'
ObjectIdNames[8331] = 'OpcUa_XmlSchema_SoftwareCertificate_DataTypeVersion'
ObjectIdNames[8332] = 'OpcUa_XmlSchema_SoftwareCertificate_DictionaryFragment'
ObjectIdNames[8333] = 'OpcUa_XmlSchema_SignedSoftwareCertificate'
ObjectIdNames[8334] = 'OpcUa_XmlSchema_SignedSoftwareCertificate_DataTypeVersion'
ObjectIdNames[8335] = 'OpcUa_XmlSchema_SignedSoftwareCertificate_DictionaryFragment'
ObjectIdNames[8363] = 'OpcUa_XmlSchema_AddNodesItem'
ObjectIdNames[8364] = 'OpcUa_XmlSchema_AddNodesItem_DataTypeVersion'
ObjectIdNames[8365] = 'OpcUa_XmlSchema_AddNodesItem_DictionaryFragment'
ObjectIdNames[8366] = 'OpcUa_XmlSchema_AddReferencesItem'
ObjectIdNames[8367] = 'OpcUa_XmlSchema_AddReferencesItem_DataTypeVersion'
ObjectIdNames[8368] = 'OpcUa_XmlSchema_AddReferencesItem_DictionaryFragment'
ObjectIdNames[8369] = 'OpcUa_XmlSchema_DeleteNodesItem'
ObjectIdNames[8370] = 'OpcUa_XmlSchema_DeleteNodesItem_DataTypeVersion'
ObjectIdNames[8371] = 'OpcUa_XmlSchema_DeleteNodesItem_DictionaryFragment'
ObjectIdNames[8372] = 'OpcUa_XmlSchema_DeleteReferencesItem'
ObjectIdNames[8373] = 'OpcUa_XmlSchema_DeleteReferencesItem_DataTypeVersion'
ObjectIdNames[8374] = 'OpcUa_XmlSchema_DeleteReferencesItem_DictionaryFragment'
ObjectIdNames[8417] = 'OpcUa_XmlSchema_RegisteredServer'
ObjectIdNames[8418] = 'OpcUa_XmlSchema_RegisteredServer_DataTypeVersion'
ObjectIdNames[8419] = 'OpcUa_XmlSchema_RegisteredServer_DictionaryFragment'
ObjectIdNames[8564] = 'OpcUa_XmlSchema_ContentFilterElement'
ObjectIdNames[8565] = 'OpcUa_XmlSchema_ContentFilterElement_DataTypeVersion'
ObjectIdNames[8566] = 'OpcUa_XmlSchema_ContentFilterElement_DictionaryFragment'
ObjectIdNames[8567] = 'OpcUa_XmlSchema_ContentFilter'
ObjectIdNames[8568] = 'OpcUa_XmlSchema_ContentFilter_DataTypeVersion'
ObjectIdNames[8569] = 'OpcUa_XmlSchema_ContentFilter_DictionaryFragment'
ObjectIdNames[8570] = 'OpcUa_XmlSchema_FilterOperand'
ObjectIdNames[8571] = 'OpcUa_XmlSchema_FilterOperand_DataTypeVersion'
ObjectIdNames[8572] = 'OpcUa_XmlSchema_FilterOperand_DictionaryFragment'
ObjectIdNames[8573] = 'OpcUa_XmlSchema_ElementOperand'
ObjectIdNames[8574] = 'OpcUa_XmlSchema_ElementOperand_DataTypeVersion'
ObjectIdNames[8575] = 'OpcUa_XmlSchema_ElementOperand_DictionaryFragment'
ObjectIdNames[8576] = 'OpcUa_XmlSchema_LiteralOperand'
ObjectIdNames[8577] = 'OpcUa_XmlSchema_LiteralOperand_DataTypeVersion'
ObjectIdNames[8578] = 'OpcUa_XmlSchema_LiteralOperand_DictionaryFragment'
ObjectIdNames[8579] = 'OpcUa_XmlSchema_AttributeOperand'
ObjectIdNames[8580] = 'OpcUa_XmlSchema_AttributeOperand_DataTypeVersion'
ObjectIdNames[8581] = 'OpcUa_XmlSchema_AttributeOperand_DictionaryFragment'
ObjectIdNames[8582] = 'OpcUa_XmlSchema_SimpleAttributeOperand'
ObjectIdNames[8583] = 'OpcUa_XmlSchema_SimpleAttributeOperand_DataTypeVersion'
ObjectIdNames[8584] = 'OpcUa_XmlSchema_SimpleAttributeOperand_DictionaryFragment'
ObjectIdNames[8639] = 'OpcUa_XmlSchema_HistoryEvent'
ObjectIdNames[8640] = 'OpcUa_XmlSchema_HistoryEvent_DataTypeVersion'
ObjectIdNames[8641] = 'OpcUa_XmlSchema_HistoryEvent_DictionaryFragment'
ObjectIdNames[8702] = 'OpcUa_XmlSchema_MonitoringFilter'
ObjectIdNames[8703] = 'OpcUa_XmlSchema_MonitoringFilter_DataTypeVersion'
ObjectIdNames[8704] = 'OpcUa_XmlSchema_MonitoringFilter_DictionaryFragment'
ObjectIdNames[8708] = 'OpcUa_XmlSchema_EventFilter'
ObjectIdNames[8709] = 'OpcUa_XmlSchema_EventFilter_DataTypeVersion'
ObjectIdNames[8710] = 'OpcUa_XmlSchema_EventFilter_DictionaryFragment'
ObjectIdNames[8711] = 'OpcUa_XmlSchema_AggregateConfiguration'
ObjectIdNames[8712] = 'OpcUa_XmlSchema_AggregateConfiguration_DataTypeVersion'
ObjectIdNames[8713] = 'OpcUa_XmlSchema_AggregateConfiguration_DictionaryFragment'
ObjectIdNames[8807] = 'OpcUa_XmlSchema_HistoryEventFieldList'
ObjectIdNames[8808] = 'OpcUa_XmlSchema_HistoryEventFieldList_DataTypeVersion'
ObjectIdNames[8809] = 'OpcUa_XmlSchema_HistoryEventFieldList_DictionaryFragment'
ObjectIdNames[8843] = 'OpcUa_XmlSchema_RedundantServerDataType'
ObjectIdNames[8844] = 'OpcUa_XmlSchema_RedundantServerDataType_DataTypeVersion'
ObjectIdNames[8845] = 'OpcUa_XmlSchema_RedundantServerDataType_DictionaryFragment'
ObjectIdNames[8846] = 'OpcUa_XmlSchema_SamplingIntervalDiagnosticsDataType'
ObjectIdNames[8847] = 'OpcUa_XmlSchema_SamplingIntervalDiagnosticsDataType_DataTypeVersion'
ObjectIdNames[8848] = 'OpcUa_XmlSchema_SamplingIntervalDiagnosticsDataType_DictionaryFragment'
ObjectIdNames[8849] = 'OpcUa_XmlSchema_ServerDiagnosticsSummaryDataType'
ObjectIdNames[8850] = 'OpcUa_XmlSchema_ServerDiagnosticsSummaryDataType_DataTypeVersion'
ObjectIdNames[8851] = 'OpcUa_XmlSchema_ServerDiagnosticsSummaryDataType_DictionaryFragment'
ObjectIdNames[8852] = 'OpcUa_XmlSchema_ServerStatusDataType'
ObjectIdNames[8853] = 'OpcUa_XmlSchema_ServerStatusDataType_DataTypeVersion'
ObjectIdNames[8854] = 'OpcUa_XmlSchema_ServerStatusDataType_DictionaryFragment'
ObjectIdNames[8855] = 'OpcUa_XmlSchema_SessionDiagnosticsDataType'
ObjectIdNames[8856] = 'OpcUa_XmlSchema_SessionDiagnosticsDataType_DataTypeVersion'
ObjectIdNames[8857] = 'OpcUa_XmlSchema_SessionDiagnosticsDataType_DictionaryFragment'
ObjectIdNames[8858] = 'OpcUa_XmlSchema_SessionSecurityDiagnosticsDataType'
ObjectIdNames[8859] = 'OpcUa_XmlSchema_SessionSecurityDiagnosticsDataType_DataTypeVersion'
ObjectIdNames[8860] = 'OpcUa_XmlSchema_SessionSecurityDiagnosticsDataType_DictionaryFragment'
ObjectIdNames[8861] = 'OpcUa_XmlSchema_ServiceCounterDataType'
ObjectIdNames[8862] = 'OpcUa_XmlSchema_ServiceCounterDataType_DataTypeVersion'
ObjectIdNames[8863] = 'OpcUa_XmlSchema_ServiceCounterDataType_DictionaryFragment'
ObjectIdNames[8864] = 'OpcUa_XmlSchema_SubscriptionDiagnosticsDataType'
ObjectIdNames[8865] = 'OpcUa_XmlSchema_SubscriptionDiagnosticsDataType_DataTypeVersion'
ObjectIdNames[8866] = 'OpcUa_XmlSchema_SubscriptionDiagnosticsDataType_DictionaryFragment'
ObjectIdNames[8867] = 'OpcUa_XmlSchema_ModelChangeStructureDataType'
ObjectIdNames[8868] = 'OpcUa_XmlSchema_ModelChangeStructureDataType_DataTypeVersion'
ObjectIdNames[8869] = 'OpcUa_XmlSchema_ModelChangeStructureDataType_DictionaryFragment'
ObjectIdNames[8870] = 'OpcUa_XmlSchema_SemanticChangeStructureDataType'
ObjectIdNames[8871] = 'OpcUa_XmlSchema_SemanticChangeStructureDataType_DataTypeVersion'
ObjectIdNames[8872] = 'OpcUa_XmlSchema_SemanticChangeStructureDataType_DictionaryFragment'
ObjectIdNames[8873] = 'OpcUa_XmlSchema_Range'
ObjectIdNames[8874] = 'OpcUa_XmlSchema_Range_DataTypeVersion'
ObjectIdNames[8875] = 'OpcUa_XmlSchema_Range_DictionaryFragment'
ObjectIdNames[8876] = 'OpcUa_XmlSchema_EUInformation'
ObjectIdNames[8877] = 'OpcUa_XmlSchema_EUInformation_DataTypeVersion'
ObjectIdNames[8878] = 'OpcUa_XmlSchema_EUInformation_DictionaryFragment'
ObjectIdNames[8879] = 'OpcUa_XmlSchema_Annotation'
ObjectIdNames[8880] = 'OpcUa_XmlSchema_Annotation_DataTypeVersion'
ObjectIdNames[8881] = 'OpcUa_XmlSchema_Annotation_DictionaryFragment'
ObjectIdNames[8882] = 'OpcUa_XmlSchema_ProgramDiagnosticDataType'
ObjectIdNames[8883] = 'OpcUa_XmlSchema_ProgramDiagnosticDataType_DataTypeVersion'
ObjectIdNames[8884] = 'OpcUa_XmlSchema_ProgramDiagnosticDataType_DictionaryFragment'
ObjectIdNames[8888] = 'SubscriptionDiagnosticsType_MaxLifetimeCount'
ObjectIdNames[8889] = 'SubscriptionDiagnosticsType_LatePublishRequestCount'
ObjectIdNames[8890] = 'SubscriptionDiagnosticsType_CurrentKeepAliveCount'
ObjectIdNames[8891] = 'SubscriptionDiagnosticsType_CurrentLifetimeCount'
ObjectIdNames[8892] = 'SubscriptionDiagnosticsType_UnacknowledgedMessageCount'
ObjectIdNames[8893] = 'SubscriptionDiagnosticsType_DiscardedMessageCount'
ObjectIdNames[8894] = 'SubscriptionDiagnosticsType_MonitoredItemCount'
ObjectIdNames[8895] = 'SubscriptionDiagnosticsType_DisabledMonitoredItemCount'
ObjectIdNames[8896] = 'SubscriptionDiagnosticsType_MonitoringQueueOverflowCount'
ObjectIdNames[8897] = 'SubscriptionDiagnosticsType_NextSequenceNumber'
ObjectIdNames[8898] = 'SessionDiagnosticsObjectType_SessionDiagnostics_TotalRequestCount'
ObjectIdNames[8900] = 'SessionDiagnosticsVariableType_TotalRequestCount'
ObjectIdNames[8902] = 'SubscriptionDiagnosticsType_EventQueueOverFlowCount'
ObjectIdNames[8912] = 'TimeZoneDataType'
ObjectIdNames[8913] = 'TimeZoneDataType_Encoding_DefaultXml'
ObjectIdNames[8914] = 'OpcUa_BinarySchema_TimeZoneDataType'
ObjectIdNames[8915] = 'OpcUa_BinarySchema_TimeZoneDataType_DataTypeVersion'
ObjectIdNames[8916] = 'OpcUa_BinarySchema_TimeZoneDataType_DictionaryFragment'
ObjectIdNames[8917] = 'TimeZoneDataType_Encoding_DefaultBinary'
ObjectIdNames[8918] = 'OpcUa_XmlSchema_TimeZoneDataType'
ObjectIdNames[8919] = 'OpcUa_XmlSchema_TimeZoneDataType_DataTypeVersion'
ObjectIdNames[8920] = 'OpcUa_XmlSchema_TimeZoneDataType_DictionaryFragment'
ObjectIdNames[8927] = 'AuditConditionRespondEventType'
ObjectIdNames[8928] = 'AuditConditionRespondEventType_EventId'
ObjectIdNames[8929] = 'AuditConditionRespondEventType_EventType'
ObjectIdNames[8930] = 'AuditConditionRespondEventType_SourceNode'
ObjectIdNames[8931] = 'AuditConditionRespondEventType_SourceName'
ObjectIdNames[8932] = 'AuditConditionRespondEventType_Time'
ObjectIdNames[8933] = 'AuditConditionRespondEventType_ReceiveTime'
ObjectIdNames[8934] = 'AuditConditionRespondEventType_LocalTime'
ObjectIdNames[8935] = 'AuditConditionRespondEventType_Message'
ObjectIdNames[8936] = 'AuditConditionRespondEventType_Severity'
ObjectIdNames[8937] = 'AuditConditionRespondEventType_ActionTimeStamp'
ObjectIdNames[8938] = 'AuditConditionRespondEventType_Status'
ObjectIdNames[8939] = 'AuditConditionRespondEventType_ServerId'
ObjectIdNames[8940] = 'AuditConditionRespondEventType_ClientAuditEntryId'
ObjectIdNames[8941] = 'AuditConditionRespondEventType_ClientUserId'
ObjectIdNames[8942] = 'AuditConditionRespondEventType_MethodId'
ObjectIdNames[8943] = 'AuditConditionRespondEventType_InputArguments'
ObjectIdNames[8944] = 'AuditConditionAcknowledgeEventType'
ObjectIdNames[8945] = 'AuditConditionAcknowledgeEventType_EventId'
ObjectIdNames[8946] = 'AuditConditionAcknowledgeEventType_EventType'
ObjectIdNames[8947] = 'AuditConditionAcknowledgeEventType_SourceNode'
ObjectIdNames[8948] = 'AuditConditionAcknowledgeEventType_SourceName'
ObjectIdNames[8949] = 'AuditConditionAcknowledgeEventType_Time'
ObjectIdNames[8950] = 'AuditConditionAcknowledgeEventType_ReceiveTime'
ObjectIdNames[8951] = 'AuditConditionAcknowledgeEventType_LocalTime'
ObjectIdNames[8952] = 'AuditConditionAcknowledgeEventType_Message'
ObjectIdNames[8953] = 'AuditConditionAcknowledgeEventType_Severity'
ObjectIdNames[8954] = 'AuditConditionAcknowledgeEventType_ActionTimeStamp'
ObjectIdNames[8955] = 'AuditConditionAcknowledgeEventType_Status'
ObjectIdNames[8956] = 'AuditConditionAcknowledgeEventType_ServerId'
ObjectIdNames[8957] = 'AuditConditionAcknowledgeEventType_ClientAuditEntryId'
ObjectIdNames[8958] = 'AuditConditionAcknowledgeEventType_ClientUserId'
ObjectIdNames[8959] = 'AuditConditionAcknowledgeEventType_MethodId'
ObjectIdNames[8960] = 'AuditConditionAcknowledgeEventType_InputArguments'
ObjectIdNames[8961] = 'AuditConditionConfirmEventType'
ObjectIdNames[8962] = 'AuditConditionConfirmEventType_EventId'
ObjectIdNames[8963] = 'AuditConditionConfirmEventType_EventType'
ObjectIdNames[8964] = 'AuditConditionConfirmEventType_SourceNode'
ObjectIdNames[8965] = 'AuditConditionConfirmEventType_SourceName'
ObjectIdNames[8966] = 'AuditConditionConfirmEventType_Time'
ObjectIdNames[8967] = 'AuditConditionConfirmEventType_ReceiveTime'
ObjectIdNames[8968] = 'AuditConditionConfirmEventType_LocalTime'
ObjectIdNames[8969] = 'AuditConditionConfirmEventType_Message'
ObjectIdNames[8970] = 'AuditConditionConfirmEventType_Severity'
ObjectIdNames[8971] = 'AuditConditionConfirmEventType_ActionTimeStamp'
ObjectIdNames[8972] = 'AuditConditionConfirmEventType_Status'
ObjectIdNames[8973] = 'AuditConditionConfirmEventType_ServerId'
ObjectIdNames[8974] = 'AuditConditionConfirmEventType_ClientAuditEntryId'
ObjectIdNames[8975] = 'AuditConditionConfirmEventType_ClientUserId'
ObjectIdNames[8976] = 'AuditConditionConfirmEventType_MethodId'
ObjectIdNames[8977] = 'AuditConditionConfirmEventType_InputArguments'
ObjectIdNames[8995] = 'TwoStateVariableType'
ObjectIdNames[8996] = 'TwoStateVariableType_Id'
ObjectIdNames[8997] = 'TwoStateVariableType_Name'
ObjectIdNames[8998] = 'TwoStateVariableType_Number'
ObjectIdNames[8999] = 'TwoStateVariableType_EffectiveDisplayName'
ObjectIdNames[9000] = 'TwoStateVariableType_TransitionTime'
ObjectIdNames[9001] = 'TwoStateVariableType_EffectiveTransitionTime'
ObjectIdNames[9002] = 'ConditionVariableType'
ObjectIdNames[9003] = 'ConditionVariableType_SourceTimestamp'
ObjectIdNames[9004] = 'HasTrueSubState'
ObjectIdNames[9005] = 'HasFalseSubState'
ObjectIdNames[9006] = 'HasCondition'
ObjectIdNames[9007] = 'ConditionRefreshMethodType'
ObjectIdNames[9008] = 'ConditionRefreshMethodType_InputArguments'
ObjectIdNames[9009] = 'ConditionType_ConditionName'
ObjectIdNames[9010] = 'ConditionType_BranchId'
ObjectIdNames[9011] = 'ConditionType_EnabledState'
ObjectIdNames[9012] = 'ConditionType_EnabledState_Id'
ObjectIdNames[9013] = 'ConditionType_EnabledState_Name'
ObjectIdNames[9014] = 'ConditionType_EnabledState_Number'
ObjectIdNames[9015] = 'ConditionType_EnabledState_EffectiveDisplayName'
ObjectIdNames[9016] = 'ConditionType_EnabledState_TransitionTime'
ObjectIdNames[9017] = 'ConditionType_EnabledState_EffectiveTransitionTime'
ObjectIdNames[9018] = 'ConditionType_EnabledState_TrueState'
ObjectIdNames[9019] = 'ConditionType_EnabledState_FalseState'
ObjectIdNames[9020] = 'ConditionType_Quality'
ObjectIdNames[9021] = 'ConditionType_Quality_SourceTimestamp'
ObjectIdNames[9022] = 'ConditionType_LastSeverity'
ObjectIdNames[9023] = 'ConditionType_LastSeverity_SourceTimestamp'
ObjectIdNames[9024] = 'ConditionType_Comment'
ObjectIdNames[9025] = 'ConditionType_Comment_SourceTimestamp'
ObjectIdNames[9026] = 'ConditionType_ClientUserId'
ObjectIdNames[9027] = 'ConditionType_Enable'
ObjectIdNames[9028] = 'ConditionType_Disable'
ObjectIdNames[9029] = 'ConditionType_AddComment'
ObjectIdNames[9030] = 'ConditionType_AddComment_InputArguments'
ObjectIdNames[9031] = 'DialogResponseMethodType'
ObjectIdNames[9032] = 'DialogResponseMethodType_InputArguments'
ObjectIdNames[9033] = 'DialogConditionType_ConditionName'
ObjectIdNames[9034] = 'DialogConditionType_BranchId'
ObjectIdNames[9035] = 'DialogConditionType_EnabledState'
ObjectIdNames[9036] = 'DialogConditionType_EnabledState_Id'
ObjectIdNames[9037] = 'DialogConditionType_EnabledState_Name'
ObjectIdNames[9038] = 'DialogConditionType_EnabledState_Number'
ObjectIdNames[9039] = 'DialogConditionType_EnabledState_EffectiveDisplayName'
ObjectIdNames[9040] = 'DialogConditionType_EnabledState_TransitionTime'
ObjectIdNames[9041] = 'DialogConditionType_EnabledState_EffectiveTransitionTime'
ObjectIdNames[9042] = 'DialogConditionType_EnabledState_TrueState'
ObjectIdNames[9043] = 'DialogConditionType_EnabledState_FalseState'
ObjectIdNames[9044] = 'DialogConditionType_Quality'
ObjectIdNames[9045] = 'DialogConditionType_Quality_SourceTimestamp'
ObjectIdNames[9046] = 'DialogConditionType_LastSeverity'
ObjectIdNames[9047] = 'DialogConditionType_LastSeverity_SourceTimestamp'
ObjectIdNames[9048] = 'DialogConditionType_Comment'
ObjectIdNames[9049] = 'DialogConditionType_Comment_SourceTimestamp'
ObjectIdNames[9050] = 'DialogConditionType_ClientUserId'
ObjectIdNames[9051] = 'DialogConditionType_Enable'
ObjectIdNames[9052] = 'DialogConditionType_Disable'
ObjectIdNames[9053] = 'DialogConditionType_AddComment'
ObjectIdNames[9054] = 'DialogConditionType_AddComment_InputArguments'
ObjectIdNames[9055] = 'DialogConditionType_DialogState'
ObjectIdNames[9056] = 'DialogConditionType_DialogState_Id'
ObjectIdNames[9057] = 'DialogConditionType_DialogState_Name'
ObjectIdNames[9058] = 'DialogConditionType_DialogState_Number'
ObjectIdNames[9059] = 'DialogConditionType_DialogState_EffectiveDisplayName'
ObjectIdNames[9060] = 'DialogConditionType_DialogState_TransitionTime'
ObjectIdNames[9061] = 'DialogConditionType_DialogState_EffectiveTransitionTime'
ObjectIdNames[9062] = 'DialogConditionType_DialogState_TrueState'
ObjectIdNames[9063] = 'DialogConditionType_DialogState_FalseState'
ObjectIdNames[9064] = 'DialogConditionType_ResponseOptionSet'
ObjectIdNames[9065] = 'DialogConditionType_DefaultResponse'
ObjectIdNames[9066] = 'DialogConditionType_OkResponse'
ObjectIdNames[9067] = 'DialogConditionType_CancelResponse'
ObjectIdNames[9068] = 'DialogConditionType_LastResponse'
ObjectIdNames[9069] = 'DialogConditionType_Respond'
ObjectIdNames[9070] = 'DialogConditionType_Respond_InputArguments'
ObjectIdNames[9071] = 'AcknowledgeableConditionType_ConditionName'
ObjectIdNames[9072] = 'AcknowledgeableConditionType_BranchId'
ObjectIdNames[9073] = 'AcknowledgeableConditionType_EnabledState'
ObjectIdNames[9074] = 'AcknowledgeableConditionType_EnabledState_Id'
ObjectIdNames[9075] = 'AcknowledgeableConditionType_EnabledState_Name'
ObjectIdNames[9076] = 'AcknowledgeableConditionType_EnabledState_Number'
ObjectIdNames[9077] = 'AcknowledgeableConditionType_EnabledState_EffectiveDisplayName'
ObjectIdNames[9078] = 'AcknowledgeableConditionType_EnabledState_TransitionTime'
ObjectIdNames[9079] = 'AcknowledgeableConditionType_EnabledState_EffectiveTransitionTime'
ObjectIdNames[9080] = 'AcknowledgeableConditionType_EnabledState_TrueState'
ObjectIdNames[9081] = 'AcknowledgeableConditionType_EnabledState_FalseState'
ObjectIdNames[9082] = 'AcknowledgeableConditionType_Quality'
ObjectIdNames[9083] = 'AcknowledgeableConditionType_Quality_SourceTimestamp'
ObjectIdNames[9084] = 'AcknowledgeableConditionType_LastSeverity'
ObjectIdNames[9085] = 'AcknowledgeableConditionType_LastSeverity_SourceTimestamp'
ObjectIdNames[9086] = 'AcknowledgeableConditionType_Comment'
ObjectIdNames[9087] = 'AcknowledgeableConditionType_Comment_SourceTimestamp'
ObjectIdNames[9088] = 'AcknowledgeableConditionType_ClientUserId'
ObjectIdNames[9089] = 'AcknowledgeableConditionType_Enable'
ObjectIdNames[9090] = 'AcknowledgeableConditionType_Disable'
ObjectIdNames[9091] = 'AcknowledgeableConditionType_AddComment'
ObjectIdNames[9092] = 'AcknowledgeableConditionType_AddComment_InputArguments'
ObjectIdNames[9093] = 'AcknowledgeableConditionType_AckedState'
ObjectIdNames[9094] = 'AcknowledgeableConditionType_AckedState_Id'
ObjectIdNames[9095] = 'AcknowledgeableConditionType_AckedState_Name'
ObjectIdNames[9096] = 'AcknowledgeableConditionType_AckedState_Number'
ObjectIdNames[9097] = 'AcknowledgeableConditionType_AckedState_EffectiveDisplayName'
ObjectIdNames[9098] = 'AcknowledgeableConditionType_AckedState_TransitionTime'
ObjectIdNames[9099] = 'AcknowledgeableConditionType_AckedState_EffectiveTransitionTime'
ObjectIdNames[9100] = 'AcknowledgeableConditionType_AckedState_TrueState'
ObjectIdNames[9101] = 'AcknowledgeableConditionType_AckedState_FalseState'
ObjectIdNames[9102] = 'AcknowledgeableConditionType_ConfirmedState'
ObjectIdNames[9103] = 'AcknowledgeableConditionType_ConfirmedState_Id'
ObjectIdNames[9104] = 'AcknowledgeableConditionType_ConfirmedState_Name'
ObjectIdNames[9105] = 'AcknowledgeableConditionType_ConfirmedState_Number'
ObjectIdNames[9106] = 'AcknowledgeableConditionType_ConfirmedState_EffectiveDisplayName'
ObjectIdNames[9107] = 'AcknowledgeableConditionType_ConfirmedState_TransitionTime'
ObjectIdNames[9108] = 'AcknowledgeableConditionType_ConfirmedState_EffectiveTransitionTime'
ObjectIdNames[9109] = 'AcknowledgeableConditionType_ConfirmedState_TrueState'
ObjectIdNames[9110] = 'AcknowledgeableConditionType_ConfirmedState_FalseState'
ObjectIdNames[9111] = 'AcknowledgeableConditionType_Acknowledge'
ObjectIdNames[9112] = 'AcknowledgeableConditionType_Acknowledge_InputArguments'
ObjectIdNames[9113] = 'AcknowledgeableConditionType_Confirm'
ObjectIdNames[9114] = 'AcknowledgeableConditionType_Confirm_InputArguments'
ObjectIdNames[9115] = 'ShelvedStateMachineType_UnshelveTime'
ObjectIdNames[9116] = 'AlarmConditionType_ConditionName'
ObjectIdNames[9117] = 'AlarmConditionType_BranchId'
ObjectIdNames[9118] = 'AlarmConditionType_EnabledState'
ObjectIdNames[9119] = 'AlarmConditionType_EnabledState_Id'
ObjectIdNames[9120] = 'AlarmConditionType_EnabledState_Name'
ObjectIdNames[9121] = 'AlarmConditionType_EnabledState_Number'
ObjectIdNames[9122] = 'AlarmConditionType_EnabledState_EffectiveDisplayName'
ObjectIdNames[9123] = 'AlarmConditionType_EnabledState_TransitionTime'
ObjectIdNames[9124] = 'AlarmConditionType_EnabledState_EffectiveTransitionTime'
ObjectIdNames[9125] = 'AlarmConditionType_EnabledState_TrueState'
ObjectIdNames[9126] = 'AlarmConditionType_EnabledState_FalseState'
ObjectIdNames[9127] = 'AlarmConditionType_Quality'
ObjectIdNames[9128] = 'AlarmConditionType_Quality_SourceTimestamp'
ObjectIdNames[9129] = 'AlarmConditionType_LastSeverity'
ObjectIdNames[9130] = 'AlarmConditionType_LastSeverity_SourceTimestamp'
ObjectIdNames[9131] = 'AlarmConditionType_Comment'
ObjectIdNames[9132] = 'AlarmConditionType_Comment_SourceTimestamp'
ObjectIdNames[9133] = 'AlarmConditionType_ClientUserId'
ObjectIdNames[9134] = 'AlarmConditionType_Enable'
ObjectIdNames[9135] = 'AlarmConditionType_Disable'
ObjectIdNames[9136] = 'AlarmConditionType_AddComment'
ObjectIdNames[9137] = 'AlarmConditionType_AddComment_InputArguments'
ObjectIdNames[9138] = 'AlarmConditionType_AckedState'
ObjectIdNames[9139] = 'AlarmConditionType_AckedState_Id'
ObjectIdNames[9140] = 'AlarmConditionType_AckedState_Name'
ObjectIdNames[9141] = 'AlarmConditionType_AckedState_Number'
ObjectIdNames[9142] = 'AlarmConditionType_AckedState_EffectiveDisplayName'
ObjectIdNames[9143] = 'AlarmConditionType_AckedState_TransitionTime'
ObjectIdNames[9144] = 'AlarmConditionType_AckedState_EffectiveTransitionTime'
ObjectIdNames[9145] = 'AlarmConditionType_AckedState_TrueState'
ObjectIdNames[9146] = 'AlarmConditionType_AckedState_FalseState'
ObjectIdNames[9147] = 'AlarmConditionType_ConfirmedState'
ObjectIdNames[9148] = 'AlarmConditionType_ConfirmedState_Id'
ObjectIdNames[9149] = 'AlarmConditionType_ConfirmedState_Name'
ObjectIdNames[9150] = 'AlarmConditionType_ConfirmedState_Number'
ObjectIdNames[9151] = 'AlarmConditionType_ConfirmedState_EffectiveDisplayName'
ObjectIdNames[9152] = 'AlarmConditionType_ConfirmedState_TransitionTime'
ObjectIdNames[9153] = 'AlarmConditionType_ConfirmedState_EffectiveTransitionTime'
ObjectIdNames[9154] = 'AlarmConditionType_ConfirmedState_TrueState'
ObjectIdNames[9155] = 'AlarmConditionType_ConfirmedState_FalseState'
ObjectIdNames[9156] = 'AlarmConditionType_Acknowledge'
ObjectIdNames[9157] = 'AlarmConditionType_Acknowledge_InputArguments'
ObjectIdNames[9158] = 'AlarmConditionType_Confirm'
ObjectIdNames[9159] = 'AlarmConditionType_Confirm_InputArguments'
ObjectIdNames[9160] = 'AlarmConditionType_ActiveState'
ObjectIdNames[9161] = 'AlarmConditionType_ActiveState_Id'
ObjectIdNames[9162] = 'AlarmConditionType_ActiveState_Name'
ObjectIdNames[9163] = 'AlarmConditionType_ActiveState_Number'
ObjectIdNames[9164] = 'AlarmConditionType_ActiveState_EffectiveDisplayName'
ObjectIdNames[9165] = 'AlarmConditionType_ActiveState_TransitionTime'
ObjectIdNames[9166] = 'AlarmConditionType_ActiveState_EffectiveTransitionTime'
ObjectIdNames[9167] = 'AlarmConditionType_ActiveState_TrueState'
ObjectIdNames[9168] = 'AlarmConditionType_ActiveState_FalseState'
ObjectIdNames[9169] = 'AlarmConditionType_SuppressedState'
ObjectIdNames[9170] = 'AlarmConditionType_SuppressedState_Id'
ObjectIdNames[9171] = 'AlarmConditionType_SuppressedState_Name'
ObjectIdNames[9172] = 'AlarmConditionType_SuppressedState_Number'
ObjectIdNames[9173] = 'AlarmConditionType_SuppressedState_EffectiveDisplayName'
ObjectIdNames[9174] = 'AlarmConditionType_SuppressedState_TransitionTime'
ObjectIdNames[9175] = 'AlarmConditionType_SuppressedState_EffectiveTransitionTime'
ObjectIdNames[9176] = 'AlarmConditionType_SuppressedState_TrueState'
ObjectIdNames[9177] = 'AlarmConditionType_SuppressedState_FalseState'
ObjectIdNames[9178] = 'AlarmConditionType_ShelvingState'
ObjectIdNames[9179] = 'AlarmConditionType_ShelvingState_CurrentState'
ObjectIdNames[9180] = 'AlarmConditionType_ShelvingState_CurrentState_Id'
ObjectIdNames[9181] = 'AlarmConditionType_ShelvingState_CurrentState_Name'
ObjectIdNames[9182] = 'AlarmConditionType_ShelvingState_CurrentState_Number'
ObjectIdNames[9183] = 'AlarmConditionType_ShelvingState_CurrentState_EffectiveDisplayName'
ObjectIdNames[9184] = 'AlarmConditionType_ShelvingState_LastTransition'
ObjectIdNames[9185] = 'AlarmConditionType_ShelvingState_LastTransition_Id'
ObjectIdNames[9186] = 'AlarmConditionType_ShelvingState_LastTransition_Name'
ObjectIdNames[9187] = 'AlarmConditionType_ShelvingState_LastTransition_Number'
ObjectIdNames[9188] = 'AlarmConditionType_ShelvingState_LastTransition_TransitionTime'
ObjectIdNames[9189] = 'AlarmConditionType_ShelvingState_UnshelveTime'
ObjectIdNames[9211] = 'AlarmConditionType_ShelvingState_Unshelve'
ObjectIdNames[9212] = 'AlarmConditionType_ShelvingState_OneShotShelve'
ObjectIdNames[9213] = 'AlarmConditionType_ShelvingState_TimedShelve'
ObjectIdNames[9214] = 'AlarmConditionType_ShelvingState_TimedShelve_InputArguments'
ObjectIdNames[9215] = 'AlarmConditionType_SuppressedOrShelved'
ObjectIdNames[9216] = 'AlarmConditionType_MaxTimeShelved'
ObjectIdNames[9217] = 'LimitAlarmType_ConditionName'
ObjectIdNames[9218] = 'LimitAlarmType_BranchId'
ObjectIdNames[9219] = 'LimitAlarmType_EnabledState'
ObjectIdNames[9220] = 'LimitAlarmType_EnabledState_Id'
ObjectIdNames[9221] = 'LimitAlarmType_EnabledState_Name'
ObjectIdNames[9222] = 'LimitAlarmType_EnabledState_Number'
ObjectIdNames[9223] = 'LimitAlarmType_EnabledState_EffectiveDisplayName'
ObjectIdNames[9224] = 'LimitAlarmType_EnabledState_TransitionTime'
ObjectIdNames[9225] = 'LimitAlarmType_EnabledState_EffectiveTransitionTime'
ObjectIdNames[9226] = 'LimitAlarmType_EnabledState_TrueState'
ObjectIdNames[9227] = 'LimitAlarmType_EnabledState_FalseState'
ObjectIdNames[9228] = 'LimitAlarmType_Quality'
ObjectIdNames[9229] = 'LimitAlarmType_Quality_SourceTimestamp'
ObjectIdNames[9230] = 'LimitAlarmType_LastSeverity'
ObjectIdNames[9231] = 'LimitAlarmType_LastSeverity_SourceTimestamp'
ObjectIdNames[9232] = 'LimitAlarmType_Comment'
ObjectIdNames[9233] = 'LimitAlarmType_Comment_SourceTimestamp'
ObjectIdNames[9234] = 'LimitAlarmType_ClientUserId'
ObjectIdNames[9235] = 'LimitAlarmType_Enable'
ObjectIdNames[9236] = 'LimitAlarmType_Disable'
ObjectIdNames[9237] = 'LimitAlarmType_AddComment'
ObjectIdNames[9238] = 'LimitAlarmType_AddComment_InputArguments'
ObjectIdNames[9239] = 'LimitAlarmType_AckedState'
ObjectIdNames[9240] = 'LimitAlarmType_AckedState_Id'
ObjectIdNames[9241] = 'LimitAlarmType_AckedState_Name'
ObjectIdNames[9242] = 'LimitAlarmType_AckedState_Number'
ObjectIdNames[9243] = 'LimitAlarmType_AckedState_EffectiveDisplayName'
ObjectIdNames[9244] = 'LimitAlarmType_AckedState_TransitionTime'
ObjectIdNames[9245] = 'LimitAlarmType_AckedState_EffectiveTransitionTime'
ObjectIdNames[9246] = 'LimitAlarmType_AckedState_TrueState'
ObjectIdNames[9247] = 'LimitAlarmType_AckedState_FalseState'
ObjectIdNames[9248] = 'LimitAlarmType_ConfirmedState'
ObjectIdNames[9249] = 'LimitAlarmType_ConfirmedState_Id'
ObjectIdNames[9250] = 'LimitAlarmType_ConfirmedState_Name'
ObjectIdNames[9251] = 'LimitAlarmType_ConfirmedState_Number'
ObjectIdNames[9252] = 'LimitAlarmType_ConfirmedState_EffectiveDisplayName'
ObjectIdNames[9253] = 'LimitAlarmType_ConfirmedState_TransitionTime'
ObjectIdNames[9254] = 'LimitAlarmType_ConfirmedState_EffectiveTransitionTime'
ObjectIdNames[9255] = 'LimitAlarmType_ConfirmedState_TrueState'
ObjectIdNames[9256] = 'LimitAlarmType_ConfirmedState_FalseState'
ObjectIdNames[9257] = 'LimitAlarmType_Acknowledge'
ObjectIdNames[9258] = 'LimitAlarmType_Acknowledge_InputArguments'
ObjectIdNames[9259] = 'LimitAlarmType_Confirm'
ObjectIdNames[9260] = 'LimitAlarmType_Confirm_InputArguments'
ObjectIdNames[9261] = 'LimitAlarmType_ActiveState'
ObjectIdNames[9262] = 'LimitAlarmType_ActiveState_Id'
ObjectIdNames[9263] = 'LimitAlarmType_ActiveState_Name'
ObjectIdNames[9264] = 'LimitAlarmType_ActiveState_Number'
ObjectIdNames[9265] = 'LimitAlarmType_ActiveState_EffectiveDisplayName'
ObjectIdNames[9266] = 'LimitAlarmType_ActiveState_TransitionTime'
ObjectIdNames[9267] = 'LimitAlarmType_ActiveState_EffectiveTransitionTime'
ObjectIdNames[9268] = 'LimitAlarmType_ActiveState_TrueState'
ObjectIdNames[9269] = 'LimitAlarmType_ActiveState_FalseState'
ObjectIdNames[9270] = 'LimitAlarmType_SuppressedState'
ObjectIdNames[9271] = 'LimitAlarmType_SuppressedState_Id'
ObjectIdNames[9272] = 'LimitAlarmType_SuppressedState_Name'
ObjectIdNames[9273] = 'LimitAlarmType_SuppressedState_Number'
ObjectIdNames[9274] = 'LimitAlarmType_SuppressedState_EffectiveDisplayName'
ObjectIdNames[9275] = 'LimitAlarmType_SuppressedState_TransitionTime'
ObjectIdNames[9276] = 'LimitAlarmType_SuppressedState_EffectiveTransitionTime'
ObjectIdNames[9277] = 'LimitAlarmType_SuppressedState_TrueState'
ObjectIdNames[9278] = 'LimitAlarmType_SuppressedState_FalseState'
ObjectIdNames[9279] = 'LimitAlarmType_ShelvingState'
ObjectIdNames[9280] = 'LimitAlarmType_ShelvingState_CurrentState'
ObjectIdNames[9281] = 'LimitAlarmType_ShelvingState_CurrentState_Id'
ObjectIdNames[9282] = 'LimitAlarmType_ShelvingState_CurrentState_Name'
ObjectIdNames[9283] = 'LimitAlarmType_ShelvingState_CurrentState_Number'
ObjectIdNames[9284] = 'LimitAlarmType_ShelvingState_CurrentState_EffectiveDisplayName'
ObjectIdNames[9285] = 'LimitAlarmType_ShelvingState_LastTransition'
ObjectIdNames[9286] = 'LimitAlarmType_ShelvingState_LastTransition_Id'
ObjectIdNames[9287] = 'LimitAlarmType_ShelvingState_LastTransition_Name'
ObjectIdNames[9288] = 'LimitAlarmType_ShelvingState_LastTransition_Number'
ObjectIdNames[9289] = 'LimitAlarmType_ShelvingState_LastTransition_TransitionTime'
ObjectIdNames[9290] = 'LimitAlarmType_ShelvingState_UnshelveTime'
ObjectIdNames[9312] = 'LimitAlarmType_ShelvingState_Unshelve'
ObjectIdNames[9313] = 'LimitAlarmType_ShelvingState_OneShotShelve'
ObjectIdNames[9314] = 'LimitAlarmType_ShelvingState_TimedShelve'
ObjectIdNames[9315] = 'LimitAlarmType_ShelvingState_TimedShelve_InputArguments'
ObjectIdNames[9316] = 'LimitAlarmType_SuppressedOrShelved'
ObjectIdNames[9317] = 'LimitAlarmType_MaxTimeShelved'
ObjectIdNames[9318] = 'ExclusiveLimitStateMachineType'
ObjectIdNames[9319] = 'ExclusiveLimitStateMachineType_CurrentState'
ObjectIdNames[9320] = 'ExclusiveLimitStateMachineType_CurrentState_Id'
ObjectIdNames[9321] = 'ExclusiveLimitStateMachineType_CurrentState_Name'
ObjectIdNames[9322] = 'ExclusiveLimitStateMachineType_CurrentState_Number'
ObjectIdNames[9323] = 'ExclusiveLimitStateMachineType_CurrentState_EffectiveDisplayName'
ObjectIdNames[9324] = 'ExclusiveLimitStateMachineType_LastTransition'
ObjectIdNames[9325] = 'ExclusiveLimitStateMachineType_LastTransition_Id'
ObjectIdNames[9326] = 'ExclusiveLimitStateMachineType_LastTransition_Name'
ObjectIdNames[9327] = 'ExclusiveLimitStateMachineType_LastTransition_Number'
ObjectIdNames[9328] = 'ExclusiveLimitStateMachineType_LastTransition_TransitionTime'
ObjectIdNames[9329] = 'ExclusiveLimitStateMachineType_HighHigh'
ObjectIdNames[9330] = 'ExclusiveLimitStateMachineType_HighHigh_StateNumber'
ObjectIdNames[9331] = 'ExclusiveLimitStateMachineType_High'
ObjectIdNames[9332] = 'ExclusiveLimitStateMachineType_High_StateNumber'
ObjectIdNames[9333] = 'ExclusiveLimitStateMachineType_Low'
ObjectIdNames[9334] = 'ExclusiveLimitStateMachineType_Low_StateNumber'
ObjectIdNames[9335] = 'ExclusiveLimitStateMachineType_LowLow'
ObjectIdNames[9336] = 'ExclusiveLimitStateMachineType_LowLow_StateNumber'
ObjectIdNames[9337] = 'ExclusiveLimitStateMachineType_LowLowToLow'
ObjectIdNames[9338] = 'ExclusiveLimitStateMachineType_LowToLowLow'
ObjectIdNames[9339] = 'ExclusiveLimitStateMachineType_HighHighToHigh'
ObjectIdNames[9340] = 'ExclusiveLimitStateMachineType_HighToHighHigh'
ObjectIdNames[9341] = 'ExclusiveLimitAlarmType'
ObjectIdNames[9342] = 'ExclusiveLimitAlarmType_EventId'
ObjectIdNames[9343] = 'ExclusiveLimitAlarmType_EventType'
ObjectIdNames[9344] = 'ExclusiveLimitAlarmType_SourceNode'
ObjectIdNames[9345] = 'ExclusiveLimitAlarmType_SourceName'
ObjectIdNames[9346] = 'ExclusiveLimitAlarmType_Time'
ObjectIdNames[9347] = 'ExclusiveLimitAlarmType_ReceiveTime'
ObjectIdNames[9348] = 'ExclusiveLimitAlarmType_LocalTime'
ObjectIdNames[9349] = 'ExclusiveLimitAlarmType_Message'
ObjectIdNames[9350] = 'ExclusiveLimitAlarmType_Severity'
ObjectIdNames[9351] = 'ExclusiveLimitAlarmType_ConditionName'
ObjectIdNames[9352] = 'ExclusiveLimitAlarmType_BranchId'
ObjectIdNames[9353] = 'ExclusiveLimitAlarmType_Retain'
ObjectIdNames[9354] = 'ExclusiveLimitAlarmType_EnabledState'
ObjectIdNames[9355] = 'ExclusiveLimitAlarmType_EnabledState_Id'
ObjectIdNames[9356] = 'ExclusiveLimitAlarmType_EnabledState_Name'
ObjectIdNames[9357] = 'ExclusiveLimitAlarmType_EnabledState_Number'
ObjectIdNames[9358] = 'ExclusiveLimitAlarmType_EnabledState_EffectiveDisplayName'
ObjectIdNames[9359] = 'ExclusiveLimitAlarmType_EnabledState_TransitionTime'
ObjectIdNames[9360] = 'ExclusiveLimitAlarmType_EnabledState_EffectiveTransitionTime'
ObjectIdNames[9361] = 'ExclusiveLimitAlarmType_EnabledState_TrueState'
ObjectIdNames[9362] = 'ExclusiveLimitAlarmType_EnabledState_FalseState'
ObjectIdNames[9363] = 'ExclusiveLimitAlarmType_Quality'
ObjectIdNames[9364] = 'ExclusiveLimitAlarmType_Quality_SourceTimestamp'
ObjectIdNames[9365] = 'ExclusiveLimitAlarmType_LastSeverity'
ObjectIdNames[9366] = 'ExclusiveLimitAlarmType_LastSeverity_SourceTimestamp'
ObjectIdNames[9367] = 'ExclusiveLimitAlarmType_Comment'
ObjectIdNames[9368] = 'ExclusiveLimitAlarmType_Comment_SourceTimestamp'
ObjectIdNames[9369] = 'ExclusiveLimitAlarmType_ClientUserId'
ObjectIdNames[9370] = 'ExclusiveLimitAlarmType_Enable'
ObjectIdNames[9371] = 'ExclusiveLimitAlarmType_Disable'
ObjectIdNames[9372] = 'ExclusiveLimitAlarmType_AddComment'
ObjectIdNames[9373] = 'ExclusiveLimitAlarmType_AddComment_InputArguments'
ObjectIdNames[9374] = 'ExclusiveLimitAlarmType_ConditionRefresh'
ObjectIdNames[9375] = 'ExclusiveLimitAlarmType_ConditionRefresh_InputArguments'
ObjectIdNames[9376] = 'ExclusiveLimitAlarmType_AckedState'
ObjectIdNames[9377] = 'ExclusiveLimitAlarmType_AckedState_Id'
ObjectIdNames[9378] = 'ExclusiveLimitAlarmType_AckedState_Name'
ObjectIdNames[9379] = 'ExclusiveLimitAlarmType_AckedState_Number'
ObjectIdNames[9380] = 'ExclusiveLimitAlarmType_AckedState_EffectiveDisplayName'
ObjectIdNames[9381] = 'ExclusiveLimitAlarmType_AckedState_TransitionTime'
ObjectIdNames[9382] = 'ExclusiveLimitAlarmType_AckedState_EffectiveTransitionTime'
ObjectIdNames[9383] = 'ExclusiveLimitAlarmType_AckedState_TrueState'
ObjectIdNames[9384] = 'ExclusiveLimitAlarmType_AckedState_FalseState'
ObjectIdNames[9385] = 'ExclusiveLimitAlarmType_ConfirmedState'
ObjectIdNames[9386] = 'ExclusiveLimitAlarmType_ConfirmedState_Id'
ObjectIdNames[9387] = 'ExclusiveLimitAlarmType_ConfirmedState_Name'
ObjectIdNames[9388] = 'ExclusiveLimitAlarmType_ConfirmedState_Number'
ObjectIdNames[9389] = 'ExclusiveLimitAlarmType_ConfirmedState_EffectiveDisplayName'
ObjectIdNames[9390] = 'ExclusiveLimitAlarmType_ConfirmedState_TransitionTime'
ObjectIdNames[9391] = 'ExclusiveLimitAlarmType_ConfirmedState_EffectiveTransitionTime'
ObjectIdNames[9392] = 'ExclusiveLimitAlarmType_ConfirmedState_TrueState'
ObjectIdNames[9393] = 'ExclusiveLimitAlarmType_ConfirmedState_FalseState'
ObjectIdNames[9394] = 'ExclusiveLimitAlarmType_Acknowledge'
ObjectIdNames[9395] = 'ExclusiveLimitAlarmType_Acknowledge_InputArguments'
ObjectIdNames[9396] = 'ExclusiveLimitAlarmType_Confirm'
ObjectIdNames[9397] = 'ExclusiveLimitAlarmType_Confirm_InputArguments'
ObjectIdNames[9398] = 'ExclusiveLimitAlarmType_ActiveState'
ObjectIdNames[9399] = 'ExclusiveLimitAlarmType_ActiveState_Id'
ObjectIdNames[9400] = 'ExclusiveLimitAlarmType_ActiveState_Name'
ObjectIdNames[9401] = 'ExclusiveLimitAlarmType_ActiveState_Number'
ObjectIdNames[9402] = 'ExclusiveLimitAlarmType_ActiveState_EffectiveDisplayName'
ObjectIdNames[9403] = 'ExclusiveLimitAlarmType_ActiveState_TransitionTime'
ObjectIdNames[9404] = 'ExclusiveLimitAlarmType_ActiveState_EffectiveTransitionTime'
ObjectIdNames[9405] = 'ExclusiveLimitAlarmType_ActiveState_TrueState'
ObjectIdNames[9406] = 'ExclusiveLimitAlarmType_ActiveState_FalseState'
ObjectIdNames[9407] = 'ExclusiveLimitAlarmType_SuppressedState'
ObjectIdNames[9408] = 'ExclusiveLimitAlarmType_SuppressedState_Id'
ObjectIdNames[9409] = 'ExclusiveLimitAlarmType_SuppressedState_Name'
ObjectIdNames[9410] = 'ExclusiveLimitAlarmType_SuppressedState_Number'
ObjectIdNames[9411] = 'ExclusiveLimitAlarmType_SuppressedState_EffectiveDisplayName'
ObjectIdNames[9412] = 'ExclusiveLimitAlarmType_SuppressedState_TransitionTime'
ObjectIdNames[9413] = 'ExclusiveLimitAlarmType_SuppressedState_EffectiveTransitionTime'
ObjectIdNames[9414] = 'ExclusiveLimitAlarmType_SuppressedState_TrueState'
ObjectIdNames[9415] = 'ExclusiveLimitAlarmType_SuppressedState_FalseState'
ObjectIdNames[9416] = 'ExclusiveLimitAlarmType_ShelvingState'
ObjectIdNames[9417] = 'ExclusiveLimitAlarmType_ShelvingState_CurrentState'
ObjectIdNames[9418] = 'ExclusiveLimitAlarmType_ShelvingState_CurrentState_Id'
ObjectIdNames[9419] = 'ExclusiveLimitAlarmType_ShelvingState_CurrentState_Name'
ObjectIdNames[9420] = 'ExclusiveLimitAlarmType_ShelvingState_CurrentState_Number'
ObjectIdNames[9421] = 'ExclusiveLimitAlarmType_ShelvingState_CurrentState_EffectiveDisplayName'
ObjectIdNames[9422] = 'ExclusiveLimitAlarmType_ShelvingState_LastTransition'
ObjectIdNames[9423] = 'ExclusiveLimitAlarmType_ShelvingState_LastTransition_Id'
ObjectIdNames[9424] = 'ExclusiveLimitAlarmType_ShelvingState_LastTransition_Name'
ObjectIdNames[9425] = 'ExclusiveLimitAlarmType_ShelvingState_LastTransition_Number'
ObjectIdNames[9426] = 'ExclusiveLimitAlarmType_ShelvingState_LastTransition_TransitionTime'
ObjectIdNames[9427] = 'ExclusiveLimitAlarmType_ShelvingState_UnshelveTime'
ObjectIdNames[9449] = 'ExclusiveLimitAlarmType_ShelvingState_Unshelve'
ObjectIdNames[9450] = 'ExclusiveLimitAlarmType_ShelvingState_OneShotShelve'
ObjectIdNames[9451] = 'ExclusiveLimitAlarmType_ShelvingState_TimedShelve'
ObjectIdNames[9452] = 'ExclusiveLimitAlarmType_ShelvingState_TimedShelve_InputArguments'
ObjectIdNames[9453] = 'ExclusiveLimitAlarmType_SuppressedOrShelved'
ObjectIdNames[9454] = 'ExclusiveLimitAlarmType_MaxTimeShelved'
ObjectIdNames[9455] = 'ExclusiveLimitAlarmType_LimitState'
ObjectIdNames[9456] = 'ExclusiveLimitAlarmType_LimitState_CurrentState'
ObjectIdNames[9457] = 'ExclusiveLimitAlarmType_LimitState_CurrentState_Id'
ObjectIdNames[9458] = 'ExclusiveLimitAlarmType_LimitState_CurrentState_Name'
ObjectIdNames[9459] = 'ExclusiveLimitAlarmType_LimitState_CurrentState_Number'
ObjectIdNames[9460] = 'ExclusiveLimitAlarmType_LimitState_CurrentState_EffectiveDisplayName'
ObjectIdNames[9461] = 'ExclusiveLimitAlarmType_LimitState_LastTransition'
ObjectIdNames[9462] = 'ExclusiveLimitAlarmType_LimitState_LastTransition_Id'
ObjectIdNames[9463] = 'ExclusiveLimitAlarmType_LimitState_LastTransition_Name'
ObjectIdNames[9464] = 'ExclusiveLimitAlarmType_LimitState_LastTransition_Number'
ObjectIdNames[9465] = 'ExclusiveLimitAlarmType_LimitState_LastTransition_TransitionTime'
ObjectIdNames[9478] = 'ExclusiveLimitAlarmType_HighHighLimit'
ObjectIdNames[9479] = 'ExclusiveLimitAlarmType_HighLimit'
ObjectIdNames[9480] = 'ExclusiveLimitAlarmType_LowLimit'
ObjectIdNames[9481] = 'ExclusiveLimitAlarmType_LowLowLimit'
ObjectIdNames[9482] = 'ExclusiveLevelAlarmType'
ObjectIdNames[9483] = 'ExclusiveLevelAlarmType_EventId'
ObjectIdNames[9484] = 'ExclusiveLevelAlarmType_EventType'
ObjectIdNames[9485] = 'ExclusiveLevelAlarmType_SourceNode'
ObjectIdNames[9486] = 'ExclusiveLevelAlarmType_SourceName'
ObjectIdNames[9487] = 'ExclusiveLevelAlarmType_Time'
ObjectIdNames[9488] = 'ExclusiveLevelAlarmType_ReceiveTime'
ObjectIdNames[9489] = 'ExclusiveLevelAlarmType_LocalTime'
ObjectIdNames[9490] = 'ExclusiveLevelAlarmType_Message'
ObjectIdNames[9491] = 'ExclusiveLevelAlarmType_Severity'
ObjectIdNames[9492] = 'ExclusiveLevelAlarmType_ConditionName'
ObjectIdNames[9493] = 'ExclusiveLevelAlarmType_BranchId'
ObjectIdNames[9494] = 'ExclusiveLevelAlarmType_Retain'
ObjectIdNames[9495] = 'ExclusiveLevelAlarmType_EnabledState'
ObjectIdNames[9496] = 'ExclusiveLevelAlarmType_EnabledState_Id'
ObjectIdNames[9497] = 'ExclusiveLevelAlarmType_EnabledState_Name'
ObjectIdNames[9498] = 'ExclusiveLevelAlarmType_EnabledState_Number'
ObjectIdNames[9499] = 'ExclusiveLevelAlarmType_EnabledState_EffectiveDisplayName'
ObjectIdNames[9500] = 'ExclusiveLevelAlarmType_EnabledState_TransitionTime'
ObjectIdNames[9501] = 'ExclusiveLevelAlarmType_EnabledState_EffectiveTransitionTime'
ObjectIdNames[9502] = 'ExclusiveLevelAlarmType_EnabledState_TrueState'
ObjectIdNames[9503] = 'ExclusiveLevelAlarmType_EnabledState_FalseState'
ObjectIdNames[9504] = 'ExclusiveLevelAlarmType_Quality'
ObjectIdNames[9505] = 'ExclusiveLevelAlarmType_Quality_SourceTimestamp'
ObjectIdNames[9506] = 'ExclusiveLevelAlarmType_LastSeverity'
ObjectIdNames[9507] = 'ExclusiveLevelAlarmType_LastSeverity_SourceTimestamp'
ObjectIdNames[9508] = 'ExclusiveLevelAlarmType_Comment'
ObjectIdNames[9509] = 'ExclusiveLevelAlarmType_Comment_SourceTimestamp'
ObjectIdNames[9510] = 'ExclusiveLevelAlarmType_ClientUserId'
ObjectIdNames[9511] = 'ExclusiveLevelAlarmType_Enable'
ObjectIdNames[9512] = 'ExclusiveLevelAlarmType_Disable'
ObjectIdNames[9513] = 'ExclusiveLevelAlarmType_AddComment'
ObjectIdNames[9514] = 'ExclusiveLevelAlarmType_AddComment_InputArguments'
ObjectIdNames[9515] = 'ExclusiveLevelAlarmType_ConditionRefresh'
ObjectIdNames[9516] = 'ExclusiveLevelAlarmType_ConditionRefresh_InputArguments'
ObjectIdNames[9517] = 'ExclusiveLevelAlarmType_AckedState'
ObjectIdNames[9518] = 'ExclusiveLevelAlarmType_AckedState_Id'
ObjectIdNames[9519] = 'ExclusiveLevelAlarmType_AckedState_Name'
ObjectIdNames[9520] = 'ExclusiveLevelAlarmType_AckedState_Number'
ObjectIdNames[9521] = 'ExclusiveLevelAlarmType_AckedState_EffectiveDisplayName'
ObjectIdNames[9522] = 'ExclusiveLevelAlarmType_AckedState_TransitionTime'
ObjectIdNames[9523] = 'ExclusiveLevelAlarmType_AckedState_EffectiveTransitionTime'
ObjectIdNames[9524] = 'ExclusiveLevelAlarmType_AckedState_TrueState'
ObjectIdNames[9525] = 'ExclusiveLevelAlarmType_AckedState_FalseState'
ObjectIdNames[9526] = 'ExclusiveLevelAlarmType_ConfirmedState'
ObjectIdNames[9527] = 'ExclusiveLevelAlarmType_ConfirmedState_Id'
ObjectIdNames[9528] = 'ExclusiveLevelAlarmType_ConfirmedState_Name'
ObjectIdNames[9529] = 'ExclusiveLevelAlarmType_ConfirmedState_Number'
ObjectIdNames[9530] = 'ExclusiveLevelAlarmType_ConfirmedState_EffectiveDisplayName'
ObjectIdNames[9531] = 'ExclusiveLevelAlarmType_ConfirmedState_TransitionTime'
ObjectIdNames[9532] = 'ExclusiveLevelAlarmType_ConfirmedState_EffectiveTransitionTime'
ObjectIdNames[9533] = 'ExclusiveLevelAlarmType_ConfirmedState_TrueState'
ObjectIdNames[9534] = 'ExclusiveLevelAlarmType_ConfirmedState_FalseState'
ObjectIdNames[9535] = 'ExclusiveLevelAlarmType_Acknowledge'
ObjectIdNames[9536] = 'ExclusiveLevelAlarmType_Acknowledge_InputArguments'
ObjectIdNames[9537] = 'ExclusiveLevelAlarmType_Confirm'
ObjectIdNames[9538] = 'ExclusiveLevelAlarmType_Confirm_InputArguments'
ObjectIdNames[9539] = 'ExclusiveLevelAlarmType_ActiveState'
ObjectIdNames[9540] = 'ExclusiveLevelAlarmType_ActiveState_Id'
ObjectIdNames[9541] = 'ExclusiveLevelAlarmType_ActiveState_Name'
ObjectIdNames[9542] = 'ExclusiveLevelAlarmType_ActiveState_Number'
ObjectIdNames[9543] = 'ExclusiveLevelAlarmType_ActiveState_EffectiveDisplayName'
ObjectIdNames[9544] = 'ExclusiveLevelAlarmType_ActiveState_TransitionTime'
ObjectIdNames[9545] = 'ExclusiveLevelAlarmType_ActiveState_EffectiveTransitionTime'
ObjectIdNames[9546] = 'ExclusiveLevelAlarmType_ActiveState_TrueState'
ObjectIdNames[9547] = 'ExclusiveLevelAlarmType_ActiveState_FalseState'
ObjectIdNames[9548] = 'ExclusiveLevelAlarmType_SuppressedState'
ObjectIdNames[9549] = 'ExclusiveLevelAlarmType_SuppressedState_Id'
ObjectIdNames[9550] = 'ExclusiveLevelAlarmType_SuppressedState_Name'
ObjectIdNames[9551] = 'ExclusiveLevelAlarmType_SuppressedState_Number'
ObjectIdNames[9552] = 'ExclusiveLevelAlarmType_SuppressedState_EffectiveDisplayName'
ObjectIdNames[9553] = 'ExclusiveLevelAlarmType_SuppressedState_TransitionTime'
ObjectIdNames[9554] = 'ExclusiveLevelAlarmType_SuppressedState_EffectiveTransitionTime'
ObjectIdNames[9555] = 'ExclusiveLevelAlarmType_SuppressedState_TrueState'
ObjectIdNames[9556] = 'ExclusiveLevelAlarmType_SuppressedState_FalseState'
ObjectIdNames[9557] = 'ExclusiveLevelAlarmType_ShelvingState'
ObjectIdNames[9558] = 'ExclusiveLevelAlarmType_ShelvingState_CurrentState'
ObjectIdNames[9559] = 'ExclusiveLevelAlarmType_ShelvingState_CurrentState_Id'
ObjectIdNames[9560] = 'ExclusiveLevelAlarmType_ShelvingState_CurrentState_Name'
ObjectIdNames[9561] = 'ExclusiveLevelAlarmType_ShelvingState_CurrentState_Number'
ObjectIdNames[9562] = 'ExclusiveLevelAlarmType_ShelvingState_CurrentState_EffectiveDisplayName'
ObjectIdNames[9563] = 'ExclusiveLevelAlarmType_ShelvingState_LastTransition'
ObjectIdNames[9564] = 'ExclusiveLevelAlarmType_ShelvingState_LastTransition_Id'
ObjectIdNames[9565] = 'ExclusiveLevelAlarmType_ShelvingState_LastTransition_Name'
ObjectIdNames[9566] = 'ExclusiveLevelAlarmType_ShelvingState_LastTransition_Number'
ObjectIdNames[9567] = 'ExclusiveLevelAlarmType_ShelvingState_LastTransition_TransitionTime'
ObjectIdNames[9568] = 'ExclusiveLevelAlarmType_ShelvingState_UnshelveTime'
ObjectIdNames[9590] = 'ExclusiveLevelAlarmType_ShelvingState_Unshelve'
ObjectIdNames[9591] = 'ExclusiveLevelAlarmType_ShelvingState_OneShotShelve'
ObjectIdNames[9592] = 'ExclusiveLevelAlarmType_ShelvingState_TimedShelve'
ObjectIdNames[9593] = 'ExclusiveLevelAlarmType_ShelvingState_TimedShelve_InputArguments'
ObjectIdNames[9594] = 'ExclusiveLevelAlarmType_SuppressedOrShelved'
ObjectIdNames[9595] = 'ExclusiveLevelAlarmType_MaxTimeShelved'
ObjectIdNames[9596] = 'ExclusiveLevelAlarmType_LimitState'
ObjectIdNames[9597] = 'ExclusiveLevelAlarmType_LimitState_CurrentState'
ObjectIdNames[9598] = 'ExclusiveLevelAlarmType_LimitState_CurrentState_Id'
ObjectIdNames[9599] = 'ExclusiveLevelAlarmType_LimitState_CurrentState_Name'
ObjectIdNames[9600] = 'ExclusiveLevelAlarmType_LimitState_CurrentState_Number'
ObjectIdNames[9601] = 'ExclusiveLevelAlarmType_LimitState_CurrentState_EffectiveDisplayName'
ObjectIdNames[9602] = 'ExclusiveLevelAlarmType_LimitState_LastTransition'
ObjectIdNames[9603] = 'ExclusiveLevelAlarmType_LimitState_LastTransition_Id'
ObjectIdNames[9604] = 'ExclusiveLevelAlarmType_LimitState_LastTransition_Name'
ObjectIdNames[9605] = 'ExclusiveLevelAlarmType_LimitState_LastTransition_Number'
ObjectIdNames[9606] = 'ExclusiveLevelAlarmType_LimitState_LastTransition_TransitionTime'
ObjectIdNames[9619] = 'ExclusiveLevelAlarmType_HighHighLimit'
ObjectIdNames[9620] = 'ExclusiveLevelAlarmType_HighLimit'
ObjectIdNames[9621] = 'ExclusiveLevelAlarmType_LowLimit'
ObjectIdNames[9622] = 'ExclusiveLevelAlarmType_LowLowLimit'
ObjectIdNames[9623] = 'ExclusiveRateOfChangeAlarmType'
ObjectIdNames[9624] = 'ExclusiveRateOfChangeAlarmType_EventId'
ObjectIdNames[9625] = 'ExclusiveRateOfChangeAlarmType_EventType'
ObjectIdNames[9626] = 'ExclusiveRateOfChangeAlarmType_SourceNode'
ObjectIdNames[9627] = 'ExclusiveRateOfChangeAlarmType_SourceName'
ObjectIdNames[9628] = 'ExclusiveRateOfChangeAlarmType_Time'
ObjectIdNames[9629] = 'ExclusiveRateOfChangeAlarmType_ReceiveTime'
ObjectIdNames[9630] = 'ExclusiveRateOfChangeAlarmType_LocalTime'
ObjectIdNames[9631] = 'ExclusiveRateOfChangeAlarmType_Message'
ObjectIdNames[9632] = 'ExclusiveRateOfChangeAlarmType_Severity'
ObjectIdNames[9633] = 'ExclusiveRateOfChangeAlarmType_ConditionName'
ObjectIdNames[9634] = 'ExclusiveRateOfChangeAlarmType_BranchId'
ObjectIdNames[9635] = 'ExclusiveRateOfChangeAlarmType_Retain'
ObjectIdNames[9636] = 'ExclusiveRateOfChangeAlarmType_EnabledState'
ObjectIdNames[9637] = 'ExclusiveRateOfChangeAlarmType_EnabledState_Id'
ObjectIdNames[9638] = 'ExclusiveRateOfChangeAlarmType_EnabledState_Name'
ObjectIdNames[9639] = 'ExclusiveRateOfChangeAlarmType_EnabledState_Number'
ObjectIdNames[9640] = 'ExclusiveRateOfChangeAlarmType_EnabledState_EffectiveDisplayName'
ObjectIdNames[9641] = 'ExclusiveRateOfChangeAlarmType_EnabledState_TransitionTime'
ObjectIdNames[9642] = 'ExclusiveRateOfChangeAlarmType_EnabledState_EffectiveTransitionTime'
ObjectIdNames[9643] = 'ExclusiveRateOfChangeAlarmType_EnabledState_TrueState'
ObjectIdNames[9644] = 'ExclusiveRateOfChangeAlarmType_EnabledState_FalseState'
ObjectIdNames[9645] = 'ExclusiveRateOfChangeAlarmType_Quality'
ObjectIdNames[9646] = 'ExclusiveRateOfChangeAlarmType_Quality_SourceTimestamp'
ObjectIdNames[9647] = 'ExclusiveRateOfChangeAlarmType_LastSeverity'
ObjectIdNames[9648] = 'ExclusiveRateOfChangeAlarmType_LastSeverity_SourceTimestamp'
ObjectIdNames[9649] = 'ExclusiveRateOfChangeAlarmType_Comment'
ObjectIdNames[9650] = 'ExclusiveRateOfChangeAlarmType_Comment_SourceTimestamp'
ObjectIdNames[9651] = 'ExclusiveRateOfChangeAlarmType_ClientUserId'
ObjectIdNames[9652] = 'ExclusiveRateOfChangeAlarmType_Enable'
ObjectIdNames[9653] = 'ExclusiveRateOfChangeAlarmType_Disable'
ObjectIdNames[9654] = 'ExclusiveRateOfChangeAlarmType_AddComment'
ObjectIdNames[9655] = 'ExclusiveRateOfChangeAlarmType_AddComment_InputArguments'
ObjectIdNames[9656] = 'ExclusiveRateOfChangeAlarmType_ConditionRefresh'
ObjectIdNames[9657] = 'ExclusiveRateOfChangeAlarmType_ConditionRefresh_InputArguments'
ObjectIdNames[9658] = 'ExclusiveRateOfChangeAlarmType_AckedState'
ObjectIdNames[9659] = 'ExclusiveRateOfChangeAlarmType_AckedState_Id'
ObjectIdNames[9660] = 'ExclusiveRateOfChangeAlarmType_AckedState_Name'
ObjectIdNames[9661] = 'ExclusiveRateOfChangeAlarmType_AckedState_Number'
ObjectIdNames[9662] = 'ExclusiveRateOfChangeAlarmType_AckedState_EffectiveDisplayName'
ObjectIdNames[9663] = 'ExclusiveRateOfChangeAlarmType_AckedState_TransitionTime'
ObjectIdNames[9664] = 'ExclusiveRateOfChangeAlarmType_AckedState_EffectiveTransitionTime'
ObjectIdNames[9665] = 'ExclusiveRateOfChangeAlarmType_AckedState_TrueState'
ObjectIdNames[9666] = 'ExclusiveRateOfChangeAlarmType_AckedState_FalseState'
ObjectIdNames[9667] = 'ExclusiveRateOfChangeAlarmType_ConfirmedState'
ObjectIdNames[9668] = 'ExclusiveRateOfChangeAlarmType_ConfirmedState_Id'
ObjectIdNames[9669] = 'ExclusiveRateOfChangeAlarmType_ConfirmedState_Name'
ObjectIdNames[9670] = 'ExclusiveRateOfChangeAlarmType_ConfirmedState_Number'
ObjectIdNames[9671] = 'ExclusiveRateOfChangeAlarmType_ConfirmedState_EffectiveDisplayName'
ObjectIdNames[9672] = 'ExclusiveRateOfChangeAlarmType_ConfirmedState_TransitionTime'
ObjectIdNames[9673] = 'ExclusiveRateOfChangeAlarmType_ConfirmedState_EffectiveTransitionTime'
ObjectIdNames[9674] = 'ExclusiveRateOfChangeAlarmType_ConfirmedState_TrueState'
ObjectIdNames[9675] = 'ExclusiveRateOfChangeAlarmType_ConfirmedState_FalseState'
ObjectIdNames[9676] = 'ExclusiveRateOfChangeAlarmType_Acknowledge'
ObjectIdNames[9677] = 'ExclusiveRateOfChangeAlarmType_Acknowledge_InputArguments'
ObjectIdNames[9678] = 'ExclusiveRateOfChangeAlarmType_Confirm'
ObjectIdNames[9679] = 'ExclusiveRateOfChangeAlarmType_Confirm_InputArguments'
ObjectIdNames[9680] = 'ExclusiveRateOfChangeAlarmType_ActiveState'
ObjectIdNames[9681] = 'ExclusiveRateOfChangeAlarmType_ActiveState_Id'
ObjectIdNames[9682] = 'ExclusiveRateOfChangeAlarmType_ActiveState_Name'
ObjectIdNames[9683] = 'ExclusiveRateOfChangeAlarmType_ActiveState_Number'
ObjectIdNames[9684] = 'ExclusiveRateOfChangeAlarmType_ActiveState_EffectiveDisplayName'
ObjectIdNames[9685] = 'ExclusiveRateOfChangeAlarmType_ActiveState_TransitionTime'
ObjectIdNames[9686] = 'ExclusiveRateOfChangeAlarmType_ActiveState_EffectiveTransitionTime'
ObjectIdNames[9687] = 'ExclusiveRateOfChangeAlarmType_ActiveState_TrueState'
ObjectIdNames[9688] = 'ExclusiveRateOfChangeAlarmType_ActiveState_FalseState'
ObjectIdNames[9689] = 'ExclusiveRateOfChangeAlarmType_SuppressedState'
ObjectIdNames[9690] = 'ExclusiveRateOfChangeAlarmType_SuppressedState_Id'
ObjectIdNames[9691] = 'ExclusiveRateOfChangeAlarmType_SuppressedState_Name'
ObjectIdNames[9692] = 'ExclusiveRateOfChangeAlarmType_SuppressedState_Number'
ObjectIdNames[9693] = 'ExclusiveRateOfChangeAlarmType_SuppressedState_EffectiveDisplayName'
ObjectIdNames[9694] = 'ExclusiveRateOfChangeAlarmType_SuppressedState_TransitionTime'
ObjectIdNames[9695] = 'ExclusiveRateOfChangeAlarmType_SuppressedState_EffectiveTransitionTime'
ObjectIdNames[9696] = 'ExclusiveRateOfChangeAlarmType_SuppressedState_TrueState'
ObjectIdNames[9697] = 'ExclusiveRateOfChangeAlarmType_SuppressedState_FalseState'
ObjectIdNames[9698] = 'ExclusiveRateOfChangeAlarmType_ShelvingState'
ObjectIdNames[9699] = 'ExclusiveRateOfChangeAlarmType_ShelvingState_CurrentState'
ObjectIdNames[9700] = 'ExclusiveRateOfChangeAlarmType_ShelvingState_CurrentState_Id'
ObjectIdNames[9701] = 'ExclusiveRateOfChangeAlarmType_ShelvingState_CurrentState_Name'
ObjectIdNames[9702] = 'ExclusiveRateOfChangeAlarmType_ShelvingState_CurrentState_Number'
ObjectIdNames[9703] = 'ExclusiveRateOfChangeAlarmType_ShelvingState_CurrentState_EffectiveDisplayName'
ObjectIdNames[9704] = 'ExclusiveRateOfChangeAlarmType_ShelvingState_LastTransition'
ObjectIdNames[9705] = 'ExclusiveRateOfChangeAlarmType_ShelvingState_LastTransition_Id'
ObjectIdNames[9706] = 'ExclusiveRateOfChangeAlarmType_ShelvingState_LastTransition_Name'
ObjectIdNames[9707] = 'ExclusiveRateOfChangeAlarmType_ShelvingState_LastTransition_Number'
ObjectIdNames[9708] = 'ExclusiveRateOfChangeAlarmType_ShelvingState_LastTransition_TransitionTime'
ObjectIdNames[9709] = 'ExclusiveRateOfChangeAlarmType_ShelvingState_UnshelveTime'
ObjectIdNames[9731] = 'ExclusiveRateOfChangeAlarmType_ShelvingState_Unshelve'
ObjectIdNames[9732] = 'ExclusiveRateOfChangeAlarmType_ShelvingState_OneShotShelve'
ObjectIdNames[9733] = 'ExclusiveRateOfChangeAlarmType_ShelvingState_TimedShelve'
ObjectIdNames[9734] = 'ExclusiveRateOfChangeAlarmType_ShelvingState_TimedShelve_InputArguments'
ObjectIdNames[9735] = 'ExclusiveRateOfChangeAlarmType_SuppressedOrShelved'
ObjectIdNames[9736] = 'ExclusiveRateOfChangeAlarmType_MaxTimeShelved'
ObjectIdNames[9737] = 'ExclusiveRateOfChangeAlarmType_LimitState'
ObjectIdNames[9738] = 'ExclusiveRateOfChangeAlarmType_LimitState_CurrentState'
ObjectIdNames[9739] = 'ExclusiveRateOfChangeAlarmType_LimitState_CurrentState_Id'
ObjectIdNames[9740] = 'ExclusiveRateOfChangeAlarmType_LimitState_CurrentState_Name'
ObjectIdNames[9741] = 'ExclusiveRateOfChangeAlarmType_LimitState_CurrentState_Number'
ObjectIdNames[9742] = 'ExclusiveRateOfChangeAlarmType_LimitState_CurrentState_EffectiveDisplayName'
ObjectIdNames[9743] = 'ExclusiveRateOfChangeAlarmType_LimitState_LastTransition'
ObjectIdNames[9744] = 'ExclusiveRateOfChangeAlarmType_LimitState_LastTransition_Id'
ObjectIdNames[9745] = 'ExclusiveRateOfChangeAlarmType_LimitState_LastTransition_Name'
ObjectIdNames[9746] = 'ExclusiveRateOfChangeAlarmType_LimitState_LastTransition_Number'
ObjectIdNames[9747] = 'ExclusiveRateOfChangeAlarmType_LimitState_LastTransition_TransitionTime'
ObjectIdNames[9760] = 'ExclusiveRateOfChangeAlarmType_HighHighLimit'
ObjectIdNames[9761] = 'ExclusiveRateOfChangeAlarmType_HighLimit'
ObjectIdNames[9762] = 'ExclusiveRateOfChangeAlarmType_LowLimit'
ObjectIdNames[9763] = 'ExclusiveRateOfChangeAlarmType_LowLowLimit'
ObjectIdNames[9764] = 'ExclusiveDeviationAlarmType'
ObjectIdNames[9765] = 'ExclusiveDeviationAlarmType_EventId'
ObjectIdNames[9766] = 'ExclusiveDeviationAlarmType_EventType'
ObjectIdNames[9767] = 'ExclusiveDeviationAlarmType_SourceNode'
ObjectIdNames[9768] = 'ExclusiveDeviationAlarmType_SourceName'
ObjectIdNames[9769] = 'ExclusiveDeviationAlarmType_Time'
ObjectIdNames[9770] = 'ExclusiveDeviationAlarmType_ReceiveTime'
ObjectIdNames[9771] = 'ExclusiveDeviationAlarmType_LocalTime'
ObjectIdNames[9772] = 'ExclusiveDeviationAlarmType_Message'
ObjectIdNames[9773] = 'ExclusiveDeviationAlarmType_Severity'
ObjectIdNames[9774] = 'ExclusiveDeviationAlarmType_ConditionName'
ObjectIdNames[9775] = 'ExclusiveDeviationAlarmType_BranchId'
ObjectIdNames[9776] = 'ExclusiveDeviationAlarmType_Retain'
ObjectIdNames[9777] = 'ExclusiveDeviationAlarmType_EnabledState'
ObjectIdNames[9778] = 'ExclusiveDeviationAlarmType_EnabledState_Id'
ObjectIdNames[9779] = 'ExclusiveDeviationAlarmType_EnabledState_Name'
ObjectIdNames[9780] = 'ExclusiveDeviationAlarmType_EnabledState_Number'
ObjectIdNames[9781] = 'ExclusiveDeviationAlarmType_EnabledState_EffectiveDisplayName'
ObjectIdNames[9782] = 'ExclusiveDeviationAlarmType_EnabledState_TransitionTime'
ObjectIdNames[9783] = 'ExclusiveDeviationAlarmType_EnabledState_EffectiveTransitionTime'
ObjectIdNames[9784] = 'ExclusiveDeviationAlarmType_EnabledState_TrueState'
ObjectIdNames[9785] = 'ExclusiveDeviationAlarmType_EnabledState_FalseState'
ObjectIdNames[9786] = 'ExclusiveDeviationAlarmType_Quality'
ObjectIdNames[9787] = 'ExclusiveDeviationAlarmType_Quality_SourceTimestamp'
ObjectIdNames[9788] = 'ExclusiveDeviationAlarmType_LastSeverity'
ObjectIdNames[9789] = 'ExclusiveDeviationAlarmType_LastSeverity_SourceTimestamp'
ObjectIdNames[9790] = 'ExclusiveDeviationAlarmType_Comment'
ObjectIdNames[9791] = 'ExclusiveDeviationAlarmType_Comment_SourceTimestamp'
ObjectIdNames[9792] = 'ExclusiveDeviationAlarmType_ClientUserId'
ObjectIdNames[9793] = 'ExclusiveDeviationAlarmType_Enable'
ObjectIdNames[9794] = 'ExclusiveDeviationAlarmType_Disable'
ObjectIdNames[9795] = 'ExclusiveDeviationAlarmType_AddComment'
ObjectIdNames[9796] = 'ExclusiveDeviationAlarmType_AddComment_InputArguments'
ObjectIdNames[9797] = 'ExclusiveDeviationAlarmType_ConditionRefresh'
ObjectIdNames[9798] = 'ExclusiveDeviationAlarmType_ConditionRefresh_InputArguments'
ObjectIdNames[9799] = 'ExclusiveDeviationAlarmType_AckedState'
ObjectIdNames[9800] = 'ExclusiveDeviationAlarmType_AckedState_Id'
ObjectIdNames[9801] = 'ExclusiveDeviationAlarmType_AckedState_Name'
ObjectIdNames[9802] = 'ExclusiveDeviationAlarmType_AckedState_Number'
ObjectIdNames[9803] = 'ExclusiveDeviationAlarmType_AckedState_EffectiveDisplayName'
ObjectIdNames[9804] = 'ExclusiveDeviationAlarmType_AckedState_TransitionTime'
ObjectIdNames[9805] = 'ExclusiveDeviationAlarmType_AckedState_EffectiveTransitionTime'
ObjectIdNames[9806] = 'ExclusiveDeviationAlarmType_AckedState_TrueState'
ObjectIdNames[9807] = 'ExclusiveDeviationAlarmType_AckedState_FalseState'
ObjectIdNames[9808] = 'ExclusiveDeviationAlarmType_ConfirmedState'
ObjectIdNames[9809] = 'ExclusiveDeviationAlarmType_ConfirmedState_Id'
ObjectIdNames[9810] = 'ExclusiveDeviationAlarmType_ConfirmedState_Name'
ObjectIdNames[9811] = 'ExclusiveDeviationAlarmType_ConfirmedState_Number'
ObjectIdNames[9812] = 'ExclusiveDeviationAlarmType_ConfirmedState_EffectiveDisplayName'
ObjectIdNames[9813] = 'ExclusiveDeviationAlarmType_ConfirmedState_TransitionTime'
ObjectIdNames[9814] = 'ExclusiveDeviationAlarmType_ConfirmedState_EffectiveTransitionTime'
ObjectIdNames[9815] = 'ExclusiveDeviationAlarmType_ConfirmedState_TrueState'
ObjectIdNames[9816] = 'ExclusiveDeviationAlarmType_ConfirmedState_FalseState'
ObjectIdNames[9817] = 'ExclusiveDeviationAlarmType_Acknowledge'
ObjectIdNames[9818] = 'ExclusiveDeviationAlarmType_Acknowledge_InputArguments'
ObjectIdNames[9819] = 'ExclusiveDeviationAlarmType_Confirm'
ObjectIdNames[9820] = 'ExclusiveDeviationAlarmType_Confirm_InputArguments'
ObjectIdNames[9821] = 'ExclusiveDeviationAlarmType_ActiveState'
ObjectIdNames[9822] = 'ExclusiveDeviationAlarmType_ActiveState_Id'
ObjectIdNames[9823] = 'ExclusiveDeviationAlarmType_ActiveState_Name'
ObjectIdNames[9824] = 'ExclusiveDeviationAlarmType_ActiveState_Number'
ObjectIdNames[9825] = 'ExclusiveDeviationAlarmType_ActiveState_EffectiveDisplayName'
ObjectIdNames[9826] = 'ExclusiveDeviationAlarmType_ActiveState_TransitionTime'
ObjectIdNames[9827] = 'ExclusiveDeviationAlarmType_ActiveState_EffectiveTransitionTime'
ObjectIdNames[9828] = 'ExclusiveDeviationAlarmType_ActiveState_TrueState'
ObjectIdNames[9829] = 'ExclusiveDeviationAlarmType_ActiveState_FalseState'
ObjectIdNames[9830] = 'ExclusiveDeviationAlarmType_SuppressedState'
ObjectIdNames[9831] = 'ExclusiveDeviationAlarmType_SuppressedState_Id'
ObjectIdNames[9832] = 'ExclusiveDeviationAlarmType_SuppressedState_Name'
ObjectIdNames[9833] = 'ExclusiveDeviationAlarmType_SuppressedState_Number'
ObjectIdNames[9834] = 'ExclusiveDeviationAlarmType_SuppressedState_EffectiveDisplayName'
ObjectIdNames[9835] = 'ExclusiveDeviationAlarmType_SuppressedState_TransitionTime'
ObjectIdNames[9836] = 'ExclusiveDeviationAlarmType_SuppressedState_EffectiveTransitionTime'
ObjectIdNames[9837] = 'ExclusiveDeviationAlarmType_SuppressedState_TrueState'
ObjectIdNames[9838] = 'ExclusiveDeviationAlarmType_SuppressedState_FalseState'
ObjectIdNames[9839] = 'ExclusiveDeviationAlarmType_ShelvingState'
ObjectIdNames[9840] = 'ExclusiveDeviationAlarmType_ShelvingState_CurrentState'
ObjectIdNames[9841] = 'ExclusiveDeviationAlarmType_ShelvingState_CurrentState_Id'
ObjectIdNames[9842] = 'ExclusiveDeviationAlarmType_ShelvingState_CurrentState_Name'
ObjectIdNames[9843] = 'ExclusiveDeviationAlarmType_ShelvingState_CurrentState_Number'
ObjectIdNames[9844] = 'ExclusiveDeviationAlarmType_ShelvingState_CurrentState_EffectiveDisplayName'
ObjectIdNames[9845] = 'ExclusiveDeviationAlarmType_ShelvingState_LastTransition'
ObjectIdNames[9846] = 'ExclusiveDeviationAlarmType_ShelvingState_LastTransition_Id'
ObjectIdNames[9847] = 'ExclusiveDeviationAlarmType_ShelvingState_LastTransition_Name'
ObjectIdNames[9848] = 'ExclusiveDeviationAlarmType_ShelvingState_LastTransition_Number'
ObjectIdNames[9849] = 'ExclusiveDeviationAlarmType_ShelvingState_LastTransition_TransitionTime'
ObjectIdNames[9850] = 'ExclusiveDeviationAlarmType_ShelvingState_UnshelveTime'
ObjectIdNames[9872] = 'ExclusiveDeviationAlarmType_ShelvingState_Unshelve'
ObjectIdNames[9873] = 'ExclusiveDeviationAlarmType_ShelvingState_OneShotShelve'
ObjectIdNames[9874] = 'ExclusiveDeviationAlarmType_ShelvingState_TimedShelve'
ObjectIdNames[9875] = 'ExclusiveDeviationAlarmType_ShelvingState_TimedShelve_InputArguments'
ObjectIdNames[9876] = 'ExclusiveDeviationAlarmType_SuppressedOrShelved'
ObjectIdNames[9877] = 'ExclusiveDeviationAlarmType_MaxTimeShelved'
ObjectIdNames[9878] = 'ExclusiveDeviationAlarmType_LimitState'
ObjectIdNames[9879] = 'ExclusiveDeviationAlarmType_LimitState_CurrentState'
ObjectIdNames[9880] = 'ExclusiveDeviationAlarmType_LimitState_CurrentState_Id'
ObjectIdNames[9881] = 'ExclusiveDeviationAlarmType_LimitState_CurrentState_Name'
ObjectIdNames[9882] = 'ExclusiveDeviationAlarmType_LimitState_CurrentState_Number'
ObjectIdNames[9883] = 'ExclusiveDeviationAlarmType_LimitState_CurrentState_EffectiveDisplayName'
ObjectIdNames[9884] = 'ExclusiveDeviationAlarmType_LimitState_LastTransition'
ObjectIdNames[9885] = 'ExclusiveDeviationAlarmType_LimitState_LastTransition_Id'
ObjectIdNames[9886] = 'ExclusiveDeviationAlarmType_LimitState_LastTransition_Name'
ObjectIdNames[9887] = 'ExclusiveDeviationAlarmType_LimitState_LastTransition_Number'
ObjectIdNames[9888] = 'ExclusiveDeviationAlarmType_LimitState_LastTransition_TransitionTime'
ObjectIdNames[9901] = 'ExclusiveDeviationAlarmType_HighHighLimit'
ObjectIdNames[9902] = 'ExclusiveDeviationAlarmType_HighLimit'
ObjectIdNames[9903] = 'ExclusiveDeviationAlarmType_LowLimit'
ObjectIdNames[9904] = 'ExclusiveDeviationAlarmType_LowLowLimit'
ObjectIdNames[9905] = 'ExclusiveDeviationAlarmType_SetpointNode'
ObjectIdNames[9906] = 'NonExclusiveLimitAlarmType'
ObjectIdNames[9907] = 'NonExclusiveLimitAlarmType_EventId'
ObjectIdNames[9908] = 'NonExclusiveLimitAlarmType_EventType'
ObjectIdNames[9909] = 'NonExclusiveLimitAlarmType_SourceNode'
ObjectIdNames[9910] = 'NonExclusiveLimitAlarmType_SourceName'
ObjectIdNames[9911] = 'NonExclusiveLimitAlarmType_Time'
ObjectIdNames[9912] = 'NonExclusiveLimitAlarmType_ReceiveTime'
ObjectIdNames[9913] = 'NonExclusiveLimitAlarmType_LocalTime'
ObjectIdNames[9914] = 'NonExclusiveLimitAlarmType_Message'
ObjectIdNames[9915] = 'NonExclusiveLimitAlarmType_Severity'
ObjectIdNames[9916] = 'NonExclusiveLimitAlarmType_ConditionName'
ObjectIdNames[9917] = 'NonExclusiveLimitAlarmType_BranchId'
ObjectIdNames[9918] = 'NonExclusiveLimitAlarmType_Retain'
ObjectIdNames[9919] = 'NonExclusiveLimitAlarmType_EnabledState'
ObjectIdNames[9920] = 'NonExclusiveLimitAlarmType_EnabledState_Id'
ObjectIdNames[9921] = 'NonExclusiveLimitAlarmType_EnabledState_Name'
ObjectIdNames[9922] = 'NonExclusiveLimitAlarmType_EnabledState_Number'
ObjectIdNames[9923] = 'NonExclusiveLimitAlarmType_EnabledState_EffectiveDisplayName'
ObjectIdNames[9924] = 'NonExclusiveLimitAlarmType_EnabledState_TransitionTime'
ObjectIdNames[9925] = 'NonExclusiveLimitAlarmType_EnabledState_EffectiveTransitionTime'
ObjectIdNames[9926] = 'NonExclusiveLimitAlarmType_EnabledState_TrueState'
ObjectIdNames[9927] = 'NonExclusiveLimitAlarmType_EnabledState_FalseState'
ObjectIdNames[9928] = 'NonExclusiveLimitAlarmType_Quality'
ObjectIdNames[9929] = 'NonExclusiveLimitAlarmType_Quality_SourceTimestamp'
ObjectIdNames[9930] = 'NonExclusiveLimitAlarmType_LastSeverity'
ObjectIdNames[9931] = 'NonExclusiveLimitAlarmType_LastSeverity_SourceTimestamp'
ObjectIdNames[9932] = 'NonExclusiveLimitAlarmType_Comment'
ObjectIdNames[9933] = 'NonExclusiveLimitAlarmType_Comment_SourceTimestamp'
ObjectIdNames[9934] = 'NonExclusiveLimitAlarmType_ClientUserId'
ObjectIdNames[9935] = 'NonExclusiveLimitAlarmType_Enable'
ObjectIdNames[9936] = 'NonExclusiveLimitAlarmType_Disable'
ObjectIdNames[9937] = 'NonExclusiveLimitAlarmType_AddComment'
ObjectIdNames[9938] = 'NonExclusiveLimitAlarmType_AddComment_InputArguments'
ObjectIdNames[9939] = 'NonExclusiveLimitAlarmType_ConditionRefresh'
ObjectIdNames[9940] = 'NonExclusiveLimitAlarmType_ConditionRefresh_InputArguments'
ObjectIdNames[9941] = 'NonExclusiveLimitAlarmType_AckedState'
ObjectIdNames[9942] = 'NonExclusiveLimitAlarmType_AckedState_Id'
ObjectIdNames[9943] = 'NonExclusiveLimitAlarmType_AckedState_Name'
ObjectIdNames[9944] = 'NonExclusiveLimitAlarmType_AckedState_Number'
ObjectIdNames[9945] = 'NonExclusiveLimitAlarmType_AckedState_EffectiveDisplayName'
ObjectIdNames[9946] = 'NonExclusiveLimitAlarmType_AckedState_TransitionTime'
ObjectIdNames[9947] = 'NonExclusiveLimitAlarmType_AckedState_EffectiveTransitionTime'
ObjectIdNames[9948] = 'NonExclusiveLimitAlarmType_AckedState_TrueState'
ObjectIdNames[9949] = 'NonExclusiveLimitAlarmType_AckedState_FalseState'
ObjectIdNames[9950] = 'NonExclusiveLimitAlarmType_ConfirmedState'
ObjectIdNames[9951] = 'NonExclusiveLimitAlarmType_ConfirmedState_Id'
ObjectIdNames[9952] = 'NonExclusiveLimitAlarmType_ConfirmedState_Name'
ObjectIdNames[9953] = 'NonExclusiveLimitAlarmType_ConfirmedState_Number'
ObjectIdNames[9954] = 'NonExclusiveLimitAlarmType_ConfirmedState_EffectiveDisplayName'
ObjectIdNames[9955] = 'NonExclusiveLimitAlarmType_ConfirmedState_TransitionTime'
ObjectIdNames[9956] = 'NonExclusiveLimitAlarmType_ConfirmedState_EffectiveTransitionTime'
ObjectIdNames[9957] = 'NonExclusiveLimitAlarmType_ConfirmedState_TrueState'
ObjectIdNames[9958] = 'NonExclusiveLimitAlarmType_ConfirmedState_FalseState'
ObjectIdNames[9959] = 'NonExclusiveLimitAlarmType_Acknowledge'
ObjectIdNames[9960] = 'NonExclusiveLimitAlarmType_Acknowledge_InputArguments'
ObjectIdNames[9961] = 'NonExclusiveLimitAlarmType_Confirm'
ObjectIdNames[9962] = 'NonExclusiveLimitAlarmType_Confirm_InputArguments'
ObjectIdNames[9963] = 'NonExclusiveLimitAlarmType_ActiveState'
ObjectIdNames[9964] = 'NonExclusiveLimitAlarmType_ActiveState_Id'
ObjectIdNames[9965] = 'NonExclusiveLimitAlarmType_ActiveState_Name'
ObjectIdNames[9966] = 'NonExclusiveLimitAlarmType_ActiveState_Number'
ObjectIdNames[9967] = 'NonExclusiveLimitAlarmType_ActiveState_EffectiveDisplayName'
ObjectIdNames[9968] = 'NonExclusiveLimitAlarmType_ActiveState_TransitionTime'
ObjectIdNames[9969] = 'NonExclusiveLimitAlarmType_ActiveState_EffectiveTransitionTime'
ObjectIdNames[9970] = 'NonExclusiveLimitAlarmType_ActiveState_TrueState'
ObjectIdNames[9971] = 'NonExclusiveLimitAlarmType_ActiveState_FalseState'
ObjectIdNames[9972] = 'NonExclusiveLimitAlarmType_SuppressedState'
ObjectIdNames[9973] = 'NonExclusiveLimitAlarmType_SuppressedState_Id'
ObjectIdNames[9974] = 'NonExclusiveLimitAlarmType_SuppressedState_Name'
ObjectIdNames[9975] = 'NonExclusiveLimitAlarmType_SuppressedState_Number'
ObjectIdNames[9976] = 'NonExclusiveLimitAlarmType_SuppressedState_EffectiveDisplayName'
ObjectIdNames[9977] = 'NonExclusiveLimitAlarmType_SuppressedState_TransitionTime'
ObjectIdNames[9978] = 'NonExclusiveLimitAlarmType_SuppressedState_EffectiveTransitionTime'
ObjectIdNames[9979] = 'NonExclusiveLimitAlarmType_SuppressedState_TrueState'
ObjectIdNames[9980] = 'NonExclusiveLimitAlarmType_SuppressedState_FalseState'
ObjectIdNames[9981] = 'NonExclusiveLimitAlarmType_ShelvingState'
ObjectIdNames[9982] = 'NonExclusiveLimitAlarmType_ShelvingState_CurrentState'
ObjectIdNames[9983] = 'NonExclusiveLimitAlarmType_ShelvingState_CurrentState_Id'
ObjectIdNames[9984] = 'NonExclusiveLimitAlarmType_ShelvingState_CurrentState_Name'
ObjectIdNames[9985] = 'NonExclusiveLimitAlarmType_ShelvingState_CurrentState_Number'
ObjectIdNames[9986] = 'NonExclusiveLimitAlarmType_ShelvingState_CurrentState_EffectiveDisplayName'
ObjectIdNames[9987] = 'NonExclusiveLimitAlarmType_ShelvingState_LastTransition'
ObjectIdNames[9988] = 'NonExclusiveLimitAlarmType_ShelvingState_LastTransition_Id'
ObjectIdNames[9989] = 'NonExclusiveLimitAlarmType_ShelvingState_LastTransition_Name'
ObjectIdNames[9990] = 'NonExclusiveLimitAlarmType_ShelvingState_LastTransition_Number'
ObjectIdNames[9991] = 'NonExclusiveLimitAlarmType_ShelvingState_LastTransition_TransitionTime'
ObjectIdNames[9992] = 'NonExclusiveLimitAlarmType_ShelvingState_UnshelveTime'
ObjectIdNames[10014] = 'NonExclusiveLimitAlarmType_ShelvingState_Unshelve'
ObjectIdNames[10015] = 'NonExclusiveLimitAlarmType_ShelvingState_OneShotShelve'
ObjectIdNames[10016] = 'NonExclusiveLimitAlarmType_ShelvingState_TimedShelve'
ObjectIdNames[10017] = 'NonExclusiveLimitAlarmType_ShelvingState_TimedShelve_InputArguments'
ObjectIdNames[10018] = 'NonExclusiveLimitAlarmType_SuppressedOrShelved'
ObjectIdNames[10019] = 'NonExclusiveLimitAlarmType_MaxTimeShelved'
ObjectIdNames[10020] = 'NonExclusiveLimitAlarmType_HighHighState'
ObjectIdNames[10021] = 'NonExclusiveLimitAlarmType_HighHighState_Id'
ObjectIdNames[10022] = 'NonExclusiveLimitAlarmType_HighHighState_Name'
ObjectIdNames[10023] = 'NonExclusiveLimitAlarmType_HighHighState_Number'
ObjectIdNames[10024] = 'NonExclusiveLimitAlarmType_HighHighState_EffectiveDisplayName'
ObjectIdNames[10025] = 'NonExclusiveLimitAlarmType_HighHighState_TransitionTime'
ObjectIdNames[10026] = 'NonExclusiveLimitAlarmType_HighHighState_EffectiveTransitionTime'
ObjectIdNames[10027] = 'NonExclusiveLimitAlarmType_HighHighState_TrueState'
ObjectIdNames[10028] = 'NonExclusiveLimitAlarmType_HighHighState_FalseState'
ObjectIdNames[10029] = 'NonExclusiveLimitAlarmType_HighState'
ObjectIdNames[10030] = 'NonExclusiveLimitAlarmType_HighState_Id'
ObjectIdNames[10031] = 'NonExclusiveLimitAlarmType_HighState_Name'
ObjectIdNames[10032] = 'NonExclusiveLimitAlarmType_HighState_Number'
ObjectIdNames[10033] = 'NonExclusiveLimitAlarmType_HighState_EffectiveDisplayName'
ObjectIdNames[10034] = 'NonExclusiveLimitAlarmType_HighState_TransitionTime'
ObjectIdNames[10035] = 'NonExclusiveLimitAlarmType_HighState_EffectiveTransitionTime'
ObjectIdNames[10036] = 'NonExclusiveLimitAlarmType_HighState_TrueState'
ObjectIdNames[10037] = 'NonExclusiveLimitAlarmType_HighState_FalseState'
ObjectIdNames[10038] = 'NonExclusiveLimitAlarmType_LowState'
ObjectIdNames[10039] = 'NonExclusiveLimitAlarmType_LowState_Id'
ObjectIdNames[10040] = 'NonExclusiveLimitAlarmType_LowState_Name'
ObjectIdNames[10041] = 'NonExclusiveLimitAlarmType_LowState_Number'
ObjectIdNames[10042] = 'NonExclusiveLimitAlarmType_LowState_EffectiveDisplayName'
ObjectIdNames[10043] = 'NonExclusiveLimitAlarmType_LowState_TransitionTime'
ObjectIdNames[10044] = 'NonExclusiveLimitAlarmType_LowState_EffectiveTransitionTime'
ObjectIdNames[10045] = 'NonExclusiveLimitAlarmType_LowState_TrueState'
ObjectIdNames[10046] = 'NonExclusiveLimitAlarmType_LowState_FalseState'
ObjectIdNames[10047] = 'NonExclusiveLimitAlarmType_LowLowState'
ObjectIdNames[10048] = 'NonExclusiveLimitAlarmType_LowLowState_Id'
ObjectIdNames[10049] = 'NonExclusiveLimitAlarmType_LowLowState_Name'
ObjectIdNames[10050] = 'NonExclusiveLimitAlarmType_LowLowState_Number'
ObjectIdNames[10051] = 'NonExclusiveLimitAlarmType_LowLowState_EffectiveDisplayName'
ObjectIdNames[10052] = 'NonExclusiveLimitAlarmType_LowLowState_TransitionTime'
ObjectIdNames[10053] = 'NonExclusiveLimitAlarmType_LowLowState_EffectiveTransitionTime'
ObjectIdNames[10054] = 'NonExclusiveLimitAlarmType_LowLowState_TrueState'
ObjectIdNames[10055] = 'NonExclusiveLimitAlarmType_LowLowState_FalseState'
ObjectIdNames[10056] = 'NonExclusiveLimitAlarmType_HighHighLimit'
ObjectIdNames[10057] = 'NonExclusiveLimitAlarmType_HighLimit'
ObjectIdNames[10058] = 'NonExclusiveLimitAlarmType_LowLimit'
ObjectIdNames[10059] = 'NonExclusiveLimitAlarmType_LowLowLimit'
ObjectIdNames[10060] = 'NonExclusiveLevelAlarmType'
ObjectIdNames[10061] = 'NonExclusiveLevelAlarmType_EventId'
ObjectIdNames[10062] = 'NonExclusiveLevelAlarmType_EventType'
ObjectIdNames[10063] = 'NonExclusiveLevelAlarmType_SourceNode'
ObjectIdNames[10064] = 'NonExclusiveLevelAlarmType_SourceName'
ObjectIdNames[10065] = 'NonExclusiveLevelAlarmType_Time'
ObjectIdNames[10066] = 'NonExclusiveLevelAlarmType_ReceiveTime'
ObjectIdNames[10067] = 'NonExclusiveLevelAlarmType_LocalTime'
ObjectIdNames[10068] = 'NonExclusiveLevelAlarmType_Message'
ObjectIdNames[10069] = 'NonExclusiveLevelAlarmType_Severity'
ObjectIdNames[10070] = 'NonExclusiveLevelAlarmType_ConditionName'
ObjectIdNames[10071] = 'NonExclusiveLevelAlarmType_BranchId'
ObjectIdNames[10072] = 'NonExclusiveLevelAlarmType_Retain'
ObjectIdNames[10073] = 'NonExclusiveLevelAlarmType_EnabledState'
ObjectIdNames[10074] = 'NonExclusiveLevelAlarmType_EnabledState_Id'
ObjectIdNames[10075] = 'NonExclusiveLevelAlarmType_EnabledState_Name'
ObjectIdNames[10076] = 'NonExclusiveLevelAlarmType_EnabledState_Number'
ObjectIdNames[10077] = 'NonExclusiveLevelAlarmType_EnabledState_EffectiveDisplayName'
ObjectIdNames[10078] = 'NonExclusiveLevelAlarmType_EnabledState_TransitionTime'
ObjectIdNames[10079] = 'NonExclusiveLevelAlarmType_EnabledState_EffectiveTransitionTime'
ObjectIdNames[10080] = 'NonExclusiveLevelAlarmType_EnabledState_TrueState'
ObjectIdNames[10081] = 'NonExclusiveLevelAlarmType_EnabledState_FalseState'
ObjectIdNames[10082] = 'NonExclusiveLevelAlarmType_Quality'
ObjectIdNames[10083] = 'NonExclusiveLevelAlarmType_Quality_SourceTimestamp'
ObjectIdNames[10084] = 'NonExclusiveLevelAlarmType_LastSeverity'
ObjectIdNames[10085] = 'NonExclusiveLevelAlarmType_LastSeverity_SourceTimestamp'
ObjectIdNames[10086] = 'NonExclusiveLevelAlarmType_Comment'
ObjectIdNames[10087] = 'NonExclusiveLevelAlarmType_Comment_SourceTimestamp'
ObjectIdNames[10088] = 'NonExclusiveLevelAlarmType_ClientUserId'
ObjectIdNames[10089] = 'NonExclusiveLevelAlarmType_Enable'
ObjectIdNames[10090] = 'NonExclusiveLevelAlarmType_Disable'
ObjectIdNames[10091] = 'NonExclusiveLevelAlarmType_AddComment'
ObjectIdNames[10092] = 'NonExclusiveLevelAlarmType_AddComment_InputArguments'
ObjectIdNames[10093] = 'NonExclusiveLevelAlarmType_ConditionRefresh'
ObjectIdNames[10094] = 'NonExclusiveLevelAlarmType_ConditionRefresh_InputArguments'
ObjectIdNames[10095] = 'NonExclusiveLevelAlarmType_AckedState'
ObjectIdNames[10096] = 'NonExclusiveLevelAlarmType_AckedState_Id'
ObjectIdNames[10097] = 'NonExclusiveLevelAlarmType_AckedState_Name'
ObjectIdNames[10098] = 'NonExclusiveLevelAlarmType_AckedState_Number'
ObjectIdNames[10099] = 'NonExclusiveLevelAlarmType_AckedState_EffectiveDisplayName'
ObjectIdNames[10100] = 'NonExclusiveLevelAlarmType_AckedState_TransitionTime'
ObjectIdNames[10101] = 'NonExclusiveLevelAlarmType_AckedState_EffectiveTransitionTime'
ObjectIdNames[10102] = 'NonExclusiveLevelAlarmType_AckedState_TrueState'
ObjectIdNames[10103] = 'NonExclusiveLevelAlarmType_AckedState_FalseState'
ObjectIdNames[10104] = 'NonExclusiveLevelAlarmType_ConfirmedState'
ObjectIdNames[10105] = 'NonExclusiveLevelAlarmType_ConfirmedState_Id'
ObjectIdNames[10106] = 'NonExclusiveLevelAlarmType_ConfirmedState_Name'
ObjectIdNames[10107] = 'NonExclusiveLevelAlarmType_ConfirmedState_Number'
ObjectIdNames[10108] = 'NonExclusiveLevelAlarmType_ConfirmedState_EffectiveDisplayName'
ObjectIdNames[10109] = 'NonExclusiveLevelAlarmType_ConfirmedState_TransitionTime'
ObjectIdNames[10110] = 'NonExclusiveLevelAlarmType_ConfirmedState_EffectiveTransitionTime'
ObjectIdNames[10111] = 'NonExclusiveLevelAlarmType_ConfirmedState_TrueState'
ObjectIdNames[10112] = 'NonExclusiveLevelAlarmType_ConfirmedState_FalseState'
ObjectIdNames[10113] = 'NonExclusiveLevelAlarmType_Acknowledge'
ObjectIdNames[10114] = 'NonExclusiveLevelAlarmType_Acknowledge_InputArguments'
ObjectIdNames[10115] = 'NonExclusiveLevelAlarmType_Confirm'
ObjectIdNames[10116] = 'NonExclusiveLevelAlarmType_Confirm_InputArguments'
ObjectIdNames[10117] = 'NonExclusiveLevelAlarmType_ActiveState'
ObjectIdNames[10118] = 'NonExclusiveLevelAlarmType_ActiveState_Id'
ObjectIdNames[10119] = 'NonExclusiveLevelAlarmType_ActiveState_Name'
ObjectIdNames[10120] = 'NonExclusiveLevelAlarmType_ActiveState_Number'
ObjectIdNames[10121] = 'NonExclusiveLevelAlarmType_ActiveState_EffectiveDisplayName'
ObjectIdNames[10122] = 'NonExclusiveLevelAlarmType_ActiveState_TransitionTime'
ObjectIdNames[10123] = 'NonExclusiveLevelAlarmType_ActiveState_EffectiveTransitionTime'
ObjectIdNames[10124] = 'NonExclusiveLevelAlarmType_ActiveState_TrueState'
ObjectIdNames[10125] = 'NonExclusiveLevelAlarmType_ActiveState_FalseState'
ObjectIdNames[10126] = 'NonExclusiveLevelAlarmType_SuppressedState'
ObjectIdNames[10127] = 'NonExclusiveLevelAlarmType_SuppressedState_Id'
ObjectIdNames[10128] = 'NonExclusiveLevelAlarmType_SuppressedState_Name'
ObjectIdNames[10129] = 'NonExclusiveLevelAlarmType_SuppressedState_Number'
ObjectIdNames[10130] = 'NonExclusiveLevelAlarmType_SuppressedState_EffectiveDisplayName'
ObjectIdNames[10131] = 'NonExclusiveLevelAlarmType_SuppressedState_TransitionTime'
ObjectIdNames[10132] = 'NonExclusiveLevelAlarmType_SuppressedState_EffectiveTransitionTime'
ObjectIdNames[10133] = 'NonExclusiveLevelAlarmType_SuppressedState_TrueState'
ObjectIdNames[10134] = 'NonExclusiveLevelAlarmType_SuppressedState_FalseState'
ObjectIdNames[10135] = 'NonExclusiveLevelAlarmType_ShelvingState'
ObjectIdNames[10136] = 'NonExclusiveLevelAlarmType_ShelvingState_CurrentState'
ObjectIdNames[10137] = 'NonExclusiveLevelAlarmType_ShelvingState_CurrentState_Id'
ObjectIdNames[10138] = 'NonExclusiveLevelAlarmType_ShelvingState_CurrentState_Name'
ObjectIdNames[10139] = 'NonExclusiveLevelAlarmType_ShelvingState_CurrentState_Number'
ObjectIdNames[10140] = 'NonExclusiveLevelAlarmType_ShelvingState_CurrentState_EffectiveDisplayName'
ObjectIdNames[10141] = 'NonExclusiveLevelAlarmType_ShelvingState_LastTransition'
ObjectIdNames[10142] = 'NonExclusiveLevelAlarmType_ShelvingState_LastTransition_Id'
ObjectIdNames[10143] = 'NonExclusiveLevelAlarmType_ShelvingState_LastTransition_Name'
ObjectIdNames[10144] = 'NonExclusiveLevelAlarmType_ShelvingState_LastTransition_Number'
ObjectIdNames[10145] = 'NonExclusiveLevelAlarmType_ShelvingState_LastTransition_TransitionTime'
ObjectIdNames[10146] = 'NonExclusiveLevelAlarmType_ShelvingState_UnshelveTime'
ObjectIdNames[10168] = 'NonExclusiveLevelAlarmType_ShelvingState_Unshelve'
ObjectIdNames[10169] = 'NonExclusiveLevelAlarmType_ShelvingState_OneShotShelve'
ObjectIdNames[10170] = 'NonExclusiveLevelAlarmType_ShelvingState_TimedShelve'
ObjectIdNames[10171] = 'NonExclusiveLevelAlarmType_ShelvingState_TimedShelve_InputArguments'
ObjectIdNames[10172] = 'NonExclusiveLevelAlarmType_SuppressedOrShelved'
ObjectIdNames[10173] = 'NonExclusiveLevelAlarmType_MaxTimeShelved'
ObjectIdNames[10174] = 'NonExclusiveLevelAlarmType_HighHighState'
ObjectIdNames[10175] = 'NonExclusiveLevelAlarmType_HighHighState_Id'
ObjectIdNames[10176] = 'NonExclusiveLevelAlarmType_HighHighState_Name'
ObjectIdNames[10177] = 'NonExclusiveLevelAlarmType_HighHighState_Number'
ObjectIdNames[10178] = 'NonExclusiveLevelAlarmType_HighHighState_EffectiveDisplayName'
ObjectIdNames[10179] = 'NonExclusiveLevelAlarmType_HighHighState_TransitionTime'
ObjectIdNames[10180] = 'NonExclusiveLevelAlarmType_HighHighState_EffectiveTransitionTime'
ObjectIdNames[10181] = 'NonExclusiveLevelAlarmType_HighHighState_TrueState'
ObjectIdNames[10182] = 'NonExclusiveLevelAlarmType_HighHighState_FalseState'
ObjectIdNames[10183] = 'NonExclusiveLevelAlarmType_HighState'
ObjectIdNames[10184] = 'NonExclusiveLevelAlarmType_HighState_Id'
ObjectIdNames[10185] = 'NonExclusiveLevelAlarmType_HighState_Name'
ObjectIdNames[10186] = 'NonExclusiveLevelAlarmType_HighState_Number'
ObjectIdNames[10187] = 'NonExclusiveLevelAlarmType_HighState_EffectiveDisplayName'
ObjectIdNames[10188] = 'NonExclusiveLevelAlarmType_HighState_TransitionTime'
ObjectIdNames[10189] = 'NonExclusiveLevelAlarmType_HighState_EffectiveTransitionTime'
ObjectIdNames[10190] = 'NonExclusiveLevelAlarmType_HighState_TrueState'
ObjectIdNames[10191] = 'NonExclusiveLevelAlarmType_HighState_FalseState'
ObjectIdNames[10192] = 'NonExclusiveLevelAlarmType_LowState'
ObjectIdNames[10193] = 'NonExclusiveLevelAlarmType_LowState_Id'
ObjectIdNames[10194] = 'NonExclusiveLevelAlarmType_LowState_Name'
ObjectIdNames[10195] = 'NonExclusiveLevelAlarmType_LowState_Number'
ObjectIdNames[10196] = 'NonExclusiveLevelAlarmType_LowState_EffectiveDisplayName'
ObjectIdNames[10197] = 'NonExclusiveLevelAlarmType_LowState_TransitionTime'
ObjectIdNames[10198] = 'NonExclusiveLevelAlarmType_LowState_EffectiveTransitionTime'
ObjectIdNames[10199] = 'NonExclusiveLevelAlarmType_LowState_TrueState'
ObjectIdNames[10200] = 'NonExclusiveLevelAlarmType_LowState_FalseState'
ObjectIdNames[10201] = 'NonExclusiveLevelAlarmType_LowLowState'
ObjectIdNames[10202] = 'NonExclusiveLevelAlarmType_LowLowState_Id'
ObjectIdNames[10203] = 'NonExclusiveLevelAlarmType_LowLowState_Name'
ObjectIdNames[10204] = 'NonExclusiveLevelAlarmType_LowLowState_Number'
ObjectIdNames[10205] = 'NonExclusiveLevelAlarmType_LowLowState_EffectiveDisplayName'
ObjectIdNames[10206] = 'NonExclusiveLevelAlarmType_LowLowState_TransitionTime'
ObjectIdNames[10207] = 'NonExclusiveLevelAlarmType_LowLowState_EffectiveTransitionTime'
ObjectIdNames[10208] = 'NonExclusiveLevelAlarmType_LowLowState_TrueState'
ObjectIdNames[10209] = 'NonExclusiveLevelAlarmType_LowLowState_FalseState'
ObjectIdNames[10210] = 'NonExclusiveLevelAlarmType_HighHighLimit'
ObjectIdNames[10211] = 'NonExclusiveLevelAlarmType_HighLimit'
ObjectIdNames[10212] = 'NonExclusiveLevelAlarmType_LowLimit'
ObjectIdNames[10213] = 'NonExclusiveLevelAlarmType_LowLowLimit'
ObjectIdNames[10214] = 'NonExclusiveRateOfChangeAlarmType'
ObjectIdNames[10215] = 'NonExclusiveRateOfChangeAlarmType_EventId'
ObjectIdNames[10216] = 'NonExclusiveRateOfChangeAlarmType_EventType'
ObjectIdNames[10217] = 'NonExclusiveRateOfChangeAlarmType_SourceNode'
ObjectIdNames[10218] = 'NonExclusiveRateOfChangeAlarmType_SourceName'
ObjectIdNames[10219] = 'NonExclusiveRateOfChangeAlarmType_Time'
ObjectIdNames[10220] = 'NonExclusiveRateOfChangeAlarmType_ReceiveTime'
ObjectIdNames[10221] = 'NonExclusiveRateOfChangeAlarmType_LocalTime'
ObjectIdNames[10222] = 'NonExclusiveRateOfChangeAlarmType_Message'
ObjectIdNames[10223] = 'NonExclusiveRateOfChangeAlarmType_Severity'
ObjectIdNames[10224] = 'NonExclusiveRateOfChangeAlarmType_ConditionName'
ObjectIdNames[10225] = 'NonExclusiveRateOfChangeAlarmType_BranchId'
ObjectIdNames[10226] = 'NonExclusiveRateOfChangeAlarmType_Retain'
ObjectIdNames[10227] = 'NonExclusiveRateOfChangeAlarmType_EnabledState'
ObjectIdNames[10228] = 'NonExclusiveRateOfChangeAlarmType_EnabledState_Id'
ObjectIdNames[10229] = 'NonExclusiveRateOfChangeAlarmType_EnabledState_Name'
ObjectIdNames[10230] = 'NonExclusiveRateOfChangeAlarmType_EnabledState_Number'
ObjectIdNames[10231] = 'NonExclusiveRateOfChangeAlarmType_EnabledState_EffectiveDisplayName'
ObjectIdNames[10232] = 'NonExclusiveRateOfChangeAlarmType_EnabledState_TransitionTime'
ObjectIdNames[10233] = 'NonExclusiveRateOfChangeAlarmType_EnabledState_EffectiveTransitionTime'
ObjectIdNames[10234] = 'NonExclusiveRateOfChangeAlarmType_EnabledState_TrueState'
ObjectIdNames[10235] = 'NonExclusiveRateOfChangeAlarmType_EnabledState_FalseState'
ObjectIdNames[10236] = 'NonExclusiveRateOfChangeAlarmType_Quality'
ObjectIdNames[10237] = 'NonExclusiveRateOfChangeAlarmType_Quality_SourceTimestamp'
ObjectIdNames[10238] = 'NonExclusiveRateOfChangeAlarmType_LastSeverity'
ObjectIdNames[10239] = 'NonExclusiveRateOfChangeAlarmType_LastSeverity_SourceTimestamp'
ObjectIdNames[10240] = 'NonExclusiveRateOfChangeAlarmType_Comment'
ObjectIdNames[10241] = 'NonExclusiveRateOfChangeAlarmType_Comment_SourceTimestamp'
ObjectIdNames[10242] = 'NonExclusiveRateOfChangeAlarmType_ClientUserId'
ObjectIdNames[10243] = 'NonExclusiveRateOfChangeAlarmType_Enable'
ObjectIdNames[10244] = 'NonExclusiveRateOfChangeAlarmType_Disable'
ObjectIdNames[10245] = 'NonExclusiveRateOfChangeAlarmType_AddComment'
ObjectIdNames[10246] = 'NonExclusiveRateOfChangeAlarmType_AddComment_InputArguments'
ObjectIdNames[10247] = 'NonExclusiveRateOfChangeAlarmType_ConditionRefresh'
ObjectIdNames[10248] = 'NonExclusiveRateOfChangeAlarmType_ConditionRefresh_InputArguments'
ObjectIdNames[10249] = 'NonExclusiveRateOfChangeAlarmType_AckedState'
ObjectIdNames[10250] = 'NonExclusiveRateOfChangeAlarmType_AckedState_Id'
ObjectIdNames[10251] = 'NonExclusiveRateOfChangeAlarmType_AckedState_Name'
ObjectIdNames[10252] = 'NonExclusiveRateOfChangeAlarmType_AckedState_Number'
ObjectIdNames[10253] = 'NonExclusiveRateOfChangeAlarmType_AckedState_EffectiveDisplayName'
ObjectIdNames[10254] = 'NonExclusiveRateOfChangeAlarmType_AckedState_TransitionTime'
ObjectIdNames[10255] = 'NonExclusiveRateOfChangeAlarmType_AckedState_EffectiveTransitionTime'
ObjectIdNames[10256] = 'NonExclusiveRateOfChangeAlarmType_AckedState_TrueState'
ObjectIdNames[10257] = 'NonExclusiveRateOfChangeAlarmType_AckedState_FalseState'
ObjectIdNames[10258] = 'NonExclusiveRateOfChangeAlarmType_ConfirmedState'
ObjectIdNames[10259] = 'NonExclusiveRateOfChangeAlarmType_ConfirmedState_Id'
ObjectIdNames[10260] = 'NonExclusiveRateOfChangeAlarmType_ConfirmedState_Name'
ObjectIdNames[10261] = 'NonExclusiveRateOfChangeAlarmType_ConfirmedState_Number'
ObjectIdNames[10262] = 'NonExclusiveRateOfChangeAlarmType_ConfirmedState_EffectiveDisplayName'
ObjectIdNames[10263] = 'NonExclusiveRateOfChangeAlarmType_ConfirmedState_TransitionTime'
ObjectIdNames[10264] = 'NonExclusiveRateOfChangeAlarmType_ConfirmedState_EffectiveTransitionTime'
ObjectIdNames[10265] = 'NonExclusiveRateOfChangeAlarmType_ConfirmedState_TrueState'
ObjectIdNames[10266] = 'NonExclusiveRateOfChangeAlarmType_ConfirmedState_FalseState'
ObjectIdNames[10267] = 'NonExclusiveRateOfChangeAlarmType_Acknowledge'
ObjectIdNames[10268] = 'NonExclusiveRateOfChangeAlarmType_Acknowledge_InputArguments'
ObjectIdNames[10269] = 'NonExclusiveRateOfChangeAlarmType_Confirm'
ObjectIdNames[10270] = 'NonExclusiveRateOfChangeAlarmType_Confirm_InputArguments'
ObjectIdNames[10271] = 'NonExclusiveRateOfChangeAlarmType_ActiveState'
ObjectIdNames[10272] = 'NonExclusiveRateOfChangeAlarmType_ActiveState_Id'
ObjectIdNames[10273] = 'NonExclusiveRateOfChangeAlarmType_ActiveState_Name'
ObjectIdNames[10274] = 'NonExclusiveRateOfChangeAlarmType_ActiveState_Number'
ObjectIdNames[10275] = 'NonExclusiveRateOfChangeAlarmType_ActiveState_EffectiveDisplayName'
ObjectIdNames[10276] = 'NonExclusiveRateOfChangeAlarmType_ActiveState_TransitionTime'
ObjectIdNames[10277] = 'NonExclusiveRateOfChangeAlarmType_ActiveState_EffectiveTransitionTime'
ObjectIdNames[10278] = 'NonExclusiveRateOfChangeAlarmType_ActiveState_TrueState'
ObjectIdNames[10279] = 'NonExclusiveRateOfChangeAlarmType_ActiveState_FalseState'
ObjectIdNames[10280] = 'NonExclusiveRateOfChangeAlarmType_SuppressedState'
ObjectIdNames[10281] = 'NonExclusiveRateOfChangeAlarmType_SuppressedState_Id'
ObjectIdNames[10282] = 'NonExclusiveRateOfChangeAlarmType_SuppressedState_Name'
ObjectIdNames[10283] = 'NonExclusiveRateOfChangeAlarmType_SuppressedState_Number'
ObjectIdNames[10284] = 'NonExclusiveRateOfChangeAlarmType_SuppressedState_EffectiveDisplayName'
ObjectIdNames[10285] = 'NonExclusiveRateOfChangeAlarmType_SuppressedState_TransitionTime'
ObjectIdNames[10286] = 'NonExclusiveRateOfChangeAlarmType_SuppressedState_EffectiveTransitionTime'
ObjectIdNames[10287] = 'NonExclusiveRateOfChangeAlarmType_SuppressedState_TrueState'
ObjectIdNames[10288] = 'NonExclusiveRateOfChangeAlarmType_SuppressedState_FalseState'
ObjectIdNames[10289] = 'NonExclusiveRateOfChangeAlarmType_ShelvingState'
ObjectIdNames[10290] = 'NonExclusiveRateOfChangeAlarmType_ShelvingState_CurrentState'
ObjectIdNames[10291] = 'NonExclusiveRateOfChangeAlarmType_ShelvingState_CurrentState_Id'
ObjectIdNames[10292] = 'NonExclusiveRateOfChangeAlarmType_ShelvingState_CurrentState_Name'
ObjectIdNames[10293] = 'NonExclusiveRateOfChangeAlarmType_ShelvingState_CurrentState_Number'
ObjectIdNames[10294] = 'NonExclusiveRateOfChangeAlarmType_ShelvingState_CurrentState_EffectiveDisplayName'
ObjectIdNames[10295] = 'NonExclusiveRateOfChangeAlarmType_ShelvingState_LastTransition'
ObjectIdNames[10296] = 'NonExclusiveRateOfChangeAlarmType_ShelvingState_LastTransition_Id'
ObjectIdNames[10297] = 'NonExclusiveRateOfChangeAlarmType_ShelvingState_LastTransition_Name'
ObjectIdNames[10298] = 'NonExclusiveRateOfChangeAlarmType_ShelvingState_LastTransition_Number'
ObjectIdNames[10299] = 'NonExclusiveRateOfChangeAlarmType_ShelvingState_LastTransition_TransitionTime'
ObjectIdNames[10300] = 'NonExclusiveRateOfChangeAlarmType_ShelvingState_UnshelveTime'
ObjectIdNames[10322] = 'NonExclusiveRateOfChangeAlarmType_ShelvingState_Unshelve'
ObjectIdNames[10323] = 'NonExclusiveRateOfChangeAlarmType_ShelvingState_OneShotShelve'
ObjectIdNames[10324] = 'NonExclusiveRateOfChangeAlarmType_ShelvingState_TimedShelve'
ObjectIdNames[10325] = 'NonExclusiveRateOfChangeAlarmType_ShelvingState_TimedShelve_InputArguments'
ObjectIdNames[10326] = 'NonExclusiveRateOfChangeAlarmType_SuppressedOrShelved'
ObjectIdNames[10327] = 'NonExclusiveRateOfChangeAlarmType_MaxTimeShelved'
ObjectIdNames[10328] = 'NonExclusiveRateOfChangeAlarmType_HighHighState'
ObjectIdNames[10329] = 'NonExclusiveRateOfChangeAlarmType_HighHighState_Id'
ObjectIdNames[10330] = 'NonExclusiveRateOfChangeAlarmType_HighHighState_Name'
ObjectIdNames[10331] = 'NonExclusiveRateOfChangeAlarmType_HighHighState_Number'
ObjectIdNames[10332] = 'NonExclusiveRateOfChangeAlarmType_HighHighState_EffectiveDisplayName'
ObjectIdNames[10333] = 'NonExclusiveRateOfChangeAlarmType_HighHighState_TransitionTime'
ObjectIdNames[10334] = 'NonExclusiveRateOfChangeAlarmType_HighHighState_EffectiveTransitionTime'
ObjectIdNames[10335] = 'NonExclusiveRateOfChangeAlarmType_HighHighState_TrueState'
ObjectIdNames[10336] = 'NonExclusiveRateOfChangeAlarmType_HighHighState_FalseState'
ObjectIdNames[10337] = 'NonExclusiveRateOfChangeAlarmType_HighState'
ObjectIdNames[10338] = 'NonExclusiveRateOfChangeAlarmType_HighState_Id'
ObjectIdNames[10339] = 'NonExclusiveRateOfChangeAlarmType_HighState_Name'
ObjectIdNames[10340] = 'NonExclusiveRateOfChangeAlarmType_HighState_Number'
ObjectIdNames[10341] = 'NonExclusiveRateOfChangeAlarmType_HighState_EffectiveDisplayName'
ObjectIdNames[10342] = 'NonExclusiveRateOfChangeAlarmType_HighState_TransitionTime'
ObjectIdNames[10343] = 'NonExclusiveRateOfChangeAlarmType_HighState_EffectiveTransitionTime'
ObjectIdNames[10344] = 'NonExclusiveRateOfChangeAlarmType_HighState_TrueState'
ObjectIdNames[10345] = 'NonExclusiveRateOfChangeAlarmType_HighState_FalseState'
ObjectIdNames[10346] = 'NonExclusiveRateOfChangeAlarmType_LowState'
ObjectIdNames[10347] = 'NonExclusiveRateOfChangeAlarmType_LowState_Id'
ObjectIdNames[10348] = 'NonExclusiveRateOfChangeAlarmType_LowState_Name'
ObjectIdNames[10349] = 'NonExclusiveRateOfChangeAlarmType_LowState_Number'
ObjectIdNames[10350] = 'NonExclusiveRateOfChangeAlarmType_LowState_EffectiveDisplayName'
ObjectIdNames[10351] = 'NonExclusiveRateOfChangeAlarmType_LowState_TransitionTime'
ObjectIdNames[10352] = 'NonExclusiveRateOfChangeAlarmType_LowState_EffectiveTransitionTime'
ObjectIdNames[10353] = 'NonExclusiveRateOfChangeAlarmType_LowState_TrueState'
ObjectIdNames[10354] = 'NonExclusiveRateOfChangeAlarmType_LowState_FalseState'
ObjectIdNames[10355] = 'NonExclusiveRateOfChangeAlarmType_LowLowState'
ObjectIdNames[10356] = 'NonExclusiveRateOfChangeAlarmType_LowLowState_Id'
ObjectIdNames[10357] = 'NonExclusiveRateOfChangeAlarmType_LowLowState_Name'
ObjectIdNames[10358] = 'NonExclusiveRateOfChangeAlarmType_LowLowState_Number'
ObjectIdNames[10359] = 'NonExclusiveRateOfChangeAlarmType_LowLowState_EffectiveDisplayName'
ObjectIdNames[10360] = 'NonExclusiveRateOfChangeAlarmType_LowLowState_TransitionTime'
ObjectIdNames[10361] = 'NonExclusiveRateOfChangeAlarmType_LowLowState_EffectiveTransitionTime'
ObjectIdNames[10362] = 'NonExclusiveRateOfChangeAlarmType_LowLowState_TrueState'
ObjectIdNames[10363] = 'NonExclusiveRateOfChangeAlarmType_LowLowState_FalseState'
ObjectIdNames[10364] = 'NonExclusiveRateOfChangeAlarmType_HighHighLimit'
ObjectIdNames[10365] = 'NonExclusiveRateOfChangeAlarmType_HighLimit'
ObjectIdNames[10366] = 'NonExclusiveRateOfChangeAlarmType_LowLimit'
ObjectIdNames[10367] = 'NonExclusiveRateOfChangeAlarmType_LowLowLimit'
ObjectIdNames[10368] = 'NonExclusiveDeviationAlarmType'
ObjectIdNames[10369] = 'NonExclusiveDeviationAlarmType_EventId'
ObjectIdNames[10370] = 'NonExclusiveDeviationAlarmType_EventType'
ObjectIdNames[10371] = 'NonExclusiveDeviationAlarmType_SourceNode'
ObjectIdNames[10372] = 'NonExclusiveDeviationAlarmType_SourceName'
ObjectIdNames[10373] = 'NonExclusiveDeviationAlarmType_Time'
ObjectIdNames[10374] = 'NonExclusiveDeviationAlarmType_ReceiveTime'
ObjectIdNames[10375] = 'NonExclusiveDeviationAlarmType_LocalTime'
ObjectIdNames[10376] = 'NonExclusiveDeviationAlarmType_Message'
ObjectIdNames[10377] = 'NonExclusiveDeviationAlarmType_Severity'
ObjectIdNames[10378] = 'NonExclusiveDeviationAlarmType_ConditionName'
ObjectIdNames[10379] = 'NonExclusiveDeviationAlarmType_BranchId'
ObjectIdNames[10380] = 'NonExclusiveDeviationAlarmType_Retain'
ObjectIdNames[10381] = 'NonExclusiveDeviationAlarmType_EnabledState'
ObjectIdNames[10382] = 'NonExclusiveDeviationAlarmType_EnabledState_Id'
ObjectIdNames[10383] = 'NonExclusiveDeviationAlarmType_EnabledState_Name'
ObjectIdNames[10384] = 'NonExclusiveDeviationAlarmType_EnabledState_Number'
ObjectIdNames[10385] = 'NonExclusiveDeviationAlarmType_EnabledState_EffectiveDisplayName'
ObjectIdNames[10386] = 'NonExclusiveDeviationAlarmType_EnabledState_TransitionTime'
ObjectIdNames[10387] = 'NonExclusiveDeviationAlarmType_EnabledState_EffectiveTransitionTime'
ObjectIdNames[10388] = 'NonExclusiveDeviationAlarmType_EnabledState_TrueState'
ObjectIdNames[10389] = 'NonExclusiveDeviationAlarmType_EnabledState_FalseState'
ObjectIdNames[10390] = 'NonExclusiveDeviationAlarmType_Quality'
ObjectIdNames[10391] = 'NonExclusiveDeviationAlarmType_Quality_SourceTimestamp'
ObjectIdNames[10392] = 'NonExclusiveDeviationAlarmType_LastSeverity'
ObjectIdNames[10393] = 'NonExclusiveDeviationAlarmType_LastSeverity_SourceTimestamp'
ObjectIdNames[10394] = 'NonExclusiveDeviationAlarmType_Comment'
ObjectIdNames[10395] = 'NonExclusiveDeviationAlarmType_Comment_SourceTimestamp'
ObjectIdNames[10396] = 'NonExclusiveDeviationAlarmType_ClientUserId'
ObjectIdNames[10397] = 'NonExclusiveDeviationAlarmType_Enable'
ObjectIdNames[10398] = 'NonExclusiveDeviationAlarmType_Disable'
ObjectIdNames[10399] = 'NonExclusiveDeviationAlarmType_AddComment'
ObjectIdNames[10400] = 'NonExclusiveDeviationAlarmType_AddComment_InputArguments'
ObjectIdNames[10401] = 'NonExclusiveDeviationAlarmType_ConditionRefresh'
ObjectIdNames[10402] = 'NonExclusiveDeviationAlarmType_ConditionRefresh_InputArguments'
ObjectIdNames[10403] = 'NonExclusiveDeviationAlarmType_AckedState'
ObjectIdNames[10404] = 'NonExclusiveDeviationAlarmType_AckedState_Id'
ObjectIdNames[10405] = 'NonExclusiveDeviationAlarmType_AckedState_Name'
ObjectIdNames[10406] = 'NonExclusiveDeviationAlarmType_AckedState_Number'
ObjectIdNames[10407] = 'NonExclusiveDeviationAlarmType_AckedState_EffectiveDisplayName'
ObjectIdNames[10408] = 'NonExclusiveDeviationAlarmType_AckedState_TransitionTime'
ObjectIdNames[10409] = 'NonExclusiveDeviationAlarmType_AckedState_EffectiveTransitionTime'
ObjectIdNames[10410] = 'NonExclusiveDeviationAlarmType_AckedState_TrueState'
ObjectIdNames[10411] = 'NonExclusiveDeviationAlarmType_AckedState_FalseState'
ObjectIdNames[10412] = 'NonExclusiveDeviationAlarmType_ConfirmedState'
ObjectIdNames[10413] = 'NonExclusiveDeviationAlarmType_ConfirmedState_Id'
ObjectIdNames[10414] = 'NonExclusiveDeviationAlarmType_ConfirmedState_Name'
ObjectIdNames[10415] = 'NonExclusiveDeviationAlarmType_ConfirmedState_Number'
ObjectIdNames[10416] = 'NonExclusiveDeviationAlarmType_ConfirmedState_EffectiveDisplayName'
ObjectIdNames[10417] = 'NonExclusiveDeviationAlarmType_ConfirmedState_TransitionTime'
ObjectIdNames[10418] = 'NonExclusiveDeviationAlarmType_ConfirmedState_EffectiveTransitionTime'
ObjectIdNames[10419] = 'NonExclusiveDeviationAlarmType_ConfirmedState_TrueState'
ObjectIdNames[10420] = 'NonExclusiveDeviationAlarmType_ConfirmedState_FalseState'
ObjectIdNames[10421] = 'NonExclusiveDeviationAlarmType_Acknowledge'
ObjectIdNames[10422] = 'NonExclusiveDeviationAlarmType_Acknowledge_InputArguments'
ObjectIdNames[10423] = 'NonExclusiveDeviationAlarmType_Confirm'
ObjectIdNames[10424] = 'NonExclusiveDeviationAlarmType_Confirm_InputArguments'
ObjectIdNames[10425] = 'NonExclusiveDeviationAlarmType_ActiveState'
ObjectIdNames[10426] = 'NonExclusiveDeviationAlarmType_ActiveState_Id'
ObjectIdNames[10427] = 'NonExclusiveDeviationAlarmType_ActiveState_Name'
ObjectIdNames[10428] = 'NonExclusiveDeviationAlarmType_ActiveState_Number'
ObjectIdNames[10429] = 'NonExclusiveDeviationAlarmType_ActiveState_EffectiveDisplayName'
ObjectIdNames[10430] = 'NonExclusiveDeviationAlarmType_ActiveState_TransitionTime'
ObjectIdNames[10431] = 'NonExclusiveDeviationAlarmType_ActiveState_EffectiveTransitionTime'
ObjectIdNames[10432] = 'NonExclusiveDeviationAlarmType_ActiveState_TrueState'
ObjectIdNames[10433] = 'NonExclusiveDeviationAlarmType_ActiveState_FalseState'
ObjectIdNames[10434] = 'NonExclusiveDeviationAlarmType_SuppressedState'
ObjectIdNames[10435] = 'NonExclusiveDeviationAlarmType_SuppressedState_Id'
ObjectIdNames[10436] = 'NonExclusiveDeviationAlarmType_SuppressedState_Name'
ObjectIdNames[10437] = 'NonExclusiveDeviationAlarmType_SuppressedState_Number'
ObjectIdNames[10438] = 'NonExclusiveDeviationAlarmType_SuppressedState_EffectiveDisplayName'
ObjectIdNames[10439] = 'NonExclusiveDeviationAlarmType_SuppressedState_TransitionTime'
ObjectIdNames[10440] = 'NonExclusiveDeviationAlarmType_SuppressedState_EffectiveTransitionTime'
ObjectIdNames[10441] = 'NonExclusiveDeviationAlarmType_SuppressedState_TrueState'
ObjectIdNames[10442] = 'NonExclusiveDeviationAlarmType_SuppressedState_FalseState'
ObjectIdNames[10443] = 'NonExclusiveDeviationAlarmType_ShelvingState'
ObjectIdNames[10444] = 'NonExclusiveDeviationAlarmType_ShelvingState_CurrentState'
ObjectIdNames[10445] = 'NonExclusiveDeviationAlarmType_ShelvingState_CurrentState_Id'
ObjectIdNames[10446] = 'NonExclusiveDeviationAlarmType_ShelvingState_CurrentState_Name'
ObjectIdNames[10447] = 'NonExclusiveDeviationAlarmType_ShelvingState_CurrentState_Number'
ObjectIdNames[10448] = 'NonExclusiveDeviationAlarmType_ShelvingState_CurrentState_EffectiveDisplayName'
ObjectIdNames[10449] = 'NonExclusiveDeviationAlarmType_ShelvingState_LastTransition'
ObjectIdNames[10450] = 'NonExclusiveDeviationAlarmType_ShelvingState_LastTransition_Id'
ObjectIdNames[10451] = 'NonExclusiveDeviationAlarmType_ShelvingState_LastTransition_Name'
ObjectIdNames[10452] = 'NonExclusiveDeviationAlarmType_ShelvingState_LastTransition_Number'
ObjectIdNames[10453] = 'NonExclusiveDeviationAlarmType_ShelvingState_LastTransition_TransitionTime'
ObjectIdNames[10454] = 'NonExclusiveDeviationAlarmType_ShelvingState_UnshelveTime'
ObjectIdNames[10476] = 'NonExclusiveDeviationAlarmType_ShelvingState_Unshelve'
ObjectIdNames[10477] = 'NonExclusiveDeviationAlarmType_ShelvingState_OneShotShelve'
ObjectIdNames[10478] = 'NonExclusiveDeviationAlarmType_ShelvingState_TimedShelve'
ObjectIdNames[10479] = 'NonExclusiveDeviationAlarmType_ShelvingState_TimedShelve_InputArguments'
ObjectIdNames[10480] = 'NonExclusiveDeviationAlarmType_SuppressedOrShelved'
ObjectIdNames[10481] = 'NonExclusiveDeviationAlarmType_MaxTimeShelved'
ObjectIdNames[10482] = 'NonExclusiveDeviationAlarmType_HighHighState'
ObjectIdNames[10483] = 'NonExclusiveDeviationAlarmType_HighHighState_Id'
ObjectIdNames[10484] = 'NonExclusiveDeviationAlarmType_HighHighState_Name'
ObjectIdNames[10485] = 'NonExclusiveDeviationAlarmType_HighHighState_Number'
ObjectIdNames[10486] = 'NonExclusiveDeviationAlarmType_HighHighState_EffectiveDisplayName'
ObjectIdNames[10487] = 'NonExclusiveDeviationAlarmType_HighHighState_TransitionTime'
ObjectIdNames[10488] = 'NonExclusiveDeviationAlarmType_HighHighState_EffectiveTransitionTime'
ObjectIdNames[10489] = 'NonExclusiveDeviationAlarmType_HighHighState_TrueState'
ObjectIdNames[10490] = 'NonExclusiveDeviationAlarmType_HighHighState_FalseState'
ObjectIdNames[10491] = 'NonExclusiveDeviationAlarmType_HighState'
ObjectIdNames[10492] = 'NonExclusiveDeviationAlarmType_HighState_Id'
ObjectIdNames[10493] = 'NonExclusiveDeviationAlarmType_HighState_Name'
ObjectIdNames[10494] = 'NonExclusiveDeviationAlarmType_HighState_Number'
ObjectIdNames[10495] = 'NonExclusiveDeviationAlarmType_HighState_EffectiveDisplayName'
ObjectIdNames[10496] = 'NonExclusiveDeviationAlarmType_HighState_TransitionTime'
ObjectIdNames[10497] = 'NonExclusiveDeviationAlarmType_HighState_EffectiveTransitionTime'
ObjectIdNames[10498] = 'NonExclusiveDeviationAlarmType_HighState_TrueState'
ObjectIdNames[10499] = 'NonExclusiveDeviationAlarmType_HighState_FalseState'
ObjectIdNames[10500] = 'NonExclusiveDeviationAlarmType_LowState'
ObjectIdNames[10501] = 'NonExclusiveDeviationAlarmType_LowState_Id'
ObjectIdNames[10502] = 'NonExclusiveDeviationAlarmType_LowState_Name'
ObjectIdNames[10503] = 'NonExclusiveDeviationAlarmType_LowState_Number'
ObjectIdNames[10504] = 'NonExclusiveDeviationAlarmType_LowState_EffectiveDisplayName'
ObjectIdNames[10505] = 'NonExclusiveDeviationAlarmType_LowState_TransitionTime'
ObjectIdNames[10506] = 'NonExclusiveDeviationAlarmType_LowState_EffectiveTransitionTime'
ObjectIdNames[10507] = 'NonExclusiveDeviationAlarmType_LowState_TrueState'
ObjectIdNames[10508] = 'NonExclusiveDeviationAlarmType_LowState_FalseState'
ObjectIdNames[10509] = 'NonExclusiveDeviationAlarmType_LowLowState'
ObjectIdNames[10510] = 'NonExclusiveDeviationAlarmType_LowLowState_Id'
ObjectIdNames[10511] = 'NonExclusiveDeviationAlarmType_LowLowState_Name'
ObjectIdNames[10512] = 'NonExclusiveDeviationAlarmType_LowLowState_Number'
ObjectIdNames[10513] = 'NonExclusiveDeviationAlarmType_LowLowState_EffectiveDisplayName'
ObjectIdNames[10514] = 'NonExclusiveDeviationAlarmType_LowLowState_TransitionTime'
ObjectIdNames[10515] = 'NonExclusiveDeviationAlarmType_LowLowState_EffectiveTransitionTime'
ObjectIdNames[10516] = 'NonExclusiveDeviationAlarmType_LowLowState_TrueState'
ObjectIdNames[10517] = 'NonExclusiveDeviationAlarmType_LowLowState_FalseState'
ObjectIdNames[10518] = 'NonExclusiveDeviationAlarmType_HighHighLimit'
ObjectIdNames[10519] = 'NonExclusiveDeviationAlarmType_HighLimit'
ObjectIdNames[10520] = 'NonExclusiveDeviationAlarmType_LowLimit'
ObjectIdNames[10521] = 'NonExclusiveDeviationAlarmType_LowLowLimit'
ObjectIdNames[10522] = 'NonExclusiveDeviationAlarmType_SetpointNode'
ObjectIdNames[10523] = 'DiscreteAlarmType'
ObjectIdNames[10524] = 'DiscreteAlarmType_EventId'
ObjectIdNames[10525] = 'DiscreteAlarmType_EventType'
ObjectIdNames[10526] = 'DiscreteAlarmType_SourceNode'
ObjectIdNames[10527] = 'DiscreteAlarmType_SourceName'
ObjectIdNames[10528] = 'DiscreteAlarmType_Time'
ObjectIdNames[10529] = 'DiscreteAlarmType_ReceiveTime'
ObjectIdNames[10530] = 'DiscreteAlarmType_LocalTime'
ObjectIdNames[10531] = 'DiscreteAlarmType_Message'
ObjectIdNames[10532] = 'DiscreteAlarmType_Severity'
ObjectIdNames[10533] = 'DiscreteAlarmType_ConditionName'
ObjectIdNames[10534] = 'DiscreteAlarmType_BranchId'
ObjectIdNames[10535] = 'DiscreteAlarmType_Retain'
ObjectIdNames[10536] = 'DiscreteAlarmType_EnabledState'
ObjectIdNames[10537] = 'DiscreteAlarmType_EnabledState_Id'
ObjectIdNames[10538] = 'DiscreteAlarmType_EnabledState_Name'
ObjectIdNames[10539] = 'DiscreteAlarmType_EnabledState_Number'
ObjectIdNames[10540] = 'DiscreteAlarmType_EnabledState_EffectiveDisplayName'
ObjectIdNames[10541] = 'DiscreteAlarmType_EnabledState_TransitionTime'
ObjectIdNames[10542] = 'DiscreteAlarmType_EnabledState_EffectiveTransitionTime'
ObjectIdNames[10543] = 'DiscreteAlarmType_EnabledState_TrueState'
ObjectIdNames[10544] = 'DiscreteAlarmType_EnabledState_FalseState'
ObjectIdNames[10545] = 'DiscreteAlarmType_Quality'
ObjectIdNames[10546] = 'DiscreteAlarmType_Quality_SourceTimestamp'
ObjectIdNames[10547] = 'DiscreteAlarmType_LastSeverity'
ObjectIdNames[10548] = 'DiscreteAlarmType_LastSeverity_SourceTimestamp'
ObjectIdNames[10549] = 'DiscreteAlarmType_Comment'
ObjectIdNames[10550] = 'DiscreteAlarmType_Comment_SourceTimestamp'
ObjectIdNames[10551] = 'DiscreteAlarmType_ClientUserId'
ObjectIdNames[10552] = 'DiscreteAlarmType_Enable'
ObjectIdNames[10553] = 'DiscreteAlarmType_Disable'
ObjectIdNames[10554] = 'DiscreteAlarmType_AddComment'
ObjectIdNames[10555] = 'DiscreteAlarmType_AddComment_InputArguments'
ObjectIdNames[10556] = 'DiscreteAlarmType_ConditionRefresh'
ObjectIdNames[10557] = 'DiscreteAlarmType_ConditionRefresh_InputArguments'
ObjectIdNames[10558] = 'DiscreteAlarmType_AckedState'
ObjectIdNames[10559] = 'DiscreteAlarmType_AckedState_Id'
ObjectIdNames[10560] = 'DiscreteAlarmType_AckedState_Name'
ObjectIdNames[10561] = 'DiscreteAlarmType_AckedState_Number'
ObjectIdNames[10562] = 'DiscreteAlarmType_AckedState_EffectiveDisplayName'
ObjectIdNames[10563] = 'DiscreteAlarmType_AckedState_TransitionTime'
ObjectIdNames[10564] = 'DiscreteAlarmType_AckedState_EffectiveTransitionTime'
ObjectIdNames[10565] = 'DiscreteAlarmType_AckedState_TrueState'
ObjectIdNames[10566] = 'DiscreteAlarmType_AckedState_FalseState'
ObjectIdNames[10567] = 'DiscreteAlarmType_ConfirmedState'
ObjectIdNames[10568] = 'DiscreteAlarmType_ConfirmedState_Id'
ObjectIdNames[10569] = 'DiscreteAlarmType_ConfirmedState_Name'
ObjectIdNames[10570] = 'DiscreteAlarmType_ConfirmedState_Number'
ObjectIdNames[10571] = 'DiscreteAlarmType_ConfirmedState_EffectiveDisplayName'
ObjectIdNames[10572] = 'DiscreteAlarmType_ConfirmedState_TransitionTime'
ObjectIdNames[10573] = 'DiscreteAlarmType_ConfirmedState_EffectiveTransitionTime'
ObjectIdNames[10574] = 'DiscreteAlarmType_ConfirmedState_TrueState'
ObjectIdNames[10575] = 'DiscreteAlarmType_ConfirmedState_FalseState'
ObjectIdNames[10576] = 'DiscreteAlarmType_Acknowledge'
ObjectIdNames[10577] = 'DiscreteAlarmType_Acknowledge_InputArguments'
ObjectIdNames[10578] = 'DiscreteAlarmType_Confirm'
ObjectIdNames[10579] = 'DiscreteAlarmType_Confirm_InputArguments'
ObjectIdNames[10580] = 'DiscreteAlarmType_ActiveState'
ObjectIdNames[10581] = 'DiscreteAlarmType_ActiveState_Id'
ObjectIdNames[10582] = 'DiscreteAlarmType_ActiveState_Name'
ObjectIdNames[10583] = 'DiscreteAlarmType_ActiveState_Number'
ObjectIdNames[10584] = 'DiscreteAlarmType_ActiveState_EffectiveDisplayName'
ObjectIdNames[10585] = 'DiscreteAlarmType_ActiveState_TransitionTime'
ObjectIdNames[10586] = 'DiscreteAlarmType_ActiveState_EffectiveTransitionTime'
ObjectIdNames[10587] = 'DiscreteAlarmType_ActiveState_TrueState'
ObjectIdNames[10588] = 'DiscreteAlarmType_ActiveState_FalseState'
ObjectIdNames[10589] = 'DiscreteAlarmType_SuppressedState'
ObjectIdNames[10590] = 'DiscreteAlarmType_SuppressedState_Id'
ObjectIdNames[10591] = 'DiscreteAlarmType_SuppressedState_Name'
ObjectIdNames[10592] = 'DiscreteAlarmType_SuppressedState_Number'
ObjectIdNames[10593] = 'DiscreteAlarmType_SuppressedState_EffectiveDisplayName'
ObjectIdNames[10594] = 'DiscreteAlarmType_SuppressedState_TransitionTime'
ObjectIdNames[10595] = 'DiscreteAlarmType_SuppressedState_EffectiveTransitionTime'
ObjectIdNames[10596] = 'DiscreteAlarmType_SuppressedState_TrueState'
ObjectIdNames[10597] = 'DiscreteAlarmType_SuppressedState_FalseState'
ObjectIdNames[10598] = 'DiscreteAlarmType_ShelvingState'
ObjectIdNames[10599] = 'DiscreteAlarmType_ShelvingState_CurrentState'
ObjectIdNames[10600] = 'DiscreteAlarmType_ShelvingState_CurrentState_Id'
ObjectIdNames[10601] = 'DiscreteAlarmType_ShelvingState_CurrentState_Name'
ObjectIdNames[10602] = 'DiscreteAlarmType_ShelvingState_CurrentState_Number'
ObjectIdNames[10603] = 'DiscreteAlarmType_ShelvingState_CurrentState_EffectiveDisplayName'
ObjectIdNames[10604] = 'DiscreteAlarmType_ShelvingState_LastTransition'
ObjectIdNames[10605] = 'DiscreteAlarmType_ShelvingState_LastTransition_Id'
ObjectIdNames[10606] = 'DiscreteAlarmType_ShelvingState_LastTransition_Name'
ObjectIdNames[10607] = 'DiscreteAlarmType_ShelvingState_LastTransition_Number'
ObjectIdNames[10608] = 'DiscreteAlarmType_ShelvingState_LastTransition_TransitionTime'
ObjectIdNames[10609] = 'DiscreteAlarmType_ShelvingState_UnshelveTime'
ObjectIdNames[10631] = 'DiscreteAlarmType_ShelvingState_Unshelve'
ObjectIdNames[10632] = 'DiscreteAlarmType_ShelvingState_OneShotShelve'
ObjectIdNames[10633] = 'DiscreteAlarmType_ShelvingState_TimedShelve'
ObjectIdNames[10634] = 'DiscreteAlarmType_ShelvingState_TimedShelve_InputArguments'
ObjectIdNames[10635] = 'DiscreteAlarmType_SuppressedOrShelved'
ObjectIdNames[10636] = 'DiscreteAlarmType_MaxTimeShelved'
ObjectIdNames[10637] = 'OffNormalAlarmType'
ObjectIdNames[10638] = 'OffNormalAlarmType_EventId'
ObjectIdNames[10639] = 'OffNormalAlarmType_EventType'
ObjectIdNames[10640] = 'OffNormalAlarmType_SourceNode'
ObjectIdNames[10641] = 'OffNormalAlarmType_SourceName'
ObjectIdNames[10642] = 'OffNormalAlarmType_Time'
ObjectIdNames[10643] = 'OffNormalAlarmType_ReceiveTime'
ObjectIdNames[10644] = 'OffNormalAlarmType_LocalTime'
ObjectIdNames[10645] = 'OffNormalAlarmType_Message'
ObjectIdNames[10646] = 'OffNormalAlarmType_Severity'
ObjectIdNames[10647] = 'OffNormalAlarmType_ConditionName'
ObjectIdNames[10648] = 'OffNormalAlarmType_BranchId'
ObjectIdNames[10649] = 'OffNormalAlarmType_Retain'
ObjectIdNames[10650] = 'OffNormalAlarmType_EnabledState'
ObjectIdNames[10651] = 'OffNormalAlarmType_EnabledState_Id'
ObjectIdNames[10652] = 'OffNormalAlarmType_EnabledState_Name'
ObjectIdNames[10653] = 'OffNormalAlarmType_EnabledState_Number'
ObjectIdNames[10654] = 'OffNormalAlarmType_EnabledState_EffectiveDisplayName'
ObjectIdNames[10655] = 'OffNormalAlarmType_EnabledState_TransitionTime'
ObjectIdNames[10656] = 'OffNormalAlarmType_EnabledState_EffectiveTransitionTime'
ObjectIdNames[10657] = 'OffNormalAlarmType_EnabledState_TrueState'
ObjectIdNames[10658] = 'OffNormalAlarmType_EnabledState_FalseState'
ObjectIdNames[10659] = 'OffNormalAlarmType_Quality'
ObjectIdNames[10660] = 'OffNormalAlarmType_Quality_SourceTimestamp'
ObjectIdNames[10661] = 'OffNormalAlarmType_LastSeverity'
ObjectIdNames[10662] = 'OffNormalAlarmType_LastSeverity_SourceTimestamp'
ObjectIdNames[10663] = 'OffNormalAlarmType_Comment'
ObjectIdNames[10664] = 'OffNormalAlarmType_Comment_SourceTimestamp'
ObjectIdNames[10665] = 'OffNormalAlarmType_ClientUserId'
ObjectIdNames[10666] = 'OffNormalAlarmType_Enable'
ObjectIdNames[10667] = 'OffNormalAlarmType_Disable'
ObjectIdNames[10668] = 'OffNormalAlarmType_AddComment'
ObjectIdNames[10669] = 'OffNormalAlarmType_AddComment_InputArguments'
ObjectIdNames[10670] = 'OffNormalAlarmType_ConditionRefresh'
ObjectIdNames[10671] = 'OffNormalAlarmType_ConditionRefresh_InputArguments'
ObjectIdNames[10672] = 'OffNormalAlarmType_AckedState'
ObjectIdNames[10673] = 'OffNormalAlarmType_AckedState_Id'
ObjectIdNames[10674] = 'OffNormalAlarmType_AckedState_Name'
ObjectIdNames[10675] = 'OffNormalAlarmType_AckedState_Number'
ObjectIdNames[10676] = 'OffNormalAlarmType_AckedState_EffectiveDisplayName'
ObjectIdNames[10677] = 'OffNormalAlarmType_AckedState_TransitionTime'
ObjectIdNames[10678] = 'OffNormalAlarmType_AckedState_EffectiveTransitionTime'
ObjectIdNames[10679] = 'OffNormalAlarmType_AckedState_TrueState'
ObjectIdNames[10680] = 'OffNormalAlarmType_AckedState_FalseState'
ObjectIdNames[10681] = 'OffNormalAlarmType_ConfirmedState'
ObjectIdNames[10682] = 'OffNormalAlarmType_ConfirmedState_Id'
ObjectIdNames[10683] = 'OffNormalAlarmType_ConfirmedState_Name'
ObjectIdNames[10684] = 'OffNormalAlarmType_ConfirmedState_Number'
ObjectIdNames[10685] = 'OffNormalAlarmType_ConfirmedState_EffectiveDisplayName'
ObjectIdNames[10686] = 'OffNormalAlarmType_ConfirmedState_TransitionTime'
ObjectIdNames[10687] = 'OffNormalAlarmType_ConfirmedState_EffectiveTransitionTime'
ObjectIdNames[10688] = 'OffNormalAlarmType_ConfirmedState_TrueState'
ObjectIdNames[10689] = 'OffNormalAlarmType_ConfirmedState_FalseState'
ObjectIdNames[10690] = 'OffNormalAlarmType_Acknowledge'
ObjectIdNames[10691] = 'OffNormalAlarmType_Acknowledge_InputArguments'
ObjectIdNames[10692] = 'OffNormalAlarmType_Confirm'
ObjectIdNames[10693] = 'OffNormalAlarmType_Confirm_InputArguments'
ObjectIdNames[10694] = 'OffNormalAlarmType_ActiveState'
ObjectIdNames[10695] = 'OffNormalAlarmType_ActiveState_Id'
ObjectIdNames[10696] = 'OffNormalAlarmType_ActiveState_Name'
ObjectIdNames[10697] = 'OffNormalAlarmType_ActiveState_Number'
ObjectIdNames[10698] = 'OffNormalAlarmType_ActiveState_EffectiveDisplayName'
ObjectIdNames[10699] = 'OffNormalAlarmType_ActiveState_TransitionTime'
ObjectIdNames[10700] = 'OffNormalAlarmType_ActiveState_EffectiveTransitionTime'
ObjectIdNames[10701] = 'OffNormalAlarmType_ActiveState_TrueState'
ObjectIdNames[10702] = 'OffNormalAlarmType_ActiveState_FalseState'
ObjectIdNames[10703] = 'OffNormalAlarmType_SuppressedState'
ObjectIdNames[10704] = 'OffNormalAlarmType_SuppressedState_Id'
ObjectIdNames[10705] = 'OffNormalAlarmType_SuppressedState_Name'
ObjectIdNames[10706] = 'OffNormalAlarmType_SuppressedState_Number'
ObjectIdNames[10707] = 'OffNormalAlarmType_SuppressedState_EffectiveDisplayName'
ObjectIdNames[10708] = 'OffNormalAlarmType_SuppressedState_TransitionTime'
ObjectIdNames[10709] = 'OffNormalAlarmType_SuppressedState_EffectiveTransitionTime'
ObjectIdNames[10710] = 'OffNormalAlarmType_SuppressedState_TrueState'
ObjectIdNames[10711] = 'OffNormalAlarmType_SuppressedState_FalseState'
ObjectIdNames[10712] = 'OffNormalAlarmType_ShelvingState'
ObjectIdNames[10713] = 'OffNormalAlarmType_ShelvingState_CurrentState'
ObjectIdNames[10714] = 'OffNormalAlarmType_ShelvingState_CurrentState_Id'
ObjectIdNames[10715] = 'OffNormalAlarmType_ShelvingState_CurrentState_Name'
ObjectIdNames[10716] = 'OffNormalAlarmType_ShelvingState_CurrentState_Number'
ObjectIdNames[10717] = 'OffNormalAlarmType_ShelvingState_CurrentState_EffectiveDisplayName'
ObjectIdNames[10718] = 'OffNormalAlarmType_ShelvingState_LastTransition'
ObjectIdNames[10719] = 'OffNormalAlarmType_ShelvingState_LastTransition_Id'
ObjectIdNames[10720] = 'OffNormalAlarmType_ShelvingState_LastTransition_Name'
ObjectIdNames[10721] = 'OffNormalAlarmType_ShelvingState_LastTransition_Number'
ObjectIdNames[10722] = 'OffNormalAlarmType_ShelvingState_LastTransition_TransitionTime'
ObjectIdNames[10723] = 'OffNormalAlarmType_ShelvingState_UnshelveTime'
ObjectIdNames[10745] = 'OffNormalAlarmType_ShelvingState_Unshelve'
ObjectIdNames[10746] = 'OffNormalAlarmType_ShelvingState_OneShotShelve'
ObjectIdNames[10747] = 'OffNormalAlarmType_ShelvingState_TimedShelve'
ObjectIdNames[10748] = 'OffNormalAlarmType_ShelvingState_TimedShelve_InputArguments'
ObjectIdNames[10749] = 'OffNormalAlarmType_SuppressedOrShelved'
ObjectIdNames[10750] = 'OffNormalAlarmType_MaxTimeShelved'
ObjectIdNames[10751] = 'TripAlarmType'
ObjectIdNames[10752] = 'TripAlarmType_EventId'
ObjectIdNames[10753] = 'TripAlarmType_EventType'
ObjectIdNames[10754] = 'TripAlarmType_SourceNode'
ObjectIdNames[10755] = 'TripAlarmType_SourceName'
ObjectIdNames[10756] = 'TripAlarmType_Time'
ObjectIdNames[10757] = 'TripAlarmType_ReceiveTime'
ObjectIdNames[10758] = 'TripAlarmType_LocalTime'
ObjectIdNames[10759] = 'TripAlarmType_Message'
ObjectIdNames[10760] = 'TripAlarmType_Severity'
ObjectIdNames[10761] = 'TripAlarmType_ConditionName'
ObjectIdNames[10762] = 'TripAlarmType_BranchId'
ObjectIdNames[10763] = 'TripAlarmType_Retain'
ObjectIdNames[10764] = 'TripAlarmType_EnabledState'
ObjectIdNames[10765] = 'TripAlarmType_EnabledState_Id'
ObjectIdNames[10766] = 'TripAlarmType_EnabledState_Name'
ObjectIdNames[10767] = 'TripAlarmType_EnabledState_Number'
ObjectIdNames[10768] = 'TripAlarmType_EnabledState_EffectiveDisplayName'
ObjectIdNames[10769] = 'TripAlarmType_EnabledState_TransitionTime'
ObjectIdNames[10770] = 'TripAlarmType_EnabledState_EffectiveTransitionTime'
ObjectIdNames[10771] = 'TripAlarmType_EnabledState_TrueState'
ObjectIdNames[10772] = 'TripAlarmType_EnabledState_FalseState'
ObjectIdNames[10773] = 'TripAlarmType_Quality'
ObjectIdNames[10774] = 'TripAlarmType_Quality_SourceTimestamp'
ObjectIdNames[10775] = 'TripAlarmType_LastSeverity'
ObjectIdNames[10776] = 'TripAlarmType_LastSeverity_SourceTimestamp'
ObjectIdNames[10777] = 'TripAlarmType_Comment'
ObjectIdNames[10778] = 'TripAlarmType_Comment_SourceTimestamp'
ObjectIdNames[10779] = 'TripAlarmType_ClientUserId'
ObjectIdNames[10780] = 'TripAlarmType_Enable'
ObjectIdNames[10781] = 'TripAlarmType_Disable'
ObjectIdNames[10782] = 'TripAlarmType_AddComment'
ObjectIdNames[10783] = 'TripAlarmType_AddComment_InputArguments'
ObjectIdNames[10784] = 'TripAlarmType_ConditionRefresh'
ObjectIdNames[10785] = 'TripAlarmType_ConditionRefresh_InputArguments'
ObjectIdNames[10786] = 'TripAlarmType_AckedState'
ObjectIdNames[10787] = 'TripAlarmType_AckedState_Id'
ObjectIdNames[10788] = 'TripAlarmType_AckedState_Name'
ObjectIdNames[10789] = 'TripAlarmType_AckedState_Number'
ObjectIdNames[10790] = 'TripAlarmType_AckedState_EffectiveDisplayName'
ObjectIdNames[10791] = 'TripAlarmType_AckedState_TransitionTime'
ObjectIdNames[10792] = 'TripAlarmType_AckedState_EffectiveTransitionTime'
ObjectIdNames[10793] = 'TripAlarmType_AckedState_TrueState'
ObjectIdNames[10794] = 'TripAlarmType_AckedState_FalseState'
ObjectIdNames[10795] = 'TripAlarmType_ConfirmedState'
ObjectIdNames[10796] = 'TripAlarmType_ConfirmedState_Id'
ObjectIdNames[10797] = 'TripAlarmType_ConfirmedState_Name'
ObjectIdNames[10798] = 'TripAlarmType_ConfirmedState_Number'
ObjectIdNames[10799] = 'TripAlarmType_ConfirmedState_EffectiveDisplayName'
ObjectIdNames[10800] = 'TripAlarmType_ConfirmedState_TransitionTime'
ObjectIdNames[10801] = 'TripAlarmType_ConfirmedState_EffectiveTransitionTime'
ObjectIdNames[10802] = 'TripAlarmType_ConfirmedState_TrueState'
ObjectIdNames[10803] = 'TripAlarmType_ConfirmedState_FalseState'
ObjectIdNames[10804] = 'TripAlarmType_Acknowledge'
ObjectIdNames[10805] = 'TripAlarmType_Acknowledge_InputArguments'
ObjectIdNames[10806] = 'TripAlarmType_Confirm'
ObjectIdNames[10807] = 'TripAlarmType_Confirm_InputArguments'
ObjectIdNames[10808] = 'TripAlarmType_ActiveState'
ObjectIdNames[10809] = 'TripAlarmType_ActiveState_Id'
ObjectIdNames[10810] = 'TripAlarmType_ActiveState_Name'
ObjectIdNames[10811] = 'TripAlarmType_ActiveState_Number'
ObjectIdNames[10812] = 'TripAlarmType_ActiveState_EffectiveDisplayName'
ObjectIdNames[10813] = 'TripAlarmType_ActiveState_TransitionTime'
ObjectIdNames[10814] = 'TripAlarmType_ActiveState_EffectiveTransitionTime'
ObjectIdNames[10815] = 'TripAlarmType_ActiveState_TrueState'
ObjectIdNames[10816] = 'TripAlarmType_ActiveState_FalseState'
ObjectIdNames[10817] = 'TripAlarmType_SuppressedState'
ObjectIdNames[10818] = 'TripAlarmType_SuppressedState_Id'
ObjectIdNames[10819] = 'TripAlarmType_SuppressedState_Name'
ObjectIdNames[10820] = 'TripAlarmType_SuppressedState_Number'
ObjectIdNames[10821] = 'TripAlarmType_SuppressedState_EffectiveDisplayName'
ObjectIdNames[10822] = 'TripAlarmType_SuppressedState_TransitionTime'
ObjectIdNames[10823] = 'TripAlarmType_SuppressedState_EffectiveTransitionTime'
ObjectIdNames[10824] = 'TripAlarmType_SuppressedState_TrueState'
ObjectIdNames[10825] = 'TripAlarmType_SuppressedState_FalseState'
ObjectIdNames[10826] = 'TripAlarmType_ShelvingState'
ObjectIdNames[10827] = 'TripAlarmType_ShelvingState_CurrentState'
ObjectIdNames[10828] = 'TripAlarmType_ShelvingState_CurrentState_Id'
ObjectIdNames[10829] = 'TripAlarmType_ShelvingState_CurrentState_Name'
ObjectIdNames[10830] = 'TripAlarmType_ShelvingState_CurrentState_Number'
ObjectIdNames[10831] = 'TripAlarmType_ShelvingState_CurrentState_EffectiveDisplayName'
ObjectIdNames[10832] = 'TripAlarmType_ShelvingState_LastTransition'
ObjectIdNames[10833] = 'TripAlarmType_ShelvingState_LastTransition_Id'
ObjectIdNames[10834] = 'TripAlarmType_ShelvingState_LastTransition_Name'
ObjectIdNames[10835] = 'TripAlarmType_ShelvingState_LastTransition_Number'
ObjectIdNames[10836] = 'TripAlarmType_ShelvingState_LastTransition_TransitionTime'
ObjectIdNames[10837] = 'TripAlarmType_ShelvingState_UnshelveTime'
ObjectIdNames[10859] = 'TripAlarmType_ShelvingState_Unshelve'
ObjectIdNames[10860] = 'TripAlarmType_ShelvingState_OneShotShelve'
ObjectIdNames[10861] = 'TripAlarmType_ShelvingState_TimedShelve'
ObjectIdNames[10862] = 'TripAlarmType_ShelvingState_TimedShelve_InputArguments'
ObjectIdNames[10863] = 'TripAlarmType_SuppressedOrShelved'
ObjectIdNames[10864] = 'TripAlarmType_MaxTimeShelved'
ObjectIdNames[11093] = 'AuditConditionShelvingEventType'
ObjectIdNames[11094] = 'AuditConditionShelvingEventType_EventId'
ObjectIdNames[11095] = 'AuditConditionShelvingEventType_EventType'
ObjectIdNames[11096] = 'AuditConditionShelvingEventType_SourceNode'
ObjectIdNames[11097] = 'AuditConditionShelvingEventType_SourceName'
ObjectIdNames[11098] = 'AuditConditionShelvingEventType_Time'
ObjectIdNames[11099] = 'AuditConditionShelvingEventType_ReceiveTime'
ObjectIdNames[11100] = 'AuditConditionShelvingEventType_LocalTime'
ObjectIdNames[11101] = 'AuditConditionShelvingEventType_Message'
ObjectIdNames[11102] = 'AuditConditionShelvingEventType_Severity'
ObjectIdNames[11103] = 'AuditConditionShelvingEventType_ActionTimeStamp'
ObjectIdNames[11104] = 'AuditConditionShelvingEventType_Status'
ObjectIdNames[11105] = 'AuditConditionShelvingEventType_ServerId'
ObjectIdNames[11106] = 'AuditConditionShelvingEventType_ClientAuditEntryId'
ObjectIdNames[11107] = 'AuditConditionShelvingEventType_ClientUserId'
ObjectIdNames[11108] = 'AuditConditionShelvingEventType_MethodId'
ObjectIdNames[11109] = 'AuditConditionShelvingEventType_InputArguments'
ObjectIdNames[11110] = 'TwoStateVariableType_TrueState'
ObjectIdNames[11111] = 'TwoStateVariableType_FalseState'
ObjectIdNames[11112] = 'ConditionType_ConditionClassId'
ObjectIdNames[11113] = 'ConditionType_ConditionClassName'
ObjectIdNames[11114] = 'DialogConditionType_ConditionClassId'
ObjectIdNames[11115] = 'DialogConditionType_ConditionClassName'
ObjectIdNames[11116] = 'AcknowledgeableConditionType_ConditionClassId'
ObjectIdNames[11117] = 'AcknowledgeableConditionType_ConditionClassName'
ObjectIdNames[11118] = 'AlarmConditionType_ConditionClassId'
ObjectIdNames[11119] = 'AlarmConditionType_ConditionClassName'
ObjectIdNames[11120] = 'AlarmConditionType_InputNode'
ObjectIdNames[11121] = 'LimitAlarmType_ConditionClassId'
ObjectIdNames[11122] = 'LimitAlarmType_ConditionClassName'
ObjectIdNames[11123] = 'LimitAlarmType_InputNode'
ObjectIdNames[11124] = 'LimitAlarmType_HighHighLimit'
ObjectIdNames[11125] = 'LimitAlarmType_HighLimit'
ObjectIdNames[11126] = 'LimitAlarmType_LowLimit'
ObjectIdNames[11127] = 'LimitAlarmType_LowLowLimit'
ObjectIdNames[11128] = 'ExclusiveLimitAlarmType_ConditionClassId'
ObjectIdNames[11129] = 'ExclusiveLimitAlarmType_ConditionClassName'
ObjectIdNames[11130] = 'ExclusiveLimitAlarmType_InputNode'
ObjectIdNames[11131] = 'ExclusiveLevelAlarmType_ConditionClassId'
ObjectIdNames[11132] = 'ExclusiveLevelAlarmType_ConditionClassName'
ObjectIdNames[11133] = 'ExclusiveLevelAlarmType_InputNode'
ObjectIdNames[11134] = 'ExclusiveRateOfChangeAlarmType_ConditionClassId'
ObjectIdNames[11135] = 'ExclusiveRateOfChangeAlarmType_ConditionClassName'
ObjectIdNames[11136] = 'ExclusiveRateOfChangeAlarmType_InputNode'
ObjectIdNames[11137] = 'ExclusiveDeviationAlarmType_ConditionClassId'
ObjectIdNames[11138] = 'ExclusiveDeviationAlarmType_ConditionClassName'
ObjectIdNames[11139] = 'ExclusiveDeviationAlarmType_InputNode'
ObjectIdNames[11140] = 'NonExclusiveLimitAlarmType_ConditionClassId'
ObjectIdNames[11141] = 'NonExclusiveLimitAlarmType_ConditionClassName'
ObjectIdNames[11142] = 'NonExclusiveLimitAlarmType_InputNode'
ObjectIdNames[11143] = 'NonExclusiveLevelAlarmType_ConditionClassId'
ObjectIdNames[11144] = 'NonExclusiveLevelAlarmType_ConditionClassName'
ObjectIdNames[11145] = 'NonExclusiveLevelAlarmType_InputNode'
ObjectIdNames[11146] = 'NonExclusiveRateOfChangeAlarmType_ConditionClassId'
ObjectIdNames[11147] = 'NonExclusiveRateOfChangeAlarmType_ConditionClassName'
ObjectIdNames[11148] = 'NonExclusiveRateOfChangeAlarmType_InputNode'
ObjectIdNames[11149] = 'NonExclusiveDeviationAlarmType_ConditionClassId'
ObjectIdNames[11150] = 'NonExclusiveDeviationAlarmType_ConditionClassName'
ObjectIdNames[11151] = 'NonExclusiveDeviationAlarmType_InputNode'
ObjectIdNames[11152] = 'DiscreteAlarmType_ConditionClassId'
ObjectIdNames[11153] = 'DiscreteAlarmType_ConditionClassName'
ObjectIdNames[11154] = 'DiscreteAlarmType_InputNode'
ObjectIdNames[11155] = 'OffNormalAlarmType_ConditionClassId'
ObjectIdNames[11156] = 'OffNormalAlarmType_ConditionClassName'
ObjectIdNames[11157] = 'OffNormalAlarmType_InputNode'
ObjectIdNames[11158] = 'OffNormalAlarmType_NormalState'
ObjectIdNames[11159] = 'TripAlarmType_ConditionClassId'
ObjectIdNames[11160] = 'TripAlarmType_ConditionClassName'
ObjectIdNames[11161] = 'TripAlarmType_InputNode'
ObjectIdNames[11162] = 'TripAlarmType_NormalState'
ObjectIdNames[11163] = 'BaseConditionClassType'
ObjectIdNames[11164] = 'ProcessConditionClassType'
ObjectIdNames[11165] = 'MaintenanceConditionClassType'
ObjectIdNames[11166] = 'SystemConditionClassType'
ObjectIdNames[11168] = 'HistoricalDataConfigurationType_AggregateConfiguration_TreatUncertainAsBad'
ObjectIdNames[11169] = 'HistoricalDataConfigurationType_AggregateConfiguration_PercentDataBad'
ObjectIdNames[11170] = 'HistoricalDataConfigurationType_AggregateConfiguration_PercentDataGood'
ObjectIdNames[11171] = 'HistoricalDataConfigurationType_AggregateConfiguration_UseSlopedExtrapolation'
ObjectIdNames[11172] = 'HistoryServerCapabilitiesType_AggregateFunctions'
ObjectIdNames[11187] = 'AggregateConfigurationType'
ObjectIdNames[11188] = 'AggregateConfigurationType_TreatUncertainAsBad'
ObjectIdNames[11189] = 'AggregateConfigurationType_PercentDataBad'
ObjectIdNames[11190] = 'AggregateConfigurationType_PercentDataGood'
ObjectIdNames[11191] = 'AggregateConfigurationType_UseSlopedExtrapolation'
ObjectIdNames[11192] = 'HistoryServerCapabilities'
ObjectIdNames[11193] = 'HistoryServerCapabilities_AccessHistoryDataCapability'
ObjectIdNames[11196] = 'HistoryServerCapabilities_InsertDataCapability'
ObjectIdNames[11197] = 'HistoryServerCapabilities_ReplaceDataCapability'
ObjectIdNames[11198] = 'HistoryServerCapabilities_UpdateDataCapability'
ObjectIdNames[11199] = 'HistoryServerCapabilities_DeleteRawCapability'
ObjectIdNames[11200] = 'HistoryServerCapabilities_DeleteAtTimeCapability'
ObjectIdNames[11201] = 'HistoryServerCapabilities_AggregateFunctions'
ObjectIdNames[11202] = 'HAConfiguration'
ObjectIdNames[11203] = 'HAConfiguration_AggregateConfiguration'
ObjectIdNames[11204] = 'HAConfiguration_AggregateConfiguration_TreatUncertainAsBad'
ObjectIdNames[11205] = 'HAConfiguration_AggregateConfiguration_PercentDataBad'
ObjectIdNames[11206] = 'HAConfiguration_AggregateConfiguration_PercentDataGood'
ObjectIdNames[11207] = 'HAConfiguration_AggregateConfiguration_UseSlopedExtrapolation'
ObjectIdNames[11208] = 'HAConfiguration_Stepped'
ObjectIdNames[11209] = 'HAConfiguration_Definition'
ObjectIdNames[11210] = 'HAConfiguration_MaxTimeInterval'
ObjectIdNames[11211] = 'HAConfiguration_MinTimeInterval'
ObjectIdNames[11212] = 'HAConfiguration_ExceptionDeviation'
ObjectIdNames[11213] = 'HAConfiguration_ExceptionDeviationFormat'
ObjectIdNames[11214] = 'Annotations'
ObjectIdNames[11215] = 'HistoricalEventFilter'
ObjectIdNames[11216] = 'ModificationInfo'
ObjectIdNames[11217] = 'HistoryModifiedData'
ObjectIdNames[11218] = 'ModificationInfo_Encoding_DefaultXml'
ObjectIdNames[11219] = 'HistoryModifiedData_Encoding_DefaultXml'
ObjectIdNames[11226] = 'ModificationInfo_Encoding_DefaultBinary'
ObjectIdNames[11227] = 'HistoryModifiedData_Encoding_DefaultBinary'
ObjectIdNames[11234] = 'HistoryUpdateType'
ObjectIdNames[11238] = 'MultiStateValueDiscreteType'
ObjectIdNames[11239] = 'MultiStateValueDiscreteType_Definition'
ObjectIdNames[11240] = 'MultiStateValueDiscreteType_ValuePrecision'
ObjectIdNames[11241] = 'MultiStateValueDiscreteType_EnumValues'
ObjectIdNames[11242] = 'HistoryServerCapabilities_AccessHistoryEventsCapability'
ObjectIdNames[11268] = 'HistoryServerCapabilitiesType_MaxReturnDataValues'
ObjectIdNames[11269] = 'HistoryServerCapabilitiesType_MaxReturnEventValues'
ObjectIdNames[11270] = 'HistoryServerCapabilitiesType_InsertAnnotationCapability'
ObjectIdNames[11273] = 'HistoryServerCapabilities_MaxReturnDataValues'
ObjectIdNames[11274] = 'HistoryServerCapabilities_MaxReturnEventValues'
ObjectIdNames[11275] = 'HistoryServerCapabilities_InsertAnnotationCapability'
ObjectIdNames[11278] = 'HistoryServerCapabilitiesType_InsertEventCapability'
ObjectIdNames[11279] = 'HistoryServerCapabilitiesType_ReplaceEventCapability'
ObjectIdNames[11280] = 'HistoryServerCapabilitiesType_UpdateEventCapability'
ObjectIdNames[11281] = 'HistoryServerCapabilities_InsertEventCapability'
ObjectIdNames[11282] = 'HistoryServerCapabilities_ReplaceEventCapability'
ObjectIdNames[11283] = 'HistoryServerCapabilities_UpdateEventCapability'
ObjectIdNames[11285] = 'AggregateFunction_TimeAverage2'
ObjectIdNames[11286] = 'AggregateFunction_Minimum2'
ObjectIdNames[11287] = 'AggregateFunction_Maximum2'
ObjectIdNames[11288] = 'AggregateFunction_Range2'
ObjectIdNames[11292] = 'AggregateFunction_WorstQuality2'
ObjectIdNames[11293] = 'PerformUpdateType'
ObjectIdNames[11295] = 'UpdateStructureDataDetails'
ObjectIdNames[11296] = 'UpdateStructureDataDetails_Encoding_DefaultXml'
ObjectIdNames[11300] = 'UpdateStructureDataDetails_Encoding_DefaultBinary'
ObjectIdNames[11304] = 'AggregateFunction_Total2'
ObjectIdNames[11305] = 'AggregateFunction_MinimumActualTime2'
ObjectIdNames[11306] = 'AggregateFunction_MaximumActualTime2'
ObjectIdNames[11307] = 'AggregateFunction_DurationInStateZero'
ObjectIdNames[11308] = 'AggregateFunction_DurationInStateNonZero'
ObjectIdNames[11312] = 'Server_ServerRedundancy_CurrentServerId'
ObjectIdNames[11313] = 'Server_ServerRedundancy_RedundantServerArray'
ObjectIdNames[11314] = 'Server_ServerRedundancy_ServerUriArray'
ObjectIdNames[11322] = 'ShelvedStateMachineType_UnshelvedToTimedShelved_TransitionNumber'
ObjectIdNames[11323] = 'ShelvedStateMachineType_UnshelvedToOneShotShelved_TransitionNumber'
ObjectIdNames[11324] = 'ShelvedStateMachineType_TimedShelvedToUnshelved_TransitionNumber'
ObjectIdNames[11325] = 'ShelvedStateMachineType_TimedShelvedToOneShotShelved_TransitionNumber'
ObjectIdNames[11326] = 'ShelvedStateMachineType_OneShotShelvedToUnshelved_TransitionNumber'
ObjectIdNames[11327] = 'ShelvedStateMachineType_OneShotShelvedToTimedShelved_TransitionNumber'
ObjectIdNames[11340] = 'ExclusiveLimitStateMachineType_LowLowToLow_TransitionNumber'
ObjectIdNames[11341] = 'ExclusiveLimitStateMachineType_LowToLowLow_TransitionNumber'
ObjectIdNames[11342] = 'ExclusiveLimitStateMachineType_HighHighToHigh_TransitionNumber'
ObjectIdNames[11343] = 'ExclusiveLimitStateMachineType_HighToHighHigh_TransitionNumber'
ObjectIdNames[11426] = 'AggregateFunction_StandardDeviationSample'
ObjectIdNames[11427] = 'AggregateFunction_StandardDeviationPopulation'
ObjectIdNames[11428] = 'AggregateFunction_VarianceSample'
ObjectIdNames[11429] = 'AggregateFunction_VariancePopulation'
ObjectIdNames[11432] = 'EnumStrings'
ObjectIdNames[11433] = 'ValueAsText'
ObjectIdNames[11436] = 'ProgressEventType'
ObjectIdNames[11437] = 'ProgressEventType_EventId'
ObjectIdNames[11438] = 'ProgressEventType_EventType'
ObjectIdNames[11439] = 'ProgressEventType_SourceNode'
ObjectIdNames[11440] = 'ProgressEventType_SourceName'
ObjectIdNames[11441] = 'ProgressEventType_Time'
ObjectIdNames[11442] = 'ProgressEventType_ReceiveTime'
ObjectIdNames[11443] = 'ProgressEventType_LocalTime'
ObjectIdNames[11444] = 'ProgressEventType_Message'
ObjectIdNames[11445] = 'ProgressEventType_Severity'
ObjectIdNames[11446] = 'SystemStatusChangeEventType'
ObjectIdNames[11447] = 'SystemStatusChangeEventType_EventId'
ObjectIdNames[11448] = 'SystemStatusChangeEventType_EventType'
ObjectIdNames[11449] = 'SystemStatusChangeEventType_SourceNode'
ObjectIdNames[11450] = 'SystemStatusChangeEventType_SourceName'
ObjectIdNames[11451] = 'SystemStatusChangeEventType_Time'
ObjectIdNames[11452] = 'SystemStatusChangeEventType_ReceiveTime'
ObjectIdNames[11453] = 'SystemStatusChangeEventType_LocalTime'
ObjectIdNames[11454] = 'SystemStatusChangeEventType_Message'
ObjectIdNames[11455] = 'SystemStatusChangeEventType_Severity'
ObjectIdNames[11456] = 'TransitionVariableType_EffectiveTransitionTime'
ObjectIdNames[11457] = 'FiniteTransitionVariableType_EffectiveTransitionTime'
ObjectIdNames[11458] = 'StateMachineType_LastTransition_EffectiveTransitionTime'
ObjectIdNames[11459] = 'FiniteStateMachineType_LastTransition_EffectiveTransitionTime'
ObjectIdNames[11460] = 'TransitionEventType_Transition_EffectiveTransitionTime'
ObjectIdNames[11461] = 'MultiStateValueDiscreteType_ValueAsText'
ObjectIdNames[11462] = 'ProgramTransitionEventType_Transition_EffectiveTransitionTime'
ObjectIdNames[11463] = 'ProgramTransitionAuditEventType_Transition_EffectiveTransitionTime'
ObjectIdNames[11464] = 'ProgramStateMachineType_LastTransition_EffectiveTransitionTime'
ObjectIdNames[11465] = 'ShelvedStateMachineType_LastTransition_EffectiveTransitionTime'
ObjectIdNames[11466] = 'AlarmConditionType_ShelvingState_LastTransition_EffectiveTransitionTime'
ObjectIdNames[11467] = 'LimitAlarmType_ShelvingState_LastTransition_EffectiveTransitionTime'
ObjectIdNames[11468] = 'ExclusiveLimitStateMachineType_LastTransition_EffectiveTransitionTime'
ObjectIdNames[11469] = 'ExclusiveLimitAlarmType_ShelvingState_LastTransition_EffectiveTransitionTime'
ObjectIdNames[11470] = 'ExclusiveLimitAlarmType_LimitState_LastTransition_EffectiveTransitionTime'
ObjectIdNames[11471] = 'ExclusiveLevelAlarmType_ShelvingState_LastTransition_EffectiveTransitionTime'
ObjectIdNames[11472] = 'ExclusiveLevelAlarmType_LimitState_LastTransition_EffectiveTransitionTime'
ObjectIdNames[11473] = 'ExclusiveRateOfChangeAlarmType_ShelvingState_LastTransition_EffectiveTransitionTime'
ObjectIdNames[11474] = 'ExclusiveRateOfChangeAlarmType_LimitState_LastTransition_EffectiveTransitionTime'
ObjectIdNames[11475] = 'ExclusiveDeviationAlarmType_ShelvingState_LastTransition_EffectiveTransitionTime'
ObjectIdNames[11476] = 'ExclusiveDeviationAlarmType_LimitState_LastTransition_EffectiveTransitionTime'
ObjectIdNames[11477] = 'NonExclusiveLimitAlarmType_ShelvingState_LastTransition_EffectiveTransitionTime'
ObjectIdNames[11478] = 'NonExclusiveLevelAlarmType_ShelvingState_LastTransition_EffectiveTransitionTime'
ObjectIdNames[11479] = 'NonExclusiveRateOfChangeAlarmType_ShelvingState_LastTransition_EffectiveTransitionTime'
ObjectIdNames[11480] = 'NonExclusiveDeviationAlarmType_ShelvingState_LastTransition_EffectiveTransitionTime'
ObjectIdNames[11481] = 'DiscreteAlarmType_ShelvingState_LastTransition_EffectiveTransitionTime'
ObjectIdNames[11482] = 'OffNormalAlarmType_ShelvingState_LastTransition_EffectiveTransitionTime'
ObjectIdNames[11483] = 'TripAlarmType_ShelvingState_LastTransition_EffectiveTransitionTime'
ObjectIdNames[11485] = 'AuditActivateSessionEventType_SecureChannelId'
ObjectIdNames[11487] = 'OptionSetType'
ObjectIdNames[11488] = 'OptionSetType_OptionSetValues'
ObjectIdNames[11489] = 'ServerType_GetMonitoredItems'
ObjectIdNames[11490] = 'ServerType_GetMonitoredItems_InputArguments'
ObjectIdNames[11491] = 'ServerType_GetMonitoredItems_OutputArguments'
ObjectIdNames[11492] = 'Server_GetMonitoredItems'
ObjectIdNames[11493] = 'Server_GetMonitoredItems_InputArguments'
ObjectIdNames[11494] = 'Server_GetMonitoredItems_OutputArguments'
ObjectIdNames[11495] = 'GetMonitoredItemsMethodType'
ObjectIdNames[11496] = 'GetMonitoredItemsMethodType_InputArguments'
ObjectIdNames[11497] = 'GetMonitoredItemsMethodType_OutputArguments'
ObjectIdNames[11498] = 'MaxStringLength'
ObjectIdNames[11499] = 'HistoricalDataConfigurationType_StartOfArchive'
ObjectIdNames[11500] = 'HistoricalDataConfigurationType_StartOfOnlineArchive'
ObjectIdNames[11501] = 'HistoryServerCapabilitiesType_DeleteEventCapability'
ObjectIdNames[11502] = 'HistoryServerCapabilities_DeleteEventCapability'
ObjectIdNames[11503] = 'HAConfiguration_StartOfArchive'
ObjectIdNames[11504] = 'HAConfiguration_StartOfOnlineArchive'
ObjectIdNames[11505] = 'AggregateFunction_StartBound'
ObjectIdNames[11506] = 'AggregateFunction_EndBound'
ObjectIdNames[11507] = 'AggregateFunction_DeltaBounds'
ObjectIdNames[11508] = 'ModellingRule_OptionalPlaceholder'
ObjectIdNames[11509] = 'ModellingRule_OptionalPlaceholder_NamingRule'
ObjectIdNames[11510] = 'ModellingRule_MandatoryPlaceholder'
ObjectIdNames[11511] = 'ModellingRule_MandatoryPlaceholder_NamingRule'
ObjectIdNames[11512] = 'MaxArrayLength'
ObjectIdNames[11513] = 'EngineeringUnits'
ObjectIdNames[11514] = 'ServerType_ServerCapabilities_MaxArrayLength'
ObjectIdNames[11515] = 'ServerType_ServerCapabilities_MaxStringLength'
ObjectIdNames[11516] = 'ServerType_ServerCapabilities_OperationLimits'
ObjectIdNames[11517] = 'ServerType_ServerCapabilities_OperationLimits_MaxNodesPerRead'
ObjectIdNames[11519] = 'ServerType_ServerCapabilities_OperationLimits_MaxNodesPerWrite'
ObjectIdNames[11521] = 'ServerType_ServerCapabilities_OperationLimits_MaxNodesPerMethodCall'
ObjectIdNames[11522] = 'ServerType_ServerCapabilities_OperationLimits_MaxNodesPerBrowse'
ObjectIdNames[11523] = 'ServerType_ServerCapabilities_OperationLimits_MaxNodesPerRegisterNodes'
ObjectIdNames[11524] = 'ServerType_ServerCapabilities_OperationLimits_MaxNodesPerTranslateBrowsePathsToNodeIds'
ObjectIdNames[11525] = 'ServerType_ServerCapabilities_OperationLimits_MaxNodesPerNodeManagement'
ObjectIdNames[11526] = 'ServerType_ServerCapabilities_OperationLimits_MaxMonitoredItemsPerCall'
ObjectIdNames[11527] = 'ServerType_Namespaces'
ObjectIdNames[11528] = 'ServerType_Namespaces_AddressSpaceFile'
ObjectIdNames[11529] = 'ServerType_Namespaces_AddressSpaceFile_Size'
ObjectIdNames[11532] = 'ServerType_Namespaces_AddressSpaceFile_OpenCount'
ObjectIdNames[11533] = 'ServerType_Namespaces_AddressSpaceFile_Open'
ObjectIdNames[11534] = 'ServerType_Namespaces_AddressSpaceFile_Open_InputArguments'
ObjectIdNames[11535] = 'ServerType_Namespaces_AddressSpaceFile_Open_OutputArguments'
ObjectIdNames[11536] = 'ServerType_Namespaces_AddressSpaceFile_Close'
ObjectIdNames[11537] = 'ServerType_Namespaces_AddressSpaceFile_Close_InputArguments'
ObjectIdNames[11538] = 'ServerType_Namespaces_AddressSpaceFile_Read'
ObjectIdNames[11539] = 'ServerType_Namespaces_AddressSpaceFile_Read_InputArguments'
ObjectIdNames[11540] = 'ServerType_Namespaces_AddressSpaceFile_Read_OutputArguments'
ObjectIdNames[11541] = 'ServerType_Namespaces_AddressSpaceFile_Write'
ObjectIdNames[11542] = 'ServerType_Namespaces_AddressSpaceFile_Write_InputArguments'
ObjectIdNames[11543] = 'ServerType_Namespaces_AddressSpaceFile_GetPosition'
ObjectIdNames[11544] = 'ServerType_Namespaces_AddressSpaceFile_GetPosition_InputArguments'
ObjectIdNames[11545] = 'ServerType_Namespaces_AddressSpaceFile_GetPosition_OutputArguments'
ObjectIdNames[11546] = 'ServerType_Namespaces_AddressSpaceFile_SetPosition'
ObjectIdNames[11547] = 'ServerType_Namespaces_AddressSpaceFile_SetPosition_InputArguments'
ObjectIdNames[11548] = 'ServerType_Namespaces_AddressSpaceFile_ExportNamespace'
ObjectIdNames[11549] = 'ServerCapabilitiesType_MaxArrayLength'
ObjectIdNames[11550] = 'ServerCapabilitiesType_MaxStringLength'
ObjectIdNames[11551] = 'ServerCapabilitiesType_OperationLimits'
ObjectIdNames[11552] = 'ServerCapabilitiesType_OperationLimits_MaxNodesPerRead'
ObjectIdNames[11554] = 'ServerCapabilitiesType_OperationLimits_MaxNodesPerWrite'
ObjectIdNames[11556] = 'ServerCapabilitiesType_OperationLimits_MaxNodesPerMethodCall'
ObjectIdNames[11557] = 'ServerCapabilitiesType_OperationLimits_MaxNodesPerBrowse'
ObjectIdNames[11558] = 'ServerCapabilitiesType_OperationLimits_MaxNodesPerRegisterNodes'
ObjectIdNames[11559] = 'ServerCapabilitiesType_OperationLimits_MaxNodesPerTranslateBrowsePathsToNodeIds'
ObjectIdNames[11560] = 'ServerCapabilitiesType_OperationLimits_MaxNodesPerNodeManagement'
ObjectIdNames[11561] = 'ServerCapabilitiesType_OperationLimits_MaxMonitoredItemsPerCall'
ObjectIdNames[11562] = 'ServerCapabilitiesType_VendorCapability'
ObjectIdNames[11564] = 'OperationLimitsType'
ObjectIdNames[11565] = 'OperationLimitsType_MaxNodesPerRead'
ObjectIdNames[11567] = 'OperationLimitsType_MaxNodesPerWrite'
ObjectIdNames[11569] = 'OperationLimitsType_MaxNodesPerMethodCall'
ObjectIdNames[11570] = 'OperationLimitsType_MaxNodesPerBrowse'
ObjectIdNames[11571] = 'OperationLimitsType_MaxNodesPerRegisterNodes'
ObjectIdNames[11572] = 'OperationLimitsType_MaxNodesPerTranslateBrowsePathsToNodeIds'
ObjectIdNames[11573] = 'OperationLimitsType_MaxNodesPerNodeManagement'
ObjectIdNames[11574] = 'OperationLimitsType_MaxMonitoredItemsPerCall'
ObjectIdNames[11575] = 'FileType'
ObjectIdNames[11576] = 'FileType_Size'
ObjectIdNames[11579] = 'FileType_OpenCount'
ObjectIdNames[11580] = 'FileType_Open'
ObjectIdNames[11581] = 'FileType_Open_InputArguments'
ObjectIdNames[11582] = 'FileType_Open_OutputArguments'
ObjectIdNames[11583] = 'FileType_Close'
ObjectIdNames[11584] = 'FileType_Close_InputArguments'
ObjectIdNames[11585] = 'FileType_Read'
ObjectIdNames[11586] = 'FileType_Read_InputArguments'
ObjectIdNames[11587] = 'FileType_Read_OutputArguments'
ObjectIdNames[11588] = 'FileType_Write'
ObjectIdNames[11589] = 'FileType_Write_InputArguments'
ObjectIdNames[11590] = 'FileType_GetPosition'
ObjectIdNames[11591] = 'FileType_GetPosition_InputArguments'
ObjectIdNames[11592] = 'FileType_GetPosition_OutputArguments'
ObjectIdNames[11593] = 'FileType_SetPosition'
ObjectIdNames[11594] = 'FileType_SetPosition_InputArguments'
ObjectIdNames[11595] = 'AddressSpaceFileType'
ObjectIdNames[11596] = 'AddressSpaceFileType_Size'
ObjectIdNames[11599] = 'AddressSpaceFileType_OpenCount'
ObjectIdNames[11600] = 'AddressSpaceFileType_Open'
ObjectIdNames[11601] = 'AddressSpaceFileType_Open_InputArguments'
ObjectIdNames[11602] = 'AddressSpaceFileType_Open_OutputArguments'
ObjectIdNames[11603] = 'AddressSpaceFileType_Close'
ObjectIdNames[11604] = 'AddressSpaceFileType_Close_InputArguments'
ObjectIdNames[11605] = 'AddressSpaceFileType_Read'
ObjectIdNames[11606] = 'AddressSpaceFileType_Read_InputArguments'
ObjectIdNames[11607] = 'AddressSpaceFileType_Read_OutputArguments'
ObjectIdNames[11608] = 'AddressSpaceFileType_Write'
ObjectIdNames[11609] = 'AddressSpaceFileType_Write_InputArguments'
ObjectIdNames[11610] = 'AddressSpaceFileType_GetPosition'
ObjectIdNames[11611] = 'AddressSpaceFileType_GetPosition_InputArguments'
ObjectIdNames[11612] = 'AddressSpaceFileType_GetPosition_OutputArguments'
ObjectIdNames[11613] = 'AddressSpaceFileType_SetPosition'
ObjectIdNames[11614] = 'AddressSpaceFileType_SetPosition_InputArguments'
ObjectIdNames[11615] = 'AddressSpaceFileType_ExportNamespace'
ObjectIdNames[11616] = 'NamespaceMetadataType'
ObjectIdNames[11617] = 'NamespaceMetadataType_NamespaceUri'
ObjectIdNames[11618] = 'NamespaceMetadataType_NamespaceVersion'
ObjectIdNames[11619] = 'NamespaceMetadataType_NamespacePublicationDate'
ObjectIdNames[11620] = 'NamespaceMetadataType_IsNamespaceSubset'
ObjectIdNames[11621] = 'NamespaceMetadataType_StaticNodeIdIdentifierTypes'
ObjectIdNames[11622] = 'NamespaceMetadataType_StaticNumericNodeIdRange'
ObjectIdNames[11623] = 'NamespaceMetadataType_StaticStringNodeIdPattern'
ObjectIdNames[11624] = 'NamespaceMetadataType_NamespaceFile'
ObjectIdNames[11625] = 'NamespaceMetadataType_NamespaceFile_Size'
ObjectIdNames[11628] = 'NamespaceMetadataType_NamespaceFile_OpenCount'
ObjectIdNames[11629] = 'NamespaceMetadataType_NamespaceFile_Open'
ObjectIdNames[11630] = 'NamespaceMetadataType_NamespaceFile_Open_InputArguments'
ObjectIdNames[11631] = 'NamespaceMetadataType_NamespaceFile_Open_OutputArguments'
ObjectIdNames[11632] = 'NamespaceMetadataType_NamespaceFile_Close'
ObjectIdNames[11633] = 'NamespaceMetadataType_NamespaceFile_Close_InputArguments'
ObjectIdNames[11634] = 'NamespaceMetadataType_NamespaceFile_Read'
ObjectIdNames[11635] = 'NamespaceMetadataType_NamespaceFile_Read_InputArguments'
ObjectIdNames[11636] = 'NamespaceMetadataType_NamespaceFile_Read_OutputArguments'
ObjectIdNames[11637] = 'NamespaceMetadataType_NamespaceFile_Write'
ObjectIdNames[11638] = 'NamespaceMetadataType_NamespaceFile_Write_InputArguments'
ObjectIdNames[11639] = 'NamespaceMetadataType_NamespaceFile_GetPosition'
ObjectIdNames[11640] = 'NamespaceMetadataType_NamespaceFile_GetPosition_InputArguments'
ObjectIdNames[11641] = 'NamespaceMetadataType_NamespaceFile_GetPosition_OutputArguments'
ObjectIdNames[11642] = 'NamespaceMetadataType_NamespaceFile_SetPosition'
ObjectIdNames[11643] = 'NamespaceMetadataType_NamespaceFile_SetPosition_InputArguments'
ObjectIdNames[11644] = 'NamespaceMetadataType_NamespaceFile_ExportNamespace'
ObjectIdNames[11645] = 'NamespacesType'
ObjectIdNames[11646] = 'NamespacesType_NamespaceIdentifier'
ObjectIdNames[11647] = 'NamespacesType_NamespaceIdentifier_NamespaceUri'
ObjectIdNames[11648] = 'NamespacesType_NamespaceIdentifier_NamespaceVersion'
ObjectIdNames[11649] = 'NamespacesType_NamespaceIdentifier_NamespacePublicationDate'
ObjectIdNames[11650] = 'NamespacesType_NamespaceIdentifier_IsNamespaceSubset'
ObjectIdNames[11651] = 'NamespacesType_NamespaceIdentifier_StaticNodeIdIdentifierTypes'
ObjectIdNames[11652] = 'NamespacesType_NamespaceIdentifier_StaticNumericNodeIdRange'
ObjectIdNames[11653] = 'NamespacesType_NamespaceIdentifier_StaticStringNodeIdPattern'
ObjectIdNames[11654] = 'NamespacesType_NamespaceIdentifier_NamespaceFile'
ObjectIdNames[11655] = 'NamespacesType_NamespaceIdentifier_NamespaceFile_Size'
ObjectIdNames[11658] = 'NamespacesType_NamespaceIdentifier_NamespaceFile_OpenCount'
ObjectIdNames[11659] = 'NamespacesType_NamespaceIdentifier_NamespaceFile_Open'
ObjectIdNames[11660] = 'NamespacesType_NamespaceIdentifier_NamespaceFile_Open_InputArguments'
ObjectIdNames[11661] = 'NamespacesType_NamespaceIdentifier_NamespaceFile_Open_OutputArguments'
ObjectIdNames[11662] = 'NamespacesType_NamespaceIdentifier_NamespaceFile_Close'
ObjectIdNames[11663] = 'NamespacesType_NamespaceIdentifier_NamespaceFile_Close_InputArguments'
ObjectIdNames[11664] = 'NamespacesType_NamespaceIdentifier_NamespaceFile_Read'
ObjectIdNames[11665] = 'NamespacesType_NamespaceIdentifier_NamespaceFile_Read_InputArguments'
ObjectIdNames[11666] = 'NamespacesType_NamespaceIdentifier_NamespaceFile_Read_OutputArguments'
ObjectIdNames[11667] = 'NamespacesType_NamespaceIdentifier_NamespaceFile_Write'
ObjectIdNames[11668] = 'NamespacesType_NamespaceIdentifier_NamespaceFile_Write_InputArguments'
ObjectIdNames[11669] = 'NamespacesType_NamespaceIdentifier_NamespaceFile_GetPosition'
ObjectIdNames[11670] = 'NamespacesType_NamespaceIdentifier_NamespaceFile_GetPosition_InputArguments'
ObjectIdNames[11671] = 'NamespacesType_NamespaceIdentifier_NamespaceFile_GetPosition_OutputArguments'
ObjectIdNames[11672] = 'NamespacesType_NamespaceIdentifier_NamespaceFile_SetPosition'
ObjectIdNames[11673] = 'NamespacesType_NamespaceIdentifier_NamespaceFile_SetPosition_InputArguments'
ObjectIdNames[11674] = 'NamespacesType_NamespaceIdentifier_NamespaceFile_ExportNamespace'
ObjectIdNames[11675] = 'NamespacesType_AddressSpaceFile'
ObjectIdNames[11676] = 'NamespacesType_AddressSpaceFile_Size'
ObjectIdNames[11679] = 'NamespacesType_AddressSpaceFile_OpenCount'
ObjectIdNames[11680] = 'NamespacesType_AddressSpaceFile_Open'
ObjectIdNames[11681] = 'NamespacesType_AddressSpaceFile_Open_InputArguments'
ObjectIdNames[11682] = 'NamespacesType_AddressSpaceFile_Open_OutputArguments'
ObjectIdNames[11683] = 'NamespacesType_AddressSpaceFile_Close'
ObjectIdNames[11684] = 'NamespacesType_AddressSpaceFile_Close_InputArguments'
ObjectIdNames[11685] = 'NamespacesType_AddressSpaceFile_Read'
ObjectIdNames[11686] = 'NamespacesType_AddressSpaceFile_Read_InputArguments'
ObjectIdNames[11687] = 'NamespacesType_AddressSpaceFile_Read_OutputArguments'
ObjectIdNames[11688] = 'NamespacesType_AddressSpaceFile_Write'
ObjectIdNames[11689] = 'NamespacesType_AddressSpaceFile_Write_InputArguments'
ObjectIdNames[11690] = 'NamespacesType_AddressSpaceFile_GetPosition'
ObjectIdNames[11691] = 'NamespacesType_AddressSpaceFile_GetPosition_InputArguments'
ObjectIdNames[11692] = 'NamespacesType_AddressSpaceFile_GetPosition_OutputArguments'
ObjectIdNames[11693] = 'NamespacesType_AddressSpaceFile_SetPosition'
ObjectIdNames[11694] = 'NamespacesType_AddressSpaceFile_SetPosition_InputArguments'
ObjectIdNames[11695] = 'NamespacesType_AddressSpaceFile_ExportNamespace'
ObjectIdNames[11696] = 'SystemStatusChangeEventType_SystemState'
ObjectIdNames[11697] = 'SamplingIntervalDiagnosticsType_SampledMonitoredItemsCount'
ObjectIdNames[11698] = 'SamplingIntervalDiagnosticsType_MaxSampledMonitoredItemsCount'
ObjectIdNames[11699] = 'SamplingIntervalDiagnosticsType_DisabledMonitoredItemsSamplingCount'
ObjectIdNames[11701] = 'OptionSetType_BitMask'
ObjectIdNames[11702] = 'Server_ServerCapabilities_MaxArrayLength'
ObjectIdNames[11703] = 'Server_ServerCapabilities_MaxStringLength'
ObjectIdNames[11704] = 'Server_ServerCapabilities_OperationLimits'
ObjectIdNames[11705] = 'Server_ServerCapabilities_OperationLimits_MaxNodesPerRead'
ObjectIdNames[11707] = 'Server_ServerCapabilities_OperationLimits_MaxNodesPerWrite'
ObjectIdNames[11709] = 'Server_ServerCapabilities_OperationLimits_MaxNodesPerMethodCall'
ObjectIdNames[11710] = 'Server_ServerCapabilities_OperationLimits_MaxNodesPerBrowse'
ObjectIdNames[11711] = 'Server_ServerCapabilities_OperationLimits_MaxNodesPerRegisterNodes'
ObjectIdNames[11712] = 'Server_ServerCapabilities_OperationLimits_MaxNodesPerTranslateBrowsePathsToNodeIds'
ObjectIdNames[11713] = 'Server_ServerCapabilities_OperationLimits_MaxNodesPerNodeManagement'
ObjectIdNames[11714] = 'Server_ServerCapabilities_OperationLimits_MaxMonitoredItemsPerCall'
ObjectIdNames[11715] = 'Server_Namespaces'
ObjectIdNames[11716] = 'Server_Namespaces_AddressSpaceFile'
ObjectIdNames[11717] = 'Server_Namespaces_AddressSpaceFile_Size'
ObjectIdNames[11720] = 'Server_Namespaces_AddressSpaceFile_OpenCount'
ObjectIdNames[11721] = 'Server_Namespaces_AddressSpaceFile_Open'
ObjectIdNames[11722] = 'Server_Namespaces_AddressSpaceFile_Open_InputArguments'
ObjectIdNames[11723] = 'Server_Namespaces_AddressSpaceFile_Open_OutputArguments'
ObjectIdNames[11724] = 'Server_Namespaces_AddressSpaceFile_Close'
ObjectIdNames[11725] = 'Server_Namespaces_AddressSpaceFile_Close_InputArguments'
ObjectIdNames[11726] = 'Server_Namespaces_AddressSpaceFile_Read'
ObjectIdNames[11727] = 'Server_Namespaces_AddressSpaceFile_Read_InputArguments'
ObjectIdNames[11728] = 'Server_Namespaces_AddressSpaceFile_Read_OutputArguments'
ObjectIdNames[11729] = 'Server_Namespaces_AddressSpaceFile_Write'
ObjectIdNames[11730] = 'Server_Namespaces_AddressSpaceFile_Write_InputArguments'
ObjectIdNames[11731] = 'Server_Namespaces_AddressSpaceFile_GetPosition'
ObjectIdNames[11732] = 'Server_Namespaces_AddressSpaceFile_GetPosition_InputArguments'
ObjectIdNames[11733] = 'Server_Namespaces_AddressSpaceFile_GetPosition_OutputArguments'
ObjectIdNames[11734] = 'Server_Namespaces_AddressSpaceFile_SetPosition'
ObjectIdNames[11735] = 'Server_Namespaces_AddressSpaceFile_SetPosition_InputArguments'
ObjectIdNames[11736] = 'Server_Namespaces_AddressSpaceFile_ExportNamespace'
ObjectIdNames[11737] = 'BitFieldMaskDataType'
ObjectIdNames[11738] = 'OpenMethodType'
ObjectIdNames[11739] = 'OpenMethodType_InputArguments'
ObjectIdNames[11740] = 'OpenMethodType_OutputArguments'
ObjectIdNames[11741] = 'CloseMethodType'
ObjectIdNames[11742] = 'CloseMethodType_InputArguments'
ObjectIdNames[11743] = 'ReadMethodType'
ObjectIdNames[11744] = 'ReadMethodType_InputArguments'
ObjectIdNames[11745] = 'ReadMethodType_OutputArguments'
ObjectIdNames[11746] = 'WriteMethodType'
ObjectIdNames[11747] = 'WriteMethodType_InputArguments'
ObjectIdNames[11748] = 'GetPositionMethodType'
ObjectIdNames[11749] = 'GetPositionMethodType_InputArguments'
ObjectIdNames[11750] = 'GetPositionMethodType_OutputArguments'
ObjectIdNames[11751] = 'SetPositionMethodType'
ObjectIdNames[11752] = 'SetPositionMethodType_InputArguments'
ObjectIdNames[11753] = 'SystemOffNormalAlarmType'
ObjectIdNames[11754] = 'SystemOffNormalAlarmType_EventId'
ObjectIdNames[11755] = 'SystemOffNormalAlarmType_EventType'
ObjectIdNames[11756] = 'SystemOffNormalAlarmType_SourceNode'
ObjectIdNames[11757] = 'SystemOffNormalAlarmType_SourceName'
ObjectIdNames[11758] = 'SystemOffNormalAlarmType_Time'
ObjectIdNames[11759] = 'SystemOffNormalAlarmType_ReceiveTime'
ObjectIdNames[11760] = 'SystemOffNormalAlarmType_LocalTime'
ObjectIdNames[11761] = 'SystemOffNormalAlarmType_Message'
ObjectIdNames[11762] = 'SystemOffNormalAlarmType_Severity'
ObjectIdNames[11763] = 'SystemOffNormalAlarmType_ConditionClassId'
ObjectIdNames[11764] = 'SystemOffNormalAlarmType_ConditionClassName'
ObjectIdNames[11765] = 'SystemOffNormalAlarmType_ConditionName'
ObjectIdNames[11766] = 'SystemOffNormalAlarmType_BranchId'
ObjectIdNames[11767] = 'SystemOffNormalAlarmType_Retain'
ObjectIdNames[11768] = 'SystemOffNormalAlarmType_EnabledState'
ObjectIdNames[11769] = 'SystemOffNormalAlarmType_EnabledState_Id'
ObjectIdNames[11770] = 'SystemOffNormalAlarmType_EnabledState_Name'
ObjectIdNames[11771] = 'SystemOffNormalAlarmType_EnabledState_Number'
ObjectIdNames[11772] = 'SystemOffNormalAlarmType_EnabledState_EffectiveDisplayName'
ObjectIdNames[11773] = 'SystemOffNormalAlarmType_EnabledState_TransitionTime'
ObjectIdNames[11774] = 'SystemOffNormalAlarmType_EnabledState_EffectiveTransitionTime'
ObjectIdNames[11775] = 'SystemOffNormalAlarmType_EnabledState_TrueState'
ObjectIdNames[11776] = 'SystemOffNormalAlarmType_EnabledState_FalseState'
ObjectIdNames[11777] = 'SystemOffNormalAlarmType_Quality'
ObjectIdNames[11778] = 'SystemOffNormalAlarmType_Quality_SourceTimestamp'
ObjectIdNames[11779] = 'SystemOffNormalAlarmType_LastSeverity'
ObjectIdNames[11780] = 'SystemOffNormalAlarmType_LastSeverity_SourceTimestamp'
ObjectIdNames[11781] = 'SystemOffNormalAlarmType_Comment'
ObjectIdNames[11782] = 'SystemOffNormalAlarmType_Comment_SourceTimestamp'
ObjectIdNames[11783] = 'SystemOffNormalAlarmType_ClientUserId'
ObjectIdNames[11784] = 'SystemOffNormalAlarmType_Disable'
ObjectIdNames[11785] = 'SystemOffNormalAlarmType_Enable'
ObjectIdNames[11786] = 'SystemOffNormalAlarmType_AddComment'
ObjectIdNames[11787] = 'SystemOffNormalAlarmType_AddComment_InputArguments'
ObjectIdNames[11788] = 'SystemOffNormalAlarmType_ConditionRefresh'
ObjectIdNames[11789] = 'SystemOffNormalAlarmType_ConditionRefresh_InputArguments'
ObjectIdNames[11790] = 'SystemOffNormalAlarmType_AckedState'
ObjectIdNames[11791] = 'SystemOffNormalAlarmType_AckedState_Id'
ObjectIdNames[11792] = 'SystemOffNormalAlarmType_AckedState_Name'
ObjectIdNames[11793] = 'SystemOffNormalAlarmType_AckedState_Number'
ObjectIdNames[11794] = 'SystemOffNormalAlarmType_AckedState_EffectiveDisplayName'
ObjectIdNames[11795] = 'SystemOffNormalAlarmType_AckedState_TransitionTime'
ObjectIdNames[11796] = 'SystemOffNormalAlarmType_AckedState_EffectiveTransitionTime'
ObjectIdNames[11797] = 'SystemOffNormalAlarmType_AckedState_TrueState'
ObjectIdNames[11798] = 'SystemOffNormalAlarmType_AckedState_FalseState'
ObjectIdNames[11799] = 'SystemOffNormalAlarmType_ConfirmedState'
ObjectIdNames[11800] = 'SystemOffNormalAlarmType_ConfirmedState_Id'
ObjectIdNames[11801] = 'SystemOffNormalAlarmType_ConfirmedState_Name'
ObjectIdNames[11802] = 'SystemOffNormalAlarmType_ConfirmedState_Number'
ObjectIdNames[11803] = 'SystemOffNormalAlarmType_ConfirmedState_EffectiveDisplayName'
ObjectIdNames[11804] = 'SystemOffNormalAlarmType_ConfirmedState_TransitionTime'
ObjectIdNames[11805] = 'SystemOffNormalAlarmType_ConfirmedState_EffectiveTransitionTime'
ObjectIdNames[11806] = 'SystemOffNormalAlarmType_ConfirmedState_TrueState'
ObjectIdNames[11807] = 'SystemOffNormalAlarmType_ConfirmedState_FalseState'
ObjectIdNames[11808] = 'SystemOffNormalAlarmType_Acknowledge'
ObjectIdNames[11809] = 'SystemOffNormalAlarmType_Acknowledge_InputArguments'
ObjectIdNames[11810] = 'SystemOffNormalAlarmType_Confirm'
ObjectIdNames[11811] = 'SystemOffNormalAlarmType_Confirm_InputArguments'
ObjectIdNames[11812] = 'SystemOffNormalAlarmType_ActiveState'
ObjectIdNames[11813] = 'SystemOffNormalAlarmType_ActiveState_Id'
ObjectIdNames[11814] = 'SystemOffNormalAlarmType_ActiveState_Name'
ObjectIdNames[11815] = 'SystemOffNormalAlarmType_ActiveState_Number'
ObjectIdNames[11816] = 'SystemOffNormalAlarmType_ActiveState_EffectiveDisplayName'
ObjectIdNames[11817] = 'SystemOffNormalAlarmType_ActiveState_TransitionTime'
ObjectIdNames[11818] = 'SystemOffNormalAlarmType_ActiveState_EffectiveTransitionTime'
ObjectIdNames[11819] = 'SystemOffNormalAlarmType_ActiveState_TrueState'
ObjectIdNames[11820] = 'SystemOffNormalAlarmType_ActiveState_FalseState'
ObjectIdNames[11821] = 'SystemOffNormalAlarmType_InputNode'
ObjectIdNames[11822] = 'SystemOffNormalAlarmType_SuppressedState'
ObjectIdNames[11823] = 'SystemOffNormalAlarmType_SuppressedState_Id'
ObjectIdNames[11824] = 'SystemOffNormalAlarmType_SuppressedState_Name'
ObjectIdNames[11825] = 'SystemOffNormalAlarmType_SuppressedState_Number'
ObjectIdNames[11826] = 'SystemOffNormalAlarmType_SuppressedState_EffectiveDisplayName'
ObjectIdNames[11827] = 'SystemOffNormalAlarmType_SuppressedState_TransitionTime'
ObjectIdNames[11828] = 'SystemOffNormalAlarmType_SuppressedState_EffectiveTransitionTime'
ObjectIdNames[11829] = 'SystemOffNormalAlarmType_SuppressedState_TrueState'
ObjectIdNames[11830] = 'SystemOffNormalAlarmType_SuppressedState_FalseState'
ObjectIdNames[11831] = 'SystemOffNormalAlarmType_ShelvingState'
ObjectIdNames[11832] = 'SystemOffNormalAlarmType_ShelvingState_CurrentState'
ObjectIdNames[11833] = 'SystemOffNormalAlarmType_ShelvingState_CurrentState_Id'
ObjectIdNames[11834] = 'SystemOffNormalAlarmType_ShelvingState_CurrentState_Name'
ObjectIdNames[11835] = 'SystemOffNormalAlarmType_ShelvingState_CurrentState_Number'
ObjectIdNames[11836] = 'SystemOffNormalAlarmType_ShelvingState_CurrentState_EffectiveDisplayName'
ObjectIdNames[11837] = 'SystemOffNormalAlarmType_ShelvingState_LastTransition'
ObjectIdNames[11838] = 'SystemOffNormalAlarmType_ShelvingState_LastTransition_Id'
ObjectIdNames[11839] = 'SystemOffNormalAlarmType_ShelvingState_LastTransition_Name'
ObjectIdNames[11840] = 'SystemOffNormalAlarmType_ShelvingState_LastTransition_Number'
ObjectIdNames[11841] = 'SystemOffNormalAlarmType_ShelvingState_LastTransition_TransitionTime'
ObjectIdNames[11842] = 'SystemOffNormalAlarmType_ShelvingState_LastTransition_EffectiveTransitionTime'
ObjectIdNames[11843] = 'SystemOffNormalAlarmType_ShelvingState_UnshelveTime'
ObjectIdNames[11844] = 'SystemOffNormalAlarmType_ShelvingState_Unshelve'
ObjectIdNames[11845] = 'SystemOffNormalAlarmType_ShelvingState_OneShotShelve'
ObjectIdNames[11846] = 'SystemOffNormalAlarmType_ShelvingState_TimedShelve'
ObjectIdNames[11847] = 'SystemOffNormalAlarmType_ShelvingState_TimedShelve_InputArguments'
ObjectIdNames[11848] = 'SystemOffNormalAlarmType_SuppressedOrShelved'
ObjectIdNames[11849] = 'SystemOffNormalAlarmType_MaxTimeShelved'
ObjectIdNames[11850] = 'SystemOffNormalAlarmType_NormalState'
ObjectIdNames[11851] = 'AuditConditionCommentEventType_Comment'
ObjectIdNames[11852] = 'AuditConditionRespondEventType_SelectedResponse'
ObjectIdNames[11853] = 'AuditConditionAcknowledgeEventType_Comment'
ObjectIdNames[11854] = 'AuditConditionConfirmEventType_Comment'
ObjectIdNames[11855] = 'AuditConditionShelvingEventType_ShelvingTime'
ObjectIdNames[11856] = 'AuditProgramTransitionEventType'
ObjectIdNames[11857] = 'AuditProgramTransitionEventType_EventId'
ObjectIdNames[11858] = 'AuditProgramTransitionEventType_EventType'
ObjectIdNames[11859] = 'AuditProgramTransitionEventType_SourceNode'
ObjectIdNames[11860] = 'AuditProgramTransitionEventType_SourceName'
ObjectIdNames[11861] = 'AuditProgramTransitionEventType_Time'
ObjectIdNames[11862] = 'AuditProgramTransitionEventType_ReceiveTime'
ObjectIdNames[11863] = 'AuditProgramTransitionEventType_LocalTime'
ObjectIdNames[11864] = 'AuditProgramTransitionEventType_Message'
ObjectIdNames[11865] = 'AuditProgramTransitionEventType_Severity'
ObjectIdNames[11866] = 'AuditProgramTransitionEventType_ActionTimeStamp'
ObjectIdNames[11867] = 'AuditProgramTransitionEventType_Status'
ObjectIdNames[11868] = 'AuditProgramTransitionEventType_ServerId'
ObjectIdNames[11869] = 'AuditProgramTransitionEventType_ClientAuditEntryId'
ObjectIdNames[11870] = 'AuditProgramTransitionEventType_ClientUserId'
ObjectIdNames[11871] = 'AuditProgramTransitionEventType_MethodId'
ObjectIdNames[11872] = 'AuditProgramTransitionEventType_InputArguments'
ObjectIdNames[11873] = 'AuditProgramTransitionEventType_OldStateId'
ObjectIdNames[11874] = 'AuditProgramTransitionEventType_NewStateId'
ObjectIdNames[11875] = 'AuditProgramTransitionEventType_TransitionNumber'
ObjectIdNames[11876] = 'HistoricalDataConfigurationType_AggregateFunctions'
ObjectIdNames[11877] = 'HAConfiguration_AggregateFunctions'
ObjectIdNames[11878] = 'NodeClass_EnumValues'
ObjectIdNames[11879] = 'InstanceNode'
ObjectIdNames[11880] = 'TypeNode'
ObjectIdNames[11881] = 'NodeAttributesMask_EnumValues'
ObjectIdNames[11882] = 'AttributeWriteMask_EnumValues'
ObjectIdNames[11883] = 'BrowseResultMask_EnumValues'
ObjectIdNames[11884] = 'HistoryUpdateType_EnumValues'
ObjectIdNames[11885] = 'PerformUpdateType_EnumValues'
ObjectIdNames[11886] = 'EnumeratedTestType_EnumValues'
ObjectIdNames[11887] = 'InstanceNode_Encoding_DefaultXml'
ObjectIdNames[11888] = 'TypeNode_Encoding_DefaultXml'
ObjectIdNames[11889] = 'InstanceNode_Encoding_DefaultBinary'
ObjectIdNames[11890] = 'TypeNode_Encoding_DefaultBinary'
ObjectIdNames[11891] = 'SessionDiagnosticsObjectType_SessionDiagnostics_UnauthorizedRequestCount'
ObjectIdNames[11892] = 'SessionDiagnosticsVariableType_UnauthorizedRequestCount'
ObjectIdNames[11939] = 'OpenFileMode'
ObjectIdNames[11940] = 'OpenFileMode_EnumValues'
ObjectIdNames[11941] = 'ModelChangeStructureVerbMask'
ObjectIdNames[11942] = 'ModelChangeStructureVerbMask_EnumValues'
ObjectIdNames[11943] = 'EndpointUrlListDataType'
ObjectIdNames[11944] = 'NetworkGroupDataType'
ObjectIdNames[11945] = 'NonTransparentNetworkRedundancyType'
ObjectIdNames[11946] = 'NonTransparentNetworkRedundancyType_RedundancySupport'
ObjectIdNames[11947] = 'NonTransparentNetworkRedundancyType_ServerUriArray'
ObjectIdNames[11948] = 'NonTransparentNetworkRedundancyType_ServerNetworkGroups'
ObjectIdNames[11949] = 'EndpointUrlListDataType_Encoding_DefaultXml'
ObjectIdNames[11950] = 'NetworkGroupDataType_Encoding_DefaultXml'
ObjectIdNames[11951] = 'OpcUa_XmlSchema_EndpointUrlListDataType'
ObjectIdNames[11952] = 'OpcUa_XmlSchema_EndpointUrlListDataType_DataTypeVersion'
ObjectIdNames[11953] = 'OpcUa_XmlSchema_EndpointUrlListDataType_DictionaryFragment'
ObjectIdNames[11954] = 'OpcUa_XmlSchema_NetworkGroupDataType'
ObjectIdNames[11955] = 'OpcUa_XmlSchema_NetworkGroupDataType_DataTypeVersion'
ObjectIdNames[11956] = 'OpcUa_XmlSchema_NetworkGroupDataType_DictionaryFragment'
ObjectIdNames[11957] = 'EndpointUrlListDataType_Encoding_DefaultBinary'
ObjectIdNames[11958] = 'NetworkGroupDataType_Encoding_DefaultBinary'
ObjectIdNames[11959] = 'OpcUa_BinarySchema_EndpointUrlListDataType'
ObjectIdNames[11960] = 'OpcUa_BinarySchema_EndpointUrlListDataType_DataTypeVersion'
ObjectIdNames[11961] = 'OpcUa_BinarySchema_EndpointUrlListDataType_DictionaryFragment'
ObjectIdNames[11962] = 'OpcUa_BinarySchema_NetworkGroupDataType'
ObjectIdNames[11963] = 'OpcUa_BinarySchema_NetworkGroupDataType_DataTypeVersion'
ObjectIdNames[11964] = 'OpcUa_BinarySchema_NetworkGroupDataType_DictionaryFragment'
ObjectIdNames[12021] = 'ArrayItemType'
ObjectIdNames[12022] = 'ArrayItemType_Definition'
ObjectIdNames[12023] = 'ArrayItemType_ValuePrecision'
ObjectIdNames[12024] = 'ArrayItemType_InstrumentRange'
ObjectIdNames[12025] = 'ArrayItemType_EURange'
ObjectIdNames[12026] = 'ArrayItemType_EngineeringUnits'
ObjectIdNames[12027] = 'ArrayItemType_Title'
ObjectIdNames[12028] = 'ArrayItemType_AxisScaleType'
ObjectIdNames[12029] = 'YArrayItemType'
ObjectIdNames[12030] = 'YArrayItemType_Definition'
ObjectIdNames[12031] = 'YArrayItemType_ValuePrecision'
ObjectIdNames[12032] = 'YArrayItemType_InstrumentRange'
ObjectIdNames[12033] = 'YArrayItemType_EURange'
ObjectIdNames[12034] = 'YArrayItemType_EngineeringUnits'
ObjectIdNames[12035] = 'YArrayItemType_Title'
ObjectIdNames[12036] = 'YArrayItemType_AxisScaleType'
ObjectIdNames[12037] = 'YArrayItemType_XAxisDefinition'
ObjectIdNames[12038] = 'XYArrayItemType'
ObjectIdNames[12039] = 'XYArrayItemType_Definition'
ObjectIdNames[12040] = 'XYArrayItemType_ValuePrecision'
ObjectIdNames[12041] = 'XYArrayItemType_InstrumentRange'
ObjectIdNames[12042] = 'XYArrayItemType_EURange'
ObjectIdNames[12043] = 'XYArrayItemType_EngineeringUnits'
ObjectIdNames[12044] = 'XYArrayItemType_Title'
ObjectIdNames[12045] = 'XYArrayItemType_AxisScaleType'
ObjectIdNames[12046] = 'XYArrayItemType_XAxisDefinition'
ObjectIdNames[12047] = 'ImageItemType'
ObjectIdNames[12048] = 'ImageItemType_Definition'
ObjectIdNames[12049] = 'ImageItemType_ValuePrecision'
ObjectIdNames[12050] = 'ImageItemType_InstrumentRange'
ObjectIdNames[12051] = 'ImageItemType_EURange'
ObjectIdNames[12052] = 'ImageItemType_EngineeringUnits'
ObjectIdNames[12053] = 'ImageItemType_Title'
ObjectIdNames[12054] = 'ImageItemType_AxisScaleType'
ObjectIdNames[12055] = 'ImageItemType_XAxisDefinition'
ObjectIdNames[12056] = 'ImageItemType_YAxisDefinition'
ObjectIdNames[12057] = 'CubeItemType'
ObjectIdNames[12058] = 'CubeItemType_Definition'
ObjectIdNames[12059] = 'CubeItemType_ValuePrecision'
ObjectIdNames[12060] = 'CubeItemType_InstrumentRange'
ObjectIdNames[12061] = 'CubeItemType_EURange'
ObjectIdNames[12062] = 'CubeItemType_EngineeringUnits'
ObjectIdNames[12063] = 'CubeItemType_Title'
ObjectIdNames[12064] = 'CubeItemType_AxisScaleType'
ObjectIdNames[12065] = 'CubeItemType_XAxisDefinition'
ObjectIdNames[12066] = 'CubeItemType_YAxisDefinition'
ObjectIdNames[12067] = 'CubeItemType_ZAxisDefinition'
ObjectIdNames[12068] = 'NDimensionArrayItemType'
ObjectIdNames[12069] = 'NDimensionArrayItemType_Definition'
ObjectIdNames[12070] = 'NDimensionArrayItemType_ValuePrecision'
ObjectIdNames[12071] = 'NDimensionArrayItemType_InstrumentRange'
ObjectIdNames[12072] = 'NDimensionArrayItemType_EURange'
ObjectIdNames[12073] = 'NDimensionArrayItemType_EngineeringUnits'
ObjectIdNames[12074] = 'NDimensionArrayItemType_Title'
ObjectIdNames[12075] = 'NDimensionArrayItemType_AxisScaleType'
ObjectIdNames[12076] = 'NDimensionArrayItemType_AxisDefinition'
ObjectIdNames[12077] = 'AxisScaleEnumeration'
ObjectIdNames[12078] = 'AxisScaleEnumeration_EnumStrings'
ObjectIdNames[12079] = 'AxisInformation'
ObjectIdNames[12080] = 'XVType'
ObjectIdNames[12081] = 'AxisInformation_Encoding_DefaultXml'
ObjectIdNames[12082] = 'XVType_Encoding_DefaultXml'
ObjectIdNames[12083] = 'OpcUa_XmlSchema_AxisInformation'
ObjectIdNames[12084] = 'OpcUa_XmlSchema_AxisInformation_DataTypeVersion'
ObjectIdNames[12085] = 'OpcUa_XmlSchema_AxisInformation_DictionaryFragment'
ObjectIdNames[12086] = 'OpcUa_XmlSchema_XVType'
ObjectIdNames[12087] = 'OpcUa_XmlSchema_XVType_DataTypeVersion'
ObjectIdNames[12088] = 'OpcUa_XmlSchema_XVType_DictionaryFragment'
ObjectIdNames[12089] = 'AxisInformation_Encoding_DefaultBinary'
ObjectIdNames[12090] = 'XVType_Encoding_DefaultBinary'
ObjectIdNames[12091] = 'OpcUa_BinarySchema_AxisInformation'
ObjectIdNames[12092] = 'OpcUa_BinarySchema_AxisInformation_DataTypeVersion'
ObjectIdNames[12093] = 'OpcUa_BinarySchema_AxisInformation_DictionaryFragment'
ObjectIdNames[12094] = 'OpcUa_BinarySchema_XVType'
ObjectIdNames[12095] = 'OpcUa_BinarySchema_XVType_DataTypeVersion'
ObjectIdNames[12096] = 'OpcUa_BinarySchema_XVType_DictionaryFragment'
ObjectIdNames[12097] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder'
ObjectIdNames[12098] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics'
ObjectIdNames[12099] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_SessionId'
ObjectIdNames[12100] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_SessionName'
ObjectIdNames[12101] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_ClientDescription'
ObjectIdNames[12102] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_ServerUri'
ObjectIdNames[12103] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_EndpointUrl'
ObjectIdNames[12104] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_LocaleIds'
ObjectIdNames[12105] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_ActualSessionTimeout'
ObjectIdNames[12106] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_MaxResponseMessageSize'
ObjectIdNames[12107] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_ClientConnectionTime'
ObjectIdNames[12108] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_ClientLastContactTime'
ObjectIdNames[12109] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_CurrentSubscriptionsCount'
ObjectIdNames[12110] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_CurrentMonitoredItemsCount'
ObjectIdNames[12111] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_CurrentPublishRequestsInQueue'
ObjectIdNames[12112] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_TotalRequestCount'
ObjectIdNames[12113] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_UnauthorizedRequestCount'
ObjectIdNames[12114] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_ReadCount'
ObjectIdNames[12115] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_HistoryReadCount'
ObjectIdNames[12116] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_WriteCount'
ObjectIdNames[12117] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_HistoryUpdateCount'
ObjectIdNames[12118] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_CallCount'
ObjectIdNames[12119] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_CreateMonitoredItemsCount'
ObjectIdNames[12120] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_ModifyMonitoredItemsCount'
ObjectIdNames[12121] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_SetMonitoringModeCount'
ObjectIdNames[12122] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_SetTriggeringCount'
ObjectIdNames[12123] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_DeleteMonitoredItemsCount'
ObjectIdNames[12124] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_CreateSubscriptionCount'
ObjectIdNames[12125] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_ModifySubscriptionCount'
ObjectIdNames[12126] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_SetPublishingModeCount'
ObjectIdNames[12127] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_PublishCount'
ObjectIdNames[12128] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_RepublishCount'
ObjectIdNames[12129] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_TransferSubscriptionsCount'
ObjectIdNames[12130] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_DeleteSubscriptionsCount'
ObjectIdNames[12131] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_AddNodesCount'
ObjectIdNames[12132] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_AddReferencesCount'
ObjectIdNames[12133] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_DeleteNodesCount'
ObjectIdNames[12134] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_DeleteReferencesCount'
ObjectIdNames[12135] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_BrowseCount'
ObjectIdNames[12136] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_BrowseNextCount'
ObjectIdNames[12137] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_TranslateBrowsePathsToNodeIdsCount'
ObjectIdNames[12138] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_QueryFirstCount'
ObjectIdNames[12139] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_QueryNextCount'
ObjectIdNames[12140] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_RegisterNodesCount'
ObjectIdNames[12141] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionDiagnostics_UnregisterNodesCount'
ObjectIdNames[12142] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionSecurityDiagnostics'
ObjectIdNames[12143] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionSecurityDiagnostics_SessionId'
ObjectIdNames[12144] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionSecurityDiagnostics_ClientUserIdOfSession'
ObjectIdNames[12145] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionSecurityDiagnostics_ClientUserIdHistory'
ObjectIdNames[12146] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionSecurityDiagnostics_AuthenticationMechanism'
ObjectIdNames[12147] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionSecurityDiagnostics_Encoding'
ObjectIdNames[12148] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionSecurityDiagnostics_TransportProtocol'
ObjectIdNames[12149] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionSecurityDiagnostics_SecurityMode'
ObjectIdNames[12150] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionSecurityDiagnostics_SecurityPolicyUri'
ObjectIdNames[12151] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SessionSecurityDiagnostics_ClientCertificate'
ObjectIdNames[12152] = 'SessionsDiagnosticsSummaryType_SessionPlaceholder_SubscriptionDiagnosticsArray'
ObjectIdNames[12153] = 'ServerType_ServerCapabilities_OperationLimits_MaxNodesPerHistoryReadData'
ObjectIdNames[12154] = 'ServerType_ServerCapabilities_OperationLimits_MaxNodesPerHistoryReadEvents'
ObjectIdNames[12155] = 'ServerType_ServerCapabilities_OperationLimits_MaxNodesPerHistoryUpdateData'
ObjectIdNames[12156] = 'ServerType_ServerCapabilities_OperationLimits_MaxNodesPerHistoryUpdateEvents'
ObjectIdNames[12157] = 'ServerCapabilitiesType_OperationLimits_MaxNodesPerHistoryReadData'
ObjectIdNames[12158] = 'ServerCapabilitiesType_OperationLimits_MaxNodesPerHistoryReadEvents'
ObjectIdNames[12159] = 'ServerCapabilitiesType_OperationLimits_MaxNodesPerHistoryUpdateData'
ObjectIdNames[12160] = 'ServerCapabilitiesType_OperationLimits_MaxNodesPerHistoryUpdateEvents'
ObjectIdNames[12161] = 'OperationLimitsType_MaxNodesPerHistoryReadData'
ObjectIdNames[12162] = 'OperationLimitsType_MaxNodesPerHistoryReadEvents'
ObjectIdNames[12163] = 'OperationLimitsType_MaxNodesPerHistoryUpdateData'
ObjectIdNames[12164] = 'OperationLimitsType_MaxNodesPerHistoryUpdateEvents'
ObjectIdNames[12165] = 'Server_ServerCapabilities_OperationLimits_MaxNodesPerHistoryReadData'
ObjectIdNames[12166] = 'Server_ServerCapabilities_OperationLimits_MaxNodesPerHistoryReadEvents'
ObjectIdNames[12167] = 'Server_ServerCapabilities_OperationLimits_MaxNodesPerHistoryUpdateData'
ObjectIdNames[12168] = 'Server_ServerCapabilities_OperationLimits_MaxNodesPerHistoryUpdateEvents'
ObjectIdNames[12169] = 'NamingRuleType_EnumValues'
ObjectIdNames[12170] = 'ViewVersion'
ObjectIdNames[12171] = 'ComplexNumberType'
ObjectIdNames[12172] = 'DoubleComplexNumberType'
ObjectIdNames[12173] = 'ComplexNumberType_Encoding_DefaultXml'
ObjectIdNames[12174] = 'DoubleComplexNumberType_Encoding_DefaultXml'
ObjectIdNames[12175] = 'OpcUa_XmlSchema_ComplexNumberType'
ObjectIdNames[12176] = 'OpcUa_XmlSchema_ComplexNumberType_DataTypeVersion'
ObjectIdNames[12177] = 'OpcUa_XmlSchema_ComplexNumberType_DictionaryFragment'
ObjectIdNames[12178] = 'OpcUa_XmlSchema_DoubleComplexNumberType'
ObjectIdNames[12179] = 'OpcUa_XmlSchema_DoubleComplexNumberType_DataTypeVersion'
ObjectIdNames[12180] = 'OpcUa_XmlSchema_DoubleComplexNumberType_DictionaryFragment'
ObjectIdNames[12181] = 'ComplexNumberType_Encoding_DefaultBinary'
ObjectIdNames[12182] = 'DoubleComplexNumberType_Encoding_DefaultBinary'
ObjectIdNames[12183] = 'OpcUa_BinarySchema_ComplexNumberType'
ObjectIdNames[12184] = 'OpcUa_BinarySchema_ComplexNumberType_DataTypeVersion'
ObjectIdNames[12185] = 'OpcUa_BinarySchema_ComplexNumberType_DictionaryFragment'
ObjectIdNames[12186] = 'OpcUa_BinarySchema_DoubleComplexNumberType'
ObjectIdNames[12187] = 'OpcUa_BinarySchema_DoubleComplexNumberType_DataTypeVersion'
ObjectIdNames[12188] = 'OpcUa_BinarySchema_DoubleComplexNumberType_DictionaryFragment'
ObjectIdNames[12189] = 'ServerOnNetwork'
ObjectIdNames[12190] = 'FindServersOnNetworkRequest'
ObjectIdNames[12191] = 'FindServersOnNetworkResponse'
ObjectIdNames[12193] = 'RegisterServer2Request'
ObjectIdNames[12194] = 'RegisterServer2Response'
ObjectIdNames[12195] = 'ServerOnNetwork_Encoding_DefaultXml'
ObjectIdNames[12196] = 'FindServersOnNetworkRequest_Encoding_DefaultXml'
ObjectIdNames[12197] = 'FindServersOnNetworkResponse_Encoding_DefaultXml'
ObjectIdNames[12199] = 'RegisterServer2Request_Encoding_DefaultXml'
ObjectIdNames[12200] = 'RegisterServer2Response_Encoding_DefaultXml'
ObjectIdNames[12201] = 'OpcUa_XmlSchema_ServerOnNetwork'
ObjectIdNames[12202] = 'OpcUa_XmlSchema_ServerOnNetwork_DataTypeVersion'
ObjectIdNames[12203] = 'OpcUa_XmlSchema_ServerOnNetwork_DictionaryFragment'
ObjectIdNames[12207] = 'ServerOnNetwork_Encoding_DefaultBinary'
ObjectIdNames[12208] = 'FindServersOnNetworkRequest_Encoding_DefaultBinary'
ObjectIdNames[12209] = 'FindServersOnNetworkResponse_Encoding_DefaultBinary'
ObjectIdNames[12211] = 'RegisterServer2Request_Encoding_DefaultBinary'
ObjectIdNames[12212] = 'RegisterServer2Response_Encoding_DefaultBinary'
ObjectIdNames[12213] = 'OpcUa_BinarySchema_ServerOnNetwork'
ObjectIdNames[12214] = 'OpcUa_BinarySchema_ServerOnNetwork_DataTypeVersion'
ObjectIdNames[12215] = 'OpcUa_BinarySchema_ServerOnNetwork_DictionaryFragment'
ObjectIdNames[12502] = 'ProgressEventType_Context'
ObjectIdNames[12503] = 'ProgressEventType_Progress'
ObjectIdNames[12504] = 'KerberosIdentityToken'
ObjectIdNames[12505] = 'KerberosIdentityToken_Encoding_DefaultXml'
ObjectIdNames[12506] = 'OpcUa_XmlSchema_KerberosIdentityToken'
ObjectIdNames[12507] = 'OpcUa_XmlSchema_KerberosIdentityToken_DataTypeVersion'
ObjectIdNames[12508] = 'OpcUa_XmlSchema_KerberosIdentityToken_DictionaryFragment'
ObjectIdNames[12509] = 'KerberosIdentityToken_Encoding_DefaultBinary'
ObjectIdNames[12510] = 'OpcUa_BinarySchema_KerberosIdentityToken'
ObjectIdNames[12511] = 'OpcUa_BinarySchema_KerberosIdentityToken_DataTypeVersion'
ObjectIdNames[12512] = 'OpcUa_BinarySchema_KerberosIdentityToken_DictionaryFragment'
ObjectIdNames[12513] = 'OpenWithMasksMethodType'
ObjectIdNames[12514] = 'OpenWithMasksMethodType_InputArguments'
ObjectIdNames[12515] = 'OpenWithMasksMethodType_OutputArguments'
ObjectIdNames[12516] = 'CloseAndUpdateMethodType'
ObjectIdNames[12517] = 'CloseAndUpdateMethodType_OutputArguments'
ObjectIdNames[12518] = 'AddCertificateMethodType'
ObjectIdNames[12519] = 'AddCertificateMethodType_InputArguments'
ObjectIdNames[12520] = 'RemoveCertificateMethodType'
ObjectIdNames[12521] = 'RemoveCertificateMethodType_InputArguments'
ObjectIdNames[12522] = 'TrustListType'
ObjectIdNames[12523] = 'TrustListType_Size'
ObjectIdNames[12526] = 'TrustListType_OpenCount'
ObjectIdNames[12527] = 'TrustListType_Open'
ObjectIdNames[12528] = 'TrustListType_Open_InputArguments'
ObjectIdNames[12529] = 'TrustListType_Open_OutputArguments'
ObjectIdNames[12530] = 'TrustListType_Close'
ObjectIdNames[12531] = 'TrustListType_Close_InputArguments'
ObjectIdNames[12532] = 'TrustListType_Read'
ObjectIdNames[12533] = 'TrustListType_Read_InputArguments'
ObjectIdNames[12534] = 'TrustListType_Read_OutputArguments'
ObjectIdNames[12535] = 'TrustListType_Write'
ObjectIdNames[12536] = 'TrustListType_Write_InputArguments'
ObjectIdNames[12537] = 'TrustListType_GetPosition'
ObjectIdNames[12538] = 'TrustListType_GetPosition_InputArguments'
ObjectIdNames[12539] = 'TrustListType_GetPosition_OutputArguments'
ObjectIdNames[12540] = 'TrustListType_SetPosition'
ObjectIdNames[12541] = 'TrustListType_SetPosition_InputArguments'
ObjectIdNames[12542] = 'TrustListType_LastUpdateTime'
ObjectIdNames[12543] = 'TrustListType_OpenWithMasks'
ObjectIdNames[12544] = 'TrustListType_OpenWithMasks_InputArguments'
ObjectIdNames[12545] = 'TrustListType_OpenWithMasks_OutputArguments'
ObjectIdNames[12546] = 'TrustListType_CloseAndUpdate'
ObjectIdNames[12547] = 'TrustListType_CloseAndUpdate_OutputArguments'
ObjectIdNames[12548] = 'TrustListType_AddCertificate'
ObjectIdNames[12549] = 'TrustListType_AddCertificate_InputArguments'
ObjectIdNames[12550] = 'TrustListType_RemoveCertificate'
ObjectIdNames[12551] = 'TrustListType_RemoveCertificate_InputArguments'
ObjectIdNames[12552] = 'TrustListMasks'
ObjectIdNames[12553] = 'TrustListMasks_EnumValues'
ObjectIdNames[12554] = 'TrustListDataType'
ObjectIdNames[12555] = 'CertificateGroupType'
ObjectIdNames[12556] = 'CertificateType'
ObjectIdNames[12557] = 'ApplicationCertificateType'
ObjectIdNames[12558] = 'HttpsCertificateType'
ObjectIdNames[12559] = 'RsaMinApplicationCertificateType'
ObjectIdNames[12560] = 'RsaSha256ApplicationCertificateType'
ObjectIdNames[12561] = 'TrustListUpdatedAuditEventType'
ObjectIdNames[12562] = 'TrustListUpdatedAuditEventType_EventId'
ObjectIdNames[12563] = 'TrustListUpdatedAuditEventType_EventType'
ObjectIdNames[12564] = 'TrustListUpdatedAuditEventType_SourceNode'
ObjectIdNames[12565] = 'TrustListUpdatedAuditEventType_SourceName'
ObjectIdNames[12566] = 'TrustListUpdatedAuditEventType_Time'
ObjectIdNames[12567] = 'TrustListUpdatedAuditEventType_ReceiveTime'
ObjectIdNames[12568] = 'TrustListUpdatedAuditEventType_LocalTime'
ObjectIdNames[12569] = 'TrustListUpdatedAuditEventType_Message'
ObjectIdNames[12570] = 'TrustListUpdatedAuditEventType_Severity'
ObjectIdNames[12571] = 'TrustListUpdatedAuditEventType_ActionTimeStamp'
ObjectIdNames[12572] = 'TrustListUpdatedAuditEventType_Status'
ObjectIdNames[12573] = 'TrustListUpdatedAuditEventType_ServerId'
ObjectIdNames[12574] = 'TrustListUpdatedAuditEventType_ClientAuditEntryId'
ObjectIdNames[12575] = 'TrustListUpdatedAuditEventType_ClientUserId'
ObjectIdNames[12576] = 'TrustListUpdatedAuditEventType_MethodId'
ObjectIdNames[12577] = 'TrustListUpdatedAuditEventType_InputArguments'
ObjectIdNames[12578] = 'UpdateCertificateMethodType'
ObjectIdNames[12579] = 'UpdateCertificateMethodType_InputArguments'
ObjectIdNames[12580] = 'UpdateCertificateMethodType_OutputArguments'
ObjectIdNames[12581] = 'ServerConfigurationType'
ObjectIdNames[12583] = 'ServerConfigurationType_SupportedPrivateKeyFormats'
ObjectIdNames[12584] = 'ServerConfigurationType_MaxTrustListSize'
ObjectIdNames[12585] = 'ServerConfigurationType_MulticastDnsEnabled'
ObjectIdNames[12616] = 'ServerConfigurationType_UpdateCertificate'
ObjectIdNames[12617] = 'ServerConfigurationType_UpdateCertificate_InputArguments'
ObjectIdNames[12618] = 'ServerConfigurationType_UpdateCertificate_OutputArguments'
ObjectIdNames[12620] = 'CertificateUpdatedAuditEventType'
ObjectIdNames[12621] = 'CertificateUpdatedAuditEventType_EventId'
ObjectIdNames[12622] = 'CertificateUpdatedAuditEventType_EventType'
ObjectIdNames[12623] = 'CertificateUpdatedAuditEventType_SourceNode'
ObjectIdNames[12624] = 'CertificateUpdatedAuditEventType_SourceName'
ObjectIdNames[12625] = 'CertificateUpdatedAuditEventType_Time'
ObjectIdNames[12626] = 'CertificateUpdatedAuditEventType_ReceiveTime'
ObjectIdNames[12627] = 'CertificateUpdatedAuditEventType_LocalTime'
ObjectIdNames[12628] = 'CertificateUpdatedAuditEventType_Message'
ObjectIdNames[12629] = 'CertificateUpdatedAuditEventType_Severity'
ObjectIdNames[12630] = 'CertificateUpdatedAuditEventType_ActionTimeStamp'
ObjectIdNames[12631] = 'CertificateUpdatedAuditEventType_Status'
ObjectIdNames[12632] = 'CertificateUpdatedAuditEventType_ServerId'
ObjectIdNames[12633] = 'CertificateUpdatedAuditEventType_ClientAuditEntryId'
ObjectIdNames[12634] = 'CertificateUpdatedAuditEventType_ClientUserId'
ObjectIdNames[12635] = 'CertificateUpdatedAuditEventType_MethodId'
ObjectIdNames[12636] = 'CertificateUpdatedAuditEventType_InputArguments'
ObjectIdNames[12637] = 'ServerConfiguration'
ObjectIdNames[12639] = 'ServerConfiguration_SupportedPrivateKeyFormats'
ObjectIdNames[12640] = 'ServerConfiguration_MaxTrustListSize'
ObjectIdNames[12641] = 'ServerConfiguration_MulticastDnsEnabled'
ObjectIdNames[12642] = 'ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList'
ObjectIdNames[12643] = 'ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_Size'
ObjectIdNames[12646] = 'ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_OpenCount'
ObjectIdNames[12647] = 'ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_Open'
ObjectIdNames[12648] = 'ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_Open_InputArguments'
ObjectIdNames[12649] = 'ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_Open_OutputArguments'
ObjectIdNames[12650] = 'ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_Close'
ObjectIdNames[12651] = 'ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_Close_InputArguments'
ObjectIdNames[12652] = 'ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_Read'
ObjectIdNames[12653] = 'ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_Read_InputArguments'
ObjectIdNames[12654] = 'ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_Read_OutputArguments'
ObjectIdNames[12655] = 'ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_Write'
ObjectIdNames[12656] = 'ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_Write_InputArguments'
ObjectIdNames[12657] = 'ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_GetPosition'
ObjectIdNames[12658] = 'ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_GetPosition_InputArguments'
ObjectIdNames[12659] = 'ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_GetPosition_OutputArguments'
ObjectIdNames[12660] = 'ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_SetPosition'
ObjectIdNames[12661] = 'ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_SetPosition_InputArguments'
ObjectIdNames[12662] = 'ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_LastUpdateTime'
ObjectIdNames[12663] = 'ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_OpenWithMasks'
ObjectIdNames[12664] = 'ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_OpenWithMasks_InputArguments'
ObjectIdNames[12665] = 'ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_OpenWithMasks_OutputArguments'
ObjectIdNames[12666] = 'ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_CloseAndUpdate'
ObjectIdNames[12667] = 'ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_CloseAndUpdate_OutputArguments'
ObjectIdNames[12668] = 'ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_AddCertificate'
ObjectIdNames[12669] = 'ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_AddCertificate_InputArguments'
ObjectIdNames[12670] = 'ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_RemoveCertificate'
ObjectIdNames[12671] = 'ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_RemoveCertificate_InputArguments'
ObjectIdNames[12676] = 'TrustListDataType_Encoding_DefaultXml'
ObjectIdNames[12677] = 'OpcUa_XmlSchema_TrustListDataType'
ObjectIdNames[12678] = 'OpcUa_XmlSchema_TrustListDataType_DataTypeVersion'
ObjectIdNames[12679] = 'OpcUa_XmlSchema_TrustListDataType_DictionaryFragment'
ObjectIdNames[12680] = 'TrustListDataType_Encoding_DefaultBinary'
ObjectIdNames[12681] = 'OpcUa_BinarySchema_TrustListDataType'
ObjectIdNames[12682] = 'OpcUa_BinarySchema_TrustListDataType_DataTypeVersion'
ObjectIdNames[12683] = 'OpcUa_BinarySchema_TrustListDataType_DictionaryFragment'
ObjectIdNames[12684] = 'ServerType_Namespaces_AddressSpaceFile_Writable'
ObjectIdNames[12685] = 'ServerType_Namespaces_AddressSpaceFile_UserWritable'
ObjectIdNames[12686] = 'FileType_Writable'
ObjectIdNames[12687] = 'FileType_UserWritable'
ObjectIdNames[12688] = 'AddressSpaceFileType_Writable'
ObjectIdNames[12689] = 'AddressSpaceFileType_UserWritable'
ObjectIdNames[12690] = 'NamespaceMetadataType_NamespaceFile_Writable'
ObjectIdNames[12691] = 'NamespaceMetadataType_NamespaceFile_UserWritable'
ObjectIdNames[12692] = 'NamespacesType_NamespaceIdentifier_NamespaceFile_Writable'
ObjectIdNames[12693] = 'NamespacesType_NamespaceIdentifier_NamespaceFile_UserWritable'
ObjectIdNames[12694] = 'NamespacesType_AddressSpaceFile_Writable'
ObjectIdNames[12695] = 'NamespacesType_AddressSpaceFile_UserWritable'
ObjectIdNames[12696] = 'Server_Namespaces_AddressSpaceFile_Writable'
ObjectIdNames[12697] = 'Server_Namespaces_AddressSpaceFile_UserWritable'
ObjectIdNames[12698] = 'TrustListType_Writable'
ObjectIdNames[12699] = 'TrustListType_UserWritable'
ObjectIdNames[12704] = 'CloseAndUpdateMethodType_InputArguments'
ObjectIdNames[12705] = 'TrustListType_CloseAndUpdate_InputArguments'
ObjectIdNames[12708] = 'ServerConfigurationType_ServerCapabilities'
ObjectIdNames[12710] = 'ServerConfiguration_ServerCapabilities'
ObjectIdNames[12712] = 'OpcUa_XmlSchema_RelativePathElement'
ObjectIdNames[12713] = 'OpcUa_XmlSchema_RelativePathElement_DataTypeVersion'
ObjectIdNames[12714] = 'OpcUa_XmlSchema_RelativePathElement_DictionaryFragment'
ObjectIdNames[12715] = 'OpcUa_XmlSchema_RelativePath'
ObjectIdNames[12716] = 'OpcUa_XmlSchema_RelativePath_DataTypeVersion'
ObjectIdNames[12717] = 'OpcUa_XmlSchema_RelativePath_DictionaryFragment'
ObjectIdNames[12718] = 'OpcUa_BinarySchema_RelativePathElement'
ObjectIdNames[12719] = 'OpcUa_BinarySchema_RelativePathElement_DataTypeVersion'
ObjectIdNames[12720] = 'OpcUa_BinarySchema_RelativePathElement_DictionaryFragment'
ObjectIdNames[12721] = 'OpcUa_BinarySchema_RelativePath'
ObjectIdNames[12722] = 'OpcUa_BinarySchema_RelativePath_DataTypeVersion'
ObjectIdNames[12723] = 'OpcUa_BinarySchema_RelativePath_DictionaryFragment'
ObjectIdNames[12731] = 'ServerConfigurationType_CreateSigningRequest'
ObjectIdNames[12732] = 'ServerConfigurationType_CreateSigningRequest_InputArguments'
ObjectIdNames[12733] = 'ServerConfigurationType_CreateSigningRequest_OutputArguments'
ObjectIdNames[12734] = 'ServerConfigurationType_ApplyChanges'
ObjectIdNames[12737] = 'ServerConfiguration_CreateSigningRequest'
ObjectIdNames[12738] = 'ServerConfiguration_CreateSigningRequest_InputArguments'
ObjectIdNames[12739] = 'ServerConfiguration_CreateSigningRequest_OutputArguments'
ObjectIdNames[12740] = 'ServerConfiguration_ApplyChanges'
ObjectIdNames[12741] = 'CreateSigningRequestMethodType'
ObjectIdNames[12742] = 'CreateSigningRequestMethodType_InputArguments'
ObjectIdNames[12743] = 'CreateSigningRequestMethodType_OutputArguments'
ObjectIdNames[12745] = 'OptionSetValues'
ObjectIdNames[12746] = 'ServerType_SetSubscriptionDurable'
ObjectIdNames[12747] = 'ServerType_SetSubscriptionDurable_InputArguments'
ObjectIdNames[12748] = 'ServerType_SetSubscriptionDurable_OutputArguments'
ObjectIdNames[12749] = 'Server_SetSubscriptionDurable'
ObjectIdNames[12750] = 'Server_SetSubscriptionDurable_InputArguments'
ObjectIdNames[12751] = 'Server_SetSubscriptionDurable_OutputArguments'
ObjectIdNames[12752] = 'SetSubscriptionDurableMethodType'
ObjectIdNames[12753] = 'SetSubscriptionDurableMethodType_InputArguments'
ObjectIdNames[12754] = 'SetSubscriptionDurableMethodType_OutputArguments'
ObjectIdNames[12755] = 'OptionSet'
ObjectIdNames[12756] = 'Union'
ObjectIdNames[12757] = 'OptionSet_Encoding_DefaultXml'
ObjectIdNames[12758] = 'Union_Encoding_DefaultXml'
ObjectIdNames[12759] = 'OpcUa_XmlSchema_OptionSet'
ObjectIdNames[12760] = 'OpcUa_XmlSchema_OptionSet_DataTypeVersion'
ObjectIdNames[12761] = 'OpcUa_XmlSchema_OptionSet_DictionaryFragment'
ObjectIdNames[12762] = 'OpcUa_XmlSchema_Union'
ObjectIdNames[12763] = 'OpcUa_XmlSchema_Union_DataTypeVersion'
ObjectIdNames[12764] = 'OpcUa_XmlSchema_Union_DictionaryFragment'
ObjectIdNames[12765] = 'OptionSet_Encoding_DefaultBinary'
ObjectIdNames[12766] = 'Union_Encoding_DefaultBinary'
ObjectIdNames[12767] = 'OpcUa_BinarySchema_OptionSet'
ObjectIdNames[12768] = 'OpcUa_BinarySchema_OptionSet_DataTypeVersion'
ObjectIdNames[12769] = 'OpcUa_BinarySchema_OptionSet_DictionaryFragment'
ObjectIdNames[12770] = 'OpcUa_BinarySchema_Union'
ObjectIdNames[12771] = 'OpcUa_BinarySchema_Union_DataTypeVersion'
ObjectIdNames[12772] = 'OpcUa_BinarySchema_Union_DictionaryFragment'
ObjectIdNames[12773] = 'GetRejectedListMethodType'
ObjectIdNames[12774] = 'GetRejectedListMethodType_OutputArguments'
ObjectIdNames[12775] = 'ServerConfigurationType_GetRejectedList'
ObjectIdNames[12776] = 'ServerConfigurationType_GetRejectedList_OutputArguments'
ObjectIdNames[12777] = 'ServerConfiguration_GetRejectedList'
ObjectIdNames[12778] = 'ServerConfiguration_GetRejectedList_OutputArguments'
ObjectIdNames[12779] = 'SamplingIntervalDiagnosticsArrayType_SamplingIntervalDiagnostics'
ObjectIdNames[12780] = 'SamplingIntervalDiagnosticsArrayType_SamplingIntervalDiagnostics_SamplingInterval'
ObjectIdNames[12781] = 'SamplingIntervalDiagnosticsArrayType_SamplingIntervalDiagnostics_SampledMonitoredItemsCount'
ObjectIdNames[12782] = 'SamplingIntervalDiagnosticsArrayType_SamplingIntervalDiagnostics_MaxSampledMonitoredItemsCount'
ObjectIdNames[12783] = 'SamplingIntervalDiagnosticsArrayType_SamplingIntervalDiagnostics_DisabledMonitoredItemsSamplingCount'
ObjectIdNames[12784] = 'SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics'
ObjectIdNames[12785] = 'SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_SessionId'
ObjectIdNames[12786] = 'SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_SubscriptionId'
ObjectIdNames[12787] = 'SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_Priority'
ObjectIdNames[12788] = 'SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_PublishingInterval'
ObjectIdNames[12789] = 'SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_MaxKeepAliveCount'
ObjectIdNames[12790] = 'SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_MaxLifetimeCount'
ObjectIdNames[12791] = 'SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_MaxNotificationsPerPublish'
ObjectIdNames[12792] = 'SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_PublishingEnabled'
ObjectIdNames[12793] = 'SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_ModifyCount'
ObjectIdNames[12794] = 'SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_EnableCount'
ObjectIdNames[12795] = 'SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_DisableCount'
ObjectIdNames[12796] = 'SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_RepublishRequestCount'
ObjectIdNames[12797] = 'SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_RepublishMessageRequestCount'
ObjectIdNames[12798] = 'SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_RepublishMessageCount'
ObjectIdNames[12799] = 'SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_TransferRequestCount'
ObjectIdNames[12800] = 'SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_TransferredToAltClientCount'
ObjectIdNames[12801] = 'SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_TransferredToSameClientCount'
ObjectIdNames[12802] = 'SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_PublishRequestCount'
ObjectIdNames[12803] = 'SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_DataChangeNotificationsCount'
ObjectIdNames[12804] = 'SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_EventNotificationsCount'
ObjectIdNames[12805] = 'SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_NotificationsCount'
ObjectIdNames[12806] = 'SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_LatePublishRequestCount'
ObjectIdNames[12807] = 'SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_CurrentKeepAliveCount'
ObjectIdNames[12808] = 'SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_CurrentLifetimeCount'
ObjectIdNames[12809] = 'SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_UnacknowledgedMessageCount'
ObjectIdNames[12810] = 'SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_DiscardedMessageCount'
ObjectIdNames[12811] = 'SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_MonitoredItemCount'
ObjectIdNames[12812] = 'SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_DisabledMonitoredItemCount'
ObjectIdNames[12813] = 'SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_MonitoringQueueOverflowCount'
ObjectIdNames[12814] = 'SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_NextSequenceNumber'
ObjectIdNames[12815] = 'SubscriptionDiagnosticsArrayType_SubscriptionDiagnostics_EventQueueOverFlowCount'
ObjectIdNames[12816] = 'SessionDiagnosticsArrayType_SessionDiagnostics'
ObjectIdNames[12817] = 'SessionDiagnosticsArrayType_SessionDiagnostics_SessionId'
ObjectIdNames[12818] = 'SessionDiagnosticsArrayType_SessionDiagnostics_SessionName'
ObjectIdNames[12819] = 'SessionDiagnosticsArrayType_SessionDiagnostics_ClientDescription'
ObjectIdNames[12820] = 'SessionDiagnosticsArrayType_SessionDiagnostics_ServerUri'
ObjectIdNames[12821] = 'SessionDiagnosticsArrayType_SessionDiagnostics_EndpointUrl'
ObjectIdNames[12822] = 'SessionDiagnosticsArrayType_SessionDiagnostics_LocaleIds'
ObjectIdNames[12823] = 'SessionDiagnosticsArrayType_SessionDiagnostics_ActualSessionTimeout'
ObjectIdNames[12824] = 'SessionDiagnosticsArrayType_SessionDiagnostics_MaxResponseMessageSize'
ObjectIdNames[12825] = 'SessionDiagnosticsArrayType_SessionDiagnostics_ClientConnectionTime'
ObjectIdNames[12826] = 'SessionDiagnosticsArrayType_SessionDiagnostics_ClientLastContactTime'
ObjectIdNames[12827] = 'SessionDiagnosticsArrayType_SessionDiagnostics_CurrentSubscriptionsCount'
ObjectIdNames[12828] = 'SessionDiagnosticsArrayType_SessionDiagnostics_CurrentMonitoredItemsCount'
ObjectIdNames[12829] = 'SessionDiagnosticsArrayType_SessionDiagnostics_CurrentPublishRequestsInQueue'
ObjectIdNames[12830] = 'SessionDiagnosticsArrayType_SessionDiagnostics_TotalRequestCount'
ObjectIdNames[12831] = 'SessionDiagnosticsArrayType_SessionDiagnostics_UnauthorizedRequestCount'
ObjectIdNames[12832] = 'SessionDiagnosticsArrayType_SessionDiagnostics_ReadCount'
ObjectIdNames[12833] = 'SessionDiagnosticsArrayType_SessionDiagnostics_HistoryReadCount'
ObjectIdNames[12834] = 'SessionDiagnosticsArrayType_SessionDiagnostics_WriteCount'
ObjectIdNames[12835] = 'SessionDiagnosticsArrayType_SessionDiagnostics_HistoryUpdateCount'
ObjectIdNames[12836] = 'SessionDiagnosticsArrayType_SessionDiagnostics_CallCount'
ObjectIdNames[12837] = 'SessionDiagnosticsArrayType_SessionDiagnostics_CreateMonitoredItemsCount'
ObjectIdNames[12838] = 'SessionDiagnosticsArrayType_SessionDiagnostics_ModifyMonitoredItemsCount'
ObjectIdNames[12839] = 'SessionDiagnosticsArrayType_SessionDiagnostics_SetMonitoringModeCount'
ObjectIdNames[12840] = 'SessionDiagnosticsArrayType_SessionDiagnostics_SetTriggeringCount'
ObjectIdNames[12841] = 'SessionDiagnosticsArrayType_SessionDiagnostics_DeleteMonitoredItemsCount'
ObjectIdNames[12842] = 'SessionDiagnosticsArrayType_SessionDiagnostics_CreateSubscriptionCount'
ObjectIdNames[12843] = 'SessionDiagnosticsArrayType_SessionDiagnostics_ModifySubscriptionCount'
ObjectIdNames[12844] = 'SessionDiagnosticsArrayType_SessionDiagnostics_SetPublishingModeCount'
ObjectIdNames[12845] = 'SessionDiagnosticsArrayType_SessionDiagnostics_PublishCount'
ObjectIdNames[12846] = 'SessionDiagnosticsArrayType_SessionDiagnostics_RepublishCount'
ObjectIdNames[12847] = 'SessionDiagnosticsArrayType_SessionDiagnostics_TransferSubscriptionsCount'
ObjectIdNames[12848] = 'SessionDiagnosticsArrayType_SessionDiagnostics_DeleteSubscriptionsCount'
ObjectIdNames[12849] = 'SessionDiagnosticsArrayType_SessionDiagnostics_AddNodesCount'
ObjectIdNames[12850] = 'SessionDiagnosticsArrayType_SessionDiagnostics_AddReferencesCount'
ObjectIdNames[12851] = 'SessionDiagnosticsArrayType_SessionDiagnostics_DeleteNodesCount'
ObjectIdNames[12852] = 'SessionDiagnosticsArrayType_SessionDiagnostics_DeleteReferencesCount'
ObjectIdNames[12853] = 'SessionDiagnosticsArrayType_SessionDiagnostics_BrowseCount'
ObjectIdNames[12854] = 'SessionDiagnosticsArrayType_SessionDiagnostics_BrowseNextCount'
ObjectIdNames[12855] = 'SessionDiagnosticsArrayType_SessionDiagnostics_TranslateBrowsePathsToNodeIdsCount'
ObjectIdNames[12856] = 'SessionDiagnosticsArrayType_SessionDiagnostics_QueryFirstCount'
ObjectIdNames[12857] = 'SessionDiagnosticsArrayType_SessionDiagnostics_QueryNextCount'
ObjectIdNames[12858] = 'SessionDiagnosticsArrayType_SessionDiagnostics_RegisterNodesCount'
ObjectIdNames[12859] = 'SessionDiagnosticsArrayType_SessionDiagnostics_UnregisterNodesCount'
ObjectIdNames[12860] = 'SessionSecurityDiagnosticsArrayType_SessionSecurityDiagnostics'
ObjectIdNames[12861] = 'SessionSecurityDiagnosticsArrayType_SessionSecurityDiagnostics_SessionId'
ObjectIdNames[12862] = 'SessionSecurityDiagnosticsArrayType_SessionSecurityDiagnostics_ClientUserIdOfSession'
ObjectIdNames[12863] = 'SessionSecurityDiagnosticsArrayType_SessionSecurityDiagnostics_ClientUserIdHistory'
ObjectIdNames[12864] = 'SessionSecurityDiagnosticsArrayType_SessionSecurityDiagnostics_AuthenticationMechanism'
ObjectIdNames[12865] = 'SessionSecurityDiagnosticsArrayType_SessionSecurityDiagnostics_Encoding'
ObjectIdNames[12866] = 'SessionSecurityDiagnosticsArrayType_SessionSecurityDiagnostics_TransportProtocol'
ObjectIdNames[12867] = 'SessionSecurityDiagnosticsArrayType_SessionSecurityDiagnostics_SecurityMode'
ObjectIdNames[12868] = 'SessionSecurityDiagnosticsArrayType_SessionSecurityDiagnostics_SecurityPolicyUri'
ObjectIdNames[12869] = 'SessionSecurityDiagnosticsArrayType_SessionSecurityDiagnostics_ClientCertificate'
ObjectIdNames[12871] = 'ServerType_ResendData'
ObjectIdNames[12872] = 'ServerType_ResendData_InputArguments'
ObjectIdNames[12873] = 'Server_ResendData'
ObjectIdNames[12874] = 'Server_ResendData_InputArguments'
ObjectIdNames[12875] = 'ResendDataMethodType'
ObjectIdNames[12876] = 'ResendDataMethodType_InputArguments'
ObjectIdNames[12877] = 'NormalizedString'
ObjectIdNames[12878] = 'DecimalString'
ObjectIdNames[12879] = 'DurationString'
ObjectIdNames[12880] = 'TimeString'
ObjectIdNames[12881] = 'DateString'
ObjectIdNames[12882] = 'ServerType_EstimatedReturnTime'
ObjectIdNames[12883] = 'ServerType_RequestServerStateChange'
ObjectIdNames[12884] = 'ServerType_RequestServerStateChange_InputArguments'
ObjectIdNames[12885] = 'Server_EstimatedReturnTime'
ObjectIdNames[12886] = 'Server_RequestServerStateChange'
ObjectIdNames[12887] = 'Server_RequestServerStateChange_InputArguments'
ObjectIdNames[12888] = 'RequestServerStateChangeMethodType'
ObjectIdNames[12889] = 'RequestServerStateChangeMethodType_InputArguments'
ObjectIdNames[12890] = 'DiscoveryConfiguration'
ObjectIdNames[12891] = 'MdnsDiscoveryConfiguration'
ObjectIdNames[12892] = 'DiscoveryConfiguration_Encoding_DefaultXml'
ObjectIdNames[12893] = 'MdnsDiscoveryConfiguration_Encoding_DefaultXml'
ObjectIdNames[12894] = 'OpcUa_XmlSchema_DiscoveryConfiguration'
ObjectIdNames[12895] = 'OpcUa_XmlSchema_DiscoveryConfiguration_DataTypeVersion'
ObjectIdNames[12896] = 'OpcUa_XmlSchema_DiscoveryConfiguration_DictionaryFragment'
ObjectIdNames[12897] = 'OpcUa_XmlSchema_MdnsDiscoveryConfiguration'
ObjectIdNames[12898] = 'OpcUa_XmlSchema_MdnsDiscoveryConfiguration_DataTypeVersion'
ObjectIdNames[12899] = 'OpcUa_XmlSchema_MdnsDiscoveryConfiguration_DictionaryFragment'
ObjectIdNames[12900] = 'DiscoveryConfiguration_Encoding_DefaultBinary'
ObjectIdNames[12901] = 'MdnsDiscoveryConfiguration_Encoding_DefaultBinary'
ObjectIdNames[12902] = 'OpcUa_BinarySchema_DiscoveryConfiguration'
ObjectIdNames[12903] = 'OpcUa_BinarySchema_DiscoveryConfiguration_DataTypeVersion'
ObjectIdNames[12904] = 'OpcUa_BinarySchema_DiscoveryConfiguration_DictionaryFragment'
ObjectIdNames[12905] = 'OpcUa_BinarySchema_MdnsDiscoveryConfiguration'
ObjectIdNames[12906] = 'OpcUa_BinarySchema_MdnsDiscoveryConfiguration_DataTypeVersion'
ObjectIdNames[12907] = 'OpcUa_BinarySchema_MdnsDiscoveryConfiguration_DictionaryFragment'
ObjectIdNames[12908] = 'MaxByteStringLength'
ObjectIdNames[12909] = 'ServerType_ServerCapabilities_MaxByteStringLength'
ObjectIdNames[12910] = 'ServerCapabilitiesType_MaxByteStringLength'
ObjectIdNames[12911] = 'Server_ServerCapabilities_MaxByteStringLength'
ObjectIdNames[12912] = 'ConditionType_ConditionRefresh2'
ObjectIdNames[12913] = 'ConditionType_ConditionRefresh2_InputArguments'
ObjectIdNames[12914] = 'ConditionRefresh2MethodType'
ObjectIdNames[12915] = 'ConditionRefresh2MethodType_InputArguments'
ObjectIdNames[12916] = 'DialogConditionType_ConditionRefresh2'
ObjectIdNames[12917] = 'DialogConditionType_ConditionRefresh2_InputArguments'
ObjectIdNames[12918] = 'AcknowledgeableConditionType_ConditionRefresh2'
ObjectIdNames[12919] = 'AcknowledgeableConditionType_ConditionRefresh2_InputArguments'
ObjectIdNames[12984] = 'AlarmConditionType_ConditionRefresh2'
ObjectIdNames[12985] = 'AlarmConditionType_ConditionRefresh2_InputArguments'
ObjectIdNames[12986] = 'LimitAlarmType_ConditionRefresh2'
ObjectIdNames[12987] = 'LimitAlarmType_ConditionRefresh2_InputArguments'
ObjectIdNames[12988] = 'ExclusiveLimitAlarmType_ConditionRefresh2'
ObjectIdNames[12989] = 'ExclusiveLimitAlarmType_ConditionRefresh2_InputArguments'
ObjectIdNames[12990] = 'NonExclusiveLimitAlarmType_ConditionRefresh2'
ObjectIdNames[12991] = 'NonExclusiveLimitAlarmType_ConditionRefresh2_InputArguments'
ObjectIdNames[12992] = 'NonExclusiveLevelAlarmType_ConditionRefresh2'
ObjectIdNames[12993] = 'NonExclusiveLevelAlarmType_ConditionRefresh2_InputArguments'
ObjectIdNames[12994] = 'ExclusiveLevelAlarmType_ConditionRefresh2'
ObjectIdNames[12995] = 'ExclusiveLevelAlarmType_ConditionRefresh2_InputArguments'
ObjectIdNames[12996] = 'NonExclusiveDeviationAlarmType_ConditionRefresh2'
ObjectIdNames[12997] = 'NonExclusiveDeviationAlarmType_ConditionRefresh2_InputArguments'
ObjectIdNames[12998] = 'ExclusiveDeviationAlarmType_ConditionRefresh2'
ObjectIdNames[12999] = 'ExclusiveDeviationAlarmType_ConditionRefresh2_InputArguments'
ObjectIdNames[13000] = 'NonExclusiveRateOfChangeAlarmType_ConditionRefresh2'
ObjectIdNames[13001] = 'NonExclusiveRateOfChangeAlarmType_ConditionRefresh2_InputArguments'
ObjectIdNames[13002] = 'ExclusiveRateOfChangeAlarmType_ConditionRefresh2'
ObjectIdNames[13003] = 'ExclusiveRateOfChangeAlarmType_ConditionRefresh2_InputArguments'
ObjectIdNames[13004] = 'DiscreteAlarmType_ConditionRefresh2'
ObjectIdNames[13005] = 'DiscreteAlarmType_ConditionRefresh2_InputArguments'
ObjectIdNames[13006] = 'OffNormalAlarmType_ConditionRefresh2'
ObjectIdNames[13007] = 'OffNormalAlarmType_ConditionRefresh2_InputArguments'
ObjectIdNames[13008] = 'SystemOffNormalAlarmType_ConditionRefresh2'
ObjectIdNames[13009] = 'SystemOffNormalAlarmType_ConditionRefresh2_InputArguments'
ObjectIdNames[13010] = 'TripAlarmType_ConditionRefresh2'
ObjectIdNames[13011] = 'TripAlarmType_ConditionRefresh2_InputArguments'
ObjectIdNames[13225] = 'CertificateExpirationAlarmType'
ObjectIdNames[13226] = 'CertificateExpirationAlarmType_EventId'
ObjectIdNames[13227] = 'CertificateExpirationAlarmType_EventType'
ObjectIdNames[13228] = 'CertificateExpirationAlarmType_SourceNode'
ObjectIdNames[13229] = 'CertificateExpirationAlarmType_SourceName'
ObjectIdNames[13230] = 'CertificateExpirationAlarmType_Time'
ObjectIdNames[13231] = 'CertificateExpirationAlarmType_ReceiveTime'
ObjectIdNames[13232] = 'CertificateExpirationAlarmType_LocalTime'
ObjectIdNames[13233] = 'CertificateExpirationAlarmType_Message'
ObjectIdNames[13234] = 'CertificateExpirationAlarmType_Severity'
ObjectIdNames[13235] = 'CertificateExpirationAlarmType_ConditionClassId'
ObjectIdNames[13236] = 'CertificateExpirationAlarmType_ConditionClassName'
ObjectIdNames[13237] = 'CertificateExpirationAlarmType_ConditionName'
ObjectIdNames[13238] = 'CertificateExpirationAlarmType_BranchId'
ObjectIdNames[13239] = 'CertificateExpirationAlarmType_Retain'
ObjectIdNames[13240] = 'CertificateExpirationAlarmType_EnabledState'
ObjectIdNames[13241] = 'CertificateExpirationAlarmType_EnabledState_Id'
ObjectIdNames[13242] = 'CertificateExpirationAlarmType_EnabledState_Name'
ObjectIdNames[13243] = 'CertificateExpirationAlarmType_EnabledState_Number'
ObjectIdNames[13244] = 'CertificateExpirationAlarmType_EnabledState_EffectiveDisplayName'
ObjectIdNames[13245] = 'CertificateExpirationAlarmType_EnabledState_TransitionTime'
ObjectIdNames[13246] = 'CertificateExpirationAlarmType_EnabledState_EffectiveTransitionTime'
ObjectIdNames[13247] = 'CertificateExpirationAlarmType_EnabledState_TrueState'
ObjectIdNames[13248] = 'CertificateExpirationAlarmType_EnabledState_FalseState'
ObjectIdNames[13249] = 'CertificateExpirationAlarmType_Quality'
ObjectIdNames[13250] = 'CertificateExpirationAlarmType_Quality_SourceTimestamp'
ObjectIdNames[13251] = 'CertificateExpirationAlarmType_LastSeverity'
ObjectIdNames[13252] = 'CertificateExpirationAlarmType_LastSeverity_SourceTimestamp'
ObjectIdNames[13253] = 'CertificateExpirationAlarmType_Comment'
ObjectIdNames[13254] = 'CertificateExpirationAlarmType_Comment_SourceTimestamp'
ObjectIdNames[13255] = 'CertificateExpirationAlarmType_ClientUserId'
ObjectIdNames[13256] = 'CertificateExpirationAlarmType_Disable'
ObjectIdNames[13257] = 'CertificateExpirationAlarmType_Enable'
ObjectIdNames[13258] = 'CertificateExpirationAlarmType_AddComment'
ObjectIdNames[13259] = 'CertificateExpirationAlarmType_AddComment_InputArguments'
ObjectIdNames[13260] = 'CertificateExpirationAlarmType_ConditionRefresh'
ObjectIdNames[13261] = 'CertificateExpirationAlarmType_ConditionRefresh_InputArguments'
ObjectIdNames[13262] = 'CertificateExpirationAlarmType_ConditionRefresh2'
ObjectIdNames[13263] = 'CertificateExpirationAlarmType_ConditionRefresh2_InputArguments'
ObjectIdNames[13264] = 'CertificateExpirationAlarmType_AckedState'
ObjectIdNames[13265] = 'CertificateExpirationAlarmType_AckedState_Id'
ObjectIdNames[13266] = 'CertificateExpirationAlarmType_AckedState_Name'
ObjectIdNames[13267] = 'CertificateExpirationAlarmType_AckedState_Number'
ObjectIdNames[13268] = 'CertificateExpirationAlarmType_AckedState_EffectiveDisplayName'
ObjectIdNames[13269] = 'CertificateExpirationAlarmType_AckedState_TransitionTime'
ObjectIdNames[13270] = 'CertificateExpirationAlarmType_AckedState_EffectiveTransitionTime'
ObjectIdNames[13271] = 'CertificateExpirationAlarmType_AckedState_TrueState'
ObjectIdNames[13272] = 'CertificateExpirationAlarmType_AckedState_FalseState'
ObjectIdNames[13273] = 'CertificateExpirationAlarmType_ConfirmedState'
ObjectIdNames[13274] = 'CertificateExpirationAlarmType_ConfirmedState_Id'
ObjectIdNames[13275] = 'CertificateExpirationAlarmType_ConfirmedState_Name'
ObjectIdNames[13276] = 'CertificateExpirationAlarmType_ConfirmedState_Number'
ObjectIdNames[13277] = 'CertificateExpirationAlarmType_ConfirmedState_EffectiveDisplayName'
ObjectIdNames[13278] = 'CertificateExpirationAlarmType_ConfirmedState_TransitionTime'
ObjectIdNames[13279] = 'CertificateExpirationAlarmType_ConfirmedState_EffectiveTransitionTime'
ObjectIdNames[13280] = 'CertificateExpirationAlarmType_ConfirmedState_TrueState'
ObjectIdNames[13281] = 'CertificateExpirationAlarmType_ConfirmedState_FalseState'
ObjectIdNames[13282] = 'CertificateExpirationAlarmType_Acknowledge'
ObjectIdNames[13283] = 'CertificateExpirationAlarmType_Acknowledge_InputArguments'
ObjectIdNames[13284] = 'CertificateExpirationAlarmType_Confirm'
ObjectIdNames[13285] = 'CertificateExpirationAlarmType_Confirm_InputArguments'
ObjectIdNames[13286] = 'CertificateExpirationAlarmType_ActiveState'
ObjectIdNames[13287] = 'CertificateExpirationAlarmType_ActiveState_Id'
ObjectIdNames[13288] = 'CertificateExpirationAlarmType_ActiveState_Name'
ObjectIdNames[13289] = 'CertificateExpirationAlarmType_ActiveState_Number'
ObjectIdNames[13290] = 'CertificateExpirationAlarmType_ActiveState_EffectiveDisplayName'
ObjectIdNames[13291] = 'CertificateExpirationAlarmType_ActiveState_TransitionTime'
ObjectIdNames[13292] = 'CertificateExpirationAlarmType_ActiveState_EffectiveTransitionTime'
ObjectIdNames[13293] = 'CertificateExpirationAlarmType_ActiveState_TrueState'
ObjectIdNames[13294] = 'CertificateExpirationAlarmType_ActiveState_FalseState'
ObjectIdNames[13295] = 'CertificateExpirationAlarmType_InputNode'
ObjectIdNames[13296] = 'CertificateExpirationAlarmType_SuppressedState'
ObjectIdNames[13297] = 'CertificateExpirationAlarmType_SuppressedState_Id'
ObjectIdNames[13298] = 'CertificateExpirationAlarmType_SuppressedState_Name'
ObjectIdNames[13299] = 'CertificateExpirationAlarmType_SuppressedState_Number'
ObjectIdNames[13300] = 'CertificateExpirationAlarmType_SuppressedState_EffectiveDisplayName'
ObjectIdNames[13301] = 'CertificateExpirationAlarmType_SuppressedState_TransitionTime'
ObjectIdNames[13302] = 'CertificateExpirationAlarmType_SuppressedState_EffectiveTransitionTime'
ObjectIdNames[13303] = 'CertificateExpirationAlarmType_SuppressedState_TrueState'
ObjectIdNames[13304] = 'CertificateExpirationAlarmType_SuppressedState_FalseState'
ObjectIdNames[13305] = 'CertificateExpirationAlarmType_ShelvingState'
ObjectIdNames[13306] = 'CertificateExpirationAlarmType_ShelvingState_CurrentState'
ObjectIdNames[13307] = 'CertificateExpirationAlarmType_ShelvingState_CurrentState_Id'
ObjectIdNames[13308] = 'CertificateExpirationAlarmType_ShelvingState_CurrentState_Name'
ObjectIdNames[13309] = 'CertificateExpirationAlarmType_ShelvingState_CurrentState_Number'
ObjectIdNames[13310] = 'CertificateExpirationAlarmType_ShelvingState_CurrentState_EffectiveDisplayName'
ObjectIdNames[13311] = 'CertificateExpirationAlarmType_ShelvingState_LastTransition'
ObjectIdNames[13312] = 'CertificateExpirationAlarmType_ShelvingState_LastTransition_Id'
ObjectIdNames[13313] = 'CertificateExpirationAlarmType_ShelvingState_LastTransition_Name'
ObjectIdNames[13314] = 'CertificateExpirationAlarmType_ShelvingState_LastTransition_Number'
ObjectIdNames[13315] = 'CertificateExpirationAlarmType_ShelvingState_LastTransition_TransitionTime'
ObjectIdNames[13316] = 'CertificateExpirationAlarmType_ShelvingState_LastTransition_EffectiveTransitionTime'
ObjectIdNames[13317] = 'CertificateExpirationAlarmType_ShelvingState_UnshelveTime'
ObjectIdNames[13318] = 'CertificateExpirationAlarmType_ShelvingState_Unshelve'
ObjectIdNames[13319] = 'CertificateExpirationAlarmType_ShelvingState_OneShotShelve'
ObjectIdNames[13320] = 'CertificateExpirationAlarmType_ShelvingState_TimedShelve'
ObjectIdNames[13321] = 'CertificateExpirationAlarmType_ShelvingState_TimedShelve_InputArguments'
ObjectIdNames[13322] = 'CertificateExpirationAlarmType_SuppressedOrShelved'
ObjectIdNames[13323] = 'CertificateExpirationAlarmType_MaxTimeShelved'
ObjectIdNames[13324] = 'CertificateExpirationAlarmType_NormalState'
ObjectIdNames[13325] = 'CertificateExpirationAlarmType_ExpirationDate'
ObjectIdNames[13326] = 'CertificateExpirationAlarmType_CertificateType'
ObjectIdNames[13327] = 'CertificateExpirationAlarmType_Certificate'
ObjectIdNames[13340] = 'ServerType_Namespaces_AddressSpaceFile_MimeType'
ObjectIdNames[13341] = 'FileType_MimeType'
ObjectIdNames[13342] = 'CreateDirectoryMethodType'
ObjectIdNames[13343] = 'CreateDirectoryMethodType_InputArguments'
ObjectIdNames[13344] = 'CreateDirectoryMethodType_OutputArguments'
ObjectIdNames[13345] = 'CreateFileMethodType'
ObjectIdNames[13346] = 'CreateFileMethodType_InputArguments'
ObjectIdNames[13347] = 'CreateFileMethodType_OutputArguments'
ObjectIdNames[13348] = 'DeleteFileMethodType'
ObjectIdNames[13349] = 'DeleteFileMethodType_InputArguments'
ObjectIdNames[13350] = 'MoveOrCopyMethodType'
ObjectIdNames[13351] = 'MoveOrCopyMethodType_InputArguments'
ObjectIdNames[13352] = 'MoveOrCopyMethodType_OutputArguments'
ObjectIdNames[13353] = 'FileDirectoryType'
ObjectIdNames[13354] = 'FileDirectoryType_xFileDirectoryNamex'
ObjectIdNames[13355] = 'FileDirectoryType_xFileDirectoryNamex_CreateDirectory'
ObjectIdNames[13356] = 'FileDirectoryType_xFileDirectoryNamex_CreateDirectory_InputArguments'
ObjectIdNames[13357] = 'FileDirectoryType_xFileDirectoryNamex_CreateDirectory_OutputArguments'
ObjectIdNames[13358] = 'FileDirectoryType_xFileDirectoryNamex_CreateFile'
ObjectIdNames[13359] = 'FileDirectoryType_xFileDirectoryNamex_CreateFile_InputArguments'
ObjectIdNames[13360] = 'FileDirectoryType_xFileDirectoryNamex_CreateFile_OutputArguments'
ObjectIdNames[13361] = 'FileDirectoryType_xFileDirectoryNamex_Delete'
ObjectIdNames[13362] = 'FileDirectoryType_xFileDirectoryNamex_Delete_InputArguments'
ObjectIdNames[13363] = 'FileDirectoryType_xFileDirectoryNamex_MoveOrCopy'
ObjectIdNames[13364] = 'FileDirectoryType_xFileDirectoryNamex_MoveOrCopy_InputArguments'
ObjectIdNames[13365] = 'FileDirectoryType_xFileDirectoryNamex_MoveOrCopy_OutputArguments'
ObjectIdNames[13366] = 'FileDirectoryType_xFileNamex'
ObjectIdNames[13367] = 'FileDirectoryType_xFileNamex_Size'
ObjectIdNames[13368] = 'FileDirectoryType_xFileNamex_Writable'
ObjectIdNames[13369] = 'FileDirectoryType_xFileNamex_UserWritable'
ObjectIdNames[13370] = 'FileDirectoryType_xFileNamex_OpenCount'
ObjectIdNames[13371] = 'FileDirectoryType_xFileNamex_MimeType'
ObjectIdNames[13372] = 'FileDirectoryType_xFileNamex_Open'
ObjectIdNames[13373] = 'FileDirectoryType_xFileNamex_Open_InputArguments'
ObjectIdNames[13374] = 'FileDirectoryType_xFileNamex_Open_OutputArguments'
ObjectIdNames[13375] = 'FileDirectoryType_xFileNamex_Close'
ObjectIdNames[13376] = 'FileDirectoryType_xFileNamex_Close_InputArguments'
ObjectIdNames[13377] = 'FileDirectoryType_xFileNamex_Read'
ObjectIdNames[13378] = 'FileDirectoryType_xFileNamex_Read_InputArguments'
ObjectIdNames[13379] = 'FileDirectoryType_xFileNamex_Read_OutputArguments'
ObjectIdNames[13380] = 'FileDirectoryType_xFileNamex_Write'
ObjectIdNames[13381] = 'FileDirectoryType_xFileNamex_Write_InputArguments'
ObjectIdNames[13382] = 'FileDirectoryType_xFileNamex_GetPosition'
ObjectIdNames[13383] = 'FileDirectoryType_xFileNamex_GetPosition_InputArguments'
ObjectIdNames[13384] = 'FileDirectoryType_xFileNamex_GetPosition_OutputArguments'
ObjectIdNames[13385] = 'FileDirectoryType_xFileNamex_SetPosition'
ObjectIdNames[13386] = 'FileDirectoryType_xFileNamex_SetPosition_InputArguments'
ObjectIdNames[13387] = 'FileDirectoryType_CreateDirectory'
ObjectIdNames[13388] = 'FileDirectoryType_CreateDirectory_InputArguments'
ObjectIdNames[13389] = 'FileDirectoryType_CreateDirectory_OutputArguments'
ObjectIdNames[13390] = 'FileDirectoryType_CreateFile'
ObjectIdNames[13391] = 'FileDirectoryType_CreateFile_InputArguments'
ObjectIdNames[13392] = 'FileDirectoryType_CreateFile_OutputArguments'
ObjectIdNames[13393] = 'FileDirectoryType_Delete'
ObjectIdNames[13394] = 'FileDirectoryType_Delete_InputArguments'
ObjectIdNames[13395] = 'FileDirectoryType_MoveOrCopy'
ObjectIdNames[13396] = 'FileDirectoryType_MoveOrCopy_InputArguments'
ObjectIdNames[13397] = 'FileDirectoryType_MoveOrCopy_OutputArguments'
ObjectIdNames[13398] = 'AddressSpaceFileType_MimeType'
ObjectIdNames[13399] = 'NamespaceMetadataType_NamespaceFile_MimeType'
ObjectIdNames[13400] = 'NamespacesType_NamespaceIdentifier_NamespaceFile_MimeType'
ObjectIdNames[13401] = 'NamespacesType_AddressSpaceFile_MimeType'
ObjectIdNames[13402] = 'Server_Namespaces_AddressSpaceFile_MimeType'
ObjectIdNames[13403] = 'TrustListType_MimeType'
ObjectIdNames[13599] = 'CertificateGroupType_TrustList'
ObjectIdNames[13600] = 'CertificateGroupType_TrustList_Size'
ObjectIdNames[13601] = 'CertificateGroupType_TrustList_Writable'
ObjectIdNames[13602] = 'CertificateGroupType_TrustList_UserWritable'
ObjectIdNames[13603] = 'CertificateGroupType_TrustList_OpenCount'
ObjectIdNames[13604] = 'CertificateGroupType_TrustList_MimeType'
ObjectIdNames[13605] = 'CertificateGroupType_TrustList_Open'
ObjectIdNames[13606] = 'CertificateGroupType_TrustList_Open_InputArguments'
ObjectIdNames[13607] = 'CertificateGroupType_TrustList_Open_OutputArguments'
ObjectIdNames[13608] = 'CertificateGroupType_TrustList_Close'
ObjectIdNames[13609] = 'CertificateGroupType_TrustList_Close_InputArguments'
ObjectIdNames[13610] = 'CertificateGroupType_TrustList_Read'
ObjectIdNames[13611] = 'CertificateGroupType_TrustList_Read_InputArguments'
ObjectIdNames[13612] = 'CertificateGroupType_TrustList_Read_OutputArguments'
ObjectIdNames[13613] = 'CertificateGroupType_TrustList_Write'
ObjectIdNames[13614] = 'CertificateGroupType_TrustList_Write_InputArguments'
ObjectIdNames[13615] = 'CertificateGroupType_TrustList_GetPosition'
ObjectIdNames[13616] = 'CertificateGroupType_TrustList_GetPosition_InputArguments'
ObjectIdNames[13617] = 'CertificateGroupType_TrustList_GetPosition_OutputArguments'
ObjectIdNames[13618] = 'CertificateGroupType_TrustList_SetPosition'
ObjectIdNames[13619] = 'CertificateGroupType_TrustList_SetPosition_InputArguments'
ObjectIdNames[13620] = 'CertificateGroupType_TrustList_LastUpdateTime'
ObjectIdNames[13621] = 'CertificateGroupType_TrustList_OpenWithMasks'
ObjectIdNames[13622] = 'CertificateGroupType_TrustList_OpenWithMasks_InputArguments'
ObjectIdNames[13623] = 'CertificateGroupType_TrustList_OpenWithMasks_OutputArguments'
ObjectIdNames[13624] = 'CertificateGroupType_TrustList_CloseAndUpdate'
ObjectIdNames[13625] = 'CertificateGroupType_TrustList_CloseAndUpdate_InputArguments'
ObjectIdNames[13626] = 'CertificateGroupType_TrustList_CloseAndUpdate_OutputArguments'
ObjectIdNames[13627] = 'CertificateGroupType_TrustList_AddCertificate'
ObjectIdNames[13628] = 'CertificateGroupType_TrustList_AddCertificate_InputArguments'
ObjectIdNames[13629] = 'CertificateGroupType_TrustList_RemoveCertificate'
ObjectIdNames[13630] = 'CertificateGroupType_TrustList_RemoveCertificate_InputArguments'
ObjectIdNames[13631] = 'CertificateGroupType_CertificateTypes'
ObjectIdNames[13735] = 'CertificateUpdatedAuditEventType_CertificateGroup'
ObjectIdNames[13736] = 'CertificateUpdatedAuditEventType_CertificateType'
ObjectIdNames[13737] = 'ServerConfiguration_UpdateCertificate'
ObjectIdNames[13738] = 'ServerConfiguration_UpdateCertificate_InputArguments'
ObjectIdNames[13739] = 'ServerConfiguration_UpdateCertificate_OutputArguments'
ObjectIdNames[13813] = 'CertificateGroupFolderType'
ObjectIdNames[13814] = 'CertificateGroupFolderType_DefaultApplicationGroup'
ObjectIdNames[13815] = 'CertificateGroupFolderType_DefaultApplicationGroup_TrustList'
ObjectIdNames[13816] = 'CertificateGroupFolderType_DefaultApplicationGroup_TrustList_Size'
ObjectIdNames[13817] = 'CertificateGroupFolderType_DefaultApplicationGroup_TrustList_Writable'
ObjectIdNames[13818] = 'CertificateGroupFolderType_DefaultApplicationGroup_TrustList_UserWritable'
ObjectIdNames[13819] = 'CertificateGroupFolderType_DefaultApplicationGroup_TrustList_OpenCount'
ObjectIdNames[13820] = 'CertificateGroupFolderType_DefaultApplicationGroup_TrustList_MimeType'
ObjectIdNames[13821] = 'CertificateGroupFolderType_DefaultApplicationGroup_TrustList_Open'
ObjectIdNames[13822] = 'CertificateGroupFolderType_DefaultApplicationGroup_TrustList_Open_InputArguments'
ObjectIdNames[13823] = 'CertificateGroupFolderType_DefaultApplicationGroup_TrustList_Open_OutputArguments'
ObjectIdNames[13824] = 'CertificateGroupFolderType_DefaultApplicationGroup_TrustList_Close'
ObjectIdNames[13825] = 'CertificateGroupFolderType_DefaultApplicationGroup_TrustList_Close_InputArguments'
ObjectIdNames[13826] = 'CertificateGroupFolderType_DefaultApplicationGroup_TrustList_Read'
ObjectIdNames[13827] = 'CertificateGroupFolderType_DefaultApplicationGroup_TrustList_Read_InputArguments'
ObjectIdNames[13828] = 'CertificateGroupFolderType_DefaultApplicationGroup_TrustList_Read_OutputArguments'
ObjectIdNames[13829] = 'CertificateGroupFolderType_DefaultApplicationGroup_TrustList_Write'
ObjectIdNames[13830] = 'CertificateGroupFolderType_DefaultApplicationGroup_TrustList_Write_InputArguments'
ObjectIdNames[13831] = 'CertificateGroupFolderType_DefaultApplicationGroup_TrustList_GetPosition'
ObjectIdNames[13832] = 'CertificateGroupFolderType_DefaultApplicationGroup_TrustList_GetPosition_InputArguments'
ObjectIdNames[13833] = 'CertificateGroupFolderType_DefaultApplicationGroup_TrustList_GetPosition_OutputArguments'
ObjectIdNames[13834] = 'CertificateGroupFolderType_DefaultApplicationGroup_TrustList_SetPosition'
ObjectIdNames[13835] = 'CertificateGroupFolderType_DefaultApplicationGroup_TrustList_SetPosition_InputArguments'
ObjectIdNames[13836] = 'CertificateGroupFolderType_DefaultApplicationGroup_TrustList_LastUpdateTime'
ObjectIdNames[13837] = 'CertificateGroupFolderType_DefaultApplicationGroup_TrustList_OpenWithMasks'
ObjectIdNames[13838] = 'CertificateGroupFolderType_DefaultApplicationGroup_TrustList_OpenWithMasks_InputArguments'
ObjectIdNames[13839] = 'CertificateGroupFolderType_DefaultApplicationGroup_TrustList_OpenWithMasks_OutputArguments'
ObjectIdNames[13840] = 'CertificateGroupFolderType_DefaultApplicationGroup_TrustList_CloseAndUpdate'
ObjectIdNames[13841] = 'CertificateGroupFolderType_DefaultApplicationGroup_TrustList_CloseAndUpdate_InputArguments'
ObjectIdNames[13842] = 'CertificateGroupFolderType_DefaultApplicationGroup_TrustList_CloseAndUpdate_OutputArguments'
ObjectIdNames[13843] = 'CertificateGroupFolderType_DefaultApplicationGroup_TrustList_AddCertificate'
ObjectIdNames[13844] = 'CertificateGroupFolderType_DefaultApplicationGroup_TrustList_AddCertificate_InputArguments'
ObjectIdNames[13845] = 'CertificateGroupFolderType_DefaultApplicationGroup_TrustList_RemoveCertificate'
ObjectIdNames[13846] = 'CertificateGroupFolderType_DefaultApplicationGroup_TrustList_RemoveCertificate_InputArguments'
ObjectIdNames[13847] = 'CertificateGroupFolderType_DefaultApplicationGroup_CertificateTypes'
ObjectIdNames[13848] = 'CertificateGroupFolderType_DefaultHttpsGroup'
ObjectIdNames[13849] = 'CertificateGroupFolderType_DefaultHttpsGroup_TrustList'
ObjectIdNames[13850] = 'CertificateGroupFolderType_DefaultHttpsGroup_TrustList_Size'
ObjectIdNames[13851] = 'CertificateGroupFolderType_DefaultHttpsGroup_TrustList_Writable'
ObjectIdNames[13852] = 'CertificateGroupFolderType_DefaultHttpsGroup_TrustList_UserWritable'
ObjectIdNames[13853] = 'CertificateGroupFolderType_DefaultHttpsGroup_TrustList_OpenCount'
ObjectIdNames[13854] = 'CertificateGroupFolderType_DefaultHttpsGroup_TrustList_MimeType'
ObjectIdNames[13855] = 'CertificateGroupFolderType_DefaultHttpsGroup_TrustList_Open'
ObjectIdNames[13856] = 'CertificateGroupFolderType_DefaultHttpsGroup_TrustList_Open_InputArguments'
ObjectIdNames[13857] = 'CertificateGroupFolderType_DefaultHttpsGroup_TrustList_Open_OutputArguments'
ObjectIdNames[13858] = 'CertificateGroupFolderType_DefaultHttpsGroup_TrustList_Close'
ObjectIdNames[13859] = 'CertificateGroupFolderType_DefaultHttpsGroup_TrustList_Close_InputArguments'
ObjectIdNames[13860] = 'CertificateGroupFolderType_DefaultHttpsGroup_TrustList_Read'
ObjectIdNames[13861] = 'CertificateGroupFolderType_DefaultHttpsGroup_TrustList_Read_InputArguments'
ObjectIdNames[13862] = 'CertificateGroupFolderType_DefaultHttpsGroup_TrustList_Read_OutputArguments'
ObjectIdNames[13863] = 'CertificateGroupFolderType_DefaultHttpsGroup_TrustList_Write'
ObjectIdNames[13864] = 'CertificateGroupFolderType_DefaultHttpsGroup_TrustList_Write_InputArguments'
ObjectIdNames[13865] = 'CertificateGroupFolderType_DefaultHttpsGroup_TrustList_GetPosition'
ObjectIdNames[13866] = 'CertificateGroupFolderType_DefaultHttpsGroup_TrustList_GetPosition_InputArguments'
ObjectIdNames[13867] = 'CertificateGroupFolderType_DefaultHttpsGroup_TrustList_GetPosition_OutputArguments'
ObjectIdNames[13868] = 'CertificateGroupFolderType_DefaultHttpsGroup_TrustList_SetPosition'
ObjectIdNames[13869] = 'CertificateGroupFolderType_DefaultHttpsGroup_TrustList_SetPosition_InputArguments'
ObjectIdNames[13870] = 'CertificateGroupFolderType_DefaultHttpsGroup_TrustList_LastUpdateTime'
ObjectIdNames[13871] = 'CertificateGroupFolderType_DefaultHttpsGroup_TrustList_OpenWithMasks'
ObjectIdNames[13872] = 'CertificateGroupFolderType_DefaultHttpsGroup_TrustList_OpenWithMasks_InputArguments'
ObjectIdNames[13873] = 'CertificateGroupFolderType_DefaultHttpsGroup_TrustList_OpenWithMasks_OutputArguments'
ObjectIdNames[13874] = 'CertificateGroupFolderType_DefaultHttpsGroup_TrustList_CloseAndUpdate'
ObjectIdNames[13875] = 'CertificateGroupFolderType_DefaultHttpsGroup_TrustList_CloseAndUpdate_InputArguments'
ObjectIdNames[13876] = 'CertificateGroupFolderType_DefaultHttpsGroup_TrustList_CloseAndUpdate_OutputArguments'
ObjectIdNames[13877] = 'CertificateGroupFolderType_DefaultHttpsGroup_TrustList_AddCertificate'
ObjectIdNames[13878] = 'CertificateGroupFolderType_DefaultHttpsGroup_TrustList_AddCertificate_InputArguments'
ObjectIdNames[13879] = 'CertificateGroupFolderType_DefaultHttpsGroup_TrustList_RemoveCertificate'
ObjectIdNames[13880] = 'CertificateGroupFolderType_DefaultHttpsGroup_TrustList_RemoveCertificate_InputArguments'
ObjectIdNames[13881] = 'CertificateGroupFolderType_DefaultHttpsGroup_CertificateTypes'
ObjectIdNames[13882] = 'CertificateGroupFolderType_DefaultUserTokenGroup'
ObjectIdNames[13883] = 'CertificateGroupFolderType_DefaultUserTokenGroup_TrustList'
ObjectIdNames[13884] = 'CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_Size'
ObjectIdNames[13885] = 'CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_Writable'
ObjectIdNames[13886] = 'CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_UserWritable'
ObjectIdNames[13887] = 'CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_OpenCount'
ObjectIdNames[13888] = 'CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_MimeType'
ObjectIdNames[13889] = 'CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_Open'
ObjectIdNames[13890] = 'CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_Open_InputArguments'
ObjectIdNames[13891] = 'CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_Open_OutputArguments'
ObjectIdNames[13892] = 'CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_Close'
ObjectIdNames[13893] = 'CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_Close_InputArguments'
ObjectIdNames[13894] = 'CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_Read'
ObjectIdNames[13895] = 'CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_Read_InputArguments'
ObjectIdNames[13896] = 'CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_Read_OutputArguments'
ObjectIdNames[13897] = 'CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_Write'
ObjectIdNames[13898] = 'CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_Write_InputArguments'
ObjectIdNames[13899] = 'CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_GetPosition'
ObjectIdNames[13900] = 'CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_GetPosition_InputArguments'
ObjectIdNames[13901] = 'CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_GetPosition_OutputArguments'
ObjectIdNames[13902] = 'CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_SetPosition'
ObjectIdNames[13903] = 'CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_SetPosition_InputArguments'
ObjectIdNames[13904] = 'CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_LastUpdateTime'
ObjectIdNames[13905] = 'CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_OpenWithMasks'
ObjectIdNames[13906] = 'CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_OpenWithMasks_InputArguments'
ObjectIdNames[13907] = 'CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_OpenWithMasks_OutputArguments'
ObjectIdNames[13908] = 'CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_CloseAndUpdate'
ObjectIdNames[13909] = 'CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_CloseAndUpdate_InputArguments'
ObjectIdNames[13910] = 'CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_CloseAndUpdate_OutputArguments'
ObjectIdNames[13911] = 'CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_AddCertificate'
ObjectIdNames[13912] = 'CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_AddCertificate_InputArguments'
ObjectIdNames[13913] = 'CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_RemoveCertificate'
ObjectIdNames[13914] = 'CertificateGroupFolderType_DefaultUserTokenGroup_TrustList_RemoveCertificate_InputArguments'
ObjectIdNames[13915] = 'CertificateGroupFolderType_DefaultUserTokenGroup_CertificateTypes'
ObjectIdNames[13916] = 'CertificateGroupFolderType_xCertificateGroupx'
ObjectIdNames[13917] = 'CertificateGroupFolderType_xCertificateGroupx_TrustList'
ObjectIdNames[13918] = 'CertificateGroupFolderType_xCertificateGroupx_TrustList_Size'
ObjectIdNames[13919] = 'CertificateGroupFolderType_xCertificateGroupx_TrustList_Writable'
ObjectIdNames[13920] = 'CertificateGroupFolderType_xCertificateGroupx_TrustList_UserWritable'
ObjectIdNames[13921] = 'CertificateGroupFolderType_xCertificateGroupx_TrustList_OpenCount'
ObjectIdNames[13922] = 'CertificateGroupFolderType_xCertificateGroupx_TrustList_MimeType'
ObjectIdNames[13923] = 'CertificateGroupFolderType_xCertificateGroupx_TrustList_Open'
ObjectIdNames[13924] = 'CertificateGroupFolderType_xCertificateGroupx_TrustList_Open_InputArguments'
ObjectIdNames[13925] = 'CertificateGroupFolderType_xCertificateGroupx_TrustList_Open_OutputArguments'
ObjectIdNames[13926] = 'CertificateGroupFolderType_xCertificateGroupx_TrustList_Close'
ObjectIdNames[13927] = 'CertificateGroupFolderType_xCertificateGroupx_TrustList_Close_InputArguments'
ObjectIdNames[13928] = 'CertificateGroupFolderType_xCertificateGroupx_TrustList_Read'
ObjectIdNames[13929] = 'CertificateGroupFolderType_xCertificateGroupx_TrustList_Read_InputArguments'
ObjectIdNames[13930] = 'CertificateGroupFolderType_xCertificateGroupx_TrustList_Read_OutputArguments'
ObjectIdNames[13931] = 'CertificateGroupFolderType_xCertificateGroupx_TrustList_Write'
ObjectIdNames[13932] = 'CertificateGroupFolderType_xCertificateGroupx_TrustList_Write_InputArguments'
ObjectIdNames[13933] = 'CertificateGroupFolderType_xCertificateGroupx_TrustList_GetPosition'
ObjectIdNames[13934] = 'CertificateGroupFolderType_xCertificateGroupx_TrustList_GetPosition_InputArguments'
ObjectIdNames[13935] = 'CertificateGroupFolderType_xCertificateGroupx_TrustList_GetPosition_OutputArguments'
ObjectIdNames[13936] = 'CertificateGroupFolderType_xCertificateGroupx_TrustList_SetPosition'
ObjectIdNames[13937] = 'CertificateGroupFolderType_xCertificateGroupx_TrustList_SetPosition_InputArguments'
ObjectIdNames[13938] = 'CertificateGroupFolderType_xCertificateGroupx_TrustList_LastUpdateTime'
ObjectIdNames[13939] = 'CertificateGroupFolderType_xCertificateGroupx_TrustList_OpenWithMasks'
ObjectIdNames[13940] = 'CertificateGroupFolderType_xCertificateGroupx_TrustList_OpenWithMasks_InputArguments'
ObjectIdNames[13941] = 'CertificateGroupFolderType_xCertificateGroupx_TrustList_OpenWithMasks_OutputArguments'
ObjectIdNames[13942] = 'CertificateGroupFolderType_xCertificateGroupx_TrustList_CloseAndUpdate'
ObjectIdNames[13943] = 'CertificateGroupFolderType_xCertificateGroupx_TrustList_CloseAndUpdate_InputArguments'
ObjectIdNames[13944] = 'CertificateGroupFolderType_xCertificateGroupx_TrustList_CloseAndUpdate_OutputArguments'
ObjectIdNames[13945] = 'CertificateGroupFolderType_xCertificateGroupx_TrustList_AddCertificate'
ObjectIdNames[13946] = 'CertificateGroupFolderType_xCertificateGroupx_TrustList_AddCertificate_InputArguments'
ObjectIdNames[13947] = 'CertificateGroupFolderType_xCertificateGroupx_TrustList_RemoveCertificate'
ObjectIdNames[13948] = 'CertificateGroupFolderType_xCertificateGroupx_TrustList_RemoveCertificate_InputArguments'
ObjectIdNames[13949] = 'CertificateGroupFolderType_xCertificateGroupx_CertificateTypes'
ObjectIdNames[13950] = 'ServerConfigurationType_CertificateGroups'
ObjectIdNames[13951] = 'ServerConfigurationType_CertificateGroups_DefaultApplicationGroup'
ObjectIdNames[13952] = 'ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList'
ObjectIdNames[13953] = 'ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_Size'
ObjectIdNames[13954] = 'ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_Writable'
ObjectIdNames[13955] = 'ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_UserWritable'
ObjectIdNames[13956] = 'ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_OpenCount'
ObjectIdNames[13957] = 'ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_MimeType'
ObjectIdNames[13958] = 'ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_Open'
ObjectIdNames[13959] = 'ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_Open_InputArguments'
ObjectIdNames[13960] = 'ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_Open_OutputArguments'
ObjectIdNames[13961] = 'ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_Close'
ObjectIdNames[13962] = 'ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_Close_InputArguments'
ObjectIdNames[13963] = 'ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_Read'
ObjectIdNames[13964] = 'ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_Read_InputArguments'
ObjectIdNames[13965] = 'ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_Read_OutputArguments'
ObjectIdNames[13966] = 'ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_Write'
ObjectIdNames[13967] = 'ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_Write_InputArguments'
ObjectIdNames[13968] = 'ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_GetPosition'
ObjectIdNames[13969] = 'ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_GetPosition_InputArguments'
ObjectIdNames[13970] = 'ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_GetPosition_OutputArguments'
ObjectIdNames[13971] = 'ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_SetPosition'
ObjectIdNames[13972] = 'ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_SetPosition_InputArguments'
ObjectIdNames[13973] = 'ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_LastUpdateTime'
ObjectIdNames[13974] = 'ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_OpenWithMasks'
ObjectIdNames[13975] = 'ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_OpenWithMasks_InputArguments'
ObjectIdNames[13976] = 'ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_OpenWithMasks_OutputArguments'
ObjectIdNames[13977] = 'ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_CloseAndUpdate'
ObjectIdNames[13978] = 'ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_CloseAndUpdate_InputArguments'
ObjectIdNames[13979] = 'ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_CloseAndUpdate_OutputArguments'
ObjectIdNames[13980] = 'ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_AddCertificate'
ObjectIdNames[13981] = 'ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_AddCertificate_InputArguments'
ObjectIdNames[13982] = 'ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_RemoveCertificate'
ObjectIdNames[13983] = 'ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_TrustList_RemoveCertificate_InputArguments'
ObjectIdNames[13984] = 'ServerConfigurationType_CertificateGroups_DefaultApplicationGroup_CertificateTypes'
ObjectIdNames[13985] = 'ServerConfigurationType_CertificateGroups_DefaultHttpsGroup'
ObjectIdNames[13986] = 'ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList'
ObjectIdNames[13987] = 'ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_Size'
ObjectIdNames[13988] = 'ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_Writable'
ObjectIdNames[13989] = 'ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_UserWritable'
ObjectIdNames[13990] = 'ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_OpenCount'
ObjectIdNames[13991] = 'ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_MimeType'
ObjectIdNames[13992] = 'ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_Open'
ObjectIdNames[13993] = 'ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_Open_InputArguments'
ObjectIdNames[13994] = 'ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_Open_OutputArguments'
ObjectIdNames[13995] = 'ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_Close'
ObjectIdNames[13996] = 'ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_Close_InputArguments'
ObjectIdNames[13997] = 'ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_Read'
ObjectIdNames[13998] = 'ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_Read_InputArguments'
ObjectIdNames[13999] = 'ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_Read_OutputArguments'
ObjectIdNames[14000] = 'ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_Write'
ObjectIdNames[14001] = 'ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_Write_InputArguments'
ObjectIdNames[14002] = 'ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_GetPosition'
ObjectIdNames[14003] = 'ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_GetPosition_InputArguments'
ObjectIdNames[14004] = 'ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_GetPosition_OutputArguments'
ObjectIdNames[14005] = 'ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_SetPosition'
ObjectIdNames[14006] = 'ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_SetPosition_InputArguments'
ObjectIdNames[14007] = 'ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_LastUpdateTime'
ObjectIdNames[14008] = 'ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_OpenWithMasks'
ObjectIdNames[14009] = 'ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_OpenWithMasks_InputArguments'
ObjectIdNames[14010] = 'ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_OpenWithMasks_OutputArguments'
ObjectIdNames[14011] = 'ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_CloseAndUpdate'
ObjectIdNames[14012] = 'ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_CloseAndUpdate_InputArguments'
ObjectIdNames[14013] = 'ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_CloseAndUpdate_OutputArguments'
ObjectIdNames[14014] = 'ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_AddCertificate'
ObjectIdNames[14015] = 'ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_AddCertificate_InputArguments'
ObjectIdNames[14016] = 'ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_RemoveCertificate'
ObjectIdNames[14017] = 'ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_TrustList_RemoveCertificate_InputArguments'
ObjectIdNames[14018] = 'ServerConfigurationType_CertificateGroups_DefaultHttpsGroup_CertificateTypes'
ObjectIdNames[14019] = 'ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup'
ObjectIdNames[14020] = 'ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList'
ObjectIdNames[14021] = 'ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_Size'
ObjectIdNames[14022] = 'ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_Writable'
ObjectIdNames[14023] = 'ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_UserWritable'
ObjectIdNames[14024] = 'ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_OpenCount'
ObjectIdNames[14025] = 'ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_MimeType'
ObjectIdNames[14026] = 'ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_Open'
ObjectIdNames[14027] = 'ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_Open_InputArguments'
ObjectIdNames[14028] = 'ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_Open_OutputArguments'
ObjectIdNames[14029] = 'ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_Close'
ObjectIdNames[14030] = 'ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_Close_InputArguments'
ObjectIdNames[14031] = 'ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_Read'
ObjectIdNames[14032] = 'ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_Read_InputArguments'
ObjectIdNames[14033] = 'ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_Read_OutputArguments'
ObjectIdNames[14034] = 'ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_Write'
ObjectIdNames[14035] = 'ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_Write_InputArguments'
ObjectIdNames[14036] = 'ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_GetPosition'
ObjectIdNames[14037] = 'ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_GetPosition_InputArguments'
ObjectIdNames[14038] = 'ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_GetPosition_OutputArguments'
ObjectIdNames[14039] = 'ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_SetPosition'
ObjectIdNames[14040] = 'ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_SetPosition_InputArguments'
ObjectIdNames[14041] = 'ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_LastUpdateTime'
ObjectIdNames[14042] = 'ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_OpenWithMasks'
ObjectIdNames[14043] = 'ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_OpenWithMasks_InputArguments'
ObjectIdNames[14044] = 'ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_OpenWithMasks_OutputArguments'
ObjectIdNames[14045] = 'ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_CloseAndUpdate'
ObjectIdNames[14046] = 'ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_CloseAndUpdate_InputArguments'
ObjectIdNames[14047] = 'ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_CloseAndUpdate_OutputArguments'
ObjectIdNames[14048] = 'ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_AddCertificate'
ObjectIdNames[14049] = 'ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_AddCertificate_InputArguments'
ObjectIdNames[14050] = 'ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_RemoveCertificate'
ObjectIdNames[14051] = 'ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_TrustList_RemoveCertificate_InputArguments'
ObjectIdNames[14052] = 'ServerConfigurationType_CertificateGroups_DefaultUserTokenGroup_CertificateTypes'
ObjectIdNames[14053] = 'ServerConfiguration_CertificateGroups'
ObjectIdNames[14088] = 'ServerConfiguration_CertificateGroups_DefaultHttpsGroup'
ObjectIdNames[14089] = 'ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList'
ObjectIdNames[14090] = 'ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_Size'
ObjectIdNames[14091] = 'ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_Writable'
ObjectIdNames[14092] = 'ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_UserWritable'
ObjectIdNames[14093] = 'ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_OpenCount'
ObjectIdNames[14094] = 'ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_MimeType'
ObjectIdNames[14095] = 'ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_Open'
ObjectIdNames[14096] = 'ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_Open_InputArguments'
ObjectIdNames[14097] = 'ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_Open_OutputArguments'
ObjectIdNames[14098] = 'ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_Close'
ObjectIdNames[14099] = 'ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_Close_InputArguments'
ObjectIdNames[14100] = 'ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_Read'
ObjectIdNames[14101] = 'ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_Read_InputArguments'
ObjectIdNames[14102] = 'ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_Read_OutputArguments'
ObjectIdNames[14103] = 'ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_Write'
ObjectIdNames[14104] = 'ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_Write_InputArguments'
ObjectIdNames[14105] = 'ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_GetPosition'
ObjectIdNames[14106] = 'ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_GetPosition_InputArguments'
ObjectIdNames[14107] = 'ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_GetPosition_OutputArguments'
ObjectIdNames[14108] = 'ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_SetPosition'
ObjectIdNames[14109] = 'ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_SetPosition_InputArguments'
ObjectIdNames[14110] = 'ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_LastUpdateTime'
ObjectIdNames[14111] = 'ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_OpenWithMasks'
ObjectIdNames[14112] = 'ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_OpenWithMasks_InputArguments'
ObjectIdNames[14113] = 'ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_OpenWithMasks_OutputArguments'
ObjectIdNames[14114] = 'ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_CloseAndUpdate'
ObjectIdNames[14115] = 'ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_CloseAndUpdate_InputArguments'
ObjectIdNames[14116] = 'ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_CloseAndUpdate_OutputArguments'
ObjectIdNames[14117] = 'ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_AddCertificate'
ObjectIdNames[14118] = 'ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_AddCertificate_InputArguments'
ObjectIdNames[14119] = 'ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_RemoveCertificate'
ObjectIdNames[14120] = 'ServerConfiguration_CertificateGroups_DefaultHttpsGroup_TrustList_RemoveCertificate_InputArguments'
ObjectIdNames[14121] = 'ServerConfiguration_CertificateGroups_DefaultHttpsGroup_CertificateTypes'
ObjectIdNames[14122] = 'ServerConfiguration_CertificateGroups_DefaultUserTokenGroup'
ObjectIdNames[14123] = 'ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList'
ObjectIdNames[14124] = 'ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_Size'
ObjectIdNames[14125] = 'ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_Writable'
ObjectIdNames[14126] = 'ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_UserWritable'
ObjectIdNames[14127] = 'ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_OpenCount'
ObjectIdNames[14128] = 'ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_MimeType'
ObjectIdNames[14129] = 'ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_Open'
ObjectIdNames[14130] = 'ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_Open_InputArguments'
ObjectIdNames[14131] = 'ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_Open_OutputArguments'
ObjectIdNames[14132] = 'ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_Close'
ObjectIdNames[14133] = 'ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_Close_InputArguments'
ObjectIdNames[14134] = 'ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_Read'
ObjectIdNames[14135] = 'ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_Read_InputArguments'
ObjectIdNames[14136] = 'ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_Read_OutputArguments'
ObjectIdNames[14137] = 'ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_Write'
ObjectIdNames[14138] = 'ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_Write_InputArguments'
ObjectIdNames[14139] = 'ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_GetPosition'
ObjectIdNames[14140] = 'ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_GetPosition_InputArguments'
ObjectIdNames[14141] = 'ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_GetPosition_OutputArguments'
ObjectIdNames[14142] = 'ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_SetPosition'
ObjectIdNames[14143] = 'ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_SetPosition_InputArguments'
ObjectIdNames[14144] = 'ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_LastUpdateTime'
ObjectIdNames[14145] = 'ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_OpenWithMasks'
ObjectIdNames[14146] = 'ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_OpenWithMasks_InputArguments'
ObjectIdNames[14147] = 'ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_OpenWithMasks_OutputArguments'
ObjectIdNames[14148] = 'ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_CloseAndUpdate'
ObjectIdNames[14149] = 'ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_CloseAndUpdate_InputArguments'
ObjectIdNames[14150] = 'ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_CloseAndUpdate_OutputArguments'
ObjectIdNames[14151] = 'ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_AddCertificate'
ObjectIdNames[14152] = 'ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_AddCertificate_InputArguments'
ObjectIdNames[14153] = 'ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_RemoveCertificate'
ObjectIdNames[14154] = 'ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_TrustList_RemoveCertificate_InputArguments'
ObjectIdNames[14155] = 'ServerConfiguration_CertificateGroups_DefaultUserTokenGroup_CertificateTypes'
ObjectIdNames[14156] = 'ServerConfiguration_CertificateGroups_DefaultApplicationGroup'
ObjectIdNames[14157] = 'ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_Writable'
ObjectIdNames[14158] = 'ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_UserWritable'
ObjectIdNames[14159] = 'ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_MimeType'
ObjectIdNames[14160] = 'ServerConfiguration_CertificateGroups_DefaultApplicationGroup_TrustList_CloseAndUpdate_InputArguments'
ObjectIdNames[14161] = 'ServerConfiguration_CertificateGroups_DefaultApplicationGroup_CertificateTypes'
ObjectIdNames[14413] = 'AuditCreateSessionEventType_SessionId'
ObjectIdNames[14414] = 'AuditUrlMismatchEventType_SessionId'
ObjectIdNames[14415] = 'Server_ServerRedundancy_ServerNetworkGroups'
